Package: lintr
Title: A 'Linter' for R Code
Version: 3.4.0
Authors@R: c(
    person("Jim", "Hester", , role = "aut"),
    person("Florent", "Angly", role = "aut",
           comment = c(GitHub = "fangly")),
    person("Russ", "Hyde", role = "aut"),
    person("Michael", "Chirico", email = "michaelchirico4@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-0787-087X")),
    person("Kun", "Ren", role = "aut"),
    person("Alexander", "Rosenstock", role = "aut",
           comment = c(GitHub = "AshesITR")),
    person("Indrajeet", "Patil", email = "patilindrajeet.science@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0003-1995-6531")),
    person("Hugo", "Gruson", role = "aut",
           comment = c(ORCID = "0000-0002-4094-1476")),
    person("Emmanuel", "Ferdman", role = "aut")
  )
Description: Checks adherence to a given style, syntax errors and possible
    semantic issues.  Supports on the fly checking of R code edited with
    'RStudio IDE', 'Emacs', 'Vim', 'Sublime Text', 'Atom' and 'Visual
    Studio Code'.
License: MIT + file LICENSE
URL: https://lintr.r-lib.org, https://github.com/r-lib/lintr
BugReports: https://github.com/r-lib/lintr/issues
Depends: R (>= 4.1.0)
Imports: cli (>= 3.4.0), codetools, digest, glue, knitr, rex, stats,
        utils, xfun, xml2 (>= 1.3.4), xmlparsedata (>= 1.0.5)
Suggests: bookdown, cyclocomp, jsonlite, patrick (>= 0.2.0), rlang,
        rmarkdown, rstudioapi (>= 0.2), testthat (>= 3.3.0), tibble,
        tufte, withr (>= 2.5.0)
Enhances: data.table
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/Needs/development: pkgload, testthat, patrick
Config/testthat/edition: 3
Encoding: UTF-8
Collate: 'make_linter_from_xpath.R' 'xp_utils.R' 'utils.R' 'AAA.R'
        'T_and_F_symbol_linter.R' 'absolute_path_linter.R' 'actions.R'
        'addins.R' 'all_equal_linter.R' 'any_duplicated_linter.R'
        'any_is_na_linter.R' 'assignment_linter.R' 'backport_linter.R'
        'boolean_arithmetic_linter.R' 'brace_linter.R' 'cache.R'
        'class_equals_linter.R' 'coalesce_linter.R' 'commas_linter.R'
        'commented_code_linter.R' 'comparison_negation_linter.R'
        'condition_call_linter.R' 'condition_message_linter.R'
        'conjunct_test_linter.R' 'consecutive_assertion_linter.R'
        'consecutive_mutate_linter.R' 'cyclocomp_linter.R'
        'declared_functions.R' 'deprecated.R' 'download_file_linter.R'
        'duplicate_argument_linter.R' 'empty_assignment_linter.R'
        'equals_na_linter.R' 'exclude.R' 'expect_comparison_linter.R'
        'expect_identical_linter.R' 'expect_length_linter.R'
        'expect_lint.R' 'expect_named_linter.R' 'expect_not_linter.R'
        'expect_null_linter.R' 'expect_s3_class_linter.R'
        'expect_s4_class_linter.R' 'expect_true_false_linter.R'
        'expect_type_linter.R' 'extract.R' 'fixed_regex_linter.R'
        'for_loop_index_linter.R' 'function_argument_linter.R'
        'function_left_parentheses_linter.R' 'function_return_linter.R'
        'get_source_expressions.R' 'ids_with_token.R'
        'if_not_else_linter.R' 'if_switch_linter.R'
        'ifelse_censor_linter.R' 'implicit_assignment_linter.R'
        'implicit_integer_linter.R' 'indentation_linter.R'
        'infix_spaces_linter.R' 'inner_combine_linter.R'
        'is_lint_level.R' 'is_numeric_linter.R'
        'keyword_quote_linter.R' 'length_levels_linter.R'
        'length_test_linter.R' 'lengths_linter.R'
        'library_call_linter.R' 'line_length_linter.R' 'lint.R'
        'linter_tag_docs.R' 'linter_tags.R' 'lintr-package.R'
        'list2df_linter.R' 'list_comparison_linter.R'
        'literal_coercion_linter.R' 'make_linter_from_regex.R'
        'matrix_apply_linter.R' 'methods.R' 'missing_argument_linter.R'
        'missing_package_linter.R' 'namespace.R' 'namespace_linter.R'
        'nested_ifelse_linter.R' 'nested_pipe_linter.R'
        'nonportable_path_linter.R' 'shared_constants.R'
        'nrow_subset_linter.R' 'numeric_leading_zero_linter.R'
        'nzchar_linter.R' 'object_length_linter.R'
        'object_name_linter.R' 'object_overwrite_linter.R'
        'object_usage_linter.R' 'one_call_pipe_linter.R'
        'outer_negation_linter.R' 'package_hooks_linter.R'
        'paren_body_linter.R' 'paste_linter.R' 'path_utils.R'
        'pipe_call_linter.R' 'pipe_consistency_linter.R'
        'pipe_continuation_linter.R' 'pipe_return_linter.R'
        'print_linter.R' 'quotes_linter.R' 'redundant_equals_linter.R'
        'redundant_ifelse_linter.R' 'regex_subset_linter.R'
        'rep_len_linter.R' 'repeat_linter.R' 'return_linter.R'
        'routine_registration_linter.R' 'sample_int_linter.R'
        'scalar_in_linter.R' 'semicolon_linter.R' 'seq_linter.R'
        'settings.R' 'settings_utils.R' 'sort_linter.R'
        'source_utils.R' 'spaces_inside_linter.R'
        'spaces_left_parentheses_linter.R' 'sprintf_linter.R'
        'stopifnot_all_linter.R' 'string_boundary_linter.R'
        'strings_as_factors_linter.R' 'system_file_linter.R'
        'terminal_close_linter.R' 'todo_comment_linter.R'
        'trailing_blank_lines_linter.R' 'trailing_whitespace_linter.R'
        'tree_utils.R' 'undesirable_function_linter.R'
        'undesirable_operator_linter.R'
        'unnecessary_concatenation_linter.R'
        'unnecessary_lambda_linter.R' 'unnecessary_nesting_linter.R'
        'unnecessary_placeholder_linter.R' 'unreachable_code_linter.R'
        'unused_import_linter.R' 'use_lintr.R' 'vector_logic_linter.R'
        'which_grepl_linter.R' 'whitespace_linter.R' 'with.R'
        'with_id.R' 'xml_nodes_to_lints.R' 'xml_utils.R'
        'yoda_test_linter.R' 'zzz.R'
Language: en-US
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-15 23:20:21 UTC; chiricom
Author: Jim Hester [aut],
  Florent Angly [aut] (GitHub: fangly),
  Russ Hyde [aut],
  Michael Chirico [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-0787-087X>),
  Kun Ren [aut],
  Alexander Rosenstock [aut] (GitHub: AshesITR),
  Indrajeet Patil [aut] (ORCID: <https://orcid.org/0000-0003-1995-6531>),
  Hugo Gruson [aut] (ORCID: <https://orcid.org/0000-0002-4094-1476>),
  Emmanuel Ferdman [aut]
Maintainer: Michael Chirico <michaelchirico4@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-16 00:10:02 UTC

Package: cp1252
Type: Package
Title: What the Package Does (Title Case)
Version: 0.1.0
Author: Who wrote it
Maintainer: The package maintainer <yourself@somewhere.net>
Description: More about what it does (maybe more than one line)
    Use four spaces when indenting paragraphs within the Description.
License: What license is it under?
Encoding: UTF-8
LazyData: true
Roxygen: list(markdown = TRUE)

Package: package
Title: What the Package Does (One Line, Title Case)
Version: 0.0.0.9000
Authors@R: 
    person(given = "First",
           family = "Last",
           role = c("aut", "cre"),
           email = "first.last@example.com",
           comment = c(ORCID = "YOUR-ORCID-ID"))
Description: What the package does (one paragraph).
License: What license it uses
Encoding: UTF-8
Imports: lintr
LazyData: true
Roxygen: list(markdown = TRUE)

Package: assignmentLinter
Type: Package
Title: What the package does (short line)
Version: 1.0
Date: 2019-12-17
Author: Who wrote it
Maintainer: Who to complain to <yourfault@somewhere.net>
Description: More about what it does (maybe more than one line)
License: What license is it under?
Suggests: 
    testthat (>= 3.0.0)
Config/testthat/edition: 3

# Generated by roxygen2: do not edit by hand

S3method("[",lints)
S3method(as.data.frame,lints)
S3method(data.table::as.data.table,lints)
S3method(format,lint)
S3method(format,lints)
S3method(names,lints)
S3method(print,lint)
S3method(print,lints)
S3method(split,lints)
S3method(summary,lints)
S3method(tibble::as_tibble,lints)
export(Lint)
export(Linter)
export(T_and_F_symbol_linter)
export(absolute_path_linter)
export(all_equal_linter)
export(all_linters)
export(all_undesirable_functions)
export(all_undesirable_operators)
export(any_duplicated_linter)
export(any_is_na_linter)
export(assignment_linter)
export(available_linters)
export(available_tags)
export(backport_linter)
export(boolean_arithmetic_linter)
export(brace_linter)
export(checkstyle_output)
export(class_equals_linter)
export(clear_cache)
export(coalesce_linter)
export(commas_linter)
export(commented_code_linter)
export(comparison_negation_linter)
export(condition_call_linter)
export(condition_message_linter)
export(conjunct_test_linter)
export(consecutive_assertion_linter)
export(consecutive_mutate_linter)
export(cyclocomp_linter)
export(default_linters)
export(default_settings)
export(default_undesirable_functions)
export(default_undesirable_operators)
export(download_file_linter)
export(duplicate_argument_linter)
export(empty_assignment_linter)
export(equals_na_linter)
export(expect_comparison_linter)
export(expect_identical_linter)
export(expect_length_linter)
export(expect_lint)
export(expect_lint_free)
export(expect_named_linter)
export(expect_no_lint)
export(expect_not_linter)
export(expect_null_linter)
export(expect_s3_class_linter)
export(expect_s4_class_linter)
export(expect_true_false_linter)
export(expect_type_linter)
export(fixed_regex_linter)
export(for_loop_index_linter)
export(function_argument_linter)
export(function_left_parentheses_linter)
export(function_return_linter)
export(get_r_string)
export(get_source_expressions)
export(gitlab_output)
export(ids_with_token)
export(if_not_else_linter)
export(if_switch_linter)
export(ifelse_censor_linter)
export(implicit_assignment_linter)
export(implicit_integer_linter)
export(indentation_linter)
export(infix_spaces_linter)
export(inner_combine_linter)
export(is_lint_level)
export(is_numeric_linter)
export(keyword_quote_linter)
export(length_levels_linter)
export(length_test_linter)
export(lengths_linter)
export(library_call_linter)
export(line_length_linter)
export(lint)
export(lint_dir)
export(lint_package)
export(linters_with_defaults)
export(linters_with_tags)
export(list2df_linter)
export(list_comparison_linter)
export(literal_coercion_linter)
export(make_linter_from_function_xpath)
export(make_linter_from_xpath)
export(matrix_apply_linter)
export(missing_argument_linter)
export(missing_package_linter)
export(modify_defaults)
export(namespace_linter)
export(nested_ifelse_linter)
export(nested_pipe_linter)
export(nonportable_path_linter)
export(nrow_subset_linter)
export(numeric_leading_zero_linter)
export(nzchar_linter)
export(object_length_linter)
export(object_name_linter)
export(object_overwrite_linter)
export(object_usage_linter)
export(one_call_pipe_linter)
export(outer_negation_linter)
export(package_hooks_linter)
export(paren_body_linter)
export(paste_linter)
export(pipe_call_linter)
export(pipe_consistency_linter)
export(pipe_continuation_linter)
export(pipe_return_linter)
export(print_linter)
export(quotes_linter)
export(redundant_equals_linter)
export(redundant_ifelse_linter)
export(regex_subset_linter)
export(rep_len_linter)
export(repeat_linter)
export(return_linter)
export(routine_registration_linter)
export(sample_int_linter)
export(sarif_output)
export(scalar_in_linter)
export(semicolon_linter)
export(seq_linter)
export(sort_linter)
export(spaces_inside_linter)
export(spaces_left_parentheses_linter)
export(sprintf_linter)
export(stopifnot_all_linter)
export(string_boundary_linter)
export(strings_as_factors_linter)
export(system_file_linter)
export(terminal_close_linter)
export(todo_comment_linter)
export(trailing_blank_lines_linter)
export(trailing_whitespace_linter)
export(undesirable_function_linter)
export(undesirable_operator_linter)
export(unnecessary_concatenation_linter)
export(unnecessary_lambda_linter)
export(unnecessary_nesting_linter)
export(unnecessary_placeholder_linter)
export(unreachable_code_linter)
export(unused_import_linter)
export(use_lintr)
export(vector_logic_linter)
export(which_grepl_linter)
export(whitespace_linter)
export(with_id)
export(xml_nodes_to_lints)
export(xp_call_name)
export(yoda_test_linter)
importFrom(cli,cli_abort)
importFrom(cli,cli_inform)
importFrom(cli,cli_warn)
importFrom(cli,qty)
importFrom(glue,glue)
importFrom(glue,glue_collapse)
importFrom(rex,character_class)
importFrom(rex,re_matches)
importFrom(rex,re_substitutes)
importFrom(rex,regex)
importFrom(rex,rex)
importFrom(stats,complete.cases)
importFrom(stats,na.omit)
importFrom(tools,R_user_dir)
importFrom(utils,capture.output)
importFrom(utils,getParseData)
importFrom(utils,globalVariables)
importFrom(utils,head)
importFrom(utils,relist)
importFrom(utils,tail)
importFrom(xml2,xml_attr)
importFrom(xml2,xml_attrs)
importFrom(xml2,xml_child)
importFrom(xml2,xml_children)
importFrom(xml2,xml_find_all)
importFrom(xml2,xml_find_chr)
importFrom(xml2,xml_find_first)
importFrom(xml2,xml_find_lgl)
importFrom(xml2,xml_find_num)
importFrom(xml2,xml_missing)
importFrom(xml2,xml_name)
importFrom(xml2,xml_parent)
importFrom(xml2,xml_text)

# Generated by roxygen2: do not edit by hand

S3method("names<-",my_custom_class)
S3method(drink_me,data.frame)
S3method(drink_me,default)
S3method(drink_me,list)
S3method(eat_me,liiiiiiiiiiiiiiiiiiiiiiiiiiist)
S3method(head,my_s3_object)
export(drink_me)
export(eat_me)
importFrom(utils,head)

---
output:
  rmarkdown::github_document:
    pandoc_args: ["--wrap=none"]
---

# lintr <img src="man/figures/logo.png" align="right" width="240" />

[![R build status](https://github.com/r-lib/lintr/workflows/R-CMD-check/badge.svg)](https://github.com/r-lib/lintr/actions)
[![codecov.io](https://codecov.io/gh/r-lib/lintr/branch/main/graphs/badge.svg)](https://app.codecov.io/gh/r-lib/lintr?branch=main)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/lintr)](https://cran.r-project.org/package=lintr)
[![lifecycle](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![DOI](https://joss.theoj.org/papers/10.21105/joss.07240/status.svg)](https://doi.org/10.21105/joss.07240)

`{lintr}` provides [static code analysis for R](https://en.wikipedia.org/wiki/Static_program_analysis). It checks for adherence to a given style, identifying syntax errors and possible semantic issues, then reports them to you so you can take action.

For example, given a file `bad.R` with the following contents:


``` r
my_func = function(x){
    if(x > 0){
        print("Positive")
    }
    else {
        print("Not positive")
    }
    return(T)
}
```

Running `lintr::lint()` on this file would produce the following output:
```R
lintr::lint("bad.R")
```


```
bad.R:1:9: style: [assignment_linter] Use one of <-, <<- for assignment, not =.
my_func = function(x){
        ^
bad.R:1:22: style: [brace_linter] There should be a space before an opening curly brace.
my_func = function(x){
                     ^
bad.R:1:22: style: [paren_body_linter] Put a space between a right parenthesis and a body expression.
my_func = function(x){
                     ^
bad.R:2:4: style: [indentation_linter] Indentation should be 2 spaces but is 4 spaces.
    if(x > 0){
  ~^
bad.R:2:7: style: [spaces_left_parentheses_linter] Place a space before left parenthesis, except in a function call.
    if(x > 0){
      ^
bad.R:2:14: style: [brace_linter] There should be a space before an opening curly brace.
    if(x > 0){
             ^
bad.R:2:14: style: [paren_body_linter] Put a space between a right parenthesis and a body expression.
    if(x > 0){
             ^
bad.R:3:8: style: [indentation_linter] Indentation should be 4 spaces but is 8 spaces.
        print("Positive")
    ~~~^
bad.R:4:4: style: [indentation_linter] Indentation should be 2 spaces but is 4 spaces.
    }
  ~^
bad.R:5:5: style: [brace_linter] `else` should come on the same line as the previous `}`.
    else {
    ^~~~
bad.R:6:8: style: [indentation_linter] Indentation should be 4 spaces but is 8 spaces.
        print("Not positive")
    ~~~^
bad.R:7:4: style: [indentation_linter] Indentation should be 2 spaces but is 4 spaces.
    }
  ~^
bad.R:8:5: style: [return_linter] Use implicit return behavior; explicit return() is not needed.
    return(T)
    ^~~~~~
bad.R:8:13: style: [T_and_F_symbol_linter] Use TRUE instead of the symbol T.
    return(T)
           ~^
bad.R:10:1: style: [trailing_blank_lines_linter] Remove trailing blank lines.

^
```

`{lintr}` is complementary to [the `{styler}` package](https://github.com/r-lib/styler) which automatically restyles code, eliminating some of the problems that `{lintr}` can detect.

## Installation

Install the stable version from CRAN:

```R
install.packages("lintr")
```

Or the development version from GitHub:

```R
# install.packages("remotes")
remotes::install_github("r-lib/lintr")
```

## Usage

And then you can create a configuration file and run selected linters:

```R
lintr::use_lintr(type = "tidyverse")

# in a project:
lintr::lint_dir()

# in a package:
lintr::lint_package()
```

To see a list of linters included for each configuration:

```R
# tidyverse (default)
names(lintr::linters_with_defaults())

# full
names(lintr::all_linters())
```

### Setting up GitHub Actions

`{usethis}` provides helper functions to generate lint workflows for GitHub Actions:

```R
# in a project:
usethis::use_github_action("lint-project")

# in a package:
usethis::use_github_action("lint")
```

You can also run lintr during continuous integration or within your IDE or text editor. See `vignette("continuous-integration")` and `vignette("editors")` for more details.

Without further configuration, this will run the [default linters](https://lintr.r-lib.org/reference/default_linters.html). See `vignette("lintr")` to learn how to modify these defaults.

## Code of Conduct

Please note that the lintr project is released with a [Contributor Code of Conduct](https://www.contributor-covenant.org/version/2/1/code_of_conduct/). By contributing to this project, you agree to abide by its terms.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----show_default_settings, echo = FALSE--------------------------------------
default_settings <- lintr::default_settings
default_settings$linters <- "`lintr::default_linters`"
default_settings$exclusions <- "(empty)"

make_string <- function(x) {
  if (inherits(x, "regex")) {
    paste0("regex: `", x, "`")
  } else {
    as.character(x)
  }
}

defaults_table <- data.frame(default = vapply(default_settings, make_string, character(1L)))

# avoid conflict when loading lintr in echo=TRUE cell below
rm(default_settings)

knitr::kable(defaults_table)

## ----show_linter_defaults, echo = FALSE---------------------------------------
library(lintr) # needed here for formalArgs

default_linters <- lintr::default_linters
linters_with_args <- lapply(
  setNames(nm = intersect(names(default_linters), lintr::available_linters(tags = "configurable")$linter)),
  formalArgs
)

make_setting_string <- function(linter_name) {
  linter_args <- linters_with_args[[linter_name]]
  if (is.null(linter_args)) {
    return("")
  }

  arglist <- vapply(linter_args, function(arg) {
    env <- environment(default_linters[[linter_name]])
    deparse(env[[arg]])
  }, character(1L))

  paste0(linter_args, " = ", arglist, collapse = ", ")
}

defaults_table <- data.frame(
  row.names = names(default_linters),
  settings = vapply(names(default_linters), make_setting_string, character(1L))
)

knitr::kable(defaults_table)

## ----show_tags----------------------------------------------------------------
lintr::available_tags(packages = "lintr")

## ----show_tag_linters---------------------------------------------------------
linters <- lintr::linters_with_tags(tags = c("package_development", "readability"))
names(linters)

## ----show_all_linter_names----------------------------------------------------
names(lintr::all_linters())

## ----programmatic_lintr-------------------------------------------------------
library(lintr)
linter_name <- "assignment_linter"

show_lint <- function(l) {
  lint_df <- as.data.frame(l)
  print(lint_df[, c("line_number", "message", "linter")])
}
hline <- function() cat(strrep("-", getOption("width") - 5L), "\n", sep = "")

withr::with_tempfile("tmp", {
  writeLines("a = 1", tmp)

  # linter column is just 'get'
  show_lint(lint(tmp, linters = get(linter_name)()))
  hline()

  this_linter <- get(linter_name)()
  attr(this_linter, "name") <- linter_name
  # linter column is 'assignment_linter'
  show_lint(lint(tmp, linters = this_linter))
  hline()

  # more concise alternative for this case: use eval(call(.))
  show_lint(lint(tmp, linters = eval(call(linter_name))))
})

## ----show_long_line_lint, echo = FALSE----------------------------------------
lint("X = 42L # -------------- this comment overflows the default 80 chars line length.\n",
  parse_settings = FALSE
)

## ----show_nolint, echo = FALSE------------------------------------------------
lint("X = 42L # nolint ------ this comment overflows the default 80 chars line length.\n",
  parse_settings = FALSE
)

## ----show_long_line_lint_not_skipped, echo = FALSE----------------------------
lint("X = 42L # nolint: object_name_linter. this comment overflows the default 80 chars line length.\n",
  parse_settings = FALSE
)

## ----show_nolint_multiple, echo = FALSE---------------------------------------
lint(
  paste(
    "X = 42L",
    "# nolint: object_name_linter, line_length_linter. this comment overflows the default 80 chars line length.\n"
  ),
  parse_settings = FALSE
)

## ----show_nolint_abbrev, echo = FALSE-----------------------------------------
lint(
  paste(
    "X = 42L",
    "# nolint: object_name, line_len. this comment still overflows the default 80 chars line length.\n"
  ),
  parse_settings = FALSE
)

## ----show_comment_code_lint, echo = FALSE-------------------------------------
lint("# x <- 42L\n# print(x)\n", parse_settings = FALSE)

## ----show_comment_code_nolint, echo = FALSE-----------------------------------
lint("# nolint start: commented_code_linter.\n# x <- 42L\n# print(x)\n# nolint end\n",
  parse_settings = FALSE
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

in_pkgdown <- identical(Sys.getenv("IN_PKGDOWN"), "true")

maybe_still <- function(url) {
  if (in_pkgdown) {
    url
  } else {
    gsub("\\.gif$", "-still.gif", url)
  }
}

## ----echo = FALSE, results = 'asis'-------------------------------------------
if (!in_pkgdown) {
  cat(
    "Note: This vignette is best viewed [online](https://lintr.r-lib.org/articles/editors.html),",
    "where we can render full animations of editor flows.\n"
  )
}

control_char_regex <- rex(one_of(intToUtf8(seq.int(0L, 31L), multiple = TRUE)))
# control chars (non-printing)

win32_bad_punct_regex <- rex(one_of(
  "*", "?", "\"", "<", ">", "|", ":", "/", "\\"
))
win32_good_punct_regex <- rex(one_of(
  "!", "#", "$", "%", "&", "'", "(", ")", "+", ",", "-", ".", ";",
  "=", "@", "[", "]", "^", "_", "`", "{", "}", "~"
))
# win32BadPunct + win32AllowedPunct = [:punct:]

unsafe_char_regex <- rex(or(control_char_regex, win32_bad_punct_regex))

safe_char_regex <- rex(or(alnum, " ", win32_good_punct_regex))
# alphanumeric chars (including UTF-8), spaces, and win32-allowed punctuation

portable_char_regex <- rex(character_class("A-Za-z0-9_\\-."))
# ASCII letters, digits, dot, hyphen and underscore

protocol_regex <- rex(one_or_more(letter), "://")

root_regex <- list(
  unix  = rex(start, "/"),
  tilde = rex(start, "~", zero_or_more(portable_char_regex), zero_or_more("/")),
  win32 = rex(start, letter, ":", maybe(one_of("/", "\\"))),
  unc   = rex(start, "\\\\", maybe(one_or_more(portable_char_regex), maybe("\\")))
)

root_path_regex <- rex(start, or(root_regex), end)

# styler: off
absolute_path_regex <- rex(or(
  root_regex[["unix"]]  %if_next_isnt% one_of(space, "/"),
  root_regex[["tilde"]] %if_next_isnt% one_of(space, quote),
  root_regex[["win32"]] %if_next_isnt% one_of(space, "/", "\\"),
  root_regex[["unc"]]   %if_next_isnt% one_of(space, "\\", quote)
))

relative_path_regex <- rex(
  start %if_next_isnt% or(quote, absolute_path_regex, protocol_regex),
  or(                                                    # not an absolute path or protocol, and then
    group(one_or_more(safe_char_regex), or("/", "\\")),  # (back)slash-separated paths
    group(dot, maybe(dot), end)                          # or single or double dot
  )                                                      # (paths without slash or dots left out)
  %if_next_isnt% quote
)
# styler: on

path_regex <- rex(or(absolute_path_regex, relative_path_regex))

is_absolute_path <- function(path) {
  re_matches(path, absolute_path_regex)
}

is_root_path <- function(path) {
  re_matches(path, root_path_regex)
}

is_path <- function(path) {
  re_matches(path, path_regex)
}

is_valid_path <- function(paths, lax = FALSE) {
  # Given a character vector of paths, return FALSE for directory or file having valid characters.
  # On Windows, invalid chars are all control chars and: * ? " < > | :
  # On Unix, all characters are valid, except when lax=TRUE (use same invalid chars as Windows).
  as.logical(
    Map(
      function(dirs, is_win32) {
        if (!is_win32 && !lax) {
          return(TRUE)
        }
        if (length(dirs) > 0L && is_root_path(dirs[[1L]])) {
          dirs <- tail(dirs, -1L) # remove root element ("/", "C:", or "\\")
        }
        !any(re_matches(dirs, unsafe_char_regex))
      },
      split_paths(paths),
      re_matches(paths, rex(or(list(root_regex[["win32"]], root_regex[["unc"]], "\\"))))
    )
  )
}

is_long_path <- function(path) {
  # Take a character vector of paths and determine if they are "long enough"
  # TRUE , e.g.: "./foo", "C:\\foo", "foo/bar"
  # FALSE, e.g.: "/",  "\\", "n/a", "/foo", "foo/"
  re_matches(
    re_substitutes(path, ":", ""),
    rex(at_least(safe_char_regex, 1L), one_of("/", "\\"), at_least(safe_char_regex, 2L))
  )
}

is_valid_long_path <- function(path, lax = FALSE) {
  # Convenience function to avoid linting short paths and those unlikely to be valid paths
  ret <- is_valid_path(path, lax)
  if (lax) {
    ret <- ret & is_long_path(path)
  }
  ret
}


split_paths <- function(path)  Map(split_path, strsplit(path, "/|\\\\"), substr(path, 1L, 1L))

split_path <- function(dirs, prefix) {
  # add root dir if needed
  nonempty_dirs <- nzchar(dirs)
  i <- match(TRUE, nonempty_dirs, nomatch = length(dirs) + 1L) - 1L
  if (i > 0L) {
    dirs <- c(strrep(prefix, i), tail(dirs, -i))
  }

  # add // to protocols (like http, smb, ...)
  if (length(dirs) > 0L && grepl("..:$", dirs[[1L]])) {
    dirs[[1L]] <- paste0(dirs[[1L]], "//")
  }

  # remove empty dirs
  dirs[nzchar(dirs)]
}

#' Simple wrapper around normalizePath to ensure forward slash on Windows
#' https://github.com/r-lib/lintr/pull/2613
#' @noRd
# nolint next: undesirable_function_name_linter, object_name_linter.
normalize_path <- function(path, mustWork = NA) normalizePath(path = path, winslash = "/", mustWork = mustWork)

#' @include utils.R
path_linter_factory <- function(path_function, message, linter, name = linter_auto_name()) {
  force(name)
  Linter(name = name, linter_level = "expression", function(source_expression) {
    lapply(
      ids_with_token(source_expression, "STR_CONST"),
      function(id) {
        token <- with_id(source_expression, id)
        path <- get_r_string(token$text)
        if (path_function(path)) {
          path_start <- token[["col1"]] + 1L
          path_end <- token[["col2"]] - 1L
          Lint(
            filename = source_expression[["filename"]],
            line_number = token[["line1"]],
            column_number = path_start,
            type = "warning",
            message = message,
            line = source_expression[["lines"]][[as.character(token[["line1"]])]],
            ranges = list(c(path_start, path_end))
          )
        }
      }
    )
  })
}

test_that("use_lintr works as expected", {
  tmp <- withr::local_tempdir()

  lintr_file <- use_lintr(path = tmp)
  expect_true(file.exists(lintr_file))

  # check that newly created file is in the root directory
  expect_identical(
    normalize_path(lintr_file),
    file.path(normalize_path(tmp), ".lintr")
  )

  # can't generate if a .lintr already exists
  expect_error(use_lintr(path = tmp), "Found an existing configuration")

  # check that `read_settings()` works with the generated file
  # this can be checked by checking lintr runs successfully
  lints <- lint_dir(tmp)
  expect_length(lints, 0L)
})

test_that("use_lintr with type = full also works", {
  tmp <- withr::local_tempdir()

  # type = "full" also works with read_settings()
  lintr_file <- use_lintr(path = tmp, type = "full")
  expect_true(file.exists(lintr_file))

  # check that newly created file is in the root directory
  expect_identical(
    normalize_path(lintr_file),
    file.path(normalize_path(tmp), ".lintr")
  )

  skip_if_not_installed("cyclocomp") # avoid warning
  lints <- lint_dir(tmp)
  expect_length(lints, 0L)
})

test_that("No .Rbuildignore is created of packages", {
  tmp <- withr::local_tempdir()

  lintr_file <- use_lintr(path = tmp, type = "full")
  expect_false(file.exists(file.path(tmp, ".Rbuildignore")))
})

test_that("No .Rbuildignore is filled outside of packages", {
  tmp <- withr::local_tempdir()
  ignore_path <- file.path(tmp, ".Rbuildignore")
  file.create(ignore_path)

  lintr_file <- use_lintr(path = tmp, type = "full")
  expect_identical(readLines(ignore_path), character())
})

test_that("No .Rbuildignore is filled if pattern already present", {
  tmp <- withr::local_tempdir()
  writeLines(
    "Package: test",
    file.path(tmp, "DESCRIPTION")
  )
  ignore_path <- file.path(tmp, ".Rbuildignore")
  ignore <- c("^fu$", "^\\.lintr$", "^bar$")
  writeLines(ignore, ignore_path)

  expect_message(
    {
      lintr_file <- use_lintr(path = tmp, type = "full")
    },
    "^\\.lintr$ is already ignored in `.Rbuildignore`.",
    fixed = TRUE
  )
  expect_identical(readLines(ignore_path), ignore)
})

test_that("use_lintr creates the correct regex", {
  tmp <- withr::local_tempdir()
  writeLines(
    "Package: test",
    file.path(tmp, "DESCRIPTION")
  )
  ignore_path <- file.path(tmp, ".Rbuildignore")
  writeLines(
    c("^fu$", "^bar$"),
    ignore_path
  )

  expect_message(
    {
      lintr_file <- use_lintr(path = tmp, type = "full")
    },
    regexp = rex::rex("Added ^\\.lintr$ to `.Rbuildignore`")
  )
  expect_identical(readLines(ignore_path), c("^fu$", "^bar$", "^\\.lintr$"))
})

#' Require usage of `lengths()` where possible
#'
#' [base::lengths()] is a function that was added to base R in version 3.2.0 to
#'   get the length of each element of a list. It is equivalent to
#'   `sapply(x, length)`, but faster and more readable.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "sapply(x, length)",
#'   linters = lengths_linter()
#' )
#'
#' lint(
#'   text = "vapply(x, length, integer(1L))",
#'   linters = lengths_linter()
#' )
#'
#' lint(
#'   text = "purrr::map_int(x, length)",
#'   linters = lengths_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "lengths(x)",
#'   linters = lengths_linter()
#' )
#'
#' @evalRd rd_tags("lengths_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
lengths_linter <- make_linter_from_function_xpath(
  function_names = c("sapply", "vapply", "map_int", "map_dbl"),
  xpath = "parent::expr[expr/SYMBOL[text() = 'length']]",
  lint_message = "Use lengths() to find the length of each element in a list."
)

in_github_actions <- function() {
  identical(Sys.getenv("GITHUB_ACTIONS"), "true")
}

in_pkgdown <- function() {
  identical(Sys.getenv("IN_PKGDOWN"), "true")
}

# Output logging commands for any lints found
github_actions_log_lints <- function(lints, project_dir = "") {
  for (x in lints) {
    if (nzchar(project_dir)) {
      x$filename <- file.path(project_dir, x$filename)
    }
    file_line_col <- sprintf(
      "file=%s,line=%s,col=%s", x$filename, x$line_number, x$column_number
    )
    cat(sprintf(
      "::warning %s::%s,[%s] %s\n",
      file_line_col, file_line_col, x$linter, x$message
    ), sep = "")
  }
}

# Regression test for #737 via expecting clean to be lint-free

#' drink_me
#' @description empty
#'
#' @export
drink_me <- function(x, ...) {
  UseMethod("drink_me")
}

#' drink_me for most things
#' @export
drink_me.default <- function(x, ...) {
  1
}

#' drink_me for lists
#' @export
drink_me.list <- function(x, ...) {
  NULL
}

#' drink_me for data.frames
#' @export
drink_me.data.frame <- function(x, ...) {
  NULL
}

#' head on my_s3_object
#' @importFrom utils head
#' @export
head.my_s3_object <- function(x, ...) {
  NULL
}

#' assign names for my_custom_class
#' @export
`names<-.my_custom_class` <- function(x, value) {
  NULL
}

#' Defined S3 generic in R/eat_me.R
#' Tests #1808
#' @export
eat_me.liiiiiiiiiiiiiiiiiiiiiiiiiiist <- function(x, ...) {
  NULL
}

#' Trailing blank lines linter
#'
#' Check that there are no trailing blank lines in source code.
#'
#' @examples
#' # will produce lints
#' f <- tempfile()
#' cat("x <- 1\n\n", file = f)
#' writeLines(readChar(f, file.size(f)))
#' lint(
#'   filename = f,
#'   linters = trailing_blank_lines_linter()
#' )
#' unlink(f)
#'
#' # okay
#' cat("x <- 1\n", file = f)
#' writeLines(readChar(f, file.size(f)))
#' lint(
#'   filename = f,
#'   linters = trailing_blank_lines_linter()
#' )
#' unlink(f)
#'
#' @evalRd rd_tags("trailing_blank_lines_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
trailing_blank_lines_linter <- function() {
  Linter(function(source_expression) {
    blanks <- re_matches(
      source_expression$file_lines,
      rex(start, any_spaces, end)
    )

    line_number <- length(source_expression$file_lines)
    lints <- list()
    while (line_number > 0L && (is.na(blanks[[line_number]]) || isTRUE(blanks[[line_number]]))) {
      if (!is.na(blanks[[line_number]])) {
        lints[[length(lints) + 1L]] <- Lint(
          filename = source_expression$filename,
          line_number = line_number,
          column_number = 1L,
          type = "style",
          message = "Remove trailing blank lines.",
          line = source_expression$file_lines[[line_number]]
        )
      }
      line_number <- line_number - 1L
    }

    if (identical(source_expression$terminal_newline, FALSE)) { # could use isFALSE, but needs backports
      last_line <- tail(source_expression$file_lines, 1L)
      line_width <- nchar(last_line)
      if (is.na(line_width)) line_width <- 0L

      lints[[length(lints) + 1L]] <- Lint(
        filename = source_expression$filename,
        line_number = length(source_expression$file_lines),
        column_number = line_width + 1L,
        type = "style",
        message = "Add a terminal newline.",
        line = last_line
      )
    }

    lints
  })
}

#' Block usage of return() in magrittr pipelines
#'
#' [return()] inside a magrittr pipeline does not actually execute `return()`
#'   like you'd expect:
#'   ```r
#'   bad_usage <- function(x) {
#'     x %>%
#'       return()
#'     FALSE
#'   }
#'   ```
#'
#'   `bad_usage(TRUE)` will return `FALSE`! It will technically work "as expected"
#'   if this is the final statement in the function body, but such usage is misleading.
#'   Instead, assign the pipe outcome to a variable and return that.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "function(x) x %>% return()",
#'   linters = pipe_return_linter()
#' )
#'
#' # okay
#' code <- "function(x) {\n  y <- sum(x)\n  return(y)\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = pipe_return_linter()
#' )
#'
#' @evalRd rd_tags("pipe_return_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
pipe_return_linter <- make_linter_from_xpath(
  # NB: Native pipe disallows this at the parser level, so there's no need
  #   to lint in valid R code.
  xpath = "
  //SPECIAL[text() = '%>%']
    /following-sibling::expr[expr/SYMBOL_FUNCTION_CALL[text() = 'return']]
  ",
  lint_message = paste(
    "Avoid return() as the final step of a magrittr pipeline. ",
    "Instead, assign the output of the pipeline to a well-named object and return that."
  )
)

#' Block statements like if (!A) x else y
#'
#' `if (!A) x else y` is the same as `if (A) y else x`, but the latter is
#'   easier to reason about in the `else` case. The former requires
#'   double negation that can be avoided by switching the statement order.
#'
#' This only applies in the simple `if/else` case. Statements like
#'   `if (!A) x else if (B) y else z` don't always have a simpler or
#'   more readable form.
#'
#' It also applies to [ifelse()] and the package equivalents
#'   `dplyr::if_else()` and `data.table::fifelse()`.
#'
#' @param exceptions Character vector of calls to exclude from linting.
#'   By default, [is.null()], [is.na()], and [missing()] are excluded
#'   given the common idiom `!is.na(x)` as "x is present".
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "if (!A) x else y",
#'   linters = if_not_else_linter()
#' )
#'
#' lint(
#'   text = "if (!A) x else if (!B) y else z",
#'   linters = if_not_else_linter()
#' )
#'
#' lint(
#'   text = "ifelse(!is_treatment, x, y)",
#'   linters = if_not_else_linter()
#' )
#'
#' lint(
#'   text = "if (!is.null(x)) x else 2",
#'   linters = if_not_else_linter(exceptions = character())
#' )
#'
#' # okay
#' lint(
#'   text = "if (A) x else y",
#'   linters = if_not_else_linter()
#' )
#'
#' lint(
#'   text = "if (!A) x else if (B) z else y",
#'   linters = if_not_else_linter()
#' )
#'
#' lint(
#'   text = "ifelse(is_treatment, y, x)",
#'   linters = if_not_else_linter()
#' )
#'
#' lint(
#'   text = "if (!is.null(x)) x else 2",
#'   linters = if_not_else_linter()
#' )
#'
#' @evalRd rd_tags("if_not_else_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
if_not_else_linter <- function(exceptions = c("is.null", "is.na", "missing")) {
  if_xpath <- glue("
  //IF[following-sibling::ELSE[not(following-sibling::expr[IF])]]
    /following-sibling::expr[1][
      OP-EXCLAMATION
      and not(expr[expr[SYMBOL_FUNCTION_CALL[{ xp_text_in_table(exceptions) }]]])
    ]
  ")

  ifelse_xpath <- glue("
  parent::expr[expr[
    position() = 2
    and OP-EXCLAMATION
    and not(expr[
      OP-EXCLAMATION
      or expr/SYMBOL_FUNCTION_CALL[{ xp_text_in_table(exceptions) }]
    ])
  ]]")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content
    ifelse_calls <- source_expression$xml_find_function_calls(ifelse_funs)

    if_expr <- xml_find_all_(xml, if_xpath)
    if_lints <- xml_nodes_to_lints(
      if_expr,
      source_expression = source_expression,
      lint_message = "Prefer `if (A) x else y` to the less-readable `if (!A) y else x` in a simple if/else statement.",
      type = "warning"
    )

    ifelse_expr <- xml_find_all_(ifelse_calls, ifelse_xpath)
    ifelse_call <- xp_call_name(ifelse_expr)
    ifelse_lints <- xml_nodes_to_lints(
      ifelse_expr,
      source_expression = source_expression,
      lint_message = sprintf(
        "Prefer `%1$s(A, x, y)` to the less-readable `%1$s(!A, y, x)`.",
        ifelse_call
      ),
      type = "warning"
    )

    c(if_lints, ifelse_lints)
  })
}

#' Block single-call magrittr pipes
#'
#' Prefer using a plain call instead of a pipe with only one call,
#'   i.e. `1:10 %>% sum()` should instead be `sum(1:10)`. Note that
#'   calls in the first `%>%` argument count. `rowSums(x) %>% max()` is OK
#'   because there are two total calls (`rowSums()` and `max()`).
#'
#' Note also that un-"called" steps are *not* counted, since they should
#'   be calls (see [pipe_call_linter()]).
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "(1:10) %>% sum()",
#'   linters = one_call_pipe_linter()
#' )
#'
#' lint(
#'   text = "DT %>% .[grp == 'a', sum(v)]",
#'   linters = one_call_pipe_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "rowSums(x) %>% mean()",
#'   linters = one_call_pipe_linter()
#' )
#'
#' lint(
#'   text = "DT[src == 'a', .N, by = grp] %>% .[N > 10]",
#'   linters = one_call_pipe_linter()
#' )
#'
#' # assignment pipe is exempted
#' lint(
#'   text = "DF %<>% mutate(a = 2)",
#'   linters = one_call_pipe_linter()
#' )
#'
#' @evalRd rd_tags("one_call_pipe_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/pipes.html#short-pipes>
#' @export
one_call_pipe_linter <- function() {
  # exception for assignment pipe per #2330
  pipes_cond <- xp_text_in_table(setdiff(magrittr_pipes, "%<>%"))

  # preceding-sibling::SPECIAL: if there are ever two pipes, don't lint
  # OP-LEFT-BRACKET/LBB: accept DT[...] %>% .[...] as a two-call pipe,
  #   (but not DT %>% .[...])
  # parent::expr/SPECIAL: make sure we are at the top of a pipeline
  # count(): any call anywhere else in the AST within the pipe expression
  xpath <- glue("
  (//SPECIAL[{pipes_cond}] | //PIPE)[
    not(preceding-sibling::expr[1]/*[self::SPECIAL[{pipes_cond}] or self::PIPE])
    and (
      not(following-sibling::expr[OP-LEFT-BRACKET or LBB])
      or not(preceding-sibling::expr[OP-LEFT-BRACKET or LBB])
    )
  ]
    /parent::expr[
      not(parent::expr/*[self::SPECIAL[{ pipes_cond }] or self::PIPE])
      and count(.//SYMBOL_FUNCTION_CALL) <= 1
    ]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)
    pipe <- xml_find_chr_(bad_expr, "string(SPECIAL | PIPE)")

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste0("Avoid pipe ", pipe, " for expressions with only a single call."),
      type = "warning"
    )
  })
}

#' Raise lints for several common poor usages of `paste()`
#'
#' @description
#'
#' The following issues are linted by default by this linter
#'   (see arguments for which can be de-activated optionally):
#'
#'  1. Block usage of [base::paste()] with `sep = ""`. [base::paste0()] is a faster, more concise alternative;
#'     this is valid unless `paste` occurs inside [base::expression], which according to [grDevices::plotmath]
#'     does not support the `sep` argument.
#'  2. Block usage of `paste()` or `paste0()` with `collapse = ", "`. [toString()] is a direct
#'     wrapper for this, and alternatives like [glue::glue_collapse()] might give better messages for humans.
#'  3. Block usage of `paste0()` that supplies `sep=` -- this is not a formal argument to `paste0`, and
#'     is likely to be a mistake.
#'  4. Block usage of `paste()` / `paste0()` combined with [rep()] that could be replaced by
#'     [base::strrep()]. `strrep()` can handle the task of building a block of repeated strings
#'     (e.g. often used to build "horizontal lines" for messages). This is both more readable and
#'     skips the (likely small) overhead of putting two strings into the global string cache when only one is needed.
#'
#'     Only target scalar usages -- `strrep` can handle more complicated cases (e.g. `strrep(letters, 26:1)`,
#'     but those aren't as easily translated from a `paste(collapse=)` call.
#'  5. Block usage of `paste()` or `paste0()` to collapse the output of [base::deparse()], e.g.
#'     `paste(deparse(x), collapse = " ")`. [base::deparse1()] is a more readable equivalent,
#'     i.e. `deparse1(x)`.
#'
#' @evalRd rd_tags("paste_linter")
#' @param allow_empty_sep Logical, default `FALSE`. If `TRUE`, usage of
#'   `paste()` with `sep = ""` is not linted.
#' @param allow_to_string Logical, default `FALSE`. If `TRUE`, usage of
#'   `paste()` and `paste0()` with `collapse = ", "` is not linted.
#' @param allow_file_path String, one of `"never"`, `"double_slash"`, or `"always"`; `"double_slash"` by default.
#'   If `"always"`, usage of `paste()` and `paste0()` to construct file paths is not linted. If `"double_slash"`,
#'   strings containing consecutive forward slashes will not lint. The main use case here is for URLs -- "paths" like
#'   `"https://"` will not induce lints, since constructing them with `file.path()` might be deemed unnatural.
#'   Lastly, if `"never"`, strings with consecutive forward slashes will also lint. Note that `"//"` is never linted
#'   when it comes at the beginning or end of the input, to avoid requiring empty inputs like
#'  `file.path("", ...)` or `file.path(..., "")`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'paste("a", "b", sep = "")',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'paste(c("a", "b"), collapse = ", ")',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'paste0(c("a", "b"), sep = " ")',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'paste0(rep("*", 10L), collapse = "")',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'paste0("http://site.com/", path)',
#'   linters = paste_linter(allow_file_path = "never")
#' )
#'
#' lint(
#'   text = 'paste0(x, collapse = "")',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'expression(paste("a", "b", sep = ""))',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'paste(deparse(x), collapse = " ")',
#'   linters = paste_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'paste0("a", "b")',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'paste("a", "b", sep = "")',
#'   linters = paste_linter(allow_empty_sep = TRUE)
#' )
#'
#' lint(
#'   text = 'toString(c("a", "b"))',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'paste(c("a", "b"), collapse = ", ")',
#'   linters = paste_linter(allow_to_string = TRUE)
#' )
#'
#' lint(
#'   text = 'paste(c("a", "b"))',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'strrep("*", 10L)',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'paste0(year, "/", month, "/", day)',
#'   linters = paste_linter(allow_file_path = "always")
#' )
#'
#' lint(
#'   text = 'paste0("http://site.com/", path)',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'paste(x, collapse = "")',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = 'expression(paste("a", "b"))',
#'   linters = paste_linter()
#' )
#'
#' lint(
#'   text = "deparse1(x)",
#'   linters = paste_linter()
#' )
#'
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
paste_linter <- function(allow_empty_sep = FALSE,
                         allow_to_string = FALSE,
                         allow_file_path = c("double_slash", "always", "never")) {
  allow_file_path <- match.arg(allow_file_path)
  check_file_paths <- allow_file_path %in% c("double_slash", "never")

  ancestor_expr_cond <-
    "parent::expr/ancestor-or-self::expr/preceding-sibling::expr/SYMBOL_FUNCTION_CALL[text() = 'expression']"
  paste_sep_xpath <- glue("
  following-sibling::SYMBOL_SUB[text() = 'sep' and following-sibling::expr[1][STR_CONST] and not({ancestor_expr_cond})]
    /parent::expr
  ")
  expression_paste_sep_xpath <- glue("
  following-sibling::SYMBOL_SUB[text() = 'sep' and following-sibling::expr[1][STR_CONST] and {ancestor_expr_cond}]
    /parent::expr
  ")

  to_string_xpath <- "
  parent::expr[
    count(expr) = 3
    and SYMBOL_SUB[text() = 'collapse']/following-sibling::expr[1][STR_CONST]
  ]"

  paste0_sep_xpath <- "
  following-sibling::SYMBOL_SUB[text() = 'sep']
    /parent::expr
  "

  paste_strrep_xpath <- "
  self::*[
    count(following-sibling::expr) = 2
    and following-sibling::expr[1][expr[1][SYMBOL_FUNCTION_CALL[text() = 'rep']] and expr[2][STR_CONST]]
    and following-sibling::SYMBOL_SUB[text() = 'collapse']
  ]
    /parent::expr
  "

  # Type II: paste0(x, "/", y, "/", z)
  #   NB: some conditions require evaluating the R string, only a few can be done in pure XPath. See below.
  paste0_file_path_xpath <- xp_strip_comments("
  parent::expr[
    (: exclude paste0(x) :)
    count(expr) > 2
    (: An expression matching _any_ of these conditions is _not_ a file path :)
    and not(
      (: Any numeric input :)
      expr/NUM_CONST
      (: A call using collapse= :)
      or SYMBOL_SUB[text() = 'collapse']
      (: Consecutive non-strings like paste0(x, y) :)
      or expr[(SYMBOL or expr) and following-sibling::expr[1][SYMBOL or expr]]
    )
  ]")

  empty_paste_note <-
    'Note that paste() converts empty inputs to "", whereas file.path() leaves it empty.'

  paste0_collapse_xpath <- glue("
  parent::expr[
    SYMBOL_SUB[text() = 'collapse']
    and count(expr) =
      3 - count(preceding-sibling::*[self::PIPE or self::SPECIAL[{ xp_text_in_table(magrittr_pipes) }]])
    and not(expr/SYMBOL[text() = '...'])
  ]")

  # Skip collapse = NULL (does not collapse, so not deparse1()) and deparse() supplied
  #   as the collapse separator rather than as the collapsed vector.
  deparse1_xpath <- "
  parent::expr[
    count(expr) = 3
    and SYMBOL_SUB[text() = 'collapse']
    and not(SYMBOL_SUB[text() = 'collapse']/following-sibling::expr[1]/NULL_CONST)
    and expr[
      expr[1]/SYMBOL_FUNCTION_CALL[text() = 'deparse']
      and not(preceding-sibling::*[not(self::COMMENT)][1][self::EQ_SUB])
    ]
  ]"

  Linter(linter_level = "expression", function(source_expression) {
    paste_calls <- source_expression$xml_find_function_calls("paste")
    paste0_calls <- source_expression$xml_find_function_calls("paste0")
    both_calls <- combine_nodesets(paste_calls, paste0_calls)

    optional_lints <- list()

    # Both of these look for paste(..., sep = "..."), differing in which 'sep' is linted,
    #   so run the expensive XPath search/R parse only once
    if (!allow_empty_sep || check_file_paths) {
      paste_sep_expr <- xml_find_all_(paste_calls, paste_sep_xpath)
      paste_sep_value <- get_r_string(paste_sep_expr, xpath = "./SYMBOL_SUB[text() = 'sep']/following-sibling::expr[1]")
    }

    ## check if we are inside an expression()
    expression_paste_sep_expr <- xml_find_all_(paste_calls, expression_paste_sep_xpath)
    optional_lints <- c(optional_lints, xml_nodes_to_lints(
      expression_paste_sep_expr,
      source_expression = source_expression,
      lint_message = "inside expression(...), paste does not accept a 'sep' argument.",
      type = "warning"
    ))

    if (!allow_empty_sep) {
      optional_lints <- c(optional_lints, xml_nodes_to_lints(
        paste_sep_expr[!nzchar(paste_sep_value)],
        source_expression = source_expression,
        lint_message = 'paste0(...) is better than paste(..., sep = "").',
        type = "warning"
      ))
    }

    if (!allow_to_string) {
      # 3 expr: the function call, the argument, and collapse=
      to_string_expr <- xml_find_all_(both_calls, to_string_xpath)
      collapse_value <- get_r_string(
        to_string_expr,
        xpath = "./SYMBOL_SUB[text() = 'collapse']/following-sibling::expr[1]"
      )

      optional_lints <- c(optional_lints, xml_nodes_to_lints(
        to_string_expr[collapse_value == ", "],
        source_expression = source_expression,
        lint_message = paste(
          'toString(.) is more expressive than paste(., collapse = ", ").',
          "Note also glue::glue_collapse() and and::and()",
          "for constructing human-readable / translation-friendly lists"
        ),
        type = "warning"
      ))
    }

    paste0_sep_expr <- xml_find_all_(paste0_calls, paste0_sep_xpath)
    paste0_sep_lints <- xml_nodes_to_lints(
      paste0_sep_expr,
      source_expression = source_expression,
      lint_message = "sep= is not a formal argument to paste0(); did you mean to use paste(), or collapse=?",
      type = "warning"
    )

    paste_strrep_expr <- xml_find_all_(both_calls, paste_strrep_xpath)
    collapse_arg <- get_r_string(paste_strrep_expr, "SYMBOL_SUB/following-sibling::expr[1]/STR_CONST")
    paste_strrep_expr <- paste_strrep_expr[!nzchar(collapse_arg)]
    paste_call <- xp_call_name(paste_strrep_expr)
    paste_strrep_lints <- xml_nodes_to_lints(
      paste_strrep_expr,
      source_expression = source_expression,
      lint_message = sprintf('strrep(x, times) is better than %s(rep(x, times), collapse = "").', paste_call),
      type = "warning"
    )

    paste0_collapse_expr <- xml_find_all_(paste0_calls, paste0_collapse_xpath)
    paste0_collapse_lints <- xml_nodes_to_lints(
      paste0_collapse_expr,
      source_expression = source_expression,
      lint_message = "Use paste(), not paste0(), to collapse a character vector when sep= is not used.",
      type = "warning"
    )

    deparse1_expr <- xml_find_all_(both_calls, deparse1_xpath)
    deparse1_lints <- xml_nodes_to_lints(
      deparse1_expr,
      source_expression = source_expression,
      lint_message = "Use deparse1(x) instead of paste(deparse(x), collapse = ...).",
      type = "warning"
    )

    if (check_file_paths) {
      paste_sep_slash_expr <- paste_sep_expr[paste_sep_value == "/"]
      optional_lints <- c(optional_lints, xml_nodes_to_lints(
        # in addition to paste(..., sep = "/") we ensure collapse= is not present
        paste_sep_slash_expr[is.na(xml_find_first_(paste_sep_slash_expr, "./SYMBOL_SUB[text() = 'collapse']"))],
        source_expression = source_expression,
        lint_message = paste(
          'Construct file paths with file.path(...) instead of paste(..., sep = "/").',
          'If you are using paste(sep = "/") to construct a date,',
          "consider using format() or lubridate helpers instead.",
          empty_paste_note
        ),
        type = "warning"
      ))

      paste0_file_path_expr <- xml_find_all_(paste0_calls, paste0_file_path_xpath)
      is_file_path <-
        !vapply(paste0_file_path_expr, check_is_not_file_path, logical(1L), allow_file_path = allow_file_path)
      optional_lints <- c(optional_lints, xml_nodes_to_lints(
        paste0_file_path_expr[is_file_path],
        source_expression = source_expression,
        lint_message = paste(
          'Construct file paths with file.path(...) instead of paste0(x, "/", y, "/", z).',
          empty_paste_note
        ),
        type = "warning"
      ))
    }

    c(optional_lints, paste0_sep_lints, paste_strrep_lints, paste0_collapse_lints, deparse1_lints)
  })
}

check_is_not_file_path <- function(expr, allow_file_path) {
  arguments <- xml_find_all_(expr, "expr[position() > 1]")

  is_string <- !is.na(xml_find_first_(arguments, "STR_CONST"))
  string_values <- character(length(arguments))
  string_values[is_string] <- get_r_string(arguments[is_string])
  not_start_slash <- which(!startsWith(string_values, "/"))
  not_end_slash <- which(!endsWith(string_values, "/"))

  if (allow_file_path == "double_slash") {
    check_double_slash <- function(str) any(grepl("//", str, fixed = TRUE))
  } else {
    # always skip on strings starting/ending with //, since switching to
    #   file.path() would require file.path("", ...) or file.path(..., "")
    check_double_slash <- function(str) any(grepl("^//|//$", str))
  }

  # First input is '/', meaning file.path() would need to start with ''
  string_values[1L] == "/" ||
    # Last input is '/', meaning file.path() would need to end with ''
    string_values[length(string_values)] == "/" ||
    check_double_slash(string_values) ||
    # A string not ending with /, followed by non-string,
    #   or a string not starting with /, preceded by a non-string
    !all(is_string[c(not_end_slash + 1L, not_start_slash - 1L)], na.rm = TRUE) ||
    # A string not starting with / preceded by a string not ending with /
    any(not_start_slash %in% (not_end_slash + 1L))
}

#' Brace linter
#'
#' Perform various style checks related to placement and spacing of curly braces:
#'  - Opening curly braces are never on their own line and are always followed by a newline.
#'  - Opening curly braces have a space before them.
#'  - Closing curly braces are on their own line unless they are followed by an `else`.
#'  - Closing curly braces in `if` conditions are on the same line as the corresponding `else`.
#'  - Either both or neither branch in `if`/`else` use curly braces, i.e., either both branches use `{...}` or neither
#'    does.
#'  - Function bodies are wrapped in curly braces.
#'
#' @param allow_single_line If `TRUE`, allow an open and closed curly pair on the same line.
#' @param function_bodies When to require function bodies to be wrapped in curly braces. One of
#'   - `"always"` to require braces around all function bodies, including inline functions,
#'   - `"not_inline"` to require braces when a function body does not start on the same line as its signature,
#'   - `"multi_line"` (the default) to require braces when a function definition spans multiple lines,
#'   - `"never"` to never require braces in function bodies.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "f <- function() { 1 }",
#'   linters = brace_linter()
#' )
#'
#' writeLines("if (TRUE) {\n return(1) }")
#' lint(
#'   text = "if (TRUE) {\n return(1) }",
#'   linters = brace_linter()
#' )
#'
#' # okay
#' writeLines("f <- function() {\n  1\n}")
#' lint(
#'   text = "f <- function() {\n  1\n}",
#'   linters = brace_linter()
#' )
#'
#' writeLines("if (TRUE) { \n return(1) \n}")
#' lint(
#'   text = "if (TRUE) { \n return(1) \n}",
#'   linters = brace_linter()
#' )
#'
#' # customizing using arguments
#' writeLines("if (TRUE) { return(1) }")
#' lint(
#'   text = "if (TRUE) { return(1) }",
#'   linters = brace_linter(allow_single_line = TRUE)
#' )
#' @evalRd rd_tags("brace_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#indenting>
#' - <https://style.tidyverse.org/syntax.html#if-statements>
#' @export
brace_linter <- function(allow_single_line = FALSE,
                         function_bodies = c("multi_line", "always", "not_inline", "never")) {
  function_bodies <- match.arg(function_bodies)

  xp_cond_open <- xp_and(c(
    # matching } is on same line
    if (isTRUE(allow_single_line)) {
      "(@line1 != following-sibling::OP-LEFT-BRACE/@line1)"
    },
    # double curly
    "not(
      (@line1 = parent::expr/preceding-sibling::OP-LEFT-BRACE/@line1)
      or (@line1 = following-sibling::expr/OP-LEFT-BRACE/@line1)
    )",
    # allow `(`, `,` and `%>%` on preceding line
    #
    # note that '{' is not supported in RHS call of base-R's native pipe (`|>`),
    # so no exception needs to be made for this operator
    glue("not(
      @line1 > parent::expr/preceding-sibling::*[not(self::COMMENT)][1][
        self::OP-LEFT-PAREN
        or self::OP-COMMA
        or (self::SPECIAL and ({xp_text_in_table(magrittr_pipes)}) )
      ]/@line2
    )")
  ))

  # TODO(#1103): if c_style_braces is TRUE, invert the preceding-sibling condition
  xp_open_curly <- glue("//OP-LEFT-BRACE[
    { xp_cond_open }
    and (
      not(@line1 = parent::expr/preceding-sibling::*/@line2)
      or @line1 = following-sibling::*[1][not(self::COMMENT or self::OP-RIGHT-BRACE)]/@line1
    )
  ]")

  xp_open_preceding <- "parent::expr/preceding-sibling::*[1][self::OP-RIGHT-PAREN or self::ELSE or self::REPEAT]"

  xp_paren_brace <- glue("//OP-LEFT-BRACE[
    @line1 = { xp_open_preceding }/@line1
    and @col1 = { xp_open_preceding }/@col2 + 1
  ]")

  xp_cond_closed <- xp_and(c(
    # matching { is on same line
    if (isTRUE(allow_single_line)) {
      "(@line1 != preceding-sibling::OP-LEFT-BRACE/@line1)"
    },
    # immediately followed by ",", "]" or ")"
    "not(
      @line1 = ancestor::expr/following-sibling::*[1][
        self::OP-COMMA or self::OP-RIGHT-BRACKET or self::OP-RIGHT-PAREN
      ]
        /@line1
    )",
    # double curly
    "not(
      (@line1 = parent::expr/following-sibling::OP-RIGHT-BRACE/@line1)
      or (@line1 = preceding-sibling::expr/OP-RIGHT-BRACE/@line1)
    )"
  ))

  # TODO(#1103): if c_style_braces is TRUE, skip the not(ELSE) condition
  xp_closed_curly <- glue("//OP-RIGHT-BRACE[
    { xp_cond_closed }
    and (
      (@line1 = preceding-sibling::*[1][not(self::OP-LEFT-BRACE)]/@line2)
      or (@line1 = parent::expr/following-sibling::*[not(self::COMMENT)][1][not(self::ELSE)]/@line1)
    )
  ]")

  xp_else_closed_curly <- "preceding-sibling::IF/following-sibling::expr[2]/OP-RIGHT-BRACE"
  # need to (?) repeat previous_curly_path since != will return true if there is
  #   no such node. ditto for approach with not(@line1 = ...).
  # TODO(#1103): if c_style_braces is TRUE, this needs to be @line2 + 1
  xp_else_same_line <- glue("//ELSE[{xp_else_closed_curly} and @line1 != {xp_else_closed_curly}/@line2]")

  if (function_bodies != "never") {
    xp_cond_function_brace <- switch(
      function_bodies,
      always = "1",
      multi_line = "@line1 != @line2",
      not_inline = "@line1 != expr/@line1"
    )

    xp_function_brace <- glue(
      "(//FUNCTION | //OP-LAMBDA)/parent::expr[{xp_cond_function_brace} and not(expr/OP-LEFT-BRACE)]"
    )

    msg_function_brace <- switch(
      function_bodies,
      always = "Wrap function bodies in curly braces.",
      multi_line = "Wrap multi-line function bodies in curly braces.",
      not_inline = "Wrap function bodies starting on a new line in curly braces."
    )
  }

  # if (x) { ... } else if (y) { ... } else { ... } is OK; fully exact pairing
  #   of if/else would require this to be
  #   if (x) { ... } else { if (y) { ... } else { ... } } since there's no
  #   elif operator/token in R, which is pretty unseemly
  xp_if_else_match_brace <- "
  //IF[
    following-sibling::expr[2][OP-LEFT-BRACE]
    and
      following-sibling::ELSE
        /following-sibling::expr[1][not(OP-LEFT-BRACE or IF/following-sibling::expr[2][OP-LEFT-BRACE])]
  ]

  |

  //ELSE[
    following-sibling::expr[1][OP-LEFT-BRACE]
    and preceding-sibling::IF/following-sibling::expr[2][not(OP-LEFT-BRACE)]
  ]
  "

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    lints <- list()

    lints <- c(
      lints,
      xml_nodes_to_lints(
        xml_find_all_(xml, xp_open_curly),
        source_expression = source_expression,
        lint_message =
          "Opening curly braces should never go on their own line and should always be followed by a new line."
      )
    )

    lints <- c(
      lints,
      xml_nodes_to_lints(
        xml_find_all_(xml, xp_paren_brace),
        source_expression = source_expression,
        lint_message = "There should be a space before an opening curly brace."
      )
    )

    lints <- c(
      lints,
      xml_nodes_to_lints(
        xml_find_all_(xml, xp_closed_curly),
        source_expression = source_expression,
        lint_message =
          "Closing curly-braces should always be on their own line, unless they are followed by an else."
      )
    )

    lints <- c(
      lints,
      xml_nodes_to_lints(
        xml_find_all_(xml, xp_else_same_line),
        source_expression = source_expression,
        lint_message = "`else` should come on the same line as the previous `}`."
      )
    )
    if (function_bodies != "never") {
      lints <- c(
        lints,
        xml_nodes_to_lints(
          xml_find_all_(xml, xp_function_brace),
          source_expression = source_expression,
          lint_message = msg_function_brace
        )
      )
    }

    lints <- c(
      lints,
      xml_nodes_to_lints(
        xml_find_all_(xml, xp_if_else_match_brace),
        source_expression = source_expression,
        lint_message = "Either both or neither branch in `if`/`else` should use curly braces."
      )
    )

    lints
  })
}

declared_s3_generics <- function(x) {
  # Top level expression which assigns to a symbol
  #   and is an S3 Generic (contains call to UseMethod)
  # Retrieve assigned name of the function
  xpath <- "/exprlist
    /*[
      (LEFT_ASSIGN or EQ_ASSIGN)
      and expr[FUNCTION or OP-LAMBDA]
      and .//SYMBOL_FUNCTION_CALL[text() = 'UseMethod']
    ]
    /expr
    /SYMBOL
  "

  xml_text(xml_find_all_(x, xpath))
}

#' Parsed sourced file from a filename
#'
#' This object is given as input to each linter.
#'
#' @details
#' The file is read using the `encoding` setting.
#' This setting is found by taking the first valid result from the following locations
#'
#' 1. The `encoding` key from the usual lintr configuration settings.
#' 2. The `Encoding` field from a Package `DESCRIPTION` file in a parent directory.
#' 3. The `Encoding` field from an R Project `.Rproj` file in a parent directory.
#' 4. `"UTF-8"` as a fallback.
#'
#' @param filename the file to be parsed.
#' @param lines a character vector of lines.
#'   If `NULL`, then `filename` will be read.
#' @return A `list` with three components:
#' \describe{
#'   \item{expressions}{a `list` of
#'   `n+1` objects. The first `n` elements correspond to each expression in
#'   `filename`, and consist of a list of 8 elements:
#'   \itemize{
#'     \item{`filename` (`character`) the name of the file.}
#'     \item{`line` (`integer`) the line in the file where this expression begins.}
#'     \item{`column` (`integer`) the column in the file where this expression begins.}
#'     \item{`lines` (named `character`) vector of all lines spanned by this
#'           expression, named with the corresponding line numbers.}
#'     \item{`parsed_content` (`data.frame`) as given by [utils::getParseData()] for this expression.}
#'     \item{`xml_parsed_content` (`xml_document`) the XML parse tree of this expression as given by
#'           [xmlparsedata::xml_parse_data()].}
#'     \item{`content` (`character`) the same as `lines` as a single string (not split across lines).}
#'     \item{`xml_find_function_calls(function_names)` (`function`) a function that returns all `SYMBOL_FUNCTION_CALL`
#'           XML nodes from `xml_parsed_content` with specified function names.}
#'   }
#'
#'   The final element of `expressions` is a list corresponding to the full file
#'   consisting of 7 elements:
#'   \itemize{
#'     \item{`filename` (`character`) the name of this file.}
#'     \item{`file_lines` (`character`) the [readLines()] output for this file.}
#'     \item{`content` (`character`) for .R files, the same as `file_lines`;
#'           for .Rmd or .qmd scripts, this is the extracted R source code (as text).}
#'     \item{`full_parsed_content` (`data.frame`) as given by
#'           [utils::getParseData()] for the full content.}
#'     \item{`full_xml_parsed_content` (`xml_document`) the XML parse tree of all
#'           expressions as given by [xmlparsedata::xml_parse_data()].}
#'     \item{`terminal_newline` (`logical`) records whether `filename` has a terminal
#'           newline (as determined by [readLines()] producing a corresponding warning).}
#'     \item{`xml_find_function_calls(function_names)` (`function`) a function that returns all `SYMBOL_FUNCTION_CALL`
#'           XML nodes from `full_xml_parsed_content` with specified function names.}
#'   }
#'   }
#'   \item{error}{A `Lint` object describing any parsing error.}
#'   \item{warning}{A `lints` object describing any parsing warning.}
#'   \item{lines}{The [readLines()] output for this file.}
#' }
#'
#' @examples
#' tmp <- tempfile()
#' writeLines(c("x <- 1", "y <- x + 1"), tmp)
#' get_source_expressions(tmp)
#' unlink(tmp)
#' @export
get_source_expressions <- function(filename, lines = NULL) {
  source_expression <- srcfile(filename, encoding = settings$encoding)

  # Ensure English locale for terminal newline and zero-length variable warning messages
  old_lang <- set_lang("en")
  on.exit(reset_lang(old_lang))

  encoding <- if (is.null(lines)) settings$encoding
  source_expression$lines <- ensure_utf8(lines %||% read_lines(filename), encoding)

  # Only regard explicit attribute terminal_newline=FALSE as FALSE and all other cases (e.g. NULL or TRUE) as TRUE.
  terminal_newline <- !isFALSE(attr(source_expression$lines, "terminal_newline", exact = TRUE))

  e <- w <- NULL
  source_expression$lines <- extract_r_source(
    filename = source_expression$filename,
    lines = source_expression$lines,
    error = \(e) lint_rmd_error(e, source_expression)
  )
  names(source_expression$lines) <- seq_along(source_expression$lines)
  source_expression$content <- get_content(source_expression$lines)
  parsed_content <- get_source_expression(source_expression, error = \(e) lint_parse_error(e, source_expression))

  # Currently no way to distinguish the source of the warning
  #   from the message itself, so we just grep the source for the
  #   exact string generating the warning; de-dupe in case of
  #   multiple exact matches like '1e-3L; 1e-3L'.
  # See https://bugs.r-project.org/show_bug.cgi?id=18863.
  w <- lint_parse_warnings(w, parsed_content, source_expression)

  if (is_lint(e) && (is.na(e$line) || !nzchar(e$line) || e$message == "unexpected end of input")) {
    # Don't create expression list if it's unreliable (invalid encoding or unhandled parse error)
    return(list(expressions = list(), error = e, warning = w, lines = source_expression$lines))
  }

  top_level_map <- generate_top_level_map(parsed_content)
  xml_parsed_content <- safe_parse_to_xml(parsed_content)

  expressions <- lapply(
    X = top_level_expressions(parsed_content),
    FUN = get_single_source_expression,
    parsed_content,
    source_expression,
    filename,
    top_level_map
  )

  expressions <- maybe_append_expression_xml(expressions, xml_parsed_content)

  # add global expression
  expressions[[length(expressions) + 1L]] <- list(
    filename = filename,
    file_lines = source_expression$lines,
    content = source_expression$lines,
    full_parsed_content = parsed_content,
    full_xml_parsed_content = xml_parsed_content,
    xml_find_function_calls = build_xml_find_function_calls(xml_parsed_content),
    terminal_newline = terminal_newline
  )

  list(expressions = expressions, error = e, warning = w, lines = source_expression$lines)
}

lint_parse_error <- function(e, source_expression) {
  # R >= 4.3.0
  if (inherits(e, "parseError")) {
    return(lint_parse_error_r43(e, source_expression))
  }

  message_info <- re_matches(
    e$message,
    rex(
      except_some_of(":"),
      ":",
      capture(name = "line", digits),
      ":",
      capture(name = "column", digits),
      ":",
      space,
      capture(name = "message", anything),
      "\n"
    )
  )

  if (!is.na(message_info$line)) {
    return(lint_parse_error_r42(message_info, source_expression))
  }

  # an error that does not use R_ParseErrorMsg
  lint_parse_error_nonstandard(e, source_expression)
}

#' Currently no way to distinguish the source of the warning
#'   from the message itself, so we just grep the source for the
#'   exact string generating the warning; de-dupe in case of
#'   multiple exact matches like '1e-3L; 1e-3L'
#' @noRd
lint_parse_warnings <- function(w, parsed_content, source_expression) {
  if (!length(w)) {
    return(w)
  }
  flatten_lints(lapply(unique(w), lint_parse_warning, parsed_content, source_expression))
}

#' The set of parser warnings seems pretty stable, but as
#'   long as they don't generate sourceref hints, we're
#'   stuck with this somewhat-manual approach.
#' @noRd
lint_parse_warning <- function(w, parsed_content, source_expression) {
  for (lint_re in parser_warning_regexes) {
    bad_txt <- re_matches(w, lint_re)$txt[1L]
    # at most one regex matches the warning
    if (is.na(bad_txt)) {
      next
    } else {
      break
    }
  }
  # use the parse tree to avoid baroque matches to comments, strings
  hits <- parsed_content[with(parsed_content, token == "NUM_CONST" & text == bad_txt), ]
  lapply(seq_len(nrow(hits)), function(ii) {
    Lint(
      filename = source_expression$filename,
      line_number = hits$line1[ii],
      column_number = hits$col1[ii],
      type = "warning",
      message = w,
      line = source_expression$lines[[as.character(hits$line1[ii])]],
      ranges = list(c(hits$col1[ii], hits$col2[ii]))
    )
  })
}

parser_warning_regexes <- list(
  int_with_decimal =
    rex("integer literal ", capture(anything, name = "txt"), " contains decimal; using numeric value"),
  nonint_with_l =
    rex("non-integer value ", capture(anything, name = "txt"), " qualified with L; using numeric value"),
  unneeded_decimal =
    rex("integer literal ", capture(anything, name = "txt"), " contains unnecessary decimal point")
)

#' Ensure a string is valid for printing
#'
#' Helper to ensure a valid string is provided as line where necessary.
#' Handles two cases:
#'
#'  1. NA lines, as generated by running outside of code blocks in Rmd documents
#'  2. invalid string lines, as generated by invalid encoding settings
#'
#' [nchar()] can detect both cases: 1. returns `NA_integer_`, 2. errors
#'
#' @param line Possibly misencoded or NA line
#'
#' @return `line` if it's a valid non-NA string, otherwise an empty string.
#'
#' @noRd
fixup_line <- function(line) {
  nchars <- tryCatch(nchar(line, type = "chars"), error = \(e) NA_integer_)
  if (is.na(nchars)) {
    ""
  } else {
    line
  }
}

#' Convert an R >= 4.3.0 classed parseError with metadata into a lint
#'
#' @param e A parse error of class `parseError` generated by R >= 4.3.0
#' @param source_expression The source expression that generated the parse error
#'
#' @return A [Lint()] based on the metadata attached to `e`.
#'
#' @noRd
lint_parse_error_r43 <- function(e, source_expression) {
  msg <- re_substitutes(e$message, rex(" (", except_some_of(")"), ")", end), "")
  line_number <- e$lineno
  column <- e$colno
  substr(msg, 1L, 1L) <- toupper(substr(msg, 1L, 1L))
  msg <- paste0(msg, ".")

  if (inherits(e, "invalidMBCS")) {
    msg <- paste(msg, "Is the encoding correct?")
  }

  if (column == 0L) {
    line_number <- line_number - 1L
  }

  # Safely handle invalid location info
  if (line_number < 1L || line_number > length(source_expression$lines)) {
    # nocov start
    line_number <- 1L
    column <- 1L
    # nocov end
  }

  line <- fixup_line(source_expression$lines[[line_number]])
  if (column == 0L || column > nchar(line) + 1L) {
    column <- nchar(line) + 1L
  }

  Lint(
    filename = source_expression$filename,
    line_number = line_number,
    column_number = column,
    type = "error",
    message = msg,
    line = line
  )
}

#' Convert a R < 4.3.0 standard parse error message into a lint
#'
#' @param message_info Match of the structured parse error message regex, matched in `lint_parse_error()`
#' @param source_expression The source expression that generated the parse error
#'
#' @return A [Lint()] based on text mining of the error message captured by `message_info`,
#'
#' @noRd
lint_parse_error_r42 <- function(message_info, source_expression) {
  line_number <- as.integer(message_info$line)
  column_number <- as.integer(message_info$column)

  # If the column number is zero it means the error really occurred at the
  # end of the previous line
  if (column_number %==% 0L) {
    line_number <- line_number - 1L
    line <- fixup_line(source_expression$lines[[line_number]])
    column_number <- nchar(line) + 1L
  } else {
    line <- fixup_line(source_expression$lines[[line_number]])
  }

  Lint(
    filename = source_expression$filename,
    line_number = line_number,
    column_number = column_number,
    type = "error",
    message = message_info$message,
    line = line
  )
}

#' Convert a R < 4.3.0 non-standard parse error message into a lint
#'
#' Uses a hand-crafted regex and some heuristics to find location information, falling back to Line 1, Column 1 if all
#' attempts fail.
#'
#' @param e A parse error that doesn't fit the standard `message_info` regex used for most parse errors in R < 4.3.0
#' @param source_expression The source expression that generated the parse error
#'
#' @return A [Lint()] based on trying to extract information from the error message of `e`.
#'
#' @noRd
lint_parse_error_nonstandard <- function(e, source_expression) {
  if (grepl("invalid multibyte character in parser at line", e$message, fixed = TRUE)) {
    # nocov start: platform-specific
    l <- as.integer(re_matches(
      e$message,
      rex("invalid multibyte character in parser at line ", capture(name = "line", digits))
    )$line)
    # Invalid encoding in source code
    return(
      Lint(
        filename = source_expression$filename,
        line_number = l,
        column_number = 1L,
        type = "error",
        message = "Invalid multibyte character in parser. Is the encoding correct?",
        line = fixup_line(source_expression$lines[[l]])
      )
    )
    # nocov end
  } else if (grepl("invalid multibyte string, element", e$message, fixed = TRUE)) {
    # Invalid encoding, will break even re_matches() below, so we need to handle this first.
    return(
      Lint(
        filename = source_expression$filename,
        line_number = 1L,
        column_number = 1L,
        type = "error",
        message = "Invalid multibyte string. Is the encoding correct?",
        line = ""
      )
    )
  } else if (grepl("repeated formal argument", e$message, fixed = TRUE)) {
    # nocov start: only throws on certain versions; tested on L389
    matches <- re_matches(
      e$message,
      rex(
        "repeated formal argument '",
        capture(name = "symbol", anything),
        "' on line ",
        capture(name = "line", digits)
      )
    )
    sym <- matches$symbol
    l <- as.integer(matches$line)
    # Repeated formal argument 'sym' on line l
    return(
      Lint(
        filename = source_expression$filename,
        line_number = l,
        column_number = 1L,
        type = "error",
        message = sprintf("Repeated formal argument '%s'.", sym),
        line = fixup_line(source_expression$lines[[l]])
      )
    )
    # nocov end
  }

  # Hand-crafted regex to parse all error messages generated by the R parser code for R < 4.3.
  # The error messages can be found in src/main/gram.c and src/main/character.c in the R code.
  # This code produces a list of all possible error messages:
  #
  # nolint start: commented_code_linter.
  # parser_files <- c("src/main/gram.c", "src/main/character.c")
  #
  # lines <- unlist(lapply(
  #   file.path("https://raw.githubusercontent.com/wch/r-source/trunk", parser_files),
  #   readLines
  # ))
  # error_calls <- grep("error(_(", lines, fixed = TRUE, value = TRUE)
  # error_formats <- trimws(gsub("^.*error\\(_\\(\"(.+)\".+", "\\1", error_calls))
  # error_formats <- unique(error_formats)
  # nolint end
  parse_error_rx <- rex(
    start,
    capture(anything, name = "msg_1"),
    or(" at ", " on ", " ("),
    "line ",
    capture(digits, name = "line"),
    maybe(")"),
    capture(anything, name = "msg_2"),
    end
  )

  if (grepl(parse_error_rx, e$message, perl = TRUE)) {
    # nocov start: only throws on certain versions. Tested on L396: Nul character not allowed
    rx_match <- re_matches(
      e$message,
      parse_error_rx
    )
    l <- as.integer(rx_match$line)
    # Sometimes the parser "line" runs one past the last line
    l <- pmin(l, length(source_expression$lines))

    msg <- paste0(rx_match$msg_1, if (nzchar(rx_match$msg_2)) " ", rx_match$msg_2, ".")
    substr(msg, 1L, 1L) <- toupper(substr(msg, 1L, 1L))

    return(
      Lint(
        filename = source_expression$filename,
        line_number = l,
        column_number = 1L,
        type = "error",
        message = msg,
        line = source_expression$lines[[l]]
      )
    )
    # nocov end
  }

  message_info <- re_matches(
    e$message,
    rex(
      single_quotes, capture(name = "name", anything), single_quotes,
      anything,
      double_quotes, capture(name = "starting", anything), double_quotes
    )
  )

  loc <- re_matches(source_expression$content, rex(message_info$starting), locations = TRUE)
  line_location <- loc[!is.na(loc$start) & !is.na(loc$end), ]

  if (nrow(line_location) == 0L) {
    if (grepl("attempt to use zero-length variable name", e$message, fixed = TRUE)) {
      # empty symbol: ``, ``(), ''(), ""(), fun(''=42), fun(""=42), fun(a=1,""=42)
      loc <- re_matches(
        source_expression$content,
        rex(
          "``" %or%
            list(or("''", '""'), any_spaces, "(") %or%
            list(or("(", ","), any_spaces, or("''", '""'), any_spaces, "=")
        ),
        options = "multi-line",
        locations = TRUE
      )
      loc <- loc[!is.na(loc$start) & !is.na(loc$end), ]
      if (nrow(loc) > 0L) {
        line_location <- loc[1L, ]
      }
    } else {
      # nocov start
      return(
        Lint(
          filename = source_expression$filename,
          line_number = 1L,
          column_number = 1L,
          type = "error",
          message = e$message,
          line = ""
        )
      )
      # nocov end
    }
  }

  newline_locs <- get_newline_locs(source_expression$content)
  line_number <- which(newline_locs >= line_location$start)[1L] - 1L
  column_number <- line_location$start - newline_locs[line_number]

  Lint(
    filename = source_expression$filename,
    line_number = line_number,
    column_number = column_number,
    type = "error",
    message = e$message,
    line = source_expression$lines[[line_number]]
  )
}

lint_rmd_error <- function(e, source_expression) {
  message_info <- re_matches(
    e$message,
    rex(
      except_some_of(":"),
      ":",
      capture(name = "line", digits),
      ":",
      capture(name = "column", digits),
      ":",
      space,
      capture(name = "message", anything),
      "\n"
    )
  )

  line_number <- as.integer(message_info$line)
  column_number <- as.integer(message_info$column)

  Lint(
    filename = source_expression$filename,
    line_number = line_number,
    column_number = column_number,
    type = "error",
    message = message_info$message,
    line = source_expression$lines[line_number]
  )
}

get_single_source_expression <- function(loc,
                                         parsed_content,
                                         source_expression,
                                         filename,
                                         top_level_map) {
  line_nums <- parsed_content$line1[loc]:parsed_content$line2[loc]
  expr_lines <- source_expression$lines[line_nums]
  names(expr_lines) <- line_nums
  content <- get_content(source_expression$lines, parsed_content[loc, ])
  id <- parsed_content$id[loc]
  pc <- parsed_content[which(top_level_map == id), ]
  list(
    filename = filename,
    line = parsed_content[loc, "line1"],
    column = parsed_content[loc, "col1"],
    lines = expr_lines,
    parsed_content = pc,
    xml_parsed_content = xml_missing(),
    # Placeholder for xml_find_function_calls, if needed (e.g. on R <= 4.0.5 with input source "\\")
    xml_find_function_calls = build_xml_find_function_calls(xml_missing()),
    content = content
  )
}

get_source_expression <- function(source_expression, error = identity) {
  parse_error <- FALSE

  env <- parent.frame() # nolint: object_usage_linter. Used below.
  # https://adv-r.hadley.nz/conditions.html
  parsed_content <- withCallingHandlers(
    tryCatch(
      parse(
        text = source_expression$content,
        srcfile = source_expression,
        keep.source = TRUE
      ),
      error = error
    ),
    warning = function(w) {
      env$w <- c(env$w, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )

  if (is_error(parsed_content) || is_lint(parsed_content)) {
    assign("e", parsed_content, envir = parent.frame())
    parse_error <- TRUE
  }

  # Triggers an error if the lines contain invalid characters.
  parsed_content <- tryCatch(
    nchar(source_expression$content, type = "chars"),
    error = error
  )

  if (is_error(parsed_content) || is_lint(parsed_content)) {
    # Let parse errors take precedence over encoding problems
    if (!parse_error) assign("e", parsed_content, envir = parent.frame())
    return() # parsed_content is unreliable if encoding is invalid
  }

  source_expression$parsed_content <- parsed_content
  fix_octal_escapes(fix_eq_assigns(fix_tab_indentations(source_expression)), source_expression$lines)
}

maybe_append_expression_xml <- function(expressions, xml_parsed_content) {
  if (is.null(xml_parsed_content) || is.na(xml_parsed_content)) {
    return(expressions)
  }
  expression_xmls <- lapply(
    xml_find_all_(xml_parsed_content, "/exprlist/*"),
    \(top_level_expr) xml2::xml_add_parent(xml2::xml_new_root(top_level_expr), "exprlist")
  )
  for (i in seq_along(expressions)) {
    expressions[[i]]$xml_parsed_content <- expression_xmls[[i]]
    expressions[[i]]$xml_find_function_calls <- build_xml_find_function_calls(expression_xmls[[i]])
  }
  expressions
}

get_newline_locs <- function(x) {
  newline_search <- re_matches(x, rex("\n"), locations = TRUE, global = TRUE)[[1L]]$start
  c(
    0L,
    if (!is.na(newline_search[1L])) newline_search,
    nchar(x) + 1L
  )
}

# Fix column numbers when there are tabs
# getParseData() counts 1 tab as a variable number of spaces instead of one:
# https://github.com/wch/r-source/blame/e7401b68ab0e032fce3e376aaca9a5431619b2b4/src/main/gram.y#L512
# The number of spaces is so that the code is brought to the next 8-character indentation level e.g.:
#   "1\t;"          --> "1       ;"
#   "12\t;"         --> "12      ;"
#   "123\t;"        --> "123     ;"
#   "1234\t;"       --> "1234    ;"
#   "12345\t;"      --> "12345   ;"
#   "123456\t;"     --> "123456  ;"
#   "1234567\t;"    --> "1234567 ;"
#   "12345678\t;"   --> "12345678        ;"
#   "123456789\t;"  --> "123456789       ;"
#   "1234567890\t;" --> "1234567890      ;"
fix_tab_indentations <- function(source_expression) {
  parse_data <- getParseData(source_expression)

  if (is.null(parse_data)) {
    return(NULL)
  }

  tab_cols <- gregexpr("\t", source_expression[["lines"]], fixed = TRUE)
  names(tab_cols) <- seq_along(tab_cols)
  matched_lines <- vapply(tab_cols, \(line_match) !is.na(line_match[1L]) && line_match[1L] > 0L, logical(1L))
  if (!any(matched_lines)) {
    return(parse_data)
  }
  tab_cols <- tab_cols[matched_lines]

  fix_columns <- c("line1", "line2", "col1", "col2")
  parse_data[, fix_columns] <- fix_tab_columns(parse_data[, fix_columns], tab_cols)
  parse_data
}

fix_tab_columns <- function(parse_content, tab_cols) {
  dat <- cbind(
    c(parse_content$line1, parse_content$line2),
    c(parse_content$col1, parse_content$col2)
  )
  lines <- as.integer(names(tab_cols))
  for (i in seq_along(tab_cols)) {
    is_curr_line <- dat[, 1L] == lines[[i]]
    if (!any(is_curr_line)) {
      next
    }
    tab_col <- tab_cols[[i]]
    line_tab_offsets <- tab_offsets(tab_col)
    for (j in seq_along(tab_col)) {
      is_line_to_change <- is_curr_line & dat[, 2L] > tab_col[[j]]
      if (any(is_line_to_change)) {
        dat[is_line_to_change, 2L] <- dat[is_line_to_change, 2L] - line_tab_offsets[[j]]
      }
    }
  }
  dat
}

tab_offsets <- function(tab_columns) {
  outer_env <- new.env(parent = emptyenv())
  outer_env$cum_offset <- 0L
  vapply(
    tab_columns - 1L,
    function(tab_idx) {
      # nolint next: object_overwrite_linter. 'offset' is a perfect name here.
      offset <- 7L - (tab_idx + outer_env$cum_offset) %% 8L # using a tab width of 8 characters
      outer_env$cum_offset <- outer_env$cum_offset + offset
      offset
    },
    integer(1L),
    USE.NAMES = FALSE
  )
}

# This function wraps equal assign expressions in a parent expression so they
# are the same as the corresponding <- expression
fix_eq_assigns <- function(pc) {
  if (is.null(pc) || any(c("equal_assign", "expr_or_assign_or_help") %in% pc$token)) {
    return(pc)
  }

  eq_assign_locs <- which(pc$token == "EQ_ASSIGN")
  # check whether the equal-assignment is the final entry
  if (length(eq_assign_locs) == 0L || tail(eq_assign_locs, 1L) == nrow(pc)) {
    return(pc)
  }

  prev_locs <- vapply(eq_assign_locs, prev_with_parent, pc = pc, integer(1L))
  next_locs <- vapply(eq_assign_locs, next_with_parent, pc = pc, integer(1L))
  expr_locs <- prev_locs != lag(next_locs)
  expr_locs[is.na(expr_locs)] <- TRUE

  id_itr <- max(pc$id)

  true_locs <- which(expr_locs)
  n_expr <- length(true_locs)

  supplemental_content <- data.frame(
    line1 = integer(n_expr),
    col1 = integer(n_expr),
    line2 = integer(n_expr),
    col2 = integer(n_expr),
    id = integer(n_expr),
    parent = integer(n_expr),
    token = character(n_expr),
    terminal = logical(n_expr),
    text = character(n_expr)
  )

  for (i in seq_len(n_expr)) {
    start_loc <- true_locs[i]
    end_loc <- true_locs[i]

    prev_loc <- prev_locs[start_loc]
    next_loc <- next_locs[end_loc]

    id_itr <- id_itr + 1L
    supplemental_content[i, ] <- list(
      pc[prev_loc, "line1"],
      pc[prev_loc, "col1"],
      pc[next_loc, "line2"],
      pc[next_loc, "col2"],
      id_itr,
      pc[eq_assign_locs[true_locs[i]], "parent"],
      "expr", # R now uses "equal_assign"
      FALSE,
      ""
    )

    new_parent_locs <- c(
      prev_locs[start_loc:end_loc],
      eq_assign_locs[start_loc:end_loc],
      next_locs[start_loc:end_loc],
      next_loc
    )
    pc[new_parent_locs, "parent"] <- id_itr
  }
  rownames(supplemental_content) <- supplemental_content$id
  res <- rbind(pc, supplemental_content)
  res[order(res$line1, res$col1, res$line2, res$col2, res$id), ]
}

step_with_parent <- function(pc, loc, offset) {
  id <- pc$id[loc]
  parent_id <- pc$parent[loc]

  with_parent <- pc[pc$parent == parent_id, ]
  with_parent <- with_parent[order(with_parent$line1, with_parent$col1, with_parent$line2, with_parent$col2), ]

  loc <- which(with_parent$id == id)

  which(pc$id == with_parent$id[loc + offset])
}

prev_with_parent <- function(pc, loc) step_with_parent(pc, loc, offset = -1L)
next_with_parent <- function(pc, loc) step_with_parent(pc, loc, offset = 1L)

top_level_expressions <- function(pc) {
  if (is.null(pc)) {
    return(integer(0L))
  }
  which(pc$parent <= 0L)
}

# workaround for bad parse data bug for octal escapes
#   https://bugs.r-project.org/show_bug.cgi?id=18323
fix_octal_escapes <- function(pc, lines) {
  # subset first to prevent using nchar() on MBCS input
  is_str_const <- pc$token == "STR_CONST"
  str_const <- pc[is_str_const, ]
  str_const_mismatch <- str_const$col2 - str_const$col1 != nchar(str_const$text) - 1L
  if (!any(str_const_mismatch)) {
    return(pc)
  }
  str_const <- str_const[str_const_mismatch, ]
  out <- character(nrow(str_const))
  single_line <- str_const$line1 == str_const$line2
  out[single_line] <- substr(
    lines[str_const$line1[single_line]],
    str_const$col1[single_line],
    str_const$col2[single_line]
  )
  for (ii in which(!single_line)) {
    out[ii] <- paste(
      c(
        substring(lines[str_const$line1[ii]], str_const$col1[ii]),
        if (str_const$line1[ii] < str_const$line2[ii] - 1L) {
          lines[(str_const$line1[ii] + 1L):(str_const$line2[ii] - 1L)]
        },
        substr(lines[str_const$line2[ii]], 1L, str_const$col2[ii])
      ),
      collapse = "\n"
    )
  }
  pc$text[is_str_const][str_const_mismatch] <- out
  pc
}

test_that("with_id works as expected", {
  source_expression <- get_source_expressions("tmp.R", "a <- 42L")$expressions[[1L]]
  ref <- with_id(
    source_expression = source_expression,
    ids_with_token(source_expression = source_expression, value = "expr")
  )
  expect_identical(ref, source_expression$parsed_content[c(1L, 3L, 6L), ])
  expect_identical(ref$token, rep_len("expr", nrow(ref)))
})

test_that("it creates basic lints", {
  code <- "before   %+%   after"
  tmpfile <- withr::local_tempfile(lines = code)
  expr <- get_source_expressions(tmpfile)$expressions[[2L]]
  xml <- expr$full_xml_parsed_content
  node <- xml2::xml_find_first(xml, "//SPECIAL")

  l <- xml_nodes_to_lints(
    xml = node,
    source_expression = expr,
    lint_message = "lint_msg",
    type = "warning"
  )

  expect_s3_class(l, "lint")
  expect_identical(l$filename, tmpfile)
  expect_identical(l$line_number, 1L)
  expect_identical(l$column_number, as.integer(xml2::xml_attr(node, "col1")))
  expect_identical(l$type, "warning")
  expect_identical(l$message, "lint_msg")
  expect_identical(l$line, code)
  expect_identical(l$ranges, list(as.integer(c(xml2::xml_attr(node, "col1"), xml2::xml_attr(node, "col2")))))

  # location XPaths work
  l_before_col <- xml_nodes_to_lints(
    xml = node,
    source_expression = expr,
    lint_message = "lint_msg",
    column_number_xpath = "number(./preceding-sibling::*[1]/@col2 + 1)",
    range_start_xpath = "number(./preceding-sibling::*[1]/@col1)",
    range_end_xpath = "number(./following-sibling::*[1]/@col2)"
  )
  expect_identical(l_before_col$column_number, nchar("before") + 1L)
  expect_identical(l_before_col$ranges, list(c(1L, nchar(code))))

  # Vectorization works
  ll <- xml_nodes_to_lints(
    xml = xml2::xml_find_all(xml, "//expr"),
    source_expression = expr,
    lint_message = "lint_msg"
  )
  expect_s3_class(ll, "lints")
  expect_length(ll, 3L)
  expect_s3_class(ll[[1L]], "lint")

  # Also for plain lists of xml_nodes, usually obtained by c(nodeset_a, nodeset_b)
  ll_unclassed <- xml_nodes_to_lints(
    xml = unclass(xml2::xml_find_all(xml, "//expr")),
    source_expression = expr,
    lint_message = "lint_msg"
  )
  expect_s3_class(ll_unclassed, "lints")
  expect_length(ll_unclassed, 3L)
  expect_s3_class(ll_unclassed[[1L]], "lint")

  # lint_message as a vector
  ll_msgvec <- xml_nodes_to_lints(
    xml = xml2::xml_find_all(xml, "//SYMBOL"),
    source_expression = expr,
    lint_message = letters[1L:2L]
  )
  expect_identical(ll_msgvec[[1L]]$message, "a")
  expect_identical(ll_msgvec[[2L]]$message, "b")
})

test_that("it handles multi-line lints correctly", {
  code <- c("before %+%", "  after")
  tmpfile <- withr::local_tempfile(lines = code)
  expr <- get_source_expressions(tmpfile)$expressions[[2L]]
  xml <- expr$full_xml_parsed_content
  node <- xml2::xml_find_first(xml, "/exprlist/expr")

  l <- xml_nodes_to_lints(
    xml = node,
    source_expression = expr,
    lint_message = "lint_msg"
  )

  expect_s3_class(l, "lint")
  expect_identical(l$filename, tmpfile)
  expect_identical(l$line_number, 1L)
  expect_identical(l$column_number, as.integer(xml2::xml_attr(node, "col1")))
  expect_identical(l$type, "style")
  expect_identical(l$message, "lint_msg")
  expect_identical(l$line, code[[1L]])
  expect_identical(l$ranges, list(as.integer(c(xml2::xml_attr(node, "col1"), nchar(code[1L])))))
})

test_that("it doesn't produce invalid lints", {
  # Part of #1427
  code <- "before   %+%   after"
  tmpfile <- withr::local_tempfile(lines = code)
  expr <- get_source_expressions(tmpfile)$expressions[[2L]]
  xml <- expr$full_xml_parsed_content
  node <- xml2::xml_find_first(xml, "//SPECIAL")

  # We can produce invalid location xpaths by requiring any non-existent node
  xp_column_number <- "number(./preceding-sibling::*[1]/@col2 + 1)"
  xp_range_start <- "number(./preceding-sibling::*[1]/@col1)"
  xp_range_end <- "number(./following-sibling::*[1]/@col2)"
  xp_invalid <- "number(./DOES-NOT-EXIST/@col1)"

  expect_warning(
    {
      lints1 <- xml_nodes_to_lints(
        xml = node,
        source_expression = expr,
        lint_message = "lint_msg",
        column_number_xpath = xp_column_number,
        range_start_xpath = xp_invalid,
        range_end_xpath = xp_range_end
      )
    },
    "Defaulting to start of line",
    fixed = TRUE
  )

  expect_identical(lints1[["column_number"]], nchar("before") + 1L)
  expect_identical(lints1[["ranges"]], list(c(1L, nchar(code))))

  expect_warning(
    {
      lints2 <- xml_nodes_to_lints(
        xml = node,
        source_expression = expr,
        lint_message = "lint_msg",
        column_number_xpath = xp_column_number,
        range_start_xpath = xp_range_start,
        range_end_xpath = xp_invalid
      )
    },
    "Defaulting to width 1",
    fixed = TRUE
  )

  expect_identical(lints2[["column_number"]], nchar("before") + 1L)
  expect_identical(lints2[["ranges"]], list(c(1L, 1L)))

  expect_warning(
    {
      lints3 <- xml_nodes_to_lints(
        xml = node,
        source_expression = expr,
        lint_message = "lint_msg",
        column_number_xpath = xp_invalid,
        range_start_xpath = xp_range_start,
        range_end_xpath = xp_range_end
      )
    },
    "Defaulting to start of range",
    fixed = TRUE
  )

  expect_identical(lints3[["column_number"]], 1L)
  expect_identical(lints3[["ranges"]], list(c(1L, nchar(code))))
})

test_that("erroneous input errors", {
  expect_error(xml_nodes_to_lints(1L), "Expected an <xml_nodeset>", fixed = TRUE)
})

test_that("ids_with_token works as expected", {
  source_expression <- get_source_expressions("tmp.R", "a <- 42L")$expressions[[1L]]
  ref <- ids_with_token(source_expression = source_expression, value = "expr")
  expect_identical(ref, c(1L, 3L, 6L))
  expect_identical(source_expression$parsed_content$token[ref], rep_len("expr", length(ref)))
})

#' Lint common mistakes/style issues cropping up from return statements
#'
#' `return(x <- ...)` is either distracting (because `x` is ignored), or
#'   confusing (because assigning to `x` has some side effect that is muddled
#'   by the dual-purpose expression).
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "foo <- function(x) return(y <- x + 1)",
#'   linters = function_return_linter()
#' )
#'
#' lint(
#'   text = "foo <- function(x) return(x <<- x + 1)",
#'   linters = function_return_linter()
#' )
#'
#' writeLines("e <- new.env() \nfoo <- function(x) return(e$val <- x + 1)")
#' lint(
#'   text = "e <- new.env() \nfoo <- function(x) return(e$val <- x + 1)",
#'   linters = function_return_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "foo <- function(x) return(x + 1)",
#'   linters = function_return_linter()
#' )
#'
#' code_lines <- "
#' foo <- function(x) {
#'   x <<- x + 1
#'   return(x)
#' }
#' "
#' lint(
#'   text = code_lines,
#'   linters = function_return_linter()
#' )
#'
#' code_lines <- "
#' e <- new.env()
#' foo <- function(x) {
#'   e$val <- x + 1
#'   return(e$val)
#' }
#' "
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = function_return_linter()
#' )
#'
#' @evalRd rd_tags("function_return_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
function_return_linter <- make_linter_from_function_xpath(
  function_names = "return",
  xpath = "parent::expr/expr[LEFT_ASSIGN or RIGHT_ASSIGN]",
  lint_message = "Move the assignment outside of the return() clause, or skip assignment altogether."
)

#' Block assigning any variables whose name clashes with a `base` R function
#'
#' Re-using existing names creates a risk of subtle error best avoided.
#'   Avoiding this practice also encourages using better, more descriptive names.
#'
#' @param packages Character vector of packages to search for names that should
#'   be avoided. Defaults to the most common default packages: base, stats,
#'   utils, tools, methods, graphics, and grDevices.
#' @param allow_names Character vector of object names to ignore, i.e., which
#'   are allowed to collide with exports from `packages`.
#'
#' @examples
#' # will produce lints
#' code <- "function(x) {\n  data <- x\n  data\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = object_overwrite_linter()
#' )
#'
#' code <- "function(x) {\n  lint <- 'fun'\n  lint\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = object_overwrite_linter(packages = "lintr")
#' )
#'
#' # okay
#' code <- "function(x) {\n  data('mtcars')\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = object_overwrite_linter()
#' )
#'
#' code <- "function(x) {\n  data <- x\n  data\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = object_overwrite_linter(packages = "base")
#' )
#'
#' # names in function signatures are ignored
#' lint(
#'   text = "function(data) data <- subset(data, x > 0)",
#'   linters = object_overwrite_linter()
#' )
#'
#' @evalRd rd_tags("object_overwrite_linter")
#' @seealso
#'  - [linters] for a complete list of linters available in lintr.
#'  - <https://style.tidyverse.org/syntax.html#object-names>
#' @export
object_overwrite_linter <- function(
  packages = c("base", "stats", "utils", "tools", "methods", "graphics", "grDevices"),
  allow_names = character()
) {
  for (package in packages) {
    if (!requireNamespace(package, quietly = TRUE)) {
      cli_abort("Package {.pkg {package}} is required, but not available.")
    }
  }
  pkg_exports <- lapply(
    packages,
    # .__C__ etc.: drop 150+ "virtual" names since they are very unlikely to appear anyway
    \(pkg) setdiff(grep("^[.]__[A-Z]__", getNamespaceExports(pkg), value = TRUE, invert = TRUE), allow_names)
  )
  pkg_exports <- data.frame(
    package = rep(packages, lengths(pkg_exports)),
    name = unlist(pkg_exports)
  )

  # Take the first among duplicate names, e.g. 'plot' resolves to base::plot, not graphics::plot
  pkg_exports <- pkg_exports[!duplicated(pkg_exports$name), ]

  # test that the symbol doesn't match an argument name in the function
  # NB: data.table := has parse token LEFT_ASSIGN as well
  # ancestor::* for '=' assignment
  xpath_assignments <- glue("
    (//SYMBOL | //STR_CONST)[
      not(text() = ancestor::*/preceding-sibling::SYMBOL_FORMALS/text())
    ]/
      parent::expr[
        count(*) = 1
        and (
          following-sibling::LEFT_ASSIGN[text() != ':=']
          or following-sibling::EQ_ASSIGN
          or preceding-sibling::RIGHT_ASSIGN
        )
        and ancestor::*[
          (self::expr or self::expr_or_assign_or_help or self::equal_assign)
          and (preceding-sibling::FUNCTION or preceding-sibling::OP-LAMBDA)
        ]
      ]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    assigned_exprs <- xml_find_all_(xml, xpath_assignments)
    assigned_symbols <- get_r_string(assigned_exprs, "SYMBOL|STR_CONST")
    is_quoted <- startsWith(assigned_symbols, "`")
    assigned_symbols[is_quoted] <- substr(assigned_symbols[is_quoted], 2L, nchar(assigned_symbols[is_quoted]) - 1L)
    is_bad <- assigned_symbols %in% pkg_exports$name
    source_pkg <- pkg_exports$package[match(assigned_symbols[is_bad], pkg_exports$name)]
    lint_message <- sprintf(
      "'%s' is an exported object from package '%s'. Avoid re-using such symbols.",
      assigned_symbols[is_bad], source_pkg
    )

    xml_nodes_to_lints(
      assigned_exprs[is_bad],
      source_expression = source_expression,
      lint_message = lint_message,
      type = "warning"
    )
  })
}

#' Return linter
#'
#' This linter checks functions' [return()] expressions.
#'
#' @param return_style Character string naming the return style. `"implicit"`,
#'   the default, enforces the Tidyverse guide recommendation to leave terminal
#'   returns implicit. `"explicit"` style requires that `return()` always be
#'   explicitly supplied.
#' @param allow_implicit_else Logical, default `TRUE`. If `FALSE`, functions with a terminal
#'   `if` clause must always have an `else` clause, making the `NULL` alternative explicit
#'   if necessary. Similarly, functions with terminal [switch()] statements must have an
#'   explicit default case.
#' @param return_functions Character vector of functions that are accepted as terminal calls
#'   when `return_style = "explicit"`. These are in addition to exit functions
#'   from base that are always allowed: [stop()], [q()], [quit()], [invokeRestart()],
#'   `tryInvokeRestart()`, [UseMethod()], [NextMethod()], [standardGeneric()],
#'   [callNextMethod()], [.C()], [.Call()], [.External()], and [.Fortran()].
#' @param except,except_regex Character vector of functions that are not checked when
#'   `return_style = "explicit"`. These are in addition to namespace hook functions
#'   that are never checked: `.onLoad()`, `.onUnload()`, `.onAttach()`, `.onDetach()`,
#'   `.Last.lib()`, `.First()` and `.Last()`. `except` matches function names exactly,
#'   while `except_regex` does exclusion by pattern matching with [rex::re_matches()].
#'
#' @examples
#' # will produce lints
#' code <- "function(x) {\n  return(x + 1)\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = return_linter()
#' )
#'
#' code <- "function(x) {\n  x + 1\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = return_linter(return_style = "explicit")
#' )
#'
#' code <- "function(x) {\n  if (x > 0) 2\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = return_linter(allow_implicit_else = FALSE)
#' )
#'
#' # okay
#' code <- "function(x) {\n  x + 1\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = return_linter()
#' )
#'
#' code <- "function(x) {\n  return(x + 1)\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = return_linter(return_style = "explicit")
#' )
#'
#' code <- "function(x) {\n  if (x > 0) 2 else NULL\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = return_linter(allow_implicit_else = FALSE)
#' )
#'
#' @evalRd rd_tags("return_linter")
#' @seealso
#'  - [linters] for a complete list of linters available in lintr.
#'  - <https://style.tidyverse.org/functions.html?q=return#return>
#' @export
return_linter <- function(
  return_style = c("implicit", "explicit"),
  allow_implicit_else = TRUE,
  return_functions = NULL,
  except = NULL,
  except_regex = NULL
) {
  return_style <- match.arg(return_style)

  check_except <- !allow_implicit_else || return_style == "explicit"
  # We defer building the XPath strings in this case since we can't build the
  #   pattern-based "except" logic directly into the XPath (because of v1.0)
  defer_except <- check_except && !is.null(except_regex)

  if (check_except) {
    except_xpath_fmt <- "parent::expr[not(
      preceding-sibling::expr/SYMBOL[{ xp_text_in_table(except) }]
    )]"
    except <- union(special_funs, except)
    if (!defer_except) except_xpath <- glue(except_xpath_fmt, except = except)
  }

  if (return_style == "implicit") {
    # nolint next: object_usage. False positive.
    body_xpath <- "(//FUNCTION | //OP-LAMBDA)/following-sibling::expr[last()]"
    params <- list(
      implicit = TRUE,
      type = "style",
      lint_xpath = "SYMBOL_FUNCTION_CALL[text() = 'return']",
      lint_message = "Use implicit return behavior; explicit return() is not needed."
    )
  } else {
    base_return_functions <- c(
      # Normal calls
      "return", "stop", "q", "quit",
      "invokeRestart", "tryInvokeRestart",

      # Functions related to S3 methods
      "UseMethod", "NextMethod",

      # Functions related to S4 methods
      "standardGeneric", "callNextMethod",

      # Functions related to C interfaces
      ".C", ".Call", ".External", ".Fortran"
    )

    return_functions <- union(base_return_functions, return_functions)

    body_xpath_fmt <- "
    (//FUNCTION | //OP-LAMBDA)[{ except_xpath }]
      /following-sibling::expr[last()][OP-LEFT-BRACE and expr[last()]/@line1 != @line1]
      /expr[last()]
    "
    if (defer_except) {
      function_name_xpath <- "(//FUNCTION | //OP-LAMBDA)/parent::expr/preceding-sibling::expr/SYMBOL"
    } else {
      body_xpath <- glue(body_xpath_fmt, except_xpath = except_xpath)
    }

    params <- list(
      implicit = FALSE,
      type = "warning",
      lint_xpath = glue("self::*[not(
        (self::expr | following-sibling::SPECIAL[text() = '%>%']/following-sibling::expr/expr[1])
          /SYMBOL_FUNCTION_CALL[{ xp_text_in_table(return_functions) }]
      )]"),
      lint_message = "All functions must have an explicit return()."
    )
  }

  params$allow_implicit_else <- allow_implicit_else

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content
    if (defer_except) {
      assigned_functions <- xml_text(xml_find_all_(xml, function_name_xpath))
      except <-
        union(except, assigned_functions[re_matches_logical(assigned_functions, except_regex)])
      except_xpath <- glue(except_xpath_fmt, except = except)
      body_xpath <- glue(body_xpath_fmt, except_xpath = except_xpath)
    }

    body_expr <- xml_find_all_(xml, body_xpath)

    params$source_expression <- source_expression

    if (params$implicit && !params$allow_implicit_else) {
      # can't incorporate this into the body_xpath for implicit return style,
      #   since we still lint explicit returns for except= functions.
      allow_implicit_else <- is.na(xml_find_first_(body_expr, except_xpath))
    } else {
      allow_implicit_else <- rep(params$allow_implicit_else, length(body_expr))
    }
    # nested_return_lints not "vectorized" due to xml_children()
    Map(
      function(expr, allow_implicit_else) {
        params$allow_implicit_else <- allow_implicit_else
        nested_return_lints(expr, params)
      },
      body_expr, allow_implicit_else
    )
  })
}

nested_return_lints <- function(expr, params) {
  child_expr <- xml_children(expr)
  if (length(child_expr) == 0L) {
    return(list())
  }
  names(child_expr) <- xml_name_(child_expr)

  if (names(child_expr)[1L] == "OP-LEFT-BRACE") {
    brace_return_lints(child_expr, expr, params)
  } else if (names(child_expr)[1L] == "IF") {
    if_return_lints(child_expr, expr, params)
  } else if (!is.na(xml_find_first_(expr, "expr/SYMBOL_FUNCTION_CALL[text() = 'switch']"))) {
    switch_return_lints(child_expr, expr, params)
  } else {
    xml_nodes_to_lints(
      xml_find_first_(child_expr[[1L]], params$lint_xpath),
      source_expression = params$source_expression,
      lint_message = params$lint_message,
      type = params$type
    )
  }
}

brace_return_lints <- function(child_expr, expr, params) {
  expr_idx <- which(names(child_expr) %in% c("expr", "equal_assign", "expr_or_assign_or_help"))
  if (length(expr_idx) == 0L) { # empty brace expression {}
    if (params$implicit) {
      return(list())
    } else {
      return(list(xml_nodes_to_lints(
        expr,
        source_expression = params$source_expression,
        lint_message = params$lint_message,
        type = params$type
      )))
    }
  }
  nested_return_lints(child_expr[[tail(expr_idx, 1L)]], params)
}

if_return_lints <- function(child_expr, expr, params) {
  expr_idx <- which(names(child_expr) %in% c("expr", "equal_assign", "expr_or_assign_or_help"))
  return_lints <- lapply(child_expr[expr_idx[-1L]], nested_return_lints, params)
  if (params$allow_implicit_else || length(expr_idx) == 3L) {
    return(return_lints)
  }
  implicit_else_lints <- list(xml_nodes_to_lints(
    expr,
    source_expression = params$source_expression,
    lint_message = "All functions with terminal if statements must have a corresponding terminal else clause.",
    type = "warning"
  ))
  c(return_lints, implicit_else_lints)
}

switch_return_lints <- function(child_expr, expr, params) {
  # equal_assign/expr_or_assign_or_help not possible here
  expr_idx <- which(names(child_expr) == "expr")
  # switch(x, ...) | expr[1]: switch; expr[2]: x. Drop the first two, check usage in ...
  return_lints <- lapply(child_expr[tail(expr_idx, -2L)], nested_return_lints, params)
  # in addition to the two <expr> dropped above, a third unmatched <expr> would be the default case.
  if (params$allow_implicit_else || length(expr_idx) - sum(names(child_expr) == "EQ_SUB") == 3L) {
    return(return_lints)
  }
  implicit_else_lints <- list(xml_nodes_to_lints(
    expr,
    source_expression = params$source_expression,
    lint_message = "All functions with terminal switch statements must have a terminal default clause.",
    type = "warning"
  ))
  c(return_lints, implicit_else_lints)
}

#' Pipe consistency linter
#'
#' Check that the recommended pipe operator is used, or more conservatively that
#'   pipes are consistent by file.
#'
#' @param pipe Which pipe operator is valid (either `"%>%"` or `"|>"`). The default
#'   is the native pipe (`|>`). `"auto"` will instead
#'   only enforce consistency, i.e., that in any given file there is only one pipe.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "1:3 |> mean() %>% as.character()",
#'   linters = pipe_consistency_linter()
#' )
#'
#' lint(
#'   text = "1:3 %>% mean() %>% as.character()",
#'   linters = pipe_consistency_linter("|>")
#' )
#'
#' # okay
#' lint(
#'   text = "1:3 |> mean() |> as.character()",
#'   linters = pipe_consistency_linter()
#' )
#'
#' lint(
#'   text = "1:3 %>% mean() %>% as.character()",
#'   linters = pipe_consistency_linter("%>%")
#' )
#' @evalRd rd_tags("pipe_consistency_linter")
#' @seealso
#'  - [linters] for a complete list of linters available in lintr.
#'  - <https://style.tidyverse.org/pipes.html#magrittr>
#' @export
pipe_consistency_linter <- function(pipe = c("|>", "%>%", "auto")) {
  pipe <- match.arg(pipe)

  xpath_magrittr <- glue("//SPECIAL[{ xp_text_in_table(magrittr_pipes) }]")
  xpath_native <- "//PIPE"

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    match_magrittr <- xml_find_all_(xml, xpath_magrittr)
    match_native <- xml_find_all_(xml, xpath_native)

    n_magrittr <- length(match_magrittr)
    n_native <- length(match_native)

    if (pipe == "auto" && n_magrittr > 0L && n_native > 0L) {
      xml_nodes_to_lints(
        xml = c(match_magrittr, match_native),
        source_expression = source_expression,
        lint_message = paste(
          "Stick to one pipe operator; found", n_magrittr, "instances of %>% and", n_native, "instances of |>."
        ),
        type = "style"
      )
    } else if (pipe == "%>%" && n_native > 0L) {
      xml_nodes_to_lints(
        xml = match_native,
        source_expression = source_expression,
        lint_message = "Use the %>% pipe operator instead of the |> pipe operator.",
        type = "style"
      )
    } else if (pipe == "|>" && n_magrittr > 0L) {
      lint_message <- sprintf("Use the |> pipe operator instead of the %s pipe operator.", xml_text(match_magrittr))
      xml_nodes_to_lints(
        xml = match_magrittr,
        source_expression = source_expression,
        lint_message = lint_message,
        type = "style"
      )
    } else {
      list()
    }
  })
}

#' Convert an XML node or nodeset into a Lint
#'
#' Convenience function for converting nodes matched by XPath-based
#'   linter logic into a [Lint()] object to return.
#'
#' @details
#' The location XPaths, `column_number_xpath`, `range_start_xpath` and `range_end_xpath` are evaluated using
#' [xml2::xml_find_num()] and will usually be of the form `"number(./relative/xpath)"`.
#' Note that the location line number cannot be changed and lints spanning multiple lines will ignore `range_end_xpath`.
#' `column_number_xpath` and `range_start_xpath` are assumed to always refer to locations on the starting line of the
#' `xml` node.
#'
#' @inheritParams lint-s3
#' @param xml An `xml_node` object (to generate one `Lint`) or an
#'   `xml_nodeset` object (to generate several `Lint`s), e.g. as returned by
#'   [xml2::xml_find_all()] or [xml2::xml_find_first()] or a
#'   list of `xml_node` objects.
#' @param source_expression A source expression object, e.g. as
#'   returned typically by [lint()], or more generally
#'   by [get_source_expressions()].
#' @param lint_message The message to be included as the `message`
#'   to the `Lint` object. If `lint_message` is a character vector the same length as `xml`,
#'   the `i`-th lint will be given the `i`-th message.
#' @param column_number_xpath XPath expression to return the column number location of the lint.
#'   Defaults to the start of the range matched by `range_start_xpath`. See details for more information.
#' @param range_start_xpath XPath expression to return the range start location of the lint.
#'   Defaults to the start of the expression matched by `xml`. See details for more information.
#' @param range_end_xpath XPath expression to return the range end location of the lint.
#'   Defaults to the end of the expression matched by `xml`. See details for more information.
#'
#' @return For `xml_node`s, a `lint`. For `xml_nodeset`s, `lints` (a list of `lint`s).
#' @export
xml_nodes_to_lints <- function(xml, source_expression, lint_message,
                               type = c("style", "warning", "error"),
                               column_number_xpath = range_start_xpath,
                               range_start_xpath = "number(./@col1)",
                               range_end_xpath = "number(./@col2)") {
  if (length(xml) == 0L) {
    return(list())
  }
  if (is_nodeset_like(xml)) {
    lints <- .mapply(
      xml_nodes_to_lints,
      dots = list(xml = xml, lint_message = lint_message),
      MoreArgs = list(
        source_expression = source_expression,
        type = type,
        column_number_xpath = column_number_xpath,
        range_start_xpath = range_start_xpath,
        range_end_xpath = range_end_xpath
      )
    )
    class(lints) <- "lints"
    return(lints)
  } else if (!is_node(xml)) {
    cli_abort(c(
      x = "Expected an {.cls xml_nodeset}, a {.cls list} of xml_nodes, or an {.cls xml_node}.",
      i = "Instead got {.obj_type_friendly {xml}}."
    ))
  }
  type <- match.arg(type, c("style", "warning", "error"))
  line1 <- xml_attr_(xml, "line1")
  col1 <- xp_find_location(xml, range_start_xpath)
  if (is.na(col1)) {
    cli_warn(c(
      x = "Could not find range start for lint.",
      i = "Defaulting to start of line."
    ))
    col1 <- 1L
  }

  lines <- source_expression[["lines"]]
  if (is.null(lines)) lines <- source_expression[["file_lines"]]

  if (xml_attr_(xml, "line2") == line1) {
    col2 <- xp_find_location(xml, range_end_xpath)
    if (is.na(col2)) {
      cli_warn(c(
        x = "Could not find range end for lint.",
        i = "Defaulting to width 1."
      ))
      col2 <- col1
    }
  } else {
    col2 <- nchar(lines[[line1]])
  }

  column_number <- xp_find_location(xml, column_number_xpath)
  if (is.na(column_number)) {
    cli_warn(c(
      x = "Could not find location for lint.",
      i = "Defaulting to start of range."
    ))
    column_number <- col1
  }

  Lint(
    filename = source_expression$filename,
    line_number = as.integer(line1),
    column_number = column_number,
    type = type,
    message = lint_message,
    line = lines[[line1]],
    ranges = list(c(col1, col2))
  )
}

#' Require usage of `expect_length(x, n)` over `expect_equal(length(x), n)`
#'
#' [testthat::expect_length()] exists specifically for testing the [length()] of
#'   an object. [testthat::expect_equal()] can also be used for such tests,
#'   but it is better to use the tailored function instead.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "expect_equal(length(x), 2L)",
#'   linters = expect_length_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "expect_length(x, 2L)",
#'   linters = expect_length_linter()
#' )
#'
#' @evalRd rd_tags("expect_length_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
expect_length_linter <- function() {
  # TODO(#2465): also catch expect_true(length(x) == 1)
  xpath <- "
  following-sibling::expr[
    expr[1][SYMBOL_FUNCTION_CALL[text() = 'length']]
    and (position() = 1 or preceding-sibling::expr[NUM_CONST])
  ]
    /parent::expr[not(SYMBOL_SUB[text() = 'info' or contains(text(), 'label')])]
  "

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(c("expect_equal", "expect_identical"))
    bad_expr <- xml_find_all_(xml_calls, xpath)

    matched_function <- xp_call_name(bad_expr)
    lint_message <- sprintf("expect_length(x, n) is better than %s(length(x), n)", matched_function)
    xml_nodes_to_lints(bad_expr, source_expression, lint_message, type = "warning")
  })
}

#' Require usage of `expect_identical(x, y)` where appropriate
#'
#' This linter enforces the usage of [testthat::expect_identical()] as the
#'   default expectation for comparisons in a testthat suite. `expect_true(identical(x, y))`
#'   is an equivalent but unadvised method of the same test. Further,
#'   [testthat::expect_equal()] should only be used when `expect_identical()`
#'   is inappropriate, i.e., when `x` and `y` need only be *numerically
#'   equivalent* instead of fully identical (in which case, provide the
#'   `tolerance=` argument to `expect_equal()` explicitly). This also applies
#'   when it's inconvenient to check full equality (e.g., names can be ignored,
#'   in which case `ignore_attr = "names"` should be supplied to
#'   `expect_equal()` (or, for 2nd edition, `check.attributes = FALSE`).
#'
#' @section Exceptions:
#'
#' The linter allows `expect_equal()` in three circumstances:
#'   1. A named argument is set (e.g. `ignore_attr` or `tolerance`)
#'   2. Comparison is made to an explicit decimal, e.g.
#'      `expect_equal(x, 1.0)` (implicitly setting `tolerance`)
#'   3. `...` is passed (wrapper functions which might set
#'      arguments such as `ignore_attr` or `tolerance`)
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "expect_equal(x, y)",
#'   linters = expect_identical_linter()
#' )
#'
#' lint(
#'   text = "expect_true(identical(x, y))",
#'   linters = expect_identical_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "expect_identical(x, y)",
#'   linters = expect_identical_linter()
#' )
#'
#' lint(
#'   text = "expect_equal(x, y, check.attributes = FALSE)",
#'   linters = expect_identical_linter()
#' )
#'
#' lint(
#'   text = "expect_equal(x, y, tolerance = 1e-6)",
#'   linters = expect_identical_linter()
#' )
#'
#' @evalRd rd_tags("expect_identical_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
expect_identical_linter <- function() {
  # outline:
  #   - skip when any named argument is set. most commonly this
  #     is check.attributes (for 2e tests) or one of the ignore_*
  #     arguments (for 3e tests). This will generate some false
  #     negatives, but will be much easier to maintain.
  #   - skip cases like expect_equal(x, 1.02) or the constant vector version
  #     where a numeric constant indicates inexact testing is preferable
  #   - skip calls using dots (`...`); see tests
  non_integer <- glue("
    NUM_CONST[contains(text(), '.')]
    or (
      OP-MINUS
      and count(expr) = 1
      and expr[NUM_CONST[contains(text(), '.')]]
    )
  ")

  expect_equal_xpath <- glue("
  self::*[not(
      following-sibling::EQ_SUB
      or following-sibling::expr[
        (
          expr[1][SYMBOL_FUNCTION_CALL[text() = 'c']]
          and expr[{non_integer}]
        ) or (
          {non_integer}
        ) or (
          OP-MINUS
          and count(expr) = 1
          and expr[
            expr[1][SYMBOL_FUNCTION_CALL[text() = 'c']]
            and expr[{non_integer}]
          ]
        ) or (
          SYMBOL[text() = '...']
        )
      ]
    )]
    /parent::expr
  ")
  expect_true_xpath <- "
  following-sibling::expr[1][expr[1]/SYMBOL_FUNCTION_CALL[text() = 'identical']]
    /parent::expr
  "
  Linter(linter_level = "expression", function(source_expression) {
    expect_equal_calls <- source_expression$xml_find_function_calls("expect_equal")
    expect_true_calls <- source_expression$xml_find_function_calls("expect_true")
    bad_expr <- c(
      xml_find_all_(expect_equal_calls, expect_equal_xpath),
      xml_find_all_(expect_true_calls, expect_true_xpath)
    )

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste(
        "Use expect_identical(x, y) by default; resort to expect_equal() only when needed,",
        "e.g. when setting ignore_attr= or tolerance=."
      ),
      type = "warning"
    )
  })
}

#' Require usage of `expect_s4_class(x, k)` over `expect_true(is(x, k))`
#'
#' [testthat::expect_s4_class()] exists specifically for testing the class
#'   of S4 objects. [testthat::expect_true()] can also be used for such tests,
#'   but it is better to use the tailored function instead.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'expect_true(is(x, "Matrix"))',
#'   linters = expect_s4_class_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'expect_s4_class(x, "Matrix")',
#'   linters = expect_s4_class_linter()
#' )
#'
#' @evalRd rd_tags("expect_s4_class_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - [expect_s3_class_linter()]
#' @export
expect_s4_class_linter <- function() {
  # require 2 expressions because methods::is(x) alone is a valid call, even
  #   though the character output wouldn't make any sense for expect_true().
  xpath <- "
  following-sibling::expr[1][count(expr) = 3 and expr[1][SYMBOL_FUNCTION_CALL[text() = 'is']]]
    /parent::expr[not(SYMBOL_SUB[text() = 'info' or text() = 'label'])]
  "

  Linter(linter_level = "expression", function(source_expression) {
    # TODO(#2423): also catch expect_{equal,identical}(methods::is(x), k).
    #   this seems empirically rare, but didn't check many S4-heavy packages.

    xml_calls <- source_expression$xml_find_function_calls("expect_true")
    bad_expr <- xml_find_all_(xml_calls, xpath)

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste(
        "expect_s4_class(x, k) is better than expect_true(is(x, k)).",
        "Note also expect_s3_class() available for testing S3 objects."
      ),
      type = "warning"
    )
  })
}

#' Enforce usage of scalar logical operators in conditional statements
#'
#' Usage of `&` in conditional statements is error-prone and inefficient.
#'   `condition` in `if (condition) expr` must always be of length 1, in which
#'   case `&&` is to be preferred. Ditto for `|` vs. `||`.
#'
#' This linter covers inputs to `if()` and `while()` conditions and to
#'   [testthat::expect_true()] and [testthat::expect_false()].
#'
#' Note that because `&` and `|` are generics, it is possible that
#'   `&&` / `||` are not perfect substitutes because `&` is doing
#'   method dispatch in an incompatible way.
#'
#' Moreover, be wary of code that may have side effects, most commonly
#'   assignments. Consider `if ((a <- foo(x)) | (b <- bar(y))) { ... }`
#'   vs. `if ((a <- foo(x)) || (b <- bar(y))) { ... }`. Because `||` exits
#'   early, if `a` is `TRUE`,  the second condition will never be evaluated
#'   and `b` will not be assigned. Such usage is not allowed by the Tidyverse
#'   style guide, and the code can easily be refactored by pulling the
#'   assignment outside the condition, so using `||` is still preferable.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "if (TRUE & FALSE) 1",
#'   linters = vector_logic_linter()
#' )
#'
#' lint(
#'   text = "if (TRUE && (TRUE | FALSE)) 4",
#'   linters = vector_logic_linter()
#' )
#'
#' lint(
#'   text = "filter(x, A && B)",
#'   linters = vector_logic_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "if (TRUE && FALSE) 1",
#'   linters = vector_logic_linter()
#' )
#'
#' lint(
#'   text = "if (TRUE && (TRUE || FALSE)) 4",
#'   linters = vector_logic_linter()
#' )
#'
#' lint(
#'   text = "filter(x, A & B)",
#'   linters = vector_logic_linter()
#' )
#'
#' @evalRd rd_tags("vector_logic_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#if-statements>
#' @export
vector_logic_linter <- function() {
  # ensures the expr is in the cond part of `if/while (cond) expr` --
  #   if on the XML parse tree is structured like
  #   <expr>
  #     <IF> | <expr><SYMBOL_FUNCTION_CALL>
  #     <OP-LEFT-PAREN>
  #     <expr> ... </expr> # <- loop condition
  #     <OP-RIGHT-PAREN>
  #     <expr> ... </expr> # <- evaluation; includes BRACEs if present
  #     <ELSE>             # (here & below is optional)
  #     <expr> ... </expr>
  #  </expr>
  #  we _don't_ want to match anything on the second expr.
  #  not(call[not(lambda)]): skip unless _every_ call ancestor has a lambda
  condition_xpath <- "
  (//AND | //OR)[
    ancestor::expr[
      not(preceding-sibling::OP-RIGHT-PAREN)
      and preceding-sibling::*[
        self::IF
        or self::WHILE
        or self::expr/SYMBOL_FUNCTION_CALL[text() = 'expect_true' or text() = 'expect_false']
      ]
    ]
    and not(ancestor::expr[
      (
        preceding-sibling::expr[last()][SYMBOL_FUNCTION_CALL[not(text() = 'expect_true' or text() = 'expect_false')]]
        or preceding-sibling::OP-LEFT-BRACKET
      )
      and not(descendant-or-self::expr[FUNCTION or OP-LAMBDA])
    ])
    and not(parent::expr/expr[
      STR_CONST
      or expr/SYMBOL_FUNCTION_CALL[text() = 'as.raw' or text() = 'as.octmode' or text() = 'as.hexmode']
    ])
  ]
  "

  subset_xpath <- "
  self::*[not(SYMBOL_PACKAGE[text() = 'stats'])]
    /parent::expr
    //expr[
      (AND2 or OR2)
      and not(preceding-sibling::expr[last()]/SYMBOL_FUNCTION_CALL[not(text() = 'subset' or text() = 'filter')])
      and not(preceding-sibling::OP-LEFT-BRACKET)
      and not(preceding-sibling::*[not(self::COMMENT)][2][self::SYMBOL_SUB and text() = 'circular'])
      and not(ancestor::expr[FUNCTION or OP-LAMBDA])
    ]
    /*[not(self::COMMENT)][2]
  "

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content
    xml_call <- source_expression$xml_find_function_calls(c("subset", "filter"))

    condition_expr <- xml_find_all_(xml, condition_xpath)
    condition_op <- xml_text(condition_expr)
    condition_lints <- xml_nodes_to_lints(
      condition_expr,
      source_expression = source_expression,
      lint_message = sprintf("Use `%s` in conditional expressions.", strrep(condition_op, 2L)),
      type = "warning"
    )

    subset_expr <- xml_find_all_(xml_call, subset_xpath)
    subset_op <- xml_text(subset_expr)
    subset_lints <- xml_nodes_to_lints(
      subset_expr,
      source_expression = source_expression,
      lint_message = sprintf("Use `%s` in subsetting expressions.", substr(subset_op, 1L, 1L)),
      type = "warning"
    )

    c(condition_lints, subset_lints)
  })
}

#' Check for a common mistake where a size check like 'length' is applied in the wrong place
#'
#' Usage like `length(x == 0)` is a mistake. If you intended to check `x` is empty,
#'   use `length(x) == 0`. Other mistakes are possible, but running `length()` on the
#'   outcome of a logical comparison is never the best choice.
#'
#' The linter also checks for similar usage with `nrow()`, `ncol()`, `NROW()`, and `NCOL()`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "length(x == 0)",
#'   linters = length_test_linter()
#' )
#'
#' lint(
#'   text = "nrow(x > 0) || ncol(x > 0)",
#'   linters = length_test_linter()
#' )
#'
#' lint(
#'   text = "NROW(x == 1) && NCOL(y == 1)",
#'   linters = length_test_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "length(x) > 0",
#'   linters = length_test_linter()
#' )
#'
#' lint(
#'   text = "nrow(x) > 0 || ncol(x) > 0",
#'   linters = length_test_linter()
#' )
#'
#' lint(
#'   text = "NROW(x) == 1 && NCOL(y) == 1",
#'   linters = length_test_linter()
#' )
#'
#' @evalRd rd_tags("class_equals_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
length_test_linter <- function() {
  xpath <- glue("
  following-sibling::expr[{ xp_or(infix_metadata$xml_tag[infix_metadata$comparator]) }]
    /parent::expr
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(c("length", "nrow", "ncol", "NROW", "NCOL"))
    bad_expr <- xml_find_all_(xml_calls, xpath)
    bad_expr <- strip_comments_from_subtree(bad_expr)

    matched_function <- xp_call_name(bad_expr)
    expr_parts <- vapply(
      lapply(bad_expr, xml_find_all_, "expr[2]/*[not(self::COMMENT)]"),
      xml_text,
      character(3L)
    )
    lint_message <- sprintf(
      "Checking the %s of a logical vector is likely a mistake. Did you mean `%s(%s) %s %s`?",
      matched_function, matched_function, expr_parts[1L, ], expr_parts[2L, ], expr_parts[3L, ]
    )

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = lint_message,
      type = "warning"
    )
  })
}

#' Block unnecessary quoting in calls
#'
#' Any valid symbol can be used as a keyword argument to an R function call.
#'   Sometimes, it is necessary to quote (or backtick) an argument that is
#'   not an otherwise valid symbol (e.g. creating a vector whose names have
#'   spaces); besides this edge case, quoting should not be done.
#'
#' The most common source of violation for this is creating named vectors,
#'   lists, or data.frame-alikes, but it can be observed in other calls as well.
#'
#' Similar reasoning applies to extractions with `$` or `@`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'data.frame("a" = 1)',
#'   linters = keyword_quote_linter()
#' )
#'
#' lint(
#'   text = "data.frame(`a` = 1)",
#'   linters = keyword_quote_linter()
#' )
#'
#' lint(
#'   text = 'my_list$"key"',
#'   linters = keyword_quote_linter()
#' )
#'
#' lint(
#'   text = 's4obj@"key"',
#'   linters = keyword_quote_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "data.frame(`a b` = 1)",
#'   linters = keyword_quote_linter()
#' )
#'
#' lint(
#'   text = "my_list$`a b`",
#'   linters = keyword_quote_linter()
#' )
#'
#' @evalRd rd_tags("keyword_quote_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
keyword_quote_linter <- function() {
  # Check if a string could be assigned as an R variable.
  #
  # See [make.names()] for the description of syntactically valid names in R.
  is_valid_r_name <- function(x) make.names(x) == x

  # NB: xml2 uses xpath 1.0 which doesn't support matches() for regex, so we
  #   have to jump out of xpath to complete this lint.
  # It's also a bit tough to get the escaping through R and then xpath to
  #   work as intended, hence the rather verbose declaration here.
  quote_cond <- xp_or(
    "starts-with(text(), '\"')",
    "starts-with(text(), '`')",
    'starts-with(text(), "\'")'
  )
  # SYMBOL_SUB for backticks, STR_CONST for quoted names
  call_arg_xpath <- glue("
  parent::expr
    /*[(self::SYMBOL_SUB or self::STR_CONST) and {quote_cond}]
  ")

  # also exclude $ or @, which are handled below
  assignment_candidate_cond <- "
    not(OP-DOLLAR or OP-AT)
    and (STR_CONST or SYMBOL[starts-with(text(), '`')])
  "
  assignment_xpath <- glue("
  (//EQ_ASSIGN | //LEFT_ASSIGN[text() != ':='])
    /preceding-sibling::expr[{ assignment_candidate_cond }]
  | //RIGHT_ASSIGN/following-sibling::expr[{ assignment_candidate_cond }]
  ")

  extraction_xpath <- "
    (//OP-DOLLAR | //OP-AT)/following-sibling::STR_CONST
    | //OP-DOLLAR/following-sibling::SYMBOL[starts-with(text(), '`')]
    | //OP-AT/following-sibling::SLOT[starts-with(text(), '`')]
  "

  no_quote_msg <- "Use backticks to create non-syntactic names, not quotes."
  clarification <- "i.e., if the name is not a valid R symbol (see ?make.names)."

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content
    xml_calls <- source_expression$xml_find_function_calls(NULL)

    call_arg_expr <- xml_find_all_(xml_calls, call_arg_xpath)

    invalid_call_quoting <- is_valid_r_name(get_r_string(call_arg_expr))

    call_arg_lints <- xml_nodes_to_lints(
      call_arg_expr[invalid_call_quoting],
      source_expression = source_expression,
      lint_message = paste("Only quote named arguments to functions if necessary,", clarification),
      type = "warning"
    )

    assignment_expr <- xml_find_all_(xml, assignment_xpath)

    invalid_assignment_quoting <- is_valid_r_name(get_r_string(assignment_expr))
    # NB: XPath is such that there is exactly 1 node per match, making xml_children() ideal.
    #   xml_child() gets it wrong for 0 (an error) and >1 match.
    assignment_to_string <- xml_name_(xml_children(assignment_expr)) == "STR_CONST"

    string_assignment_lints <- xml_nodes_to_lints(
      assignment_expr[assignment_to_string & !invalid_assignment_quoting],
      source_expression = source_expression,
      lint_message = no_quote_msg,
      type = "warning"
    )

    assignment_lints <- xml_nodes_to_lints(
      assignment_expr[invalid_assignment_quoting],
      source_expression = source_expression,
      lint_message = paste("Only quote targets of assignment if necessary,", clarification),
      type = "warning"
    )

    extraction_expr <- xml_find_all_(xml, extraction_xpath)

    invalid_extraction_quoting <- is_valid_r_name(get_r_string(extraction_expr))
    extraction_of_string <- xml_name_(extraction_expr) == "STR_CONST"

    string_extraction_lints <- xml_nodes_to_lints(
      extraction_expr[extraction_of_string & !invalid_extraction_quoting],
      source_expression = source_expression,
      lint_message = no_quote_msg,
      type = "warning"
    )

    extraction_expr <- extraction_expr[invalid_extraction_quoting]
    extractor <- xml_find_chr_(extraction_expr, "string(preceding-sibling::*[not(self::COMMENT)][1])")
    gen_extractor <- ifelse(extractor == "$", "[[", "slot()")

    extraction_lints <- xml_nodes_to_lints(
      extraction_expr,
      source_expression = source_expression,
      lint_message = paste(
        "Only quote targets of extraction with", extractor, "if necessary,", clarification,
        "Use backticks to create non-syntactic names, or use", gen_extractor, "to extract by string."
      ),
      type = "warning"
    )

    c(call_arg_lints, string_assignment_lints, assignment_lints, string_extraction_lints, extraction_lints)
  })
}

#' Default linters
#'
#' @description List of default linters for [lint()]. Use
#' [linters_with_defaults()] to customize it. Most of the default linters
#' are based on [the tidyverse style guide](https://style.tidyverse.org/).
#'
#' The set of default linters is as follows (any parameterized linters, e.g., `line_length_linter` use their default
#' argument(s), see `?<linter_name>` for details):
#'
#' @evalRd rd_linters("default")
#' @seealso [linters] for a complete list of linters available in lintr.
#'
#' @export
default_linters <- modify_defaults(
  defaults = list(),
  assignment_linter(),
  brace_linter(),
  commas_linter(),
  commented_code_linter(),
  equals_na_linter(),
  function_left_parentheses_linter(),
  indentation_linter(),
  infix_spaces_linter(),
  line_length_linter(),
  object_length_linter(),
  object_name_linter(),
  object_usage_linter(),
  paren_body_linter(),
  pipe_consistency_linter(),
  pipe_continuation_linter(),
  quotes_linter(),
  return_linter(),
  semicolon_linter(),
  seq_linter(),
  spaces_inside_linter(),
  spaces_left_parentheses_linter(),
  T_and_F_symbol_linter(),
  trailing_blank_lines_linter(),
  trailing_whitespace_linter(),
  vector_logic_linter(),
  whitespace_linter()
)

#' Default undesirable functions and operators
#'
#' Lists of function names and operators for [undesirable_function_linter()] and [undesirable_operator_linter()].
#' There is a list for the default elements and another that contains all available elements.
#' Use [modify_defaults()] to produce a custom list.
#'
#' @evalRd c(
#'   "\\details{",
#'   rd_undesirable_functions(),
#'   "",
#'   rd_undesirable_operators(),
#'   "}"
#' )
#'
#' @format A named list of character strings.
#' @rdname default_undesirable_functions
#' @export
all_undesirable_functions <- modify_defaults(
  defaults = list(),
  attach =
    "use roxygen2's @importFrom statement in packages, or `::` in scripts. attach() modifies the global search path",
  browser =
    "remove this likely leftover from debugging. It pauses execution when run",
  debug = paste(
    "remove this likely leftover from debugging.",
    "It traps a function and causes execution to pause when that function is run"
  ),
  debugcall = paste(
    "remove this likely leftover from debugging.",
    "It traps a function and causes execution to pause when that function is run"
  ),
  debugonce = paste(
    "remove this likely leftover from debugging.",
    "It traps a function and causes execution to pause when that function is run"
  ),
  detach = paste(
    "avoid modifying the global search path.",
    "Detaching environments from the search path is rarely necessary in production code"
  ),
  ifelse = paste(
    "use an `if`/`else` block for scalar logic,",
    "or use dplyr::if_else()/data.table::fifelse() for type-stable vectorized logic"
  ),
  .libPaths = paste(
    "use withr::with_libpaths() for a temporary change",
    "instead of permanently modifying the library location"
  ),
  library = paste(
    "use roxygen2's @importFrom statement in packages and `::` in scripts,",
    "instead of modifying the global search path"
  ),
  loadNamespace =
    "use the return value of requireNamespace() instead to provide an easy way to signal failures",
  mapply =
    "use Map() to guarantee a list is returned and simplify accordingly",
  options =
    "use withr::with_options() for a temporary change instead of permanently modifying the session options",
  par =
    "use withr::with_par() for a temporary change instead of permanently modifying the graphics device parameters",
  require = paste(
    "use roxygen2's @importFrom statement in packages and library() or `::` in scripts,",
    "instead of modifying the global search path"
  ),
  sapply =
    "use vapply() with an appropriate `FUN.VALUE=` argument to obtain type-stable simplification",
  setwd =
    "use withr::with_dir() for a temporary change instead of modifying the global working directory",
  sink =
    "use withr::with_sink() for a temporary redirection instead of permanently redirecting output",
  source = paste(
    "manage dependencies through packages.",
    "source() loads code into the global environment unless `local = TRUE` is used,",
    "which can cause hard-to-predict behavior"
  ),
  structure =
    "Use `class<-`, `names<-`, and `attr<-` to set attributes",
  substring =
    "use substr() with appropriate `stop=` value.",
  Sys.setenv =
    "use withr::with_envvar() for a temporary change instead of permanently modifying global environment variables",
  Sys.setlocale =
    "use withr::with_locale() for a temporary change instead of permanently modifying the session locale",
  trace = paste(
    "remove this likely leftover from debugging.",
    "It traps a function and causes execution of arbitrary code when that function is run"
  ),
  undebug = paste(
    "remove this likely leftover from debugging.",
    "It is only useful for interactive debugging with debug()"
  ),
  untrace = paste(
    "remove this likely leftover from debugging.",
    "It is only useful for interactive debugging with trace()"
  )
)

#' @rdname default_undesirable_functions
#' @format NULL
#' @export
default_undesirable_functions <- all_undesirable_functions[names(all_undesirable_functions) %in% c(
  "attach",
  "browser",
  "debug",
  "debugcall",
  "debugonce",
  "detach",
  ".libPaths",
  "library",
  "mapply",
  "options",
  "par",
  "require",
  "sapply",
  "setwd",
  "sink",
  "source",
  "structure",
  "Sys.setenv",
  "Sys.setlocale",
  "Sys.unsetenv",
  "trace",
  "undebug",
  "untrace",
  NULL
)]

# nocov start
rd_auto_link <- function(x) {
  x <- unlist(x)
  x <- gsub(R"{([a-zA-Z0-9.]+)::([a-zA-Z0-9._]+)\(\)}", R"(\\code{\\link[\1:\2]{\1::\2()}})", x)
  x <- gsub(R"{([^:a-zA-Z0-9._])([a-zA-Z0-9._]+)\(\)}", R"(\1\\code{\\link[=\2]{\2()}})", x)
  x <- gsub("`([^`]+)`", R"(\\code{\1})", x)
  x
}

rd_undesirable_functions <- function() {
  alternatives <- rd_auto_link(default_undesirable_functions)

  c(
    "The following functions are sometimes regarded as undesirable:",
    "\\itemize{",
    sprintf(
      R"(\item \code{\link[=%1$s]{%1$s()}} As an alternative, %2$s.)",
      names(default_undesirable_functions), alternatives
    ),
    "}"
  )
}
# nocov end

#' @rdname default_undesirable_functions
#' @format NULL
#' @export
all_undesirable_operators <- modify_defaults(
  defaults = list(),
  ":::" = paste(
    "It accesses non-exported functions inside packages. Code relying on these is likely to break in",
    "future versions of the package because the functions are not part of the public interface and may be",
    "changed or removed by the maintainers without notice. Use public functions via `::` instead."
  ),
  "<<-" = paste(
    "It assigns outside the current environment in a way that can be hard to reason about.",
    "Prefer fully-encapsulated functions wherever possible, or, if necessary, assign to a specific",
    "environment with assign(). Recall that you can create an environment at the desired scope with",
    "new.env()."
  ),
  "->>" = paste(
    "It assigns outside the current environment in a way that can be hard to reason about.",
    "Prefer fully-encapsulated functions wherever possible, or, if necessary, assign to a specific",
    "environment with assign(). Recall that you can create an environment at the desired scope with",
    "new.env()."
  )
)

#' @rdname default_undesirable_functions
#' @format NULL
#' @export
default_undesirable_operators <- all_undesirable_operators[names(all_undesirable_operators) %in% c(
  ":::",
  "<<-",
  "->>",
  NULL
)]

# nocov start
rd_undesirable_operators <- function() {
  op_link_map <- c(
    `:::` = "\\link[base:ns-dblcolon]{:::}",
    `<<-` = "\\link[base:assignOps]{<<-}",
    `->>` = "\\link[base:assignOps]{->>}"
  )
  op <- names(default_undesirable_operators)

  alternatives <- rd_auto_link(default_undesirable_operators)

  c(
    "The following operators are sometimes regarded as undesirable:",
    "\\itemize{",
    sprintf(
      "\\item \\code{%1$s}. %2$s",
      op_link_map[op], alternatives
    ),
    "}"
  )
}
# nocov end

#' Default lintr settings
#'
#' @description
#' The default settings consist of
#'
#'  - `linters`: a list of default linters (see [default_linters()])
#'  - `encoding`: the character encoding assumed for the file
#'  - `exclude`: pattern used to exclude a line of code
#'  - `exclude_start`, `exclude_end`: patterns used to mark start and end of the code block to exclude
#'  - `exclude_linter`, `exclude_linter_sep`: patterns used to exclude linters
#'  - `exclusions`: a list of exclusions, see [exclude()] for a complete description of valid values.
#'  - `cache_directory`: location of cache directory
#'  - `error_on_lint`: decides if error should be produced when any lints are found
#'
#' There are no settings without defaults, i.e., this list describes every valid setting.
#'
#' @examples
#' # available settings
#' names(default_settings)
#'
#' # linters included by default
#' names(default_settings$linters)
#'
#' # default values for a few of the other settings
#' default_settings[c(
#'   "encoding",
#'   "exclude",
#'   "exclude_start",
#'   "exclude_end",
#'   "exclude_linter",
#'   "exclude_linter_sep",
#'   "exclusions",
#'   "error_on_lint"
#' )]
#'
#' @seealso [read_settings()], [default_linters]
#' @aliases settings config lintr-config lintr-settings .lintr
#' @export
default_settings <- NULL

settings <- new.env(parent = emptyenv())

# nocov start
logical_env <- function(x, unset = "") {
  res <- as.logical(Sys.getenv(x, unset = unset))
  if (is.na(res)) {
    return(NULL)
  }
  res
}

.onLoad <- function(libname, pkgname) {
  op <- options()
  op_lintr <- list(
    lintr.linter_file = Sys.getenv("R_LINTR_LINTER_FILE", ".lintr")
  )
  toset <- !(names(op_lintr) %in% names(op))
  if (any(toset)) options(op_lintr[toset])

  utils::assignInMyNamespace("default_settings", list(
    linters = default_linters,
    encoding = "UTF-8",
    exclude = rex("#", any_spaces, "nolint"),
    exclude_next = rex("#", any_spaces, "nolint next"),
    exclude_start = rex("#", any_spaces, "nolint start"),
    exclude_end = rex("#", any_spaces, "nolint end"),
    exclude_linter = rex(
      start, any_spaces, ":", any_spaces,
      capture(
        name = "linters",
        zero_or_more(one_or_more(none_of(",.")), any_spaces, ",", any_spaces),
        one_or_more(none_of(",."))
      ), "."
    ),
    exclude_linter_sep = rex(any_spaces, ",", any_spaces),
    exclusions = list(),
    cache_directory = R_user_dir("lintr", "cache"),
    error_on_lint = logical_env("LINTR_ERROR_ON_LINT", unset = FALSE)
  ))

  reset_settings()
}
# nocov end

#' Undesirable operator linter
#'
#' Report the use of undesirable operators, e.g. \code{\link[base:ns-dblcolon]{:::}} or
#' [`<<-`][base::assignOps] and suggest an alternative.
#'
#' @param op Character vector of undesirable operators. Input can be any of three types:
#'   - Unnamed entries must be a character string specifying an undesirable operator.
#'   - For named entries, the name specifies the undesirable operator.
#'     + If the entry is a character string, it is used as a description of
#'       why a given operator is undesirable
#'     + Otherwise, entries should be missing (`NA`)
#'   A generic message that the named operator is undesirable is used if no
#'     specific description is provided.
#'   Input can also be a list of character strings for convenience.
#'
#'   Defaults to [default_undesirable_operators]. To make small customizations to this list,
#'   use [modify_defaults()].
#' @param call_is_undesirable Logical, default `TRUE`. Should lints also be produced
#'   for prefix-style usage of the operators provided in `op`?
#'
#' @examples
#' # defaults for which functions are considered undesirable
#' names(default_undesirable_operators)
#'
#' # will produce lints
#' lint(
#'   text = "a <<- log(10)",
#'   linters = undesirable_operator_linter()
#' )
#'
#' lint(
#'   text = "mtcars$wt",
#'   linters = undesirable_operator_linter(op = c("$" = "As an alternative, use the `[[` accessor."))
#' )
#'
#' lint(
#'   text = "`:::`(utils, hasName)",
#'   linters = undesirable_operator_linter()
#' )
#'
#' lint(
#'   text = "mtcars$wt",
#'   linters = undesirable_operator_linter("$")
#' )
#'
#' # okay
#' lint(
#'   text = "a <- log(10)",
#'   linters = undesirable_operator_linter()
#' )
#' lint(
#'   text = 'mtcars[["wt"]]',
#'   linters = undesirable_operator_linter(op = c("$" = NA))
#' )
#'
#' lint(
#'   text = 'mtcars[["wt"]]',
#'   linters = undesirable_operator_linter(op = c("$" = "As an alternative, use the `[[` accessor."))
#' )
#'
#' lint(
#'   text = "`:::`(utils, hasName)",
#'   linters = undesirable_operator_linter(call_is_undesirable = FALSE)
#' )
#'
#' lint(
#'   text = 'mtcars[["wt"]]',
#'   linters = undesirable_operator_linter("$")
#' )
#'
#' @evalRd rd_tags("undesirable_operator_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
undesirable_operator_linter <- function(op = default_undesirable_operators,
                                        call_is_undesirable = TRUE) {
  if (is.list(op)) op <- unlist(op)
  if (!is.logical(call_is_undesirable)) {
    cli_abort("{.arg call_is_undesirable} must be a logical, not {.obj_type_friendly {call_is_undesirable}}.")
  }
  # allow (uncoerced->implicitly logical) 'NA'
  if (length(op) == 0L || !(is.character(op) || all(is.na(op)))) {
    cli_abort("{.arg op} must be a non-empty character vector.")
  }

  nm <- names2(op)
  implicit_idx <- !nzchar(nm)
  if (any(implicit_idx)) {
    names(op)[implicit_idx] <- op[implicit_idx]
    is.na(op) <- implicit_idx
  }
  if (anyNA(names(op))) {
    missing_idx <- which(is.na(names(op))) # nolint: object_usage_linter. False positive.
    cli_abort(paste(
      "Unnamed elements of {.arg op} must not be missing,",
      "but {.val {missing_idx}} {qty(length(missing_idx))} {?is/are}."
    ))
  }

  # infix must be handled individually below; non-assignment `=` are always OK
  operator_nodes <- infix_metadata$xml_tag_exact[
    infix_metadata$string_value %in% setdiff(names(op), "%%") &
      !infix_metadata$xml_tag %in% c("EQ_SUB", "EQ_FORMALS")
  ]

  is_infix <- startsWith(names(op), "%")
  if (any(is_infix)) {
    operator_nodes <- c(operator_nodes, sprintf("SPECIAL[text() = '%s']", names(op)[is_infix]))
  }

  if (length(operator_nodes) == 0L) {
    cli_abort("Did not recognize any valid operators in request for: {.str {names(op)}}")
  }

  infix_xpath <- paste(paste0("//", operator_nodes), collapse = " | ")

  quoted_op <- paste0("`", names(op), "`")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    infix_expr <- xml_find_all_(xml, infix_xpath)

    all_expr <- infix_expr
    operator <- xml_text(infix_expr)
    if (call_is_undesirable) {
      op_calls <- source_expression$xml_find_function_calls(quoted_op)
      all_expr <- combine_nodesets(all_expr, op_calls)
      operator <- c(operator, gsub("^`|`$", "", xml_text(op_calls)))
    }

    lint_message <- sprintf("Avoid undesirable operator `%s`.", operator)
    alternative <- op[operator]
    has_alternative <- !is.na(alternative)
    lint_message[has_alternative] <- paste(lint_message[has_alternative], alternative[has_alternative])

    xml_nodes_to_lints(
      all_expr,
      source_expression,
      lint_message,
      type = "warning"
    )
  })
}

#' Absolute path linter
#'
#' Check that no absolute paths are used (e.g. "/var", "C:\\System", "~/docs").
#'
#' @param lax Less stringent linting, leading to fewer false positives.
#' If `TRUE`, only lint path strings, which
#'
#' * contain at least two path elements, with one having at least two characters and
#' * contain only alphanumeric chars (including UTF-8), spaces, and win32-allowed punctuation
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'R"(/blah/file.txt)"',
#'   linters = absolute_path_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'R"(./blah)"',
#'   linters = absolute_path_linter()
#' )
#'
#' @evalRd rd_tags("absolute_path_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - [nonportable_path_linter()]
#' @export
absolute_path_linter <- function(lax = TRUE) {
  path_linter_factory(
    path_function = function(path) {
      is_absolute_path(path) && is_valid_long_path(path, lax)
    },
    message = "Do not use absolute paths."
  )
}

#' Modify lintr defaults
#'
#' Modify a list of defaults by name, allowing for replacement, deletion and addition of new elements.
#'
#' @param ... arguments of elements to change. If unnamed, the argument is automatically named.
#' If the named argument already exists in `defaults`, it is replaced by the new element.
#' If it does not exist, it is added. If the value is `NULL`, the element is removed.
#' @param defaults named list of elements to modify.
#' @return A modified list of elements, sorted by name. To achieve this sort in a platform-independent way, two
#'   transformations are applied to the names: (1) replace `_` with `0` and (2) convert [tolower()].
#'
#' @seealso
#' - [linters_with_defaults] for basing off lintr's set of default linters.
#' - [all_linters] for basing off all available linters in lintr.
#' - [linters_with_tags] for basing off tags attached to linters, possibly across multiple packages.
#' - [available_linters] to get a data frame of available linters.
#' - [linters] for a complete list of linters available in lintr.
#'
#' @examples
#' # custom list of undesirable functions:
#' #    remove `sapply` (using `NULL`)
#' #    add `cat` (with an accompanying message),
#' #    add `print` (unnamed, i.e. with no accompanying message)
#' #    add `source` (as taken from `all_undesirable_functions`)
#' my_undesirable_functions <- modify_defaults(
#'   defaults = default_undesirable_functions,
#'   sapply = NULL, "cat" = "No cat allowed", "print", all_undesirable_functions[["source"]]
#' )
#'
#' # list names of functions specified as undesirable
#' names(my_undesirable_functions)
#' @export
modify_defaults <- function(defaults, ...) {
  if (missing(defaults)) {
    cli_abort("{.arg defaults} is a required argument, but is missing.")
  }
  if (!(is.list(defaults) || is.environment(defaults)) || !all(nzchar(names2(defaults)))) {
    cli_abort("{.arg defaults} must be a named list or environment, not {.obj_type_friendly {defaults}}.")
  }
  vals <- list(...)
  nms <- names2(vals)
  missing_index <- !nzchar(nms, keepNA = TRUE)
  if (any(missing_index)) {
    nms[missing_index] <- guess_names(..., missing_index = missing_index)
  }

  to_null <- vapply(vals, is.null, logical(1L))
  if (!all(nms[to_null] %in% names(defaults))) {
    bad_nms <- setdiff(nms[to_null], names(defaults)) # nolint: object_usage_linter. TODO(#2252).
    cli_warn(c(
      i = "Trying to remove {.field {bad_nms}}, which {?is/are} not in {.arg defaults}."
    ))
  }

  is.na(vals) <- nms == vals
  for (ii in seq_along(nms)) defaults[[nms[ii]]] <- vals[[ii]]
  if (is.environment(defaults)) defaults <- as.list(defaults)

  res <- defaults[!vapply(defaults, is.null, logical(1L))]
  res <- res[platform_independent_order(names(res))]
  res
}

#' Create a tag-based linter configuration
#'
#' Make a new list based on all linters provided by `packages` and tagged with `tags`.
#' The result of this function is meant to be passed to the `linters` argument of `lint()`,
#' or to be put in your configuration file.
#'
#' @param ... Arguments of elements to change. If unnamed, the argument is automatically named.
#' If the named argument already exists in the list of linters, it is replaced by the new element.
#' If it does not exist, it is added. If the value is `NULL`, the linter is removed.
#' @inheritParams available_linters
#'
#' @return A modified list of linters.
#' @seealso
#' - [linters_with_defaults] for basing off lintr's set of default linters.
#' - [all_linters] for basing off all available linters in lintr.
#' - [available_linters] to get a data frame of available linters.
#' - [linters] for a complete list of linters available in lintr.
#'
#' @examples
#' # `linters_with_defaults()` and `linters_with_tags("default")` are the same:
#' all.equal(linters_with_defaults(), linters_with_tags("default"))
#'
#' # Get all linters useful for package development
#' linters <- linters_with_tags(tags = c("package_development", "style"))
#' names(linters)
#'
#' # Get all linters tagged as "default" from lintr and mypkg
#' if (FALSE) {
#'   linters_with_tags("default", packages = c("lintr", "mypkg"))
#' }
#' @export
linters_with_tags <- function(tags, ..., packages = "lintr", exclude_tags = "deprecated") {
  if (missing(tags)) {
    cli_abort("{.arg tags} was not specified. Available tags: {available_tags()}")
  }
  if (!is.character(tags) && !is.null(tags)) {
    cli_abort("{.arg tags} must be a character vector, or {.code NULL}, not {.obj_type_friendly {tags}}.")
  }
  tagged_linter_env <- new.env()

  for (package in packages) {
    ns_exports <- getNamespaceExports(package)
    available <- available_linters(packages = package, tags = tags, exclude_tags = exclude_tags)
    if (nrow(available) > 0L) {
      if (!all(available$linter %in% ns_exports)) {
        missing_linters <- setdiff(available$linter, ns_exports) # nolint: object_usage_linter. TODO(#2252).
        cli_abort(c(
          x = "Can't find linters {.fn {missing_linters}}.",
          i = "These are advertised by {.fn available_linters}, but are not exported by package {.pkg {package}}."
        ))
      }
      for (linter in available$linter) lazily_assign_linter_(linter, package, tagged_linter_env)
    }
  }

  modify_defaults(..., defaults = tagged_linter_env)
}

#' Avoid call_linter_factory up-front to delay displaying warnings until needed.
#'   This e.g. allows modify_defaults() to skip warnings from linters that aren't needed.
#' NB: this helper has the very subtle effect of ensuring 'linter' is associated correctly;
#'   an earlier attempt had this logic directly in a loop in linters_with_tags, but
#'   that results in the value of 'linter' matching that of the loop index after the loop
#'   completes, rather than what its value was when 'delayedAssign()' is called. local({})
#'   _can_ work, but requires the befuddling line 'linter <- linter' to ensure that the
#'   local() environment retains a copy of that variable; the formals of a helper
#'   have the same effect.
#' @noRd
lazily_assign_linter_ <- function(linter, package, env) {
  linter_factory <- get(linter, envir = getNamespace(package), inherits = FALSE)
  delayedAssign(linter, call_linter_factory(linter_factory, linter, package), assign.env = env)
}

#' Create a linter configuration based on all available linters
#'
#' @inheritParams linters_with_tags
#'
#' @examples
#' names(all_linters())
#'
#' @seealso
#' - [linters_with_defaults] for basing off lintr's set of default linters.
#' - [linters_with_tags] for basing off tags attached to linters, possibly across multiple packages.
#' - [available_linters] to get a data frame of available linters.
#' - [linters] for a complete list of linters available in lintr.
#' @export
all_linters <- function(..., packages = "lintr") {
  linters_with_tags(tags = NULL, packages = packages, ...)
}

#' Create a linter configuration based on defaults
#'
#' Make a new list based on \pkg{lintr}'s default linters.
#' The result of this function is meant to be passed to the `linters` argument of `lint()`,
#' or to be put in your configuration file.
#'
#' @param defaults Default list of linters to modify. Must be named.
#' @inheritParams linters_with_tags
#' @examples
#' # When using interactively you will usually pass the result onto `lint` or `lint_package()`
#' f <- tempfile()
#' writeLines("my_slightly_long_variable_name <- 2.3", f)
#' lint(f, linters = linters_with_defaults(line_length_linter = line_length_linter(120L)))
#' unlink(f)
#'
#' # the default linter list with a different line length cutoff
#' my_linters <- linters_with_defaults(line_length_linter = line_length_linter(120L))
#'
#' # omit the argument name if you are just using different arguments
#' my_linters <- linters_with_defaults(defaults = my_linters, object_name_linter("camelCase"))
#'
#' # remove assignment checks (with NULL), add absolute path checks
#' my_linters <- linters_with_defaults(
#'   defaults = my_linters,
#'   assignment_linter = NULL,
#'   absolute_path_linter()
#' )
#'
#' # checking the included linters
#' names(my_linters)
#'
#' @seealso
#' - [linters_with_tags] for basing off tags attached to linters, possibly across multiple packages.
#' - [all_linters] for basing off all available linters in lintr.
#' - [available_linters] to get a data frame of available linters.
#' - [linters] for a complete list of linters available in lintr.
#' @export
linters_with_defaults <- function(..., defaults = default_linters) {
  modify_defaults(..., defaults = defaults)
}

#' @keywords internal
#' @noRd
call_linter_factory <- function(linter_factory, linter_name, package) {
  linter <- tryCatch(
    linter_factory(),
    error = function(e) {
      cli_abort(
        "Could not create linter with {.fun {package}::{linter_name}}.",
        parent = e
      )
    }
  )
  # Otherwise, all linters would be called "linter_factory".
  attr(linter, "name") <- linter_name
  linter
}

#' @keywords internal
#' @noRd
guess_names <- function(..., missing_index) {
  arguments <- as.character(eval(substitute(alist(...)[missing_index])))
  # foo_linter(x=1) => "foo"
  # var[["foo"]]    => "foo"
  # strip call: foo_linter(x=1) --> foo_linter
  # NB: Very long input might have newlines which are not caught
  #  by . in a perl regex; see #774
  arguments <- re_substitutes(arguments, rex("(", anything), "", options = "s")
  # strip extractors: pkg::foo_linter, var[["foo_linter"]] --> foo_linter
  arguments <- re_substitutes(arguments, rex(start, anything, '["' %or% "::"), "")
  re_substitutes(arguments, rex('"]', anything, end), "")
}

test_that("modify_defaults produces error with missing or incorrect defaults", {
  expect_error(
    modify_defaults(),
    "`defaults` is a required argument, but is missing",
    fixed = TRUE
  )
  expect_error(
    modify_defaults("assignment_linter"),
    "`defaults` must be a named list",
    fixed = TRUE
  )
})

test_that("linters_with_tags produces error with incorrect tags", {
  expect_error(linters_with_tags(1L:4L), "`tags` must be a character vector, or `NULL`", fixed = TRUE)
})

test_that("linters_with_defaults works as expected with unnamed args", {
  # assignment_linter is in defaults, so output doesn't change
  expect_named(linters_with_defaults(assignment_linter), names(linters_with_defaults()))
})

test_that("linters_with_defaults warns on unused NULLs", {
  expect_warning(linters_with_defaults(not_a_default = NULL), rex::rex("which is not in `defaults`."))
  expect_warning(
    linters_with_defaults(not_a_default = NULL, also_not_default = NULL),
    rex::rex("which are not in `defaults`.")
  )
})

test_that("linters_with_tags() verifies the output of available_linters()", {
  local_mocked_bindings(
    available_linters = function(...) {
      data.frame(linter = c("fake_linter", "very_fake_linter"), package = "lintr", tags = "")
    }
  )
  expect_error(
    linters_with_tags(NULL),
    "Can't find linters `fake_linter()` and `very_fake_linter()`",
    fixed = TRUE
  )
})

test_that("all default linters are tagged default", {
  expect_named(linters_with_defaults(), available_linters(tags = "default")$linter)

  # covr modifies package functions causing differing deparse() results even for identical anonymous functions.
  # This happens because default_linters is generated at build time and thus not modifiable by covr, whereas
  # linters_with_tags() constructs the linters at runtime.
  skip_on_covr()

  expect_identical(linters_with_tags("default"), linters_with_defaults())
  expect_length(linters_with_tags(NULL, exclude_tags = available_tags()), 0L)

  # Check that above test also trips on default arguments.
  expect_identical(
    all.equal(linters_with_tags("default"), linters_with_defaults(line_length_linter(120L))),
    c(
      'Component "line_length_linter": Component "general_msg": 1 string mismatch',
      'Component "line_length_linter": Component "length": Mean relative difference: 0.5'
    )
  )
})

test_that("can instantiate all linters without arguments", {
  skip_if_not_installed("cyclocomp") # avoid warning

  all_linters <- linters_with_tags(tags = NULL)

  expect_type(all_linters, "list")
  expect_length(all_linters, nrow(available_linters()))

  really_all_linters <- suppressWarnings(linters_with_tags(tags = NULL, exclude_tags = NULL))
  expect_type(really_all_linters, "list")
  expect_length(really_all_linters, nrow(available_linters(exclude_tags = NULL)))
})

test_that("modify_defaults works", {
  my_default <- list(a = 1L, b = 2L, c = 3L)
  expect_identical(modify_defaults(defaults = my_default), my_default)
  expect_identical(modify_defaults(defaults = my_default, a = 2L), list(a = 2L, b = 2L, c = 3L))
  expect_identical(modify_defaults(defaults = my_default, c = NULL), list(a = 1L, b = 2L))

  # auto-sorts
  expect_identical(modify_defaults(defaults = list(b = 2L, a = 1L), c = 3L), my_default)
})

test_that("linters_with_defaults(default = .) is supported with a deprecation warning", {
  # the same warning is not triggered in modify_defaults
  expect_silent({
    linters <- modify_defaults(defaults = list(), default = list(), whitespace_linter())
  })
  expect_named(linters, c("default", "whitespace_linter"))

  # if default= is explicitly provided alongside defaults=, assume that was intentional
  default <- Linter(\(.) list())
  expect_silent({
    linters <- linters_with_defaults(defaults = list(), default = default)
  })
  expect_named(linters, "default")
})

test_that("all_linters contains all available linters", {
  skip_if_not_installed("cyclocomp") # avoid warning

  all_linters <- all_linters(packages = "lintr")

  expect_identical(linters_with_tags(NULL, packages = "lintr"), all_linters)
  expect_length(all_linters, nrow(available_linters()))
})

test_that("all_linters respects ellipsis argument", {
  skip_if_not_installed("cyclocomp") # avoid warning

  expect_identical(
    linters_with_tags(tags = NULL, implicit_integer_linter = NULL),
    all_linters(packages = "lintr", implicit_integer_linter = NULL)
  )
})

test_that("Excluding cyclocomp linter avoids a warning", {
  local_mocked_bindings(
    requireNamespace = function(pkg, ...) pkg != "cyclocomp" || base::requireNamespace(pkg, ...)
  )

  expect_silent(all_linters(cyclocomp_linter = NULL))
  expect_silent(linters_with_tags("configurable", cyclocomp_linter = NULL))
})

test_that("cyclocomp_linter does warn as intended", {
  local_mocked_bindings(
    requireNamespace = function(pkg, ...) pkg != "cyclocomp" && base::requireNamespace(pkg, ...)
  )

  expect_warning(linters_with_tags("configurable"), "cyclocomp::cyclocomp")
})

#' Implicit integer linter
#'
#' Check that integers are explicitly typed using the form `1L` instead of `1`.
#'
#' @param allow_colon Logical, default `FALSE`. If `TRUE`, expressions involving `:`
#'   won't throw a lint regardless of whether the inputs are implicitly integers.
#' @examples
#' # will produce lints
#' lint(
#'   text = "x <- 1",
#'   linters = implicit_integer_linter()
#' )
#'
#' lint(
#'   text = "x[2]",
#'   linters = implicit_integer_linter()
#' )
#'
#' lint(
#'   text = "1:10",
#'   linters = implicit_integer_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x <- 1.0",
#'   linters = implicit_integer_linter()
#' )
#'
#' lint(
#'   text = "x <- 1L",
#'   linters = implicit_integer_linter()
#' )
#'
#' lint(
#'   text = "x[2L]",
#'   linters = implicit_integer_linter()
#' )
#'
#' lint(
#'   text = "1:10",
#'   linters = implicit_integer_linter(allow_colon = TRUE)
#' )
#'
#' @evalRd rd_tags("implicit_integer_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
implicit_integer_linter <- function(allow_colon = FALSE) {
  if (allow_colon) {
    xpath <- "//NUM_CONST[not(parent::expr/parent::expr[
      OP-COLON
      or (OP-MINUS and parent::expr/OP-COLON)
    ])]"
  } else {
    xpath <- "//NUM_CONST"
  }
  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    number_expr <- xml_find_all_(xml, xpath)
    number <- xml_text(number_expr)
    lint_idx <- is_implicit_integer(number)
    number_expr <- number_expr[lint_idx]
    number <- number[lint_idx]
    is_negative <- !is.na(xml_find_first_(number_expr, "parent::expr/preceding-sibling::OP-MINUS"))

    lint_message <-
      sprintf("Use %1$dL or %1$d.0 to avoid implicit integers.", ((-1L)^is_negative) * as.integer(number))

    xml_nodes_to_lints(
      number_expr,
      source_expression = source_expression,
      lint_message = lint_message,
      type = "style",
      column_number_xpath = "number(./@col2 + 1)", # mark at end
      range_end_xpath = "number(./@col2 + 1)" # end after number for easy fixing (enter "L" or ".0")
    )
  })
}

is_implicit_integer <- function(s) {
  # styler: off
  is_implicit <- !re_matches(s, rex(or(
    group(start, upper),  # Inf, NaN and logicals (TRUE, FALSE, NA, NA_*)
    group(start, "0x"),   # hexadecimals, e.g. 0x0f, 0x1p+0 or 0x1.ec6666666p+6
    dot,                  # doubles with a decimal point, e.g. 1.0 or 0.945
    group("i", end),      # complex type, e.g. 1+3i
    group("L", end)       # integer type, e.g. 62L or 1e3L
  )))
  # styler: on

  # only keep numbers that are within the range representable by R, and that are whole
  n <- as.double(s[is_implicit])
  is_implicit[is_implicit] <- abs(n) <= .Machine[["integer.max"]] & floor(n) == n
  is_implicit
}

#' Require correct `sprintf()` calls
#'
#' Check for an inconsistent number of arguments or arguments with incompatible types (for literal arguments) in
#' [sprintf()] calls.
#'
#' [gettextf()] calls are also included, since `gettextf()` is a thin wrapper around `sprintf()`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'sprintf("hello %s %s %d", x, y)',
#'   linters = sprintf_linter()
#' )
#'
#' lint(
#'   text = 'sprintf("hello")',
#'   linters = sprintf_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'sprintf("hello %s %s %d", x, y, z)',
#'   linters = sprintf_linter()
#' )
#'
#' lint(
#'   text = 'sprintf("hello %s %s %d", x, y, ...)',
#'   linters = sprintf_linter()
#' )
#'
#' @evalRd rd_tags("sprintf_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
sprintf_linter <- function() {
  call_xpath <- "
  parent::expr[
    not(expr/SYMBOL[text() = '...'])
  ]"

  pipes <- setdiff(magrittr_pipes, "%$%")
  in_pipe_xpath <- glue(
    "self::expr[
      preceding-sibling::*[not(self::COMMENT)][1][self::PIPE or self::SPECIAL[{ xp_text_in_table(pipes) }]]
    ]"
  )

  is_missing <- function(x) is.symbol(x) && !nzchar(x)

  # Zap sprintf() call to contain only constants
  #
  # Set all extra arguments to 0L if they aren't a constant
  #
  # @param parsed_expr A parsed `sprintf()` call
  #
  # @return A `sprintf()` call with all non-constants replaced by `0L`
  # (which is compatible with all sprintf format specifiers)
  zap_extra_args <- function(parsed_expr) {
    if ("fmt" %in% names(parsed_expr)) {
      fmt_loc <- which(names(parsed_expr) == "fmt")
    } else {
      fmt_loc <- 2L
    }

    if (length(parsed_expr) >= 3L) {
      for (i in setdiff(seq_along(parsed_expr), c(1L, fmt_loc))) {
        if (!is_missing(parsed_expr[[i]]) && !is.atomic(parsed_expr[[i]])) {
          parsed_expr[[i]] <- 0L
        }
      }
    }
    parsed_expr
  }

  # Anticipate warnings of a sprintf() call
  #
  # Try running a static sprintf() call to determine whether it will produce warnings or errors due to format
  # misspecification
  #
  # @param xml An XML node representing a `sprintf()` call (i.e. the `<expr>` node containing the call)
  #
  # @return A string, either `NA_character_` or the text of generated errors and warnings from the `sprintf()` call when
  # replacing all dynamic components by 0, which is compatible with all format specifiers.
  capture_sprintf_warning <- function(xml) {
    parsed_expr <- xml2lang(xml)
    # convert x %>% sprintf(...) to sprintf(x, ...)
    if (length(xml_find_first_(xml, in_pipe_xpath)) > 0L) {
      arg_names <- names(parsed_expr)
      arg_idx <- 2L:length(parsed_expr)
      parsed_expr[arg_idx + 1L] <- parsed_expr[arg_idx]
      names(parsed_expr)[arg_idx + 1L] <- arg_names[arg_idx]
      parsed_expr[[2L]] <- xml2lang(xml_find_first_(xml, "preceding-sibling::*[not(self::COMMENT)][2]"))
      names(parsed_expr)[2L] <- ""
    }
    parsed_expr <- zap_extra_args(parsed_expr)
    res <- tryCatch(eval(parsed_expr, envir = baseenv()), warning = identity, error = identity)
    if (inherits(res, "condition")) {
      conditionMessage(res)
    } else {
      NA_character_
    }
  }

  Linter(linter_level = "file", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(c("sprintf", "gettextf"))
    sprintf_calls <- xml_find_all_(xml_calls, call_xpath)
    in_pipeline <- !is.na(xml_find_first_(sprintf_calls, in_pipe_xpath))

    fmt_by_name <- get_r_string(
      sprintf_calls,
      "SYMBOL_SUB[text() = 'fmt']/following-sibling::expr[1]/STR_CONST"
    )
    fmt_by_pos <- ifelse(
      in_pipeline,
      get_r_string(sprintf_calls, "preceding-sibling::*[not(self::COMMENT)][2]/STR_CONST"),
      get_r_string(sprintf_calls, "OP-LEFT-PAREN/following-sibling::expr[1]/STR_CONST")
    )

    fmt <- ifelse(!is.na(fmt_by_name), fmt_by_name, fmt_by_pos)
    constant_fmt <- !is.na(fmt) & !grepl("%", gsub("%%", "", fmt, fixed = TRUE), fixed = TRUE)

    fct_name <- xp_call_name(sprintf_calls)

    constant_fmt_lint <- xml_nodes_to_lints(
      sprintf_calls[constant_fmt],
      source_expression = source_expression,
      lint_message = sprintf("%s call can be removed when a constant string is provided.", fct_name[constant_fmt]),
      type = "warning"
    )

    templated_sprintf_calls <- sprintf_calls[!constant_fmt & !is.na(fmt)]
    sprintf_warning <- vapply(templated_sprintf_calls, capture_sprintf_warning, character(1L))

    has_warning <- !is.na(sprintf_warning)
    invalid_sprintf_lint <- xml_nodes_to_lints(
      templated_sprintf_calls[has_warning],
      source_expression = source_expression,
      lint_message = sprintf_warning[has_warning],
      type = "warning"
    )

    c(constant_fmt_lint, invalid_sprintf_lint)
  })
}

#' Check that imported packages are actually used
#'
#' @param interpret_glue If `TRUE`, interpret [glue::glue()] calls to avoid false positives caused by local variables
#'   which are only used in a glue expression.
#' @param allow_ns_usage Suppress lints for packages only used via namespace.
#' This is `FALSE` by default because `pkg::fun()` doesn't require `library(pkg)`.
#' You can use [requireNamespace("pkg")][requireNamespace()] to ensure a package is
#' installed without attaching it.
#' @param except_packages Character vector of packages that are ignored.
#' These are usually attached for their side effects.
#'
#' @examples
#' # will produce lints
#' code_lines <- "library(dplyr)\n1 + 1"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unused_import_linter()
#' )
#'
#' code_lines <- "library(dplyr)\ndplyr::tibble(a = 1)"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unused_import_linter()
#' )
#'
#' # okay
#' code_lines <- "library(dplyr)\ntibble(a = 1)"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unused_import_linter()
#' )
#'
#' code_lines <- "library(dplyr)\ndplyr::tibble(a = 1)"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unused_import_linter(allow_ns_usage = TRUE)
#' )
#'
#' @evalRd rd_tags("unused_import_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
unused_import_linter <- function(allow_ns_usage = FALSE,
                                 except_packages = c("bit64", "data.table", "tidyverse"),
                                 interpret_glue = TRUE) {
  # Get dataset names lazy-loaded by imported packages
  get_datasets <- function(pkg) {
    results <- utils::data(package = pkg)$results
    items <- results[, "Item"]
    # e.g. 'state.abb (state)' in 'datasets'
    gsub("\\s*\\([^)]*\\)$", "", items)
  }

  import_xpath <- "
  parent::expr[
    expr[2][STR_CONST]
    or not(SYMBOL_SUB[
      text() = 'character.only' and
      following-sibling::expr[1][NUM_CONST[text() = 'TRUE'] or SYMBOL[text() = 'T']]
    ])
  ]"

  xp_used_functions <- "SYMBOL_FUNCTION_CALL[not(preceding-sibling::NS_GET)]"
  xp_used_symbols <- paste(
    "//SYMBOL[not(
      parent::expr/preceding-sibling::expr[last()]/SYMBOL_FUNCTION_CALL[text() = 'library' or text() = 'require']
    )]",
    "//SPECIAL",
    sep = " | "
  )

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content
    library_calls <- source_expression$xml_find_function_calls(c("library", "require"))
    all_calls <- source_expression$xml_find_function_calls(NULL)

    import_exprs <- xml_find_all_(library_calls, import_xpath)

    if (length(import_exprs) == 0L) {
      return(list())
    }
    imported_pkgs <- xml_find_chr_(import_exprs, "string(expr[STR_CONST|SYMBOL])")
    # as.character(parse(...)) returns one entry per expression
    imported_pkgs <- as.character(parse(text = imported_pkgs, keep.source = FALSE))

    used_symbols <- unique(c(
      xml_text(xml_find_all_(all_calls, xp_used_functions)),
      xml_text(xml_find_all_(xml, xp_used_symbols)),
      extract_glued_symbols(xml, interpret_glue = interpret_glue)
    ))

    is_used <- vapply(
      imported_pkgs,
      function(pkg) {
        # Skip excepted packages and packages that are not installed
        if (pkg %in% except_packages || !requireNamespace(pkg, quietly = TRUE)) {
          return(TRUE)
        }

        package_exports <- getNamespaceExports(pkg) # functions
        dataset_exports <- get_datasets(pkg) # datasets

        any(package_exports %in% used_symbols) || any(dataset_exports %in% used_symbols)
      },
      logical(1L)
    )

    # TODO(#2480): Only call //SYMBOL_PACKAGE once.
    is_ns_used <- vapply(
      imported_pkgs,
      function(pkg) {
        ns_usage <- xml_find_first_(xml, paste0("//SYMBOL_PACKAGE[text() = '", pkg, "']"))
        !identical(ns_usage, xml_missing())
      },
      logical(1L)
    )

    is_unused <- !is_used
    if (allow_ns_usage) {
      is_unused[is_ns_used] <- FALSE
    }

    import_exprs <- import_exprs[is_unused]

    unused_packages <- get_r_string(import_exprs, xpath = "expr[STR_CONST | SYMBOL]")
    lint_message <- ifelse(
      is_ns_used[is_unused][unused_packages],
      paste0(
        "Don't attach package '", unused_packages, "', which is only used by namespace. ",
        "Check that it is installed using loadNamespace() instead."
      ),
      paste0("Package '", unused_packages, "' is attached but never used.")
    )
    xml_nodes_to_lints(import_exprs, source_expression, lint_message, type = "warning")
  })
}

#' Require usage of correctly-typed literals over literal coercions
#'
#' `as.integer(1)` (or `rlang::int(1)`) is the same as `1L` but the latter is
#' more concise and gets typed correctly at compilation.
#'
#' The same applies to missing sentinels like `NA` -- typically, it is not
#'   necessary to specify the storage type of `NA`, but when it is, prefer
#'   using the typed version (e.g. `NA_real_`) instead of a coercion
#'   (like `as.numeric(NA)`).
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "int(1)",
#'   linters = literal_coercion_linter()
#' )
#'
#' lint(
#'   text = "as.character(NA)",
#'   linters = literal_coercion_linter()
#' )
#'
#' lint(
#'   text = "rlang::lgl(1L)",
#'   linters = literal_coercion_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "1L",
#'   linters = literal_coercion_linter()
#' )
#'
#' lint(
#'   text = "NA_character_",
#'   linters = literal_coercion_linter()
#' )
#'
#' lint(
#'   text = "TRUE",
#'   linters = literal_coercion_linter()
#' )
#'
#' @evalRd rd_tags("literal_coercion_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
literal_coercion_linter <- function() {
  rlang_coercers <- c("lgl", "int", "dbl", "chr")
  coercers <- c(
    # base coercers
    paste0("as.", c("logical", "integer", "numeric", "double", "character")),
    rlang_coercers
  )

  # notes for clarification:
  #  - as.integer(1e6) is arguably easier to read than 1000000L
  #  - in x$"abc", the "abc" STR_CONST is at the top level, so exclude OP-DOLLAR (ditto OP-AT)
  #  - need condition against STR_CONST w/ EQ_SUB to skip quoted keyword arguments (see tests)
  #  - for {rlang} coercers, both `int(1)` and `int(1, )` need to be linted
  not_extraction_or_scientific <- "
    not(OP-DOLLAR or OP-AT)
    and (
      NUM_CONST[not(contains(translate(text(), 'E', 'e'), 'e'))]
      or STR_CONST[not(following-sibling::*[not(self::COMMENT)][1][self::EQ_SUB])]
    )
  "
  xpath <- glue("
  parent::expr[
    count(expr) = 2
    and expr[2][ {not_extraction_or_scientific} ]
  ]")

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(coercers)
    bad_expr <- xml_find_all_(xml_calls, xpath)

    coercer <- xp_call_name(bad_expr)
    # tiptoe around the fact that we don't require {rlang}
    is_rlang_coercer <- coercer %in% rlang_coercers
    if (any(is_rlang_coercer) && !requireNamespace("rlang", quietly = TRUE)) {
      # nocov start: test suite will have 'rlang' available
      # NB: we _could_ do some extreme customization where each lint
      #   gets a message according to whether the coercer is from rlang,
      #   but this seems like overkill. Just use a generic message and move on.
      lint_message <- paste(
        "Use literals directly where possible, instead of coercion.",
        "c.f. 1L instead of as.integer(1) or rlang::int(1), or NA_real_ instead of as.numeric(NA).",
        "NB: this message can be improved to show a specific replacement if 'rlang' is installed."
      )
      # nocov end
    } else {
      bad_expr <- strip_comments_from_subtree(bad_expr)
      # duplicate, unless we add 'rlang::' and it wasn't there originally
      coercion_str <- report_str <- xml_text(bad_expr)
      if (any(is_rlang_coercer) && !("package:rlang" %in% search())) {
        needs_prefix <- is_rlang_coercer & !startsWith(coercion_str, "rlang::")
        coercion_str[needs_prefix] <- paste0("rlang::", coercion_str[needs_prefix])
      }
      # the linter logic & rlang requirement should ensure that it's safe to run eval() here;
      #   suppressWarnings() is for cases like 'as.integer("a")' which have an NA result, #2566.
      # TODO(#2473): Avoid a recommendation like '1' that clashes with implicit_integer_linter().
      literal_equivalent_str <- vapply(
        str2expression(coercion_str),
        \(expr) deparse1(suppressWarnings(eval(expr))),
        character(1L)
      )
      lint_message <- sprintf(
        "Use %s instead of %s, i.e., use literals directly where possible, instead of coercion.",
        literal_equivalent_str, report_str
      )
    }

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = lint_message,
      type = "warning"
    )
  })
}

#' Warn about invalid usage of `all.equal()`
#'
#' [all.equal()] returns `TRUE` in the absence of differences but return a
#' character string (not `FALSE`) in the presence of differences.
#' Usage of `all.equal()` without wrapping it in `isTRUE()` in `if` clauses, or
#' preceded by the negation operator `!`, are thus likely to generate unexpected
#' errors if the compared objects have differences.
#' An alternative is to use `identical()` to compare vector of strings or when
#' exact equality is expected.
#'
#' @examples
#' # lints
#' lint(
#'   text = 'if (all.equal(a, b)) message("equal")',
#'   linters = all_equal_linter()
#' )
#'
#' lint(
#'   text = '!all.equal(a, b)',
#'   linters = all_equal_linter()
#' )
#'
#' lint(
#'   text = 'isFALSE(all.equal(a, b))',
#'   linters = all_equal_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'if (isTRUE(all.equal(a, b))) message("equal")',
#'   linters = all_equal_linter()
#' )
#'
#' lint(
#'   text = '!identical(a, b)',
#'   linters = all_equal_linter()
#' )
#'
#' lint(
#'   text = "!isTRUE(all.equal(a, b))",
#'   linters = all_equal_linter()
#' )
#'
#' @evalRd rd_tags("all_equal_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
all_equal_linter <- function() {

  Linter(linter_level = "expression", function(source_expression) {
    all_equal_calls <- source_expression$xml_find_function_calls("all.equal")

    dangerous_unwrapped_all_equal <- xml_find_all_(
      all_equal_calls,
      "parent::expr[
        preceding-sibling::*[not(self::COMMENT)][2][self::IF or self::WHILE]
        or preceding-sibling::*[not(self::COMMENT)][1][self::OP-EXCLAMATION]
      ]"
    )

    has_tolerance_arg <- !is.na(
      xml_find_first_(dangerous_unwrapped_all_equal, "SYMBOL_SUB[text() = 'tolerance']")
    )

    unwrapped_all_equal_lints <- xml_nodes_to_lints(
      dangerous_unwrapped_all_equal,
      source_expression = source_expression,
      lint_message = ifelse(
        has_tolerance_arg,
        "Wrap all.equal() in isTRUE().",
        "Wrap all.equal() in isTRUE(), or replace it by identical() if no tolerance is required."
      ),
      type = "warning"
    )

    is_false_all_equal <- xml_find_all_(
      all_equal_calls,
      "parent::expr[preceding-sibling::expr[1]/SYMBOL_FUNCTION_CALL/text() = 'isFALSE']"
    )

    is_false_all_equal_lints <- xml_nodes_to_lints(
      is_false_all_equal,
      source_expression = source_expression,
      lint_message =
        "Use !isTRUE() to check for differences in all.equal(). isFALSE(all.equal()) always returns FALSE.",
      type = "warning"
    )

    c(unwrapped_all_equal_lints, is_false_all_equal_lints)
  })
}

has_description <- function(path) {
  desc_info <- file.info(file.path(path, "DESCRIPTION"))
  !is.na(desc_info$size) && desc_info$size > 0.0 && !desc_info$isdir
}

has_rproj <- function(path) {
  length(Sys.glob(file.path(path, "*.Rproj"))) > 0L
}

find_package <- function(path, allow_rproj = FALSE, max_depth = 2L) {
  path <- normalize_path(path, mustWork = !allow_rproj)
  if (allow_rproj) {
    found <- function(path) has_description(path) || has_rproj(path)
  } else {
    found <- function(path) has_description(path)
  }

  depth <- max_depth
  while (!found(path)) {
    path <- dirname(path)
    if (is_root(path) || depth <= 0L) {
      return(NULL)
    }
    depth <- depth - 1L
  }
  path
}

find_rproj_at <- function(path) {
  head(Sys.glob(file.path(path, "*.Rproj")), n = 1L)
}

is_root <- function(path) {
  identical(path, dirname(path))
}

is_directory <- function(filename) isTRUE(file.info(filename)$isdir)

#' Return the first of a vector of files that exists.
#'
#' Avoid running 'expensive' [file.exists()] for the full vector,
#'   since typically the first entries will lead to early exit.
#' TODO(#2204): check if the implementation should be simpler
#' @noRd
first_exists <- function(files) {
  for (file in files) {
    if (file.exists(file)) {
      return(file)
    }
  }
  NULL
}

find_config <- function(filename) {
  if (is.null(filename)) {
    return(NULL)
  }
  linter_file <- lintr_option("linter_file")

  ## if users changed lintr.linter_file, return immediately.
  if (is_absolute_path(linter_file) && file.exists(linter_file)) {
    return(linter_file)
  }

  path <- if (is_directory(filename)) {
    filename
  } else {
    dirname(filename)
  }

  path <- normalize_path(path, mustWork = FALSE)

  # NB: This vector specifies a priority order for where to find the configs,
  # i.e. the first location where a config exists is chosen and configs which
  # may exist in subsequent directories are ignored
  file_locations <- c(
    # Local (incl. parent) directories
    find_local_config(path, linter_file),
    # User directory
    # cf: rstudio@bc9b6a5 SessionRSConnect.R#L32
    file.path(Sys.getenv("HOME", unset = "~"), linter_file),
    # Next check for a global config file
    file.path(R_user_dir("lintr", which = "config"), "config")
  )

  first_exists(file_locations)
}

find_local_config <- function(path, config_file) {
  # R config gets precedence
  configs_to_check <- c(paste0(config_file, ".R"), config_file)
  repeat {
    guesses_in_dir <- c(
      file.path(path, configs_to_check),
      file.path(path, ".github", "linters", configs_to_check)
    )
    found <- first_exists(guesses_in_dir)
    if (!is.null(found)) {
      return(found)
    }
    path <- dirname(path)
    if (is_root(path)) {
      return(character())
    }
  }
}

pkg_name <- function(path = find_package()) {
  if (is.null(path)) {
    return(NULL)
  }
  nm <- read.dcf(file.path(path, "DESCRIPTION"), fields = "Package")[1L]
  if (!is.na(nm)) nm
}

#' Spaces before parentheses linter
#'
#' Check that all left parentheses have a space before them unless they are in a function call.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "if(TRUE) x else y",
#'   linters = spaces_left_parentheses_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "if (TRUE) x else y",
#'   linters = spaces_left_parentheses_linter()
#' )
#'
#' @evalRd rd_tags("spaces_left_parentheses_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#parentheses>
#' - [function_left_parentheses_linter()]
#' @export
spaces_left_parentheses_linter <- function() {
  file_level_xpath <- "//OP-LEFT-PAREN[@start - 1 = ancestor::expr/preceding-sibling::OP-SEMICOLON/@end]"

  # apply the lint by requiring a gap in three cases:
  #   (1) if/while loop conditions, e.g. 'if(x>2) { }', including 'else('
  #   (2) for loop conditions, e.g. 'for(i in 1:5) { }' [very similar to (1) in code but different in XML]
  #   (3) non-unary infix operators, e.g. 'x&(y | z)', and including commas, braces, e.g. 'c(a,(a+b))'

  # -1 on LHS because, when RHS matches nothing, +1 tricks the condition into returning true...
  #   while @start will always be there, the @end may not be.
  if_while_cond <- "@start - 1 = preceding-sibling::*[self::IF or self::WHILE]/@end"
  for_cond <- "@start - 1 = parent::forcond/preceding-sibling::FOR/@end"

  # see infix_spaces_linter.R; preceding-sibling::* is needed for unary operators where '-(a)' is ok
  unary_nodes <- infix_metadata[infix_metadata$unary, "xml_tag"]
  unary_selves <- paste0("self::", unary_nodes, "[preceding-sibling::*]", collapse = " or ")
  binary_nodes <- c(
    infix_metadata[infix_metadata$low_precedence & !infix_metadata$unary, "xml_tag"],
    "OP-COMMA", "OP-LEFT-BRACE", "ELSE", "IN"
  )
  binary_selves <- paste0("self::", binary_nodes, collapse = " or ")
  # preceding-symbol::* catches (1) function definitions and (2) function calls
  # ancestor::expr needed for nested RHS expressions, e.g. 'y1<-(abs(yn)>90)*1'
  infix_cond <- sprintf(
    "not(preceding-sibling::*) and (@start - 1 = ancestor::expr/preceding-sibling::*[%s]/@end)",
    paste(unary_selves, "or", binary_selves)
  )
  expression_level_xpath <- sprintf(
    "//OP-LEFT-PAREN[(%s) or (%s) or (%s)]",
    if_while_cond, for_cond, infix_cond
  )

  Linter(function(source_expression) {
    # else/if structure for trees that are missing XML content (both global & local)
    if (is_lint_level(source_expression, "file")) {
      # 'x = 1;(x + 2)' can't be detected from the expression-level tree
      xml <- source_expression$full_xml_parsed_content
      xpath <- file_level_xpath
    } else {
      xml <- source_expression$xml_parsed_content
      xpath <- expression_level_xpath
    }

    bad_paren <- xml_find_all_(xml, xpath)

    xml_nodes_to_lints(
      bad_paren,
      source_expression,
      lint_message = "Place a space before left parenthesis, except in a function call.",
      type = "style"
    )
  })
}

#' Avoid implicit assignment in function calls
#'
#' Assigning inside function calls makes the code difficult to read, and should
#' be avoided, except for functions that capture side-effects (e.g. [utils::capture.output()]).
#'
#' @param except A character vector of functions to be excluded from linting.
#' @param allow_lazy logical, default `FALSE`. If `TRUE`, assignments that only
#'   trigger conditionally (e.g. in the RHS of `&&` or `||` expressions) are skipped.
#' @param allow_scoped Logical, default `FALSE`. If `TRUE`, "scoped assignments",
#'   where the object is assigned in the statement beginning a branch and used only
#'   within that branch, are skipped.
#' @param allow_paren_print Logical, default `FALSE`. If `TRUE`, assignments using
#'   `(` for auto-printing at the top-level are not linted.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "if (x <- 1L) TRUE",
#'   linters = implicit_assignment_linter()
#' )
#'
#' lint(
#'   text = "mean(x <- 1:4)",
#'   linters = implicit_assignment_linter()
#' )
#'
#' lint(
#'   text = "(x <- 1)",
#'   linters = implicit_assignment_linter()
#' )
#'
#'
#' # okay
#' lines <- "x <- 1L\nif (x) TRUE"
#' writeLines(lines)
#' lint(
#'   text = lines,
#'   linters = implicit_assignment_linter()
#' )
#'
#' lines <- "x <- 1:4\nmean(x)"
#' writeLines(lines)
#' lint(
#'   text = lines,
#'   linters = implicit_assignment_linter()
#' )
#'
#' lint(
#'   text = "A && (B <- foo(A))",
#'   linters = implicit_assignment_linter(allow_lazy = TRUE)
#' )
#'
#' lines <- c(
#'   "if (any(idx <- x < 0)) {",
#'   "  stop('negative elements: ', toString(which(idx)))",
#'   "}"
#' )
#' writeLines(lines)
#' lint(
#'   text = lines,
#'   linters = implicit_assignment_linter(allow_scoped = TRUE)
#' )
#'
#' lint(
#'   text = "(x <- 1)",
#'   linters = implicit_assignment_linter(allow_paren_print = TRUE)
#' )
#'
#' @evalRd rd_tags("implicit_assignment_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#assignment>
#'
#' @export
implicit_assignment_linter <- function(except = c("bquote", "expression", "expr", "quo", "quos", "quote"),
                                       allow_lazy = FALSE,
                                       allow_scoped = FALSE,
                                       allow_paren_print = FALSE) {
  if (!is.null(except) && !is.character(except)) {
    cli_abort("{.arg except} must be a character vector or NULL, not {.obj_type_friendly {except}}.")
  }

  if (length(except) > 0L) {
    exceptions <- xp_text_in_table(except)
    xpath_exceptions <- glue("SYMBOL_FUNCTION_CALL[ not({exceptions}) ]")
  } else {
    xpath_exceptions <- "SYMBOL_FUNCTION_CALL"
  }

  # The walrus operator `:=` is also `LEFT_ASSIGN`, but is not a relevant operator
  # to be considered for the present linter.
  assignments <- paste(
    "//LEFT_ASSIGN[text() != ':=']",
    "//RIGHT_ASSIGN",
    sep = " | "
  )

  xpath <- glue("
    ({assignments})
      /parent::expr[
        preceding-sibling::*[not(self::COMMENT)][2][self::IF or self::WHILE]
        or parent::forcond
        or preceding-sibling::expr/{xpath_exceptions}
        or parent::expr/*[1][self::OP-LEFT-PAREN]
      ]
  ")

  if (allow_lazy) {
    xpath <- paste0(xpath, "[not(ancestor::expr/preceding-sibling::*[self::AND2 or self::OR2])]")
  }
  if (allow_scoped) {
    # force 2nd preceding to ensure we're in the loop condition, not the loop expression
    in_branch_cond <- "ancestor-or-self::expr[preceding-sibling::*[not(self::COMMENT)][2][self::IF or self::WHILE]]"
    xpath <- paste0(
      xpath,
      # _if_ we're in an IF/WHILE branch, lint if the assigned SYMBOL appears anywhere later on.
      glue("[not({in_branch_cond}) or expr[1]/SYMBOL = {in_branch_cond}/parent::expr/following::SYMBOL]")
    )
  }

  implicit_message <- paste(
    "Avoid implicit assignments in function calls.",
    "For example, instead of `if (x <- 1L) { ... }`, write `x <- 1L; if (x) { ... }`."
  )

  print_message <- "Call print() explicitly instead of relying on implicit printing behavior via '('."

  Linter(linter_level = "file", function(source_expression) {
    # need the full file to also catch usages at the top level
    xml <- source_expression$full_xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)

    print_only <- !is.na(xml_find_first_(bad_expr, "parent::expr[parent::exprlist and *[1][self::OP-LEFT-PAREN]]"))
    if (allow_paren_print) {
      bad_expr <- bad_expr[!print_only]
    }

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = ifelse(print_only, print_message, implicit_message),
      type = "warning"
    )
  })
}

#' Equality check with NA linter
#'
#' Check for `x == NA`, `x != NA` and `x %in% NA`. Such usage is almost surely incorrect --
#' checks for missing values should be done with [is.na()].
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x == NA",
#'   linters = equals_na_linter()
#' )
#'
#' lint(
#'   text = "x != NA",
#'   linters = equals_na_linter()
#' )
#'
#' lint(
#'   text = "x %in% NA",
#'   linters = equals_na_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "is.na(x)",
#'   linters = equals_na_linter()
#' )
#'
#' lint(
#'   text = "!is.na(x)",
#'   linters = equals_na_linter()
#' )
#'
#' @evalRd rd_tags("equals_na_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
equals_na_linter <- function() {
  na_table <- xp_text_in_table(c("NA", "NA_integer_", "NA_real_", "NA_complex_", "NA_character_"))

  xpath <- glue("
  //NUM_CONST[ {na_table} ]
    /parent::expr
    /parent::expr[EQ or NE]
  |
  //SPECIAL[text() = '%in%' and following-sibling::expr/NUM_CONST[ {na_table} ]]
    /parent::expr
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)
    op <- xml_find_first_(bad_expr, "EQ | NE | SPECIAL")

    xml_nodes_to_lints(
      bad_expr,
      source_expression,
      lint_message = sprintf("Use is.na() instead of x %s NA", xml_text(op)),
      type = "warning"
    )
  })
}

#' Block usage of all() within stopifnot()
#'
#' `stopifnot(A)` actually checks `all(A)` "under the hood" if `A` is a vector,
#'   and produces a better error message than `stopifnot(all(A))` does.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "stopifnot(all(x > 0))",
#'   linters = stopifnot_all_linter()
#' )
#'
#' lint(
#'   text = "stopifnot(y > 3, all(x < 0))",
#'   linters = stopifnot_all_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "stopifnot(is.null(x) || all(x > 0))",
#'   linters = stopifnot_all_linter()
#' )
#'
#' lint(
#'   text = "assert_that(all(x > 0))",
#'   linters = stopifnot_all_linter()
#' )
#'
#' @evalRd rd_tags("stopifnot_all_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
stopifnot_all_linter <- make_linter_from_function_xpath(
  function_names = "stopifnot",
  xpath = "
  parent::expr
    /expr[expr/SYMBOL_FUNCTION_CALL[text() = 'all']]
  ",
  lint_message = paste(
    "Use stopifnot(x) instead of stopifnot(all(x)).",
    "stopifnot(x) runs all() 'under the hood' and provides a better error message in case of failure."
  )
)

# Each of the default linters should throw at least one lint on this file

# assignment
# function_left_parentheses
# brace_linter
# commas
# paren_brace
f = function (x,y = 1){}

# return_linter
g <- function(x) {
  return(x + 1)
}

# commented_code
# some <- commented("out code")

# equals_na
# brace_linter
# indentation
# infix_spaces
# line_length
# object_length
# object_name
# object_usage
# open_curly
# T_and_F_symbol
someComplicatedFunctionWithALongCamelCaseName <- function(x)
{
    y <- 1
  if (1 > 2 && 2 > 3 && 3 > 4 && 4 > 5 && 5*10 > 6 && 5 > 6 && 6 > 7 && x == NA) {T} else F
}

# vector_logic
if (1 & 2) FALSE else TRUE

# function_brace
my_metric <- function(x)
  sum(x) + prod(x)

# no_tab
# pipe_consistency
# pipe_continuation
# seq_linter
# spaces_inside
# indentation
x <- 1:10
x[ 2]
1:length(x) %>% lapply(function(x) x*2) %>%
	head()

# single_quotes
message('single_quotes')

# spaces_left_parentheses
# trailing_whitespace
# semicolon
x <- 42; y <- 2 +(1:10) 

# trailing_blank_lines

#' Exclude lines or files from linting
#'
#' @param lints that need to be filtered.
#' @param exclusions manually specified exclusions
#' @param linter_names character vector of names of the active linters, used for parsing inline exclusions.
#' @param ... additional arguments passed to [parse_exclusions()]
#' @eval c(
#'   # we use @eval for the details section to avoid a literal nolint exclusion tag with non-existing linter names
#'   # those produce a warning from [parse_exclusions()] otherwise. See #1219 for details.
#'   "@details",
#'   "Exclusions can be specified in three different ways.",
#'   "",
#'   "1. Single line in the source file. default: `# nolint`, possibly followed by a listing of linters to exclude.",
#'   "   If the listing is missing, all linters are excluded on that line. The default listing format is",
#'   paste(
#'     "   `#",
#'     "nolint: linter_name, linter2_name.`. There may not be anything between the colon and the line exclusion tag"
#'   ),
#'   "   and the listing must be terminated with a full stop (`.`) for the linter list to be respected.",
#'   "2. Line range in the source file. default: `# nolint start`, `# nolint end`. `# nolint start` accepts linter",
#'   "   lists in the same form as `# nolint`.",
#'   "3. Exclusions parameter, a list with named and/or unnamed entries. ",
#'   "   Outer elements have the following characteristics:",
#'   "   1. Unnamed elements specify filenames or directories.",
#'   "   2. Named elements are a vector or list of line numbers, with `Inf` indicating 'all lines'.",
#'   "      The name gives a path relative to the config.",
#'   "      1. Unnamed elements denote exclusion of all linters in the given path or directory.",
#'   "      2. Named elements, where the name specifies a linter, denote exclusion for that linter.",
#'   "   For convenience, a vector can be used in place of a list whenever it would not introduce ambiguity, e.g.",
#'   "   a character vector of files to exclude or a vector of lines to exclude.",
#'   "",
#'   "   Note also that:",
#'   "   1. All paths are interpreted as globs ([Sys.glob()]), so that e.g. `*` does pattern expansion.",
#'   "   2. For `exclusions` provided _via_ a config file, all paths are taken as",
#'   "      relative _to the config file itself_, not the path where `lint()` is run.",
#'   NULL
#' )
#'
#' @keywords internal
exclude <- function(lints, exclusions = settings$exclusions, linter_names = NULL, ...) {
  if (length(lints) <= 0L) {
    return(lints)
  }

  lint_df <- as.data.frame(lints)

  filenames <- unique(lint_df$filename)
  source_exclusions <- lapply(filenames, parse_exclusions, linter_names = linter_names, ...)
  names(source_exclusions) <- filenames


  exclusions <- normalize_exclusions(c(source_exclusions, exclusions))
  to_exclude <- vapply(
    seq_len(nrow(lint_df)),
    function(i) {
      filename <- lint_df$filename[i]
      filename %in% names(exclusions) &&
        is_excluded(lint_df$line_number[i], lint_df$linter[i], exclusions[[filename]])
    },
    logical(1L)
  )

  if (any(to_exclude)) {
    lints <- lints[!to_exclude]
  }

  lints
}

is_excluded <- function(line_number, linter, file_exclusion) {
  excluded_lines <- unlist(file_exclusion[names2(file_exclusion) %in% c("", linter)])
  Inf %in% excluded_lines || line_number %in% excluded_lines
}

is_excluded_file <- function(file_exclusion) {
  any(vapply(
    file_exclusion[!nzchar(names2(file_exclusion))],
    \(full_exclusion) Inf %in% full_exclusion,
    logical(1L)
  ))
}

line_info <- function(line_numbers, type = c("start", "end")) {
  type <- match.arg(type)
  range_word <- paste0("range ", type, if (length(line_numbers) != 1L) "s")
  n <- length(line_numbers)
  if (n == 0L) {
    paste("0", range_word)
  } else if (n == 1L) {
    paste0("1 ", range_word, " (line ", line_numbers, ")")
  } else {
    paste0(n, " ", range_word, " (lines ", toString(line_numbers), ")")
  }
}

#' read a source file and parse all the excluded lines from it
#'
#' @param file R source file
#' @param exclude Regular expression used to mark lines to exclude.
#' @param exclude_next Regular expression used to mark lines immediately preceding excluded lines.
#' @param exclude_start Regular expression used to mark the start of an excluded range.
#' @param exclude_end Regular expression used to mark the end of an excluded range.
#' @param exclude_linter Regular expression used to capture a list of to-be-excluded linters immediately following a
#' `exclude` or `exclude_start` marker.
#' @param exclude_linter_sep Regular expression used to split a linter list into individual linter names for exclusion.
#' @param lines A character vector of the content lines of `file`.
#' @param linter_names Names of active linters.
#'
#' @return A possibly named list of excluded lines, possibly for specific linters.
#' @keywords internal
parse_exclusions <- function(file,
                             exclude = settings$exclude,
                             exclude_next = settings$exclude_next,
                             exclude_start = settings$exclude_start,
                             exclude_end = settings$exclude_end,
                             exclude_linter = settings$exclude_linter,
                             exclude_linter_sep = settings$exclude_linter_sep,
                             lines = NULL,
                             linter_names = NULL) {
  lines <- lines %||% ensure_utf8(read_lines(file), settings$encoding)

  exclusions <- list()

  if (is_tainted(lines)) {
    # Invalid encoding. Don't parse exclusions.
    return(list())
  }

  start_locations <- re_matches_locations(lines, exclude_start)[, "end"] + 1L
  end_locations <- re_matches_locations(lines, exclude_end)[, "start"]
  starts <- which(!is.na(start_locations))
  ends <- which(!is.na(end_locations))

  if (length(starts) > 0L) {
    if (length(starts) != length(ends)) {
      starts_msg <- line_info(starts, type = "start") # nolint: object_usage_linter. TODO(#2252).
      ends_msg <- line_info(ends, type = "end") # nolint: object_usage_linter. TODO(#2252).
      cli_abort(c(
        i = "Equal number of line starts and ends expected for exclusion from linting.",
        x = "{.file {file}} has {.strong {starts_msg}} and {.strong {ends_msg}}."
      ))
    }

    for (i in seq_along(starts)) {
      excluded_lines <- seq(starts[i], ends[i])
      linters_string <- substring(lines[starts[i]], start_locations[starts[i]])
      linters_string <- re_matches(linters_string, exclude_linter)[, 1L]

      exclusions <- add_exclusions(exclusions, excluded_lines, linters_string, exclude_linter_sep, linter_names)
    }
  }

  next_locations <- re_matches_locations(lines, exclude_next)[, "end"] + 1L
  nexts <- which(!is.na(next_locations))

  nolint_locations <- re_matches_locations(lines, exclude)[, "end"] + 1L
  nolints <- which(!is.na(nolint_locations))

  # Disregard nolint tags if they also match nolint next / start / end
  nolints <- setdiff(nolints, c(nexts, starts, ends))

  for (nolint in nolints) {
    linters_string <- get_linters_string(lines[nolint], nolint_locations[nolint], exclude_linter)
    exclusions <- add_exclusions(exclusions, nolint, linters_string, exclude_linter_sep, linter_names)
  }
  for (nextt in nexts) {
    linters_string <- get_linters_string(lines[nextt], next_locations[nextt], exclude_linter)
    exclusions <- add_exclusions(exclusions, nextt + 1L, linters_string, exclude_linter_sep, linter_names)
  }

  exclusions[] <- lapply(exclusions, \(lines) sort(unique(lines)))

  exclusions
}

get_linters_string <- function(line, loc, exclude_linter) {
  linters_string <- substring(line, loc)
  re_matches(linters_string, exclude_linter)[, 1L]
}

add_excluded_lines <- function(exclusions, excluded_lines, excluded_linters) {
  for (linter in excluded_linters) {
    if (linter %in% names2(exclusions)) {
      i <- which(names2(exclusions) %in% linter)
      exclusions[[i]] <- c(exclusions[[i]], excluded_lines)
    } else {
      exclusions <- c(exclusions, list(excluded_lines))
      if (nzchar(linter)) {
        if (is.null(names(exclusions))) {
          # Repair names if linter == "" is the first exclusion added.
          names(exclusions) <- ""
        }
        names(exclusions)[length(exclusions)] <- linter
      }
    }
  }

  exclusions
}

add_exclusions <- function(exclusions, lines, linters_string, exclude_linter_sep, linter_names) {
  # No match for linter list: Add to global excludes
  if (is.na(linters_string)) {
    exclusions <- add_excluded_lines(exclusions, lines, "")
  } else {
    # Matched a linter list: only add excluded lines for the listed linters.
    excluded_linters <- strsplit(linters_string, exclude_linter_sep)[[1L]]
    if (!is.null(linter_names)) {
      idxs <- pmatch(excluded_linters, linter_names, duplicates.ok = TRUE)
      matched <- !is.na(idxs)
      if (!all(matched)) {
        bad <- excluded_linters[!matched] # nolint: object_usage_linter. TODO(#2252).
        cli_warn(c(
          x = "Could not find linter{?s} named {.field {bad}} in the list of active linters.",
          i = "Make sure the linter is uniquely identified by the given name or prefix."
        ))
      }
      excluded_linters[matched] <- linter_names[idxs[matched]]
    }
    exclusions <- add_excluded_lines(exclusions, lines, excluded_linters)
  }
  exclusions
}

#' Normalize lint exclusions
#'
#' @param x Exclusion specification
#'  - A character vector of filenames or directories relative to `root`. Interpreted as globs, see [Sys.glob()].
#'  - A named list of integers specifying lines to be excluded per file
#'  - A named list of named lists specifying linters and lines to be excluded for the linters per file.
#' @param normalize_path Should the names of the returned exclusion list be normalized paths?
#'   If `FALSE`, they will be relative to `root`.
#' @param root Base directory for relative filename resolution.
#' @param pattern If non-NULL, only exclude files in excluded directories if they match
#'   `pattern`. Passed to [list.files()] if a directory is excluded.
#'
#' @return A named list of file exclusions.
#'   The names of the list specify the filenames to be excluded.
#'
#' Each file exclusion is a possibly named list containing line numbers to exclude, or the sentinel `Inf` for
#'   completely excluded files. If the an entry is named, the exclusions only take effect for the linter with
#'   the same name.
#'
#' If `normalize_path` is `TRUE`, file names will be normalized relative to `root`.
#'   Otherwise the paths are left as provided (relative to `root` or absolute). That also means
#'   existence is not checked.
#'
#' @keywords internal
normalize_exclusions <- function(x, normalize_path = TRUE,
                                 root = getwd(),
                                 pattern = NULL) {
  if (is.null(x) || length(x) <= 0L) {
    return(list())
  }

  x <- as.list(x)
  unnamed <- !nzchar(names2(x))
  if (any(unnamed)) {
    bad <- vapply(
      seq_along(x),
      function(i) {
        unnamed[i] && (!is.character(x[[i]]) || length(x[[i]]) != 1L)
      },
      logical(1L)
    )

    if (any(bad)) {
      cli_abort(c(
        i = "Full file exclusions must be {.cls character} vectors of length 1.",
        x = "Items at the following indices are not: {.val {which(bad)}}."
      ))
    }
    # Normalize unnamed entries to list(<filename> = list(Inf), ...)
    names(x)[unnamed] <- x[unnamed]
    x[unnamed] <- rep_len(list(list(Inf)), sum(unnamed))
  }

  full_line_exclusions <- !vapply(x, is.list, logical(1L))

  if (any(full_line_exclusions)) {
    # must be integer or numeric vectors
    are_numeric <- vapply(x, is.numeric, logical(1L))
    bad <- full_line_exclusions & !are_numeric

    if (any(bad)) {
      cli_abort(c(
        i = "Full line exclusions must be {.cls numeric} or {.cls integer} vectors.",
        x = "Items at the following indices are not: {.val {which(bad)}}."
      ))
    }

    # Normalize list(<filename> = c(<lines>)) to
    # list(<filename> = list(c(<lines>)))
    x[full_line_exclusions] <- lapply(x[full_line_exclusions], list)
  }

  paths <- names(x)
  rel_path <- !is_absolute_path(paths)
  paths[rel_path] <- file.path(root, paths[rel_path])

  globbed_paths <- lapply(paths, Sys.glob)
  n_files <- lengths(globbed_paths)
  # restore unmatched globs
  if (!normalize_path) {
    empty <- n_files == 0L
    globbed_paths[empty] <- as.list(names(x)[empty])
    n_files[empty] <- 1L
  }
  x <- rep(x, n_files)
  paths <- unlist(globbed_paths)
  names(x) <- paths

  is_dir <- dir.exists(paths)
  if (any(is_dir)) {
    dirs <- names(x)[is_dir]
    x <- x[!is_dir]
    all_file_names <- unlist(lapply(
      dirs,
      function(dir) {
        dir_path <- if (is_absolute_path(dir)) dir else file.path(root, dir)
        files <- list.files(
          path = dir_path,
          pattern = pattern,
          recursive = TRUE
        )
        file.path(dir, files) # non-normalized relative paths
      }
    ))

    # Only exclude file if there is no more specific exclusion already
    all_file_names <- setdiff(all_file_names, names(x))

    dir_exclusions <- rep_len(list(Inf), length(all_file_names))
    names(dir_exclusions) <- all_file_names
    x <- c(x, dir_exclusions)
  }

  if (normalize_path) {
    paths <- names(x)
    # specify relative paths w.r.t. root
    rel_path <- !is_absolute_path(paths)
    paths[rel_path] <- file.path(root, paths[rel_path])
    names(x) <- paths
    x <- x[file.exists(paths)] # remove exclusions for non-existing files
    names(x) <- normalize_path(names(x)) # get full path for remaining files
  }
  x |>
    remove_empty() |>
    remove_file_duplicates() |>
    remove_linter_duplicates() |>
    remove_line_duplicates()
}

# Combines file exclusions for identical files.
remove_file_duplicates <- function(x) {
  unique_names <- unique(names(x))

  ## check for duplicate files
  if (length(unique_names) < length(names(x))) {
    x <- lapply(
      unique_names,
      function(name) {
        vals <- unname(x[names(x) == name])
        do.call(c, vals)
      }
    )

    names(x) <- unique_names
  }

  x
}

# Removes duplicate line information for each linter within each file.
remove_line_duplicates <- function(x) {
  x[] <- lapply(x, function(ex) {
    ex[] <- lapply(ex, unique)
    ex
  })

  x
}

# Combines line exclusions for identical linters within each file.
remove_linter_duplicates <- function(x) {
  x[] <- lapply(x, function(ex) {
    unique_linters <- unique(names2(ex))

    if (length(unique_linters) < length(ex)) {
      ex <- lapply(unique_linters, function(linter) {
        excluded_lines <- unlist(ex[names2(ex) == linter])
        if (Inf %in% excluded_lines) {
          Inf
        } else {
          excluded_lines
        }
      })

      if (!identical(unique_linters, "")) {
        names(ex) <- unique_linters
      }
    }

    ex
  })

  x
}

# Removes linter exclusions without lines and files without any linter exclusions.
remove_empty <- function(x) {
  x[] <- lapply(x, \(ex) ex[lengths(ex) > 0L])
  x[lengths(x) > 0L]
}

#' Require usage of `colSums(x)` or `rowSums(x)` over `apply(x, ., sum)`
#'
#' [colSums()] and [rowSums()] are clearer and more performant alternatives to
#' `apply(x, 2, sum)` and `apply(x, 1, sum)` respectively in the case of 2D
#' arrays, or matrices
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "apply(x, 1, sum)",
#'   linters = matrix_apply_linter()
#' )
#'
#' lint(
#'   text = "apply(x, 2, sum)",
#'   linters = matrix_apply_linter()
#' )
#'
#' lint(
#'   text = "apply(x, 2, sum, na.rm = TRUE)",
#'   linters = matrix_apply_linter()
#' )
#'
#' lint(
#'   text = "apply(x, 2:4, sum)",
#'   linters = matrix_apply_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "rowSums(x)",
#'   linters = matrix_apply_linter()
#' )
#'
#' lint(
#'   text = "colSums(x)",
#'   linters = matrix_apply_linter()
#' )
#'
#' lint(
#'   text = "colSums(x, na.rm = TRUE)",
#'   linters = matrix_apply_linter()
#' )
#'
#' lint(
#'   text = "rowSums(colSums(x), dims = 3)",
#'   linters = matrix_apply_linter()
#' )
#'
#'
#' @evalRd rd_tags("matrix_apply_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
matrix_apply_linter <- function() {
  # mean() and sum() have very different signatures so we treat them separately.
  # sum() takes values to sum over via ..., has just one extra argument and is not a generic
  # mean() is a generic, takes values to average via a single argument, and can have extra arguments
  #
  # Currently supported values for MARGIN: scalar numeric and vector of contiguous values created by : (OP-COLON)
  sums_xpath <- "
  following-sibling::expr[
    NUM_CONST or OP-COLON/preceding-sibling::expr[NUM_CONST]/following-sibling::expr[NUM_CONST]
    and (position() = 2)
  ]
    /following-sibling::expr[
      SYMBOL[text() = 'sum']
      and (position() = 1)
    ]
    /parent::expr
  "

  # Since mean() is a generic, we make sure that we only lint cases with arguments
  # supported by colMeans() and rowMeans(), i.e., na.rm
  means_xpath <- "
  following-sibling::expr[
    NUM_CONST or OP-COLON/preceding-sibling::expr[NUM_CONST]/following-sibling::expr[NUM_CONST]
    and (position() = 2)
  ]
    /following-sibling::expr[
      SYMBOL[text() = 'mean']
      and (position() = 1)
    ]
    /parent::expr[
      count(expr) = 4
      or (count(expr) = 5 and SYMBOL_SUB[text() = 'na.rm'])
    ]
  "

  xpath <- glue("{sums_xpath} | {means_xpath}")

  # This doesn't handle the case when MARGIN and FUN are named and in a different position
  # but this should be relatively rate
  variable_xpath <- "expr[position() = 2]"
  margin_xpath <- "expr[position() = 3]"
  fun_xpath <- "expr[position() = 4]"

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls("apply")
    bad_expr <- xml_find_all_(xml_calls, xpath)
    bad_expr <- strip_comments_from_subtree(bad_expr)

    variable <- xml_text(xml_find_all_(bad_expr, variable_xpath))

    fun <- xml_text(xml_find_all_(bad_expr, fun_xpath))
    fun <- tools::toTitleCase(fun)

    margin <- xml_find_all_(bad_expr, margin_xpath)

    narm_val <- xml_text(
      xml_find_first_(bad_expr, "SYMBOL_SUB[text() = 'na.rm']/following-sibling::expr")
    )

    recos <- Map(craft_colsums_rowsums_msg, variable, margin, fun, narm_val)

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = sprintf("Use %1$s rather than %2$s", recos, get_r_string(bad_expr)),
      type = "warning"
    )
  })
}

craft_colsums_rowsums_msg <- function(variable, margin, fun, narm_val) {
  if (is.na(xml_find_first_(margin, "OP-COLON"))) {
    l1 <- xml_text(margin)
    l2 <- NULL
  } else {
    l1 <- xml_text(xml_find_first_(margin, "expr[1]"))
    l2 <- xml_text(xml_find_first_(margin, "expr[2]"))
  }

  # See #1764 for details about various cases. In short:
  # - If apply(., 1:l2, sum) -> rowSums(., dims = l2)
  # - If apply(., l1:l2, sum) -> rowSums(colSums(., dims = l1 - 1), dims = l2 - l1 + 1)
  # - This last case can be simplified to a simple colSums() call if l2 = length(dim(.))
  # - dims argument can be dropped if equals to 1. This notably is the case for matrices
  if (is.null(l2)) {
    l2 <- l1
  }

  # We don't want warnings when converted as NAs
  l1 <- suppressWarnings(as.integer(re_substitutes(l1, "L$", "")))
  l2 <- suppressWarnings(as.integer(re_substitutes(l2, "L$", "")))

  if (!is.na(narm_val)) {
    narm <- glue(", na.rm = {narm_val}")
  } else {
    narm <- ""
  }

  if (identical(l1, 1L)) {
    reco <- glue("row{fun}s({variable}{narm}, dims = {l2})")
  } else {
    reco <- glue(
      "row{fun}s(col{fun}s({variable}{narm}, dims = {l1 - 1}), dims = {l2 - l1 + 1})",
      " or ",
      "col{fun}s({variable}{narm}, dims = {l1 - 1}) if {variable} has {l2} dimensions"
    )
  }

  # It's easier to remove this after the fact, rather than having never ending if/elses
  reco <- gsub(", dims = 1", "", reco, fixed = TRUE)

  reco
}

# Parse namespace files and return imports exports, methods
namespace_imports <- function(path = find_package(".")) {
  namespace_data <- tryCatch(
    parseNamespaceFile(basename(path), package.lib = file.path(path, "..")),
    error = \(e) NULL
  )

  if (length(namespace_data$imports) == 0L) {
    return(empty_namespace_data())
  }

  do.call(rbind, lapply(namespace_data$imports, safe_get_exports))
}

# this loads the namespaces, but is the easiest way to do it
# test package availability to avoid failing out as in #1360
#   typically, users are running this on their own package directories and thus
#   will have the namespace dependencies installed, but we can't guarantee this.
safe_get_exports <- function(ns) {
  # check package exists for both import(x) and importFrom(x, y) usages
  if (!requireNamespace(ns[[1L]], quietly = TRUE)) {
    return(empty_namespace_data())
  }

  # importFrom directives appear as list(ns, imported_funs)
  if (length(ns) > 1L) {
    return(data.frame(pkg = ns[[1L]], fun = ns[[2L]]))
  }

  # relevant only if there are any exported objects
  fun <- getNamespaceExports(ns)
  if (length(fun) > 0L) {
    data.frame(pkg = ns, fun = fun)
  }
}

empty_namespace_data <- function() {
  data.frame(pkg = character(), fun = character())
}

# filter namespace_imports() for S3 generics
# this loads all imported namespaces
imported_s3_generics <- function(ns_imports) {
  # `NROW()` for the `NULL` case of 0-export dependencies (cf. #1503)
  is_generic <- vapply(
    seq_len(NROW(ns_imports)),
    function(i) {
      fun_obj <- get(ns_imports$fun[i], envir = asNamespace(ns_imports$pkg[i]))
      is.function(fun_obj) && is_s3_generic(fun_obj)
    },
    logical(1L)
  )

  ns_imports[is_generic, ]
}

exported_s3_generics <- function(path = find_package(".")) {
  namespace_data <- tryCatch(
    parseNamespaceFile(basename(path), package.lib = file.path(path, "..")),
    error = \(e) NULL
  )

  if (length(namespace_data$S3methods) == 0L || nrow(namespace_data$S3methods) == 0L) {
    return(empty_namespace_data())
  }

  data.frame(pkg = basename(path), fun = unique(namespace_data$S3methods[, 1L]))
}

is_s3_generic <- function(fun) {
  # Inspired by `utils::isS3stdGeneric`, though it will detect functions that
  # have `useMethod()` in places other than the first expression.
  bdexpr <- body(fun)
  while (is.call(bdexpr) && bdexpr[[1L]] == "{") bdexpr <- bdexpr[[length(bdexpr)]]
  ret <- is.call(bdexpr) && identical(bdexpr[[1L]], as.name("UseMethod"))
  if (ret) {
    names(ret) <- bdexpr[[2L]]
  }
  ret
}

.base_s3_generics <- unique(c(
  names(.knownS3Generics),
  .S3_methods_table[, 1L],
  # Contains S3 generic groups, see ?base::groupGeneric and src/library/base/R/zzz.R
  ls(.GenericArgsEnv)
))

#' Repeat linter
#'
#' Check that `while (TRUE)` is not used for infinite loops. While this is valid
#' R code, using `repeat {}` is more explicit.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "while (TRUE) { }",
#'   linters = repeat_linter()
#' )
#'
#'
#' # okay
#' lint(
#'   text = "repeat { }",
#'   linters = repeat_linter()
#' )
#'
#' @evalRd rd_tags("repeat_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
repeat_linter <- function() {
  xpath <- "//WHILE[following-sibling::expr[1]/NUM_CONST[text() = 'TRUE']]/parent::expr"

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    lints <- xml_find_all_(xml, xpath)

    xml_nodes_to_lints(
      lints,
      source_expression = source_expression,
      lint_message = "Use 'repeat' instead of 'while (TRUE)' for infinite loops."
    )
  })
}

#' Block comparison of class with `==`
#'
#' Usage like `class(x) == "character"` is prone to error since class in R
#'   is in general a vector. The correct version for S3 classes is [inherits()]:
#'   `inherits(x, "character")`. Often, class `k` will have an `is.` equivalent,
#'   for example [is.character()] or [is.data.frame()].
#'
#' Similar reasoning applies for `class(x) %in% "character"`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'is_lm <- class(x) == "lm"',
#'   linters = class_equals_linter()
#' )
#'
#' lint(
#'   text = 'if ("lm" %in% class(x)) is_lm <- TRUE',
#'   linters = class_equals_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'is_lm <- inherits(x, "lm")',
#'   linters = class_equals_linter()
#' )
#'
#' lint(
#'   text = 'if (inherits(x, "lm")) is_lm <- TRUE',
#'   linters = class_equals_linter()
#' )
#'
#' @evalRd rd_tags("class_equals_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
class_equals_linter <- function() {
  xpath <- "
  parent::expr
    /parent::expr[
      not(preceding-sibling::OP-LEFT-BRACKET)
      and (EQ or NE or SPECIAL[text() = '%in%'])
    ]
  "

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls("class")
    bad_expr <- xml_find_all_(xml_calls, xpath)

    operator <- xml_find_chr_(bad_expr, "string(*[2])")
    lint_message <- paste0(
      "Use inherits(x, 'class-name'), is.<class> for S3 classes, ",
      "or is(x, 'S4Class') for S4 classes, ",
      "instead of comparing class(x) with ", operator, "."
    )
    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = lint_message,
      type = "warning"
    )
  })
}

#' Trailing whitespace linter
#'
#' Check that there are no space characters at the end of source lines.
#'
#' @param allow_empty_lines Suppress lints for lines that contain only whitespace.
#' @param allow_in_strings Suppress lints for trailing whitespace in string constants.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x <- 1.2  ",
#'   linters = trailing_whitespace_linter()
#' )
#'
#' code_lines <- "a <- TRUE\n \nb <- FALSE"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = trailing_whitespace_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x <- 1.2",
#'   linters = trailing_whitespace_linter()
#' )
#'
#' lint(
#'   text = "x <- 1.2  # comment about this assignment",
#'   linters = trailing_whitespace_linter()
#' )
#'
#' code_lines <- "a <- TRUE\n \nb <- FALSE"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = trailing_whitespace_linter(allow_empty_lines = TRUE)
#' )
#'
#' @evalRd rd_tags("trailing_whitespace_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
trailing_whitespace_linter <- function(allow_empty_lines = FALSE, allow_in_strings = TRUE) {
  Linter(linter_level = "file", function(source_expression) {
    res <- re_matches(
      source_expression$file_lines,
      rex(blanks, end),
      locations = TRUE
    )

    if (isTRUE(allow_empty_lines)) {
      bad_lines <- which(res$start > 1L)
    } else {
      bad_lines <- which(!is.na(res$start))
    }

    if (isTRUE(allow_in_strings) && !is.null(source_expression$full_xml_parsed_content)) {
      all_str_consts <- xml_find_all_(source_expression$full_xml_parsed_content, "//STR_CONST")
      start_lines <- as.integer(xml_attr_(all_str_consts, "line1"))
      end_lines <- as.integer(xml_attr_(all_str_consts, "line2"))

      is_in_str <- vapply(bad_lines, \(ln) any(start_lines <= ln & ln < end_lines), logical(1L))
      bad_lines <- bad_lines[!is_in_str]
    }

    lapply(
      bad_lines,
      function(line) {
        Lint(
          filename = source_expression$filename,
          line_number = line,
          column_number = res$start[[line]],
          type = "style",
          message = "Remove trailing whitespace.",
          line = source_expression$file_lines[[line]],
          ranges = list(c(res$start[[line]], res$end[[line]]))
        )
      }
    )
  })
}

# nolint start: assign.
a = 2
# nolint end

# nolint start: s. warn (and lint) because of non-unique identifier
x <- 42; y <- 2
# nolint end

# nolint start: bogus_linter. warn because of nonexistent identifier

# nolint end

# nolint: hocus_pocus, bogus.

#' Line length linter
#'
#' Check that the line length of both comments and code is less than `length`.
#'
#' @param length Maximum line length allowed. Default is `80L` (Hollerith limit).
#' @param ignore_string_bodies Logical, default `FALSE`. If `TRUE`, the contents
#'   of string literals are ignored. The quotes themselves are included, so this
#'   mainly affects wide multiline strings, e.g. SQL queries.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = strrep("x", 23L),
#'   linters = line_length_linter(length = 20L)
#' )
#'
#' # the trailing ' is counted towards line length, so this still lints
#' lint(
#'   text = "'a long single-line string'",
#'   linters = line_length_linter(length = 15L, ignore_string_bodies = TRUE)
#' )
#'
#' lines <- paste(
#'   "query <- '",
#'   "  SELECT *",
#'   "  FROM MyTable",
#'   "  WHERE profit > 0",
#'   "'",
#'   sep = "\n"
#' )
#' writeLines(lines)
#' lint(
#'   text = lines,
#'   linters = line_length_linter(length = 10L)
#' )
#'
#' # okay
#' lint(
#'   text = strrep("x", 21L),
#'   linters = line_length_linter(length = 40L)
#' )
#'
#' lines <- paste(
#'   "paste(",
#'   "  'a long',",
#'   "  'single-line',",
#'   "  'string'",
#'   ")",
#'   sep = "\n"
#' )
#' writeLines(lines)
#' lint(
#'   text = lines,
#'   linters = line_length_linter(length = 15L, ignore_string_bodies = TRUE)
#' )
#'
#' lines <- paste(
#'   "query <- '",
#'   "  SELECT *",
#'   "  FROM MyTable",
#'   "  WHERE profit > 0",
#'   "'",
#'   sep = "\n"
#' )
#' writeLines(lines)
#' lint(
#'   text = lines,
#'   linters = line_length_linter(length = 10L, ignore_string_bodies = TRUE)
#' )
#'
#' @evalRd rd_tags("line_length_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#long-lines>
#' @export
line_length_linter <- function(length = 80L, ignore_string_bodies = FALSE) {
  general_msg <- paste("Lines should not be more than", length, "characters.")

  Linter(linter_level = "file", function(source_expression) {
    # Only go over complete file
    line_lengths <- nchar(source_expression$file_lines)
    long_lines <- which(line_lengths > length)

    if (ignore_string_bodies) {
      in_string_body_idx <-
        is_in_string_body(source_expression$full_parsed_content, length, long_lines)
      long_lines <- long_lines[!in_string_body_idx]
    }

    Map(
      function(long_line, line_length) {
        Lint(
          filename = source_expression$filename,
          line_number = long_line,
          column_number = length + 1L,
          type = "style",
          message = paste(general_msg, "This line is", line_length, "characters."),
          line = source_expression$file_lines[long_line],
          ranges = list(c(1L, line_length))
        )
      },
      long_lines,
      line_lengths[long_lines]
    )
  })
}

is_in_string_body <- function(parse_data, max_length, long_idx) {
  str_idx <- parse_data$token == "STR_CONST"
  if (!any(str_idx)) {
    return(rep(FALSE, length(long_idx)))
  }
  str_data <- parse_data[str_idx, ]
  if (all(str_data$line1 == str_data$line2)) {
    return(rep(FALSE, length(long_idx)))
  }
  # right delimiter just ends at 'col2', but 'col1' takes some sleuthing
  str_data$line1_width <- nchar(vapply(
    strsplit(str_data$text, "\n", fixed = TRUE),
    \(x) x[1L],
    FUN.VALUE = character(1L),
    USE.NAMES = FALSE
  ))
  str_data$col1_end <- str_data$col1 + str_data$line1_width
  vapply(
    long_idx,
    function(line) {
      # strictly inside a multi-line string body
      if (any(str_data$line1 < line & str_data$line2 > line)) {
        return(TRUE)
      }
      on_line1_idx <- str_data$line1 == line
      if (any(on_line1_idx)) {
        return(max(str_data$col1_end[on_line1_idx]) <= max_length)
      }
      # use parse data to capture possible trailing expressions on this line
      on_line2_idx <- parse_data$line2 == line
      any(on_line2_idx) && max(parse_data$col2[on_line2_idx]) <= max_length
    },
    logical(1L)
  )
}

#' Undesirable function linter
#'
#' Report the use of undesirable functions and suggest an alternative.
#'
#' @param fun Character vector of undesirable function names. Input can be any of three types:
#'   - Unnamed entries must be a character string specifying an undesirable function.
#'   - For named entries, the name specifies the undesirable function.
#'     + If the entry is a character string, it is used as a description of
#'       why a given function is undesirable
#'     + Otherwise, entries should be missing (`NA`)
#'   A generic message that the named function is undesirable is used if no
#'     specific description is provided.
#'   Input can also be a list of character strings for convenience.
#'
#'   Defaults to [default_undesirable_functions]. To make small customizations to this list,
#'   use [modify_defaults()].
#' @param symbol_is_undesirable Whether to consider the use of an undesirable function
#'   name as a symbol undesirable or not.
#'
#' @examples
#' # defaults for which functions are considered undesirable
#' names(default_undesirable_functions)
#'
#' # will produce lints
#' lint(
#'   text = "sapply(x, mean)",
#'   linters = undesirable_function_linter()
#' )
#'
#' lint(
#'   text = "log10(x)",
#'   linters = undesirable_function_linter(fun = c("log10" = NA))
#' )
#'
#' lint(
#'   text = "log10(x)",
#'   linters = undesirable_function_linter(fun = c("log10" = "use log()"))
#' )
#'
#' lint(
#'   text = 'dir <- "path/to/a/directory"',
#'   linters = undesirable_function_linter(fun = c("dir" = NA))
#' )
#'
#'
#' lint(
#'   text = 'dir <- "path/to/a/directory"',
#'   linters = undesirable_function_linter(fun = "dir")
#' )
#'
#' # okay
#' lint(
#'   text = "vapply(x, mean, FUN.VALUE = numeric(1))",
#'   linters = undesirable_function_linter()
#' )
#'
#' lint(
#'   text = "log(x, base = 10)",
#'   linters = undesirable_function_linter(fun = c("log10" = "use log()"))
#' )
#'
#' lint(
#'   text = 'dir <- "path/to/a/directory"',
#'   linters = undesirable_function_linter(fun = c("dir" = NA), symbol_is_undesirable = FALSE)
#' )
#'
#' lint(
#'   text = 'dir <- "path/to/a/directory"',
#'   linters = undesirable_function_linter(fun = "dir", symbol_is_undesirable = FALSE)
#' )
#'
#' @evalRd rd_tags("undesirable_function_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
undesirable_function_linter <- function(fun = default_undesirable_functions,
                                        symbol_is_undesirable = TRUE) {
  if (is.list(fun)) fun <- unlist(fun)
  if (!is.logical(symbol_is_undesirable)) {
    cli_abort("{.arg symbol_is_undesirable} must be a logical, not {.obj_type_friendly {symbol_is_undesirable}}.")
  }
  # allow (uncoerced->implicitly logical) 'NA'
  if (length(fun) == 0L || !(is.character(fun) || all(is.na(fun)))) {
    cli_abort("{.arg fun} must be a non-empty character vector.")
  }

  nm <- names2(fun)
  implicit_idx <- !nzchar(nm)
  if (any(implicit_idx)) {
    names(fun)[implicit_idx] <- fun[implicit_idx]
    is.na(fun) <- implicit_idx
  }
  if (anyNA(names(fun))) {
    missing_idx <- which(is.na(names(fun))) # nolint: object_usage_linter. False positive.
    cli_abort(paste(
      "Unnamed elements of {.arg fun} must not be missing,",
      "but {.val {missing_idx}} {qty(length(missing_idx))} {?is/are}."
    ))
  }

  xp_condition <- xp_and(
    paste0(
      "not(parent::expr/preceding-sibling::expr[last()][SYMBOL_FUNCTION_CALL[",
      xp_text_in_table(c("library", "require")),
      "]])"
    ),
    "not(parent::expr[OP-DOLLAR or OP-AT])"
  )

  if (symbol_is_undesirable) {
    symbol_xpath <- glue("//SYMBOL[({xp_text_in_table(names(fun))}) and {xp_condition}]")
  }
  xpath <- glue("SYMBOL_FUNCTION_CALL[{xp_condition}]")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content
    xml_calls <- source_expression$xml_find_function_calls(names(fun))

    matched_nodes <- xml_find_all_(xml_calls, xpath)
    if (symbol_is_undesirable) {
      matched_nodes <- combine_nodesets(matched_nodes, xml_find_all_(xml, symbol_xpath))
    }

    fun_names <- get_r_string(matched_nodes)

    msgs <- vapply(
      stats::setNames(nm = unique(fun_names)),
      function(fun_name) {
        msg <- sprintf('Avoid undesirable function "%s".', fun_name)
        alternative <- fun[[fun_name]]
        if (!is.na(alternative)) {
          msg <- paste(msg, sprintf("As an alternative, %s.", alternative))
        }
        msg
      },
      character(1L)
    )

    xml_nodes_to_lints(
      matched_nodes,
      source_expression = source_expression,
      lint_message = unname(msgs[fun_names])
    )
  })
}

#' Parenthesis before body linter
#'
#' Check that there is a space between right parenthesis and a body expression.
#'
#' @evalRd rd_tags("paren_body_linter")
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "function(x)x + 1",
#'   linters = paren_body_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "function(x) x + 1",
#'   linters = paren_body_linter()
#' )
#'
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#parentheses>
#' @export
paren_body_linter <- make_linter_from_xpath(
  # careful to do recursive search to the less common OP-RIGHT-PAREN
  #   and forcond nodes (vs. //expr) for performance -- there can
  #   be O(100K) <expr> nodes but in all but pathological examples,
  #   these other nodes will only be a small fraction of this amount.
  # note also that <forcond> only has one following-sibling::expr.
  xpath = "
  //OP-RIGHT-PAREN[
    @end = following-sibling::expr[1]/@start - 1
    and @line1 = following-sibling::expr[1]/@line1
    and (
      preceding-sibling::FUNCTION
      or preceding-sibling::OP-LAMBDA
      or preceding-sibling::IF
      or preceding-sibling::WHILE
    )
  ]
  |
  //forcond[
    @line1 = following-sibling::expr/@line2
    and OP-RIGHT-PAREN/@col1 = following-sibling::expr/@col1 - 1
  ]
    /OP-RIGHT-PAREN
  ",
  lint_message = "Put a space between a right parenthesis and a body expression.",
  type = "style"
)

#' Commented code linter
#'
#' Check that there is no commented code outside roxygen blocks.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "# x <- 1",
#'   linters = commented_code_linter()
#' )
#'
#' lint(
#'   text = "x <- f() # g()",
#'   linters = commented_code_linter()
#' )
#'
#' lint(
#'   text = "x + y # + z[1, 2]",
#'   linters = commented_code_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x <- 1; x <- f(); x + y",
#'   linters = commented_code_linter()
#' )
#'
#' lint(
#'   text = "#' x <- 1",
#'   linters = commented_code_linter()
#' )
#'
#' @evalRd rd_tags("commented_code_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
commented_code_linter <- function() {
  ops <- list(
    "+",
    # "-",
    "=",
    "==",
    "!=",
    "<=",
    ">=",
    "<-",
    "<<-",
    "<",
    ">",
    "->",
    "->>",
    "%%",
    "/",
    "^",
    "*",
    "**",
    "|",
    "||",
    "&",
    "&&",
    rex("%", except_any_of("%"), "%")
  )

  code_candidate_regex <- rex(
    some_of("#"),
    any_spaces,
    capture(
      name = "code",
      anything,
      or(
        some_of("{}[]"), # code-like parentheses
        or(ops), # any operator
        group(graphs, "(", anything, ")"), # a function call
        group("!", alphas) # a negation
      ),
      anything
    )
  )

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    all_comment_nodes <- xml_find_all_(xml, "//COMMENT")
    all_comments <- xml_text(all_comment_nodes)
    code_candidates <- re_matches(all_comments, code_candidate_regex, global = FALSE, locations = TRUE)
    extracted_code <- code_candidates[, "code"]
    # ignore trailing ',' or pipes ('|>', '%>%') when testing for parsability
    extracted_code <- re_substitutes(extracted_code, rex(or(",", "|>", "%>%"), any_spaces, end), "")
    extracted_code <- re_substitutes(extracted_code, rex(start, any_spaces, ","), "")

    is_parsable <- which(vapply(extracted_code, parsable, logical(1L)))

    lint_list <- xml_nodes_to_lints(
      all_comment_nodes[is_parsable],
      source_expression = source_expression,
      lint_message = "Remove commented code."
    )

    # Location info needs updating
    for (i in seq_along(lint_list)) {
      rng <- lint_list[[i]]$ranges[[1L]]

      rng[2L] <- rng[1L] + code_candidates[is_parsable[i], "code.end"] - 1L
      rng[1L] <- rng[1L] + code_candidates[is_parsable[i], "code.start"] - 1L

      lint_list[[i]]$column_number <- rng[1L]
      lint_list[[i]]$ranges <- list(rng)
    }

    lint_list
  })
}

# is given text parsable
parsable <- function(x) {
  if (anyNA(x)) {
    return(FALSE)
  }
  res <- try_silently(parse(text = x))
  !inherits(res, "try-error")
}

#' Pipe call linter
#'
#' Force explicit calls in magrittr pipes, e.g., `1:3 %>% sum()` instead of `1:3 %>% sum`.
#' Note that native pipe always requires a function call, i.e. `1:3 |> sum` will produce an error.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "1:3 %>% mean %>% as.character",
#'   linters = pipe_call_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "1:3 %>% mean() %>% as.character()",
#'   linters = pipe_call_linter()
#' )
#'
#' @evalRd rd_tags("pipe_call_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
pipe_call_linter <- function() {
  # NB: the text() here shows up as %&gt;% but that's not matched, %>% is
  # NB: use *[1][self::SYMBOL] to ensure the first element is SYMBOL, otherwise
  #       we include expressions like x %>% .$col
  pipes <- setdiff(magrittr_pipes, "%$%")
  xpath <- glue("//SPECIAL[{ xp_text_in_table(pipes) }]/following-sibling::expr[*[1][self::SYMBOL]]")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)
    pipe <- xml_text(xml_find_first_(bad_expr, "preceding-sibling::SPECIAL[1]"))

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message =
        sprintf("Use explicit calls in magrittr pipes, i.e., `a %1$s foo` should be `a %1$s foo()`.", pipe),
      type = "warning"
    )
  })
}

#' Recommend direct usage of `data.frame()` to create a data.frame from a list
#'
#' [base::list2DF()] is the preferred way to turn a list of columns into a data.frame.
#'   Note that it doesn't support recycling; if that's required, use [data.frame()].
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "do.call(cbind.data.frame, x)",
#'   linters = list2df_linter()
#' )
#'
#' lint(
#'   text = "do.call('cbind.data.frame', x)",
#'   linters = list2df_linter()
#' )
#'
#' lint(
#'   text = "do.call(cbind.data.frame, list(a = 1, b = 1:10))",
#'   linters = list2df_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "list2df(x)",
#'   linters = list2df_linter()
#' )
#'
#' lint(
#'   text = "data.frame(list(a = 1, b = 1:10))",
#'   linters = list2df_linter()
#' )
#'
#' @evalRd rd_tags("list2df_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
list2df_linter <- function() {

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls("do.call")

    xml_calls_nolambda <- xml_find_all_(
      xml_calls,
      "./following-sibling::expr[1][SYMBOL or STR_CONST]"
    )

    bad_expr <- xml_calls_nolambda[
      get_r_string(xml_calls_nolambda) == "cbind.data.frame"
    ]

    xml_nodes_to_lints(
      # nolint next: undesirable_function_name_linter.
      xml_parent(bad_expr),
      source_expression = source_expression,
      lint_message = paste(
        "Use `list2DF(lst)` instead of `do.call(cbind.data.frame, lst)`.",
        "If recycling is required, use `data.frame(lst)`."
      ),
      type = "warning"
    )
  })
}

make_linter_from_regex <- function(regex,
                                   lint_type,
                                   lint_msg) {
  function() { # nocov: only run at namespace load time
    Linter(linter_level = "file", function(source_expression) {
      all_matches <- re_matches(
        source_expression[["file_lines"]],
        regex,
        locations = TRUE,
        global = FALSE
      )
      all_matches <- all_matches[!is.na(all_matches$start), ]
      all_matches$line_number <- as.integer(rownames(all_matches))

      matches_by_row <- split(all_matches, seq_len(nrow(all_matches)))

      lints <- lapply(matches_by_row, function(.match) {
        if (is_match_covered(.match, source_expression)) {
          return()
        }
        Lint(
          filename = source_expression[["filename"]],
          line_number = .match$line_number,
          type = lint_type,
          message = lint_msg,
          line = source_expression[["file_lines"]][[rownames(.match)]],
          ranges = list(c(.match$start, .match$end))
        )
      })
      lints[lengths(lints) > 0L]
    })
  } # nocov: ditto opening brace
}

#' Determine if a regex match is covered by an expression in a source_expression
#'
#' @param match The position where a regex match was observed.
#'   It must have the following elements: `"start"`, `"end"`, and `"line_number"`.
#' @param source_expression A source_expression.
#' @param token_type Restrict analysis to tokens of this type, for example,
#'   with `token_type = "STR_CONST"` you can check that a regex match occurs
#'   within a string.
#' @noRd
is_match_covered <- function(match, source_expression, token_type = "STR_CONST") {
  line_number <- match$line_number
  pc <- source_expression[["full_parsed_content"]]
  if (!is.null(token_type)) {
    pc <- pc[pc[["token"]] == token_type, ]
  }
  covering_rows <- pc[["line1"]] <= line_number & pc[["line2"]] >= line_number
  pc_cover <- pc[covering_rows, ]

  any_single_line_covers <- function() {
    x <- pc_cover[pc_cover[["line1"]] == pc_cover[["line2"]], ]
    any(
      x[["col1"]] <= match[["start"]] & x[["col2"]] >= match[["end"]]
    )
  }

  any_multi_line_covers <- function() {
    x <- pc_cover[pc_cover[["line1"]] < pc_cover[["line2"]], ]
    any(
      (x[["line1"]] < line_number & x[["line2"]] > line_number) |
        (x[["line1"]] == line_number & x[["col1"]] <= match[["start"]]) |
        (x[["line2"]] == line_number & x[["col2"]] >= match[["end"]])
    )
  }

  any_single_line_covers() || any_multi_line_covers()
}

#' Lint expectation
#'
#' These are expectation functions to test specified linters on sample code in the `testthat` testing framework.
#' * `expect_lint` asserts that specified lints are generated.
#' * `expect_no_lint` asserts that no lints are generated.
#'
#' @param content A character vector for the file content to be linted, each vector element representing a line of
#'   text.
#' @param checks Checks to be performed:
#' \describe{
#'   \item{NULL}{check that no lints are returned.}
#'   \item{single string or regex object}{check that the single lint returned has a matching message.}
#'   \item{named list}{check that the single lint returned has fields that match. Accepted fields are the same as those
#'     taken by [Lint()].}
#'   \item{list of named lists}{for each of the multiple lints returned, check that it matches the checks in the
#'     corresponding named list (as described in the point above).}
#' }
#' Named vectors are also accepted instead of named lists, but this is a compatibility feature that
#'   is not recommended for new code.
#' @param ... Arguments passed to [lint()], e.g. the linters or cache to use.
#' @param file If not `NULL`, read content from the specified file rather than from `content`.
#' @param language Temporarily override Rs `LANGUAGE` envvar, controlling localization of base R error messages.
#'   This makes testing them reproducible on all systems irrespective of their native R language setting.
#' @param ignore_order Logical, default `FALSE`. If `TRUE`, the order of the `checks` does not matter, e.g.
#'   lints with higher line numbers can come before those with lower line numbers, and the order of linters
#'   affecting the same line is also irrelevant.
#' @return `NULL`, invisibly.
#' @examples
#' # no expected lint
#' expect_no_lint("a", trailing_blank_lines_linter())
#'
#' # one expected lint
#' expect_lint("a\n", "trailing blank", trailing_blank_lines_linter())
#' expect_lint("a\n", list(message = "trailing blank", line_number = 2), trailing_blank_lines_linter())
#'
#' # several expected lints
#' expect_lint("a\n\n", list("trailing blank", "trailing blank"), trailing_blank_lines_linter())
#' expect_lint(
#'   "a\n\n",
#'   list(
#'     list(message = "trailing blank", line_number = 2),
#'     list(message = "trailing blank", line_number = 3)
#'   ),
#'   trailing_blank_lines_linter()
#' )
#' @export
expect_lint <- function(content, checks, ..., file = NULL, language = "en", ignore_order = FALSE) {
  require_testthat()

  old_lang <- set_lang(language)
  on.exit(reset_lang(old_lang))

  if (is.null(file)) on.exit(unlink(file), add = TRUE)
  file <- maybe_write_content(file, content) # NB: the lint consistency fuzz suite anchors here.

  lints <- lint(file, ...)
  n_lints <- length(lints)
  lint_str <- if (n_lints) paste(c("", lints), collapse = "\n") else ""

  wrong_number_fmt <- "got %d lints instead of %d%s"
  if (is.null(checks)) {
    if (n_lints != 0L) {
      fail_msg <- sprintf(wrong_number_fmt, n_lints, 0L, lint_str)
      return(testthat::fail(fail_msg))
    }
    return(testthat::succeed())
  }

  if (!is.list(checks) || !is.null(names(checks))) { # vector or named list
    checks <- list(checks)
  }
  checks[] <- lapply(checks, fix_names, "message")

  if (n_lints != length(checks)) {
    fail_msg <- sprintf(wrong_number_fmt, n_lints, length(checks), lint_str)
    return(testthat::fail(fail_msg))
  }

  if (ignore_order) {
    lint_order <- with(as.data.frame(lints), order(line_number, column_number, linter))
    lints <- lints[lint_order]

    check_order <- order(
      vapply(checks, \(x) x$line_number %||% 0L, FUN.VALUE = integer(1L)),
      vapply(checks, \(x) x$column_number %||% 0L, FUN.VALUE = integer(1L)),
      vapply(checks, \(x) x$linter %||% "", FUN.VALUE = character(1L))
    )
    checks <- checks[check_order]
  }

  expect_lint_impl_(lints, checks)

  invisible(lints)
}

#' NB: must _not_ succeed(), should only fail() or abort()
#' @noRd
expect_lint_impl_ <- function(lints, checks, trace_env = rlang::caller_env()) {
  itr <- 0L
  # valid fields are those from Lint(), plus 'linter'
  lint_fields <- c(names(formals(Lint)), "linter")

  for (i in seq_along(lints)) {
    lint <- lints[[i]]
    check <- checks[[i]]

    itr <- itr + 1L

    for (field in names(check)) {
      if (!field %in% lint_fields) {
        cli_abort(c(
          x = "Check {.val {itr}} has an invalid field: {.field {field}}.",
          i = "Valid fields are: {.field {lint_fields}}."
        ))
      }
      check_field <- check[[field]]
      value <- lint[[field]]
      ok <- if (field == "message") {
        re_matches_logical(value, check_field)
      } else {
        isTRUE(all.equal(value, check_field))
      }
      if (!ok) {
        fail_msg <- sprintf(
          "check #%d: %s %s did not match %s",
          # deparse ensures that NULL, list(), etc are handled gracefully
          itr, field, deparse(value), deparse(check_field)
        )
        return(testthat::fail(fail_msg, trace_env = trace_env))
      }
    }
  }

  testthat::succeed()
}

#' @rdname expect_lint
#' @export
expect_no_lint <- function(content, ..., file = NULL, language = "en") {
  require_testthat()
  # nolint next: expect_no_lint_linter.
  expect_lint(content, NULL, ..., file = file, language = language)
}

maybe_write_content <- function(file, lines) {
  if (!is.null(file)) {
    return(file)
  }
  tmp <- tempfile()
  con <- file(tmp, encoding = "UTF-8")
  on.exit(close(con))
  writeLines(lines, con = con, sep = "\n")
  tmp
}

#' Test that the package is lint free
#'
#' This function is a thin wrapper around lint_package that simply tests there are no
#' lints in the package. It can be used to ensure that your tests fail if the package
#' contains lints.
#'
#' @param ... arguments passed to [lint_package()]
#' @export
expect_lint_free <- function(...) {
  require_testthat()

  testthat::skip_on_cran()
  testthat::skip_on_covr()

  lints <- lint_package(...)
  has_lints <- length(lints) > 0L

  lint_output <- NULL
  if (has_lints) {
    lint_output <- format(lints)
  }

  if (has_lints) {
    return(testthat::fail(paste0("Not lint free\n", lint_output)))
  }
  testthat::succeed()
}

# Helper function to check if testthat is installed.
require_testthat <- function() {
  parent_call <- sys.call(-1L)[[1L]]
  # supported: foo() or lintr::foo(). Undefined behavior otherwise.
  # nolint next: object_usage_linter. TODO(#2252): Remove this.
  name <- as.character(if (is.name(parent_call)) parent_call else parent_call[[3L]])
  if (!requireNamespace("testthat", quietly = TRUE)) {
    cli_abort(
      "{.fun {name}} is designed to work within the {.pkg testthat} testing framework, which could not be loaded."
    )
  }
}

test_that("default styles are linted correctly", {
  linters <- list(
    symbols_linter = object_name_linter("symbols"),
    CamelCase_linter = object_name_linter("CamelCase"),
    camelCase_linter = object_name_linter("camelCase"),
    snake_case_linter = object_name_linter("snake_case"),
    SNAKE_CASE_linter = object_name_linter("SNAKE_CASE"),
    dotted.case_linter = object_name_linter("dotted.case"),
    lowercase_linter = object_name_linter("lowercase"),
    UPPERCASE_linter = object_name_linter("UPPERCASE")
  )
  ok_lines <- list(
    CamelCase_linter = c(3L, 10L, 11L, 13L, 21L, 27L),
    SNAKE_CASE_linter = c(3L, 7L, 11L, 13L, 15L, 18L, 24L, 27L),
    UPPERCASE_linter = c(3L, 11L, 13L, 27L),
    camelCase_linter = c(1L, 2L, 8L, 9L, 12L, 22L, 26L),
    dotted.case_linter = c(1L, 2L, 8L, 12L, 19L, 25L, 26L),
    lowercase_linter = c(1L, 2L, 8L, 12L, 26L),
    snake_case_linter = c(1L, 2L, 6L, 8L, 12L, 14L, 17L, 23L, 26L),
    symbols_linter = 29L:30L
  )
  # build list(list(linter = L1, line_number = I1), list(linter = L1, line_number = I2), ...)
  #   from the list of OK lines
  expected_lints <- unlist(
    Map(
      function(nm, ok) {
        lapply(
          setdiff(1L:30L, ok),
          \(bad) list(linter = nm, line_number = bad)
        )
      },
      names(ok_lines),
      ok_lines
    ),
    recursive = FALSE,
    use.names = FALSE
  )
  expect_lint(
    trim_some("
      x <-               #  1
        .x <-            #  2
        X <-             #  3
        x. <-            #  4
        X. <-            #  5
        x_ <-            #  6
        X_ <-            #  7
        xy <-            #  8
        xY <-            #  9
        Xy <-            # 10
        XY <-            # 11
        x1 <-            # 12
        X1 <-            # 13
        x_y <-           # 14
        X_Y <-           # 15
        X.Y <-           # 16
        x_2 <-           # 17
        X_2 <-           # 18
        x.2 <-           # 19
        X.2 <-           # 20
        IHave1Cat <-     # 21
        iHave1Cat <-     # 22
        i_have_1_cat <-  # 23
        I_HAVE_1_CAT <-  # 24
        i.have.1.cat <-  # 25
        ihave1cat <-     # 26
        IHAVE1CAT <      # 27
        I.HAVE_ONECAT <- # 28
        . <-             # 29
        `%^%` <-         # 30
        NULL
    "),
    expected_lints,
    linters,
    ignore_order = TRUE
  )
})

test_that("linter ignores some objects", {
  # names for which style check is ignored
  expect_no_lint("`%X%` <- t", object_name_linter("SNAKE_CASE")) # operator
  expect_no_lint("`%x%` <- t", object_name_linter("snake_case")) # operator
  expect_no_lint("`t.test` <- t", object_name_linter("UPPERCASE")) # std pkg
  expect_no_lint(".Deprecated('x')", object_name_linter("lowercase")) # std pkg
  expect_no_lint("print.foo <- t", object_name_linter("CamelCase")) # S3 generic
  expect_no_lint("names.foo <- t", object_name_linter("CamelCase")) # int generic
  expect_no_lint("sapply(x,f,USE.NAMES=T)", object_name_linter("snake_case")) # defined elsewhere
  expect_no_lint(".onLoad <- function(...) TRUE", object_name_linter("snake_case")) # namespace hooks, #500
  expect_no_lint(".First <- function(...) TRUE", object_name_linter("snake_case")) # namespace hooks
  expect_no_lint("`%++%` <- `+`", object_name_linter("symbols")) # all-symbol operator
  expect_no_lint("`%<-%` <- `+`", object_name_linter("symbols")) # all-symbol operator #495
  # S3 group generic, #1841
  expect_no_lint(
    "`==.snake_case` <- function(a, b) unclass(a) == unclass(b)",
    object_name_linter("snake_case")
  )
})

test_that("linter returns correct linting", {
  lint_msg <- "Variable and function name style should match camelCase."
  linter <- object_name_linter("camelCase")

  expect_no_lint("myObject <- 123", linter)
  expect_no_lint("`myObject` <- 123", linter)
  expect_lint("my.confused_NAME <- 1;", list(message = lint_msg, line_number = 1L, column_number = 1L), linter)
  expect_lint("1 ->> read.data.frame;", list(message = lint_msg, line_number = 1L, column_number = 7L), linter)
  expect_lint(
    "object_name_linter <- function(...) {}",
    list(message = lint_msg, line_number = 1L, column_number = 1L), linter
  )

  expect_lint(
    "Z = sapply('function', function(x=function(x){1}, b.a.z=F, ...){identity(b.a.z)}, USE.NAMES=TRUE)",
    list(
      list(message = lint_msg, line_number = 1L, column_number = 1L),
      list(message = lint_msg, line_number = 1L, column_number = 51L)
    ),
    linter
  )

  expect_no_lint("blah", linter)
  expect_no_lint("invokeRestartInteractively", linter)
  expect_no_lint("camelCase", linter)
  expect_no_lint("camelCase()", linter)
  expect_no_lint("pack::camelCase", linter)
  expect_no_lint("pack:::camelCase", linter)
  expect_no_lint("a(camelCase = 1)", linter)
  expect_no_lint("a$b <- 1", linter)
})

test_that("linter accepts vector of styles", {
  lint_msg <- "Variable and function name style should match camelCase or dotted.case."
  linter <- object_name_linter(styles = c("camelCase", "dotted.case"))

  expect_lint(
    "var.one <- 1\nvarTwo <- 2\nvar_three <- 3",
    list(message = lint_msg, line_number = 3L, column_number = 1L),
    linter
  )
})

test_that("dollar subsetting only lints the first expression", {
  # Regression test for #582
  linter <- object_name_linter()
  lint_msg <- rex::rex("Variable and function name style should match snake_case or symbols.")

  expect_no_lint("my_var$MY_COL <- 42L", linter)
  expect_lint("MY_VAR$MY_COL <- 42L", lint_msg, linter)
  expect_no_lint("my_var@MY_SUB <- 42L", linter)
  expect_lint("MY_VAR@MY_SUB <- 42L", lint_msg, linter)
})

patrick::with_parameters_test_that(
  "nested extraction only lints on the first symbol",
  expect_lint(
    sprintf("%s%sMY_SUB%sMY_COL <- 42L", if (should_lint) "MY_VAR" else "my_var", op1, op2),
    if (should_lint) rex::rex("Variable and function name style should match snake_case or symbols."),
    object_name_linter()
  ),
  .cases = within(
    expand.grid(should_lint = c(TRUE, FALSE), op1 = c("$", "@"), op2 = c("$", "@"), stringsAsFactors = FALSE),
    {
      .test_name <- sprintf("(should lint? %s, op1=%s, op2=%s)", should_lint, op1, op2)
    }
  )
)

test_that("assignment targets of compound lhs are correctly identified", {
  linter <- object_name_linter()
  lint_msg <- "Variable and function name style should match snake_case or symbols."

  # (recursive) [, $, and [[ subsetting
  expect_no_lint("good_name[badName] <- badName2", linter)
  expect_no_lint("good_name[1L][badName] <- badName2", linter)
  expect_no_lint("good_name[[badName]] <- badName2", linter)
  expect_no_lint("good_name[[1L]][[badName]] <- badName2", linter)
  expect_no_lint("good_name[[fun(badName)]] <- badName2", linter)
  expect_no_lint("good_name[[badName]]$badName2 <- badName3", linter)
  expect_no_lint("good_name$badName[[badName2]][badName3]$badName4 <- badName5", linter)

  expect_lint("badName[badName] <- badName2", lint_msg, linter)
  expect_lint("badName[1L][badName] <- badName2", lint_msg, linter)
  expect_lint("badName[[badName]] <- badName2", lint_msg, linter)
  expect_lint("badName[[1L]][[badName]] <- badName2", lint_msg, linter)
  expect_lint("badName[[fun(badName)]] <- badName2", lint_msg, linter)
  expect_lint("badName[[badName]]$badName2 <- badName3", lint_msg, linter)
  expect_lint("badName$badName[[badName2]][badName3]$badName4 <- badName5", lint_msg, linter)

  # setters
  expect_lint("setter(badName) <- good_name", lint_msg, linter)
  expect_no_lint("setter(good_name[[badName]]) <- good_name2", linter)

  # quotation
  expect_no_lint('"good_name" <- 42', linter)
  expect_lint('"badName" <- 42', lint_msg, linter)
  expect_no_lint("'good_name' <- 42", linter)
  expect_lint("'badName' <- 42", lint_msg, linter)
  expect_no_lint("`good_name` <- 42", linter)
  expect_lint("`badName` <- 42", lint_msg, linter)

  # subsetting with quotation
  expect_no_lint("good_name$\"badName\" <- 42", linter)
  expect_no_lint("good_name$'badName' <- 42", linter)
  expect_lint("badName$\"good_name\" <- 42", lint_msg, linter)
  expect_lint("badName$'good_name' <- 42", lint_msg, linter)
  expect_lint("`badName`$\"good_name\" <- 42", lint_msg, linter)
  expect_lint("`badName`$'good_name' <- 42", lint_msg, linter)
})

test_that("object_name_linter won't fail if an imported namespace is unavailable", {
  expect_length(
    lint_package(test_path("dummy_packages", "missing_dep"), linters = object_name_linter(), parse_settings = FALSE),
    3L
  )
})

test_that("object_name_linter supports custom regexes", {
  # disables default styles
  linter <- object_name_linter(
    regexes = c(shinyModule = rex::rex(start, lower, zero_or_more(alnum), "UI" %or% "Server", end))
  )
  msg <- rex::rex("Variable and function name style should match shinyModule.")
  linter2 <- object_name_linter(
    styles = c("snake_case", "symbols"),
    regexes = c(shinyModule = rex::rex(start, lower, zero_or_more(alnum), "UI" %or% "Server", end))
  )
  msg2 <- rex::rex("Variable and function name style should match snake_case, symbols or shinyModule.")

  # Can't allow 0 styles
  expect_error(
    object_name_linter(NULL),
    rex::rex("At least one style must be specified using `styles` or `regexes`.")
  )

  expect_lint(
    trim_some('
      snake_case <- 42L
      "%+%" <- function(...) ..1 + ..2

      myModuleUI <- function(id) {
        # blah
      }

      myModuleServer <- function(id) {
        # blah
      }

      myBadName <- 20L
    '),
    list(
      list(line_number = 1L, message = msg),
      list(line_number = 2L, message = msg),
      # argument "id" is linted if we only allow shinyModule names
      list(line_number = 4L, column_number = 24L, message = msg),
      list(line_number = 8L, column_number = 28L, message = msg),
      list(line_number = 12L, message = msg)
    ),
    linter
  )

  expect_lint(
    trim_some('
      snake_case <- 42L
      "%+%" <- function(...) ..1 + ..2

      myModuleUI <- function(id) {
        # blah
      }

      myModuleServer <- function(id) {
        # blah
      }

      myBadName <- 20L
    '),
    list(line_number = 12L, message = msg2),
    linter2
  )

  # Default regex naming works
  expect_lint(
    trim_some("
      a <- 42L
      b <- 1L
      c <- 2L
    "),
    list(line_number = 3L, message = rex::rex("Variable and function name style should match /^a$/ or /^b$/.")),
    object_name_linter(regexes = c("^a$", "^b$"))
  )

  expect_lint(
    trim_some("
      a <- 42L
      b <- 1L
      c <- 2L
    "),
    list(line_number = 3L, message = rex::rex("Variable and function name style should match a or /^b$/.")),
    object_name_linter(regexes = c(a = "^a$", "^b$"))
  )
})

test_that("complex LHS of := doesn't cause false positive", {
  # "_l" would be included under previous logic which tried ancestor::expr[ASSIGN] for STR_CONST,
  #   but only parent::expr[ASSIGN] is needed for strings.
  expect_no_lint('dplyr::mutate(df, !!paste0(v, "_l") := df$a * 2)', object_name_linter())
})

test_that("function shorthand also lints", {
  expect_lint("aBc <- \\() NULL", "function name style", object_name_linter())
})

test_that("capture groups in style are fine", {
  expect_no_lint("a <- 1\nab <- 2", object_name_linter(regexes = c(capture = "^(a)")))
  expect_no_lint("ab <- 1\nabc <- 2", object_name_linter(regexes = c(capture = "^(a)(b)")))
})

test_that("rlang name injection is handled", {
  linter <- object_name_linter()

  expect_no_lint('tibble("{name}" := 2)', linter)
  expect_no_lint('x %>% mutate("{name}" := 2)', linter)
  expect_lint('DT[, "{name}" := 2]', "style should match snake_case", linter)
})

test_that("literals in assign() and setGeneric() are checked", {
  linter <- object_name_linter()
  lint_msg <- rex::rex("Variable and function name style should match")

  expect_no_lint("assign('good_name', 2, env)", linter)
  expect_no_lint("assign('good_name', 'badName', env)", linter)
  expect_lint("assign('badName', 2, env)", lint_msg, linter)
  expect_lint('assign("badName", 2, env)', lint_msg, linter)

  expect_no_lint("setGeneric('good_name', function(x) x)", linter)
  expect_no_lint("setGeneric('good_name', function(x) x, package = 'badName')", linter)
  expect_lint("setGeneric('badName', function(x) x)", lint_msg, linter)
  expect_lint('setGeneric("badName", function(x) x)', lint_msg, linter)

  # Ditto for keyword arguments
  expect_no_lint("assign(x = 'good_name', 2, env)", linter)
  expect_no_lint("assign(envir = 'badEnvName', nm, 2)", linter)
  expect_no_lint("assign(envir = 'badEnvName', 'good_name', 2)", linter)
  expect_no_lint("assign(envir = 'badEnvName', x = 'good_name', 2)", linter)
  expect_lint("assign(x = 'badName', 2, env)", lint_msg, linter)
  expect_lint("assign(envir = 'good_env_name', 'badName', 2)", lint_msg, linter)
  expect_lint("assign(envir = 'good_env_name', x = 'badName', 2)", lint_msg, linter)

  # adversarial comments
  expect_lint(
    trim_some("
      assign(envir = # comment
      'good_env_name', 'badName', 2)
    "),
    lint_msg,
    linter
  )
})

test_that("generics assigned with '=' or <<- are registered", {
  linter <- object_name_linter()

  expect_no_lint(
    trim_some("
      f = function(x) {
        UseMethod('f')
      }
      g <<- function(x) {
        UseMethod('f')
      }
      f.default <- function(x) {}
      g.default <- function(x) {}
    "),
    linter
  )
})

test_that("pipe_call_linter skips allowed usages", {
  linter <- pipe_call_linter()

  expect_no_lint("a %>% foo()", linter)
  expect_no_lint("a %>% foo(x)", linter)
  expect_no_lint("b %>% { foo(., ., .) }", linter)
  expect_no_lint("a %>% foo() %>% bar()", linter)

  # ensure it works across lines too
  lines <- trim_some("
  a %>%
    foo() %>%
    bar()
  ")
  expect_no_lint(lines, linter)

  # symbol extraction is OK (don't force extract2(), e.g.)
  expect_no_lint("a %>% .$y %>% mean()", linter)

  # more complicated expressions don't pick up on nested symbols
  lines <- trim_some("
  x %>% {
    tmp <- .
    bla <- foo %>% unrelated_stuff(tmp)
    my_combination_fun(tmp, bla)
  }
  ")
  expect_no_lint(lines, linter)

  # extraction pipe uses RHS symbols
  expect_no_lint("a %$% b", linter)
})

test_that("pipe_call_linter blocks simple disallowed usages", {
  expect_lint(
    "x %>% foo",
    "Use explicit calls in magrittr pipes",
    pipe_call_linter()
  )

  expect_lint(
    "x %>% foo() %>% bar",
    "Use explicit calls in magrittr pipes",
    pipe_call_linter()
  )

  expect_lint(
    "x %>% foo %>% bar()",
    "Use explicit calls in magrittr pipes",
    pipe_call_linter()
  )

  lines <- trim_some("
  a %>%
    foo %>%
    bar()
  ")
  expect_lint(
    lines,
    "Use explicit calls in magrittr pipes",
    pipe_call_linter()
  )
})

local({
  pipes <- pipes(exclude = c("%$%", "|>"))
  linter <- pipe_call_linter()
  patrick::with_parameters_test_that(
    "All pipe operators are caught",
    {
      expect_no_lint(sprintf("a %s foo()", pipe), linter)
      expect_lint(sprintf("a %s foo", pipe), sprintf("`a %s foo`", pipe), linter)
    },
    pipe = pipes,
    .test_name = names(pipes)
  )
})

test_that("Multiple lints give custom messages", {
  expect_lint(
    trim_some("
      a %>% b
      c %T>% d
    "),
    list(
      list(message = "%>%", line_number = 1L),
      list(message = "%T>%", line_number = 2L)
    ),
    pipe_call_linter()
  )
})

test_that("stopifnot_all_linter skips allowed usages", {
  expect_no_lint("stopifnot(all(x) || any(y))", stopifnot_all_linter())
})

test_that("stopifnot_all_linter blocks simple disallowed usages", {
  linter <- stopifnot_all_linter()
  lint_msg <- rex::rex("stopifnot(x) runs all() 'under the hood'")

  expect_lint("stopifnot(all(A))", list(lint_msg, column_number = 11L), linter)
  expect_lint("stopifnot(x, y, all(z))", list(lint_msg, column_number = 17L), linter)

  expect_lint(
    trim_some("{
      stopifnot(all(x), all(y),
        all(z)
      )
      stopifnot(a > 0, b < 0, all(c == 0))
    }"),
    list(
      list(lint_msg, line_number = 2L, column_number = 13L),
      list(lint_msg, line_number = 2L, column_number = 21L),
      list(lint_msg, line_number = 3L, column_number = 5L),
      list(lint_msg, line_number = 5L, column_number = 27L)
    ),
    linter
  )
})

# fuzzer disable: comment_injection
test_that("brace_linter lints braces correctly", {
  open_curly_msg <- rex::rex(
    "Opening curly braces should never go on their own line"
  )
  closed_curly_msg <- rex::rex(
    "Closing curly-braces should always be on their own line, ",
    "unless they are followed by an else."
  )

  linter <- brace_linter()
  expect_no_lint("blah", linter)
  expect_no_lint("a <- function() {\n}", linter)
  expect_no_lint("a <- function() {  \n}", linter)

  expect_lint("a <- function() { 1 }", list(open_curly_msg, closed_curly_msg), linter)
  # allowed by allow_single_line
  expect_no_lint("a <- function() { 1 }", brace_linter(allow_single_line = TRUE))

  expect_lint(
    trim_some("
      a <- if(1) {
        1} else {
        2
      }
    "),
    closed_curly_msg,
    linter
  )
  expect_lint(
    trim_some("
      a <- if(1) {
        1
      } else {
        2}
    "),
    closed_curly_msg,
    linter
  )

  expect_lint(
    trim_some("
      a <- if(1) {
        1} else {
        2}
    "),
    list(
      closed_curly_msg,
      closed_curly_msg
    ),
    linter
  )

  # }) is allowed
  expect_no_lint("eval(bquote({\n...\n}))", linter)
  # }] is too
  expect_no_lint("df[, {\n...\n}]", linter)

  # }, is allowed
  expect_no_lint(
    trim_some("
      fun({
        statements
      }, param)"),
    linter
  )
  expect_no_lint(
    trim_some("
      fun(function(a) {
        statements
      }, param)"),
    linter
  )

  # ,<\n>{ is allowed
  expect_no_lint(
    trim_some("
      switch(
        x,
        'a' = do_something(x),
        'b' = do_another(x),
        {
          do_first(x)
          do_second(x)
        }
      )
    "),
    linter
  )

  # a comment before ,<\n>{ is allowed
  expect_no_lint(
    trim_some("
      switch(
        x,
        'a' = do_something(x),
        'b' = do_another(x), # comment
        {
          do_first(x)
          do_second(x)
        }
      )
    "),
    linter
  )

  # a comment before <\n>{ is allowed
  expect_no_lint(
    trim_some("
      switch(stat,
      o = {
        x <- 0.01
      },
      # else
      {
        x <- 2
      }
    )
    "),
    linter
  )

  # a comment after '}' is allowed
  expect_no_lint(
    trim_some("
      switch(
        x,
        'a' = do_something(x),
        'b' = do_another(x),
        {
          do_first(x)
          do_second(x)
        } # comment
      )
    "),
    brace_linter()
  )

  expect_no_lint(
    trim_some("
      fun(
        'This is very very very long text.',
        {
          message('This is the code.')
          message('It\\'s stupid, but proves my point.')
        }
      )
    "),
    linter
  )

  # (\n{ is allowed optionally
  expect_no_lint(
    trim_some("
      tryCatch(
        {
          print(1)
        },
        error = function(err) {
        }
      )
    "),
    linter
  )

  # {{ }} is allowed
  expect_no_lint("{{ x }}", linter)

  expect_no_lint(
    trim_some("
      pkg_name <- function(path = find_package()) {
        if (is.null(path)) {
          return(NULL)
        } else {
          read.dcf(file.path(path, \"DESCRIPTION\"), fields = \"Package\")[1]
        }
      }
    "),
    linter
  )

  expect_lint(
    "a <- function()
     # comment
    {
      1
    }",
    open_curly_msg,
    linter
  )

  expect_lint("a <- function()\n{\n  1 \n}", open_curly_msg, linter)
  expect_lint("a <- function()\n    {\n  1 \n}", open_curly_msg, linter)
  expect_lint("a <- function()\n\t{\n  1 \n}", open_curly_msg, linter)

  # trailing comments are allowed
  expect_no_lint(
    trim_some('
      if ("P" != "NP") { # what most people expect
        print("Cryptomania is possible")
      }
    '),
    linter
  )
})

test_that("brace_linter lints spaces before open braces", {
  linter <- brace_linter()
  lint_msg <- rex::rex("There should be a space before an opening curly brace.")

  expect_lint(
    "blah <- function(){\n}",
    list(
      message = lint_msg,
      column_number = 19L
    ),
    linter
  )

  expect_lint(
    "\nblah <- function(){\n\n\n}",
    list(
      message = lint_msg,
      column_number = 19L
    ),
    linter
  )

  # should also lint if/else
  expect_lint(
    "a <- if (a){\n} else{\n}",
    list(
      list(message = lint_msg, line_number = 1L, column_number = 12L),
      list(message = lint_msg, line_number = 2L, column_number = 7L)
    ),
    linter
  )

  # should lint repeat{
  expect_lint(
    "repeat{\nblah\n}",
    list(message = lint_msg, line_number = 1L, column_number = 7L),
    linter
  )

  # should ignore strings and comments, as in regexes:
  expect_no_lint("grepl('(iss){2}', 'Mississippi')", linter)
  expect_no_lint(
    "x <- 123 # don't flag (paren){brace} if inside a comment",
    linter
  )
  # should not be thrown when the brace lies on subsequent line
  expect_lint(
    trim_some("
      x <- function()
                     {2}
    "),
    list(
      rex::rex("Opening curly braces should never go on their own line"),
      rex::rex("Closing curly-braces should always be on their own line")
    ), # , but not lint_msg
    linter
  )
})

test_that("brace_linter lints else correctly", {
  linter <- brace_linter()
  expect_no_lint("if (TRUE) 1 else 2", linter)
  expect_no_lint("if (TRUE) 1", linter)

  lines_brace <- trim_some("
    if (TRUE) {
      1
    } else {
      2
    }
  ")
  expect_no_lint(lines_brace, linter)

  # such usage is also not allowed by the style guide, but test anyway
  lines_unbrace <- trim_some("
    foo <- function(x) {
      if (TRUE)
        1
      else
        2
    }
  ")
  expect_no_lint(lines_unbrace, linter)

  lines <- trim_some("
    foo <- function(x) {
      if (x) {
        1
      }
      else {
        2
      }
    }
  ")
  expect_lint(
    lines,
    rex::rex("`else` should come on the same line as the previous `}`."),
    linter
  )
})

test_that("brace_linter lints function expressions correctly", {
  msg_always <- rex::rex("Wrap function bodies in curly braces.")
  msg_multi_line <- rex::rex("Wrap multi-line function bodies in curly braces.")
  msg_not_inline <- rex::rex("Wrap function bodies starting on a new line in curly braces.")
  msgs_open_close <- list(
    rex::rex("Opening curly braces should never go on their own line and should always be followed by a new line."),
    rex::rex("Closing curly-braces should always be on their own line, unless they are followed by an else.")
  )

  linter_always <- brace_linter(function_bodies = "always")
  linter_multi_line <- brace_linter(function_bodies = "multi_line")
  linter_not_inline <- brace_linter(function_bodies = "not_inline")
  linter_never <- brace_linter(function_bodies = "never")

  lines <- trim_some("
    function(x) {
      x + 4
    }
  ")
  expect_no_lint(lines, linter_always)
  expect_no_lint(lines, linter_multi_line)
  expect_no_lint(lines, linter_not_inline)
  expect_no_lint(lines, linter_never)

  expect_lint("function(x) { x + 4 }", msgs_open_close, linter_always)
  expect_lint("function(x) { x + 4 }", msgs_open_close, linter_multi_line)
  expect_lint("function(x) { x + 4 }", msgs_open_close, linter_not_inline)
  expect_lint("function(x) { x + 4 }", msgs_open_close, linter_never)
  # function_bodies = "always" should only prohibit inline functions with allow_single_line = FALSE (the default):
  expect_no_lint(
    "function(x) { x + 4 }",
    brace_linter(allow_single_line = TRUE, function_bodies = "always")
  )

  expect_lint("function(x) x + 4", msg_always, linter_always)
  expect_no_lint("function(x) x + 4", linter_multi_line)
  expect_no_lint("function(x) x + 4", linter_not_inline)
  expect_no_lint("function(x) x + 4", linter_never)

  lines <- trim_some("
    function(x) x +
      4
  ")
  expect_lint(lines, msg_always, linter_always)
  expect_lint(lines, msg_multi_line, linter_multi_line)
  expect_no_lint(lines, linter_not_inline)
  expect_no_lint(lines, linter_never)

  lines <- trim_some("
    function(x)
      x + 4
  ")
  expect_lint(lines, msg_always, linter_always)
  expect_lint(lines, msg_multi_line, linter_multi_line)
  expect_lint(lines, msg_not_inline, linter_not_inline)
  expect_no_lint(lines, linter_never)

  # missing newline after opening brace; closing brace not on sep line
  lines <- trim_some("
    foo <- function(x) { x +
      4 }
  ")
  expect_lint(lines, msgs_open_close, linter_always)
  expect_lint(lines, msgs_open_close, linter_multi_line)
  expect_lint(lines, msgs_open_close, linter_not_inline)
  expect_lint(lines, msgs_open_close, linter_never)

  # fn body wrapped in additional unneeded parentheses
  lines <- trim_some("
    foo <- function(x) ({
      x + 1
    })
  ")
  expect_lint(lines, msg_always, linter_always)
  expect_lint(lines, msg_multi_line, linter_multi_line)
  expect_no_lint(lines, linter_not_inline)
  expect_no_lint(lines, linter_never)
})

test_that("brace_linter lints if/else matching braces correctly", {
  linter <- brace_linter()

  expect_no_lint("if (TRUE) 1 else 2", linter)
  expect_no_lint("if (TRUE) 1", linter)

  lines_brace <- trim_some("
    if (TRUE) {
      1
    } else {
      2
    }
  ")
  expect_no_lint(lines_brace, linter)

  # such usage is also not allowed by the style guide, but test anyway
  lines_unbrace <- trim_some("
    foo <- function(x) {
      if (TRUE)
        1
      else
        2
    }
  ")
  expect_no_lint(lines_unbrace, linter)

  # else if is OK
  lines_else_if <- trim_some("
    if (x) {
     1
    } else if (y) {
     2
    } else {
     3
    }
  ")
  expect_no_lint(lines_else_if, linter)

  lines_if <- trim_some("
    foo <- function(x) {
      if (x) {
        1
      } else 2
    }
  ")
  expect_lint(
    lines_if,
    rex::rex("Either both or neither branch in `if`/`else` should use curly braces."),
    linter
  )

  lines_else <- trim_some("
    foo <- function(x) {
      if (x) 1 else {
        2
      }
    }
  ")
  expect_lint(
    lines_else,
    rex::rex("Either both or neither branch in `if`/`else` should use curly braces."),
    linter
  )
})

# Keep up to date with https://github.com/tidyverse/style/issues/191
test_that("empty brace expressions are always allowed inline", {
  linter <- brace_linter()
  linter_allow <- brace_linter(allow_single_line = TRUE)
  lint_msg <- rex::rex("Opening curly braces")

  expect_no_lint("while (FALSE) {}", linter)
  expect_no_lint("while (FALSE) { }", linter)
  # only applies when `{` is "attached" to the preceding token on the same line
  expect_lint("while (FALSE)\n{}", lint_msg, linter)
  expect_lint("while (FALSE)\n{ }", lint_msg, linter)
  expect_no_lint("while (FALSE) {}", linter_allow)
  expect_no_lint("while (FALSE) { }", linter_allow)
})

test_that("formula syntax is linted properly", {
  linter <- brace_linter()
  lint_msg_open <- rex::rex("Opening curly braces should never go on their own line")
  lint_msg_closed <- rex::rex("Closing curly-braces should always be on their own line")

  expect_no_lint(
    trim_some("
      map(
        .x = 1:4,
        .f = ~ {
                .x + 1
              }
      )"),
    linter
  )

  expect_lint(
    trim_some("
      map(
        .x = 1:4,
        .f = ~ {.x + 1}
      )"),
    list(
      list(message = lint_msg_open, line_number = 3L, column_number = 10L),
      list(message = lint_msg_closed, line_number = 3L, column_number = 17L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      map(
        .x = 1:4,
        .f = ~ { .x + 1
              }
      )"),
    list(
      list(message = lint_msg_open, line_number = 3L, column_number = 10L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      map(
        .x = 1:4,
        .f = ~ {
                .x + 1}
      )"),
    list(
      list(message = lint_msg_closed, line_number = 4L, column_number = 17L)
    ),
    linter
  )
})

test_that("code with pipes is handled correctly", {
  linter <- brace_linter()
  lint_msg_open <- rex::rex("Opening curly braces should never go on their own line")
  lint_msg_closed <- rex::rex("Closing curly-braces should always be on their own line")

  expect_no_lint(
    trim_some("
      out <- lapply(stuff, function(i) {
        do_something(i)
      }) %>% unlist()
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      1:4 %!>% {
          sum(.)
        }
    "),
    linter
  )

  # %>%\n{ is allowed
  expect_no_lint(
    trim_some("
      1:4 %T>%
        {
          sum(.)
        }
    "),
    linter
  )

  expect_lint(
    trim_some("
      xx %<>% { sum(.)
        }
    "),
    list(
      list(message = lint_msg_open, line_number = 1L, column_number = 9L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      x %>%
        {
          uvwxyz }
    "),
    list(
      list(message = lint_msg_closed, line_number = 3L, column_number = 12L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      1:4 %>%
        { sum(.) }
    "),
    list(
      list(message = lint_msg_closed, line_number = 2L, column_number = 12L)
    ),
    linter
  )

  expect_lint(
    "1:4 %>% { sum(.) }",
    list(
      list(message = lint_msg_open, line_number = 1L, column_number = 9L),
      list(message = lint_msg_closed, line_number = 1L, column_number = 18L)
    ),
    linter
  )

  expect_no_lint(
    trim_some("
      out <- lapply(stuff, function(i) {
        do_something(i)
      }) |> unlist()
    "),
    linter
  )

  expect_lint(
    "local({ 1:4 |> sum() })",
    list(
      list(message = lint_msg_open, line_number = 1L, column_number = 7L)
    ),
    linter
  )
})

test_that("function shorthand is treated like 'full' function", {
  linter <- brace_linter()

  expect_no_lint("a <- \\() {  \n}", linter)
  expect_lint(
    trim_some("
      x <- \\()
              {2}
    "),
    list(
      rex::rex("Opening curly braces should never go on their own line"),
      rex::rex("Closing curly-braces should always be on their own line")
    ),
    linter
  )
})
# fuzzer enable: comment_injection

test_that("nested_pipe_linter skips allowed usages", {
  linter <- nested_pipe_linter()

  expect_no_lint("a %>% b() %>% c()", linter)

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        out <- a %>% b()
        return(out)
      }
    "),
    linter
  )

  # pipes fitting on one line can be ignored
  expect_no_lint( # nofuzz: comment_injection
    "bind_rows(a %>% select(b), c %>% select(b))",
    linter
  )

  # switch outputs are OK
  expect_no_lint("switch(x, a = x %>% foo())", linter)
  # final position is an output position
  expect_no_lint("switch(x, a = x, x %>% foo())", linter)

  # inline switch inputs are not linted
  expect_no_lint( # nofuzz: comment_injection
    trim_some("
      switch(
        x %>% foo(),
        a = x
      )
    "),
    linter
  )
})

patrick::with_parameters_test_that(
  "allow_outer_calls defaults are ignored by default",
  expect_no_lint(
    trim_some(sprintf(outer_call, fmt = "
      %s(
        x %%>%%
          foo()
      )
    ")),
    nested_pipe_linter()
  ),
  .test_name = c("try", "tryCatch", "withCallingHandlers"),
  outer_call = c("try", "tryCatch", "withCallingHandlers")
)

test_that("nested_pipe_linter blocks simple disallowed usages", {
  linter <- nested_pipe_linter()
  linter_inline <- nested_pipe_linter(allow_inline = FALSE)
  lint_msg <- rex::rex("Don't nest pipes inside other calls.")

  expect_lint(
    "bind_rows(a %>% select(b), c %>% select(b))",
    list(lint_msg, lint_msg),
    linter_inline
  )

  expect_lint(
    trim_some("
      print(
        a %>%
          filter(b > c)
      )
    "),
    lint_msg,
    linter
  )

  # switch inputs are linted
  expect_lint(
    trim_some("
      switch(
        x %>%
          foo(),
        a = x
      )
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      switch(
        x %>% foo(),
        a = x
      )
    "),
    lint_msg,
    linter_inline
  )
})

test_that("allow_outer_calls= argument works", {
  expect_lint(
    trim_some("
      try(
        x %>%
          foo()
      )
    "),
    rex::rex("Don't nest pipes inside other calls."),
    nested_pipe_linter(allow_outer_calls = character())
  )

  expect_no_lint(
    trim_some("
      print(
        x %>%
          foo()
      )
    "),
    nested_pipe_linter(allow_outer_calls = "print")
  )
})

test_that("Native pipes are handled as well", {
  linter <- nested_pipe_linter()
  linter_inline <- nested_pipe_linter(allow_inline = FALSE)
  lint_msg <- rex::rex("Don't nest pipes inside other calls.")

  expect_no_lint( # nofuzz: comment_injection
    "bind_rows(a |> select(b), c |> select(b))",
    linter
  )
  expect_lint(
    "bind_rows(a |> select(b), c |> select(b))",
    list(lint_msg, lint_msg),
    linter_inline
  )

  expect_lint(
    trim_some("
      print(
        a |>
          filter(b > c)
      )
    "),
    lint_msg,
    linter
  )
})

test_that("lints vectorize", { # nofuzz: comment_injection
  lint_msg <- rex::rex("Don't nest pipes inside other calls.")

  lines <- trim_some("{
    bind_rows(
      a %>% select(b),
      c %>%
        select(d),
      e %>%
        select(f) %>%
        filter(g > 0),
      h %>% filter(i < 0)
    )
  }")
  expect_lint(
    lines,
    list(
      list(lint_msg, line_number = 4L),
      list(lint_msg, line_number = 6L)
    ),
    nested_pipe_linter()
  )

  expect_lint(
    lines,
    list(
      list(lint_msg, line_number = 3L),
      list(lint_msg, line_number = 4L),
      list(lint_msg, line_number = 6L),
      list(lint_msg, line_number = 9L)
    ),
    nested_pipe_linter(allow_inline = FALSE)
  )
})

test_that("backport_linter produces error when R version misspecified", {
  expect_error(
    lint(text = "numToBits(2)", linters = backport_linter(420L)),
    "`r_version` must be an R version number"
  )
})

test_that("backport_linter detects backwards-incompatibility", {
  # default should be current R version; all of these are included on our dependency
  expect_no_lint(".getNamespaceInfo(dir.exists(lapply(x, toTitleCase)))", backport_linter())
  expect_no_lint(".getNamespaceInfo(dir.exists(lapply(x, toTitleCase)))", backport_linter("release"))
  expect_no_lint(".getNamespaceInfo(dir.exists(lapply(x, toTitleCase)))", backport_linter("devel"))

  expect_lint(
    "numToBits(2)",
    rex::rex("numToBits (R 4.1.0) is not always available for requested dependency (R >= 4.0.0)."),
    backport_linter("4.0.0")
  )
  # symbols as well as calls
  expect_lint(
    "lapply(1:10, numToBits)",
    rex::rex("numToBits (R 4.1.0) is not always available for requested dependency (R >= 4.0.0)."),
    backport_linter("4.0.0")
  )

  expect_lint(
    trim_some("
      trimws(
        ...names()
      )
    "),
    list(
      list(rex::rex("trimws (R 3.2.0)", anything, "(R >= 3.0.0)."), line_number = 1L),
      list(rex::rex("...names (R 4.1.0)", anything, "(R >= 3.0.0)."), line_number = 2L)
    ),
    backport_linter("3.0.0")
  )

  # oldrel specification
  expect_lint(
    "grepv()",
    rex::rex("grepv (R 4.5.0) is not always available for requested dependency (R >= 4.4.3)."),
    backport_linter("oldrel")
  )

  expect_error(
    backport_linter("oldrel-99"),
    "`r_version` is not valid"
  )

  expect_lint(
    "numToBits(2)",
    rex::rex("numToBits (R 4.1.0) is not always available for requested dependency (R >= 4.0.5)."),
    backport_linter("oldrel-5")
  )
  # no interference from namespace-qualification (even of base functions)
  expect_lint(
    "base::numToBits(2)",
    rex::rex("numToBits (R 4.1.0) is not always available for requested dependency (R >= 4.0.5)."),
    backport_linter("oldrel-5")
  )

  # except is honored
  expect_no_lint(
    trim_some("
      numToBits(2)
      R_user_dir('mypkg')
    "),
    backport_linter("3.0.0", except = c("numToBits", "R_user_dir"))
  )

  # binary operators (%||%, etc.)
  expect_lint(
    "a %||% b",
    rex::rex(
      "%||% (R 4.4.0) is not always available ",
      "for requested dependency (R >= 4.3.0)."
    ),
    backport_linter("4.3.0")
  )

  # functions not mentioned in NEWS or introduced in patch releases
  expect_lint(
    "zstdfile('foo.zst')",
    rex::rex(
      "zstdfile (R 4.5.0) is not always available ",
      "for requested dependency (R >= 4.4.0)."
    ),
    backport_linter("4.4.0")
  )

  expect_lint(
    "str2lang('x + 1')",
    rex::rex(
      "str2lang (R 3.6.1) is not always available ",
      "for requested dependency (R >= 3.6.0)."
    ),
    backport_linter("3.6.0")
  )
})

test_that("backport_linter generates expected warnings", {
  tmp <- withr::local_tempfile(lines = "x <- x + 1")

  expect_warning(
    {
      l <- lint(tmp, backport_linter("2.0.0"))
    },
    'version older than "3.0.0"',
    fixed = TRUE
  )
  expect_identical(l, lint(tmp, backport_linter("3.0.0")))
})

test_that("keyword_quote_linter skips allowed usages", {
  linter <- keyword_quote_linter()

  # main use case: c()
  expect_no_lint("x <- c(1, 2, 4, 5)", linter)
  expect_no_lint("x <- c(a = 1, 2)", linter)
  expect_no_lint("x <- c(a = 1, b = 2)", linter)
  expect_no_lint("y <- c(`a b` = 1, `c d` = 2)", linter)
  expect_no_lint('y <- c("a b" = 1, "c d" = 2)', linter)
  expect_no_lint("z <- c('a b' = 1, c = 2)", linter)

  # don't catch strings as arguments
  expect_no_lint('c(A = "a")', linter)
  # don't catch unnamed arguments
  expect_no_lint('c(1, 2, "a")', linter)
  # don't get thrown off by missing arguments
  expect_no_lint("alist(`a b` =)", linter)

  # other use cases: switch() and list()
  expect_no_lint("list(a = 1, b = list(c = 2))", linter)
  expect_no_lint("list(`a b` = 1, c = 2:6)", linter)

  expect_no_lint("switch(x, a = 1, b = 2)", linter)
  expect_no_lint("switch(x, `a b` = 1, c = 2:6)", linter)
})

test_that("keyword_quote_linter blocks simple disallowed usages", {
  linter <- keyword_quote_linter()
  lint_msg <- "Only quote named arguments to functions"

  expect_lint('c("a" = 1, b = 2)', lint_msg, linter)

  expect_lint(
    "c('a' = 1, 'b' = 2)",
    list(lint_msg, lint_msg),
    linter
  )

  expect_lint(
    "c(`a` = 1, b = list(`c` = 2))",
    list(lint_msg, lint_msg),
    linter
  )

  # missing argument is caught
  expect_lint("alist(`a` = )", lint_msg, linter)

  expect_lint(
    "switch(x, `a` = c('b' = list(\"c\" = 1)))",
    list(lint_msg, lint_msg, lint_msg),
    linter
  )
})

test_that("keyword_quote_linter skips quoting on reserved words", {
  linter <- keyword_quote_linter()

  expect_no_lint("c(`next` = 1, `while` = 2)", linter)
  expect_no_lint("switch(x, `for` = 3, `TRUE` = 4)", linter)
  expect_no_lint("list('NA' = 5, 'Inf' = 6)", linter)
})

test_that("keyword_quote_linter works on more common functions", {
  linter <- keyword_quote_linter()
  lint_msg <- "Only quote named arguments to functions"

  expect_lint("data.frame('a' = 1)", lint_msg, linter)
  expect_lint("data.table('a' = 1)", lint_msg, linter)
  expect_lint("data.table::data.table('a' = 1)", lint_msg, linter)
  expect_lint("rbind('a' = 1)", lint_msg, linter)
  expect_lint("cbind('a' = 1)", lint_msg, linter)
})

test_that("keyword_quote_linter finds blocked usages in any function call", {
  expect_lint(
    "foo('a' = 1)",
    rex::rex("Only quote named arguments to functions"),
    keyword_quote_linter()
  )
})

test_that("keyword_quote_linter blocks quoted assignment targets", {
  linter <- keyword_quote_linter()
  backtick_msg <- rex::rex("Use backticks to create non-syntactic names, not quotes.")
  assign_msg <- "Only quote targets of assignment if necessary"

  expect_lint('"foo bar" <- 1', backtick_msg, linter)
  expect_lint("'foo bar' = 1", backtick_msg, linter)
  # valid choice: use backticks
  expect_no_lint("`foo bar` = 1", linter)

  expect_lint('"foo" <- 1', assign_msg, linter)
  expect_lint("'foo' = 1", assign_msg, linter)
  expect_lint("`foo` = 1", assign_msg, linter)

  # don't include data.table assignments
  expect_no_lint('DT[, "a" := 1]', linter)
  expect_no_lint("DT[, 'a' := 1]", linter)
  expect_no_lint("DT[, `a` := 1]", linter)

  # include common use cases: [<-/$ methods and infixes
  expect_lint('"$.my_class" <- function(x, key) { }', backtick_msg, linter)
  expect_lint("'Setter[<-.my_class' = function(x, idx, value) { }", backtick_msg, linter)
  expect_lint('"%nin%" <- function(x, table) !x %in% table', backtick_msg, linter)

  # right assignment
  expect_lint('1 -> "foo"', assign_msg, linter)
  expect_no_lint("1 -> foo", linter)
  expect_lint('1 -> "a b"', backtick_msg, linter)
})

test_that("keyword_quote_linter blocks quoted $, @ extractions", { # nofuzz: dollar_at
  linter <- keyword_quote_linter()
  backtick_msg <- rex::rex("Use backticks to create non-syntactic names, not quotes.")
  dollar_msg <- rex::rex("Only quote targets of extraction with $ if necessary")
  at_msg <- rex::rex("Only quote targets of extraction with @ if necessary")

  expect_lint('x$"foo bar" <- 1', backtick_msg, linter)
  expect_lint("x$'foo bar' = 1", backtick_msg, linter)
  expect_lint('x@"foo bar" <- 1', backtick_msg, linter)
  expect_lint("x@'foo bar' = 1", backtick_msg, linter)
  # valid choice: non-syntactic name with backticks
  expect_no_lint("x@`foo bar` <- 1", linter)
  expect_no_lint("x@`foo bar` = 1", linter)

  expect_lint('x$"foo" <- 1', dollar_msg, linter)
  expect_lint("x$'foo' = 1", dollar_msg, linter)
  expect_lint('x@"foo" <- 1', at_msg, linter)
  expect_lint("x@'foo' = 1", at_msg, linter)
  expect_lint("x@`foo` <- 1", at_msg, linter)
  expect_lint("x@`foo` = 1", at_msg, linter)
})

test_that("multiple lints are generated correctly", { # nofuzz: dollar_at
  linter <- keyword_quote_linter()

  expect_lint(
    trim_some('{
      foo("a" = 1)
      "b" <- 2
      x$"c"
      y@"d"
    }'),
    list(
      "Only quote named arguments",
      "Only quote targets of assignment",
      "Only quote targets of extraction with \\$",
      "Only quote targets of extraction with @"
    ),
    linter
  )

  # multiple flavors of assignment lints
  expect_lint(
    trim_some('{
      "a" <- 1
      "a b" <- 1
      `a` <- 1
      `a b` <- 1
    }'),
    list(
      "Only quote targets of assignment if necessary",
      "Use backticks to create non-syntactic names, not quotes",
      "Only quote targets of assignment if necessary"
    ),
    linter
  )

  # multiple flavors of extraction lints
  expect_lint(
    trim_some('{
      x$"a"
      x$"a b" <- 1
      x$`a` <- 1
      x$`a b` <- 1
      y@"a"
      y@"a b" <- 1
      y@`a` <- 1
      y@`a b` <- 1
    }'),
    list(
      rex::rex("Only quote targets of extraction with $ if necessary"),
      rex::rex("Use backticks to create non-syntactic names, not quotes."),
      rex::rex("Only quote targets of extraction with $ if necessary"),
      rex::rex("Only quote targets of extraction with @ if necessary"),
      rex::rex("Use backticks to create non-syntactic names, not quotes."),
      rex::rex("Only quote targets of extraction with @ if necessary")
    ),
    linter
  )
})

# fuzzer disable: comment_injection
test_that("returns the correct linting", {
  linter <- trailing_whitespace_linter()
  lint_msg <- rex::rex("Remove trailing whitespace.")

  expect_no_lint("blah", linter)

  expect_lint(
    "blah <- 1  ",
    list(message = lint_msg, column_number = 10L),
    linter
  )

  expect_lint("blah <- 1  \n'hi'", lint_msg, linter)

  expect_lint(
    "blah <- 1\n'hi'\na <- 2  ",
    list(message = lint_msg, line_number = 3L),
    linter
  )
})

test_that("also handles completely empty lines per allow_empty_lines argument", {
  linter <- trailing_whitespace_linter()
  lint_msg <- rex::rex("Remove trailing whitespace.")

  expect_lint(
    "blah <- 1\n  \n'hi'\na <- 2",
    list(message = lint_msg, line_number = 2L),
    linter
  )

  expect_lint(
    "blah <- 1  ",
    list(message = lint_msg, column_number = 10L),
    trailing_whitespace_linter(allow_empty_lines = TRUE)
  )

  expect_no_lint(
    "blah <- 1\n  \n'hi'\na <- 2",
    trailing_whitespace_linter(allow_empty_lines = TRUE)
  )
})

test_that("also handles trailing whitespace in string constants", {
  linter <- trailing_whitespace_linter()
  lint_msg <- rex::rex("Remove trailing whitespace.")

  expect_no_lint("blah <- '  \n  \n'", linter)
  # Don't exclude past the end of string
  expect_lint(
    "blah <- '  \n  \n'  ",
    list(message = lint_msg, line_number = 3L),
    linter
  )
  # can be enabled with allow_in_strings = FALSE
  expect_lint(
    "blah <- '  \n  \n'",
    list(message = lint_msg, line_number = 1L),
    trailing_whitespace_linter(allow_empty_lines = TRUE, allow_in_strings = FALSE)
  )
  expect_lint(
    "blah <- '  \n  \n'",
    list(
      list(message = lint_msg, line_number = 1L),
      list(message = lint_msg, line_number = 2L)
    ),
    trailing_whitespace_linter(allow_in_strings = FALSE)
  )
})
# fuzzer enable: comment_injection

test_that("all_equal_linter() skips allowed usages", {
  linter <- all_equal_linter()

  # Only when used in if
  expect_no_lint("all.equal(a, b)", linter)
  expect_no_lint("all.equal(a, b, tolerance = 1e-3)", linter)

  expect_no_lint("if (isTRUE(all.equal(a, b))) message('equal')", linter)
  expect_no_lint("if (!isTRUE(all.equal(a, b))) message('different')", linter)

  expect_no_lint("if (A) all.equal(x, y)", linter)
})

test_that("all_equal_linter() blocks simple disallowed usages", {
  linter <- all_equal_linter()
  lint_message <- rex::rex("Wrap all.equal() in isTRUE().")
  expect_lint("if (all.equal(a, b, tolerance = 1e-3)) message('equal')", lint_message, linter)

  lint_message <- rex::rex("Wrap all.equal() in isTRUE(), or replace it by identical()")
  expect_lint("if (all.equal(a, b)) message('equal')", lint_message, linter)
  expect_lint("!all.equal(a, b)", lint_message, linter)
  expect_lint("while (all.equal(a, b)) message('equal')", lint_message, linter)

  lint_message <- rex::rex("!isTRUE()")
  expect_lint("isFALSE(all.equal(a, b))", lint_message, linter)
})

test_that("lints vectorize", {
  lint_message <- rex::rex("Wrap all.equal() in isTRUE()")

  expect_lint(
    trim_some("{
      all.equal(a, b)
      !all.equal(a, b)
      !isTRUE(all.equal(a, b))
      if ( # test
        all.equal(a, b)) message('equal')
    }"),
    list(
      list(lint_message, line_number = 3L),
      list(lint_message, line_number = 6L)
    ),
    all_equal_linter()
  )
})

# fuzzer disable: comment_injection
test_that("spaces_left_parentheses_linter skips allowed usages", {
  linter <- spaces_left_parentheses_linter()

  expect_no_lint("blah", linter)
  expect_no_lint("print(blah)", linter)
  expect_no_lint("base::print(blah)", linter)
  expect_no_lint("base::print(blah, fun(1))", linter)
  expect_no_lint("blah <- function(blah) { }", linter)

  expect_no_lint("(1 + 1)", linter)
  expect_no_lint("(1 + 1)", linter)
  expect_no_lint("( (1 + 1) )", linter)
  expect_no_lint("if (blah) { }", linter)
  expect_no_lint("for (i in j) { }", linter)
  expect_no_lint("1 * (1 + 1)", linter)
  expect_no_lint("!(1 == 1)", linter)
  expect_no_lint("(2 - 1):(3 - 1)", linter)
  expect_no_lint("c(1, 2, 3)[(2 - 1)]", linter)
  expect_no_lint("list(1, 2, 3)[[(2 - 1)]]", linter)
  expect_no_lint("range(10)[(2 - 1):(10 - 1)]", linter)
  expect_no_lint("function(){function(){}}()()", linter)
  expect_no_lint("c(function(){})[1]()", linter)

  expect_no_lint("\"test <- function(x) { if(1 + 1) 'hi' }\"", linter)
  expect_no_lint("res <- c((mat - 1L) %*% combs + 1L)", linter)
  expect_no_lint("if (!(foo && bar || baz)) { foo }", linter)
  expect_no_lint("x^(y + z)", linter)
  expect_no_lint("x**(y + z)", linter)
  expect_no_lint("a <- -(b)", linter)

  expect_no_lint("(3^(3 + 2))", linter)
  expect_no_lint("-(!!!symb)", linter)

  expect_no_lint("'[[<-.data.frame'(object, y)", linter)
  expect_no_lint("object@data@get('input')", linter)
  expect_no_lint("x <- ~(. + y)", linter)
  # the internal newline is required to trigger the lint
  expect_no_lint("if (x > 1)\n  x <- x[-(i)]", linter)
  # these don't violate the linter, even if they are strange coding practice
  expect_no_lint("for (ii in 1:10) next()", linter)
  expect_no_lint("for (ii in 1:10) break()", linter)
})

test_that("spaces_left_parentheses_linter blocks disallowed usages", {
  linter <- spaces_left_parentheses_linter()
  lint_msg <- rex::rex("Place a space before left parenthesis, except in a function call.")

  expect_lint("if(blah) { }", lint_msg, linter)
  expect_lint("for(i in j) { }", lint_msg, linter)
  expect_lint("1*(1 + 1)", lint_msg, linter)
  expect_lint("test <- function(x) { if(1 + 1) 'hi' }", lint_msg, linter)
  expect_lint("test <- function(x) { if(`+`(1, 1)) 'hi' }", lint_msg, linter)

  # more complicated cases for parse tree
  expect_lint("y1<-(abs(yn)>90)*1", lint_msg, linter)
  expect_lint("c(a,(a+b))", lint_msg, linter)
  expect_lint("if (x>y) 1 else(2)", lint_msg, linter)
  expect_lint("y~(x+z)", lint_msg, linter)
  expect_lint("if (x>y) {(x+y) / (x-y)}", lint_msg, linter)
  expect_lint("for (ii in(1:10)) { }", lint_msg, linter)
  expect_lint("x = 1;(x + 2)*3", lint_msg, linter)
  expect_lint("foo <- function(x=(1+2)) { }", lint_msg, linter)
})

test_that("doesn't produce a warning", {
  # this contains a function with nested if-statements where the conditional includes a unary minus.
  # This specific constellation with another if-statement at the same nesting level on the other enclosing if-branch
  # caused a warning in spaces_left_parentheses_linter (e.g. 84bc3a is broken)
  complex_lines <- trim_some("
    complexity <- function(x) {
      if (x > 0.0) {
        if (x > 10.0) {
          if (x > 20.0) {
            x <- x / 2.0
          } else {
            return(x)
          }
        } else {
          return(x)
        }
      } else {
        if (x < -10.0) {
          if (x < -20.0) {
            x <- x * 2.0
          } else {
            return(x)
          }
        } else {
          return(x)
        }
      }
      x
    }
  ")

  expect_no_warning(lint(text = complex_lines, linters = spaces_left_parentheses_linter()))
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Place a space before left parenthesis, except in a function call.")

  expect_lint(
    trim_some("{
      y1<-(abs(yn)>90)*1
      for(i in j) { }
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    spaces_left_parentheses_linter()
  )
})
# fuzzer enable: comment_injection

test_that("scalar_in_linter skips allowed usages", {
  linter <- scalar_in_linter()

  expect_no_lint("x %in% y", linter)
  expect_no_lint("y %in% c('a', 'b')", linter)
  expect_no_lint("c('a', 'b') %in% x", linter)
  expect_no_lint("z %in% 1:3", linter)
  # scalars on LHS are fine (often used as `"col" %in% names(DF)`)
  expect_no_lint("3L %in% x", linter)
  # this should be is.na(x), but it more directly uses the "always TRUE/FALSE, _not_ NA"
  #   aspect of %in%, so we delegate linting here to equals_na_linter()
  expect_no_lint("x %in% NA", linter)
  expect_no_lint("x %in% NA_character_", linter)
})

test_that("scalar_in_linter blocks simple disallowed usages", {
  linter <- scalar_in_linter(in_operators = c("%chin%", "%notin%"))
  lint_msg <- rex::rex("Use comparison operators (e.g. ==, !=, etc.) to match length-1 scalars instead of")

  expect_lint("x %in% 1", lint_msg, linter)
  expect_lint("x %chin% 'a'", lint_msg, linter)
  expect_lint("x %notin% 1", lint_msg, linter)
})

test_that("scalar_in_linter blocks or skips based on configuration", {
  linter_default <- scalar_in_linter()
  linter_config <- scalar_in_linter(in_operators = "%notin%")

  lint_msg <- rex::rex("Use comparison operators (e.g. ==, !=, etc.) to match length-1 scalars instead of")

  # default
  expect_lint("x %in% 1", lint_msg, linter_default)
  expect_no_lint("x %notin% 1", linter_default)
  expect_no_lint("x %notin% y", linter_default)

  # configured
  expect_lint("x %in% 1", lint_msg, linter_config)
  expect_lint("x %notin% 1", lint_msg, linter_config)
  expect_no_lint("x %notin% y", linter_config)
})

test_that("multiple lints are generated correctly", {
  linter <- scalar_in_linter(in_operators = "%chin%")

  expect_lint(
    trim_some('{
      x %in% 1
      y %chin% "a"
    }'),
    list(
      list("%in%", line_number = 2L),
      list("%chin%", line_number = 3L)
    ),
    linter
  )
})

# fuzzer disable: comment_injection
test_that("trailing_blank_lines_linter doesn't block allowed usages", {
  linter <- trailing_blank_lines_linter()

  expect_no_lint("blah", linter)
  expect_no_lint("blah <- 1  ", linter)
  expect_no_lint("blah <- 1\nblah", linter)
  expect_no_lint("blah <- 1\nblah\n \n blah", linter)

  tmp <- withr::local_tempfile(lines = "lm(y ~ x)")
  expect_no_lint(file = tmp, linters = linter)
})

test_that("trailing_blank_lines_linter detects disallowed usages", {
  linter <- trailing_blank_lines_linter()
  lint_msg <- rex::rex("Remove trailing blank lines.")

  expect_lint("blah <- 1\n", lint_msg, linter)
  expect_lint("blah <- 1\n  ", lint_msg, linter)
  expect_lint("blah <- 1\n \n ", list(lint_msg, lint_msg), linter)
  expect_lint("blah <- 1\n\n", list(lint_msg, lint_msg), linter)
  expect_lint("blah <- 1\n\t\n", list(lint_msg, lint_msg), linter)

  # Construct a test file without terminal newline
  # cf. test-get_source_expressions.R
  tmp2 <- withr::local_tempfile()
  cat("lm(y ~ x)", file = tmp2)
  expect_lint(
    file = tmp2,
    checks = list(
      message = rex::rex("Add a terminal newline."),
      line_number = 1L,
      column_number = 10L
    ),
    linters = linter
  )
})

test_that("trailing_blank_lines_linter detects missing terminal newlines in Rmd/qmd docs", {
  linter <- trailing_blank_lines_linter()
  lint_msg <- rex::rex("Add a terminal newline")

  tmp3 <- withr::local_tempfile(fileext = ".Rmd")
  cat(
    trim_some(
      '---
      title: "Some file"
      ---

      ```{r}
      abc = 123
      ```

      ```{r child="some-file.Rmd"}
      ```'
    ),
    file = tmp3
  )
  expect_lint(
    file = tmp3,
    # We can't get 4 here because the line is NA-masked in get_source_expressions(), so no line length info exists.
    checks = list(lint_msg, line_number = 10L, column_number = 1L),
    linters = linter
  )

  # Construct an Rmd file without R code (#1415)
  tmp4 <- withr::local_tempfile(fileext = ".Rmd")
  cat(
    trim_some(
      '---
      title: "Some file"
      ---

      No code and no terminal newline'
    ),
    file = tmp4
  )
  expect_lint(
    file = tmp4,
    # We can't get 4 here because the line is NA-masked in get_source_expressions(), so no line length info exists.
    checks = list(lint_msg, line_number = 5L, column_number = 1L),
    linters = linter
  )

  # Construct a qmd file without terminal newline
  tmp5 <- withr::local_tempfile(fileext = ".qmd")
  cat(
    trim_some(
      '---
      title: "Some file"
      ---

      ```{r}
      abc = 123
      ```

      ```{r child="some-file.qmd"}
      ```'
    ),
    file = tmp5
  )
  expect_lint(
    file = tmp5,
    # We can't get 4 here because the line is NA-masked in get_source_expressions(), so no line length info exists.
    checks = list(lint_msg, line_number = 10L, column_number = 1L),
    linters = linter
  )
})

test_that("blank lines in knitr chunks produce lints", {
  linter <- trailing_blank_lines_linter()
  lint_msg <- rex::rex("Remove trailing blank lines.")

  tmp6 <- withr::local_tempfile(
    fileext = ".Rmd",
    lines = trim_some(
      '---
      title: "Some file"
      ---

      ```{r}
      abc = 123

      ```
      \n'
    )
  )

  expect_lint(
    file = tmp6,
    checks = list(lint_msg, line_number = 7L, column_number = 1L),
    linters = linter
  )

  tmp7 <- withr::local_tempfile(
    fileext = ".qmd",
    lines = trim_some(
      '---
      title: "Some file"
      ---

      ```{r}
      abc = 123



      ```
      \n'
    )
  )

  expect_lint(
    file = tmp7,
    checks = list(
      list(lint_msg, line_number = 7L, column_number = 1L),
      list(lint_msg, line_number = 8L, column_number = 1L),
      list(lint_msg, line_number = 9L, column_number = 1L)
    ),
    linters = linter
  )
})
# fuzzer enable: comment_injection

test_that("redundant_equals_linter skips allowed usages", {
  linter <- redundant_equals_linter()

  # comparisons to non-logical constants
  expect_no_lint("x == 1", linter)
  # comparison to TRUE as a string
  expect_no_lint("x != 'TRUE'", linter)
})

test_that("multiple lints return correct custom messages", {
  expect_lint(
    trim_some("
      list(
        x == TRUE,
        y != TRUE
      )
    "),
    list(
      list("Using == on a logical vector", line_number = 2L),
      list("Using != on a logical vector", line_number = 3L)
    ),
    redundant_equals_linter()
  )
})

test_that("Order doesn't matter", {
  expect_lint("TRUE == x", rex::rex("Using == on a logical vector is redundant."), redundant_equals_linter())
})

patrick::with_parameters_test_that(
  "redundant_equals_linter blocks simple disallowed usages",
  {
    lint_msg <- rex::rex(paste("Using", op, "on a logical vector is redundant."))
    code <- paste("x", op, bool)
    expect_lint(code, lint_msg, redundant_equals_linter())
  },
  .cases = tibble::tribble(
    ~.test_name, ~op,  ~bool,
    "==, TRUE",  "==", "TRUE",
    "==, FALSE", "==", "FALSE",
    "!=, TRUE",  "!=", "TRUE",
    "!=, FALSE", "!=", "FALSE"
  )
)

test_that("logic survives adversarial comments", {
  expect_lint(
    trim_some("
      list(x #
      == TRUE)
    "),
    "==",
    redundant_equals_linter()
  )
})

test_that("string_boundary_linter skips allowed grepl() usages", {
  linter <- string_boundary_linter()

  # no known start/end anchor --> no lint
  expect_no_lint("grepl(p1, x)", linter)
  # no start/end anchor --> no lint
  expect_no_lint("grepl('abc', x)", linter)
  # regex pattern --> no lint
  expect_no_lint("grepl('^[a-z]', x)", linter)
  expect_no_lint("grepl('[a-z]$', x)", linter)
  # anchor-ish but not a regex --> no lint (#2636)
  expect_no_lint(R"[grepl("\\^", x)]", linter)
  expect_no_lint(R"[grepl("\\\\^", x)]", linter)
  expect_no_lint(R"[grepl("$\\\\", x)]", linter)

  # ignore.case --> no lint
  expect_no_lint("grepl('^abc', x, ignore.case = TRUE)", linter)
  expect_no_lint("grepl('^abc', x, ignore.case = ignore.case)", linter)

  # fixed --> no lint
  expect_no_lint("grepl('^abc', x, fixed = TRUE)", linter)
  expect_no_lint("grepl('^abc', x, fixed = fixed)", linter)
})

test_that("string_boundary_linter skips allowed str_detect() usages", {
  linter <- string_boundary_linter()

  # no known start/end anchor --> no lint
  expect_no_lint("str_detect(x, p1)", linter)
  # no start/end anchor --> no lint
  expect_no_lint("str_detect(x, 'abc')", linter)
  # regex pattern --> no lint
  expect_no_lint("str_detect(x, '^[a-z]')", linter)
  expect_no_lint("str_detect(x, '[a-z]$')", linter)
})

test_that("string_boundary_linter skips allowed substr()/substring() usages", {
  linter <- string_boundary_linter()

  # no comparison operator --> no lint
  expect_no_lint("substr(x, start, end)", linter)
  # unknown indices --> no lint
  expect_no_lint("substr(x, start, end) == 'a'", linter)
  expect_no_lint("substring(x, start, end) == 'a'", linter)
  # using foo(nchar(.))
  expect_no_lint("substring(x, nchar(x) - 4, nchar(x) - 1) == 'abc'", linter)
  # using nchar(), but not of the input
  expect_no_lint("substring(x, nchar(y) - 4, nchar(y)) == 'abcd'", linter)
  # using x in nchar(), but on foo(input)
  expect_no_lint("substring(x, nchar(foo(x)) - 4, nchar(foo(x))) == 'abcd'", linter)

  # _close_ to equivalent, but not so in general -- e.g.
  #   substring(s <- "abcdefg", 2L) == "efg" is not TRUE, but endsWith(s, "efg")
  #   is. And if `s` contains strings of varying lengths, there's no equivalent.
  expect_no_lint("substring(x, 2L)", linter)
})

test_that("string_boundary_linter blocks simple disallowed grepl() usages", {
  linter <- string_boundary_linter()
  starts_message <- rex::rex("Use !is.na(x) & startsWith(x, string) to detect a fixed initial substring,")
  ends_message <- rex::rex("Use !is.na(x) & endsWith(x, string) to detect a fixed terminal substring,")

  expect_lint("grepl('^a', x)", starts_message, linter)
  # non-trivially equivalent (but still same as startsWith())
  expect_lint("grepl('^[.]', x)", starts_message, linter)
  expect_lint("grepl('a$', x)", ends_message, linter)
  # also get negation for free
  expect_lint("!grepl('a$', x)", ends_message, linter)

  # perl = TRUE doesn't matter
  expect_lint("grepl('^a', x, perl = TRUE)", starts_message, linter)
  # explicit FALSE (i.e., an explicit default) is ignored
  expect_lint("grepl('^a', x, fixed = FALSE)", starts_message, linter)
  expect_lint("grepl('^a', x, fixed = F)", starts_message, linter)
})

test_that("string_boundary_linter blocks simple disallowed str_detect() usages", {
  linter <- string_boundary_linter()

  expect_lint(
    "str_detect(x, '^a')",
    rex::rex("Use startsWith() to detect a fixed initial substring."),
    linter
  )
  expect_lint(
    "str_detect(x, 'a$')",
    rex::rex("Use endsWith() to detect a fixed terminal substring."),
    linter
  )
})

test_that("string_boundary_linter blocks disallowed substr()/substring() usage", {
  linter <- string_boundary_linter()
  starts_message <- rex::rex("Use startsWith() to detect an initial substring.")
  ends_message <- rex::rex("Use endsWith() to detect a terminal substring.")

  expect_lint("substr(x, 1L, 2L) == 'ab'", starts_message, linter)
  # end doesn't matter, just anchoring to 1L
  expect_lint("substr(x, 1L, end) == 'ab'", starts_message, linter)
  expect_lint("substring(x, nchar(x) - 4L, nchar(x)) == 'abcde'", ends_message, linter)
  # start doesn't matter, just anchoring to nchar(x)
  expect_lint("substring(x, start, nchar(x)) == 'abcde'", ends_message, linter)
  # more complicated expressions
  expect_lint("substring(colnames(x), start, nchar(colnames(x))) == 'abc'", ends_message, linter)

  # adversarial comments
  expect_lint(
    trim_some("
      substring(colnames(x), start, nchar(colnames( # comment
      x))) == 'abc'
    "),
    ends_message,
    linter
  )
})

test_that("plain ^ or $ are skipped", {
  linter <- string_boundary_linter()

  expect_no_lint('grepl("^", x)', linter)
  expect_no_lint('grepl("$", x)', linter)
})

test_that("substr inverted tests are caught as well", {
  linter <- string_boundary_linter()

  expect_lint(
    "substr(x, 1L, 2L) != 'ab'",
    rex::rex("Use startsWith() to detect an initial substring."),
    linter
  )
  expect_lint(
    "substring(x, nchar(x) - 4L, nchar(x)) != 'abcde'",
    rex::rex("Use endsWith() to detect a terminal substring."),
    linter
  )
})

test_that("raw strings are detected", {
  linter <- string_boundary_linter()

  expect_no_lint('grepl(R"(^.{3})", x)', linter)
  expect_lint(
    'grepl(R"(^abc)", x)',
    rex::rex("Use !is.na(x) & startsWith(x, string) to detect a fixed initial substring,"),
    linter
  )
})

test_that("grepl() can optionally be ignored", {
  linter <- string_boundary_linter(allow_grepl = TRUE)

  expect_no_lint("grepl('^abc', x)", linter)
  expect_no_lint("grepl('xyz$', x)", linter)
})

test_that("whole-string regex recommends ==, not {starts,ends}With()", {
  linter <- string_boundary_linter()
  lint_msg <- rex::rex("Use == to check for an exact string match.")

  expect_lint("grepl('^abc$', x)", lint_msg, linter)
  expect_lint("grepl('^a\\\\.b$', x)", lint_msg, linter)
  expect_lint("str_detect(x, '^abc$')", lint_msg, linter)
  expect_lint("str_detect(x, '^a[.]b$')", lint_msg, linter)
})

test_that("vectorization + metadata work as intended", {
  linter <- string_boundary_linter()

  expect_lint(
    trim_some("{
      substring(a, 1, 3) == 'abc'
      substring(b, nchar(b) - 3, nchar(b)) == 'defg'
      substr(c, 1, 3) == 'hij'
      substr(d, nchar(d) - 3, nchar(d)) == 'klmn'
      grepl('^abc', e)
      grepl('abc$', f)
      grepl('^abc$', g)
      str_detect(h, '^abc')
      str_detect(i, 'abc$')
      str_detect(j, '^abc$')
    }"),
    list(
      list("startsWith", line_number = 2L),
      list("endsWith", line_number = 3L),
      list("startsWith", line_number = 4L),
      list("endsWith", line_number = 5L),
      list("startsWith", line_number = 6L),
      list("endsWith", line_number = 7L),
      list("==", line_number = 8L),
      list("startsWith", line_number = 9L),
      list("endsWith", line_number = 10L),
      list("==", line_number = 11L)
    ),
    linter
  )

  # some but not all anchor-esque as in #2636
  expect_lint(
    trim_some(R"[{
      grepl("\\^", x)
      grepl("^abc", x)
    }]"),
    list("startsWith", line_number = 3L),
    linter
  )
})

test_that("system_file_linter skips allowed usages", {
  linter <- system_file_linter()

  expect_no_lint("system.file('a', 'b', 'c')", linter)
  expect_no_lint("file.path('a', 'b', 'c')", linter)
})

test_that("system_file_linter blocks simple disallowed usages", {
  linter <- system_file_linter()
  lint_msg <- rex::rex("Use the `...` argument of system.file() to expand paths")

  expect_lint("system.file(file.path('path', 'to', 'data'), package = 'foo')", lint_msg, linter)
  expect_lint("file.path(system.file(package = 'foo'), 'path', 'to', 'data')", lint_msg, linter)
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Use the `...` argument of system.file() to expand paths")

  expect_lint(
    trim_some("{
      file.path(system.file(package = 'foo'), 'bar')
      system.file(file.path('bar', 'data'), package = 'foo')
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    system_file_linter()
  )
})

test_that("coalesce_linter skips allowed usage", {
  linter <- coalesce_linter()

  expect_no_lint("if (is.null(x)) y", linter)
  expect_no_lint("if (!is.null(x)) y", linter)
  expect_no_lint("if (!is.null(x)) x", linter)
  expect_no_lint("if (is.null(x)) x", linter)
  expect_no_lint("c(if (!is.null(E)) E)", linter)
  expect_no_lint("if (is.null(x)) y else z", linter)
  expect_no_lint("if (!is.null(x)) x[1] else y", linter)
  expect_no_lint("if (is.null(x[1])) y else x[2]", linter)
  expect_no_lint("if (is.null(x)) y else {x ; z}", linter)
  expect_no_lint("if (is.null(x)) y else {x \n z}", linter)
  expect_no_lint("if (!is.null(x)) {x ; z} else y", linter)
  expect_no_lint("if (!is.null(x)) {x \n z} else y", linter)

  expect_no_lint("if (is.null(s <- foo())) y else x", linter)
  expect_no_lint("if (!is.null(s <- foo())) x else y", linter)
})

test_that("coalesce_linter blocks simple disallowed usage", {
  linter <- coalesce_linter()
  lint_msg <- rex::rex("Use x %||% y instead of if (is.null(x))")
  lint_msg_not <- rex::rex("Use x %||% y instead of if (!is.null(x))")

  expect_lint("if (is.null(x)) y else x", lint_msg, linter)
  expect_lint("if (is.null(x)) { y } else x", lint_msg, linter)
  expect_lint("if (is.null(x)) y else { x }", lint_msg, linter)
  expect_lint("if (is.null(x)) { y } else { x }", lint_msg, linter)

  expect_lint("if (is.null(x[1])) y else x[1]", lint_msg, linter)
  expect_lint("if (is.null(foo(x))) y else foo(x)", lint_msg, linter)

  expect_lint("if (!is.null(x)) x else y", lint_msg_not, linter)
  expect_lint("if (!is.null(x)) { x } else y", lint_msg_not, linter)
  expect_lint("if (!is.null(x)) x else { y }", lint_msg_not, linter)
  expect_lint("if (!is.null(x)) { x } else { y }", lint_msg_not, linter)

  expect_lint("if (!is.null(x[1])) x[1] else y", lint_msg_not, linter)
  expect_lint("if (!is.null(foo(x))) foo(x) else y", lint_msg_not, linter)

  # adversarial comments
  expect_lint(
    trim_some("
      if (!is.null(x[1])) x[ # comment
      1] else y
    "),
    lint_msg_not,
    linter
  )
})

test_that("coalesce_linter blocks usage with implicit assignment", { # nofuzz: assignment
  linter <- coalesce_linter()
  lint_msg <- rex::rex("Use x %||% y instead of if (is.null(x))")
  lint_msg_not <- rex::rex("Use x %||% y instead of if (!is.null(x))")

  expect_lint("if (is.null(s <- foo(x))) y else s", lint_msg, linter)
  expect_lint("if (is.null(s <- foo(x))) { y } else s", lint_msg, linter)
  expect_lint("if (is.null(s <- foo(x))) y else { s }", lint_msg, linter)
  expect_lint("if (is.null(s <- foo(x))) { y } else { s }", lint_msg, linter)

  expect_lint("if (!is.null(s <- foo(x))) s else y", lint_msg_not, linter)
  expect_lint("if (!is.null(s <- foo(x))) { s } else y", lint_msg_not, linter)
  expect_lint("if (!is.null(s <- foo(x))) s else { y }", lint_msg_not, linter)
  expect_lint("if (!is.null(s <- foo(x))) { s } else { y }", lint_msg_not, linter)
})

test_that("lints vectorize", { # nofuzz: assignment
  expect_lint(
    trim_some("{
      if (is.null(x)) y else x
      if (!is.null(a)) a else b
      if (is.null(s <- foo(x))) y else s
      if (!is.null(t <- bar(a))) t else b
    }"),
    list(
      list(rex::rex("if (is.null(x))"), line_number = 2L, column_number = 3L),
      list(rex::rex("if (!is.null(x))"), line_number = 3L, column_number = 3L),
      list(rex::rex("if (is.null(x))"), line_number = 4L, column_number = 3L),
      list(rex::rex("if (!is.null(x))"), line_number = 5L, column_number = 3L)
    ),
    coalesce_linter()
  )
})

test_that("consecutive_assertion_linter skips allowed usages", {
  linter <- consecutive_assertion_linter()
  expect_no_lint("stopifnot(x)", linter)
  expect_no_lint("stopifnot(x, y, z)", linter)

  # intervening expression
  expect_no_lint("stopifnot(x); y; stopifnot(z)", linter)

  # inline or potentially with gaps don't matter
  expect_no_lint(
    trim_some("
      stopifnot(x)
      y

      stopifnot(z)
    "),
    linter
  )
})

test_that("consecutive_assertion_linter blocks simple disallowed usages", {
  linter <- consecutive_assertion_linter()
  lint_msg <- rex::rex("Unify consecutive calls to stopifnot().")

  # one test of inline usage
  expect_lint(
    "stopifnot(x); stopifnot(y)",
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      stopifnot(x)

      stopifnot(y, z)
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      stopifnot(x)
      stopifnot(y)
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      stopifnot(x)
      # a comment on y
      stopifnot(y)
    "),
    lint_msg,
    linter
  )
})

test_that("assert_that usages are handled correctly too", {
  linter <- consecutive_assertion_linter()
  lint_msg <- rex::rex("Unify consecutive calls to assert_that().")

  expect_no_lint("assert_that(x)", linter)
  expect_no_lint("assertthat::assert_that(x, y, z)", linter)

  # if msg= is used, can't necessarily combine
  lines <- trim_some("
    assert_that(x, msg = 'bad x')
    assert_that(y, msg = 'bad y')
  ")
  expect_no_lint(lines, linter)

  # one test of inline usage
  expect_lint(
    "assert_that(x); assert_that(y)",
    lint_msg,
    linter
  )

  lines_gap <- trim_some("
    assert_that(x)

    assertthat::assert_that(y, z)
  ")
  expect_lint(lines_gap, lint_msg, linter)
})

test_that("Mixing test functions is fine", {
  expect_no_lint(
    trim_some("
      assert_that(x)
      stopifnot(y)
    "),
    consecutive_assertion_linter()
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      stopifnot(A)
      stopifnot(B)
      assert_that(C)
      assert_that(D)
    }"),
    list(
      list("stopifnot", line_number = 2L),
      list("assert_that", line_number = 4L)
    ),
    consecutive_assertion_linter()
  )
})

test_that("interceding = assignments aren't linted", {
  expect_no_lint(
    trim_some("{
      stopifnot(A)
      x = 1
      stopifnot(B)

      assert_that(C)
      z = 3
      assert_that(D)
    }"),
    consecutive_assertion_linter()
  )
})

test_that("conjunct_test_linter skips allowed usages of expect_true", {
  linter <- conjunct_test_linter()

  expect_no_lint("expect_true(x)", linter)
  expect_no_lint("testthat::expect_true(x, y, z)", linter)

  # more complicated expression
  expect_no_lint("expect_true(x || (y && z))", linter)
  # the same by operator precedence, though not obvious a priori
  expect_no_lint("expect_true(x || y && z)", linter)
  expect_no_lint("expect_true(x && y || z)", linter)
})

test_that("conjunct_test_linter skips allowed usages of expect_true", {
  linter <- conjunct_test_linter()

  expect_no_lint("expect_false(x)", linter)
  expect_no_lint("testthat::expect_false(x, y, z)", linter)

  # more complicated expression
  # (NB: xx && yy || zz and xx || yy && zz both parse with || first)
  expect_no_lint("expect_false(x && (y || z))", linter)
})

test_that("conjunct_test_linter blocks && conditions with expect_true()", {
  linter <- conjunct_test_linter()
  lint_msg <-
    rex::rex("Write multiple expectations like expect_true(A) and expect_true(B) instead of expect_true(A && B)")

  expect_lint("expect_true(x && y)", lint_msg, linter)
  expect_lint("expect_true(x && y && z)", lint_msg, linter)
})

test_that("conjunct_test_linter blocks || conditions with expect_false()", {
  linter <- conjunct_test_linter()
  lint_msg <-
    rex::rex("Write multiple expectations like expect_false(A) and expect_false(B) instead of expect_false(A || B)")

  expect_lint("expect_false(x || y)", lint_msg, linter)
  expect_lint("expect_false(x || y || z)", lint_msg, linter)

  # these lint because `||` is always outer by operator precedence
  expect_lint("expect_false(x || y && z)", lint_msg, linter)
  expect_lint("expect_false(x && y || z)", lint_msg, linter)
})

test_that("conjunct_test_linter skips allowed stopifnot() and assert_that() usages", {
  linter <- conjunct_test_linter()

  expect_no_lint("stopifnot(x)", linter)
  expect_no_lint("assert_that(x, y, z)", linter)

  # more complicated expression
  expect_no_lint("stopifnot(x || (y && z))", linter)
  # the same by operator precedence, though not obvious a priori
  expect_no_lint("stopifnot(x || y && z)", linter)
  expect_no_lint("assertthat::assert_that(x && y || z)", linter)
})

test_that("conjunct_test_linter blocks simple disallowed usages of stopifnot() and assert_that()", {
  linter <- conjunct_test_linter()
  lint_msg <- function(fun) rex::rex("Write multiple conditions like ", fun, "(A, B) instead of ", fun, "(A && B)")

  expect_lint("stopifnot(x && y)", lint_msg("stopifnot"), linter)
  expect_lint("stopifnot(x && y && z)", lint_msg("stopifnot"), linter)

  # assert_that() versions
  expect_lint("assert_that(x && y)", lint_msg("assert_that"), linter)
  expect_lint("assertthat::assert_that(x && y && z)", lint_msg("assert_that"), linter)
})

test_that("conjunct_test_linter's allow_named_stopifnot argument works", {
  linter <- conjunct_test_linter()

  # allowed by default
  expect_no_lint(
    "stopifnot('x must be a logical scalar' = length(x) == 1 && is.logical(x) && !is.na(x))",
    linter
  )
  # including with intervening comment
  expect_no_lint(
    trim_some("
      stopifnot('x must be a logical scalar' = # comment
        length(x) == 1 && is.logical(x) && !is.na(x)
      )
    "),
    linter
  )

  expect_lint(
    "stopifnot('x is a logical scalar' = length(x) == 1 && is.logical(x) && !is.na(x))",
    rex::rex("Write multiple conditions like stopifnot(A, B)"),
    conjunct_test_linter(allow_named_stopifnot = FALSE)
  )
})

test_that("conjunct_test_linter skips allowed usages", {
  linter <- conjunct_test_linter()

  expect_no_lint("dplyr::filter(DF, A, B)", linter)
  expect_no_lint("dplyr::filter(DF, !(A & B))", linter)
  # | is the "top-level" operator here
  expect_no_lint("dplyr::filter(DF, A & B | C)", linter)
  expect_no_lint("dplyr::filter(DF, A | B & C)", linter)
})

test_that("conjunct_test_linter blocks simple disallowed usages", {
  linter <- conjunct_test_linter()
  lint_msg <- rex::rex("Use dplyr::filter(DF, A, B) instead of dplyr::filter(DF, A & B)")

  expect_lint("dplyr::filter(DF, A & B)", lint_msg, linter)
  expect_lint("dplyr::filter(DF, A & B & C)", lint_msg, linter)

  # more common usage, in pipes
  expect_lint("DF %>% dplyr::filter(A & B)", lint_msg, linter)
})

test_that("conjunct_test_linter respects its allow_filter argument", {
  linter_always <- conjunct_test_linter(allow_filter = "always")
  linter_dplyr <- conjunct_test_linter(allow_filter = "not_dplyr")
  lint_msg <- rex::rex("Use dplyr::filter(DF, A, B) instead of dplyr::filter(DF, A & B)")

  expect_no_lint("dplyr::filter(DF, A & B)", linter_always)
  expect_no_lint("dplyr::filter(DF, A & B & C)", linter_always)
  expect_no_lint("DF %>% dplyr::filter(A & B)", linter_always)
  expect_lint("dplyr::filter(DF, A & B)", lint_msg, linter_dplyr)
  expect_lint("dplyr::filter(DF, A & B & C)", lint_msg, linter_dplyr)
  expect_lint("DF %>% dplyr::filter(A & B)", lint_msg, linter_dplyr)
  expect_no_lint("filter(DF, A & B)", linter_dplyr)
  expect_no_lint("filter(DF, A & B & C)", linter_dplyr)
  expect_no_lint("DF %>% filter(A & B)", linter_dplyr)
})

test_that("filter() is assumed to be dplyr::filter() by default, unless o/w specified", {
  linter <- conjunct_test_linter()

  expect_no_lint("stats::filter(A & B)", linter)
  expect_no_lint("ns::filter(A & B)", linter)
  expect_lint(
    "DF %>% filter(A & B)",
    rex::rex("Use dplyr::filter(DF, A, B) instead of dplyr::filter(DF, A & B)"),
    linter
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      stopifnot(A && B)
      filter(DF, A & B)
    }"),
    list(
      list("stopifnot", line_number = 2L),
      list("filter", line_number = 3L)
    ),
    conjunct_test_linter()
  )
})

test_that("sample_int_linter skips allowed usages", {
  linter <- sample_int_linter()

  expect_no_lint("sample(n, m)", linter)
  expect_no_lint("sample(n, m, TRUE)", linter)
  expect_no_lint("sample(n, m, prob = 1:n/n)", linter)
  expect_no_lint("sample(foo(x), m, TRUE)", linter)
  expect_no_lint("sample(n, replace = TRUE)", linter)

  expect_no_lint("sample(10:1, m)", linter)
})

test_that("sample_int_linter blocks simple disallowed usages", {
  linter <- sample_int_linter()
  lint_msg <- rex::rex("sample.int(n, m, ...) is preferable to sample(1:n, m, ...).")

  expect_lint("sample(1:10, 2)", lint_msg, linter)
  # also matches literal integer
  expect_lint("sample(1L:10L, 2)", lint_msg, linter)
  expect_lint("sample(1:n, 2)", lint_msg, linter)
  expect_lint("sample(1:k, replace = TRUE)", lint_msg, linter)
  expect_lint("sample(1:foo(x), prob = bar(x))", lint_msg, linter)
})

test_that("sample_int_linter blocks sample(seq_len(n), ...) as well", {
  expect_lint(
    "sample(seq_len(10), 2)",
    rex::rex("sample.int(n, m, ...) is preferable to sample(seq_len(n), m, ...)."),
    sample_int_linter()
  )
})

test_that("sample_int_linter blocks sample(seq(n)) and sample(seq(1, ...))", {
  linter <- sample_int_linter()
  lint_msg <- rex::rex("sample.int(n, m, ...) is preferable to sample(seq(n), m, ...).")

  expect_lint("sample(seq(n), 5)", lint_msg, linter)
  expect_lint("sample(seq(1, 10), 5)", lint_msg, linter)
  expect_lint("sample(seq(1, 10, by = 1), 5)", lint_msg, linter)
  expect_lint("sample(seq(1L, 10, by = 1L), 5)", lint_msg, linter)

  # lint doesn't apply when by= is used (except when set to literal 1)
  expect_no_lint("sample(seq(1, 10, by = 2), 5)", linter)
  expect_no_lint("sample(seq(1, 10, by = n), 5)", linter)
})

test_that("sample_int_linter catches literal integer/numeric in the first arg", {
  linter <- sample_int_linter()
  lint_msg <- rex::rex("sample.int(n, m, ...) is preferable to sample(n, m, ...).")

  expect_lint("sample(10L, 4)", lint_msg, linter)
  expect_lint("sample(10, 5)", lint_msg, linter)
})

test_that("sample_int_linter skips TRUE or FALSE in the first argument", {
  linter <- sample_int_linter()

  expect_no_lint("sample(replace = TRUE, letters)", linter)
  expect_no_lint("sample(replace = FALSE, letters)", linter)
})

test_that("sample_int_linter skips x$sample() usage", {
  linter <- sample_int_linter()
  lint_msg <- rex::rex("sample.int(n, m, ...) is preferable to sample(n, m, ...).")

  expect_no_lint("foo$sample(1L)", linter)
  expect_no_lint("foo$sample(1:10)", linter)
  expect_no_lint("foo$sample(seq_len(10L))", linter)
  # ditto for '@' slot extraction
  expect_no_lint("foo@sample(1L)", linter)

  # however, base::sample qualification is still caught
  expect_lint("base::sample(10L)", lint_msg, linter)

  # but also, not everything "below" a $ extraction is skipped
  expect_lint("foo$bar(sample(10L))", lint_msg, linter)
})

test_that("multiple lints are generated correctly", {
  expect_lint(
    trim_some("{
      sample(1:10, 2)
      sample(10, 2)
      sample(seq_len(10), 2)
      sample(seq(10), 2)
    }"),
    list(
      list(rex::rex("sample(1:n"), line_number = 2L, column_number = 3L),
      list(rex::rex("sample(n"), line_number = 3L, column_number = 3L),
      list(rex::rex("sample(seq_len(n)"), line_number = 4L, column_number = 3L),
      list(rex::rex("sample(seq(n)"), line_number = 5L, column_number = 3L)
    ),
    sample_int_linter()
  )
})

test_that("expect_not_linter skips allowed usages", {
  expect_no_lint("expect_true(x)", expect_not_linter())
  # NB: also applies to tinytest, but it's sufficient to test testthat
  expect_no_lint("testthat::expect_true(x)", expect_not_linter())
  expect_no_lint("expect_false(x)", expect_not_linter())
  expect_no_lint("testthat::expect_false(x)", expect_not_linter())

  # not a strict ban on !
  ##  (expect_false(x && y) is the same, but it's not clear which to prefer)
  expect_no_lint("expect_true(!x || !y)", expect_not_linter())
})

test_that("expect_not_linter blocks simple disallowed usages", {
  linter <- expect_not_linter()
  lint_msg <- rex::rex("expect_false(x) is better than expect_true(!x), and vice versa.")

  expect_lint("expect_true(!x)", lint_msg, linter)
  expect_lint("testthat::expect_true(!x)", lint_msg, linter)
  expect_lint("expect_false(!foo(x))", lint_msg, linter)
  expect_lint("testthat::expect_true(!(x && y))", lint_msg, linter)
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("expect_false(x) is better than expect_true(!x), and vice versa.")

  expect_lint(
    trim_some("{
      expect_true(!x)
      expect_false(!y)
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    expect_not_linter()
  )
})

test_that("pipe_return_linter skips allowed usages", {
  linter <- pipe_return_linter()

  normal_pipe_lines <- trim_some("
    x %>%
      filter(str > 5) %>%
      summarize(str = sum(str))
  ")
  expect_no_lint(normal_pipe_lines, linter)

  normal_function_lines <- trim_some("
    pipeline <- function(x) {
      out <- x %>%
        filter(str > 5) %>%
        summarize(str = sum(str))
      return(out)
    }
  ")
  expect_no_lint(normal_function_lines, linter)

  nested_return_lines <- trim_some("
    pipeline <- function(x) {
      x_squared <- x %>%
        sapply(function(xi) {
          return(xi ** 2)
        })
      return(x_squared)
    }
  ")
  expect_no_lint(nested_return_lines, linter)
})

test_that("pipe_return_linter blocks simple disallowed usages", {
  lines <- trim_some("
    pipeline <- function(x) {
      out <- x %>%
        filter(str > 5) %>%
        summarize(str = sum(str)) %>%
        return()
    }
  ")
  expect_lint(
    lines,
    rex::rex("Avoid return() as the final step of a magrittr pipeline"),
    pipe_return_linter()
  )
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Avoid return() as the final step of a magrittr pipeline")

  expect_lint(
    trim_some("{
      function(x) {
        x %>% return()
      }
      function(y) {
        y %>% return()
      }
    }"),
    list(
      list(lint_msg, line_number = 3L),
      list(lint_msg, line_number = 6L)
    ),
    pipe_return_linter()
  )
})

test_that("linter returns correct linting", {
  linter <- undesirable_operator_linter(op = c("$" = "As an alternative, use the `[[` accessor.", "<<-" = NA))
  msg_assign <- rex::rex("Avoid undesirable operator `<<-`.")
  msg_dollar <- rex::rex("Avoid undesirable operator `$`. As an alternative, use the `[[` accessor.")

  expect_no_lint("x <- foo:::getObj()", linter)
  expect_no_lint("cat(\"10$\")", linter)
  expect_lint(
    "a <<- log(10)",
    list(msg_assign, line_number = 1L, column_number = 3L),
    linter
  )
  expect_lint( # nofuzz: dollar_at
    "data$parsed == c(1, 2)",
    list(msg_dollar, line_number = 1L, column_number = 5L),
    linter
  )

  expect_no_lint("`%%`(10, 2)", linter)
})

test_that("undesirable_operator_linter handles '=' consistently", { # nofuzz: assignment
  linter <- undesirable_operator_linter(op = c("=" = "As an alternative, use '<-'"))

  expect_lint("a = 2L", rex::rex("Avoid undesirable operator `=`."), linter)
  expect_no_lint("lm(data = mtcars)", linter)
  expect_no_lint("function(a = 1) { }", linter)
})

test_that("undesirable_operator_linter handles infixes correctly", {
  linter_oo <- undesirable_operator_linter(list("%oo%" = NA))
  linter_mod <- undesirable_operator_linter(list("%%" = NA))
  expect_lint("a %oo% b", rex::rex("Avoid undesirable operator `%oo%`."), linter_oo)
  expect_no_lint("a %00% b", linter_oo)

  # somewhat special case: %% is in infix_metadata
  expect_lint(
    "foo(x %% y, x %/% y)",
    rex::rex("Avoid undesirable operator `%%`."),
    linter_mod
  )

  expect_lint(
    "`%%`(10, 2)",
    rex::rex("Avoid undesirable operator `%%`."),
    linter_mod
  )
})

test_that("undesirable_operator_linter vectorizes messages", {
  expect_lint(
    "x <<- c(pkg:::foo, bar %oo% baz, `->>`(a, 2))",
    list(
      rex::rex("Avoid undesirable operator `<<-`. It assigns"),
      rex::rex("Avoid undesirable operator `:::`. It accesses"),
      rex::rex("Avoid undesirable operator `%oo%`.", end),
      rex::rex("Avoid undesirable operator `->>`. It assigns")
    ),
    undesirable_operator_linter(modify_defaults(default_undesirable_operators, "%oo%" = NA))
  )
})

test_that("invalid inputs fail correctly", {
  expect_error(
    undesirable_operator_linter(c("***" = NA, NA)),
    rex::rex("Unnamed elements of `op` must not be missing", anything, "2")
  )
  expect_error(
    undesirable_operator_linter(op = NULL),
    "`op` must be a non-empty character vector",
    fixed = TRUE
  )
  expect_error(
    undesirable_operator_linter(op = character(0L)),
    "`op` must be a non-empty character vector",
    fixed = TRUE
  )

  expect_error(
    undesirable_operator_linter(c("***" = NA)),
    'Did not recognize any valid operators in request for: "***"',
    fixed = TRUE
  )
  expect_error(
    undesirable_operator_linter(c("***" = NA, "///" = NA)),
    'Did not recognize any valid operators in request for: "***" and "///"',
    fixed = TRUE
  )
})

test_that("Default recommendations can be specified multiple ways", {
  linter_na <- undesirable_operator_linter(c(`%f%` = NA))
  linter_unnamed1 <- undesirable_operator_linter("%f%")
  linter_unnamed2 <- undesirable_operator_linter(c("%f%", "%b%"))
  linter_mixed1 <- undesirable_operator_linter(c("%f%", `%b%` = "no %b%"))
  linter_mixed2 <- undesirable_operator_linter(c("%f%", `%b%` = NA))

  lint_message <- rex::rex("Avoid undesirable operator `%f%`")

  lint_str <- "a %f% b"
  expect_lint(lint_str, lint_message, linter_na)
  expect_lint(lint_str, lint_message, linter_unnamed1)
  expect_lint(lint_str, lint_message, linter_unnamed2)
  expect_lint(lint_str, lint_message, linter_mixed1)
  expect_lint(lint_str, lint_message, linter_mixed2)
})

test_that("call_is_undesirable = FALSE doesn't lint prefix notation", {
  linter <- undesirable_operator_linter(call_is_undesirable = FALSE)
  lint_message <- rex::rex("Avoid undesirable operator `:::`.")

  expect_no_lint("`:::`(utils, hasName)", linter)
  expect_no_lint("`<<-`(x, 1)", linter)

  expect_lint("utils:::hasName", lint_message, linter)
})

test_that("returns the correct linting", {
  linter <- todo_comment_linter()
  lint_msg <- rex::rex("Remove TODO comments.")

  expect_no_lint('a <- "you#need#to#fixme"', linter)
  expect_no_lint("# something todo", linter)
  expect_lint(
    "cat(x) ### fixme",
    list(lint_msg, line_number = 1L, column_number = 8L),
    linter
  )
  expect_lint(
    trim_some("
      #' fixme
      cat(x)
    "),
    list(lint_msg, line_number = 1L),
    linter
  )
  expect_lint(
    trim_some('
      x <- "1.0
      2.0 #FIXME
      3 #TODO 4"; y <- 2; z <- 3 # todo later
    '),
    list(lint_msg, line_number = 3L, column_number = 28L),
    linter
  )
  expect_lint(
    trim_some("
      function() {
        # TODO
        function() {
          # fixme
        }
      }
    "),
    list(
      list(message = lint_msg, line_number = 2L, column_number = 3L),
      list(message = lint_msg, line_number = 4L, column_number = 5L)
    ),
    linter
  )
})

test_that("except_regex= excludes valid TODO", {
  linter <- todo_comment_linter(except_regex = "TODO\\(#[0-9]+\\):")
  lint_msg <- rex::rex("Remove TODO comments.")

  expect_no_lint("foo() # TODO(#1234): Deprecate foo.", linter)
  expect_no_lint(
    trim_some("
      #' TODO(#1234): Deprecate foo.
      foo()
    "),
    linter
  )

  # Non-excepted lints
  expect_lint(
    trim_some("
      foo() # TODO()
      bar() # TODO(#567): Deprecate bar.
    "),
    list(lint_msg, line_number = 1L),
    linter
  )
  # Only TODO() is excepted
  mixed_lines <- trim_some("
    foo() # TODO(#1234): Deprecate foo.
    bar() # fixme(#567): Deprecate bar.
  ")

  expect_lint(mixed_lines, list(lint_msg, line_number = 2L), linter)
  expect_no_lint(
    mixed_lines,
    todo_comment_linter(except_regex = c("TODO\\(#[0-9]+\\):", "fixme\\(#[0-9]+\\):"))
  )

  # ignore captured groups
  expect_no_lint(
    "# TODO(a)",
    todo_comment_linter(except_regex = "TODO\\((a|abc)\\)")
  )
})

test_that("unnecessary_nesting_linter skips allowed usages", {
  linter <- unnecessary_nesting_linter()

  # parallel stops() and return()s are OK
  expect_no_lint(
    trim_some("
      if (A) {
        stop()
      } else {
        stop()
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (A) {
        return()
      } else {
        return()
      }
    "),
    linter
  )
})

test_that("Multiple if/else statements don't require unnesting", {
  # with further branches, reducing nesting might be less readable
  expect_no_lint(
    trim_some("
      if (x == 'a') {
        stop()
      } else if (x == 'b') {
        do_b()
      } else {
        stop()
      }
    "),
    unnecessary_nesting_linter()
  )
})

test_that("else-less if statements don't lint", {
  expect_no_lint(
    trim_some("
      if (x == 4) {
        msg <- 'failed'
        stop(msg)
      }
    "),
    unnecessary_nesting_linter()
  )
})

test_that("non-terminal expressions are not considered for the logic", {
  expect_no_lint(
    trim_some("
      if (x == 4) {
        x <- 5
        return(x)
      } else {
        return(x)
      }
    "),
    unnecessary_nesting_linter()
  )
})

test_that("parallels in further nesting are skipped", {
  linter <- unnecessary_nesting_linter()

  expect_no_lint(
    trim_some("
      if (length(bucket) > 1) {
        return(age)
      } else {
        age <- age / 2
        if (grepl('[0-9]', age)) {
          return(age)
        } else {
          return('unknown')
        }
      }
    "),
    linter
  )

  # same but with '='
  expect_no_lint(
    trim_some("
      if (length(bucket) > 1) {
        return(age)
      } else {
        age = age / 2
        if (grepl('[0-9]', age)) {
          return(age)
        } else {
          return('unknown')
        }
      }
    "),
    linter
  )
})

test_that("unnecessary_nesting_linter blocks if/else with one exit branch", {
  linter <- unnecessary_nesting_linter()
  linter_warning <- unnecessary_nesting_linter(branch_exit_calls = "warning")
  lint_msg <- function(exit) rex::rex("Reduce the nesting of this if/else statement", anything, exit, "()")

  expect_lint(
    trim_some("
      if (A) {
        stop()
      } else {
        B
      }
    "),
    lint_msg("stop"),
    linter
  )

  expect_lint(
    trim_some("
      if (A) {
        return()
      } else {
        B
      }
    "),
    lint_msg("return"),
    linter
  )

  # also find exits in the later branch
  expect_lint(
    trim_some("
      if (A) {
        B
      } else {
        stop()
      }
    "),
    lint_msg("stop"),
    linter
  )

  expect_lint(
    trim_some("
      if (A) {
        B
      } else {
        return()
      }
    "),
    lint_msg("return"),
    linter
  )

  expect_lint(
    trim_some("
      if (A) {
        B
      } else {
        warning()
      }
    "),
    lint_msg("warning"),
    linter_warning
  )

  stop_warning_lines <- trim_some("
    if (A) {
      stop('An error')
    } else {
      warning('A warning')
    }
  ")
  expect_lint(stop_warning_lines, lint_msg("stop"), linter)

  # Optionally consider 'warning' as an exit call --> no lint
  expect_no_lint(stop_warning_lines, linter_warning)
})

test_that("unnecessary_nesting_linter skips one-line functions", {
  linter <- unnecessary_nesting_linter()

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        return(x)
      }
    "),
    linter
  )

  # purrr anonymous functions also get skipped
  expect_no_lint(
    trim_some("
      purrr::map(x, ~ {
        .x
      })
    "),
    linter
  )

  # ditto short-hand lambda
  expect_no_lint(
    trim_some("
      foo <- \\(x) {
        return(x)
      }
    "),
    linter
  )
})

test_that("unnecessary_nesting_linter skips one-expression for loops", {
  linter <- unnecessary_nesting_linter()

  expect_no_lint(
    trim_some("
      for (i in 1:10) {
        print(i)
      }
    "),
    linter
  )

  # also for extended control flow functionality from packages
  expect_no_lint(
    trim_some("
      foreach (i = 1:10) %dopar% {
        print(i)
      }
    "),
    linter
  )
})

test_that("unnecessary_nesting_linter skips one-expression if and else clauses", {
  expect_no_lint(
    trim_some("
      if (TRUE) {
        x
      } else {
        y
      }
    "),
    unnecessary_nesting_linter()
  )
})

test_that("unnecessary_nesting_linter skips one-expression while loops", {
  expect_no_lint(
    trim_some("
      while (x < 10) {
        x <- x + 1
      }
    "),
    unnecessary_nesting_linter()
  )
})

test_that("unnecessary_nesting_linter skips one-expression repeat loops", {
  expect_no_lint(
    trim_some("
      repeat {
        x <- x + 1
      }
    "),
    unnecessary_nesting_linter()
  )
})

test_that("unnecessary_nesting_linter skips one-expression assignments by default", {
  linter <- unnecessary_nesting_linter()

  expect_no_lint(
    trim_some("
      {
        x <- foo()
      }
    "),
    linter
  )
  expect_no_lint(
    trim_some("
      {
        x = foo()
      }
    "),
    linter
  )
})

test_that("unnecessary_nesting_linter passes for multi-line braced expressions", {
  expect_no_lint(
    trim_some("
      tryCatch(
        {
          foo(x)
          bar(x)
        },
        error = identity
      )
    "),
    unnecessary_nesting_linter()
  )
})

test_that("unnecessary_nesting_linter skips if unbracing won't reduce nesting", { # nofuzz: comment_injection
  linter <- unnecessary_nesting_linter()

  expect_no_lint(
    trim_some("
      test_that('this works', {
        expect_true(TRUE)
      })
    "),
    linter
  )
  expect_no_lint(
    trim_some("
      DT[, {
        plot(x, y)
      }]
    "),
    linter
  )
  expect_no_lint(
    trim_some("
      DT[, x := {
        foo(x, y)
      }]
    "),
    linter
  )

  # NB: styler would re-style these anyway
  expect_no_lint(
    trim_some("
      tryCatch({
        foo()
      }, error = identity)
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      DT[{
        n <- .N - 1
        x[n] < y[n]
      }, j = TRUE, by = x]
    "),
    linter
  )

  # interaction of '=' assignment and 'loose' braces (?)
  expect_no_lint(
    trim_some("
      DT[
        {
          n = .N - 1
          x[n] < y[n]
        }
        , j = TRUE, by = x
      ]
    "),
    linter
  )
})

test_that("rlang's double-brace operator is skipped", {
  expect_no_lint(
    "rename(DF, col = {{ val }})",
    unnecessary_nesting_linter()
  )
})

test_that("unnecessary_nesting_linter blocks one-expression braced expressions", {
  expect_lint(
    trim_some("
      tryToCatch(
        {
          foo(x)
        },
        error = identity
      )
    "),
    rex::rex("Reduce the nesting of this statement by removing the braces {}."),
    unnecessary_nesting_linter()
  )
})

test_that("unnecessary_nesting_linter allow_assignment= argument works", {
  expect_lint(
    trim_some("
      tryToCatch(
        {
          idx <- foo(x)
        },
        error = identity
      )
    "),
    rex::rex("Reduce the nesting of this statement by removing the braces {}."),
    unnecessary_nesting_linter(allow_assignment = FALSE)
  )
})

test_that("lints vectorize", {
  lint_msg <- function(exit) rex::rex("Reduce the nesting of this if/else", anything, exit, "()")

  expect_lint(
    trim_some("{
      if (A) {
        stop('no')
      } else {
        0
      }
      if (B) {
        q('really no')
      } else {
        1
      }
    }"),
    list(
      list(lint_msg("stop"), line_number = 2L),
      list(lint_msg("q"), line_number = 7L)
    ),
    unnecessary_nesting_linter()
  )
})

test_that("unnecessary_nesting_linter skips allowed usages", {
  linter <- unnecessary_nesting_linter()

  expect_no_lint(
    trim_some("
      if (x && y) {
        1L
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      for (x in 1:3) {
        if (x && y) {
          1L
        }
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (x) {
        1L
      } else if (y) {
        2L
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (x) {
        1L
      } else {
        2L
        if (y) {
          3L
        }
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (if (x) TRUE else FALSE) {
        1L
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (x) {
        y <- x + 1L
        if (y) {
          1L
        }
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if ((x && y) || (if (x) TRUE else FALSE)) {
        1L
      }
    "),
    linter
  )

  # if there is any additional code between the inner and outer scopes, no lint
  expect_no_lint(
    trim_some("
      if (x && a) {
        y <- x + 1L
        if (y || b) {
          1L
        }
      }
    "),
    linter
  )

  # '=' operator also (which uses non-<expr> node), #2245
  expect_no_lint(
    trim_some("
      if (x && a) {
        y = x + 1L
        if (y || b) {
          1L
        }
      }
    "),
    linter
  )

  # but comments are irrelevant (they should be moved to another anchor)
  expect_lint(
    trim_some("
      if (x && a) {
        # comment1
        if (y || b) {
          1L
        }
        # comment2
      }
    "),
    "Combine this `if` statement with the one found at line 1",
    linter
  )

  expect_no_lint(
    trim_some("
      if (x) {
        if (y) {
          1L
        }
        y <- x + 1L
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (x) {
        y <- x + 1L
        if (y) {
          1L
        }
        y <- x
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (x) {
        y <- x + 1L
        {
          if (y) {
            1L
          }
        }
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (x) {
        {
           y <- x + 1L
           if (y) {
             1L
           }
        }
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (x) {
        {
          if (y) {
            1L
          }
        }
        y <- x + 1L
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (x) {
        {
          y <- x + 1L
          {
            if (y) {
              1L
            }
          }
        }
      }
    "),
    linter
  )
})

test_that("unnecessary_nesting_linter blocks disallowed usages", {
  lint_message <-
    function(l, c) rex::rex("Combine this `if` statement", anything, "line ", l, ", column ", c)
  linter <- unnecessary_nesting_linter()

  expect_lint(
    trim_some("
      if (x) {
        if (y) {
          1L
        }
      }
    "),
    lint_message(1L, 1L),
    linter
  )

  expect_lint(
    trim_some("
      if (x) {
        if (y) 1L
      }
    "),
    lint_message(1L, 1L),
    linter
  )

  expect_lint(
    trim_some("
      # comment
      # comment
      if (x && a) {
        if (y || b) {
          1L
        }
      }
    "),
    lint_message(3L, 1L),
    linter
  )

  expect_lint(
    trim_some("
      if (if (x) TRUE else FALSE) {
        if (y) {
          1L
        }
      }
    "),
    lint_message(1L, 1L),
    linter
  )

  expect_lint(
    "if (x) if (y) 1L",
    lint_message(1L, 1L),
    linter
  )

  expect_lint(
    trim_some("
      for (x in 1:3) {
        if (x) if (y) 1L
      }
    "),
    lint_message(2L, 3L),
    linter
  )

  expect_lint(
    trim_some("
      if (x) {
        if (y) {
          if (z) {
            1L
          }
        }
      }
    "),
    list(
      list(message = lint_message(1L, 1L), line_number = 2L, column_number = 3L),
      list(message = lint_message(2L, 3L), line_number = 3L, column_number = 5L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      if (a) {
        if (b) {
          if (c) {
            message('hi')
          }
          t <- 1L
        }
      }
    "),
    lint_message(1L, 1L),
    linter
  )
})

test_that("else that can drop braces is found", {
  linter <- unnecessary_nesting_linter()
  lint_msg <- rex::rex("Simplify this condition by using 'else if' instead of 'else { if.")

  expect_lint(
    trim_some("
      if (A) {
        1
      } else {
        if (B) {
          2
        } else {
          3
        }
      }
    "),
    list(lint_msg, line_number = 4L),
    linter
  )

  expect_lint(
    trim_some("
      if (A) {
        1
      } else if (B) {
        2
      } else {
        if (C) {
          3
        } else {
          4
        }
      }
    "),
    list(lint_msg, line_number = 6L),
    linter
  )

  expect_lint(
    trim_some("
      if (A) {
        1
      } else {
        if (B) {
          2
        } else {
          if (C) {
            3
          } else {
            4
          }
        }
      }
    "),
    list(
      list(lint_msg, line_number = 4L),
      list(lint_msg, line_number = 7L)
    ),
    linter
  )
})

patrick::with_parameters_test_that(
  "default allowed functions are skipped",
  expect_no_lint(sprintf("%s(x, {y}, z)", call), unnecessary_nesting_linter()),
  call = c(
    "test_that", "with_parameters_test_that",
    "switch",
    "try", "tryCatch", "withCallingHandlers",
    "quote", "bquote", "expression", "substitute",
    "observe", "observeEvent", "reactive",
    "renderCachedPlot", "renderDataTable", "renderImage", "renderPlot",
    "renderPrint", "renderTable", "renderText", "renderUI"
  )
)

test_that("allow_functions= works", { # nofuzz: comment_injection
  linter_default <- unnecessary_nesting_linter()
  linter_foo <- unnecessary_nesting_linter(allow_functions = "foo")
  expect_lint("foo(x, {y}, z)", "Reduce the nesting of this statement", linter_default)
  expect_no_lint("foo(x, {y}, z)", linter_foo)
  expect_no_lint("test_that('a', {y})", linter_default)
  expect_no_lint("that_that('b', {y})", linter_foo)
})

test_that("regex_subset_linter skips allowed usages", {
  linter <- regex_subset_linter()

  expect_no_lint("y[grepl(ptn, x)]", linter)
  expect_no_lint("x[grepl(ptn, foo(x))]", linter)
})

test_that("regex_subset_linter blocks simple disallowed usages", {
  linter <- regex_subset_linter()
  lint_msg <- rex::rex("Prefer grepv(pattern, x, ...)")

  expect_lint("x[grep(ptn, x)]", lint_msg, linter)
  expect_lint("names(y)[grepl(ptn, names(y), perl = TRUE)]", lint_msg, linter)
  expect_lint("names(foo(y))[grepl(ptn, names(foo(y)), fixed = TRUE)]", lint_msg, linter)

  # adversarial commenting
  expect_lint(
    trim_some("
      names(y #comment
      )[grepl(ptn, names(y), perl = TRUE)]
    "),
    lint_msg,
    linter
  )
})

test_that("regex_subset_linter skips grep/grepl subassignment", {
  linter <- regex_subset_linter()

  expect_no_lint("x[grep(ptn, x)] <- ''", linter)
  expect_no_lint("x[grepl(ptn, x)] <- ''", linter)
  expect_no_lint("x[grep(ptn, x, perl = TRUE)] = ''", linter)
  expect_no_lint("'' -> x[grep(ptn, x, ignore.case = TRUE)] = ''", linter)

  expect_no_lint(
    trim_some("
      x[grepl(ptn, x) # comment
      ] <- ''
    "),
    linter
  )
})

test_that("regex_subset_linter skips allowed usages for stringr equivalents", {
  linter <- regex_subset_linter()

  expect_no_lint("y[str_detect(x, ptn)]", linter)
  expect_no_lint("x[str_detect(foo(x), ptn)]", linter)
  expect_no_lint("x[str_detect(x, ptn)] <- ''", linter)
  expect_no_lint("x[str_detect(x, ptn)] <- ''", linter)
})

test_that("regex_subset_linter blocks disallowed usages for stringr equivalents", {
  linter <- regex_subset_linter()
  lint_msg <- rex::rex("Prefer stringr::str_subset(x, pattern) over")

  expect_lint("x[str_which(x, ptn)]", lint_msg, linter)
  expect_lint("names(y)[str_detect(names(y), ptn, negate = TRUE)]", lint_msg, linter)
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      x[grep(ptn, x)]
      y[str_detect(y, ptn)]
    }"),
    list(
      list(rex::rex("Prefer grep"), line_number = 2L),
      list(rex::rex("Prefer stringr::str_subset"), line_number = 3L)
    ),
    regex_subset_linter()
  )
})

test_that("unnecessary_lambda_linter skips allowed usages", {
  linter <- unnecessary_lambda_linter()
  expect_no_lint("lapply(DF, sum)", linter)
  expect_no_lint("apply(M, 1, sum, na.rm = TRUE)", linter)

  # the first argument may be ... or have a cumbersome name, so an anonymous
  #   function may be preferable (e.g. this is often the case for grep() calls)
  expect_no_lint("sapply(x, function(xi) foo(1, xi))", linter)
  expect_no_lint("sapply(x, function(xi) return(foo(1, xi)))", linter)

  # if the argument is re-used, that's also a no-go
  expect_no_lint("dendrapply(x, function(xi) foo(xi, xi))", linter)
  # at any nesting level
  expect_no_lint("parLapply(cl, x, function(xi) foo(xi, 2, bar(baz(xi))))", linter)

  # multi-expression case
  expect_no_lint("lapply(x, function(xi) { print(xi); xi^2 })", linter)
  # multi-expression, multi-line cases
  expect_no_lint(
    trim_some("
      lapply(x, function(xi) {
        print(xi); xi^2
      })
    "),
    linter
  )
  expect_no_lint(
    trim_some("
      lapply(x, function(xi) {
        print(xi)
        xi^2
      })
    "),
    linter
  )

  # This _could_ be lapply(x, `%in%`, tbl), but don't force infix into lambda
  expect_no_lint("lapply(x, function(xi) xi %in% tbl)", linter)
  # This one could not
  expect_no_lint("lapply(x, function(xi) tbl %in% xi)", linter)
  # would require multiple lapply() loops
  expect_no_lint("lapply(x, function(xi) foo(bar(xi)))", linter)
  expect_no_lint("lapply(x, function(xi) return(foo(bar(xi))))", linter)

  # extractions, #2231
  expect_no_lint("lapply(l, function(x) rle(x)$values)", linter)
  expect_no_lint('lapply(l, function(x) rle(x)["values"])', linter)
  expect_no_lint('lapply(l, function(x) rle(x)[["values"]])', linter)
  expect_no_lint("lapply(l, function(x) rle(x)@values)", linter)

  # return() extractions, #2258
  expect_no_lint("lapply(l, function(x) return(foo(x)$bar))", linter)
  expect_no_lint('lapply(l, function(x) return(rle(x)["values"]))', linter)
  expect_no_lint('lapply(l, function(x) return(rle(x)[["values"]]))', linter)
  expect_no_lint("lapply(l, function(x) return(rle(x)@values))", linter)

  # Other operators, #2247
  expect_no_lint("lapply(l, function(x) foo(x) - 1)", linter)
  expect_no_lint("lapply(l, function(x) foo(x) * 2)", linter)
  expect_no_lint("lapply(l, function(x) foo(x) ^ 3)", linter)
  expect_no_lint("lapply(l, function(x) foo(x) %% 4)", linter)

  # Don't include other lambdas, #2249
  expect_no_lint(
    trim_some('{
      lapply(x, function(e) sprintf("%o", e))
      lapply(y, function(e) paste(strlpad(e, "0", width)))
    }'),
    linter
  )

  # only call is on RHS of operator, #2310
  expect_no_lint("lapply(l, function(x) 'a' %in% names(x))", linter)
  expect_no_lint("lapply(l, function(x = 1) 'a' %in% names(x))", linter)
})

test_that("unnecessary_lambda_linter skips allowed inner comparisons", {
  linter <- unnecessary_lambda_linter()

  # lapply returns a list, so not the same, though as.list is probably
  #   a better choice
  expect_no_lint("lapply(x, function(xi) foo(xi) == 2)", linter)

  # this _may_ return a matrix, though outer is probably a better choice if so
  expect_no_lint("sapply(x, function(xi) foo(xi) == y)", linter)

  # only lint "plain" calls that can be replaced by eliminating the lambda
  expect_no_lint("sapply(x, function(xi) sum(abs(xi)) == 0)", linter)

  # negated calls, #2742
  expect_no_lint("sapply(x, function(xi) !all.equal(xi, y))", linter)
  expect_no_lint("sapply(x, \\(xi) !all.equal(xi, y))", linter)
  expect_no_lint("purrr::map(x, ~!foo(.x))", linter)

  # unary operators, #2742
  expect_no_lint("sapply(x, function(xi) -foo(xi))", linter)
  expect_no_lint("sapply(x, function(xi) +foo(xi))", linter)
  expect_no_lint("sapply(x, function(xi) ~foo(xi))", linter)

  # parenthesized/braced unary operators
  expect_no_lint("sapply(x, function(xi) (!foo(xi)))", linter)
  expect_no_lint("sapply(x, function(xi) (-foo(xi)))", linter)
  expect_no_lint("sapply(x, function(xi) { -foo(xi) })", linter)
})

test_that("unnecessary_lambda_linter blocks simple disallowed usage", {
  linter <- unnecessary_lambda_linter()

  expect_lint(
    "lapply(DF, function(x) sum(x))",
    rex::rex("Pass sum directly as a symbol to lapply()"),
    linter
  )
  expect_lint(
    "lapply(DF, function(x) return(sum(x)))",
    rex::rex("Pass sum directly as a symbol to lapply()"),
    linter
  )

  expect_lint(
    "rapply(l, function(x) is.data.frame(x))",
    rex::rex("Pass is.data.frame directly as a symbol to rapply()"),
    linter
  )

  expect_lint(
    "eapply(env, function(x) sum(x, na.rm = TRUE))",
    rex::rex("Pass sum directly as a symbol to eapply()"),
    linter
  )
  expect_lint(
    "eapply(env, function(x) return(sum(x, na.rm = TRUE)))",
    rex::rex("Pass sum directly as a symbol to eapply()"),
    linter
  )

  # parenthesized/braced valid simplifications
  expect_lint(
    "sapply(x, function(xi) (foo(xi)))",
    rex::rex("Pass foo directly as a symbol to sapply()"),
    linter
  )
  expect_lint(
    "sapply(x, function(xi) { foo(xi) })",
    rex::rex("Pass foo directly as a symbol to sapply()"),
    linter
  )
})

test_that("unnecessary_lambda_linter blocks simple disallowed usages", {
  linter <- unnecessary_lambda_linter()
  linter_allow <- unnecessary_lambda_linter(allow_comparison = TRUE)
  lint_msg <- rex::rex("Compare to a constant after calling sapply() to get", anything, "sapply(x, foo)")

  expect_lint("sapply(x, function(xi) foo(xi) == 2)", lint_msg, linter)
  expect_lint("sapply(x, function(xi) foo(xi) == 'a')", lint_msg, linter)
  expect_lint("sapply(x, function(xi) foo(xi) == 1 + 2i)", lint_msg, linter)

  expect_no_lint("sapply(x, function(xi) foo(xi) == 2)", linter_allow)
  expect_no_lint("sapply(x, function(xi) foo(xi) == 'a')", linter_allow)
  expect_no_lint("sapply(x, function(xi) foo(xi) == 1 + 2i)", linter_allow)

  # vapply counts as well
  # NB: we ignore the FUN.VALUE argument, for now
  expect_lint(
    "vapply(x, function(xi) foo(xi) == 2, logical(1L))",
    rex::rex("Compare to a constant after calling vapply()", anything, "vapply(x, foo, FUN.VALUE = <intermediate>)"),
    linter
  )
})

test_that("unnecessary_lambda_linter blocks other comparators as well", {
  linter <- unnecessary_lambda_linter()
  linter_allow <- unnecessary_lambda_linter(allow_comparison = TRUE)
  lint_msg <- rex::rex("Compare to a constant after calling sapply() to get")

  expect_lint("sapply(x, function(xi) foo(xi) >= 2)", lint_msg, linter)
  expect_lint("sapply(x, function(xi) foo(xi) != 'a')", lint_msg, linter)
  expect_lint("sapply(x, function(xi) foo(xi) < 1 + 2i)", lint_msg, linter)

  expect_no_lint("sapply(x, function(xi) foo(xi) >= 2)", linter_allow)
  expect_no_lint("sapply(x, function(xi) foo(xi) != 'a')", linter_allow)
  expect_no_lint("sapply(x, function(xi) foo(xi) < 1 + 2i)", linter_allow)
})

test_that("unnecessary_lambda_linter doesn't apply to keyword args", {
  linter <- unnecessary_lambda_linter()

  expect_no_lint("lapply(x, function(xi) data.frame(nm = xi))", linter)
  expect_no_lint("lapply(x, function(xi) return(data.frame(nm = xi)))", linter)

  # adversarially commented
  expect_no_lint(
    trim_some("
      lapply(x, function(xi) data.frame(nm = # comment
      xi))
    "),
    linter
  )
})

test_that("purrr-style anonymous functions are also caught", {
  linter <- unnecessary_lambda_linter()

  expect_no_lint("purrr::map(x, ~.x)", linter)
  expect_no_lint("purrr::map_df(x, ~lm(y, .x))", linter)
  expect_no_lint("map_dbl(x, ~foo(bar = .x))", linter)

  expect_lint(
    "purrr::map(x, ~foo(.x))",
    rex::rex("Pass foo directly as a symbol to map()"),
    linter
  )
  expect_lint(
    "purrr::map_int(x, ~foo(.x, y))",
    rex::rex("Pass foo directly as a symbol to map_int()"),
    linter
  )
  expect_lint(
    "purrr::map_vec(x, ~foo(.x, y))",
    rex::rex("Pass foo directly as a symbol to map_vec()"),
    linter
  )

  # adversarial comment
  expect_no_lint(
    trim_some("
      map_dbl(x, ~foo(bar = # comment
      .x))
    "),
    linter
  )
})

test_that("cases with braces are caught", {
  linter <- unnecessary_lambda_linter()
  lint_msg <- rex::rex("Pass print directly as a symbol to lapply()")

  expect_lint(
    "lapply(x, function(xi) { print(xi) })",
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      lapply(x, function(xi) {
        print(xi)
      })
    "),
    lint_msg,
    linter
  )

  expect_lint(
    "lapply(x, function(xi) { return(print(xi)) })",
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      lapply(x, function(xi) {
        print(xi)
      })
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      lapply(x, function(xi) {
        return(print(xi))
      })
    "),
    lint_msg,
    linter
  )

  expect_no_lint(
    trim_some("
      lapply(x, function(xi) {
        print(xi)
        xi
      })
    "),
    linter
  )

  # false positives like #2231, #2247 are avoided with braces too
  expect_no_lint("lapply(x, function(xi) { foo(xi)$bar })", linter)
  expect_no_lint("lapply(x, function(xi) { foo(xi) - 1 })", linter)

  expect_lint(
    trim_some("
      lapply(y, function(yi) {
        print(yi) # comment
      })
    "),
    lint_msg,
    linter
  )
})

test_that("function shorthand is handled", {
  linter <- unnecessary_lambda_linter()
  linter_allow <- unnecessary_lambda_linter(allow_comparison = TRUE)

  expect_lint(
    "lapply(DF, \\(x) sum(x))",
    rex::rex("Pass sum directly as a symbol to lapply()"),
    linter
  )

  lint_msg <- rex::rex("Compare to a constant after calling sapply() to get", anything, "sapply(x, foo)")
  expect_lint("sapply(x, \\(xi) foo(xi) == 2)", lint_msg, linter)
  expect_lint("sapply(x, \\(xi) foo(xi) == 'a')", lint_msg, linter)
  expect_lint("sapply(x, \\(xi) foo(xi) == 1 + 2i)", lint_msg, linter)

  expect_no_lint("sapply(x, \\(xi) foo(xi) == 2)", linter_allow)
  expect_no_lint("sapply(x, \\(xi) foo(xi) == 'a')", linter_allow)
  expect_no_lint("sapply(x, \\(xi) foo(xi) == 1 + 2i)", linter_allow)
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      sapply(x, function(xi) sd(xi))
      lapply(y, function(yi) {
        sum(yi)
      })
    }"),
    list(
      list("sd", line_number = 2L),
      list("sum", line_number = 3L)
    ),
    unnecessary_lambda_linter()
  )
})

test_that("returns the correct linting", {
  linter <- object_usage_linter()
  local_var_msg <- rex::rex("local variable", anything, "assigned but may not be used")

  expect_no_lint("blah", linter)

  expect_no_lint(
    trim_some("
      function() {
        a <- 1
        a
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      fun <- function(x) {
        fun(1)
      }
      fun2 <- function(x) {
        fun2(2)
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      fun <- function() {
        a <- 1
      }
    "),
    local_var_msg,
    linter
  )

  expect_lint(
    trim_some("
      fun <- function() {
        a <- 1
        1
      }
    "),
    local_var_msg,
    linter
  )

  expect_lint(
    trim_some("
      fun <- function() {
        a <- 1
      }
    "),
    local_var_msg,
    linter
  )

  # same, using = for assignment
  expect_lint(
    trim_some("
      fun = function() {
        a = 1
      }
    "),
    local_var_msg,
    linter
  )

  expect_lint( # nofuzz: comment_injection
    trim_some("
      fun <- function() {
        a2 <- 1
        a3
      }
    "),
    list(
      local_var_msg,
      rex::rex("no visible binding for global variable ", anything)
    ),
    linter
  )

  expect_lint(
    trim_some("
      fun <- function() {
        fnu(1)
      }
    "),
    rex::rex("no visible global function definition for ", anything),
    linter
  )

  # earlier we used n(1) but this might conflict with dplyr::n(),
  #   so switch to use an obscure symbol
  expect_lint(
    trim_some("
      fun <- function(x) {
        `__lintr_obj`(1)
      }
    "),
    rex::rex("no visible global function definition for ", anything),
    linter
  )

  # setMethod and assign also checked

  expect_lint(
    trim_some("
      assign('fun', function() {
        a <- 1
        1
      })
    "),
    local_var_msg,
    linter
  )

  expect_lint(
    trim_some("
      setMethod('plot', 'numeric', function() {
        a <- 1
        1
      })
    "),
    local_var_msg,
    linter
  )
})

test_that("replace_functions_stripped", {
  expect_lint(
    trim_some("
      fun <- function(x) {
        `__lintr_obj`(x) = 1
      }
    "),
    rex::rex("no visible global function definition for ", anything),
    object_usage_linter()
  )

  expect_lint(
    trim_some("
      fun <- function(x) {
        `__lintr_obj`(x) <- 1
      }
    "),
    rex::rex("no visible global function definition for ", anything),
    object_usage_linter()
  )
})

test_that("eval errors are ignored", {
  expect_no_lint(
    trim_some('
    setMethod("[[<-", c("stampedEnv", "character", "missing"),
      function(x) {
        x
      })
    '),
    object_usage_linter()
  )
})

test_that("calls with top level function definitions are ignored", {
  expect_no_lint(
    'tryCatch("foo", error = function(e) e)',
    object_usage_linter()
  )
})

test_that("object-usage line-numbers are relative to start-of-file", {
  expect_lint(
    trim_some("
      a <- function(y) {
        y ** 2
      }
      b <- function() {
        x
      }
    "),
    list(line_number = 5L),
    object_usage_linter()
  )
})

test_that("used symbols are detected correctly", {
  # From #666
  expect_no_lint(
    trim_some('
      foo <- data.frame(0)
      foo$bar <- 1
      zero <- function() {
        file.info("/dev/null")$size
      }
      message(zero())
    '),
    object_usage_linter()
  )

  expect_lint(
    trim_some("
      foo$bar <- 1
      zero <- function() {
        foo
      }
      message(zero())
    "),
    list("foo"),
    object_usage_linter()
  )

  # Also test deeper nesting
  expect_no_lint(
    trim_some('
      foo <- list(0)
      foo$bar$baz$goo <- 1
      zero <- function() {
        file.info("/dev/null")$size
        foo$bar
        foo$bar$baz
        foo$bar$baz$goo
      }
      message(zero())
    '),
    object_usage_linter()
  )

  # Test alternative assignment and access methods
  expect_no_lint(
    trim_some('
      foo <- list(0)
      foo[["bar"]][["baz"]][["goo"]] <- 1
      zero <- function() {
        file.info("/dev/null")$size
        foo$bar
        foo$bar$baz
        foo$bar$baz$goo
        foo[["bar"]]
        foo[[c("bar", "baz")]]
        foo[["bar"]]$baz$goo
      }
      message(zero())
    '),
    object_usage_linter()
  )


  # regression #1322
  expect_silent(expect_no_lint("assign('x', 42)", object_usage_linter()))
})

test_that("object_usage_linter finds lints spanning multiple lines", {
  # Regression test for #507
  expect_lint(
    trim_some("
      foo <- function() {
        if (unknown_function()) NULL

        if (unknown_function()) {
          NULL
        }
      }
    "),
    list(
      list(message = "unknown_function", line_number = 2L),
      list(message = "unknown_function", line_number = 4L)
    ),
    object_usage_linter()
  )

  # Linted symbol is not on the first line of the usage warning
  expect_lint(
    trim_some("
      foo <- function(x) {
        with(
          x,
          unknown_symbol
        )
      }
    "),
    list(message = "unknown_symbol", line_number = 4L, column_number = 5L),
    object_usage_linter(skip_with = FALSE)
  )

  # Even ugly names are found
  expect_lint(
    trim_some("
      foo <- function(x) {
        with(
          x,
          `\u2019regex_kill`
        )
      }
    "),
    list(line_number = 4L, column_number = 5L),
    object_usage_linter(skip_with = FALSE)
  )
})

test_that("global variable detection works", {
  old_globals <- utils::globalVariables(package = globalenv())
  utils::globalVariables("global_function", package = globalenv())
  on.exit(utils::globalVariables(old_globals, package = globalenv(), add = FALSE))

  expect_no_lint(
    trim_some("
      foo <- function() {
        if (global_function()) NULL

        if (global_function()) {
          NULL
        }
      }
    "),
    object_usage_linter()
  )
})

test_that("package detection works", {
  expect_length(
    lint_package(test_path("dummy_packages", "package"), linters = object_usage_linter(), parse_settings = FALSE),
    10L
  )
})

test_that("robust against errors", {
  expect_no_lint(
    'assign("x", unknown_function)',
    object_usage_linter()
  )
})

test_that("interprets glue expressions", {
  linter <- object_usage_linter()

  expect_no_lint(
    trim_some("
      fun <- function() {
        local_var <- 42
        glue::glue('The answer is {local_var}.')
      }
    "),
    linter
  )

  # no need for namespace-qualification
  expect_no_lint(
    trim_some("
      glue <- glue::glue # imitate this being an @import
      fun <- function() {
        local_var <- 42
        glue('The answer is {local_var}.')
      }
    "),
    linter
  )

  # multiple variables in different interpolations
  expect_no_lint(
    trim_some("
      fun <- function() {
        local_key <- 'a'
        local_value <- 123
        glue::glue('Key-value pair: {local_key}={local_value}.')
      }
    "),
    linter
  )

  # multiple variables in single interpolation
  expect_no_lint(
    trim_some("
      fun <- function() {
        local_str1 <- 'a'
        local_str2 <- 'b'
        glue::glue('With our powers combined: {paste(local_str1, local_str2)}.')
      }
    "),
    linter
  )

  # Check non-standard .open and .close
  expect_no_lint(
    trim_some("
      fun <- function() {
        local_var <- 42
        glue::glue('The answer is $[local_var].', .open = '$[', .close = ']')
      }
    "),
    linter
  )

  # Steer clear of custom .transformer and .envir constructs
  expect_lint(trim_some("
    fun <- function() {
      local_var <- 42
      glue::glue('The answer is {local_var}.', .transformer = glue::identity_transformer)
    }
  "), "local_var", linter)

  expect_lint(trim_some("
    fun <- function() {
      local_var <- 42
      e <- new.env()
      glue::glue('The answer is {local_var}.', .envir = e)
    }
  "), "local_var", linter)

  # unused is caught, glue-used is not
  expect_lint(trim_some("
    fun <- function() {
      local_var <- 42
      unused_var <- 3
      glue::glue('The answer is {local_var}.')
    }
  "), "unused_var", linter)

  # glue-only is caught with option off
  expect_lint(trim_some("
    fun <- function() {
      local_var <- 42
      glue::glue('The answer is {local_var}.')
    }
  "), "local_var", object_usage_linter(interpret_extensions = NULL))

  # call in glue is caught
  expect_lint(
    trim_some("
      fun <- function() {
        local_call <- identity
        local_unused_call <- identity
        glue::glue('{local_call(1)}')
      }
    "),
    "local_unused_call",
    linter
  )

  # ditto infix operator
  expect_no_lint(
    trim_some("
      glue <- glue::glue # imitate this being an @import
      foo <- function() {
        `%++%` <- `+`
        glue('{x %++% y}')
      }
    "),
    linter
  )
})

test_that("errors/edge cases in glue syntax don't fail lint()", {
  linter <- object_usage_linter()

  # no lint & no error, despite glue error
  expect_warning(
    expect_no_lint(
      trim_some("
        fun <- function() {
          a <- 2
          a + 1
          glue::glue('The answer is {a')
        }
      "),
      linter
    ),
    "Evaluating glue expression.*failed: Expecting '\\}'.*Please ensure correct glue syntax"
  )

  # generates a lint because the "usage" inside glue() is not detected
  expect_warning(
    expect_lint(
      trim_some("
        fun <- function() {
          a <- 2
          glue::glue('The answer is {a')
        }
      "),
      "local variable 'a'",
      linter
    ),
    "Evaluating glue expression.*failed: Expecting '\\}'"
  )

  # complete {...}, but syntax error in code -> ignore
  expect_lint(
    trim_some("
      fun <- function() {
        a <- 2
        glue::glue('The answer is {a + }')
      }
    "),
    "local variable 'a'",
    linter
  )

  # empty glue expression {}
  expect_no_lint(
    trim_some("
      fun <- function() {
        a <- 2
        glue::glue('The answer is {}: {a}')
      }
    "),
    linter
  )

  # comment inside glue range (#1919)
  expect_no_lint(
    trim_some("
      fun <- function() {
        a <- 2
        glue::glue(
          'The answer is {}: {a}' # show the answer
        )
      }
    "),
    linter
  )
})

test_that("backtick'd names in glue are handled", {
  expect_lint(
    trim_some("
      fun <- function() {
        `w` <- 2
        x <- 3
        y <- -4
        `\\`y` <- 4
        z <- -5
        `z\\`` <- 5
        glue::glue('{w}{`x`}{y}{z}')
      }
    "),
    list(
      rex::rex("local variable '`y'"),
      rex::rex("local variable 'z`'")
    ),
    object_usage_linter()
  )
})

# reported as #1088
test_that("definitions below top level are ignored (for now)", {
  expect_no_lint(
    trim_some("
      local({
        x <- 1
        f <- function() {
          x
        }
      })
    "),
    object_usage_linter()
  )
})

# reported as #1127
test_that("package imports are detected if present in file", {
  skip_if("package:xml2" %in% search())
  expect_lint(
    trim_some("
      dog <- function(url) {
        a <- read_xml(url)
        a
      }
    "),
    rex::rex("no visible global function definition for ", anything, "read_xml"),
    object_usage_linter()
  )

  expect_no_lint(
    trim_some("
      library(xml2)

      dog <- function(url) {
        a <- read_xml(url)
        a
      }
    "),
    object_usage_linter()
  )
})

test_that("fallback works", {
  expect_lint(
    trim_some("
      f <- function() {
        `non_existing_assign<-`(1, 2)
      }
    "),
    list(
      message = rex::rex("no visible global function definition for ", anything, "non_existing_assign<-"),
      line_number = 2L,
      column_number = 3L
    ),
    object_usage_linter()
  )
})

test_that("unknown infix operators give good lint metadata", {
  expect_lint(
    trim_some("
      foo <- function(x) {
        x %unknown-operator% 1
      }
    "),
    list(
      message = rex::rex("no visible global function definition for '%unknown-operator%'"),
      line_number = 2L, column_number = 5L
    ),
    object_usage_linter()
  )

  skip_if(any(c("package:rlang", "package:data.table") %in% search()))
  expect_lint(
    trim_some('
      foo <- function(x) {
        x[, "new_col" := 2L]
      }
    '),
    list(
      message = rex::rex("no visible global function definition for ':='"),
      line_number = 2L, column_number = 17L
    ),
    object_usage_linter()
  )
})

test_that("respects `skip_with` argument for `with()` expressions", {
  f <- withr::local_tempfile(
    lines = c(
      "test_fun <- function(df) {",
      "  with(df, first_var + second_var)",
      "}"
    )
  )

  expect_length(lint(f, object_usage_linter(skip_with = TRUE)), 0L)
  expect_length(lint(f, object_usage_linter(skip_with = FALSE)), 2L)
})

test_that("missing libraries don't cause issue", {
  expect_lint(
    trim_some("
      library(a.a.a.z.z.z)
      foo <- function() {
        a <- 1
        a
      }
    "),
    list(
      "Could not find exported symbols for package \"a.a.a.z.z.z\"",
      line_number = 1L
    ),
    object_usage_linter()
  )
})

test_that("messages without a quoted name are caught", {
  # regression test for #1714
  expect_lint(
    trim_some("
      foo <- function() {
        a <- ...
        a
      }
    "),
    list(
      message = "... may be used in an incorrect context",
      line_number = 2L
    ),
    object_usage_linter()
  )
})

# See #1914
test_that("symbols in formulas aren't treated as 'undefined global'", {
  linter <- object_usage_linter()

  expect_lint(
    trim_some("
      foo <- function(x) {
        lm(
          y ~ z,
          data = x[!is.na(y)]
        )
      }
    "),
    list(
      message = "no visible binding for global variable 'y'",
      line_number = 4L,
      column_number = 21L
    ),
    linter
  )

  # neither on the RHS
  expect_lint(
    trim_some("
      foo <- function(x) {
        lm(
          z ~ y,
          data = x[!is.na(y)]
        )
      }
    "),
    list(
      message = "no visible binding for global variable 'y'",
      line_number = 4L,
      column_number = 21L
    ),
    linter
  )

  # nor in nested expressions
  expect_lint(
    trim_some("
      foo <- function(x) {
        lm(
          log(log(y)) ~ z,
          data = x[!is.na(y)]
        )
      }
    "),
    list(
      message = "no visible binding for global variable 'y'",
      line_number = 4L,
      column_number = 21L
    ),
    linter
  )

  # nor as a call
  # NB: I wanted this to be s(), as in mgcv::s(), but that
  #   doesn't work in this test suite because it resolves to
  #   rex::s() since we attach that in testthat.R
  expect_lint(
    trim_some("
      foo <- function(x) {
        lm(
          y(w) ~ z,
          data = x[!is.na(y)]
        )
      }
    "),
    list(
      message = "no visible binding for global variable 'y'",
      line_number = 4L,
      column_number = 21L
    ),
    linter
  )

  # native lambda requires being in an expression to support a comment immediately after
  expect_lint(
    trim_some("
      foo <- \\ # comment
      (x) {
        lm(
          y(w) ~ z,
          data = x[!is.na(y)]
        )
      }
    "),
    "no visible",
    linter
  )
})

test_that("NSE-ish symbols after $/@ are ignored as sources for lints", {
  linter <- object_usage_linter()
  lint_msg <- "no visible binding for global variable 'column'"

  expect_lint(
    trim_some("
      foo <- function(x) {
        ggplot2::ggplot(
          x[!is.na(x$column), ],
          ggplot2::aes(x = column, fill = factor(x$grp))
        )
      }
    "),
    list(lint_msg, line_number = 4L, column_number = 22L),
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(x) {
        ggplot2::ggplot(
          x[!is.na(x@column), ],
          ggplot2::aes(x = column, fill = factor(x$grp))
        )
      }
    "),
    list(lint_msg, line_number = 4L, column_number = 22L),
    linter
  )
})

test_that("functional lambda definitions are also caught", {
  expect_lint(
    trim_some("
      fun <- \\() {
        a <- 1
      }
    "),
    rex::rex("local variable", anything, "assigned but may not be used"),
    object_usage_linter()
  )
})

test_that("messages without location info are repaired", {
  linter <- object_usage_linter()
  global_function_msg <- rex::rex("no visible global function definition for", anything)
  global_variable_msg <- rex::rex("no visible binding for global variable", anything)
  local_variable_msg <- rex::rex("local variable", anything, "assigned but may not be used")

  # regression test for #1986
  expect_lint(
    "foo <- function() no_fun()",
    list(global_function_msg, line_number = 1L, column_number = 19L),
    linter
  )

  expect_lint(
    "foo <- function(a = no_fun()) a",
    list(global_function_msg, line_number = 1L, column_number = 21L),
    linter
  )

  expect_lint(
    "foo <- function() no_global",
    list(global_variable_msg, line_number = 1L, column_number = 19L),
    linter
  )

  expect_lint(
    "foo <- function() unused_local <- 42L",
    list(local_variable_msg, line_number = 1L, column_number = 19L),
    linter
  )

  # More complex case with two lints and missing location info
  expect_lint(
    trim_some("
      foo <- function() a <-
        bar()
    "),
    list(
      list(local_variable_msg, line_number = 1L, column_number = 19L),
      list(global_function_msg, line_number = 2L, column_number = 3L)
    ),
    linter
  )
})

test_that("globals in scripts are found regardless of assignment operator", {
  linter <- object_usage_linter()

  expect_no_lint(
    trim_some("
      library(dplyr)

      global_const_eq = 5
      global_const_la <- 6
      7 -> global_const_ra

      examplefunction <- function(df) {
        df %>%
          select(dist) %>%
          mutate(power = global_const_eq + global_const_ra + global_const_la)
      }
    "),
    linter
  )
})

test_that("dplyr's .env-specified objects are marked as 'used'", {
  skip_if_not_installed("dplyr")
  skip_if_not_installed("rlang")
  linter <- object_usage_linter()

  expect_lint( # nofuzz: dollar_at
    trim_some("
      foo <- function(df) {
        source <- 1
        target <- 2
        unused <- 3

        df %>%
          dplyr::mutate(
            from = rlang::.env$source,
            to = rlang::.env[['target']]
          )
      }
    "),
    rex::rex("local variable", anything, "unused"),
    linter
  )
})

test_that("interpret_glue is defunct", {
  expect_error(
    object_usage_linter(interpret_glue = FALSE),
    rex::rex("interpret_glue", anything, "deprecated")
  )
  expect_error(
    object_usage_linter(interpret_glue = TRUE),
    rex::rex("interpret_glue", anything, "deprecated")
  )
})

test_that("Lint return on end of function", {
  expect_lint(
    trim_some("
      function() {
        return(1)
        # Test
        3 + 4
      }
    "),
    list(
      line_number = 4L,
      message = rex::rex("All functions must have an explicit return()."),
      type = "warning"
    ),
    return_linter(return_style = "explicit")
  )

  expect_lint(
    trim_some("
      function() {
        return(1)
      }
    "),
    list(
      line_number = 2L,
      message = rex::rex("Use implicit return behavior; explicit return() is not needed."),
      type = "style"
    ),
    return_linter()
  )
})

test_that("Lint return on end of lambda function", {
  expect_lint(
    trim_some("
      \\(bar) {
        5L + 3L
      }
    "),
    list(
      line_number = 2L,
      message = rex::rex("All functions must have an explicit return().")
    ),
    return_linter(return_style = "explicit")
  )

  expect_lint(
    trim_some("
      \\(bar) {
        5L + 3L
        return(1)
      }
    "),
    list(
      line_number = 3L,
      message = rex::rex("Use implicit return behavior; explicit return() is not needed.")
    ),
    return_linter()
  )
})

test_that("Do not lint if/else statements (with return) on end of function", {
  linter <- return_linter(return_style = "explicit")

  expect_no_lint(
    trim_some("
      function() {
        if (x) {
          return(4)
        } else if (y) {
          return(5)
        } else {
          return(6)
        }
      }
    "),
    linter
  )
})

test_that("Lint control statements (without return) on end of function", {
  linter <- return_linter(return_style = "explicit")
  lint_msg <- rex::rex("All functions must have an explicit return().")

  expect_lint(
    trim_some("
      function() {
        while (x > 4) {
          cat(4)
          if (x < 4) {
            return(x)
          }
        }
      }
    "),
    list(lint_msg, line_number = 2L),
    linter
  )

  expect_lint(
    trim_some("
      function() {
        repeat {
          cat(4)
        }
      }
    "),
    list(lint_msg, line_number = 2L),
    linter
  )

  expect_lint(
    trim_some("
      function() {
        for (i in 1:10) {
          cat(4)
          if (i > 11) {
            return(x)
          }
        }
      }
    "),
    list(lint_msg, line_number = 2L),
    linter
  )

  expect_lint(
    trim_some("
      function() {
        if (x == 2L){
          return(e)
        } else if (x == 3L) {
          cat(f)
        }
      }
    "),
    list(lint_msg, line_number = 5L),
    linter
  )
})

test_that("Do not lint stop on end of function", {
  expect_no_lint(
    trim_some("
      function() {
        # Test
        3 + 4
        stop(1)
      }
    "),
    return_linter(return_style = "explicit")
  )

  expect_no_lint(
    trim_some("
      function() {
        stop(1)
      }
    "),
    return_linter()
  )
})

test_that("return_linter works in simple function", {
  expect_no_lint(
    trim_some("
      foo <- function(bar) {
        return(bar)
      }
    "),
    return_linter(return_style = "explicit")
  )
})

test_that("return_linter works for using stop() instead of returning", {
  linter <- return_linter(return_style = "explicit")

  expect_no_lint(
    trim_some("
      foo <- function(bar) {
        stop('bad')
      }
    "),
    linter
  )
})

test_that("return_linter ignores expressions that aren't functions", {
  expect_no_lint("x + 1", return_linter(return_style = "explicit"))
})

test_that("return_linter ignores anonymous/inline functions", {
  expect_no_lint("lapply(rnorm(10), function(x) x + 1)", return_linter(return_style = "explicit"))
})

test_that("return_linter ignores if statements outside of functions", {
  expect_no_lint(
    trim_some("
      if (TRUE) {
        TRUE
      } else {
        FALSE
      }
    "),
    return_linter(return_style = "explicit")
  )
})

test_that("return_linter passes on multi-line functions", {
  expect_no_lint(
    trim_some("
      foo <- function(x) {
        y <- x + 1
        return(y)
      }
    "),
    return_linter(return_style = "explicit")
  )
})


test_that("return_linter identifies a simple missing return", {
  expect_lint(
    trim_some("
      foo <- function(bar) {
        bar
      }
    "),
    rex::rex("All functions must have an explicit return()."),
    return_linter(return_style = "explicit")
  )
})


test_that("return_linter finds a missing return in a 2+ line function", {
  expect_lint(
    trim_some("
      foo <- function(x) {
        y <- x + 1
        y^2
      }
    "),
    rex::rex("All functions must have an explicit return()."),
    return_linter(return_style = "explicit")
  )
})

test_that("return_linter finds a missing return despite early returns", {
  expect_lint(
    trim_some("
      foo <- function(x) {
        if (TRUE) return(TRUE)
        x <- 1 + 1
        x
      }
    "),
    rex::rex("All functions must have an explicit return()."),
    return_linter(return_style = "explicit")
  )
})


test_that("return_linter finds multiple missing returns in branches", {
  lint_msg <- rex::rex("All functions must have an explicit return().")

  expect_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          TRUE
        } else {
          FALSE
        }
      }
    "),
    list(
      list(lint_msg, line_number = 3L),
      list(lint_msg, line_number = 5L)
    ),
    return_linter(return_style = "explicit")
  )
})

test_that("return_linter works regardless of braces in final if case", {
  linter <- return_linter(return_style = "explicit")

  expect_lint(
    trim_some("
      foo <- function() {
        if (TRUE) TRUE
      }
    "),
    rex::rex("All functions must have an explicit return()."),
    linter
  )
  expect_no_lint(
    trim_some("
      foo <- function() {
        if (TRUE) return(TRUE)
      }
    "),
    linter
  )
})

test_that("return_linter finds missing return in one branch of an if", {
  linter <- return_linter(return_style = "explicit")
  lint_msg <- rex::rex("All functions must have an explicit return().")

  expect_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          return(TRUE)
        } else {
          FALSE
        }
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          TRUE
        } else {
          return(FALSE)
        }
      }
    "),
    lint_msg,
    linter
  )
})

test_that("return_linter works in nested if statements", {
  linter <- return_linter(return_style = "explicit")

  expect_no_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          return(TRUE)
        } else if (nzchar('a')) {
          return(TRUE)
        } else {
          return(FALSE)
        }
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          if (nzchar('a')) {
            TRUE
          }
        } else {
          return(FALSE)
        }
      }
    "),
    rex::rex("All functions must have an explicit return()."),
    linter
  )
})

test_that("return_linter works in multi-line nested if statements", {
  linter <- return_linter(return_style = "explicit")

  expect_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          if (nzchar('a')) {
            y <- 1 + 1
            y
          }
        } else {
          return(FALSE)
        }
      }
    "),
    rex::rex("All functions must have an explicit return()."),
    linter
  )

  expect_no_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          if (nzchar('a')) {
            y <- 1 + 1
            return(y)
          }
        } else {
          return(FALSE)
        }
      }
    "),
    linter
  )
})

test_that("return_linter works for final for loops as well", {
  linter <- return_linter(return_style = "explicit")
  lint_msg <- rex::rex("All functions must have an explicit return().")

  expect_lint(
    trim_some("
      foo <- function() {
        for (i in seq_len(10)) {
          if (i %% 2 == 0) {
            y <- 1 + 1
            return(y)
          }
        }
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function() {
        for (i in seq_len(10)) {
          if (i %% 2 == 0) {
            y <- 1 + 1
          }
        }
      }
    "),
    lint_msg,
    linter
  )
})

test_that("return_linter works for function factories", {
  linter <- return_linter(return_style = "explicit")
  lint_msg <- rex::rex("All functions must have an explicit return().")

  expect_lint(
    trim_some("
      foo <- function(x) {
        function () {
          return(x + 1)
        }
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(x) {
        function () {
          x + 1
        }
      }
    "),
    list(lint_msg, lint_msg),
    linter
  )
})

test_that("return_linter allows return()-less Rcpp wrappers", {
  expect_no_lint(
    trim_some("
      ReadCapacitorAsList <- function(file) {
        .Call(R_ReadCapacitorAsList, file)
      }
    "),
    return_linter(return_style = "explicit")
  )
})

test_that("return_linter allows return()-less namespace hook calls", {
  expect_no_lint(
    trim_some("
      .onLoad <- function(libname, pkgname) {
        do_setup()
      }
    "),
    return_linter(return_style = "explicit")
  )
})

test_that("return_linter correctly handles pipes", {
  linter <- return_linter(return_style = "explicit")

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        x %>%
          return()
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        x %>%
          mean() %>%
          return()
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        y <- rnorm(length(x))

        x %>%
          cbind(y) %>%
          return()
      }
    "),
    linter
  )
})

test_that("return_linter handles pipes in control flow", {
  linter <- return_linter(return_style = "explicit")
  lint_msg <- rex::rex("All functions must have an explicit return().")

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        if (TRUE) {
          return(invisible())
        } else {
          x %>%
            return()
        }
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(x) {
        for (i in seq_len(10)) {
          x %>%
            mean()
        }
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(x) {
        if (TRUE) {
          x %>%
            mean()
        } else {
          return(TRUE)
        }
      }
    "),
    lint_msg,
    linter
  )
})

test_that("return_linter passes on q() or quit() calls", {
  expect_no_lint(
    trim_some("
      foo <- function(x) {
        if (TRUE) {
          q('n')
        } else {
          quit('n')
        }
      }
    "),
    return_linter(return_style = "explicit")
  )
})

test_that("return_functions= argument works", {
  linter <- return_linter(return_style = "explicit", return_functions = "LOG")

  expect_no_lint(
    trim_some("
      foo <- function(bar) {
        LOG('INFO', 'bad')
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      foo <- function(bar) {
        logging::LOG('INFO', 'bad')
      }
    "),
    linter
  )
})

test_that("except= argument works", {
  expect_no_lint(
    trim_some("
      foo <- function(bar) {
        5 + 3
      }
    "),
    return_linter(return_style = "explicit", except = "foo")
  )
})

test_that("except_regex= argument works", {
  linter <- return_linter(return_style = "explicit", except_regex = "^Test")

  expect_no_lint(
    trim_some("
      TestSummary <- function() {
        context <- foo(72643424)
        expected <- data.frame(a = 2)
        checkEquals(expected, bar(context))
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      TestMyPackage <- function() {
        checkMyCustomComparator(x, y)
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      TestOuter <- function() {
        actual <- lapply(
          input,
          function(x) {
            no_return()
          }
        )
        TestInner <- function() {
          no_return()
        }
        checkEquals(TestInner(), actual)
      }
    "),
    list(rex::rex("All functions must have an explicit return()."), line_number = 5L),
    linter
  )

  # capture group doesn't cause issues, #2678
  expect_no_lint(
    trim_some("
      TestFun <- function() {
        non_return()
      }
      AssertFun <- function() {
        non_return()
      }
    "),
    return_linter(return_style = "explicit", except_regex = "^(Test|Assert)")
  )
})

test_that("except= and except_regex= combination works", {
  expect_no_lint(
    trim_some("
      foo <- function() {
        no_return()
      }
      bar <- function() {
        no_return()
      }
      abaz <- function() {
        no_return()
      }
      bbaz <- function() {
        no_return()
      }
    "),
    return_linter(return_style = "explicit", except = c("foo", "bar"), except_regex = "baz$")
  )
})

test_that("return_linter skips brace-wrapped inline functions", { # nofuzz: comment_injection
  expect_no_lint("function(x) { sum(x) }", return_linter(return_style = "explicit"))
})

test_that("return_linter skips common S4 method functions", {
  linter <- return_linter(return_style = "explicit")

  expect_no_lint(
    trim_some("
      setGeneric(
        'ReadCircuitsPBAsDataTable',
        function(pbMessageList) {
          standardGeneric('ReadCircuitsPBAsDataTable')
        }
      )
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      setMethod('initialize', 'CircuitsTopology', function(.Object, ...) {
        callNextMethod(.Object, ...)
      })
    "),
    linter
  )
})

test_that("return_functions= is not affected by namespace qualification", {
  linter <- return_linter(return_style = "explicit", return_functions = "abort")

  expect_no_lint(
    trim_some("
      foo <- function(bar) {
        abort('bad')
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      foo <- function(bar) {
        rlang::abort('bad')
      }
    "),
    linter
  )
})

test_that("return_linter skips invokeRestart(), tryInvokeRestart()", {
  linter <- return_linter(return_style = "explicit")

  expect_no_lint(
    trim_some("
      warning = function(w) {
        warn <<- append(warn, conditionMessage(w))
        invokeRestart('muffleWarning')
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      custom_warning = function(w) {
        warn <<- append(warn, conditionMessage(w))
        tryInvokeRestart('muffleCustom_warning')
      }
    "),
    linter
  )
})

# NB: x |> return() is blocked by the parser, so no need to test that.
test_that("Native pipes are handled correctly", {
  linter <- return_linter(return_style = "explicit")
  lint_msg <- rex::rex("All functions must have an explicit return().")

  expect_lint(
    trim_some("
      foo <- function(x) {
        for (i in seq_len(10)) {
          x |>
            mean()
        }
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(x) {
        if (TRUE) {
          x |>
            mean()
        } else {
          return(TRUE)
        }
      }
    "),
    lint_msg,
    linter
  )
})

test_that("return_linter works for final while/repeat loops as well", {
  linter <- return_linter(return_style = "explicit")
  lint_msg <- rex::rex("All functions must have an explicit return().")

  expect_lint(
    trim_some("
      foo <- function(x) {
        while (x > 0) {
          if (x %% 2 == 0) {
            return(x)
          }
          x <- x + sample(10, 1)
        }
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(x) {
        repeat {
          if (x == 0) {
            return(x)
          }
          x <- x - sign(x)
        }
      }
    "),
    lint_msg,
    linter
  )
})

test_that("return_linter lints `message`, `warning` and `stopifnot`", {
  linter <- return_linter(return_style = "explicit")
  lint_msg <- rex::rex("All functions must have an explicit return().")

  expect_lint(
    trim_some("
      foo <- function(bar) {
        stopifnot(bar == 'd')
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(bar) {
        message('test')
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(bar) {
        warning(test)
      }
    "),
    lint_msg,
    linter
  )
})

test_that("return_linter handles arbitrarily nested terminal statements", {
  implicit_linter <- return_linter()
  implicit_msg <- rex::rex("Use implicit return behavior; explicit return() is not needed.")
  explicit_linter <- return_linter(return_style = "explicit")
  explicit_msg <- rex::rex("All functions must have an explicit return().")

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        if (x < 0) {
          if (x == -1) {
            return(TRUE)
          }
          if (x > -10) {
            NA
          } else {
            FALSE
          }
        } else if (x == 0) {
          TRUE
        } else {
          y <- x**2
          if (y > 10) {
            z <- sin(y)
            if (z > 0) {
              FALSE
            } else {
              NA
            }
          } else {
            TRUE
          }
        }
      }
    "),
    implicit_linter
  )

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        if (x < 0) {
          if (x == -1) {
            return(TRUE)
          }
          if (x > -10) {
            return(NA)
          } else {
            return(FALSE)
          }
        } else if (x == 0) {
          return(TRUE)
        } else {
          y <- x**2
          if (y > 10) {
            z <- sin(y)
            if (z > 0) {
              return(FALSE)
            } else {
              return(NA)
            }
          } else {
            return(TRUE)
          }
        }
      }
    "),
    explicit_linter
  )

  mixed_lines <- trim_some("
    foo <- function(x) {
      if (x < 0) {
        if (x == -1) {
          return(TRUE)
        }
        if (x > -10) {
          return(NA)
        } else {
          FALSE
        }
      } else if (x == 0) {
        return(TRUE)
      } else {
        y <- x**2
        if (y > 10) {
          z <- sin(y)
          if (z > 0) {
            FALSE
          } else {
            return(NA)
          }
        } else {
          TRUE
        }
      }
    }
  ")

  expect_lint(
    mixed_lines,
    list(
      list(implicit_msg, line_number = 7L),
      list(implicit_msg, line_number = 12L),
      list(implicit_msg, line_number = 20L)
    ),
    implicit_linter
  )

  expect_lint(
    mixed_lines,
    list(
      list(explicit_msg, line_number = 9L),
      list(explicit_msg, line_number = 18L),
      list(explicit_msg, line_number = 23L)
    ),
    explicit_linter
  )
})

test_that("explicit returns in control flow are linted correctly", {
  linter <- return_linter()
  lint_msg <- rex::rex("Use implicit return behavior")

  expect_lint(
    trim_some("
      foo <- function(bar) {
        if (TRUE) {
          return(bar)
        } else {
          return(NULL)
        }
      }
    "),
    list(lint_msg, lint_msg),
    linter
  )

  expect_no_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          if (TRUE) {
            return(1)
          }
          2
        } else {
          3
        }
      }
    "),
    linter
  )
})

# inspired by grid:::draw.all
#   https://github.com/r-devel/r-svn/blob/eeff859e427b2399f1474427a531365d2672f52f/src/library/grid/R/grob.R#L1940
test_that("logic is robust to absence of '{'", {
  linter <- return_linter()

  expect_no_lint(
    trim_some("
      foo <- function() {
        if (TRUE) # comment is a neighbor of 'if'
          FALSE
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      foo <- function() {
        if (TRUE)
          FALSE
        else # cannot rely on 'else' expr being e.g. 7th
          NA
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          FALSE
        } else # cannot rely on 'else' expr being e.g. 7th
          NA
      }
    "),
    linter
  )
})

test_that("logic is robust to terminal comments under '{'", {
  implicit_linter <- return_linter()
  implicit_msg <- rex::rex("Use implicit return behavior; explicit return() is not needed.")
  explicit_linter <- return_linter(return_style = "explicit")
  explicit_msg <- rex::rex("All functions must have an explicit return().")

  expect_lint(
    trim_some("
      foo <- function() {
        return(TRUE)
        # comment
      }
    "),
    implicit_msg,
    implicit_linter
  )

  expect_no_lint(
    trim_some("
      foo <- function() {
        return(TRUE)
        # comment
      }
    "),
    explicit_linter
  )

  expect_lint(
    trim_some("
      foo <- function() {
        TRUE
        # comment
      }
    "),
    explicit_msg,
    explicit_linter
  )
})

test_that("terminal = assignment is not an error", {
  # key is this is not an <expr> node
  expect_no_lint(
    trim_some("
      foo <- function() {
        a = 1
      }
    "),
    return_linter()
  )
})

test_that("empty terminal '{' expression is handled correctly", {
  implicit_linter <- return_linter()
  implicit_msg <- rex::rex("Use implicit return behavior; explicit return() is not needed.")
  explicit_linter <- return_linter(return_style = "explicit")
  explicit_msg <- rex::rex("All functions must have an explicit return().")

  empty_inline <- "foo <- function() { }"
  expect_no_lint(empty_inline, implicit_linter)
  expect_no_lint(empty_inline, explicit_linter)

  empty_multiline <- trim_some("
    foo <- function() {
    }
  ")
  expect_no_lint(empty_multiline, implicit_linter)
  expect_no_lint(empty_multiline, explicit_linter)

  empty_comment <- trim_some("
    foo <- function() {
      # this line intentionally left blank
    }
  ")
  expect_no_lint(empty_comment, implicit_linter)
  expect_no_lint(empty_comment, explicit_linter)

  empty_if_implicit <- trim_some("
    foo <- function() {
      if (TRUE) {
      } else {
        FALSE
      }
    }
  ")
  expect_no_lint(empty_if_implicit, implicit_linter)
  expect_lint(
    empty_if_implicit,
    list(
      list(explicit_msg, line_number = 2L),
      list(explicit_msg, line_number = 4L)
    ),
    explicit_linter
  )

  empty_else_implicit <- trim_some("
    foo <- function() {
      if (TRUE) {
        FALSE
      } else {
      }
    }
  ")
  expect_no_lint(empty_else_implicit, implicit_linter)
  expect_lint(
    empty_else_implicit,
    list(
      list(explicit_msg, line_number = 3L),
      list(explicit_msg, line_number = 4L)
    ),
    explicit_linter
  )

  empty_if_explicit <- trim_some("
    foo <- function() {
      if (TRUE) {
      } else {
        return(FALSE)
      }
    }
  ")
  expect_lint(empty_if_explicit, list(implicit_msg, line_number = 4L), implicit_linter)
  expect_lint(empty_if_explicit, list(explicit_msg, line_number = 2L), explicit_linter)

  empty_else_explicit <- trim_some("
    foo <- function() {
      if (TRUE) {
        return(FALSE)
      } else {
      }
    }
  ")
  expect_lint(empty_else_explicit, list(implicit_msg, line_number = 3L), implicit_linter)
  expect_lint(empty_else_explicit, list(explicit_msg, line_number = 4L), explicit_linter)

  empty_if_else <- trim_some("
    foo <- function() {
      if (TRUE) {
      } else {
      }
    }
  ")
  expect_no_lint(empty_if_else, implicit_linter)
  expect_lint(
    empty_if_else,
    list(
      list(explicit_msg, line_number = 2L),
      list(explicit_msg, line_number = 3L)
    ),
    explicit_linter
  )

  weird <- trim_some("
    foo <- function() {
      if (TRUE) {{{{
        0
      }}}} else {
        { return(1) }
      }
    }
  ")
  expect_lint(weird, list(implicit_msg, line_number = 5L), implicit_linter)
  expect_lint(weird, list(explicit_msg, line_number = 3L), explicit_linter)
})

test_that("non-if returns are skipped under allow_implicit_else = FALSE", {
  expect_no_lint(
    trim_some("
      foo <- function(bar) {
        bar
      }
    "),
    return_linter(allow_implicit_else = FALSE)
  )
})

test_that("if/else don't throw a lint under allow_implicit_else = FALSE", {
  expect_no_lint(
    trim_some("
      foo <- function(bar) {
        if (TRUE) {
          bar
        } else {
          NULL
        }
      }
    "),
    return_linter(allow_implicit_else = FALSE)
  )
})

test_that("implicit else outside a function doesn't lint under allow_implicit_else = FALSE", {
  expect_no_lint("if(TRUE) TRUE", return_linter(allow_implicit_else = FALSE))
})

test_that("allow_implicit_else = FALSE identifies a simple implicit else", {
  expect_lint(
    trim_some("
      foo <- function(bar) {
        if (TRUE) {
          bar
        }
      }
    "),
    rex::rex("All functions with terminal if statements must have a corresponding terminal else clause"),
    return_linter(allow_implicit_else = FALSE)
  )
})

test_that("allow_implicit_else = FALSE finds implicit else with nested if+else", {
  lint_msg <- rex::rex("All functions with terminal if statements must have a corresponding terminal else clause")

  expect_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          if (TRUE) {
            FALSE
          } else {
            TRUE
          }
        }
      }
    "),
    lint_msg,
    return_linter(allow_implicit_else = FALSE)
  )

  expect_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          if (TRUE) {
            return(FALSE)
          } else {
            return(TRUE)
          }
        }
      }
    "),
    lint_msg,
    return_linter(return_style = "explicit", allow_implicit_else = FALSE)
  )
})

test_that("allow_implicit_else = FALSE works on anonymous/inline functions", {
  expect_lint(
    "lapply(rnorm(10), function(x) if (TRUE) x + 1)",
    rex::rex("All functions with terminal if statements must"),
    return_linter(allow_implicit_else = FALSE)
  )
})

test_that("side-effect functions like .onLoad ignore the lack of explicit else under allow_implicit_else = FALSE", {
  expect_no_lint(
    trim_some("
      .onAttach <- function(libname, pkgname) {
        if (TRUE) foo()
      }
    "),
    return_linter(allow_implicit_else = FALSE)
  )

  expect_no_lint(
    trim_some("
      .onAttach <- function(libname, pkgname) {
        if (TRUE) return(foo())
      }
    "),
    return_linter(return_style = "explicit", allow_implicit_else = FALSE)
  )
})

test_that("implicit else lint has the correct metadata", {
  linter <- return_linter(return_style = "explicit", allow_implicit_else = FALSE)
  lint_msg <- "All functions with terminal if statements"

  expect_lint(
    trim_some("
      foo <- function(x, y = 3) {
        if (x) {
          return(x)
        }
      }
    "),
    list(lint_msg, line_number = 2L),
    linter
  )

  expect_lint(
    trim_some("{
      foo <- function(x, y = 3) {
        if (x) {
          return(x)
        }
      }

      bar <- function(x, y = 3) {
        if (x) {
          return(x)
        }
      }

      baz <- function(x, y = 3) {
        if (x) return(x)
      }
    }"),
    list(
      list(lint_msg, line_number = 3L),
      list(lint_msg, line_number = 9L),
      list(lint_msg, line_number = 15L)
    ),
    linter
  )
})

test_that("Correct lints thrown when lacking explicit return and explicit else", {
  linter <- return_linter(return_style = "explicit", allow_implicit_else = FALSE)
  explicit_return_msg <- rex::rex("All functions must have an explicit return().")
  implicit_else_msg <- rex::rex("All functions with terminal if statements")

  expect_lint(
    trim_some("
      foo <- function(x, y = 3) {
        if (x) {
          x
        }
      }
    "),
    list(
      list(implicit_else_msg, line_number = 2L),
      list(explicit_return_msg, line_number = 3L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      function(x, y) {
        if (x) {
          1
        } else if (y) {
          2
        }
      }
    "),
    list(
      list(explicit_return_msg, line_number = 3L),
      list(implicit_else_msg, line_number = 4L),
      list(explicit_return_msg, line_number = 5L)
    ),
    linter
  )
})

test_that("Mixing exempted functions doesn't miss lints", {
  # in the current implementation, a local copy of 'params' is
  #   edited in a loop; this test ensures that behavior continues to be correct
  expect_lint(
    trim_some("{
      foo <- function() {
        1
      }

      bar <- function() {
        if (TRUE) {
          return(2)
        }
      }

      baz <- function() {
        if (TRUE) {
          3
        }
      }
    }"),
    list(
      list("Use implicit return behavior", line_number = 8L),
      list("All functions with terminal if statements", line_number = 13L)
    ),
    return_linter(allow_implicit_else = FALSE, except = "bar")
  )
})

test_that("= assignments are handled correctly", {
  implicit_linter <- return_linter(allow_implicit_else = FALSE)
  implicit_msg <- rex::rex("All functions with terminal if statements")
  explicit_linter <- return_linter(return_style = "explicit")
  explicit_msg <- rex::rex("All functions must have an explicit return().")

  expect_no_lint(
    trim_some("
      .onLoad = function() {
        1
      }
    "),
    explicit_linter
  )

  expect_no_lint(
    trim_some("
      .onLoad = function() {
        if (TRUE) 1
      }
    "),
    implicit_linter
  )

  expect_lint(
    trim_some("
      foo = function() {
        1
      }
    "),
    explicit_msg,
    explicit_linter
  )

  expect_lint(
    trim_some("
      foo = function() {
        if (TRUE) 1
      }
    "),
    implicit_msg,
    implicit_linter
  )
})

test_that("terminal switch() is handled correctly", {
  implicit_linter <- return_linter()
  implicit_msg <- rex::rex("Use implicit return behavior; explicit return() is not needed.")
  explicit_linter <- return_linter(return_style = "explicit")
  explicit_msg <- rex::rex("All functions must have an explicit return().")

  no_return_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = 1,
        b = 2
      )
    }
  ")
  expect_no_lint(no_return_lines, implicit_linter)
  expect_lint(no_return_lines, list(explicit_msg, explicit_msg), explicit_linter)

  outer_return_lines <- trim_some("
    foo <- function(x) {
      return(switch(x,
        a = 1,
        b = 2
      ))
    }
  ")
  expect_lint(outer_return_lines, implicit_msg, implicit_linter)
  expect_no_lint(outer_return_lines, explicit_linter)

  partial_return_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = return(1),
        b = 2
      )
    }
  ")
  expect_lint(partial_return_lines, implicit_msg, implicit_linter)
  expect_lint(partial_return_lines, explicit_msg, explicit_linter)

  all_return_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = return(1),
        b = return(2)
      )
    }
  ")
  expect_lint(all_return_lines, list(implicit_msg, implicit_msg), implicit_linter)
  expect_no_lint(all_return_lines, explicit_linter)

  default_all_return_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = return(1),
        return(2)
      )
    }
  ")
  expect_lint(default_all_return_lines, list(implicit_msg, implicit_msg), implicit_linter)
  expect_no_lint(default_all_return_lines, explicit_linter)

  default_no_return_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = 1,
        2
      )
    }
  ")
  expect_no_lint(default_no_return_lines, implicit_linter)
  expect_lint(default_no_return_lines, list(explicit_msg, explicit_msg), explicit_linter)

  no_return_braced_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = {
          1
          2
          3
          4
        },
        b = {
          5
          6
          7
        }
      )
    }
  ")
  expect_no_lint(no_return_braced_lines, implicit_linter)
  expect_lint(
    no_return_braced_lines,
    list(
      list(explicit_msg, line_number = 7L),
      list(explicit_msg, line_number = 12L)
    ),
    explicit_linter
  )

  all_return_braced_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = {
          1
          2
          3
          return(4)
        },
        b = {
          5
          6
          return(7)
        }
      )
    }
  ")
  expect_lint(
    all_return_braced_lines,
    list(
      list(implicit_msg, line_number = 7L),
      list(implicit_msg, line_number = 12L)
    ),
    implicit_linter
  )
  expect_no_lint(all_return_braced_lines, explicit_linter)

  early_return_braced_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = {
          1
          if (TRUE) {
            return(2)
          }
          3
          4
        },
        b = {
          5
          6
          7
        }
      )
    }
  ")
  expect_no_lint(early_return_braced_lines, implicit_linter)
  expect_lint(
    early_return_braced_lines,
    list(
      list(explicit_msg, line_number = 9L),
      list(explicit_msg, line_number = 14L)
    ),
    explicit_linter
  )

  if_no_return_braced_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = {
          1
          if (TRUE) {
            2
          } else {
            3
          }
        },
        b = {
          5
          6
          7
        }
      )
    }
  ")
  expect_no_lint(if_no_return_braced_lines, implicit_linter)
  expect_lint(
    if_no_return_braced_lines,
    list(
      list(explicit_msg, line_number = 6L),
      list(explicit_msg, line_number = 8L),
      list(explicit_msg, line_number = 14L)
    ),
    explicit_linter
  )

  if_return_braced_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = {
          1
          if (TRUE) {
            return(2)
          } else {
            return(3)
          }
        },
        b = {
          5
          6
          return(7)
        }
      )
    }
  ")
  expect_lint(
    if_return_braced_lines,
    list(
      list(implicit_msg, line_number = 6L),
      list(implicit_msg, line_number = 8L),
      list(implicit_msg, line_number = 14L)
    ),
    implicit_linter
  )
  expect_no_lint(if_return_braced_lines, explicit_linter)

  ok_exit_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = .Call(a_routine, x),
        b = .Call(b_routine, x),
        stop('invalid')
      )
    }
  ")
  expect_no_lint(ok_exit_lines, implicit_linter)
  expect_no_lint(ok_exit_lines, explicit_linter)
})

test_that("switch() default statements interact with allow_implicit_else", {
  implicit_linter <- return_linter(allow_implicit_else = FALSE)
  explicit_linter <- return_linter(allow_implicit_else = FALSE, return_style = "explicit")
  implicit_msg <- rex::rex("Use implicit return behavior; explicit return() is not needed.")
  explicit_msg <- rex::rex("All functions must have an explicit return().")
  implicit_switch_msg <- rex::rex("All functions with terminal switch statements")
  implicit_else_msg <- rex::rex("All functions with terminal if statements")

  no_default_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = 1,
        b = 2
      )
    }
  ")
  expect_lint(no_default_lines, list(implicit_switch_msg, line_number = 2L), implicit_linter)
  expect_lint(no_default_lines, list(implicit_switch_msg, explicit_msg, explicit_msg), explicit_linter)

  ifelse_default_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = 1,
        b = 2,
        if (x != 'c') {
          3
        } else {
          4
        }
      )
    }
  ")
  expect_no_lint(ifelse_default_lines, implicit_linter)
  expect_lint(ifelse_default_lines, list(explicit_msg, explicit_msg, explicit_msg, explicit_msg), explicit_linter)

  if_no_else_default_lines <- trim_some("
    foo <- function(x) {
      switch(x,
        a = 1,
        b = 2,
        if (x != 'c') {
          3
        }
      )
    }
  ")
  expect_lint(if_no_else_default_lines, list(implicit_else_msg, line_number = 5L), implicit_linter)
  expect_lint(
    if_no_else_default_lines,
    list(
      list(explicit_msg, line_number = 3L),
      list(explicit_msg, line_number = 4L),
      list(implicit_else_msg, line_number = 5L),
      list(explicit_msg, line_number = 6L)
    ),
    explicit_linter
  )
})

test_that("functions with braced expressions in formals lint correctly", {
  implicit_linter <- return_linter()
  implicit_msg <- rex::rex("Use implicit return behavior; explicit return() is not needed.")
  explicit_linter <- return_linter(return_style = "explicit")
  explicit_msg <- rex::rex("All functions must have an explicit return().")

  brace_lines_expl <- trim_some("
    foo <- function(y,
                    x = {
                      y <- sqrt(y)
                      y + 1
                    }) {
      return(x + y)
    }
  ")
  expect_lint(brace_lines_expl, implicit_msg, implicit_linter)
  expect_no_lint(brace_lines_expl, explicit_linter)

  brace_lines_impl <- trim_some("
    foo <- function(y,
                    x = {
                      y <- sqrt(y)
                      y + 1
                    }) {
      x + y
    }
  ")
  expect_no_lint(brace_lines_impl, implicit_linter)
  expect_lint(brace_lines_impl, explicit_msg, explicit_linter)

  lambda_lines_expl_expl <- trim_some("
    foo <- function(y,
                    F = function(z) {
                      return(z + 1)
                    }) {
      return(F(y) + y)
    }
  ")
  expect_lint(
    lambda_lines_expl_expl,
    list(
      list(implicit_msg, line_number = 3L),
      list(implicit_msg, line_number = 5L)
    ),
    implicit_linter
  )
  expect_no_lint(lambda_lines_expl_expl, explicit_linter)

  lambda_lines_expl_impl <- trim_some("
    foo <- function(y,
                    F = function(z) {
                      return(z + 1)
                    }) {
      F(y) + y
    }
  ")
  expect_lint(
    lambda_lines_expl_impl,
    list(implicit_msg, line_number = 3L),
    implicit_linter
  )
  expect_lint(
    lambda_lines_expl_impl,
    list(explicit_msg, line_number = 5L),
    explicit_linter
  )

  lambda_lines_impl_expl <- trim_some("
    foo <- function(y,
                    F = function(z) {
                      z + 1
                    }) {
      return(F(y) + y)
    }
  ")
  expect_lint(
    lambda_lines_impl_expl,
    list(implicit_msg, line_number = 5L),
    implicit_linter
  )
  expect_lint(
    lambda_lines_impl_expl,
    list(explicit_msg, line_number = 3L),
    explicit_linter
  )

  lambda_lines_impl_impl <- trim_some("
    foo <- function(y,
                    F = function(z) {
                      z + 1
                    }) {
      F(y) + y
    }
  ")
  expect_no_lint(lambda_lines_impl_impl, implicit_linter)
  expect_lint(
    lambda_lines_impl_impl,
    list(
      list(explicit_msg, line_number = 3L),
      list(explicit_msg, line_number = 5L)
    ),
    explicit_linter
  )
})

test_that("boolean_arithmetic_linter doesn't block allowed usages", {
  linter <- boolean_arithmetic_linter()

  expect_no_lint("!any(x == y)", linter)
  expect_no_lint("!any(grepl(pattern, x))", linter)
})

test_that("boolean_arithmetic_linter requires use of any() or !any() over length(.(<logical>))", {
  linter <- boolean_arithmetic_linter()
  lint_msg <- rex::rex("Use any() to express logical aggregations.")

  expect_lint("length(which(x == y)) == 0", lint_msg, linter)
  # anything passed to which() can be assumed to be logical
  expect_lint("length(which(is_treatment)) == 0L", lint_msg, linter)
  # regex version
  expect_lint("length(grep(pattern, x)) == 0", lint_msg, linter)

  # non-== comparisons
  expect_lint("length(which(x == y)) > 0L", lint_msg, linter)
  expect_lint("length(which(is_treatment)) < 1", lint_msg, linter)
  expect_lint("length(grep(pattern, x)) >= 1L", lint_msg, linter)
})

local({
  linter <- boolean_arithmetic_linter()
  lint_msg <- rex::rex("Use any() to express logical aggregations.")

  outer_comparisons <- c("== 0", "== 0L", "> 0L", "> 0L", ">= 1", ">= 1L")

  patrick::with_parameters_test_that(
    "sum(x {op} y) {outer} lints",
    expect_lint(sprintf("sum(x %s y) %s", op, outer), lint_msg, linter),
    .cases = expand.grid(
      op = c("==", "!=", ">", "<", ">=", "<=", "&", "|", "%in%", "%chin%"),
      outer = outer_comparisons
    )
  )

  patrick::with_parameters_test_that(
    "sum({op}(x)) == 0 lints",
    expect_lint(sprintf("sum(%s(x)) == 0", op), lint_msg, linter),
    .cases = expand.grid(
      op = c(
        "!", "xor", "grepl", "str_detect", "is.element",
        "is.na", "is.finite", "is.infinite", "is.nan",
        "duplicated", "nzchar", "startsWith", "endsWith"
      ),
      outer = outer_comparisons
    )
  )
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Use any() to express logical aggregations.")

  expect_lint(
    trim_some("{
      length(which(x == y)) > 0L
      sum(x == y) != 0
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    boolean_arithmetic_linter()
  )
})

test_that("fixed_regex_linter skips allowed usages", {
  linter <- fixed_regex_linter()

  expect_no_lint("gsub('^x', '', y)", linter)
  expect_no_lint("grep('x$', y)", linter)
  expect_no_lint("grepv('x$', y)", linter)
  expect_no_lint("sub('[a-zA-Z]', '', y)", linter)
  expect_no_lint("grepl(fmt, y)", linter)
  expect_no_lint(R"{regexec('\\s', '', y)}", linter)
  expect_no_lint("grep('a(?=b)', x, perl = TRUE)", linter)
  expect_no_lint("grep('0+1', x, perl = TRUE)", linter)
  expect_no_lint("grep('1*2', x)", linter)
  expect_no_lint("grep('a|b', x)", linter)
  expect_no_lint(R"{grep('\\[|\\]', x)}", linter)

  # if fixed=TRUE is already set, regex patterns don't matter
  expect_no_lint(R"{gsub('\\.', '', y, fixed = TRUE)}", linter)

  # ignore.case=TRUE implies regex interpretation
  expect_no_lint("gsub('abcdefg', '', y, ignore.case = TRUE)", linter)

  # char classes starting with [] might contain other characters -> not fixed
  expect_no_lint("sub('[][]', '', y)", linter)
  expect_no_lint("sub('[][ ]', '', y)", linter)
  expect_no_lint("sub('[],[]', '', y)", linter)

  # wrapper functions don't throw
  expect_no_lint("gregexpr(pattern = pattern, data, perl = TRUE, ...)", linter)
})

test_that("fixed_regex_linter blocks simple disallowed usages", {
  linter <- fixed_regex_linter()
  lint_msg <- rex::rex("This regular expression is static")

  expect_lint(R"{gsub('\\.', '', x)}", lint_msg, linter)
  expect_lint("grepl('abcdefg', x)", lint_msg, linter)
  expect_lint("gregexpr('a-z', y)", lint_msg, linter)
  expect_lint(R"{regexec('\\$', x)}", lint_msg, linter)
  expect_lint("grep('\n', x)", lint_msg, linter)
  expect_lint("grepv('\n', x)", lint_msg, linter)

  # naming the argument doesn't matter (if it's still used positionally)
  expect_lint("gregexpr(pattern = 'a-z', y)", lint_msg, linter)
})

patrick::with_parameters_test_that(
  "fixed_regex_linter is robust to unrecognized escapes error for {char}",
  {
    expect_lint(
      sprintf(R"{grep('\\%s', x)}", char),
      rex::rex("This regular expression is static"),
      fixed_regex_linter()
    )

    expect_lint(
      sprintf(R"{strsplit('a%sb', '\\%s')}", char, char),
      rex::rex("This regular expression is static"),
      fixed_regex_linter()
    )
  },
  char = c(
    "^", "$", "{", "}", "(", ")", ".", "*", "+", "?",
    "|", "[", "]", R"(\\)", "<", ">", "=", ":", ";", "/",
    "_", "-", "!", "@", "#", "%", "&", "~"
  )
)

test_that("fixed_regex_linter catches regex like [.] or [$]", {
  linter <- fixed_regex_linter()
  lint_msg <- rex::rex("This regular expression is static")

  expect_lint("grep('[.]', x)", lint_msg, linter)
  expect_lint("grepl('a[*]b', x)", lint_msg, linter)

  # also catch char classes for [ and ]
  expect_lint("gregexpr('[]]', x)", lint_msg, linter)
})

test_that("fixed_regex_linter catches null calls to strsplit as well", {
  linter <- fixed_regex_linter()

  expect_no_lint("strsplit(x, '^x')", linter)
  expect_no_lint(R"{strsplit(x, '\\s')}", linter)
  expect_no_lint("strsplit(x, 'a(?=b)', perl = TRUE)", linter)
  expect_no_lint("strsplit(x, '0+1', perl = TRUE)", linter)
  expect_no_lint("strsplit(x, 'a|b')", linter)

  expect_no_lint("tstrsplit(x, '1*2')", linter)
  expect_no_lint("tstrsplit(x, '[a-zA-Z]')", linter)
  expect_no_lint("tstrsplit(x, fmt)", linter)

  # if fixed=TRUE is already set, regex patterns don't matter
  expect_no_lint(R"{strsplit(x, '\\.', fixed = TRUE)}", linter)
  expect_no_lint(R"{strsplit(x, '\\.', fixed = T)}", linter)
})

test_that("fixed_regex_linter catches calls to strsplit as well", {
  linter <- fixed_regex_linter()
  lint_msg <- rex::rex("This regular expression is static")

  expect_lint(R"{strsplit(x, '\\.')}", lint_msg, linter)
  expect_lint("strsplit(x, '[.]')", lint_msg, linter)

  expect_lint("tstrsplit(x, 'abcdefg')", lint_msg, linter)
})

test_that("fixed_regex_linter is more exact about distinguishing \\s from \\:", {
  linter <- fixed_regex_linter()
  lint_msg <- rex::rex("This regular expression is static")

  expect_no_lint(R"{grep('\\s', '', x)}", linter)
  expect_lint(R"{grep('\\:', '', x)}", lint_msg, linter)
})

## tests for stringr functions
test_that("fixed_regex_linter skips allowed stringr usages", {
  linter <- fixed_regex_linter()

  expect_no_lint("str_replace(y, '[a-zA-Z]', '')", linter)
  expect_no_lint("str_replace_all(y, '^x', '')", linter)
  expect_no_lint("str_detect(y, fmt)", linter)
  expect_no_lint(R"{str_extract(y, '\\s')}", linter)
  expect_no_lint(R"{str_extract_all(y, '\\s')}", linter)
  expect_no_lint("str_which(x, '1*2')", linter)

  # if fixed() is already set, regex patterns don't matter
  expect_no_lint(R"{str_replace(y, fixed('\\.'), '')}", linter)

  # namespace qualification doesn't matter
  expect_no_lint("stringr::str_replace(y, stringr::fixed('abcdefg'), '')", linter)
})

test_that("fixed_regex_linter blocks simple disallowed usages of stringr functions", {
  linter <- fixed_regex_linter()
  lint_msg <- rex::rex("This regular expression is static")

  expect_lint(R"{str_replace_all(x, '\\.', '')}", lint_msg, linter)
  expect_lint("str_detect(x, 'abcdefg')", lint_msg, linter)
  expect_lint("str_locate(y, 'a-z')", lint_msg, linter)
  expect_lint(R"{str_subset(x, '\\$')}", lint_msg, linter)
  expect_lint("str_which(x, '\n')", lint_msg, linter)

  # named, positional arguments are still caught
  expect_lint("str_locate(y, pattern = 'a-z')", lint_msg, linter)
  # nor do other named arguments throw things off
  expect_lint(R"{str_starts(x, '\\.', negate = TRUE)}", lint_msg, linter)
})

test_that("fixed_regex_linter catches calls to str_split as well", {
  linter <- fixed_regex_linter()
  lint_msg <- rex::rex("This regular expression is static")

  expect_no_lint("str_split(x, '^x')", linter)
  expect_no_lint("str_split(x, fmt)", linter)

  # if fixed() is already set, regex patterns don't matter
  expect_no_lint(R"{str_split(x, fixed('\\.'))}", linter)
  expect_lint(R"{str_split(x, '\\.')}", lint_msg, linter)
  expect_lint("str_split(x, '[.]')", lint_msg, linter)
})

test_that("str_replace_all's multi-replacement version is handled", {
  linter <- fixed_regex_linter()

  # While each of the replacements is fixed, and this _could_ in principle be replaced by
  #   a pipeline where each step does one of the replacements and fixed() is used, this is overkill.
  #   Instead, ensure that no lint is returned for this case
  expect_no_lint('str_replace_all(x, c("one" = "1", "two" = "2", "three" = "3"))', linter)
  expect_no_lint('grepl(c("a" = "b"), x)', linter)
})

test_that("1- or 2-width octal escape sequences are handled", {
  linter <- fixed_regex_linter()
  lint_msg <- rex::rex("This regular expression is static")

  expect_lint('strsplit(x, "\\1")', lint_msg, linter)
})

test_that("one-character character classes with escaped characters are caught", {
  linter <- fixed_regex_linter()
  lint_msg <- rex::rex("This regular expression is static")

  expect_lint(R"{gsub('[\n]', '', x)}", lint_msg, linter)
  expect_lint(R"{gsub('[\"]', '', x)}", lint_msg, linter)
  expect_lint(R'{gsub("\\<", "x", x, perl = TRUE)}', lint_msg, linter)

  expect_lint(R"{str_split(x, '[\1]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\12]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\123]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\xa]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\x32]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\uF]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\u01]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\u012]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\u0123]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\U8]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\U1d4d7]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\u{1}]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\U{F7D5}]')}", lint_msg, linter)
  expect_lint(R"{str_split(x, '[\U{1D4D7}]')}", lint_msg, linter)
})

test_that("bracketed unicode escapes are caught", {
  linter <- fixed_regex_linter()
  lint_msg <- rex::rex("This regular expression is static")

  expect_lint('gsub("\\u{A0}", " ", out, useBytes = TRUE)', lint_msg, linter)
  expect_lint('gsub("abc\\U{A0DEF}ghi", " ", out, useBytes = TRUE)', lint_msg, linter)
  expect_lint('gsub("\\u{A0}\\U{0001d4d7}", " ", out, useBytes = TRUE)', lint_msg, linter)
})

test_that("escaped characters are handled correctly", {
  linter <- fixed_regex_linter()
  expect_no_lint(R"{gsub('\n+', '', sql)}", linter)
  expect_no_lint('gsub("\\n{2,}", "\n", D)', linter)
  expect_no_lint(R'{gsub("[\r\n]", "", x)}', linter)
  expect_no_lint(R'{gsub("\n $", "", y)}', linter)
  expect_no_lint(R'{gsub("```\n*```r*\n*", "", x)}', linter)
  expect_no_lint('strsplit(x, "(;|\n)")', linter)
  expect_no_lint(R'{strsplit(x, "(;|\n)")}', linter)
  expect_no_lint(R'{grepl("[\\W]", x, perl = TRUE)}', linter)
  expect_no_lint(R'{grepl("[\\W]", x)}', linter)
})

# make sure the logic is properly vectorized
test_that("single expression with multiple regexes is OK", {
  expect_no_lint('c(grep("^a", x), grep("b$", x))', fixed_regex_linter())
})

test_that("fixed replacements vectorize and recognize str_detect", {
  linter <- fixed_regex_linter()
  # properly vectorized
  expect_lint(
    trim_some(R"({
      grepl('abcdefg', x)
      grepl('a[.]\\.b\n', x)
    })"),
    list(
      list(rex::rex('Use "abcdefg" with fixed = TRUE'), line_number = 2L),
      list(rex::rex('Use "a..b\\n" with fixed = TRUE'), line_number = 3L)
    ),
    linter
  )

  # stringr hint works
  expect_lint(
    "str_detect(x, 'abc')",
    rex::rex('Use stringr::fixed("abc") as the pattern'),
    linter
  )
})

test_that("fixed replacement is correct with UTF-8", {
  skip_if(
    .Platform$OS.type == "windows" && !hasName(R.Version(), "crt"),
    message = "UTF-8 support is required"
  )

  expect_lint(
    "grepl('[\\U{1D4D7}]', x)",
    rex::rex('Use "\U1D4D7" with fixed = TRUE'),
    fixed_regex_linter()
  )
})

#' Generate a string with a non-printable Unicode entry robust to test environment
#'
#' Non-printable unicode behaves wildly different with `encodeString()`
#'   across R versions and platforms. Nonetheless, all of these expressions
#'   are valid replacements.
#' @noRd
robust_non_printable_unicode <- function() {
  if (.Platform$OS.type == "windows") {
    "abc\U{0a0def}ghi"
  } else {
    "abc\\U{0a0def}ghi"
  }
}

# styler: off
local({
  .cases <- tibble::tribble(
    ~.test_name,             ~regex_expr,             ~fixed_expr,
    "[.]",                   "[.]",                   ".",
    '[\\\"]',                R'([\"])',               '\\"',
    "[]]",                   "[]]",                   "]",
    R"(\\.)",                R"(\\.)",                ".",
    R"(\\:)",                R"(\\:)",                ":",
    R"(\\<)",                R"(\\<)",                "<",
    R"(\\$)",                R"(\\$)",                "$",
    R"([\1])",               R"([\1])",               R"(\001)",
    R"(\1)",                 R"(\1)",                 R"(\001)",
    R"([\12])",              R"([\12])",              R"(\n)",
    R"([\123])",             R"([\123])",             "S",
    "a[*]b",                 "a[*]b",                 "a*b",
    "abcdefg",               "abcdefg",               "abcdefg",
    "abc\\U{A0DEF}ghi",      "abc\\U{A0DEF}ghi",      robust_non_printable_unicode(),
    "a-z",                   "a-z",                   "a-z",
    R"([\n])",               R"([\n])",               R"(\n)",
    R"(\n)",                 "\n",                    R"(\n)",
    R"([\u01])",             R"([\u01])",             R"(\001)",
    R"([\u012])",            R"([\u012])",            R"(\022)",
    R"([\u0123])",           R"([\u0123])",           "\u0123",
    R"([\u{1}])",            R"([\u{1}])",            R"(\001)",
    R"([\U1d4d7])",          R"([\U1d4d7])",          "\U1D4D7",
    R"([\U{1D4D7}])",        R"([\U{1D4D7}])",        "\U1D4D7",
    R"([\U8])",              R"([\U8])",              R"(\b)",
    R"(\u{A0})",             R"(\u{A0})",             "\uA0",
    R"(\u{A0}\U{0001d4d7})", R"(\u{A0}\U{0001d4d7})", "\uA0\U1D4D7",
    R"([\uF])",              R"([\uF])",              R"(\017)",
    R"([\U{F7D5}])",         R"([\U{F7D5}])",         "\UF7D5",
    R"([\x32])",             R"([\x32])",             "2",
    R"([\xa])",              R"([\xa])",              R"(\n)"
  )
  if (.Platform$OS.type == "windows" && !hasName(R.Version(), "crt")) {
    # These require UTF-8 support
    skip_cases <- c("abc\\U{A0DEF}ghi", "[\\U1d4d7]", "[\\U{1D4D7}]", "\\u{A0}\\U{0001d4d7}")
  } else {
    skip_cases <- character()
  }
  patrick::with_parameters_test_that(
    "fixed replacements of {regex_expr} with {fixed_expr} is correct",
    {
      # TODO(google/patrick#19): handle this more cleanly by skipping up-front
      skip_if(
        regex_expr %in% skip_cases,
        sprintf("regex '%s' is not supported on this system", regex_expr)
      )
      expect_lint(
        sprintf("grepl('%s', x)", regex_expr),
        rex::rex(sprintf('Use "%s" with fixed = TRUE', fixed_expr)),
        fixed_regex_linter()
      )
    },
    .cases = .cases
  )
})
# styler: on

test_that("'unescaped' regex can optionally be skipped", {
  linter <- fixed_regex_linter(allow_unescaped = TRUE)

  expect_no_lint("grepl('a', x)", linter)
  expect_no_lint("str_detect(x, 'a')", linter)
  expect_lint("grepl('[$]', x)", rex::rex('Use "$" with fixed = TRUE'), linter)
})

local({
  linter <- fixed_regex_linter()
  lint_msg <- "This regular expression is static"
  pipes <- pipes(exclude = c("%$%", "%T>%"))

  patrick::with_parameters_test_that(
    "linter is pipe-aware",
    {
      expect_lint(paste("x", pipe, "grepl(pattern = 'a')"), lint_msg, linter)
      expect_no_lint(paste("x", pipe, "grepl(pattern = '^a')"), linter)
      expect_no_lint(paste("x", pipe, "grepl(pattern = 'a', fixed = TRUE)"), linter)
      expect_lint(paste("x", pipe, "str_detect('a')"), lint_msg, linter)
      expect_no_lint(paste("x", pipe, "str_detect('^a')"), linter)
      expect_no_lint(paste("x", pipe, "str_detect(fixed('a'))"), linter)

      expect_lint(paste("x", pipe, "gsub(pattern = 'a', replacement = '')"), lint_msg, linter)
      expect_no_lint(paste("x", pipe, "gsub(pattern = '^a', replacement = '')"), linter)
      expect_no_lint(paste("x", pipe, "gsub(pattern = 'a', replacement = '', fixed = TRUE)"), linter)
      expect_lint(paste("x", pipe, "str_replace('a', '')"), lint_msg, linter)
      expect_no_lint(paste("x", pipe, "str_replace('^a', '')"), linter)
      expect_no_lint(paste("x", pipe, "str_replace(fixed('a'), '')"), linter)
    },
    pipe = pipes,
    .test_name = names(pipes)
  )
})

test_that("pipe-aware lint logic survives adversarial comments", {
  expect_lint(
    trim_some("
      x %>%   grepl(pattern = # comment
      'a')
    "),
    "This regular expression is static",
    fixed_regex_linter()
  )
})

# fuzzer disable: dollar_at
test_that("any_duplicated_linter skips allowed usages", {
  linter <- any_duplicated_linter()

  expect_no_lint("x <- any(y)", linter)
  expect_no_lint("y <- duplicated(z)", linter)

  # extended usage of any is not covered
  expect_no_lint("any(duplicated(y), b)", linter)
  expect_no_lint("any(b, duplicated(y))", linter)
})

test_that("any_duplicated_linter blocks simple disallowed usages", {
  linter <- any_duplicated_linter()
  lint_msg <- rex::rex("anyDuplicated(x, ...) > 0 is better")

  expect_lint("any(duplicated(x))", lint_msg, linter)
  expect_lint("any(duplicated(foo(x)))", lint_msg, linter)
  # na.rm doesn't really matter for this since duplicated can't return NA
  expect_lint("any(duplicated(x), na.rm = TRUE)", lint_msg, linter)
  # also catch nested usage
  expect_lint("foo(any(duplicated(x)))", lint_msg, linter)
})

test_that("any_duplicated_linter catches length(unique()) equivalencies too", {
  linter <- any_duplicated_linter()
  lint_msg_x <- rex::rex("anyDuplicated(x) == 0L is better than length(unique(x)) == length(x)")
  lint_msg_df <- rex::rex("anyDuplicated(DF$col) == 0L is better than length(unique(DF$col)) == nrow(DF)")

  # non-matches
  ## different variable
  expect_no_lint("length(unique(x)) == length(y)", linter)
  ## different table
  expect_no_lint("length(unique(DF$x)) == nrow(DT)", linter)
  expect_no_lint("length(unique(l1$DF$x)) == nrow(l2$DF)", linter)

  # lintable usage
  expect_lint("length(unique(x)) == length(x)", lint_msg_x, linter)
  # argument order doesn't matter
  expect_lint("length(x) == length(unique(x))", lint_msg_x, linter)
  # nrow-style equivalency
  expect_lint("nrow(DF) == length(unique(DF$col))", lint_msg_df, linter)
  expect_lint("nrow(DF) == length(unique(DF[['col']]))", lint_msg_df, linter)
  # match with nesting too
  expect_lint("nrow(l$DF) == length(unique(l$DF[['col']]))", lint_msg_df, linter)

  # !=, <, and > usages are all alternative ways of writing a test for dupes
  #   technically, the direction of > / < matter, but writing
  #   length(unique(x)) > length(x) doesn't seem like it would ever happen.
  expect_lint("length(unique(x)) != length(x)", lint_msg_x, linter)
  expect_lint("length(unique(x)) < length(x)", lint_msg_x, linter)
  expect_lint("length(x) > length(unique(x))", lint_msg_x, linter)
})

test_that("dplyr & data.table equivalents are also linted", {
  linter <- any_duplicated_linter()

  expect_no_lint("uniqueN(x) == nrow(y)", linter)
  expect_no_lint("n_distinct(x) == nrow(y)", linter)
  expect_no_lint("x[, length(unique(col)) == .N()]", linter)
  # some other n function, not dplyr::n
  expect_no_lint("x %>% summarize(length(unique(col)) == n(2))", linter)
  expect_no_lint("x %>% summarize(length(unique(col)) == n)", linter)

  expect_lint("uniqueN(x) == nrow(x)", rex::rex("uniqueN(DF$col) == nrow(DF)"), linter)
  expect_lint("data.table::uniqueN(x) == nrow(x)", rex::rex("uniqueN(DF$col) == nrow(DF)"), linter)
  expect_lint("x[, length(unique(col)) == .N]", rex::rex("length(unique(x)) == .N"), linter)
  expect_lint("x[, uniqueN(col) == .N]", rex::rex("uniqueN(x) == .N"), linter)

  expect_lint("n_distinct(x) == nrow(x)", rex::rex("n_distinct(DF$col) == nrow(DF)"), linter)
  expect_lint("dplyr::n_distinct(x) == nrow(x)", rex::rex("n_distinct(DF$col) == nrow(DF)"), linter)
  expect_lint("x %>% summarize(length(unique(col)) == n())", rex::rex("length(unique(x)) == n()"), linter)
  expect_lint("x %>% summarize(n_distinct(col) == n())", rex::rex("n_distinct(x) == n()"), linter)
})

test_that("any_duplicated_linter catches expression with two types of lint", {
  linter <- any_duplicated_linter()
  lint_msg <- rex::rex("anyDuplicated(DF$col) == 0L is better than length(unique(DF$col)) == nrow(DF)")

  expect_lint(
    trim_some("{
      any(duplicated(x))
      length(unique(DF$col)) == nrow(DF)
    }"),
    list(
      list(rex::rex("anyDuplicated(x, ...) > 0 is better"), line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    linter
  )

  # ditto for different messages within the length(unique()) tests
  expect_lint(
    trim_some("{
      length(unique(x)) == length(x)
      length(unique(DF$col)) == nrow(DF)
    }"),
    list(
      list(rex::rex("anyDuplicated(x) == 0L is better than length(unique(x)) == length(x)"), line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    linter
  )
})
# fuzzer enable: dollar_at

test_that("missing_package_linter skips allowed usages", {
  linter <- missing_package_linter()

  expect_no_lint("library(stats)", linter)
  expect_no_lint('library("stats")', linter)
  expect_no_lint("library('stats')", linter)
  expect_no_lint("library(`stats`)", linter)
  expect_no_lint("library(stats, quietly)", linter)
  expect_no_lint("library(stats, quietly = TRUE)", linter)
  expect_no_lint("require(stats)", linter)
  expect_no_lint("require(stats, quietly = TRUE)", linter)
  expect_no_lint('loadNamespace("stats")', linter)
  expect_no_lint('requireNamespace("stats")', linter)
})

test_that("missing_package_linter blocks disallowed usages", {
  linter <- missing_package_linter()
  lint_msg <- rex::rex("Package 'statts' is not installed.")

  expect_lint("require(statts)", lint_msg, linter)
  expect_lint("library(statts, quietly = TRUE)", lint_msg, linter)
  expect_lint("library(statts, quietly = TRUE)", lint_msg, linter)
  expect_lint('loadNamespace("statts")', lint_msg, linter)
  expect_lint('requireNamespace("statts")', lint_msg, linter)

  expect_lint(
    trim_some("
      library(utils)
      library(statts)
    "),
    list(lint_msg, line_number = 2L, line = "library(statts)"),
    linter
  )
})

test_that("loadNamespace and requireNamespace allow plain symbols", {
  linter <- missing_package_linter()

  expect_no_lint("loadNamespace(mypkg)", linter)
  expect_no_lint("requireNamespace(mypkg)", linter)
})

test_that("character.only=TRUE case is handled", {
  linter <- missing_package_linter()

  expect_no_lint("library(statts, character.only = TRUE)", linter)
  expect_no_lint("require(statts, character.only = TRUE)", linter)
  expect_no_lint('library("stats", character.only = TRUE)', linter)

  expect_lint(
    'library("statts", character.only = TRUE)',
    rex::rex("Package 'statts' is not installed."),
    missing_package_linter()
  )
})

# fuzzer disable: pipe
test_that("pipe_consistency skips allowed usage", {
  linter <- pipe_consistency_linter()

  expect_no_lint("1:3 |> mean() |> as.character()", linter)
  # With no pipes
  expect_no_lint("x <- 1:5", linter)
  # Across multiple lines
  expect_no_lint(
    trim_some("
      1:3 |>
        mean() |>
        as.character()
    "),
    linter
  )
})

test_that("pipe_consistency lints blocked usage", {
  linter <- pipe_consistency_linter()
  lint_message <- rex::rex("Use the |> pipe operator instead of the %>% pipe operator.")


  expect_lint("1:3 %>% mean() %>% as.character()", list(lint_message, lint_message), linter)

  expect_lint(
    "1:3 |> mean() %>% as.character()",
    list(lint_message, line_number = 1L, column_number = 15L),
    linter
  )

  expect_lint(
    "1:3 %>% mean() |> as.character()",
    list(lint_message, line_number = 1L, column_number = 5L),
    linter
  )

  expect_lint(
    trim_some("
      1:3 %>%
        mean() |>
        as.character()
    "),
    list(lint_message, line_number = 1L, column_number = 5L),
    linter
  )

  expect_lint(
    "1:3 |> sort() |> mean() %>% as.character()",
    list(lint_message, line_number = 1L, column_number = 25L),
    linter
  )
})


test_that("pipe_consistency_linter works with |> argument", {
  linter <- pipe_consistency_linter(pipe = "|>")
  lint_message <- rex::rex("Use the |> pipe operator instead of the %>% pipe operator.")
  lint_message_tee <- rex::rex("Use the |> pipe operator instead of the %T>% pipe operator.")

  expect_lint(
    trim_some("
      1:3 %>%
        mean() %>%
        as.character()
    "),
    list(
      list(lint_message, line_number = 1L, column_number = 5L),
      list(lint_message, line_number = 2L, column_number = 10L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      1:3 |>
        mean() %>%
        as.character()
    "),
    list(lint_message, line_number = 2L, column_number = 10L),
    linter
  )

  expect_no_lint(
    "1:3 |> mean() |> as.character()",
    linter
  )

  expect_lint(
    trim_some("
      1:3 |>
        mean() %>%
        as.character()
    "),
    list(lint_message, line_number = 2L, column_number = 10L),
    linter
  )

  expect_lint(
    "1:3 %>% mean() %T>% print()",
    list(lint_message, lint_message_tee),
    linter
  )

  expect_lint(
    "1:3 |> mean() %T>% print()",
    list(lint_message_tee, line_number = 1L, column_number = 15L),
    linter
  )
})

test_that("pipe_consistency_linter works with %>% argument", {
  linter <- pipe_consistency_linter(pipe = "%>%")
  expected_message <- rex::rex("Use the %>% pipe operator instead of the |> pipe operator.")

  expect_lint(
    "1:3 |> mean() |> as.character()",
    list(
      list(message = expected_message, line_number = 1L, column_number = 5L),
      list(message = expected_message, line_number = 1L, column_number = 15L)
    ),
    linter
  )

  expect_lint(
    "1:3 %>% mean() |> as.character()",
    list(message = expected_message, line_number = 1L, column_number = 16L),
    linter
  )

  expect_no_lint(
    "1:3 %>% mean() %>% as.character()",
    linter
  )

  expect_lint(
    trim_some("
      1:3 %>%
        mean() |>
        as.character()
    "),
    list(message = expected_message, line_number = 2L, column_number = 10L),
    linter
  )
})

test_that("simply enforcing a consistent style is supported", {
  linter <- pipe_consistency_linter("auto")
  lint_message <- rex::rex("Stick to one pipe operator; found 1 instances of %>% and 1 instances of |>.")

  expect_no_lint("1:3 %>% mean() %>% as.character()", linter)

  expect_lint(
    "1:3 |> mean() %>% as.character()",
    list(
      list(lint_message, line_number = 1L, column_number = 5L),
      list(lint_message, line_number = 1L, column_number = 15L)
    ),
    linter
  )

  expect_lint(
    "1:3 %>% mean() |> as.character()",
    list(
      list(lint_message, line_number = 1L, column_number = 5L),
      list(lint_message, line_number = 1L, column_number = 16L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      1:3 %>%
        mean() |>
        as.character()
    "),
    list(
      list(lint_message, line_number = 1L, column_number = 5L),
      list(lint_message, line_number = 2L, column_number = 10L)
    ),
    linter
  )

  lint_message_multi <- rex::rex("Stick to one pipe operator; found 1 instances of %>% and 2 instances of |>.")
  expect_lint(
    "1:3 |> sort() |> mean() %>% as.character()",
    list(
      list(lint_message_multi, line_number = 1L, column_number = 5L),
      list(lint_message_multi, line_number = 1L, column_number = 15L),
      list(lint_message_multi, line_number = 1L, column_number = 25L)
    ),
    linter
  )

  expect_no_lint("1:3 %>% mean() %T% print()", linter)
  expect_lint(
    "1:3 |> mean() %T>% print()",
    list(
      list(lint_message, line_number = 1L, column_number = 5L),
      list(lint_message, line_number = 1L, column_number = 15L)
    ),
    linter
  )
})
# fuzzer enable: pipe

test_that("it excludes properly", {
  withr::local_options(
    lintr.exclude = "#TeSt_NoLiNt",
    lintr.exclude_start = "#TeSt_NoLiNt_StArT",
    lintr.exclude_end = "#TeSt_NoLiNt_EnD"
  )

  exclusion_sample <- normalize_path(test_path("exclusions-test"))
  withr::local_dir(withr::local_tempdir())
  file.copy(exclusion_sample, ".")

  t1 <- lint("exclusions-test")
  expect_length(t1, 8L)

  t2 <- lint("exclusions-test", exclusions = list("exclusions-test" = 4L))
  expect_length(t2, 7L)

  t3 <- lint("exclusions-test", exclusions = list("exclusions-test"))
  expect_length(t3, 0L)

  cache_path <- file.path(tempdir(), "lintr_cache")
  clear_cache("exclusions-test", cache_path)
  for (info in sprintf("caching: pass %s", 1L:4L)) {
    t4 <- lint("exclusions-test", cache = cache_path)

    expect_identical(length(t4), 8L, info = info)
  }
})

test_that("it doesn't fail when encountering misspecified encodings", {
  withr::local_options(
    lintr.exclude = "#TeSt_NoLiNt",
    lintr.exclude_start = "#TeSt_NoLiNt_StArT",
    lintr.exclude_end = "#TeSt_NoLiNt_EnD",
    lintr.linter_file = tempfile()
  )

  withr::local_dir(test_path("dummy_projects", "project"))
  expect_no_error(lint("cp1252.R"))
})

test_that("it gives the expected error message when there is only one start but no end", {
  withr::local_options(
    lintr.exclude = "#TeSt_NoLiNt",
    lintr.exclude_start = "#TeSt_NoLiNt_StArT",
    lintr.exclude_end = "#TeSt_NoLiNt_EnD",
    lintr.linter_file = tempfile()
  )
  withr::local_dir(test_path("dummy_projects", "project"))

  expect_error(
    lint("one_start_no_end.R"),
    "has 1 range start (line 3) and 0 range ends",
    fixed = TRUE
  )
})

test_that("it gives the expected error message when there is mismatch between multiple starts and ends", {
  withr::local_options(
    lintr.exclude = "#TeSt_NoLiNt",
    lintr.exclude_start = "#TeSt_NoLiNt_StArT",
    lintr.exclude_end = "#TeSt_NoLiNt_EnD",
    lintr.linter_file = tempfile()
  )
  withr::local_dir(test_path("dummy_projects", "project"))

  expect_error(
    lint("mismatched_starts_ends.R"),
    "has 3 range starts (lines 3, 7, 11) and 2 range ends (lines 1, 9)",
    fixed = TRUE
  )
})

test_that("partial matching works for exclusions but warns if no linter found", { # nofuzz: assignment comment_injection
  withr::local_dir(test_path("dummy_projects", "project"))

  expect_warning(
    expect_warning(
      expect_warning(
        expect_lint(
          file = "partially_matched_exclusions.R",
          checks = rex::rex("semicolons"),
          parse_settings = FALSE
        ),
        rex::rex("Could not find linter named ", anything, "s")
      ),
      rex::rex("Could not find linter named ", anything, "bogus_linter")
    ),
    rex::rex("Could not find linters named ", anything, "hocus_pocus", anything, "bogus")
  )
})

test_that("#1413: lint_dir properly excludes files", {
  withr::local_options(lintr.linter_file = "lintr_test_config")
  tmp <- withr::local_tempdir()
  writeLines(
    trim_some("
      linters: linters_with_defaults(
          line_length_linter(10)
        )
      exclusions: list(
          'bad.R' = list(
            1, # global exclusions are unnamed
            line_length_linter = 4:6
          )
        )
    "),
    file.path(tmp, "lintr_test_config")
  )

  writeLines(
    trim_some("
      tmp = 'value'

      # comment
      # long comment
      # long comment
      # long comment
      # comment
    "),
    file.path(tmp, "bad.R")
  )

  expect_length(lint(file.path(tmp, "bad.R")), 0L)
  expect_length(lint_dir(tmp), 0L)
})

test_that("#1442: is_excluded_files works if no global exclusions are specified", {
  withr::local_options(lintr.linter_file = "lintr_test_config")
  tmp <- withr::local_tempdir()

  writeLines(
    trim_some("
      linters: linters_with_defaults(
          line_length_linter(10)
        )
      exclusions: list(
          'bad.R' = list(
            line_length_linter = 4:6
          )
        )
    "),
    file.path(tmp, "lintr_test_config")
  )

  writeLines(
    trim_some("
      tmp = 'value'

      # comment
      # long comment
      # long comment
      # long comment
      # comment
    "),
    file.path(tmp, "bad.R")
  )

  # 3 lints: assignment_linter(), quotes_linter() and line_length_linter()
  expect_lint( # nofuzz
    file = file.path(tmp, "bad.R"),
    checks = list(
      list(linter = "assignment_linter", line_number = 1L),
      list(linter = "quotes_linter", line_number = 1L),
      list(linter = "line_length_linter", line_number = 1L)
    )
  )
  expect_length(lint_dir(tmp), 3L)
})

test_that("next-line exclusion works", { # nofuzz
  withr::local_options(
    lintr.exclude = "# NL",
    lintr.exclude_next = "# NLN",
    lintr.exclude_linter = default_settings$exclude_linter
  )

  linter <- assignment_linter()

  # blanket exclusion works
  expect_no_lint(
    trim_some("
      # NLN
      x = 1
    "),
    linter
  )

  # specific exclusion works
  expect_no_lint(
    trim_some("
      # NLN: assignment_linter.
      x = 1
    "),
    linter
  )
  expect_no_lint(
    trim_some("
      # NLN: assignment.
      x = 1
    "),
    linter
  )
  expect_lint(
    trim_some("
      # NLN: line_length_linter.
      x = 1
    "),
    rex::rex("Use <- for assignment, not =."),
    list(linter, line_length_linter())
  )

  # interaction with plain nolint
  expect_lint(
    trim_some("
      x = 1 # NLN: assignment_linter.
      x = 2
    "),
    list(rex::rex("Use <- for assignment, not =."), line_number = 1L),
    linter
  )
})

test_that("capture groups work as intended (#2831)", { # nofuzz: assignment comment_injection
  linter <- assignment_linter()
  lint_msg <- rex::rex("Use <- for assignment, not =.")

  expect_lint("x = 1", lint_msg, linters = linter, exclude = "(a)|(b)")
  expect_no_lint("a = 1", linters = linter, exclude = "(a)|(b)")
  expect_no_lint("a = 1", linters = linter, exclude = "(?:a)|(?:b)")
  expect_no_lint("a = 1", linters = linter, exclude = "(?:a)|(b)")
  expect_no_lint("b = 1", linters = linter, exclude = "(a)|(?:b)")

  # named groups
  expect_no_lint("a = 1", linters = linter, exclude = "(?<group1>a)|(?<group2>b)")
  # nested groups
  expect_no_lint("a = 1", linters = linter, exclude = "((a)|(b))")

  # exclude_start / exclude_end with capture groups
  expect_no_lint(
    trim_some("
      x = 1 # nolint start
      y = 2
      z = 3 # nolint end
    "),
    linters = linter,
    exclude_start = "(# nolint start)",
    exclude_end = "(# nolint end)"
  )

  # exclude_next with capture groups
  expect_lint(
    trim_some("
      x = 1 # nolint next
      y = 2
    "),
    list(lint_msg, line_number = 1L),
    linters = assignment_linter(),
    exclude_next = "(# nolint next)"
  )
})

test_that("marginfigure engine from tufte package doesn't cause problems", {
  skip_if_not_installed("tufte", minimum_version = "0.12.4") # for rstudio/tufte#117
  loadNamespace("tufte") # to register additional engines

  expect_lint(
    file = test_path("knitr_extended_formats", "tufte.Rmd"),
    checks = list(rex::rex("Use <- for assignment, not =."), line_number = 11L),
    default_linters,
    parse_settings = FALSE
  )
})

test_that("engines from bookdown package cause no problems", {
  skip_if_not_installed("bookdown")
  loadNamespace("bookdown") # to register additional engines

  expect_lint(
    file = test_path("knitr_extended_formats", "bookdown.Rmd"),
    checks = list(rex::rex("Use <- for assignment, not =."), line_number = 14L),
    default_linters,
    parse_settings = FALSE
  )
})

test_that("any_is_na_linter skips allowed usages", {
  linter <- any_is_na_linter()

  expect_no_lint("x <- any(y)", linter)

  expect_no_lint("y <- is.na(z)", linter)

  # extended usage of ... arguments to any is not covered
  expect_no_lint("any(is.na(y), b)", linter)
  expect_no_lint("any(b, is.na(y))", linter)

  # negation shouldn't list
  expect_no_lint("any(!is.na(x))", linter)
  expect_no_lint("any(!is.na(foo(x)))", linter)
})

test_that("any_is_na_linter blocks simple disallowed usages", {
  linter <- any_is_na_linter()
  lint_message <- rex::rex("anyNA(x) is better than any(is.na(x)).")

  expect_lint("any(is.na(x))", lint_message, linter)
  expect_lint("any(is.na(foo(x)))", lint_message, linter)
  # na.rm doesn't really matter for this since is.na can't return NA
  expect_lint("any(is.na(x), na.rm = TRUE)", lint_message, linter)
  # also catch nested usage
  expect_lint("foo(any(is.na(x)))", lint_message, linter)
})

test_that("NA %in% x is also found", {
  linter <- any_is_na_linter()
  lint_message <- rex::rex("anyNA(x) is better than NA %in% x.")

  expect_lint("NA %in% x", lint_message, linter)
  expect_lint("NA_real_ %in% x", lint_message, linter)
  expect_no_lint("NA_not_a_sentinel_ %in% x", linter)
})

test_that("lints vectorize", {
  any_message <- rex::rex("any(is.na(x))")
  in_message <- rex::rex("NA %in% x")

  expect_lint(
    trim_some("{
      any(is.na(foo(x)))
      any(is.na(y), na.rm = TRUE)
      NA %in% a
      NA_complex_ %in% b
    }"),
    list(
      list(any_message, line_number = 2L),
      list(any_message, line_number = 3L),
      list(in_message, line_number = 4L),
      list(in_message, line_number = 5L)
    ),
    any_is_na_linter()
  )
})

test_that("vector_logic_linter skips allowed usages", {
  linter <- vector_logic_linter()

  expect_no_lint("if (TRUE) 5 else if (TRUE) 2", linter)
  expect_no_lint("if (TRUE || FALSE) 1; while (TRUE && FALSE) 2", linter)

  # function calls and extractions may aggregate to scalars -- only catch
  #   usages at the highest logical level
  expect_no_lint("if (agg_function(x & y)) 1", linter)
  expect_no_lint("if (DT[x | y, cond]) 1", linter)

  # don't match potentially OK usages nested within calls
  expect_no_lint("if (TRUE && any(TRUE | FALSE)) 4", linter)
  # even if the usage is nested in those calls (b/181915948)
  expect_no_lint("if (TRUE && any(TRUE | FALSE | TRUE)) 4", linter)

  # don't match potentially OK usages in the branch itself
  expect_no_lint(
    trim_some("
      if (TRUE) {
        x | y
      }
    "),
    linter
  )

  # valid nested usage within aggregator
  expect_no_lint("testthat::expect_false(any(TRUE | TRUE))", linter)
})

test_that("vector_logic_linter blocks simple disallowed usages", {
  linter <- vector_logic_linter()

  expect_lint("if (TRUE & FALSE) 1", rex::rex("Use `&&` in conditional expressions."), linter)
  expect_lint("while (TRUE | TRUE) 2", rex::rex("Use `||` in conditional expressions."), linter)
})

test_that("vector_logic_linter detects nested conditions", {
  linter <- vector_logic_linter()

  expect_lint(
    "if (TRUE & TRUE || FALSE) 4",
    list(rex::rex("Use `&&` in conditional expressions."), column_number = 10L),
    linter
  )
  expect_lint(
    "if (TRUE && (TRUE | FALSE)) 4",
    list(rex::rex("Use `||` in conditional expressions."), column_number = 19L),
    linter
  )
})

test_that("vector_logic_linter catches usages in expect_true()/expect_false()", {
  linter <- vector_logic_linter()
  and_msg <- rex::rex("Use `&&` in conditional expressions.")
  or_msg <- rex::rex("Use `||` in conditional expressions.")

  expect_lint("expect_true(TRUE & FALSE)", and_msg, linter)
  expect_lint("expect_false(TRUE | TRUE)", or_msg, linter)

  # ditto with namespace qualification
  expect_lint("testthat::expect_true(TRUE & FALSE)", and_msg, linter)
  expect_lint("testthat::expect_false(TRUE | TRUE)", or_msg, linter)
})

test_that("vector_logic_linter doesn't get mixed up from complex usage", {
  expect_no_lint(
    trim_some("
      if (a) {
        expect_true(ok)
        x <- 2
        a | b
      }
    "),
    vector_logic_linter()
  )
})

test_that("vector_logic_linter recognizes some false positves around bitwise &/|", {
  linter <- vector_logic_linter()

  expect_no_lint("if (info & as.raw(12)) { }", linter)
  expect_no_lint("if (as.raw(12) & info) { }", linter)
  expect_no_lint("if (info | as.raw(12)) { }", linter)
  expect_no_lint("if (info & as.octmode('100')) { }", linter)
  expect_no_lint("if (info | as.octmode('011')) { }", linter)
  expect_no_lint("if (info & as.hexmode('100')) { }", linter)
  expect_no_lint("if (info | as.hexmode('011')) { }", linter)
  # implicit as.octmode() coercion
  expect_no_lint("if (info & '100') { }", linter)
  expect_no_lint("if (info | '011') { }", linter)
  expect_no_lint("if ('011' | info) { }", linter)

  # further nesting
  expect_no_lint("if ((info & as.raw(12)) == as.raw(12)) { }", linter)
  expect_no_lint("if ((info | as.raw(12)) == as.raw(12)) { }", linter)
  expect_no_lint('if ((mode & "111") != as.octmode("111")) { }', linter)
  expect_no_lint('if ((mode | "111") != as.octmode("111")) { }', linter)
  expect_no_lint('if ((mode & "111") != as.hexmode("111")) { }', linter)
  expect_no_lint('if ((mode | "111") != as.hexmode("111")) { }', linter)
})

test_that("incorrect subset/filter usage is caught", {
  linter <- vector_logic_linter()
  and_msg <- rex::rex("Use `&` in subsetting expressions")
  or_msg <- rex::rex("Use `|` in subsetting expressions")

  expect_lint("filter(x, y && z)", and_msg, linter)
  expect_lint("filter(x, y || z)", or_msg, linter)
  expect_lint("subset(x, y && z)", and_msg, linter)
  expect_lint("subset(x, y || z)", or_msg, linter)

  expect_lint("x %>% filter(y && z)", and_msg, linter)
  expect_lint("filter(x, a & b, c | d, e && f)", list(and_msg, column_number = 27L), linter)
})

test_that("native pipe usage is caught in subset/filter logic", {
  expect_lint("x |> filter(y && z)", rex::rex("Use `&` in subsetting"), vector_logic_linter())
})

test_that("subsetting logic handles nesting", {
  linter <- vector_logic_linter()
  and_msg <- rex::rex("Use `&` in subsetting expressions")
  or_msg <- rex::rex("Use `|` in subsetting expressions")

  expect_lint("filter(x, a & b || c)", or_msg, linter)
  expect_lint("filter(x, a && b | c)", and_msg, linter)

  # adversarial commenting
  expect_lint(
    trim_some("
      filter(x, a #comment
      && b | c)
    "),
    and_msg,
    linter
  )

  expect_lint(
    trim_some("
      filter(x, a && #comment
      b | c)
    "),
    and_msg,
    linter
  )

  # but not valid usage
  expect_no_lint("filter(x, y < mean(y, na.rm = AA && BB))", linter)
  expect_no_lint("subset(x, y < mean(y, na.rm = AA && BB) & y > 0)", linter)
  expect_no_lint("subset(x, y < x[y > 0, drop = AA && BB, y])", linter)
})

test_that("scalar logic in anonymous functions within filter/subset is allowed", {
  linter <- vector_logic_linter()

  expect_no_lint("filter(x, vapply(y, function(i) is.numeric(i) && mean(i) > 1, NA))", linter)
  expect_no_lint("subset(x, sapply(y, function(i) is.null(i) || length(i) == 0))", linter)
  expect_no_lint("filter(x, vapply(y, \\(i) i && z, NA))", linter)
  expect_no_lint("filter(x, Map(function(a, b) a && b, x, y))", linter)
  expect_no_lint("filter(x, vapply(y, function(i) any(sapply(i, function(j) j && z)), NA))", linter)
  expect_no_lint("filter(x, vapply(y, function(i) { is.numeric(i) && mean(i) > 1 }, NA))", linter)
})

test_that("vector logic in if conditions inside anonymous functions is still linted", {
  linter <- vector_logic_linter()
  or_msg <- rex::rex("Use `||` in conditional expressions.")
  and_msg <- rex::rex("Use `&&` in conditional expressions.")

  expect_lint(
    "subset(x, sapply(col, function(x) { if (A | B) do_ab(x) else do_other(x) }))",
    or_msg,
    linter
  )
  expect_lint(
    "filter(x, vapply(y, function(i) { if (a & b) 1 else 0 }, 1))",
    and_msg,
    linter
  )
  expect_lint(
    "subset(x, sapply(col, \\(x) { if (A | B) x }))",
    or_msg,
    linter
  )
  expect_lint(
    "subset(x, sapply(col, function(x) { while (A | B) x }))",
    or_msg,
    linter
  )
  expect_lint(
    "filter(x, vapply(y, function(i) if (A | B) 1 else 0, 1))",
    or_msg,
    linter
  )
  expect_no_lint("subset(x, sapply(col, function(x) { if (any(a | b)) x }))", linter)
})

test_that("filter() handling is conservative about stats::filter()", {
  linter <- vector_logic_linter()
  and_msg <- rex::rex("Use `&` in subsetting expressions")

  # NB: this should be invalid, filter= is a vector argument
  expect_no_lint("stats::filter(x, y && z)", linter)
  # The only logical argument to stats::filter(), exclude by keyword
  expect_no_lint("filter(x, circular = y && z)", linter)
  # But presence of circular= doesn't invalidate lint
  expect_lint("filter(x, circular = TRUE, y && z)", and_msg, linter)
  expect_lint("filter(x, y && z, circular = TRUE)", and_msg, linter)
  expect_no_lint(
    trim_some("
      filter(x, circular # comment
      = y && z)
    "),
    linter
  )
  expect_no_lint(
    trim_some("
      filter(x, circular = # comment
        y && z)
    "),
    linter
  )
  expect_no_lint(
    trim_some("
      filter(x, circular # comment
      = # comment
      y && z)
    "),
    linter
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      if (AA & BB) {}
      if (CC | DD) {}
      filter(x, EE && FF)
      subset(y, GG || HH)
    }"),
    list(
      list(rex::rex("`&&`"), line_number = 2L),
      list(rex::rex("`||`"), line_number = 3L),
      list(rex::rex("`&`"), line_number = 4L),
      list(rex::rex("`|`"), line_number = 5L)
    ),
    vector_logic_linter()
  )
})

test_that("numeric_leading_zero_linter skips allowed usages", {
  linter <- numeric_leading_zero_linter()

  expect_no_lint("a <- 0.1", linter)
  expect_no_lint("b <- -0.2", linter)
  expect_no_lint("c <- 3.0", linter)
  expect_no_lint("d <- 4L", linter)
  expect_no_lint("e <- TRUE", linter)
  expect_no_lint("f <- 0.5e6", linter)
  expect_no_lint("g <- 0x78", linter)
  expect_no_lint("h <- 0.9 + 0.1i", linter)
  expect_no_lint("h <- 0.9+0.1i", linter)
  expect_no_lint("h <- 0.9 - 0.1i", linter)
  expect_no_lint("i <- 2L + 3.4i", linter)
})

test_that("numeric_leading_zero_linter blocks simple disallowed usages", {
  linter <- numeric_leading_zero_linter()
  lint_msg <- rex::rex("Include the leading zero for fractional numeric constants.")

  expect_lint("a <- .1", lint_msg, linter)
  expect_lint("b <- -.2", lint_msg, linter)
  expect_lint("c <- .3 + 4.5i", lint_msg, linter)
  expect_lint("d <- 6.7 + .8i", lint_msg, linter)
  expect_lint("d <- 6.7+.8i", lint_msg, linter)
  expect_lint("e <- .9e10", lint_msg, linter)
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Include the leading zero for fractional numeric constants.")

  expect_lint(
    trim_some("{
      .1
      -.2
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    numeric_leading_zero_linter()
  )
})

test_that("object_overwrite_linter skips allowed usages", {
  linter <- object_overwrite_linter()

  expect_no_lint("function() DT <- data.frame(a = 1)", linter)

  # don't block names subassigned e.g. as columns or list elements
  expect_no_lint("function() x$sd <- sd(rnorm(100))", linter)

  # These virtual names are ignored to slightly reduce the search space
  expect_no_lint("function() .__C__logical <- TRUE", linter)
})

test_that("object_overwrite_linter blocks simple disallowed usages", {
  linter <- object_overwrite_linter()

  expect_lint(
    "foo <- function() data <- mtcars",
    rex::rex("'data' is an exported object from package 'utils'."),
    linter
  )

  expect_lint(
    trim_some("
      foo <- function() {
        sigma <- sd(rnorm(100))
      }
    "),
    rex::rex("'sigma' is an exported object from package 'stats'."),
    linter
  )

  # base and graphics both export 'plot' (in recent R); ensure this is no issue
  plot_pkg <- environmentName(environment(plot))
  expect_lint(
    "function() plot <- 1",
    rex::rex("'plot' is an exported object from package '", plot_pkg, "'."),
    linter
  )

  # not just the top level of the function
  expect_lint(
    trim_some("
      foo <- function() {
        if (TRUE) {
          sigma <- sd(rnorm(100))
        } else {
          all <- FALSE
        }
      }
    "),
    list(
      rex::rex("'sigma' is an exported object from package 'stats'."),
      rex::rex("'all' is an exported object from package 'base'.")
    ),
    linter
  )
})

test_that("Non-syntactic names are matched & linted (#2346)", {
  linter <- object_overwrite_linter()
  lint_msg <- rex::rex("'+' is an exported object from package 'base'.")

  expect_lint(
    trim_some("
      foo <- function() {
        `+` <- 2L
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some('
      foo <- function() {
        "+" <- 2L
      }
    '),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function() {
        '+' <- 2L
      }
    "),
    lint_msg,
    linter
  )
})

test_that("object_overwrite_linter skips any name assigned at the top level", {
  linter <- object_overwrite_linter()

  expect_no_lint("data <- mtcars", linter)
  expect_no_lint("sigma <- sd(rnorm(100))", linter)
})

test_that("object_overwrite_linter skips argument names", {
  linter <- object_overwrite_linter()

  expect_no_lint("foo <- function(data) data <- data + 1", linter)
  expect_no_lint("foo <- function(data) data = data + 1", linter)

  expect_no_lint(
    trim_some("
      bar <- function(a, b, c, sigma) {
        sigma <- a * b * c ^ sigma
      }
    "),
    linter
  )
})

test_that("object_overwrite_linter skips data.table assignments with :=", {
  expect_no_lint("foo <- function() x[, title := 4]", object_overwrite_linter())
})

test_that("object_overwrite_linter optionally accepts package names", {
  expect_no_lint("function() data <- 1", object_overwrite_linter(packages = "base"))

  expect_lint(
    "function() lint <- TRUE",
    rex::rex("'lint' is an exported object from package 'lintr'."),
    object_overwrite_linter(packages = "lintr")
  )
})

test_that("non-<- assignments are detected", {
  linter <- object_overwrite_linter()
  lint_msg <- rex::rex("'sum' is an exported object from package 'base'.")

  expect_lint("function(x) sum(x) -> sum", lint_msg, linter)
  expect_lint("function(x) sum <<- sum(x)", lint_msg, linter)
  expect_lint("function(x) sum(x) ->> sum", lint_msg, linter)
  expect_lint("function(x) sum = sum(x)", lint_msg, linter)
})

test_that("shorthand lambda is detected", {
  expect_lint("\\() data <- 1", "'data' is an exported object", object_overwrite_linter())
})

test_that("allow_names= works to ignore certain symbols", {
  expect_no_lint("function() data <- 1", object_overwrite_linter(allow_names = "data"))
})

test_that("lints vectorize", {
  lines <- trim_some("{
    foo <- function() {
      data <- 1
      var <- 2
    }
    bar <- function(data) {
      data <- data + 3
      sum <- 4
    }
  }")
  expect_lint(
    lines,
    list(
      list(rex::rex("'data' is an exported object from package 'utils'."), line_number = 3L),
      list(rex::rex("'var' is an exported object from package 'stats'."), line_number = 4L),
      list(rex::rex("'sum' is an exported object from package 'base'."), line_number = 8L)
    ),
    object_overwrite_linter()
  )

  expect_lint(
    lines,
    list(rex::rex("'var' is an exported object from package 'stats'."), line_number = 4L),
    object_overwrite_linter(packages = c("stats", "base"), allow_names = "sum")
  )
})

test_that("Validation works", {
  expect_error(object_overwrite_linter(packages = "invalid_"), "required, but not available", fixed = TRUE)
})

regexes <- list(
  assign = rex::rex("Use <- for assignment, not =."),
  local_var = rex::rex("local variable"),
  quotes = rex::rex("Only use double-quotes."),
  trailing = rex::rex("Remove trailing blank lines."),
  trailws = rex::rex("Remove trailing whitespace."),
  indent = rex::rex("Indentation should be")
)

test_that("it handles dir", {
  file_pattern <- rex::rex(".R", one_of("html", "md", "nw", "rst", "tex", "txt"))

  lints <- lint_dir(path = "knitr_formats", pattern = file_pattern, parse_settings = FALSE)

  # For every file there should be at least 1 lint
  expect_identical(
    sort(unique(names(lints))),
    sort(list.files(test_path("knitr_formats"), pattern = file_pattern))
  )
})

test_that("it handles markdown", {
  expect_lint(
    file = test_path("knitr_formats", "test.Rmd"),
    checks = list(
      list(regexes[["assign"]], line_number = 9L),
      list(regexes[["local_var"]], line_number = 22L),
      list(regexes[["assign"]], line_number = 22L),
      list(regexes[["trailing"]], line_number = 24L)
    ),
    linters = default_linters,
    parse_settings = FALSE
  )
})

test_that("it handles quarto", {
  expect_lint(
    file = test_path("knitr_formats", "test.qmd"),
    checks = list(
      list(regexes[["assign"]], line_number = 9L),
      list(regexes[["local_var"]], line_number = 22L),
      list(regexes[["assign"]], line_number = 22L),
      list(regexes[["trailing"]], line_number = 24L)
    ),
    linters = default_linters,
    parse_settings = FALSE
  )
})

test_that("it handles Sweave", {
  expect_lint(
    file = test_path("knitr_formats", "test.Rnw"),
    checks = list(
      list(regexes[["assign"]], line_number = 12L),
      list(regexes[["local_var"]], line_number = 24L),
      list(regexes[["assign"]], line_number = 24L),
      list(regexes[["trailing"]], line_number = 26L)
    ),
    linters = default_linters,
    parse_settings = FALSE
  )
})

test_that("it handles reStructuredText", {
  expect_lint(
    file = test_path("knitr_formats", "test.Rrst"),
    checks = list(
      list(regexes[["assign"]], line_number = 10L),
      list(regexes[["local_var"]], line_number = 23L),
      list(regexes[["assign"]], line_number = 23L),
      list(regexes[["trailing"]], line_number = 25L)
    ),
    linters = default_linters,
    parse_settings = FALSE
  )
})

test_that("it handles HTML", {
  expect_lint(
    file = test_path("knitr_formats", "test.Rhtml"),
    checks = list(
      list(regexes[["assign"]], line_number = 15L),
      list(regexes[["local_var"]], line_number = 27L),
      list(regexes[["assign"]], line_number = 27L),
      list(regexes[["trailing"]], line_number = 29L)
    ),
    linters = default_linters,
    parse_settings = FALSE
  )
})

test_that("it handles tex", {
  expect_lint(
    file = test_path("knitr_formats", "test.Rtex"),
    checks = list(
      list(regexes[["indent"]], line_number = 11L),
      list(regexes[["assign"]], line_number = 11L),
      list(regexes[["indent"]], line_number = 22L),
      list(regexes[["local_var"]], line_number = 23L),
      list(regexes[["assign"]], line_number = 23L),
      list(regexes[["trailing"]], line_number = 25L),
      list(regexes[["trailws"]], line_number = 25L)
    ),
    linters = default_linters,
    parse_settings = FALSE
  )
})

test_that("it handles asciidoc", {
  expect_lint(
    file = test_path("knitr_formats", "test.Rtxt"),
    checks = list(
      list(regexes[["assign"]], line_number = 9L),
      list(regexes[["local_var"]], line_number = 22L),
      list(regexes[["assign"]], line_number = 22L),
      list(regexes[["trailing"]], line_number = 24L)
    ),
    linters = default_linters,
    parse_settings = FALSE
  )
})

test_that("it does _not_ handle brew", { # nofuzz: comment_injection
  expect_lint("'<% a %>'\n",
    checks = list(
      regexes[["quotes"]],
      regexes[["trailing"]]
    ),
    default_linters
  )
})

test_that("it does _not_ error with inline \\Sexpr", {
  expect_no_lint(
    "#' text \\Sexpr{1 + 1} more text",
    default_linters
  )
})

test_that("it does lint .Rmd or .qmd file with malformed input", {
  expect_lint(
    file = test_path("knitr_malformed", "incomplete_r_block.Rmd"),
    checks = "Missing chunk end",
    linters = default_linters,
    parse_settings = FALSE
  )

  expect_lint(
    file = test_path("knitr_malformed", "incomplete_r_block.qmd"),
    checks = "Missing chunk end",
    linters = default_linters,
    parse_settings = FALSE
  )

  contents <- c(
    trim_some("
      ```{r chunk}
      lm(x ~ y)


      # some text

      ```
      bash some_script.sh
      ```
    "),
    trim_some("
      ```{r chunk-1}
      code <- 42

      # A heading
      Some text

      ```{r chunk-2}
      some_more_code <- 42
      ```
    "),
    trim_some("
      ```{r chunk-1}
      code <- 42
      ```

      # A heading
      Some text

      ```{r chunk-2}
      some_more_code <- 42
    ")
  )

  expected <- list(
    NULL, # This test case would require parsing all chunk fences, not just r chunks.
    "maybe starting at line 1",
    "maybe starting at line 8"
  )

  for (i in seq_along(contents)) {
    expect_lint(contents[[i]], expected[[i]], linters = list())
  }
})

test_that("chunkless files are fine", {
  tmp <- withr::local_tempfile(fileext = ".Rmd", lines = c(
    "---",
    "some_option: true",
    "---",
    "Some text!"
  ))
  expect_no_lint(file = tmp, linters = assignment_linter())
})

test_that("it skips eval=FALSE chunks (#1964)", {
  linter <- assignment_linter()

  expect_lint(
    trim_some("
      ```{r label, eval=FALSE}
      bad_code = 1
      ```

      ```{r}
      good_code = 2
      ```

      ```{r, eval=F}
      bad_code = 1
      ```

      ```{r, eval=TRUE}
      good_code = 3
      ```
    "),
    list(
      list(regexes[["assign"]], line_number = 6L),
      list(regexes[["assign"]], line_number = 14L)
    ),
    linter
  )

  # .qmd with #| eval: false in chunk body
  qmd_file <- withr::local_tempfile(fileext = ".qmd", lines = c(
    "```{r}",
    "#| eval: false",
    "bad_code = 1",
    "```",
    "```{r}",
    "good_code = 2",
    "```"
  ))
  expect_lint(
    file = qmd_file,
    checks = list(regexes[["assign"]], line_number = 6L),
    linter
  )

  # .Rnw with eval=FALSE
  expect_lint(
    trim_some("
      <<chunk-1, eval=FALSE>>=
      bad_code = 1
      @

      <<chunk-2, eval=TRUE>>=
      good_code = 2
      @
    "),
    list(regexes[["assign"]], line_number = 6L),
    linter
  )
})

test_that("malformed chunk options don't crash linting and fallback to evaluated", {
  linter <- assignment_linter()

  # syntax error in header options
  tmp_csv_err <- withr::local_tempfile(fileext = ".Rmd", lines = c(
    "```{r, eval=1+}",
    "bad_code = 1",
    "```"
  ))
  expect_silent(
    expect_lint(
      file = tmp_csv_err,
      checks = list(regexes[["assign"]], line_number = 2L),
      linters = linter
    )
  )

  # divide_chunk error (YAML syntax error in body options)
  tmp_yaml_err <- withr::local_tempfile(fileext = ".qmd", lines = c(
    "```{r}",
    "#| eval: {",
    "bad_code = 1",
    "```"
  ))
  expect_silent(
    expect_lint(
      file = tmp_yaml_err,
      checks = list(regexes[["assign"]], line_number = 3L),
      linters = linter
    )
  )

  # divide_chunk warning (YAML warning - not a list)
  tmp_yaml_warn <- withr::local_tempfile(fileext = ".qmd", lines = c(
    "```{r}",
    '#| "key: 1"',
    "bad_code = 1",
    "```"
  ))
  expect_silent(
    expect_lint(
      file = tmp_yaml_warn,
      checks = list(regexes[["assign"]], line_number = 3L),
      linters = linter
    )
  )
})

test_that("library_call_linter skips allowed usages", {
  linter <- library_call_linter()

  expect_no_lint(
    trim_some("
      library(dplyr)
      print('test')
    "),
    linter
  )

  expect_no_lint("print('test')", linter)

  expect_no_lint(
    trim_some("
      # comment
      library(dplyr)
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      print('test')
      # library(dplyr)
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      suppressPackageStartupMessages({
        library(dplyr)
        library(knitr)
      })
    "),
    linter
  )
})

test_that("library_call_linter warns on disallowed usages", {
  linter <- library_call_linter()
  lint_message <- rex::rex("Move all library calls to the top of the script.")

  expect_lint(
    trim_some("
      library(dplyr)
      print('test')
      library(tidyr)
    "),
    lint_message,
    linter
  )

  expect_lint(
    trim_some("
      library(dplyr)
      print('test')
      library(tidyr)
      library(purrr)
    "),
    list(
      list(lint_message, line_number = 3L, column_number = 1L),
      list(lint_message, line_number = 4L, column_number = 1L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      library(dplyr)
      print('test')
      print('test')
      library(tidyr)
    "),
    lint_message,
    linter
  )

  expect_lint(
    trim_some("
      library(dplyr)
      print('test')
      library(tidyr)
      print('test')
    "),
    lint_message,
    linter
  )

  expect_lint(
    trim_some("
      library(dplyr)
      print('test')
      library(tidyr)
      print('test')
      library(purrr)
    "),
    list(
      list(lint_message, line_number = 3L, column_number = 1L),
      list(lint_message, line_number = 5L, column_number = 1L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      library(dplyr)
      print('test')
      suppressMessages(library('lubridate', character.only = TRUE))
      suppressMessages(library(tidyr))
      print('test')
    "),
    list(
      list(rex::rex("Unify consecutive calls to suppressMessages()"), line_number = 3L),
      list(lint_message, line_number = 3L),
      list(rex::rex("Use symbols in library calls to avoid the need for 'character.only'"), line_number = 3L),
      list(lint_message, line_number = 4L)
    ),
    linter
  )
})

test_that("require() treated the same as library()", {
  linter <- library_call_linter()
  lint_message_library <- rex::rex("Move all library calls to the top of the script.")
  lint_message_require <- rex::rex("Move all require calls to the top of the script.")

  expect_no_lint(
    trim_some("
      library(dplyr)
      require(tidyr)
    "),
    linter
  )

  expect_lint(
    trim_some("
      library(dplyr)
      print(letters)
      require(tidyr)
    "),
    list(lint_message_require, line_number = 3L),
    linter
  )

  expect_lint(
    trim_some("
      library(dplyr)
      print(letters)
      library(dbplyr)
      require(tidyr)
    "),
    list(
      list(lint_message_library, line_number = 3L),
      list(lint_message_require, line_number = 4L)
    ),
    linter
  )
})

test_that("allow_preamble applies as intended", {
  linter_preamble <- library_call_linter(allow_preamble = TRUE)
  linter_no_preamble <- library_call_linter(allow_preamble = FALSE)
  lint_msg <- rex::rex("Move all library calls to the top of the script.")

  lines <- trim_some("
    opts_chunk$set(eval = FALSE)
    library(dplyr)
    library(knitr)

    print(letters)
  ")
  expect_no_lint(lines, linter_preamble)
  expect_lint(lines, list(list(line_number = 2L), list(line_number = 3L)), linter_no_preamble)

  lines <- trim_some("
    opts_chunk$set(eval = FALSE)
    suppressPackageStartupMessages({
      library(dplyr)
      library(knitr)
    })

    print(letters)
  ")
  expect_no_lint(lines, linter_preamble)
  expect_lint(lines, list(list(line_number = 3L), list(line_number = 4L)), linter_no_preamble)

  lines <- trim_some("
    opts_chunk$set(eval = FALSE)
    suppressPackageStartupMessages(library(dplyr))
    library(knitr)

    print(letters)
  ")
  expect_no_lint(lines, linter_preamble)
  expect_lint(lines, list(list(line_number = 2L), list(line_number = 3L)), linter_no_preamble)

  lines <- trim_some("
    opts_chunk$set(eval = FALSE)
    library(dplyr)
    suppressPackageStartupMessages(library(knitr))

    print(letters)
  ")
  expect_no_lint(lines, linter_preamble)
  expect_lint(lines, list(list(line_number = 2L), list(line_number = 3L)), linter_no_preamble)

  # allow S4 operation to precede library() as well, equivalently to other function calls
  lines <- trim_some("
    opts_chunk@set(eval = FALSE)
    library(dplyr)
    suppressPackageStartupMessages(library(knitr))

    print(letters)
  ")
  expect_no_lint(lines, linter_preamble)
  expect_lint(lines, list(list(line_number = 2L), list(line_number = 3L)), linter_no_preamble)

  lines <- trim_some("
    fun()
    library(moreFun)
    oops()
  ")
  expect_no_lint(lines, linter_preamble)
  expect_lint(lines, lint_msg, linter_no_preamble)
})

test_that("skips allowed usages of library()/character.only=TRUE", {
  linter <- library_call_linter()

  expect_no_lint("library(data.table)", linter)
  expect_no_lint("function(pkg) library(pkg, character.only = TRUE)", linter)
  expect_no_lint("function(pkgs) sapply(pkgs, require, character.only = TRUE)", linter)

  expect_no_lint("\\(pkg) library(pkg, character.only = TRUE)", linter)
  expect_no_lint("\\(pkgs) sapply(pkgs, require, character.only = TRUE)", linter)
})

test_that("blocks disallowed usages of strings in library()/require()", {
  linter <- library_call_linter()
  char_only_msg <- rex::rex("Use symbols in library calls", anything, "character.only")
  direct_stub <- "Call library() directly, not vectorized with "

  expect_lint(
    'library("data.table")',
    rex::rex("Use symbols, not strings, in library calls."),
    linter
  )

  expect_lint('library("data.table", character.only = TRUE)', char_only_msg, linter)
  expect_lint('suppressWarnings(library("data.table", character.only = TRUE))', char_only_msg, linter)

  expect_lint(
    "do.call(library, list(data.table))",
    rex::rex(direct_stub, "do.call()"),
    linter
  )
  expect_lint(
    'do.call("library", list(data.table))',
    rex::rex(direct_stub, "do.call()"),
    linter
  )
  expect_lint(
    'lapply("data.table", library, character.only = TRUE)',
    rex::rex(direct_stub, "lapply()"),
    linter
  )
  expect_lint(
    'purr::map("data.table", library, character.only = TRUE)',
    rex::rex(direct_stub, "map()"),
    linter
  )
})

test_that("character.only=TRUE is caught in *apply functions passed as strings", {
  linter <- library_call_linter()
  lib_msg <- rex::rex("Call library() directly, not vectorized with sapply()")
  req_msg <- rex::rex("Call require() directly, not vectorized with sapply()")

  expect_lint("sapply(pkgs, 'library', character.only = TRUE)", lib_msg, linter)
  expect_lint('sapply(pkgs, "library", character.only = TRUE)', lib_msg, linter)
  expect_lint("sapply(pkgs, 'require', character.only = TRUE)", req_msg, linter)
  expect_lint('sapply(pkgs, "require", character.only = TRUE)', req_msg, linter)
})

test_that("character.only=TRUE is caught with multiple-line source", {
  expect_lint(
    trim_some('
      suppressWarnings(library(
        "data.table",
        character.only = TRUE
      ))
    '),
    rex::rex("Use symbols in library calls", anything, "character.only"),
    library_call_linter()
  )
})

test_that("character.only=TRUE is caught inside purrr::walk as well", {
  expect_lint(
    'purr::walk("data.table", library, character.only = TRUE)',
    rex::rex("Call library() directly, not vectorized with walk()"),
    library_call_linter()
  )
})

test_that("multiple lints are generated correctly", {
  expect_lint(
    trim_some('{
      library("dplyr", character.only = TRUE)
      print("not a library call")
      require("gfile")
      sapply(pkg_list, "library", character.only = TRUE)
      purrr::walk(extra_list, require, character.only = TRUE)
    }'),
    list(
      list(message = rex::rex("library calls", anything, "character.only")),
      list(message = rex::rex("Move all require calls to the top of the script.")),
      list(message = "symbols, not strings, in require calls"),
      list(message = rex::rex("library() directly", anything, "vectorized with sapply()")),
      list(message = rex::rex("require() directly", anything, "vectorized with walk()"))
    ),
    library_call_linter()
  )
})

patrick::with_parameters_test_that(
  "library_call_linter skips allowed usages",
  {
    linter <- library_call_linter()

    expect_no_lint(sprintf("%s(x)", call), linter)
    expect_no_lint(sprintf("%s(x, y, z)", call), linter)

    # intervening expression
    expect_no_lint(sprintf("%1$s(x); y; %1$s(z)", call), linter)

    # inline or potentially with gaps don't matter
    expect_no_lint(
      trim_some(glue::glue("
        {call}(x)
        y

        stopifnot(z)
      ")),
      linter
    )

    # only suppressing calls with library()
    expect_no_lint(
      trim_some(glue::glue("
        {call}(x)
        {call}(y)
      ")),
      linter
    )
  },
  .test_name = c("suppressMessages", "suppressPackageStartupMessages"),
  call = c("suppressMessages", "suppressPackageStartupMessages")
)

patrick::with_parameters_test_that(
  "library_call_linter blocks simple disallowed usages",
  {
    linter <- library_call_linter()
    lint_msg <- sprintf("Unify consecutive calls to %s\\(\\)\\.", call)

    # one test of inline usage
    expect_lint(sprintf("%1$s(library(x)); %1$s(library(y))", call), lint_msg, linter)

    expect_lint(
      trim_some(glue::glue("
        {call}(library(x))

        {call}(library(y))
      ")),
      lint_msg,
      linter
    )

    expect_lint(
      trim_some(glue::glue("
        {call}(require(x))
        {call}(require(y))
      ")),
      lint_msg,
      linter
    )

    expect_lint(
      trim_some(glue::glue("
        {call}(library(x))
        # a comment on y
        {call}(library(y))
      ")),
      lint_msg,
      linter
    )
  },
  .test_name = c("suppressMessages", "suppressPackageStartupMessages"),
  call = c("suppressMessages", "suppressPackageStartupMessages")
)

test_that("Namespace differences are detected", {
  linter <- library_call_linter()

  # totally different namespaces
  expect_no_lint("ns::suppressMessages(library(x)); base::suppressMessages(library(y))", linter)

  # one namespaced, one not
  expect_no_lint("ns::suppressMessages(library(x)); suppressMessages(library(y))", linter)
})

test_that("Consecutive calls to different blocked calls is OK", {
  expect_no_lint(
    "suppressPackageStartupMessages(library(x)); suppressMessages(library(y))",
    library_call_linter()
  )
})

test_that("Multiple violations across different calls are caught", {
  linter <- library_call_linter()

  expect_lint(
    trim_some("
      suppressPackageStartupMessages(library(x))
      suppressPackageStartupMessages(library(x))
      suppressMessages(library(x))
      suppressMessages(library(x))
    "),
    list(
      "Unify consecutive calls to suppressPackageStartupMessages",
      "Unify consecutive calls to suppressMessages"
    ),
    linter
  )

  expect_lint(
    trim_some("
      suppressMessages(library(A))
      suppressPackageStartupMessages(library(A))
      suppressMessages(library(A))
      suppressPackageStartupMessages(library(A))
      suppressPackageStartupMessages(library(A))
    "),
    list("Unify consecutive calls to suppressPackageStartupMessages", line_number = 4L),
    linter
  )
})

test_that("returns the correct linting", {
  msg_escape_char <- rex::rex("is an unrecognized escape in character string")
  expect_lint('"\\R"', msg_escape_char)
  expect_lint('"\\A"', msg_escape_char)
  expect_lint('"\\z"', msg_escape_char)
  placeholder_linter <- function(...) NULL
  class(placeholder_linter) <- "linter"
  attr(placeholder_linter, "name") <- "null"
  expect_lint(
    "a <- 1
    function() {
    b",
    rex::rex("unexpected end of input"),
    placeholder_linter
  )

  linter <- equals_na_linter()
  expect_lint("x=", rex::rex("unexpected end of input"), linter)
  expect_lint("x += 1", rex::rex("unexpected '='"), linter)
  expect_lint("{x = }", rex::rex("unexpected '}'"), linter)
  expect_lint("x = ;", rex::rex("unexpected ';'"), linter)

  # no parsing error is expected for the equals-assignment in this code
  expect_no_lint("purrr::partial(list, 1, ... = , 2)", linter)

  # trigger error with base only, and extract it to match against
  #   what comes out from expect_lint.
  get_base_message <- function(e) {
    rex::re_substitutes(
      data = conditionMessage(e),
      pattern = rex::rex(
        list(start, "<text>:", any_digits, ":", any_digits, ": ") %or%
          list(newline, anything, newline, anything, end)
      ),
      replacement = "",
      global = TRUE
    )
  }

  expected_message <- tryCatch(parse(text = "\\"), error = get_base_message)
  expect_lint("\\", rex::rex(expected_message))

  msg_zero_length_var <- rex::rex("attempt to use zero-length variable name")
  expect_lint("``", msg_zero_length_var)
  expect_lint("``()", msg_zero_length_var)
  expect_lint("''()", msg_zero_length_var)
  expect_lint('""()', msg_zero_length_var)
  expect_lint("fun(''=42)", msg_zero_length_var)
  expect_lint('fun(""=42)', msg_zero_length_var)
  expect_lint('fun(a=1,""=42)', msg_zero_length_var)
})

test_that("lengths_linter skips allowed usages", {
  linter <- lengths_linter()

  expect_no_lint("length(x)", linter)
  expect_no_lint("function(x) length(x) + 1L", linter)
  expect_no_lint("vapply(x, fun, integer(length(y)))", linter)
  expect_no_lint("sapply(x, sqrt, simplify = length(x))", linter)

  # TODO(#1570): also throw a lint here, and for map(x, length)
  expect_no_lint("lapply(x, length)", linter)
})

test_that("lengths_linter blocks simple disallowed base usages", {
  linter <- lengths_linter()
  lint_msg <- rex::rex("Use lengths() to find the length of each element in a list.")

  expect_lint("sapply(x, length)", lint_msg, linter)
  expect_lint("sapply(x, FUN = length)", lint_msg, linter)
  expect_lint("sapply(FUN = length, x)", lint_msg, linter)

  expect_lint("vapply(x, length, integer(1L))", lint_msg, linter)
})

test_that("lengths_linter blocks simple disallowed purrr usages", {
  linter <- lengths_linter()
  lint_msg <- rex::rex("Use lengths() to find the length of each element in a list.")

  expect_lint("purrr::map_dbl(x, length)", lint_msg, linter)
  expect_lint("map_dbl(x, .f = length)", lint_msg, linter)
  expect_lint("map_dbl(.f = length, x)", lint_msg, linter)
  expect_lint("map_int(x, length)", lint_msg, linter)
})

test_that("lengths_linter blocks simple disallowed usages with pipes", {
  linter <- lengths_linter()
  lint_msg <- rex::rex("Use lengths() to find the length of each element in a list.")

  expect_lint("x |> sapply(length)", lint_msg, linter)
  expect_lint("x %>% sapply(length)", lint_msg, linter)

  expect_lint("x |> map_int(length)", lint_msg, linter)
  expect_lint("x %>% map_int(length)", lint_msg, linter)
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Use lengths() to find the length of each element in a list.")

  expect_lint(
    trim_some("{
      sapply(x, length)
      map_int(x, length)
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    lengths_linter()
  )
})

test_that("download_file_linter skips allowed usages", {
  linter <- download_file_linter()

  expect_no_lint("download.file(x, mode = 'wb')", linter)
  expect_no_lint("download.file(x, mode = 'ab')", linter)

  # 'w' or 'a' but passed to different arguments
  expect_no_lint("download.file(x, destfile = 'w', mode = 'wb')", linter)
  expect_no_lint("download.file(x, mode = 'wb', method = 'internal', quiet = TRUE, 'w')", linter)

  # if mode = "wget" or "curl", mode is ignored so we don't lint on the default
  expect_no_lint("download.file(x, method = 'curl')", linter)
  expect_no_lint("download.file(x, method = 'wget')", linter)

  expect_no_lint("download.file(x, method = 'internal', mode = 'wb')", linter)
  expect_no_lint("download.file(x, method = method, mode = 'wb')", linter)
})

test_that("download_file_linter blocks simple disallowed usages", {
  linter <- download_file_linter()
  lint_message <- rex::rex("download.file() should use mode =")

  # Case 1: implicit default (mode = "w")
  expect_lint("download.file(x)", lint_message, linter)
  expect_lint("download.file(x, method = 'internal')", lint_message, linter)

  # Case 2: non-portable mode specified by name
  expect_lint("download.file(x, mode = 'w')", lint_message, linter)
  expect_lint("download.file(x, mode = 'a')", lint_message, linter)
  expect_lint("download.file(x, method = method, mode = 'w')", lint_message, linter)

  # 'wb' passed to different argument
  expect_lint("download.file(x, mode = 'w', method = 'internal', quiet = TRUE, 'wb')", lint_message, linter)
  expect_lint("download.file(cacheOK = TRUE, destfile, method, quiet, x = 'wb')", lint_message, linter)

  # mode explicitly specified with a method that ignores it
  lint_message <- rex::rex("mode argument value is ignored")
  expect_lint("download.file(x, method = 'wget', mode = 'w')", lint_message, linter)
  expect_lint("download.file(x, method = 'curl', mode = 'wb')", lint_message, linter)
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      download.file(x, mode = 'w')
      download.file(y, mode = 'a')
      download.file(z, mode = 'wb', method = 'curl')
    }"),
    list(
      list(
        rex::rex("download.file() should use mode = 'wb'"),
        line_number = 2L
      ),
      list(
        rex::rex("download.file() should use mode = 'ab'"),
        line_number = 3L
      ),
      list(
        rex::rex("mode argument value is ignored"),
        line_number = 4L
      )
    ),
    download_file_linter()
  )
})

# fuzzer disable: comment_injection
test_that("pipe-continuation correctly handles stand-alone expressions", {
  linter <- pipe_continuation_linter()
  lint_msg <- rex::rex("Put a space before `%>%` and a new line after it,")

  # Expressions without pipes are ignored
  expect_no_lint("blah", linter)

  # Pipe expressions on a single line are ignored
  expect_no_lint("foo %>% bar() %>% baz()", linter)

  # Pipe expressions spanning multiple lines with each expression on a line are ignored
  expect_no_lint(
    trim_some("
      foo %>%
        bar() %>%
        baz()
    "),
    linter
  )

  # Pipe expressions with multiple expression on a line are linted
  expect_lint(
    trim_some("
      foo %>% bar() %>%
        baz()
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo %>% bar() %>% baz() %>%
       qux()
    "),
    lint_msg,
    linter
  )
})

test_that("pipe-continuation linter correctly handles nesting", {
  linter <- pipe_continuation_linter()
  lint_msg <- rex::rex("Put a space before `%>%` and a new line after it,")

  expect_lint(
    trim_some("
      my_fun <- function(){
        a %>% b() %>%
          c()
      }
    "),
    list(list(message = lint_msg, line_number = 2L)),
    linter
  )

  expect_lint(
    trim_some("
      my_fun <- function(){
        a %>%
          b() %>% c()
      }
    "),
    list(list(message = lint_msg, line_number = 3L)),
    linter
  )

  # but no lints here
  expect_no_lint(
    trim_some("
      1:4 %>% {
       (.) %>% sum()
      }
    "),
    linter
  )
})

test_that("pipe-continuation linter handles native pipe", {
  linter <- pipe_continuation_linter()
  lint_msg_native <- rex::rex("Put a space before `|>` and a new line after it,")
  lint_msg_magrittr <- rex::rex("Put a space before `%>%` and a new line after it,")

  expect_no_lint("foo |> bar() |> baz()", linter)
  expect_no_lint(
    trim_some("
      foo |>
        bar() |>
        baz()
    "),
    linter
  )
  expect_lint(
    trim_some("
      foo |> bar() |>
        baz()
    "),
    lint_msg_native,
    linter
  )
  # mixing pipes
  expect_lint(
    trim_some("
      foo %>% bar() |>
        baz()
    "),
    lint_msg_native,
    linter
  )
  expect_lint(
    trim_some("
      foo |> bar() %>%
        baz()
    "),
    lint_msg_magrittr,
    linter
  )
  expect_lint(
    trim_some("
      list(
        foo |> bar() |>
          baz(),
        foo %>% bar() %>%
          baz()
      )
    "),
    list(
      lint_msg_native,
      lint_msg_magrittr
    ),
    linter
  )
})

local({
  linter <- pipe_continuation_linter()
  .cases <- tibble::tribble(
    ~code_string,            ~.test_name,
    trim_some("
      my_fun <- function() {
        a %>% b()
      }
    "),                      "two on one line",
    trim_some("
      my_fun <- function() {
        a %>% b() %>% c()
      }
    "),                      "three on one line",
    trim_some("
      with(
        diamonds,
        x %>% head(10) %>% tail(5)
      )
    "),                      "three inside with()",
    trim_some("
      test_that('blah', {
        test_data <- diamonds %>% head(10) %>% tail(5)
      })
    "),                      "three inside test_that()",
    trim_some("
      {
        x <- a %>% b %>% c
        y <- c %>% b %>% a
      }
    "),                      "two different single-line pipelines",
    trim_some("
      my_fun <- function() {
        a %>%
          b() %>%
          c()
      }
    "),                      "at most one pipe-character per line"
  )
  patrick::with_parameters_test_that(
    "valid nesting is handled",
    # nolint next: unnecessary_nesting_linter. TODO(#2334): Remove this nolint.
    {
      expect_no_lint(code_string, linter)
    },
    .cases = .cases
  )
})

local({
  linter <- pipe_continuation_linter()
  pipes <- pipes()
  cases <- expand.grid(pipe1 = pipes, pipe2 = pipes, stringsAsFactors = FALSE)
  cases <- within(cases, {
    .test_name <- sprintf("(%s, %s)", pipe1, pipe2)
  })
  patrick::with_parameters_test_that(
    "Various pipes are linted correctly",
    expect_lint(
      sprintf("a %s b() %s\n  c()", pipe1, pipe2),
      rex::rex(sprintf("Put a space before `%s` and a new line after it", pipe2)),
      linter
    ),
    .cases = cases
  )
})
# fuzzer enable: comment_injection

test_that("empty_assignment_linter skips valid usage", {
  linter <- empty_assignment_linter()

  expect_no_lint("x <- { 3 + 4 }", linter)
  expect_no_lint("x <- if (x > 1) { 3 + 4 }", linter)

  # also triggers assignment_linter
  expect_no_lint("x = { 3 + 4 }", linter)
})

test_that("empty_assignment_linter blocks disallowed usages", {
  linter <- empty_assignment_linter()
  lint_msg <- rex::rex("Assign NULL explicitly or, whenever possible, allocate the empty object")

  expect_lint("xrep <- {}", lint_msg, linter)

  # assignment with equal works as well, and white space doesn't matter
  expect_lint("x = {   }", lint_msg, linter)

  # ditto right assignments
  expect_lint("{} -> x", lint_msg, linter)
  expect_lint("{} ->> x", lint_msg, linter)

  # ditto data.table-style walrus assignments
  expect_lint("x[, col := {}]", lint_msg, linter)

  # newlines also don't matter
  expect_lint("x <- {\n}", lint_msg, linter)
  expect_lint("x <- { # comment\n}", lint_msg, linter)

  # LHS of assignment doesn't matter
  expect_lint("env$obj <- {}", lint_msg, linter)
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Assign NULL explicitly or, whenever possible, allocate the empty object")

  expect_lint(
    trim_some("{
      x <- {}
      y = {}
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    empty_assignment_linter()
  )
})

test_that("unused_import_linter lints as expected", {
  linter <- unused_import_linter()

  expect_no_lint("library(dplyr)\ntibble(a = 1)", linter)
  # SYMBOL_FUNCTION_CALL usage is detected
  expect_no_lint("library(tidyverse)\ntibble(a = 1)", linter)
  # SYMBOL usage is detected
  expect_no_lint("library(dplyr)\ndo.call(tibble, args = list(a = 1))", linter)
  # SPECIAL usage is detected
  expect_no_lint( # nofuzz: pipe
    trim_some("
      library(magrittr)
      1:3 %>% mean()
    "),
    linter
  )
  # dataset is detected
  expect_no_lint("library(dplyr)\nstarwars", linter)
  expect_no_lint("library(datasets)\nstate.center", linter)

  # Missing packages are ignored
  expect_no_lint("library(not.a.package)\ntibble(a = 1)", linter)
  # SYMBOL calls with character.only = TRUE are ignored, even if the argument is a package name
  expect_no_lint("library(dplyr, character.only = TRUE)\n1 + 1", linter)

  lint_msg <- rex::rex("Package 'dplyr' is attached but never used")
  msg_ns <- rex::rex("Don't attach package 'dplyr', which is only used by namespace.")

  expect_lint("library(dplyr)\n1 + 1", lint_msg, linter)
  expect_lint("require(dplyr)\n1 + 1", lint_msg, linter)
  expect_lint("library('dplyr')\n1 + 1", lint_msg, linter)
  expect_lint("library('dplyr', character.only = TRUE)\n1 + 1", lint_msg, linter)
  # ignore namespaced usages by default, but provide custom lint message
  expect_lint("library(dplyr)\ndplyr::tibble(a = 1)", msg_ns, linter)
  expect_no_lint("library(dplyr)\ndplyr::tibble(a = 1)", unused_import_linter(allow_ns_usage = TRUE))

  # ignore packages in except_packages
  expect_no_lint("library(data.table)\n1 + 1", linter)
  expect_no_lint("library(dplyr)\n1 + 1", unused_import_linter(except_packages = "dplyr"))
})

test_that("unused_import_linter handles message vectorization", {
  skip_if_not_installed("crayon")
  expect_lint(
    trim_some("
      library(crayon)
      library(xmlparsedata)
      xmlparsedata::xml_parse_data(parse(text = 'a'))
    "),
    list(
      list(rex::rex("Package 'crayon' is attached but never used."), line_number = 1L),
      list(rex::rex("Don't attach package 'xmlparsedata', which is only used by namespace"), line_number = 2L)
    ),
    unused_import_linter()
  )
})

test_that("unused_import_linter lints packages with exports like pkg::pkg", {
  # glue::glue is an export, so don't get thrown off by the 'glue' symbol in library()
  expect_lint(
    trim_some("
      library(glue)
      1 + 1
    "),
    rex::rex("Package 'glue' is attached but never used."),
    unused_import_linter()
  )
})

test_that("glue usages are seen", {
  lint_msg <- rex::rex("Package 'xmlparsedata' is attached but never used.")

  lines <- trim_some("
    library(glue)
    library(xmlparsedata)

    glue('{ xml_parse_data() }')
  ")
  expect_no_lint(lines, unused_import_linter())
  expect_lint(lines, lint_msg, unused_import_linter(interpret_glue = FALSE))
})

test_that("for_loop_index_linter skips allowed usages", {
  linter <- for_loop_index_linter()

  expect_no_lint("for (xi in x) {}", linter)

  # this is OK, so not every symbol is problematic
  expect_no_lint("for (col in DF$col) {}", linter)
  expect_no_lint("for (col in S4@col) {}", linter)
  expect_no_lint("for (col in DT[, col]) {}", linter)

  # make sure symbol check is scoped
  expect_no_lint(
    trim_some("
      {
        for (i in 1:10) {
          42L
        }
        i <- 7L
      }
    "),
    linter
  )
})

test_that("for_loop_index_linter blocks simple disallowed usages", {
  linter <- for_loop_index_linter()
  lint_msg <- "Don't re-use any sequence symbols as the index symbol in a for loop"

  expect_lint("for (x in x) {}", lint_msg, linter)
  # these also overwrite a variable in calling scope
  expect_lint("for (x in foo(x)) {}", lint_msg, linter)
  # arbitrary nesting
  expect_lint("for (x in foo(bar(y, baz(2, x)))) {}", lint_msg, linter)
})

test_that("lints vectorize", {
  lint_msg <- "Don't re-use any sequence symbols as the index symbol in a for loop"

  expect_lint(
    trim_some("{
      for (x in x) {}
      for (y in y) {}
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    for_loop_index_linter()
  )
})

patrick::with_parameters_test_that(
  "sprintf_linter skips allowed usages",
  {
    linter <- sprintf_linter()

    # NB: using paste0, not sprintf, to avoid escaping '%d' in sprint fmt=
    expect_no_lint(paste0(call_name, "('hello %d', 1)"), linter)
    expect_no_lint(paste0(call_name, "('hello %d', x)"), linter)
    expect_no_lint(paste0(call_name, "('hello %d', x + 1)"), linter)
    expect_no_lint(paste0(call_name, "('hello %d', f(x))"), linter)
    expect_no_lint(paste0(call_name, "('hello %1$s %1$s', x)"), linter)
    expect_no_lint(paste0(call_name, "('hello %1$s %1$s %2$d', x, y)"), linter)
    expect_no_lint(paste0(call_name, "('hello %1$s %1$s %2$d %3$s', x, y, 1.5)"), linter)
  },
  .test_name = c("sprintf", "gettextf"),
  call_name = c("sprintf", "gettextf")
)

patrick::with_parameters_test_that(
  "sprintf_linter blocks disallowed usages",
  {
    linter <- sprintf_linter()
    unused_arg_msg <- "one argument not used by format"

    expect_lint(paste0(call_name, "('hello', 1)"), "constant", linter)

    expect_lint(paste0(call_name, "('hello')"), "constant", linter)
    expect_lint(paste0(call_name, "('100%% automated')"), "constant", linter)
    expect_lint(paste0(call_name, "('100%%%% automated')"), "constant", linter)
    expect_lint(paste0(call_name, "('100%%%s')"), "too few", linter)
    expect_lint(paste0(call_name, "('100%%%%s', x)"), "constant", linter)

    expect_lint(
      paste0(call_name, "('hello %d', 'a')"),
      rex::rex("invalid format '%d'; use format %s for character objects"),
      linter
    )

    expect_lint(
      paste0(call_name, "('hello %d', 1.5)"),
      rex::rex("invalid format '%d'; use format %f, %e, %g or %a for numeric objects"),
      linter
    )

    expect_lint(
      paste0(call_name, "('hello %d',)"),
      rex::rex("argument is missing, with no default"),
      linter
    )

    expect_lint(paste0(call_name, "('hello %1$s %s', 'a', 'b')"), unused_arg_msg, linter)
    expect_lint(paste0(call_name, "('hello %1$s %1$s', x, y)"), unused_arg_msg, linter)

    expect_lint(
      paste0(call_name, "('hello %1$s %1$s %3$d', x, y)"),
      rex::rex("reference to non-existent argument 3"),
      linter
    )

    expect_lint(
      paste0(call_name, "('hello %1$s %1$s %2$d %3$d', x, y, 1.5)"),
      rex::rex("invalid format '%d'; use format %f, %e, %g or %a for numeric objects"),
      linter
    )
  },
  .test_name = c("sprintf", "gettextf"),
  call_name = c("sprintf", "gettextf")
)

test_that("edge cases are detected correctly", {
  linter <- sprintf_linter()

  # works with multi-line sprintf and comments
  expect_no_lint(
    trim_some("
      sprintf(
        'test fmt %s', # this is a comment
        2
      )
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      'test fmt %s' |>   # this is a pipe comment
        sprintf(         # this is an opening comment
          2              # this is a mid-call comment
        )
    "),
    linter
  )

  expect_lint(
    trim_some("
      'test fmt' |>   # this is a pipe comment
        sprintf()
    "),
    "constant",
    linter
  )

  # dots
  expect_no_lint("sprintf('%d %d, %d', id, ...)", linter)

  # TODO(#1265) extend ... detection to at least test for too many arguments.

  # named argument fmt
  expect_no_lint("sprintf(x, fmt = 'hello %1$s %1$s')", linter)

  expect_lint(
    "sprintf(x, fmt = 'hello %1$s %1$s %3$d', y)",
    list(message = rex::rex("reference to non-existent argument 3")),
    linter
  )

  # #2131: xml2lang stripped necessary whitespace
  expect_no_lint("sprintf('%s', if (A) '' else y)", linter)

  expect_no_lint("sprintf('100%%%s', x)", linter)
})

local({
  linter <- sprintf_linter()
  unused_fmt_msg <- "too few arguments"
  unused_arg_msg <- "one argument not used by format"
  pipes <- pipes(exclude = "%$%")
  patrick::with_parameters_test_that(
    "piping into sprintf works",
    {
      expect_no_lint(paste("x", pipe, "sprintf(fmt = '%s')"), linter)
      # no fmt= specified -> this is just 'sprintf("%s", "%s%s")', which won't lint
      expect_no_lint(paste('"%s"', pipe, 'sprintf("%s%s")'), linter)
      expect_lint(paste("x", pipe, "sprintf(fmt = '%s%s')"), unused_fmt_msg, linter)

      # Cannot evaluate statically --> skip
      expect_no_lint(paste("x", pipe, 'sprintf("a")'), linter)
      # Nested pipes
      expect_lint(
        paste("'%%sb'", pipe, "sprintf('%s')", pipe, "sprintf('a')"),
        list(column_number = nchar(paste("'%%sb'", pipe, "x")), message = "constant"),
        linter
      )
      expect_lint(
        paste("x", pipe, 'sprintf(fmt = "%s")', pipe, 'sprintf(fmt = "%s%s")'),
        list(column_number = nchar(paste("x", pipe, 'sprintf(fmt = "%s")', pipe, "x")), message = unused_fmt_msg),
        linter
      )
      expect_lint(
        paste("x", pipe, 'sprintf(fmt = "%s%s")', pipe, 'sprintf(fmt = "%s")'),
        list(column_number = nchar(paste("x", pipe, "x")), message = unused_fmt_msg),
        linter
      )
    },
    pipe = pipes,
    .test_name = names(pipes)
  )
})

test_that("pipe logic survives adversarial comments", {
  linter <- sprintf_linter()

  expect_no_lint(
    trim_some("
      x |> # comment
      sprintf(fmt = '%s')
    "),
    linter
  )

  expect_no_lint(
    trim_some('
      "%s" %>% # comment
      sprintf("%s%s")
    '),
    linter
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      sprintf('%s', a, b)
      sprintf('%s%s', a)
    }"),
    list(
      list("one argument not used by format", line_number = 2L),
      list("too few arguments", line_number = 3L)
    ),
    sprintf_linter()
  )
})

test_that("other : expressions are fine", {
  linter <- seq_linter()
  expect_no_lint("1:10", linter)
  expect_no_lint("2:length(x)", linter)
  expect_no_lint("1:(length(x) || 1)", linter)
})

test_that("seq_len(...) or seq_along(...) expressions are fine", {
  linter <- seq_linter()

  expect_no_lint("seq_len(x)", linter)
  expect_no_lint("seq_along(x)", linter)

  expect_no_lint("seq(2, length(x))", linter)
  expect_no_lint("seq(length(x), 2)", linter)
})

test_that("finds seq(...) expressions", {
  linter <- seq_linter()
  lint_msg <- function(want, got) rex::rex("Use ", want, " instead of ", got)

  expect_lint(
    "seq(length(x))",
    lint_msg("seq_along(...)", "seq(length(...))"),
    linter
  )

  expect_lint(
    "seq(nrow(x))",
    lint_msg("seq_len(nrow(...))", "seq(nrow(...))"),
    linter
  )

  expect_lint(
    "rev(seq(length(x)))",
    lint_msg("seq_along(...)", "seq(length(...))"),
    linter
  )

  expect_lint(
    "rev(seq(nrow(x)))",
    lint_msg("seq_len(nrow(...))", "seq(nrow(...))"),
    linter
  )
})

test_that("finds 1:length(...) expressions", {
  linter <- seq_linter()
  lint_msg <- function(want, got) rex::rex("Use ", want, " instead of ", got)

  expect_lint(
    "1:length(x)",
    lint_msg("seq_along(...)", "1:length(...)"),
    linter
  )

  expect_lint(
    "1:nrow(x)",
    lint_msg("seq_len(nrow(...))", "1:nrow(...)"),
    linter
  )

  expect_lint(
    "1:ncol(x)",
    lint_msg("seq_len(ncol(...))", "1:ncol(...)"),
    linter
  )

  expect_lint(
    "1:NROW(x)",
    lint_msg("seq_len(NROW(...))", "1:NROW(...)"),
    linter
  )

  expect_lint(
    "1:NCOL(x)",
    lint_msg("seq_len(NCOL(...))", "1:NCOL(...)"),
    linter
  )

  expect_lint(
    "1:dim(x)[1L]",
    lint_msg("seq_len(dim(...)[1L])", "1:dim(...)[1L]"),
    linter
  )

  expect_lint(
    "1L:dim(x)[[1]]",
    rex::rex("Use seq_len", anything, "dim(...)"),
    linter
  )

  expect_lint(
    "mutate(x, .id = 1:n())",
    lint_msg("seq_len(n())", "1:n(),"),
    linter
  )

  expect_lint(
    trim_some("
      mutate(x, .id = 1:n( # comment
      ))
    "),
    lint_msg("seq_len(n())", "1:n(),"),
    linter
  )

  expect_lint(
    "x[, .id := 1:.N]",
    lint_msg("seq_len(.N)", "1:.N,"),
    linter
  )
})

test_that("1L is also bad", {
  expect_lint(
    "1L:length(x)",
    rex::rex("seq_along", anything, "1L:length(...)"),
    seq_linter()
  )
})

test_that("reverse seq is ok", {
  linter <- seq_linter()
  expect_no_lint("rev(seq_along(x))", linter)
  expect_no_lint("rev(seq_len(nrow(x)))", linter)

  expect_lint(
    "length(x):1",
    rex::rex("rev(seq_along(...))", anything, "length(...):1"),
    linter
  )
})

test_that("finds potential sequence() replacements", {
  linter <- seq_linter()
  lint_msg <- rex::rex("Use sequence()")

  expect_lint("unlist(lapply(x, seq_len))", lint_msg, linter)

  expect_lint("unlist(lapply(x, seq))", lint_msg, linter)

  # Even for prefixed purrr:: calls
  expect_lint("unlist(purrr::map(x, seq_len))", lint_msg, linter)
})

test_that("sequence() is not recommended for complex seq() calls", {
  linter <- seq_linter()

  expect_no_lint("unlist(lapply(x, seq, from = 2))", linter)
})

test_that("Message vectorization works for multiple lints", {
  linter <- seq_linter()

  expect_lint(
    trim_some("{
      1:length(x)
      1:nrow(y)
    }"),
    list(
      list(rex::rex("seq_along(...)", anything, "1:length(...)"), line_number = 2L),
      list(rex::rex("seq_len(nrow(...))", anything, "1:nrow(...)"), line_number = 3L)
    ),
    linter
  )

  expect_lint(
    trim_some("{
      seq(length(x))
      1:nrow(y)
    }"),
    list(
      list(rex::rex("seq_along(...)", anything, "seq(length(...))"), line_number = 2L),
      list(rex::rex("seq_len(nrow(...))", anything, "1:nrow(...)"), line_number = 3L)
    ),
    linter
  )

  expect_lint(
    trim_some("{
      seq(length(x))
      seq(nrow(y))
    }"),
    list(
      list(rex::rex("seq_along(...)", anything, "seq(length(...))"), line_number = 2L),
      list(rex::rex("seq_len(nrow(...))", anything, "seq(nrow(...))"), line_number = 3L)
    ),
    linter
  )

  expect_lint(
    trim_some("{
      1:NROW(x)
      seq(NCOL(y))
    }"),
    list(
      list(rex::rex("seq_len(NROW(...))", anything, "1:NROW(...)"), line_number = 2L),
      list(rex::rex("seq_len(NCOL(...))", anything, "seq(NCOL(...))"), line_number = 3L)
    ),
    linter
  )

  expect_lint(
    trim_some("{
      1:NROW(x)
      unlist(lapply(y, seq_len))
    }"),
    list(
      list(rex::rex("seq_len(NROW(...))", anything, "1:NROW(...)"), line_number = 2L),
      list(rex::rex("sequence()"), line_number = 3L)
    ),
    linter
  )
})

test_that("Message recommends rev() correctly", {
  linter <- seq_linter()

  expect_lint(".N:1", rex::rex("Use rev(seq_len(.N))"), linter)
  expect_lint("n():1", rex::rex("Use rev(seq_len(n()))"), linter)
  expect_lint("nrow(x):1", rex::rex("Use rev(seq_len(nrow(...)))"), linter)
  expect_lint("length(x):1", rex::rex("Use rev(seq_along(...))"), linter)
})

test_that("seq_len(length(x)) should be seq_along(x)", {
  linter <- seq_linter()

  expect_no_lint("seq_len(length(x) - 1)", linter)
  expect_no_lint("seq_len(2*length(x))", linter)
  expect_no_lint("seq_len(foo(length(x)))", linter)
  expect_lint("seq_len(length(x))", "seq_along", linter)
  expect_lint("seq_len(length(foo(x)))", "seq_along", linter)
})

test_that("nrow_subset_linter skips allowed usage", {
  linter <- nrow_subset_linter()

  expect_no_lint("nrow(foo(subset(x, y == z)))", linter)
  expect_no_lint("with(x, sum(y == z))", linter)
})

test_that("nrow_subset_linter blocks subset() cases", {
  expect_lint(
    "nrow(subset(x, y == z))",
    rex::rex("Use arithmetic to count the number of rows satisfying a condition"),
    nrow_subset_linter()
  )
})

test_that("nrow_subset_linter blocks filter() cases", {
  expect_lint(
    "nrow(filter(x, y == z))",
    rex::rex("Use arithmetic to count the number of rows satisfying a condition"),
    nrow_subset_linter()
  )
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Use arithmetic to count the number of rows satisfying a condition")

  expect_lint(
    trim_some("{
      nrow(subset(x, y == z))
      subset(x) %>% transform(m = 2)
      nrow(subset(a, b == c))
      x %>% filter(y == z) %>% nrow()
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 4L),
      list(lint_msg, line_number = 5L)
    ),
    nrow_subset_linter()
  )
})

test_that("linter is pipeline-aware", {
  linter <- nrow_subset_linter()
  lint_msg <- "Use arithmetic to count the number of rows satisfying a condition"

  expect_lint("x %>% subset(y == z) %>% nrow()", lint_msg, linter)
  expect_lint("x |> subset(y == z) |> nrow()", lint_msg, linter)
  expect_lint("filter(x, y == z) %>% nrow()", lint_msg, linter)
})

# fuzzer disable: assignment
test_that("implicit_assignment_linter skips allowed usages", {
  linter <- implicit_assignment_linter()

  expect_no_lint("x <- 1L", linter)
  expect_no_lint("1L -> x", linter)
  expect_no_lint("x <<- 1L", linter)
  expect_no_lint("1L ->> x", linter)
  expect_no_lint("y <- if (is.null(x)) z else x", linter)
  expect_no_lint("for (x in 1:10) x <- x + 1", linter)

  expect_no_lint("abc <- mean(1:4)", linter)
  expect_no_lint("mean(1:4) -> abc", linter)

  expect_no_lint(
    trim_some("
    x <- 1:4
    mean(x)"),
    linter
  )

  expect_no_lint(
    trim_some("
    x <- 1L
    if (x) TRUE"),
    linter
  )

  expect_no_lint(
    trim_some("
    0L -> abc
    while (abc) {
      FALSE
    }"),
    linter
  )

  expect_no_lint(
    trim_some("
    if (x > 20L) {
      x <- x / 2.0
    }"),
    linter
  )

  expect_no_lint(
    trim_some("
    i <- 1
    while (i < 6L) {
      print(i)
      i <- i + 1
    }"),
    linter
  )

  expect_no_lint(
    trim_some("
    foo <- function(x) {
      x <- x + 1
      return(x)
    }"),
    linter
  )

  expect_no_lint(
    trim_some("
    f <- function() {
      p <- g()
      p <- if (is.null(p)) x else p
    }"),
    linter
  )

  expect_no_lint(
    trim_some("
      map(
        .x = 1:4,
        .f = ~ {
          x <- .x + 1
          x
        }
      )"),
    linter
  )

  expect_no_lint(
    trim_some("
      lapply(1:4, function(x) {
        x <- x + 1
        x
      })"),
    linter
  )

  expect_no_lint(
    trim_some("
      map(1:4, \\(x) {
        x <- x + 1
        x
      })"),
    linter
  )
})

test_that("implicit_assignment_linter respects except argument", {
  expect_no_lint(
    "local({ a <- 1L })",
    implicit_assignment_linter(except = NULL)
  )

  expect_no_lint(
    "local({ a <- 1L })",
    implicit_assignment_linter(except = character(0L))
  )

  expect_lint(
    "local(a <- 1L)",
    rex::rex("Avoid implicit assignments in function calls."),
    implicit_assignment_linter(except = character(0L))
  )

  expect_lint(
    "local(a <- 1L)",
    rex::rex("Avoid implicit assignments in function calls."),
    implicit_assignment_linter(except = NULL)
  )

  expect_no_lint(
    "local(a <- 1L)",
    implicit_assignment_linter(except = "local")
  )

  expect_error(
    implicit_assignment_linter(except = 1L),
    "`except` must be a character",
    fixed = TRUE
  )
})

test_that("implicit_assignment_linter skips allowed usages with braces", {
  linter <- implicit_assignment_linter(except = character(0L))

  expect_no_lint(
    trim_some("
    foo({
      a <- 1L
    })
    "),
    linter
  )
  expect_no_lint(
    trim_some("
    output <- capture.output({
      x <- f()
    })
    "),
    linter
  )
  expect_no_lint(
    trim_some("
    quote({
      a <- 1L
    })
    "),
    linter
  )
  expect_no_lint(
    trim_some("
    bquote({
      a <- 1L
    })
    "),
    linter
  )
  expect_no_lint(
    trim_some("
    expression({
      a <- 1L
    })
    "),
    linter
  )
  expect_no_lint(
    trim_some("
    local({
      a <- 1L
    })
    "),
    linter
  )
})

test_that("implicit_assignment_linter makes exceptions for functions that capture side-effects", {
  linter <- implicit_assignment_linter()

  expect_no_lint(
    trim_some("
    test_that('my test', {
      a <- 1L
      expect_equal(a, 1L)
    })"),
    linter
  )

  # rlang
  expect_no_lint("expr(a <- 1L)", linter)
  expect_no_lint("quo(a <- 1L)", linter)
  expect_no_lint("quos(a <- 1L)", linter)
})

test_that("implicit_assignment_linter blocks disallowed usages in simple conditional statements", {
  lint_message <- rex::rex("Avoid implicit assignments in function calls.")
  linter <- implicit_assignment_linter()

  expect_lint("if (x <- 1L) TRUE", lint_message, linter)
  expect_lint("if (1L -> x) TRUE", lint_message, linter)
  expect_lint("if (x <<- 1L) TRUE", lint_message, linter)
  expect_lint("if (1L ->> x) TRUE", lint_message, linter)
  expect_lint("while (x <- 0L) FALSE", lint_message, linter)
  expect_lint("while (0L -> x) FALSE", lint_message, linter)
  expect_lint("for (x in y <- 1:10) print(x)", lint_message, linter)
  expect_lint("for (x in 1:10 -> y) print(x)", lint_message, linter)

  # adversarial commenting
  expect_lint(
    trim_some("
      while # comment
      (x <- 0L) FALSE

      while ( # comment
      x <- 0L) FALSE
    "),
    list(
      list(lint_message, line_number = 2L),
      list(lint_message, line_number = 5L)
    ),
    linter
  )
})

test_that("implicit_assignment_linter blocks disallowed usages in nested conditional statements", {
  lint_message <- rex::rex("Avoid implicit assignments in function calls.")
  linter <- implicit_assignment_linter()

  expect_lint(
    trim_some("
    while (x <- 1L) {
      if (0L -> y) FALSE
    }"),
    list(
      list(message = lint_message, line_number = 1L, column_number = 8L),
      list(message = lint_message, line_number = 2L, column_number = 7L)
    ),
    linter
  )
  expect_lint(
    trim_some("
    for (x in y <- 1:10) {
      if (0L -> y) print(x)
    }"),
    list(
      list(message = lint_message, line_number = 1L, column_number = 11L),
      list(message = lint_message, line_number = 2L, column_number = 7L)
    ),
    linter
  )
})

test_that("implicit_assignment_linter blocks disallowed usages in function calls", {
  lint_message <- rex::rex("Avoid implicit assignments in function calls.")
  linter <- implicit_assignment_linter()

  expect_lint("mean(x <- 1:4)", lint_message, linter)
  expect_lint(
    "mean(x <- (y <- 1:3) + 1L)",
    list(list(column_number = 6L), list(column_number = 12L)),
    linter
  )
  expect_lint("y <- median(x <- 1:4)", lint_message, linter)
  expect_lint("lapply(x, function(x) return(x <- x + 1))", lint_message, linter)
  expect_lint("map(x, function(x) return(x <- x + 1))", lint_message, linter)

  expect_lint("expect_warning(out <- f(-1))", lint_message, linter)
  expect_lint("expect_message(out <- f(-1))", lint_message, linter)
  expect_lint("expect_error(out <- f(-1))", lint_message, linter)
  expect_lint("expect_condition(out <- f(-1))", lint_message, linter)

  expect_lint(
    trim_some("
    foo <- function(x) {
      return(x <- x + 1)
    }"),
    lint_message,
    linter
  )
  expect_lint(
    trim_some("
    foo <- function(x) {
      if (x <- 1L) x <- 2L
      return(x <- x + 1)
    }"),
    list(
      list(message = lint_message, line_number = 2L, column_number = 7L),
      list(message = lint_message, line_number = 3L, column_number = 10L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      map(
        .x = 1:4,
        .f = ~ .x + 1 -> x
      )"),
    lint_message,
    linter
  )

  expect_lint(
    trim_some("
      map(
        .x = 1:4,
        .f = ~ (x <- .x + 1)
      )"),
    lint_message,
    linter
  )


  expect_lint(
    "foo(a <- 1, b <- 2, c <- 3)",
    list(list(column_number = 5L), list(column_number = 13L), list(column_number = 21L)),
    linter
  )
})

test_that("implicit_assignment_linter works as expected with pipes and walrus operator", {
  linter <- implicit_assignment_linter()

  expect_no_lint("data %>% mutate(a := b)", linter)
  expect_no_lint("dt %>% .[, z := x + y]", linter)
  expect_no_lint("data %<>% mutate(a := b)", linter)
  expect_no_lint("data |> mutate(a := b)", linter)

  expect_no_lint("DT[i, x := i]", linter)
})

test_that("parenthetical assignments are caught", {
  expect_lint(
    "if (A && (B <- foo())) { }",
    rex::rex("Avoid implicit assignments in function calls."),
    implicit_assignment_linter()
  )
})

test_that("allow_lazy lets lazy assignments through", {
  linter <- implicit_assignment_linter(allow_lazy = TRUE)
  lint_message <- rex::rex("Avoid implicit assignments in function calls.")

  expect_no_lint("A && (B <- foo(A))", linter)
  # || also admits laziness
  expect_no_lint("A || (B <- foo(A))", linter)
  # & and |, however, do not
  expect_lint("A & (B <- foo(A))", lint_message, linter)
  expect_lint("A | (B <- foo(A))", lint_message, linter)
  expect_no_lint("A && foo(bar(idx <- baz()))", linter)
  # LHS _is_ linted
  expect_lint("(A <- foo()) && B", lint_message, linter)
  # however we skip on _any_ RHS (even if it's later an LHS)
  # test on all &&/|| combinations to stress test operator precedence
  expect_no_lint("A && (B <- foo(A)) && C", linter)
  expect_no_lint("A && (B <- foo(A)) || C", linter)
  expect_no_lint("A || (B <- foo(A)) && C", linter)
  expect_no_lint("A || (B <- foo(A)) || C", linter)
  # &&/|| elsewhere in the tree don't matter
  expect_lint(
    trim_some("
      A && B
      foo(C <- bar())
    "),
    lint_message,
    linter
  )
})

test_that("allow_scoped skips scoped assignments", {
  linter <- implicit_assignment_linter(allow_scoped = TRUE)
  lint_message <- rex::rex("Avoid implicit assignments in function calls.")

  expect_no_lint(
    trim_some("
      if (any(idx <- x < 0)) {
        stop('negative elements: ', toString(which(idx)))
      }
    "),
    linter
  )
  expect_lint(
    trim_some("
      if (any(idx <- x < 0)) {
        stop('negative elements: ', toString(which(idx)))
      }
      print(idx)
    "),
    lint_message,
    linter
  )

  expect_no_lint("if (a <- 1) print(a)", linter)

  # only applies to the branch condition itself -- within the branch, still lint
  expect_lint(
    trim_some("
      if (TRUE) {
        foo(idx <- bar())
      }
    "),
    lint_message,
    linter
  )

  expect_no_lint(
    trim_some("
      obj <- letters
      while ((n <- length(obj)) > 0) obj <- obj[-n]
    "),
    linter
  )
  expect_lint(
    trim_some("
      obj <- letters
      while ((n <- length(obj)) > 0) obj <- obj[-n]
      if (TRUE) {
        print(n)
      }
    "),
    lint_message,
    linter
  )

  # outside of branching, doesn't matter
  expect_lint("foo(idx <- bar()); baz()", lint_message, linter)
  expect_lint("foo(x, idx <- bar()); baz()", lint_message, linter)

  # adversarial comments
  expect_no_lint(
    trim_some("
      if # comment
      (any(idx <- x < 0)) {
        stop('negative elements: ', toString(which(idx)))
      }
    "),
    linter
  )
})

test_that("interaction of allow_lazy and allow_scoped", {
  linter <- implicit_assignment_linter(allow_scoped = TRUE, allow_lazy = TRUE)

  expect_no_lint(
    trim_some("
      if (any(idx <- foo()) && BB) {
        stop('Invalid foo() output: ', toString(idx))
      }
    "),
    linter
  )
  expect_lint(
    trim_some("
      if (any(idx <- foo()) && BB) {
        stop('Invalid foo() output: ', toString(idx))
      }
      print(format(idx))
    "),
    rex::rex("Avoid implicit assignments in function calls."),
    linter
  )
  expect_no_lint(
    trim_some("
      if (AA && any(idx <- foo())) {
        stop('Invalid foo() output: ', toString(idx))
      }
      print(format(idx)) # NB: bad code! idx may not exist.
    "),
    linter
  )
})

test_that("call-less '(' mentions avoiding implicit printing", {
  linter <- implicit_assignment_linter()
  implicit_msg <- rex::rex("Avoid implicit assignments in function calls.")
  print_msg <- rex::rex("Call print() explicitly instead of relying on implicit printing behavior via '('.")

  expect_lint("(a <- foo())", print_msg, linter)

  # only for top-level assignments; withAutoprint() ignored
  expect_lint("for (xi in x) (z <- foo(xi))", implicit_msg, linter)

  # mixed messages
  expect_lint(
    trim_some("
      (a <- foo())
      bar(z <- baz(a))
    "),
    list(print_msg, implicit_msg),
    linter
  )
})

test_that("allow_paren_print allows `(` for auto printing", {
  lint_message <- rex::rex("Avoid implicit assignments in function calls.")
  linter <- implicit_assignment_linter(allow_paren_print = TRUE)
  expect_no_lint("(a <- foo())", linter)

  # Doesn't effect other cases
  lint_message <- rex::rex("Avoid implicit assignments in function calls.")
  expect_lint("if (x <- 1L) TRUE", lint_message, linter)
  expect_lint("while (x <- 0L) FALSE", lint_message, linter)
  expect_lint("for (x in 1:10 -> y) print(x)", lint_message, linter)
  expect_lint("mean(x <- 1:4)", lint_message, linter)

  # default remains as is
  print_msg <- rex::rex("Call print() explicitly instead of relying on implicit printing behavior via '('.")
  expect_lint("(a <- foo())", print_msg, implicit_assignment_linter())
})
# fuzzer enable: assignment

test_that("which_grepl_linter skips allowed usages", {
  # this _could_ be combined as p1|p2, but often it's cleaner to read this way
  expect_no_lint("which(grepl(p1, x) | grepl(p2, x))", which_grepl_linter())
})

test_that("which_grepl_linter blocks simple disallowed usages", {
  linter <- which_grepl_linter()
  lint_msg <- rex::rex("grep(pattern, x) is better than which(grepl(pattern, x)).")

  expect_lint("which(grepl('^a', x))", lint_msg, linter)
  # options also don't matter (grep has more arguments: value, invert)
  expect_lint("which(grepl('^a', x, perl = TRUE, fixed = TRUE))", lint_msg, linter)

  expect_lint(
    trim_some('{
      which(x)
      grepl(y)
      which(grepl("pt1", x))
      which(grepl("pt2", y))
    }'),
    list(
      list(lint_msg, line_number = 4L, column_number = 3L),
      list(lint_msg, line_number = 5L, column_number = 3L)
    ),
    linter
  )
})

# The lints for a given file should be the same regardless of the working
# directory

test_that("lint() results do not depend on the working directory", {
  # Helper function: run assignment_linter on a given file
  lint_assignments <- function(filename) {
    lint(filename, linters = list(assignment_linter()))
  }

  # a dummy package for use in the test
  pkg_path <- test_path("dummy_packages", "assignmentLinter")

  # put a .lintr in the package root that excludes the first line of `R/jkl.R`
  local_config("exclusions: list('R/jkl.R' = 1)", pkg_path)

  # linting the `R/jkl.R` should identify the following assignment lint on the
  # second line of the file
  expected_lines <- "mno = 789"

  # lint the file from:
  # - outside the package
  # - at the package root
  # - in the package's R/ directory

  lints_from_outside <- lint_assignments(
    file.path(pkg_path, "R", "jkl.R")
  )
  lints_from_pkg_root <- withr::with_dir(
    pkg_path,
    lint_assignments(file.path("R", "jkl.R"))
  )
  lints_from_a_subdir <- withr::with_dir(
    file.path(pkg_path, "R"),
    lint_assignments("jkl.R")
  )

  expect_identical(
    as.data.frame(lints_from_pkg_root)[["line"]],
    expected_lines
  )
  expect_identical(
    as.data.frame(lints_from_outside),
    as.data.frame(lints_from_pkg_root)
  )
  expect_identical(
    as.data.frame(lints_from_a_subdir),
    as.data.frame(lints_from_pkg_root)
  )
})

# The lints for a given file should be the same regardless of where the .lintr
# file is positioned (file-exclusions in the .lintr should be relative to the
# directory containing the .lintr)

test_that("lint() results do not depend on the position of the .lintr", {
  # .lintr config files for lint(filepath) are looked for in:
  # - the same directory as filepath
  # - the project directory
  # - the user's home directory
  lint_with_config <- function(config_string, config_dir, filename) {
    local_config(config_string, config_dir)
    lint(filename, linters = assignment_linter())
  }

  # a dummy package for use in the test
  pkg_path <- test_path("dummy_packages", "assignmentLinter")

  # we lint the file <pkg-root>/R/jkl.R using the pkg-root as working directory
  # and
  # - 1) a .lintr config in the package root,
  # - 2) a .lintr config in the source directory R/

  # The second line of jkl.R contains the following assignment lint:
  expected_lines <- "mno = 789"

  lints_with_config_at_pkg_root <- withr::with_dir(
    pkg_path,
    lint_with_config(
      config_string = "exclusions: list('R/jkl.R' = 1)",
      config_dir = ".",
      filename = file.path("R", "jkl.R")
    )
  )

  lints_with_config_in_r_dir <- withr::with_dir(
    pkg_path,
    lint_with_config(
      config_string = "exclusions: list('jkl.R' = 1)",
      config_dir = "R",
      filename = file.path("R", "jkl.R")
    )
  )

  expect_identical(
    as.data.frame(lints_with_config_at_pkg_root)[["line"]], expected_lines
  )
  expect_identical(
    as.data.frame(lints_with_config_at_pkg_root),
    as.data.frame(lints_with_config_in_r_dir),
    info = paste(
      "lints for a source file should be independent of whether the .lintr",
      "file is in the project-root or the source-file-directory"
    )
  )
})

test_that("lint uses linter names", { # nofuzz: assignment
  expect_lint(
    "a = 2",
    list(linter = "bla"),
    linters = list(bla = assignment_linter()),
    parse_settings = FALSE
  )
})

test_that("lint() results from file or text should be consistent", {
  linters <- list(assignment_linter(), infix_spaces_linter())
  lines <- c("x<-1", "x+1")
  file <- withr::local_tempfile(lines = lines)
  text <- paste(lines, collapse = "\n")
  file <- normalize_path(file)

  lint_from_file <- lint(file, linters = linters)
  lint_from_lines <- lint(linters = linters, text = lines)
  lint_from_text <- lint(linters = linters, text = text)

  # Remove file before linting to ensure that lint works and do not
  # assume that file exists when both filename and text are supplied.
  expect_identical(unlink(file), 0L)
  lint_from_text2 <- lint(file, linters = linters, text = text)

  expect_length(lint_from_file, 2L)
  expect_length(lint_from_lines, 2L)
  expect_length(lint_from_text, 2L)
  expect_length(lint_from_text2, 2L)

  expect_identical(lint_from_file, lint_from_text2)

  for (i in seq_along(lint_from_lines)) {
    lint_from_file[[i]]$filename <- ""
    lint_from_lines[[i]]$filename <- ""
    lint_from_text[[i]]$filename <- ""
  }

  expect_identical(lint_from_file, lint_from_lines)
  expect_identical(lint_from_file, lint_from_text)
})

test_that("exclusions work with custom linter names", { # nofuzz: assignment comment_injection
  expect_no_lint(
    "a = 2 # nolint: bla.",
    linters = list(bla = assignment_linter()),
    parse_settings = FALSE
  )
})

test_that("old compatibility usage errors", {
  error_msg <- rex::rex("Expected `", anything, "` to be a function of class <linter>")

  expect_error(
    expect_lint(
      "a == NA",
      "Use is.na",
      linters = equals_na_linter
    ),
    error_msg
  )

  expect_error(
    expect_lint(
      "a = 42",
      "Use <-",
      linters = assignment_linter
    ),
    error_msg
  )

  # Also within `linters_with_defaults()` (#1725)
  expect_error(
    expect_lint(
      "a = 42",
      "Use <-",
      linters = linters_with_defaults(assignment_linter)
    ),
    error_msg
  )

  expect_error(
    expect_lint(
      "a == NA",
      "Use is.na",
      linters = unclass(equals_na_linter())
    ),
    error_msg
  )

  # Trigger compatibility in auto_names()
  expect_error(
    expect_lint(
      "a == NA",
      "Use is.na",
      linters = list(unclass(equals_na_linter()))
    ),
    error_msg
  )

  expect_error(
    lint("a <- 1\n", linters = \(two, arguments) NULL),
    error_msg
  )

  expect_error(
    lint("a <- 1\n", linters = "equals_na_linter"),
    error_msg
  )
})

test_that("Linters throwing an error give a helpful error", {
  tmp_file <- withr::local_tempfile(lines = "a <- 1")
  lintr_error_msg <- "a broken linter"
  linter <- function() Linter(\(source_expression) cli_abort(lintr_error_msg))
  # NB: Some systems/setups may use e.g. symlinked files when creating under tempfile();
  #   we don't care much about that, so just check basename()
  expect_error(lint(tmp_file, linter()), lintr_error_msg, fixed = TRUE)
  expect_error(lint(tmp_file, list(broken_linter = linter())), lintr_error_msg, fixed = TRUE)
})

test_that("Linter() input is validated", {
  expect_error(Linter(1L), "`fun` must be a function taking exactly one argument", fixed = TRUE)
  expect_error(Linter(\(a, b) TRUE), "`fun` must be a function taking exactly one argument", fixed = TRUE)
})

test_that("typo in argument name gives helpful error", {
  expect_error(lint("xxx", litners = identity), "Found unknown arguments in `...`: `litners`")
})


test_that("gitlab_output() writes expected report", {
  skip_if_not_installed("jsonlite")

  tmpfile <- withr::local_tempfile()

  # zero lints: we expect an empty json array
  gitlab_output(lint(text = "", linters = infix_spaces_linter()), filename = tmpfile)
  expect_match(
    readLines(tmpfile),
    R"(\s*\[\s*\]\s*)"
  )

  # single lint
  gitlab_output(lint(text = "x<-1", linters = infix_spaces_linter()), filename = tmpfile)
  expect_identical(
    jsonlite::read_json(tmpfile),
    list(list(
      description = "Put spaces around all infix operators.",
      check_name = "infix_spaces_linter",
      fingerprint = "eb7cc117e8616bd8170fe6aa29e8b0ae849ac6c7",
      location = list(path = "<text>", lines = list(begin = 1L)),
      severity = "info"
    ))
  )

  # two lints
  gitlab_output(lint(text = c("x<-1", "y<-1"), linters = infix_spaces_linter()), filename = tmpfile)
  expect_identical(
    jsonlite::read_json(tmpfile),
    list(list(
      description = "Put spaces around all infix operators.",
      check_name = "infix_spaces_linter",
      fingerprint = "eb7cc117e8616bd8170fe6aa29e8b0ae849ac6c7",
      location = list(path = "<text>", lines = list(begin = 1L)),
      severity = "info"
    ), list(
      description = "Put spaces around all infix operators.",
      check_name = "infix_spaces_linter",
      fingerprint = "c20bd2090d08e3a5c12d670f5763ad43d233fe05",
      location = list(path = "<text>", lines = list(begin = 2L)),
      severity = "info"
    ))
  )

  expect_error(gitlab_output(NULL), "must be a <lints> object", fixed = TRUE)
})

test_that("explicit parse_settings=TRUE works for inline data", {
  withr::local_dir(tempdir())
  .lintr <- withr::local_tempfile(lines = "linters: list(assignment_linter())")
  withr::local_options(list(lintr.linter_file = .lintr))

  lint_str <- "a=1\n" # assignment lints, but not infix_spaces
  foo <- withr::local_tempfile(lines = lint_str)

  expect_length(lint(foo, parse_settings = TRUE), 1L)
  expect_length(lint(text = lint_str, parse_settings = TRUE), 1L)
  expect_length(lint(lint_str, parse_settings = TRUE), 1L)

  # parse_settings=TRUE default not picked up
  expect_length(lint(text = lint_str), 2L)
})

test_that("lint(text=) handles unmarked UTF-8 text correctly", {
  skip_if_not_utf8_locale()

  tf <- withr::local_tempfile()
  writeLines('x <- "\u00e4"', tf, useBytes = TRUE)
  text_native <- readLines(tf)

  expect_no_error(
    expect_length(lint(text = text_native, linters = absolute_path_linter()), 0L)
  )

  writeLines(c('x <- "\u00e4"', 'y <- "/absolute/path"'), tf, useBytes = TRUE)
  text_native_lint <- readLines(tf)

  expect_no_error({
    res_lint <- lint(text = text_native_lint, linters = absolute_path_linter(lax = FALSE))
  })
  expect_identical(res_lint[[1L]]$line_number, 2L)
  expect_identical(res_lint[[1L]]$message, "Do not use absolute paths.")
})

test_that("lint(text=) handles UTF-8 marked text in non-UTF-8 locale", {
  # Set LC_CTYPE to C to simulate non-UTF-8 locale
  withr::local_locale(c(LC_CTYPE = "C"))

  # Create UTF-8 marked text
  text <- 'x <- "\u00e4"'
  expect_identical(Encoding(text), "UTF-8")

  expect_no_error(
    expect_length(lint(text = text, linters = absolute_path_linter()), 0L)
  )
})

test_that("strings_as_factors_linter skips allowed usages", {
  linter <- strings_as_factors_linter()

  expect_no_lint("data.frame(1:3)", linter)
  expect_no_lint("data.frame(x = 1:3)", linter)

  expect_no_lint("data.frame(x = 'a', stringsAsFactors = TRUE)", linter)
  expect_no_lint("data.frame(x = 'a', stringsAsFactors = FALSE)", linter)
  expect_no_lint("data.frame(x = c('a', 'b'), stringsAsFactors = FALSE)", linter)

  # strings in argument names to c() don't get linted
  expect_no_lint("data.frame(x = c('a b' = 1L, 'b c' = 2L))", linter)

  # characters supplied to row.names are not affected
  expect_no_lint("data.frame(x = 1:3, row.names = c('a', 'b', 'c'))", linter)

  # ambiguous cases passes
  expect_no_lint("data.frame(x = c(xx, 'a'))", linter)
  expect_no_lint("data.frame(x = c(foo(y), 'a'))", linter)

  # adversarial comments
  expect_no_lint(
    trim_some("
      data.frame(
        x = 1:3,
        row.names # comment
        = c('a', 'b', 'c')
      )
    "),
    linter
  )
})

test_that("strings_as_factors_linter blocks simple disallowed usages", {
  linter <- strings_as_factors_linter()
  lint_msg <- "Supply an explicit value for stringsAsFactors for this code"

  expect_lint("data.frame('a')", lint_msg, linter)
  expect_lint("data.frame(c('a', 'b'))", lint_msg, linter)
  expect_lint("data.frame(x = 1:5, y = 'b')", lint_msg, linter)
  expect_lint("data.frame(x = 1:5, y, z = 'b')", lint_msg, linter)

  # catch row.names when combined with a character vector
  expect_lint("data.frame(x = c('c', 'd', 'e'), row.names = c('a', 'b', 'c'))", lint_msg, linter)
  expect_lint("data.frame(row.names = c('c', 'd', 'e'), x = c('a', 'b', 'c'))", lint_msg, linter)

  # when everything is a literal, type promotion means the column is character
  expect_lint("data.frame(x = c(TRUE, 'a'))", lint_msg, linter)
})

test_that("strings_as_factors_linters catches rep(char) usages", {
  linter <- strings_as_factors_linter()
  lint_msg <- "Supply an explicit value for stringsAsFactors for this code"

  expect_lint("data.frame(rep('a', 10L))", lint_msg, linter)
  expect_lint("data.frame(rep(c('a', 'b'), 10L))", lint_msg, linter)

  # literal char, not mixed or non-char
  expect_no_lint("data.frame(rep(1L, 10L))", linter)
  expect_no_lint("data.frame(rep(c(x, 'a'), 10L))", linter)
  # however, type promotion of literals is caught
  expect_lint("data.frame(rep(c(TRUE, 'a'), 10L))", lint_msg, linter)
})

test_that("strings_as_factors_linter catches character(), as.character() usages", {
  linter <- strings_as_factors_linter()
  lint_msg <- "Supply an explicit value for stringsAsFactors for this code"

  expect_lint("data.frame(a = character())", lint_msg, linter)
  expect_lint("data.frame(a = character(1L))", lint_msg, linter)
  expect_lint("data.frame(a = as.character(x))", lint_msg, linter)

  # but not for row.names
  expect_no_lint("data.frame(a = 1:10, row.names = as.character(1:10))", linter)
})

test_that("strings_as_factors_linter catches more functions with string output", {
  linter <- strings_as_factors_linter()
  lint_msg <- "Supply an explicit value for stringsAsFactors for this code"

  expect_lint("data.frame(a = paste(1, 2, 3))", lint_msg, linter)
  expect_lint("data.frame(a = sprintf('%d', 1:10))", lint_msg, linter)
  expect_lint("data.frame(a = format(x, just = 'right'))", lint_msg, linter)
  expect_lint("data.frame(a = formatC(x, format = '%d'))", lint_msg, linter)
  expect_lint("data.frame(a = prettyNum(x, big.mark = ','))", lint_msg, linter)
  expect_lint("data.frame(a = toString(x))", lint_msg, linter)
  expect_lint("data.frame(a = encodeString(x))", lint_msg, linter)
  # but not for row.names
  expect_no_lint("data.frame(a = 1:10, row.names = paste(1:10))", linter)
})

test_that("lints vectorize", {
  lint_msg <- "Supply an explicit value for stringsAsFactors for this code"

  expect_lint(
    trim_some("{
      data.frame('a')
      data.frame('b')
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    strings_as_factors_linter()
  )
})

test_that("expect_identical_linter skips allowed usages", {
  linter <- expect_identical_linter()

  # expect_type doesn't have an inverted version
  expect_no_lint("expect_true(identical(x, y) || identical(y, z))", linter)
  # NB: also applies to tinytest, but it's sufficient to test testthat
  expect_no_lint("testthat::expect_true(identical(x, y) || identical(y, z))", linter)

  # expect_equal calls with explicit tolerance= are OK
  expect_no_lint("expect_equal(x, y, tolerance = 1e-6)", linter)
  # ditto if the argument is passed quoted
  expect_no_lint("expect_equal(x, y, 'tolerance' = 1e-6)", linter)

  # ditto for check.attributes = FALSE
  expect_no_lint("expect_equal(x, y, check.attributes = FALSE)", linter)
})

test_that("expect_identical_linter blocks simple disallowed usages", {
  linter <- expect_identical_linter()
  lint_msg <- rex::rex("Use expect_identical(x, y) by default; resort to expect_equal() only when needed")

  expect_lint("expect_equal(x, y)", lint_msg, linter)

  # different usage to redirect to expect_identical
  expect_lint("expect_true(identical(x, y))", lint_msg, linter)
})

test_that("expect_identical_linter skips cases likely testing numeric equality", {
  linter <- expect_identical_linter()
  lint_msg <- rex::rex("Use expect_identical(x, y) by default; resort to expect_equal() only when needed")

  expect_no_lint("expect_equal(x, 1.034)", linter)
  expect_no_lint("expect_equal(x, -1.034)", linter)
  expect_no_lint("expect_equal(x, c(-1.034))", linter)
  expect_no_lint("expect_equal(x, -c(1.034))", linter)

  expect_no_lint("expect_equal(x, c(1.01, 1.02))", linter)
  # whole numbers with explicit decimals are OK, even in mixed scenarios
  expect_no_lint("expect_equal(x, c(1.0, 2))", linter)
  # plain numbers are still caught, however
  expect_lint("expect_equal(x, 1L)", lint_msg, linter)
  expect_lint("expect_equal(x, 1)", lint_msg, linter)
  # NB: TRUE is a NUM_CONST so we want test matching it, even though this test is
  #   also a violation of expect_true_false_linter()
  expect_lint("expect_equal(x, TRUE)", lint_msg, linter)

  expect_lint("expect_equal(x, 1.01 - y)", lint_msg, linter)
  expect_lint("expect_equal(x, foo() - 0.01)", lint_msg, linter)
})

test_that("expect_identical_linter skips 3e cases needing expect_equal", {
  expect_no_lint("expect_equal(x, y, ignore_attr = 'names')", expect_identical_linter())
})

# this arose where a helper function was wrapping expect_equal() and
#   some of the "allowed" arguments were being passed --> false positive
test_that("expect_identical_linter skips calls using ...", {
  expect_no_lint("expect_equal(x, y, ...)", expect_identical_linter())
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Use expect_identical(x, y) by default; resort to expect_equal() only when needed")

  expect_lint(
    trim_some("{
      expect_equal(x, 1)
      expect_true(identical(x, y))
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    expect_identical_linter()
  )
})

test_that("ifelse_censor_linter skips allowed usages", {
  linter <- ifelse_censor_linter()

  expect_no_lint("ifelse(x == 2, x, y)", linter)
  expect_no_lint("ifelse(x > 2, x, y)", linter)
})

test_that("ifelse_censor_linter blocks simple disallowed usages", {
  linter <- ifelse_censor_linter()

  expect_lint(
    "ifelse(x < 0, 0, x)",
    rex::rex("pmax(x, y) is preferable to ifelse(x < y, y, x)"),
    linter
  )

  # other equivalents to base::ifelse()
  expect_lint(
    "if_else(x < 0, 0, x)",
    rex::rex("pmax(x, y) is preferable to if_else(x < y, y, x)"),
    linter
  )
  expect_lint(
    "fifelse(x < 0, 0, x)",
    rex::rex("pmax(x, y) is preferable to fifelse(x < y, y, x)"),
    linter
  )

  # other equivalents for censoring
  expect_lint(
    "ifelse(x <= 0, 0, x)",
    rex::rex("pmax(x, y) is preferable to ifelse(x <= y, y, x)"),
    linter
  )
  expect_lint(
    "ifelse(x > 0, x, 0)",
    rex::rex("pmax(x, y) is preferable to ifelse(x > y, x, y)"),
    linter
  )
  expect_lint(
    "ifelse(x >= 0, x, 0)",
    rex::rex("pmax(x, y) is preferable to ifelse(x >= y, x, y)"),
    linter
  )

  # pairwise min/max (similar to censoring)
  expect_lint(
    "ifelse(x < y, x, y)",
    rex::rex("pmin(x, y) is preferable to ifelse(x < y, x, y)"),
    linter
  )
  expect_lint(
    "ifelse(x >= y, y, x)",
    rex::rex("pmin(x, y) is preferable to ifelse(x >= y, y, x)"),
    linter
  )

  # more complicated expression still matches
  expect_lint(
    trim_some("
      ifelse(2 + p + 104 + 1 > ncols,
            ncols, 2 + p + 104 + 1
            )
    "),
    rex::rex("pmin(x, y) is preferable to ifelse(x > y, y, x)"),
    linter
  )

  # including with comments
  expect_lint(
    trim_some("
      ifelse(2 + p + 104 + 1 #comment
      > ncols, ncols, 2 + p + 104 + 1)
    "),
    rex::rex("pmin(x, y) is preferable to ifelse(x > y, y, x)"),
    linter
  )
  expect_lint(
    trim_some("
      ifelse(2 + p + 104 + # comment
      1 > ncols, ncols, 2 + p + 104 + 1)
    "),
    rex::rex("pmin(x, y) is preferable to ifelse(x > y, y, x)"),
    linter
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      ifelse(x >= y, y, x)
      ifelse(x >= 0, x, 0)
    }"),
    list(
      list("pmin", line_number = 2L),
      list("pmax", line_number = 3L)
    ),
    ifelse_censor_linter()
  )
})

test_that("missing_argument_linter skips allowed usages", {
  linter <- missing_argument_linter()

  expect_no_lint("fun(x, a = 1)", linter)
  expect_no_lint("fun(x = 1, a = 1)", linter)
  expect_no_lint("dt[, 1]", linter)
  expect_no_lint("dt[, 'col']", linter)
  expect_no_lint("array[, , 1]", linter)
  expect_no_lint("switch(a =, b =, c = 1, 0)", linter)
  expect_no_lint("alist(a =, b =, c = 1, 0)", linter)
  expect_no_lint("pairlist(path = quote(expr = ))", linter) # #1889

  # always allow this missing usage
  expect_no_lint("foo()", linter)
  # even if there's intervening comment(s)
  expect_no_lint("foo(#\n)", linter)
  expect_no_lint("foo(\n#\n)", linter)
  expect_no_lint("foo(#\n#\n)", linter)

  expect_no_lint("test(a =, b =, c = 1, 0)", missing_argument_linter("test"))
})

test_that("missing_argument_linter blocks disallowed usages", {
  linter <- missing_argument_linter()
  lint_msg1 <- rex::rex("Missing argument 1 in function call.")
  lint_msg2 <- rex::rex("Missing argument 2 in function call.")
  lint_msg3 <- rex::rex("Missing argument 3 in function call.")
  lint_msga <- rex::rex("Missing argument 'a' in function call.")

  expect_lint("fun(, a = 1)", lint_msg1, linter)
  expect_lint(
    "f <- function(x, y) x\nf(, y = 1)\n",
    list(lint_msg1, line = "f(, y = 1)"),
    linter
  )
  expect_lint("fun(a = 1,, b = 2)", lint_msg2, linter)
  expect_lint("fun(b = 1, a =)", lint_msga, linter)
  expect_lint("fun(a = 1,)", lint_msg2, linter)
  expect_lint("fun(a = )", lint_msga, linter)

  expect_lint(
    trim_some("
      list(
        a = 1,
        b = 2,
      )
    "),
    lint_msg3,
    linter
  )

  expect_lint("stats::median(1:10, a =)", lint_msga, linter)
  expect_lint("env$get(1:10, a =)", lint_msga, linter)
  expect_lint("env@get(1:10, a =)", lint_msga, linter)

  # Fixes https://github.com/r-lib/lintr/issues/906
  # Comments should be ignored so that missing arguments could be
  # properly identified in these cases.
  expect_lint(
    trim_some("
      fun(
        1,
        2,
        # comment
      )
    "),
    lint_msg3,
    linter
  )

  expect_lint(
    trim_some("
      fun(
        # comment
        ,
        1
      )
    "),
    lint_msg1,
    linter
  )

  expect_lint(
    trim_some("
      fun(
        a = # comment
        ,
        1
      )
    "),
    lint_msga,
    linter
  )
})

test_that("except list can be empty", {
  linter <- missing_argument_linter(character())
  lint_msg <- rex::rex("Missing argument 'a' in function call.")

  expect_lint("switch(a =, b = 1, 0)", lint_msg, linter)
  expect_lint("alist(a =)", lint_msg, linter)
})

test_that("allow_trailing can allow trailing empty args also for non-excepted functions", {
  linter <- missing_argument_linter(allow_trailing = TRUE)

  expect_no_lint("fun(a = 1,)", linter)
  expect_no_lint(
    trim_some("
      fun(
        a = 1,
        # comment
      )
    "),
    linter
  )
  # ... but not if the final argument is named
  expect_lint(
    "fun(a = 1, b = )",
    rex::rex("Missing argument 'b' in function call."),
    linter
  )
})

test_that("lints vectorize", {
  linter <- missing_argument_linter()
  linter_trailing <- missing_argument_linter(allow_trailing = TRUE)
  lint_msg <- rex::rex("Missing argument in function call.")

  expect_lint(
    "foo(,,)",
    list(
      list("Missing argument 1", column_number = 5L),
      list("Missing argument 2", column_number = 6L),
      list("Missing argument 3", column_number = 7L)
    ),
    linter
  )
  expect_lint(
    "foo(,,)",
    list(
      list("Missing argument 1", column_number = 5L),
      list("Missing argument 2", column_number = 6L)
    ),
    linter_trailing
  )

  expect_lint(
    "foo(a =,,)",
    list(
      list("Missing argument 'a'", column_number = 7L),
      list("Missing argument 2", column_number = 9L),
      list("Missing argument 3", column_number = 10L)
    ),
    linter
  )
  expect_lint(
    "foo(a =,,)",
    list(
      list("Missing argument 'a'", column_number = 7L),
      list("Missing argument 2", column_number = 9L)
    ),
    linter_trailing
  )

  expect_lint(
    trim_some("{
      foo(1,)
      bar(,2)
    }"),
    list(
      list("Missing argument 2", line_number = 2L),
      list("Missing argument 1", line_number = 3L)
    ),
    linter
  )
  expect_lint(
    trim_some("{
      foo(1,)
      bar(,2)
    }"),
    list("Missing argument 1", line_number = 3L),
    linter_trailing
  )
})

test_that("expect_named_linter skips allowed usages", {
  linter <- expect_named_linter()

  # colnames(), rownames(), and dimnames() tests are not equivalent
  expect_no_lint("expect_equal(colnames(x), 'a')", linter)
  expect_no_lint("expect_equal(rownames(x), 'a')", linter)
  expect_no_lint("expect_equal(dimnames(x), 'a')", linter)

  expect_no_lint("expect_equal(nrow(x), 4L)", linter)
  # NB: also applies to tinytest, but it's sufficient to test testthat
  expect_no_lint("testthat::expect_equal(nrow(x), 4L)", linter)

  # only check the first argument. yoda tests in the second argument will be
  #   missed, but there are legitimate uses of names() in argument 2
  expect_no_lint("expect_equal(colnames(x), names(y))", linter)
})

test_that("expect_named_linter blocks simple disallowed usages", {
  linter <- expect_named_linter()
  lint_msg <- rex::rex("expect_named(x, n) is better than expect_equal(names(x), n)")

  expect_lint("expect_equal(names(x), 'a')", lint_msg, linter)
  expect_lint("testthat::expect_equal(names(DF), names(old))", lint_msg, linter)
  expect_lint("expect_equal('a', names(x))", lint_msg, linter)
})

test_that("expect_named_linter blocks expect_identical usage as well", {
  expect_lint(
    "expect_identical(names(x), 'a')",
    rex::rex("expect_named(x, n) is better than expect_identical(names(x), n)"),
    expect_named_linter()
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      expect_equal(names(x), nm)
      expect_identical(names(x), nm)
    }"),
    list(
      list("expect_equal", line_number = 2L),
      list("expect_identical", line_number = 3L)
    ),
    expect_named_linter()
  )
})

# testthat's expect_success() and expect_failure() can only handle the first expectation and are
# thus less than ideal to test expect_lint(), which can process multiple lints. If you want to test
# for failure, always put the lint check or lint field that must fail first.

# fuzzer disable: assignment
linter <- assignment_linter()
lint_msg <- "Use <- for assignment, not ="

test_that("no checks", {
  expect_success(expect_no_lint("a", linter))
  expect_success(expect_no_lint("a=1", list()))
  expect_failure(expect_no_lint("a=1", linter))
})

test_that("single check", {
  expect_failure(expect_lint(character(), lint_msg, linter))
  expect_failure(expect_lint("", lint_msg, linter))

  expect_success(expect_lint(content = "a=1", checks = lint_msg, linters = linter))
  expect_success(expect_lint("a=1", lint_msg, linter))
  expect_failure(expect_lint("a=1", "asdf", linter))
  expect_success(expect_lint("a=1", c(message = lint_msg), linter))
  expect_failure(expect_lint("a=1", c(message = NULL), linter))
  expect_success(expect_lint("a=1", list(message = lint_msg, line_number = 1L), linter))
  expect_failure(expect_lint("a=1", c(line_number = 2L, message = lint_msg), linter))

  expect_error(expect_lint("a=1", c(message = lint_msg, lineXXX = 1L), linter), "Check 1 has an invalid field: lineXXX")

  expect_failure(expect_lint("foo ()", list(ranges = list(c(2L, 2L))), function_left_parentheses_linter()))
  expect_success(expect_lint("\t1", list(ranges = list(c(1L, 1L))), whitespace_linter()))
  expect_success(expect_lint("a=1", list(message = lint_msg, line_number = 1L), linter))
  expect_failure(expect_lint("a=1", list(2L, lint_msg), linter))

  expect_success(expect_lint("1:nrow(x)", "(nrow)", seq_linter()))
})

test_that("multiple checks", { # nofuzz: comment_injection
  expect_success(
    expect_lint(file = "exclusions-test", checks = as.list(rep(lint_msg, 9L)), linters = linter, parse_settings = FALSE)
  )

  expect_success(expect_lint("a=1; b=2", list(lint_msg, lint_msg), linter))
  expect_success(expect_lint("a=1; b=2", list(c(message = lint_msg), c(message = lint_msg)), linter))
  expect_success(expect_lint("a=1; b=2", list(c(line_number = 1L), c(linter = "assignment_linter")), linter))
  expect_success(expect_lint("a=1; b=2", list(lint_msg, list(line = "a=1; b=2", type = "style")), linter))
  expect_success(expect_lint("a=1\nb=2", list(c(line_number = 1L), c(line_number = 2L)), linter))
  expect_failure(expect_lint("a=1\nb=2", list(c(line_number = 2L), c(line_number = 2L)), linter))

  expect_success(expect_lint("a=1\nb=2", list(list(line_number = 1L), list(line_number = 2L)), linter))
  expect_failure(expect_lint("a=1; b=2", list(list(line_number = 2L), list(line_number = 2L)), linter))
  expect_success(
    expect_lint("\t1\n\t2", list("tabs", list(column_number = 1L, ranges = list(c(1L, 1L)))), whitespace_linter())
  )
})

test_that("expect_lint_free works", {
  withr::local_options(
    lintr.rstudio_source_markers = FALSE,
    lintr.linter_file = "lintr_test_config"
  )
  withr::local_envvar(c(NOT_CRAN = "true", R_COVR = "false"))

  expect_lint_free(test_path("dummy_packages", "clean"))
  expect_lint_free(test_path("dummy_packages", "clean_subdir", "r"))
  expect_failure(expect_lint_free(test_path("dummy_packages", "package")))
})

test_that("expect_lint doesn't change language", {
  withr::with_envvar(c(LANGUAGE = "mr"), {
    expect_success(expect_lint("a=1", lint_msg, linter))
    expect_identical(Sys.getenv("LANGUAGE"), "mr")
  })
  withr::with_envvar(c(LANGUAGE = NA), {
    expect_success(expect_lint("a=1", lint_msg, linter))
    expect_identical(Sys.getenv("LANGUAGE", unset = NA), NA_character_)
  })
})

test_that("execution without testthat gives the right errors", {
  local_mocked_bindings(requireNamespace = \(...) FALSE)
  lint_msg <- function(nm) rex::rex("`", nm, "()` is designed to work", anything, "testthat")

  expect_error(expect_lint(), lint_msg("expect_lint"))
  expect_error(lintr::expect_lint(), lint_msg("expect_lint"))
  expect_error(expect_no_lint(), lint_msg("expect_no_lint"))
  expect_error(expect_lint_free(), lint_msg("expect_lint_free"))
})

test_that("lint order can be ignored", { # nofuzz: comment_injection
  linters <- list(assignment_linter(), infix_spaces_linter())
  expected <- lapply(linters, \(l) list(linter = attr(l, "name")))
  expect_success(expect_lint("a=1", expected, linters, ignore_order = TRUE))
  expect_success(expect_lint("a=1", rev(expected), linters, ignore_order = TRUE))

  lines <- trim_some("
    a=1
    b=2
  ")
  expected <- list(list(line_number = 1L), list(line_number = 2L))
  expect_success(expect_lint(lines, expected, assignment_linter(), ignore_order = TRUE))
  expect_success(expect_lint(lines, rev(expected), assignment_linter(), ignore_order = TRUE))

  # a fuzz test, since base R doesn't have a trivial way to do permutations
  expected <- list(
    list(linter = "assignment_linter", line_number = 1L),
    list(linter = "assignment_linter", line_number = 2L),
    list(linter = "infix_spaces_linter", line_number = 1L),
    list(linter = "infix_spaces_linter", line_number = 2L)
  )
  expect_success(expect_lint(lines, expected[sample.int(4L)], linters, ignore_order = TRUE))
})
# fuzzer enable: assignment

test_that("expect_s4_class_linter skips allowed usages", {
  linter <- expect_s4_class_linter()

  # expect_s4_class doesn't have an inverted version
  expect_no_lint("expect_true(!is(x, 'class'))", linter)
  # NB: also applies to tinytest, but it's sufficient to test testthat
  expect_no_lint("testthat::expect_s3_class(!is(x, 'class'))", linter)

  # expect_s4_class() doesn't have info= or label= arguments
  expect_no_lint("expect_true(is(x, 'SpatialPoly'), info = 'x should be SpatialPoly')", linter)
  expect_no_lint("expect_true(is(x, 'SpatialPoly'), label = 'x inheritance')", linter)
})

test_that("expect_s4_class blocks simple disallowed usages", {
  linter <- expect_s4_class_linter()
  lint_msg <- rex::rex("expect_s4_class(x, k) is better than expect_true(is(x, k))")

  expect_lint("expect_true(is(x, 'data.frame'))", lint_msg, linter)
  # namespace qualification is irrelevant
  expect_lint("testthat::expect_true(methods::is(x, 'SpatialPolygonsDataFrame'))", lint_msg, linter)
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("expect_s4_class(x, k) is better than expect_true(is(x, k))")

  expect_lint(
    trim_some("{
      expect_true(is(x, 'data.frame'))
      expect_true(is(x, 'SpatialPolygonsDataFrame'))
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    expect_s4_class_linter()
  )
})

test_that("expect_true_false_linter skips allowed usages", {
  linter <- expect_true_false_linter()
  # expect_true is a scalar test; testing logical vectors with expect_equal is OK
  expect_no_lint("expect_equal(x, c(TRUE, FALSE))", linter)

  expect_no_lint("expect_equal(x, y, ignore_attr = TRUE)", linter)

  expect_no_lint("42 %>% expect_identical(42, ignore_attr = TRUE)", linter)
  expect_no_lint("42 %>% expect_identical(42, TRUE)", linter)
  expect_no_lint("42 |> expect_identical(42, ignore_attr = TRUE)", linter)
})

test_that("expect_true_false_linter blocks simple disallowed usages", {
  linter <- expect_true_false_linter()
  lint_msg <- rex::rex("expect_true(x) is better than expect_equal(x, TRUE)")

  expect_lint("expect_equal(foo(x), TRUE)", lint_msg, linter)

  # expect_identical is treated the same as expect_equal
  expect_lint(
    "testthat::expect_identical(x, FALSE)",
    rex::rex("expect_false(x) is better than expect_identical(x, FALSE)"),
    linter
  )

  # also caught when TRUE/FALSE is the first argument
  expect_lint("expect_equal(TRUE, foo(x))", lint_msg, linter)

  expect_lint("42 %T>% expect_equal(TRUE)", lint_msg, linter)
  expect_lint("42 |> expect_equal(TRUE)", lint_msg, linter)
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      expect_equal(x, TRUE)
      expect_equal(x, FALSE)
    }"),
    list(
      list("expect_true", line_number = 2L),
      list("expect_false", line_number = 3L)
    ),
    expect_true_false_linter()
  )
})

test_that("returns the correct linting", {
  skip_if_not_installed("cyclocomp")

  cc_linter_1 <- cyclocomp_linter(1L)
  cc_linter_2 <- cyclocomp_linter(2L)
  lint_msg <- rex::rex("Reduce the cyclomatic complexity of this expression")

  expect_no_lint("if (TRUE) 1 else 2", cc_linter_2)
  expect_lint("if (TRUE) 1 else 2", lint_msg, cc_linter_1)

  expect_lint(
    "function(x) {not parsing}",
    "unexpected symbol", cc_linter_2
  )

  complex_lines <- trim_some("
    complexity <- function(x) {
      if (x > 0.0) {
        if (x > 10.0) {
          if (x > 20.0) {
            x <- x / 2.0
          } else {
            return(x)
          }
        } else {
          return(x)
        }
      } else {
        if (x < -10.0) {
          if (x < -20.0) {
            x <- x * 2.0
          } else {
            return(x)
          }
        } else {
          return(x)
        }
      }
      x
    }
  ")
  expect_lint(complex_lines, lint_msg, cc_linter_2)
  expect_lint(
    complex_lines,
    list(rex::rex("Reduce the cyclomatic complexity of this expression from 10 to at most 2."), line_number = 1L),
    cc_linter_2
  )
  expect_no_lint(complex_lines, cyclocomp_linter(10L))

  # no function involved, still lints, hence use 'expression' in the message
  expect_lint(
    trim_some("
      for (i in 1:10)
        for (j in 1:10)
          for (k in 1:10)
            i*j*k
    "),
    lint_msg,
    cyclocomp_linter(5L)
  )
})

test_that("a null linter is returned, with warning, if cyclocomp is unavailable", {
  # simple requireNamspace->FALSE won't work since expect_no_lint checks for testthat
  local_mocked_bindings(
    requireNamespace = \(pkg, ...) pkg != "cyclocomp" && base::requireNamespace(pkg, ...)
  )
  expect_warning(regexp = "Please install", fixed = TRUE, {
    linter <- cyclocomp_linter(1L)
  })

  expect_error(lint(text = "if (TRUE) 1 else 2", linters = linter), "disabled", fixed = TRUE)
})

test_that("returns the correct linting", {
  linter <- object_length_linter()
  lint_msg <- rex::rex("Variable and function names should not be longer than 30 characters.")

  expect_no_lint("blah", linter)

  expect_lint("very_very_very_very_long_variable_names_are_not_ideal <- 1", lint_msg, linter)

  expect_lint(
    "very_very_very_very_long_variable_names_are_not_ideal <<- 'foo'",
    rex::rex("Variable and function names should not be longer than 40 characters."),
    object_length_linter(length = 40L)
  )
})

# Regression tests for #871
test_that("lints S3 generics correctly", {
  linter <- object_length_linter()
  lint_msg <- rex::rex("Variable and function names should not be longer than 30 characters.")

  expect_no_lint("print.very_very_long_class_name <- 1", linter)
  expect_lint("print.very_very_very_very_long_class_name <- 1", lint_msg, linter)

  expect_lint(
    trim_some("
    very_very_very_long_generic_name <- function(x, ...) {
      UseMethod(\"very_very_very_long_generic_name\")
    }

    very_very_very_long_generic_name.short_class <- function(x, ...) {
      42L
    }

    very_very_very_long_generic_name.very_very_very_very_long_class_name <- function(x, ...) {
      2L
    }
  "),
    list(
      list(line_number = 1L),
      list(line_number = 9L)
    ),
    linter
  )
})

test_that("object_length_linter won't fail if an imported namespace is unavailable", {
  expect_length(
    lint_package(
      test_path("dummy_packages", "missing_dep"),
      linters = object_length_linter(),
      parse_settings = FALSE
    ),
    3L
  )
})

test_that("object_length_linter won't fail if dependency has no exports", {
  expect_length(
    lint_package(
      test_path("dummy_packages", "no_export_dep"),
      linters = object_length_linter(),
      parse_settings = FALSE
    ),
    1L
  )
})

test_that("function shorthand is caught", {
  expect_lint(
    "abcdefghijklm <- \\() NULL",
    "function names",
    object_length_linter(length = 10L)
  )
})

test_that("rlang name injection is handled", {
  linter <- object_length_linter(length = 10L)

  expect_no_lint("tibble('{foo() |> bar() |> baz()}' := TRUE)", linter)
  expect_lint("DT[, 'a_very_long_name' := FALSE]", "names should not be longer than 10 characters", linter)
})

test_that("literals in assign() and setGeneric() are checked", {
  linter <- object_length_linter(length = 10L)
  lint_msg <- rex::rex("Variable and function names should not be longer")

  expect_no_lint("assign('good_name', 2, env)", linter)
  expect_no_lint("assign('good_name', 'badBadBadBadName', env)", linter)
  expect_lint("assign('badBadBadBadName', 2, env)", lint_msg, linter)
  expect_lint('assign("badBadBadBadName", 2, env)', lint_msg, linter)

  expect_no_lint("setGeneric('good_name', function(x) x)", linter)
  expect_no_lint("setGeneric('good_name', function(x) x, package = 'badBadBadBadName')", linter)
  expect_lint("setGeneric('badBadBadBadName', function(x) x)", lint_msg, linter)
  expect_lint('setGeneric("badBadBadBadName", function(x) x)', lint_msg, linter)

  # Ditto for keyword arguments
  expect_no_lint("assign(x = 'good_name', 2, env)", linter)
  expect_no_lint("assign(envir = 'badBadBadBadName', nm, 2)", linter)
  expect_no_lint("assign(envir = 'badBadBadBadName', 'good_name', 2)", linter)
  expect_no_lint("assign(envir = 'badBadBadBadName', x = 'good_name', 2)", linter)
  expect_lint("assign(x = 'badBadBadBadName', 2, env)", lint_msg, linter)
  expect_lint("assign(envir = 'good_env_name', 'badBadBadBadName', 2)", lint_msg, linter)
  expect_lint("assign(envir = 'good_env_name', x = 'badBadBadBadName', 2)", lint_msg, linter)

  # adversarial comments
  expect_lint(
    trim_some("
      assign(envir = # comment
      'good_env_name', 'badBadBadBadName', 2)
    "),
    lint_msg,
    linter
  )
})

test_that("expect_comparison_linter skips allowed usages", {
  linter <- expect_comparison_linter()

  # there's no expect_ne() for this operator
  expect_no_lint("expect_true(x != y)", linter)
  # NB: also applies to tinytest, but it's sufficient to test testthat
  expect_no_lint("testthat::expect_true(x != y)", linter)

  # multiple comparisons are OK
  expect_no_lint("expect_true(x > y || x > z)", linter)

  # expect_gt() and friends don't have an info= argument
  expect_no_lint("expect_true(x > y, info = 'x is bigger than y yo')", linter)

  # expect_true() used incorrectly, and as executed the first argument is not a lint
  expect_no_lint("expect_true(is.count(n_draws), n_draws > 1)", linter)
})

test_that("expect_comparison_linter blocks simple disallowed usages", {
  linter <- expect_comparison_linter()

  expect_lint(
    "expect_true(x > y)",
    rex::rex("expect_gt(x, y) is better than expect_true(x > y)."),
    linter
  )

  # namespace qualification is irrelevant
  expect_lint(
    "testthat::expect_true(x < y)",
    rex::rex("expect_lt(x, y) is better than expect_true(x < y)."),
    linter
  )

  expect_lint(
    "expect_true(foo(x) >= y[[2]])",
    rex::rex("expect_gte(x, y) is better than expect_true(x >= y)."),
    linter
  )

  expect_lint(
    "expect_true(x <= y)",
    rex::rex("expect_lte(x, y) is better than expect_true(x <= y)."),
    linter
  )

  expect_lint(
    "expect_true(x == (y == 2))",
    rex::rex("expect_identical(x, y) is better than expect_true(x == y)."),
    linter
  )

  expect_lint(
    trim_some("
      expect_true(x # comment
      == (y == 2))
    "),
    rex::rex("expect_identical(x, y) is better than expect_true(x == y)."),
    expect_comparison_linter()
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      expect_true(x < y)
      expect_true(y > z)
    }"),
    list(
      list(rex::rex("expect_lt("), line_number = 2L),
      list(rex::rex("expect_gt("), line_number = 3L)
    ),
    expect_comparison_linter()
  )
})

test_that("if_not_else_linter skips allowed usages", {
  linter <- if_not_else_linter()

  # simple if/else statement is fine
  expect_no_lint("if (A) x else y", linter)
  # not plain negation --> OK
  expect_no_lint("if (!A || B) x else y", linter)
  # no else clause --> OK
  expect_no_lint("if (!A) x", linter)

  # nested statements are also OK
  expect_no_lint("if (!A) x else if (B) y", linter)
  expect_no_lint("if (!A) x else if (B) y else z", linter)
  expect_no_lint("if (A) x else if (B) y else if (!C) z", linter)

  # ! picked up in the evaluation statements is skipped
  expect_no_lint("if (A) !x else y", linter)
  expect_no_lint("if (A) x else !y", linter)
})

test_that("if_not_else_linter blocks simple disallowed usages", {
  linter <- if_not_else_linter()
  lint_msg <- rex::rex("Prefer `if (A) x else y`")

  expect_lint("if (!A) x else y", lint_msg, linter)
  expect_lint("if (!A) x else if (!B) y else z", lint_msg, linter)

  # ditto for more complicated expressions where ! is still the outer operator
  expect_lint("if (!x %in% 1:10) y else z", lint_msg, linter)
})

patrick::with_parameters_test_that(
  "if_not_else_linter blocks usages in ifelse() and friends as well",
  {
    linter <- if_not_else_linter()
    expect_no_lint(sprintf("%s(!A | B, x, y)", ifelse_fun), linter)
    expect_no_lint(sprintf("%s(A, !x, y)", ifelse_fun), linter)
    expect_lint(
      sprintf("%s(!A, x, y)", ifelse_fun),
      sprintf("Prefer `%s[(]A, x, y[)]` to the less-readable", ifelse_fun),
      linter
    )
    # particularly relevant for if_else()
    expect_no_lint(sprintf("%s(!!A, x, y)", ifelse_fun), linter)
  },
  .test_name = c("ifelse", "fifelse", "if_else"),
  ifelse_fun = c("ifelse", "fifelse", "if_else")
)

test_that("if_not_else_linter skips negated calls to is.null & similar", {
  linter <- if_not_else_linter()

  expect_no_lint("if (!is.null(x)) x else y", linter)
  expect_no_lint("if (!is.na(x)) x else y", linter)
  expect_no_lint("if (!missing(x)) x else y", linter)
  expect_no_lint("ifelse(!is.na(x), x, y)", linter)
})

test_that("multiple lints are generated correctly", {
  expect_lint(
    trim_some("{
        if (!A) x else B
        ifelse(!A, x, y)
        fifelse(!A, x, y)
        if_else(!A, x, y)
    }"),
    list(
      rex::rex("Prefer `if (A) x else y`"),
      "Prefer `ifelse",
      "Prefer `fifelse",
      "Prefer `if_else"
    ),
    if_not_else_linter()
  )
})

test_that("exceptions= argument works", {
  expect_lint(
    "if (!is.null(x)) x else y",
    rex::rex("Prefer `if (A) x else y`"),
    if_not_else_linter(exceptions = character())
  )

  expect_no_lint("if (!foo(x)) y else z", if_not_else_linter(exceptions = "foo"))
})

# fuzzer disable: assignment
test_that("assignment_linter skips allowed usages", {
  linter <- assignment_linter()

  expect_no_lint("blah", linter)
  expect_no_lint("blah <- 1", linter)
  expect_no_lint("blah<-1", linter)
  expect_no_lint("fun(blah=1)", linter)
})

test_that("assignment_linter blocks disallowed usages", {
  linter <- assignment_linter()
  lint_msg <- rex::rex("Use <- for assignment, not =.")

  expect_lint("blah=1", lint_msg, linter)
  expect_lint("blah = 1", lint_msg, linter)
  expect_lint("blah = fun(1)", lint_msg, linter)
  expect_lint("fun((blah = fun(1)))", lint_msg, linter)

  expect_lint(
    "blah = fun(1) {",
    list(
      lint_msg,
      c(type = "error", "unexpected")
    ),
    linter
  )
})

test_that("arguments handle <<- and ->/->> correctly", {
  linter <- assignment_linter()
  linter_yes_right <- assignment_linter(operator = c("->", "->>"))
  lint_msg_right <- rex::rex("Replace ->> by assigning to a specific environment")

  expect_lint("1 -> blah", rex::rex("Use <- for assignment, not ->."), linter)
  expect_lint("1 ->> blah", lint_msg_right, assignment_linter(operator = "<-"))

  # <<- can be unblocked optionally
  expect_lint(
    "1 <<- blah",
    rex::rex("Replace <<- by assigning to a specific environment"),
    linter
  )
  expect_no_lint("1 <<- blah", assignment_linter(operator = "<<-"))

  # blocking -> can be disabled
  expect_no_lint("1 -> blah", linter_yes_right)
  expect_no_lint("1 ->> blah", linter_yes_right)
  # we can also differentiate '->' and '->>'
  expect_lint(
    "1 ->> blah",
    lint_msg_right,
    assignment_linter(operator = c("<-", "->"))
  )

  # when user allows _some_ cascading assignment, advice should not mention the
  #   problems with cascading assignment, but focus on the specific disallowed operator.
  expect_lint(
    "1 ->> blah",
    rex::rex("Use one of <-, <<- for assignment, not ->>."),
    assignment_linter(operator = c("<-", "<<-"))
  )
  expect_lint(
    "blah <<- 1",
    rex::rex("Use one of ->, ->> for assignment, not <<-."),
    assignment_linter(operator = c("->", "->>"))
  )
})

# fuzzer disable: comment_injection
test_that("arguments handle trailing assignment operators correctly", {
  linter_default <- assignment_linter()
  linter_no_trailing <- assignment_linter(allow_trailing = FALSE)
  expect_no_lint("x <- y", linter_no_trailing)
  expect_no_lint("foo(bar = 1)", linter_no_trailing)

  expect_lint(
    trim_some("
      foo(bar =
        1)
    "),
    rex::rex("= should not be trailing at the end of a line."),
    linter_no_trailing
  )

  expect_lint(
    trim_some("
      x <-
        y
    "),
    rex::rex("<- should not be trailing"),
    linter_no_trailing
  )
  expect_lint(
    trim_some("
      x <<-
        y
    "),
    list(
      rex::rex("Replace <<- by assigning to a specific environment"),
      rex::rex("Assignment <<- should not be trailing")
    ),
    assignment_linter(operator = "<-", allow_trailing = FALSE)
  )

  expect_lint(
    trim_some("
      x <- #Test
        y
    "),
    rex::rex("<- should not be trailing"),
    linter_no_trailing
  )

  pipe_left_string <- trim_some("
    is_long <-
      is %>%
      gather(measure, value, -Species) %>%
      arrange(-value)
  ")
  expect_no_lint(pipe_left_string, linter_default)
  expect_lint(pipe_left_string, rex::rex("<- should not be trailing"), linter_no_trailing)

  pipe_right_string <- trim_some("
    is %>%
      gather(measure, value, -Species) %>%
      arrange(-value) ->
      is_long
  ")
  expect_lint(pipe_right_string, rex::rex("Use <- for assignment, not ->"), linter_default)
  expect_lint(
    pipe_right_string,
    list(
      rex::rex("Use <- for assignment, not ->"),
      rex::rex("Assignment -> should not be trailing")
    ),
    linter_no_trailing
  )
  expect_lint(
    pipe_right_string,
    rex::rex("-> should not be trailing"),
    assignment_linter(operator = "->", allow_trailing = FALSE)
  )

  expect_lint(
    trim_some("
      blah =
        42
      blh2 <-
        54
    "),
    list(
      list(message = "Use <- for assignment, not =.", line_number = 1L, column_number = 6L),
      list(message = "Assignment = should not be trailing at the end of a line", line_number = 1L, column_number = 6L),
      list(message = "Assignment <- should not be trailing at the end of a line", line_number = 3L, column_number = 6L)
    ),
    linter_no_trailing
  )

  expect_lint(
    trim_some("
      a =
        1
    "),
    "= should not be trailing",
    assignment_linter(operator = "=", allow_trailing = FALSE)
  )
})

test_that("allow_trailing interacts correctly with comments in braced expressions", {
  linter <- assignment_linter(allow_trailing = FALSE)
  expect_no_lint(
    trim_some("
    {
      x <- 1
      # blah
      y <- 2
    }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
    {
      x <- '#x'
      y <- '#y'
    }
    "),
    linter
  )

  expect_lint(
    trim_some("
    {
      x <- # blah
        'x'
    }
    "),
    list(message = "<-", line_number = 2L),
    linter
  )

  expect_no_lint(
    trim_some("
    {
      x <- '
        a string
        with an assignment <-
        at the end of the line
      '
    }
    "),
    linter
  )
})
# fuzzer enable: comment_injection

test_that("%<>% throws a lint", {
  expect_lint("x %<>% sum()", "Avoid the assignment pipe %<>%", assignment_linter())
  # regression test for #2850
  expect_lint("a <- 42", "Use %<>% for assignment", assignment_linter(operator = "%<>%"))

  expect_no_lint("x %<>% sum()", assignment_linter(operator = "%<>%"))

  # interaction with allow_trailing
  expect_lint(
    trim_some("
      x %<>%
        sum()
    "),
    list(
      "Avoid the assignment pipe %<>%",
      "Assignment %<>% should not be trailing"
    ),
    assignment_linter(allow_trailing = FALSE)
  )
})

test_that("multiple lints throw correct messages", {
  expect_lint(
    trim_some("{
      x <<- 1
      y ->> 2
      z -> 3
      x %<>% as.character()
    }"),
    list(
      list(message = "Replace <<- by assigning to a specific environment", line_number = 2L),
      list(message = "Replace ->> by assigning to a specific environment", line_number = 3L),
      list(message = "Use <- for assignment, not ->", line_number = 4L),
      list(message = "Avoid the assignment pipe %<>%", line_number = 5L)
    ),
    assignment_linter(operator = "<-")
  )
})

test_that("assignment operator can be toggled", {
  eq_linter <- assignment_linter(operator = "=")
  any_linter <- assignment_linter(operator = "any")
  lint_message <- rex("Use = for assignment, not")

  expect_no_lint("a = 1", eq_linter)
  expect_no_lint("a = 1", any_linter)

  expect_lint("a <- 1", lint_message, eq_linter)
  expect_no_lint("a <- 1", any_linter)

  expect_lint("a = 1; b <- 2", lint_message, eq_linter)
  expect_no_lint("a = 1; b <- 2", any_linter)

  expect_no_lint(
    trim_some("
      foo = function() {
        a = 1
      }
    "),
    eq_linter
  )
  expect_no_lint(
    trim_some("
      foo = function() {
        a = 1
      }
    "),
    any_linter
  )

  expect_lint(
    trim_some("
      foo = function() {
        a <- 1
      }
    "),
    list(lint_message, line_number = 2L),
    eq_linter
  )
  expect_no_lint(
    trim_some("
      foo = function() {
        a <- 1
      }
    "),
    any_linter
  )

  expect_no_lint("if ({a = TRUE}) 1", eq_linter)
  expect_no_lint("if ({a = TRUE}) 1", any_linter)

  expect_no_lint("if (a <- TRUE) 1", eq_linter)
  expect_no_lint("if (a <- TRUE) 1", any_linter)

  expect_no_lint("while ({a = TRUE}) 1", eq_linter)
  expect_no_lint("while ({a = TRUE}) 1", any_linter)

  expect_no_lint("while (a <- TRUE) 1", eq_linter)
  expect_no_lint("while (a <- TRUE) 1", any_linter)

  expect_no_lint("for (ii in {a = TRUE}) 1", eq_linter)
  expect_no_lint("for (ii in {a = TRUE}) 1", any_linter)

  expect_no_lint("for (ii in a <- TRUE) 1", eq_linter)
  expect_no_lint("for (ii in a <- TRUE) 1", any_linter)

  expect_lint(
    trim_some("
      x =
        2
      y <-
        3
    "),
    list(
      list("Assignment = should not be trailing", line_number = 1L),
      list("Assignment <- should not be trailing", line_number = 3L)
    ),
    assignment_linter(operator = "any", allow_trailing = FALSE)
  )
})

test_that("multiple lints throw correct messages when both = and <- are allowed", {
  expect_lint(
    trim_some("{
      x <<- 1
      y ->> 2
      z -> 3
      x %<>% as.character()
      foo <- 1
      bar = 2
    }"),
    list(
      list(message = "Replace <<- by assigning to a specific environment", line_number = 2L),
      list(message = "Replace ->> by assigning to a specific environment", line_number = 3L),
      list(message = "Use one of =, <- for assignment, not ->", line_number = 4L),
      list(message = "Avoid the assignment pipe %<>%", line_number = 5L)
    ),
    assignment_linter(operator = c("=", "<-"))
  )
})

test_that("multiple lints throw correct messages when = is required", {
  expect_lint(
    trim_some("{
      x <<- 1
      y ->> 2
      z -> 3
      x %<>% as.character()
      foo <- 1
      bar = 2
    }"),
    list(
      list(message = "Replace <<- by assigning to a specific environment", line_number = 2L),
      list(message = "Replace ->> by assigning to a specific environment", line_number = 3L),
      list(message = "Use = for assignment, not ->", line_number = 4L),
      list(message = "Avoid the assignment pipe %<>%", line_number = 5L),
      list(message = "Use = for assignment, not <-", line_number = 6L)
    ),
    assignment_linter(operator = "=")
  )
})

test_that("implicit '<-' assignments inside calls are ignored where top-level '<-' is disallowed", {
  linter <- assignment_linter(operator = "=")

  expect_no_lint("if (any(idx <- is.na(y))) which(idx)", linter)
  expect_no_lint("if (any(foo(idx <- is.na(y)))) which(idx)", linter)

  expect_no_lint("while (any(idx <- is.na(y))) break", linter)
  expect_no_lint("while (any(foo(idx <- is.na(y)))) break", linter)

  expect_no_lint("for (i in foo(idx <- is.na(y))) which(idx)", linter)
  expect_no_lint("for (i in foo(bar(idx <- is.na(y)))) which(idx)", linter)
})
# fuzzer enable: assignment

test_that("unreachable_code_linter works in simple function", {
  lines <- trim_some("
    foo <- function(bar) {
      return(bar)
    }
  ")
  expect_no_lint(lines, unreachable_code_linter())
})

test_that("unreachable_code_linter works in sub expressions", {
  linter <- unreachable_code_linter()
  msg <- rex::rex("Remove code and comments coming after return()")

  lines <- trim_some("
    foo <- function(bar) {
      if (bar) {
        return(bar)
        # Test comment
        while (bar) {
          return(bar)
          5 + 3
          repeat {
            return(bar)
            # Test comment
          }
        }
      } else if (bla) {
        # test
        return(5)
        # Test 2
      } else {
        return(bar)
        # Test comment
        for(i in 1:3) {
          return(bar)
          5 + 4
        }
      }
      return(bar)
      5 + 1
    }
  ")

  expect_lint(
    lines,
    list(
      list(line_number = 4L, message = msg),
      list(line_number = 7L, message = msg),
      list(line_number = 10L, message = msg),
      list(line_number = 16L, message = msg),
      list(line_number = 19L, message = msg),
      list(line_number = 22L, message = msg),
      list(line_number = 26L, message = msg)
    ),
    linter
  )

  expect_no_lint( # nofuzz: comment_injection
    trim_some("
      foo <- function(bar) {
        if (bar) {
          return(bar) # Test comment
        }
        while (bar) {
          return(bar) # 5 + 3
        }
        repeat {
          return(bar) # Test comment
        }

      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(bar) {
        if (bar) {
          return(bar); x <- 2
        } else {
          return(bar); x <- 3
        }
        while (bar) {
          return(bar); 5 + 3
        }
        repeat {
          return(bar); test()
        }
        for(i in 1:3) {
          return(bar); 5 + 4
        }
      }
    "),
    list(
      list(line_number = 3L, message = msg),
      list(line_number = 5L, message = msg),
      list(line_number = 8L, message = msg),
      list(line_number = 11L, message = msg),
      list(line_number = 14L, message = msg)
    ),
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(bar) {
        if (bar) {
          return(bar); # comment
          x <- 2
        } else {
          return(bar); # comment
          x <- 3
        }
        while (bar) {
          return(bar); # comment
          5 + 3
        }
        repeat {
          return(bar); # comment
          test()
        }
        for(i in 1:3) {
          return(bar); # comment
          5 + 4
        }
      }
    "),
    list(
      list(line_number = 4L, message = msg),
      list(line_number = 7L, message = msg),
      list(line_number = 11L, message = msg),
      list(line_number = 15L, message = msg),
      list(line_number = 19L, message = msg)
    ),
    linter
  )
})

test_that("unreachable_code_linter works with next and break in sub expressions", {
  linter <- unreachable_code_linter()
  next_msg <- rex::rex("Remove code and comments coming after `next`")
  break_msg <- rex::rex("Remove code and comments coming after `break`")

  lines <- trim_some("
    foo <- function(bar) {
      if (bar) {
        next
        # Test comment
        while (bar) {
          break
          5 + 3
          repeat {
            next
            # Test comment
          }
        }
      } else {
        next
        # test
        for(i in 1:3) {
          break
          5 + 4
        }
      }
    }
  ")

  expect_lint(
    lines,
    list(
      list(line_number = 4L, message = next_msg),
      list(line_number = 7L, message = break_msg),
      list(line_number = 10L, message = next_msg),
      list(line_number = 15L, message = next_msg),
      list(line_number = 18L, message = break_msg)
    ),
    linter
  )

  expect_no_lint( # nofuzz: comment_injection
    trim_some("
      foo <- function(bar) {
        if (bar) {
          break # Test comment
        } else {
          next # Test comment
        }
        while (bar) {
          next # 5 + 3
        }
        repeat {
          next # Test comment
        }
        for(i in 1:3) {
          break # 5 + 4
        }
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(bar) {
        if (bar) {
          next; x <- 2
        } else {
          break; x <- 3
        }
        while (bar) {
          break; 5 + 3
        }
        repeat {
          next; test()
        }
        for(i in 1:3) {
          break; 5 + 4
        }
      }
    "),
    list(
      list(line_number = 3L, message = next_msg),
      list(line_number = 5L, message = break_msg),
      list(line_number = 8L, message = break_msg),
      list(line_number = 11L, message = next_msg),
      list(line_number = 14L, message = break_msg)
    ),
    linter
  )

  # also with comments
  expect_lint(
    trim_some("
      foo <- function(bar) {
        if (bar) {
          next; # comment
          x <- 2
        } else {
          break; # comment
          x <- 3
        }
        while (bar) {
          break; # comment
          5 + 3
        }
        repeat {
          next; # comment
          test()
        }
        for(i in 1:3) {
          break; # comment
          5 + 4
        }
      }
    "),
    list(
      list(line_number = 4L, message = next_msg),
      list(line_number = 7L, message = break_msg),
      list(line_number = 11L, message = break_msg),
      list(line_number = 15L, message = next_msg),
      list(line_number = 19L, message = break_msg)
    ),
    linter
  )
})

test_that("unreachable_code_linter ignores expressions that aren't functions", {
  expect_no_lint("x + 1", unreachable_code_linter())
})

test_that("unreachable_code_linter ignores anonymous/inline functions", {
  expect_no_lint("lapply(rnorm(10), function(x) x + 1)", unreachable_code_linter())
})

test_that("unreachable_code_linter passes on multi-line functions", {
  lines <- trim_some("
    oo <- function(x) {
      y <- x + 1
      return(y)
    }
  ")
  expect_no_lint(lines, unreachable_code_linter())
})

test_that("unreachable_code_linter ignores comments on the same expression", { # nofuzz: comment_injection
  linter <- unreachable_code_linter()

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        return(
          y^2
        ) # y^3
      }
    "),
    linter
  )
})

test_that("unreachable_code_linter ignores comments on the same line", { # nofuzz: comment_injection
  lines <- trim_some("
    foo <- function(x) {
      return(y^2) # y^3
    }
  ")
  expect_no_lint(lines, unreachable_code_linter())
})

test_that("unreachable_code_linter identifies simple unreachable code", {
  lines <- trim_some("
    foo <- function(bar) {
      return(bar)
      x + 3
    }
  ")
  # testing the correct expression is linted (the first culprit line)
  expect_lint(
    lines,
    list(
      line_number = 3L,
      message = rex::rex("Remove code and comments coming after return()")
    ),
    unreachable_code_linter()
  )
})

test_that("unreachable_code_linter finds unreachable comments", {
  lines <- trim_some("
    foo <- function(x) {
      y <- x + 1
      return(y^2)
      # y^3
    }
  ")
  expect_lint(
    lines,
    rex::rex("Remove code and comments coming after return()"),
    unreachable_code_linter()
  )
})

test_that("unreachable_code_linter finds expressions in the same line", { # nofuzz: comment_injection
  msg <- rex::rex("Remove code and comments coming after return()")
  linter <- unreachable_code_linter()

  lines <- trim_some("
    foo <- function(x) {
      return(
        y^2
      ); 3 + 1
    }
  ")
  expect_lint(lines, msg, linter)

  lines <- trim_some("
    foo <- function(x) {
      return(y^2); 3 + 1
    }
  ")
  expect_lint(lines, msg, linter)

  lines <- trim_some("
    foo <- function(x) {
      return(y^2); 3 + 1 # Test
    }
  ")
  expect_lint(lines, msg, linter)
})

test_that("unreachable_code_linter finds expressions and comments after comment in return line", {
  msg <- rex::rex("Remove code and comments coming after return()")
  linter <- unreachable_code_linter()

  lines <- trim_some("
    foo <- function(x) {
      return(y^2) #Test comment
      #Test comment 2
    }
  ")
  expect_lint(lines, msg, linter)

  lines <- trim_some("
    foo <- function(x) {
      return(y^2) # Test
      3 + 1
    }
  ")
  expect_lint(lines, msg, linter)
})

test_that("unreachable_code_linter finds a double return", {
  lines <- trim_some("
    foo <- function(x) {
      return(y^2)
      return(y^3)
    }
  ")
  expect_lint(
    lines,
    rex::rex("Remove code and comments coming after return()"),
    unreachable_code_linter()
  )
})

test_that("unreachable_code_linter finds code after stop()", {
  lines <- trim_some("
    foo <- function(x) {
      y <- x + 1
      stop(y^2)
      # y^3
    }
  ")
  expect_lint(
    lines,
    rex::rex("Remove code and comments coming after stop()"),
    unreachable_code_linter()
  )
})

test_that("unreachable_code_linter ignores code after foo$stop(), which might be stopping a subprocess, for example", {
  linter <- unreachable_code_linter()

  expect_no_lint(
    trim_some("
      foo <- function(x) {
        bar <- get_process()
        bar$stop()
        TRUE
      }
    "),
    linter
  )
  expect_no_lint(
    trim_some("
      foo <- function(x) {
        bar <- get_process()
        bar@stop()
        TRUE
      }
    "),
    linter
  )
})

test_that("unreachable_code_linter ignores terminal nolint end comments", { # nofuzz: comment_injection
  linter <- unreachable_code_linter()

  withr::local_options(list(
    lintr.exclude_start = "#\\s*TestNoLintStart",
    lintr.exclude_end = "#\\s*TestNoLintEnd"
  ))

  expect_no_lint( # nofuzz
    trim_some("
      foo <- function() {
        do_something
        # TestNoLintStart: one_linter.
        a = 42
        return(a)
        # TestNoLintEnd
      }
    "),
    list(linter, one_linter = assignment_linter())
  )

  expect_no_lint(
    trim_some("
      foo <- function() {
        do_something
        # TestNoLintStart: one_linter.
        a = 42
        next
        # TestNoLintEnd
      }
    "),
    linter
  )
})

test_that("unreachable_code_linter identifies unreachable code in conditional loops", {
  linter <- unreachable_code_linter()
  msg <- rex::rex("Remove code inside a conditional loop with a deterministically false condition.")

  lines <- trim_some("
    foo <- function(bar) {
      if (FALSE) {
        x <- 3
      }
      x + 3
    }
  ")

  expect_lint(lines, list(line_number = 2L, message = msg), linter)

  lines <- trim_some("
    foo <- function(bar) {
      if (FALSE) {
        # Unlinted comment
        x <- 3
      }
      x + 3
    }
  ")

  expect_lint(lines, list(line_number = 2L, message = msg), linter)

  lines <- trim_some("
    foo <- function(bar) {
      if (bla) {
        x <- 3
      } else if (FALSE) {
        # Unlinted comment
        y <- 3
      }
      x + 3
    }
  ")

  expect_lint(lines, list(line_number = 4L, message = msg), linter)

  lines <- trim_some("
    foo <- function(bar) {
      while (FALSE) {
        x <- 3
      }
      x + 3
    }
  ")

  expect_lint(lines, list(line_number = 2L, message = msg), linter)

  lines <- trim_some("
    foo <- function(bar) {
      while (FALSE) {
        # Unlinted comment
        x <- 3
      }
      x + 3
    }
  ")

  expect_lint(lines, list(line_number = 2L, message = msg), linter)

  lines <- "while (FALSE) x <- 3"

  expect_lint(
    lines,
    list(line_number = 1L, ranges = list(c(1L, 20L)), message = msg),
    linter
  )

  lines <- "if (FALSE) x <- 3 # Test comment"

  expect_lint(
    lines,
    list(line_number = 1L, ranges = list(c(1L, 17L)), message = msg),
    linter
  )
})

test_that("unreachable_code_linter identifies unreachable code in conditional loops", {
  linter <- unreachable_code_linter()
  msg <- rex::rex("Remove code inside an else block after a deterministically true condition.")

  lines <- trim_some("
    foo <- function(bar) {
      if (TRUE) {
        x <- 3
      } else {
        # Unlinted comment
        x + 3
      }
    }
  ")

  expect_lint(lines, list(line_number = 6L, message = msg), linter)

  lines <- trim_some("
    foo <- function(bar) {
      if (TRUE) {
        x <- 3
      } else if (bar) {
        # Unlinted comment
        x + 3
      }
    }
  ")

  expect_lint(lines, list(line_number = 4L, message = msg), linter)

  expect_lint(
    "if (TRUE) x <- 3 else if (bar) x + 3",
    list(line_number = 1L, ranges = list(c(23L, 36L)), message = msg),
    linter
  )
})

test_that("unreachable_code_linter identifies unreachable code in mixed conditional loops", {
  linter <- unreachable_code_linter()
  false_msg <- rex::rex("Remove code inside a conditional loop with a deterministically false condition.")
  true_msg <- rex::rex("Remove code inside an else block after a deterministically true condition.")

  expect_lint(
    trim_some("
      function (bla) {
        if (FALSE) {
          code + 4
        }
        while (FALSE) {
          code == 3
        }
        if (TRUE) {
        } else {
          code + bla
        }
        stop('.')
        code <- 1
      }
    "),
    list(
      list(false_msg, line_number = 2L),
      list(false_msg, line_number = 5L),
      list(true_msg, line_number = 10L),
      list(rex::rex("Remove code and comments coming after stop()."), line_number = 13L)
    ),
    linter
  )

  expect_lint(
    "if (FALSE) x <- 3 else if (TRUE) x + 3 else x + 4",
    list(
      list(false_msg, line_number = 1L, ranges = list(c(1L, 49L))),
      list(
        rex::rex("Remove code inside an else block after a deterministically true condition."),
        line_number = 1L,
        ranges = list(c(45L, 49L))
      )
    ),
    linter
  )
})

test_that("function shorthand is handled", {
  expect_lint(
    trim_some("
      foo <- \\(bar) {
        return(bar)
        x + 3
      }
    "),
    list(
      line_number = 3L,
      message = rex::rex("Remove code and comments coming after return()")
    ),
    unreachable_code_linter()
  )
})

test_that("Do not lint inline else after stop", {
  expect_no_lint("if (x > 3L) stop() else x + 3", unreachable_code_linter())
})

test_that("Do not lint inline else after stop in inline function", {
  linter <- unreachable_code_linter()

  expect_no_lint("function(x) if (x > 3L) stop() else x + 3", linter)
  expect_no_lint("function(x) if (x > 3L) { stop() } else {x + 3}", linter)
})

test_that("Do not lint inline else after stop in inline lambda function", {
  linter <- unreachable_code_linter()

  expect_no_lint("\\(x) if (x > 3L) stop() else x + 3", linter)
  expect_no_lint("\\(x){ if (x > 3L) stop() else x + 3 }", linter)
})

test_that("allow_comment_regex= works", {
  withr::local_options(c(lintr.exclude_end = "#\\s*TestNoLintEnd"))

  linter_covr <- unreachable_code_linter()
  linter_xxxx <- unreachable_code_linter(allow_comment_regex = "#.*xxxx")
  linter_x1x2 <- unreachable_code_linter(allow_comment_regex = c("#x", "#y"))

  expect_no_lint(
    trim_some("
      function() {
        return(1)
        # nocov end
      }
    "),
    linter_covr
  )

  expect_no_lint(
    trim_some("
      function() {
        return(1)
        # TestNoLintEnd
        # nocov end
      }
    "),
    linter_covr
  )

  expect_no_lint(
    trim_some("
      function() {
        return(1)
        # ABCDxxxx
      }
    "),
    linter_xxxx
  )

  expect_no_lint(
    trim_some("
      function() {
        return(1)
        # TestNoLintEnd
        # ABCDxxxx
      }
    "),
    linter_xxxx
  )

  expect_no_lint(
    trim_some("
      function() {
        return(1)
        #x
      }
    "),
    linter_x1x2
  )

  expect_no_lint(
    trim_some("
      function() {
        return(1)
        #xABC
        #yDEF
      }
    "),
    linter_x1x2
  )

  # might contain capture groups, #2678
  expect_no_lint(
    trim_some("
      function() {
        stop('a')
        # a
        # ab
      }
    "),
    unreachable_code_linter(allow_comment_regex = "#\\s*(a|ab|abc)")
  )
})

test_that("allow_comment_regex= obeys covr's custom exclusion when set", {
  withr::local_options(c(
    lintr.exclude_end = "#\\s*TestNoLintEnd",
    covr.exclude_end = "#\\s*TestNoCovEnd"
  ))

  linter_covr <- unreachable_code_linter()

  expect_no_lint(
    trim_some("
      function() {
        return(1)
        # TestNoCovEnd
      }
    "),
    linter_covr
  )

  expect_no_lint(
    trim_some("
      function() {
        return(1)
        # TestNoLintEnd
        # TestNoCovEnd
      }
    "),
    linter_covr
  )
})

test_that("expect_s3_class_linter skips allowed usages", {
  linter <- expect_s3_class_linter()

  # expect_s3_class doesn't have an inverted version
  expect_no_lint("expect_true(!inherits(x, 'class'))", linter)
  # NB: also applies to tinytest, but it's sufficient to test testthat
  expect_no_lint("testthat::expect_true(!inherits(x, 'class'))", linter)

  # other is.<x> calls are not suitable for expect_s3_class in particular
  expect_no_lint("expect_true(is.na(x))", linter)

  # case where expect_s3_class() *could* be used but we don't enforce
  expect_no_lint("expect_true(is.data.table(x))", linter)

  # expect_s3_class() doesn't have info= or label= arguments
  expect_no_lint("expect_equal(class(x), k, info = 'x should have class k')", linter)
  expect_no_lint("expect_equal(class(x), k, label = 'x class')", linter)
  expect_no_lint("expect_equal(class(x), k, expected.label = 'target class')", linter)
  expect_no_lint("expect_true(is.data.frame(x), info = 'x should be a data.frame')", linter)
})

test_that("expect_s3_class_linter blocks simple disallowed usages", {
  linter <- expect_s3_class_linter()

  expect_lint(
    "expect_equal(class(x), 'data.frame')",
    rex::rex("expect_s3_class(x, k) is better than expect_equal(class(x), k)"),
    linter
  )

  # works when testing against a sequence of classes too
  expect_lint(
    "expect_equal(class(x), c('data.table', 'data.frame'))",
    rex::rex("expect_s3_class(x, k) is better than expect_equal(class(x), k)"),
    linter
  )

  # expect_identical is treated the same as expect_equal
  expect_lint(
    "testthat::expect_identical(class(x), 'lm')",
    rex::rex("expect_s3_class(x, k) is better than expect_identical(class(x), k)"),
    linter
  )

  # yoda test with string literal in first arg also caught
  expect_lint(
    "expect_equal('data.frame', class(x))",
    rex::rex("expect_s3_class(x, k) is better than expect_equal(class(x), k)"),
    linter
  )

  # different equivalent usages
  expect_lint(
    "expect_true(is.table(foo(x)))",
    rex::rex("expect_s3_class(x, k) is better than expect_true(is.<k>(x))"),
    linter
  )
  expect_lint(
    "expect_true(inherits(x, 'table'))",
    rex::rex("expect_s3_class(x, k) is better than expect_true(is.<k>(x))"),
    linter
  )
})

local({
  # test for lint errors appropriately raised for all is.<class> calls
  is_classes <- c(
    "data.frame", "factor", "numeric_version",
    "ordered", "package_version", "qr", "table",
    "relistable", "raster", "tclObj", "tkwin", "grob", "unit",
    "mts", "stepfun", "ts", "tskernel"
  )
  patrick::with_parameters_test_that(
    "expect_true(is.<base class>) is caught",
    expect_lint(
      sprintf("expect_true(is.%s(x))", is_class),
      rex::rex("expect_s3_class(x, k) is better than expect_true(is.<k>(x))"),
      expect_s3_class_linter()
    ),
    .test_name = is_classes,
    is_class = is_classes
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      expect_true(is.factor(x))
      expect_true(inherits(x, k))
      expect_equal(class(x), k)
    }"),
    list(
      list(rex::rex("is.<k>"), line_number = 2L),
      list(rex::rex("is.<k>"), line_number = 3L),
      list("expect_equal", line_number = 4L)
    ),
    expect_s3_class_linter()
  )
})

test_that("quotes_linter skips allowed usages", {
  linter <- quotes_linter()

  expect_no_lint("blah", linter)
  expect_no_lint('"blah"', linter)
  expect_no_lint('"\'blah"', linter)
  expect_no_lint('"blah\'"', linter)
  expect_no_lint('"\'blah\'"', linter)
  expect_no_lint("'\"'", linter)
  expect_no_lint("'\"blah\"'", linter)
})

test_that("quotes_linter blocks disallowed usages", {
  linter <- quotes_linter()
  lint_msg <- rex::rex("Only use double-quotes.")

  expect_lint("'blah'", lint_msg, linter)
  expect_lint("fun('blah')", lint_msg, linter)
  expect_lint("{'blah'}", lint_msg, linter)

  expect_lint(
    "
    x = 'test
    '",
    list(message = lint_msg, ranges = list(c(9L, 13L))),
    linter
  )
})

# NB: repeat of above tests with (mostly) opposite results
test_that("quotes_linter skips allowed usages of delimiter='", {
  linter <- quotes_linter(delimiter = "'")

  expect_no_lint("blah", linter)
  expect_no_lint('"\'blah"', linter)
  expect_no_lint('"blah\'"', linter)
  expect_no_lint('"\'blah\'"', linter)
  expect_no_lint("'\"'", linter)
  expect_no_lint("'\"blah\"'", linter)
  expect_no_lint("'blah'", linter)
  expect_no_lint("fun('blah')", linter)
  expect_no_lint("{'blah'}", linter)
})

test_that("quotes_linter blocks disallowed usages of delimiter='", {
  linter <- quotes_linter(delimiter = "'")
  lint_msg <- rex::rex("Only use single-quotes.")

  expect_lint('"blah"', lint_msg, linter)
  expect_lint(
    '
    x = "test
    "',
    list(message = lint_msg, ranges = list(c(9L, 13L))),
    linter
  )
})

test_that("raw strings are handled correctly", {
  linter <- quotes_linter()
  expect_no_lint('R"(hello \'\' there)"', linter)
  expect_lint("R'( whoops )'", rex::rex("Only use double-quotes."), linter)
  expect_lint("R'---[ daisy ]---'", rex::rex("Only use double-quotes."), linter)
  expect_no_lint("r'(\")'", linter)
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Only use double-quotes.")

  expect_lint(
    trim_some("{
      'abc'
      'def'
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    quotes_linter()
  )
})

test_that("if_switch_linter skips allowed usages", {
  linter <- if_switch_linter()

  # don't apply to simple if/else statements
  expect_no_lint("if (x == 'a') 1 else 2", linter)
  # don't apply to non-character conditions
  #   (NB: switch _could_ be used for integral input, but this
  #    interface is IMO a bit clunky / opaque)
  expect_no_lint("if (x == 1) 1 else 2", linter)
  # this also has a switch equivalent, but we don't both handling such
  #   complicated cases
  expect_no_lint("if (x == 'a') 1 else if (x != 'b') 2 else 3", linter)
  # multiple variables involved --> no clean change
  expect_no_lint("if (x == 'a') 1 else if (y == 'b') 2 else 3", linter)
  # multiple conditions --> no clean change
  expect_no_lint("if (is.character(x) && x == 'a') 1 else if (x == 'b') 2 else 3", linter)
  # simple cases with two conditions might be more natural
  #   without switch(); require at least three branches to trigger a lint
  expect_no_lint("if (x == 'a') 1 else if (x == 'b') 2", linter)
  # not thrown by raw strings
  expect_no_lint("if (x == 'a') 1 else if (x == R'(b)') 2", linter)
  # still no third if() clause
  expect_no_lint("if (x == 'a') 1 else if (x == 'b') 2 else 3", linter)
  # not thrown by raw strings
  expect_no_lint("if (x == 'a') 1 else if (x == R'(b)') 2 else 3", linter)

  # empty string comparisons can't use switch()
  expect_no_lint("if (x == '') 1 else if (x == 'a') 2 else if (x == 'b') 3", linter)
  expect_no_lint('if (x == "") 1 else if (x == "a") 2 else if (x == "b") 3', linter)
  expect_no_lint("if (x == 'a') 1 else if (x == '') 2 else if (x == 'b') 3", linter)
  expect_no_lint("if (x == 'a') 1 else if (x == 'b') 2 else if (x == '') 3", linter)
  expect_no_lint("if (x == 'a') 1 else if (x == 'b') 2 else if (x == 'c') 3 else if (x == '') 4", linter)
})

test_that("if_switch_linter handles raw empty strings", {
  linter <- if_switch_linter()

  # raw empty strings can't use switch() either
  expect_no_lint('if (x == R"--()--") 1 else if (x == "a") 2 else if (x == "b") 3', linter)
  expect_no_lint('if (x == R"()") 1 else if (x == R"{}") 2 else if (x == R"[]") 3', linter)
})

test_that("if_switch_linter blocks simple disallowed usages", {
  linter <- if_switch_linter()
  lint_msg <- rex::rex("Prefer switch() statements over repeated if/else equality tests")

  # anything with >= 2 equality statements is deemed switch()-worthy
  expect_lint("if (x == 'a') 1 else if (x == 'b') 2 else if (x == 'c') 3", lint_msg, linter)
  # regardless of raw strings
  expect_lint("if (x == 'a') 1 else if (x == r'(b)') 2 else if (x == R'--[c]--') 3", lint_msg, linter)
  # expressions are also OK
  expect_lint("if (foo(x) == 'a') 1 else if (foo(x) == 'b') 2 else if (foo(x) == 'c') 3", lint_msg, linter)
  # including when comments are present
  expect_lint(
    trim_some("
      if (foo(x) == 'a') 1 else if (foo(x # comment
      ) == 'b') 2 else if (foo(x) == 'c') 3
    "),
    lint_msg,
    linter
  )
})

test_that("if_switch_linter handles further nested if/else correctly", {
  linter <- if_switch_linter()

  # ensure that nested if() doesn't generate multiple lints;
  expect_lint(
    "if (x == 'a') 1 else if (x == 'b') 2 else if (x == 'c') 3 else if (x == 'd') 4",
    rex::rex("Prefer switch() statements over repeated if/else equality tests"),
    linter
  )
  # related to previous test -- if the first condition is non-`==`, the
  #   whole if/else chain is "tainted" / non-switch()-recommended.
  #   (technically, switch can work here, but the semantics are opaque)
  expect_no_lint(
    "if (x %in% c('a', 'e', 'f')) 1 else if (x == 'b') 2 else if (x == 'c') 3 else if (x == 'd') 4",
    linter
  )
})

test_that("multiple lints have right metadata", {
  lint_msg <- rex::rex("Prefer switch() statements over repeated if/else equality tests")

  expect_lint(
    trim_some("{
      if (x == 'a') {
        do_a()
      } else if (x == 'b') {
        do_b()
      } else if (x == 'c') {
        do_c()
      }
      if (y == 'A') {
        do_A()
      } else if (y == 'B') {
        do_B()
      } else if (y == 'C') {
        do_C()
      }
      if (z1 == 'a') {
        do_a()
      } else if (z2 == 'b') {
        do_b()
      } else if (z3 == 'c') {
        do_c()
      }
      if (a == '1') {
        do_1()
      } else if (a == '2') {
        do_2()
      } else if (a == '') {
        do_blank()
      }
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 9L)
    ),
    if_switch_linter()
  )
})

# fuzzer disable: comment_injection
test_that("max_branch_lines= and max_branch_expressions= arguments work", {
  max_lines2_linter <- if_switch_linter(max_branch_lines = 2L)
  max_lines4_linter <- if_switch_linter(max_branch_lines = 4L)
  max_expr2_linter <- if_switch_linter(max_branch_expressions = 2L)
  max_expr4_linter <- if_switch_linter(max_branch_expressions = 4L)
  lint_msg <- rex::rex("Prefer switch() statements over repeated if/else equality tests")

  one_per_branch_lines <- trim_some("
    if (x == 'a') {
      1
    } else if (x == 'b') {
      2
    } else if (x == 'c') {
      3
    }
  ")
  expect_lint(one_per_branch_lines, lint_msg, max_lines2_linter)
  expect_lint(one_per_branch_lines, lint_msg, max_lines4_linter)
  expect_lint(one_per_branch_lines, lint_msg, max_expr2_linter)
  expect_lint(one_per_branch_lines, lint_msg, max_expr4_linter)

  two_per_branch_lines <- trim_some("
    if (x == 'a') {
      1
      2
    } else if (x == 'b') {
      3
      4
    } else if (x == 'c') {
      5
      6
    }
  ")
  expect_lint(two_per_branch_lines, lint_msg, max_lines2_linter)
  expect_lint(two_per_branch_lines, lint_msg, max_lines4_linter)
  expect_lint(two_per_branch_lines, lint_msg, max_expr2_linter)
  expect_lint(two_per_branch_lines, lint_msg, max_expr4_linter)

  three_per_branch_lines <- trim_some("
    if (x == 'a') {
      1
      2
      3
    } else if (x == 'b') {
      4
      5
      6
    } else if (x == 'c') {
      7
      8
      9
    }
  ")
  expect_no_lint(three_per_branch_lines, max_lines2_linter)
  expect_lint(three_per_branch_lines, lint_msg, max_lines4_linter)
  expect_no_lint(three_per_branch_lines, max_expr2_linter)
  expect_lint(three_per_branch_lines, lint_msg, max_expr4_linter)

  five_per_branch_lines <- trim_some("
    if (x == 'a') {
      1
      2
      3
      4
      5
    } else if (x == 'b') {
      6
      7
      8
      9
      10
    } else if (x == 'c') {
      11
      12
      13
      14
      15
    }
  ")
  expect_no_lint(five_per_branch_lines, max_lines2_linter)
  expect_no_lint(five_per_branch_lines, max_lines4_linter)
  expect_no_lint(five_per_branch_lines, max_expr2_linter)
  expect_no_lint(five_per_branch_lines, max_expr4_linter)

  five_lines_three_expr_lines <- trim_some("
    if (x == 'a') {
      1
      2
      foo(
        x
      )
    } else if (x == 'b') {
      6
      7
      bar(
        y
      )
    } else if (x == 'c') {
      11
      12
      baz(
        z
      )
    }
  ")
  expect_no_lint(five_lines_three_expr_lines, max_lines2_linter)
  expect_no_lint(five_lines_three_expr_lines, max_lines4_linter)
  expect_no_lint(five_lines_three_expr_lines, max_expr2_linter)
  expect_lint(
    five_lines_three_expr_lines,
    list(lint_msg, line_number = 1L),
    max_expr4_linter
  )

  five_expr_three_lines_lines <- trim_some("
    if (x == 'a') {
      1
      2
      3; 4; 5
    } else if (x == 'b') {
      6
      7
      8; 9; 10
    } else if (x == 'c') {
      11
      12
      13; 14; 15
    }
  ")
  expect_no_lint(five_expr_three_lines_lines, max_lines2_linter)
  expect_lint(
    five_expr_three_lines_lines,
    list(lint_msg, line_number = 1L),
    max_lines4_linter
  )
  expect_no_lint(five_expr_three_lines_lines, max_expr2_linter)
  expect_no_lint(five_expr_three_lines_lines, max_expr4_linter)
})

test_that("max_branch_lines= and max_branch_expressions= block over-complex switch() too", {
  max_lines2_linter <- if_switch_linter(max_branch_lines = 2L)
  max_lines4_linter <- if_switch_linter(max_branch_lines = 4L)
  max_expr2_linter <- if_switch_linter(max_branch_expressions = 2L)
  max_expr4_linter <- if_switch_linter(max_branch_expressions = 4L)
  lint_msg <- rex::rex("Prefer repeated if/else statements over overly-complicated switch() statements.")

  one_per_branch_lines <- trim_some("
    switch(x,
      a = {
        1
      },
      b = {
        2
      },
      c = {
        3
      }
    )
  ")
  expect_no_lint(one_per_branch_lines, max_lines2_linter)
  expect_no_lint(one_per_branch_lines, max_lines4_linter)
  expect_no_lint(one_per_branch_lines, max_expr2_linter)
  expect_no_lint(one_per_branch_lines, max_expr4_linter)

  two_per_branch_lines <- trim_some("
    switch(x,
      a = {
        1
        2
      },
      b = {
        3
        4
      },
      c = {
        5
        6
      }
    )
  ")
  expect_no_lint(two_per_branch_lines, max_lines2_linter)
  expect_no_lint(two_per_branch_lines, max_lines4_linter)
  expect_no_lint(two_per_branch_lines, max_expr2_linter)
  expect_no_lint(two_per_branch_lines, max_expr4_linter)

  three_per_branch_lines <- trim_some("
    switch(x,
      a = {
        1
        2
        3
      },
      b = {
        4
        5
        6
      },
      c = {
        7
        8
        9
      }
    )
  ")
  expect_lint(
    three_per_branch_lines,
    list(lint_msg, line_number = 1L),
    max_lines2_linter
  )
  expect_no_lint(three_per_branch_lines, max_lines4_linter)
  expect_lint(
    three_per_branch_lines,
    list(lint_msg, line_number = 1L),
    max_expr2_linter
  )
  expect_no_lint(three_per_branch_lines, max_expr4_linter)

  five_per_branch_lines <- trim_some("
    switch(x,
      a = {
        1
        2
        3
        4
        5
      },
      b = {
        6
        7
        8
        9
        10
      },
      c = {
        11
        12
        13
        14
        15
      }
    )
  ")
  expect_lint(five_per_branch_lines, lint_msg, max_lines2_linter)
  expect_lint(five_per_branch_lines, lint_msg, max_lines4_linter)
  expect_lint(five_per_branch_lines, lint_msg, max_expr2_linter)
  expect_lint(five_per_branch_lines, lint_msg, max_expr4_linter)

  five_lines_three_expr_lines <- trim_some("
    switch(x,
      a = {
        1
        2
        foo(
          x
        )
      },
      b = {
        6
        7
        bar(
          y
        )
      },
      c = {
        11
        12
        baz(
          z
        )
      }
    )
  ")
  expect_lint(five_lines_three_expr_lines, lint_msg, max_lines2_linter)
  expect_lint(five_lines_three_expr_lines, lint_msg, max_lines4_linter)
  expect_lint(five_lines_three_expr_lines, lint_msg, max_expr2_linter)
  expect_no_lint(five_lines_three_expr_lines, max_expr4_linter)

  five_expr_three_lines_lines <- trim_some("
    switch(x,
      a = {
        1
        2
        3; 4; 5
      },
      b = {
        6
        7
        8; 9; 10
      },
      c = {
        11
        12
        13; 14; 15
      }
    )
  ")
  expect_lint(five_expr_three_lines_lines, lint_msg, max_lines2_linter)
  expect_no_lint(five_expr_three_lines_lines, max_lines4_linter)
  expect_lint(five_expr_three_lines_lines, lint_msg, max_expr2_linter)
  expect_lint(five_expr_three_lines_lines, lint_msg, max_expr4_linter)
})

test_that("max_branch_lines= and max_branch_expressions= interact correctly", {
  linter <- if_switch_linter(max_branch_lines = 5L, max_branch_expressions = 3L)
  lint_msg <- rex::rex("Prefer switch() statements over repeated if/else equality tests")

  expect_lint(
    trim_some("
      if (x == 'a') {
        1
      } else if (x == 'b') {
        2
      } else if (x == 'c') {
        3
      }
    "),
    lint_msg,
    linter
  )

  expect_no_lint(
    trim_some("
      if (x == 'a') {
        foo(
          x1,
          x2,
          x3,
          x4
        )
      } else if (x == 'b') {
        2
      } else if (x == 'c') {
        3
      }
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (x == 'a') {
        1; 2; 3; 4
      } else if (x == 'b') {
        5
      } else if (x == 'c') {
        6
      }
    "),
    linter
  )
})

test_that("max_branch_lines= and max_branch_expressions= work for a terminal 'else' branch", {
  max_lines2_linter <- if_switch_linter(max_branch_lines = 2L)
  max_expr2_linter <- if_switch_linter(max_branch_expressions = 2L)
  lint_msg <- rex::rex("Prefer repeated if/else statements over overly-complicated switch() statements.")

  else_long_lines <- trim_some("
    if (x == 'a') {
      1
    } else if (x == 'b') {
      2
    } else if (x == 'c') {
      3
    } else {
      4
      5
      6
    }
  ")
  expect_no_lint(else_long_lines, max_lines2_linter)
  expect_no_lint(else_long_lines, max_expr2_linter)

  default_long_lines <- trim_some("
    switch(x,
      a = {
        1
      },
      b = {
        2
      },
      c = {
        3
      },
      {
        4
        5
        6
      }
    )
  ")
  expect_lint(default_long_lines, lint_msg, max_lines2_linter)
  expect_lint(default_long_lines, lint_msg, max_expr2_linter)
})

test_that("max_branch_lines= and max_branch_expressions= are guided by the most complex branch", {
  max_lines2_linter <- if_switch_linter(max_branch_lines = 2L)
  max_expr2_linter <- if_switch_linter(max_branch_expressions = 2L)
  lint_msg <- rex::rex("Prefer repeated if/else statements over overly-complicated switch() statements.")

  # no lint if _any_ branch is too complex
  if_else_one_branch_lines <- trim_some("
    if (x == 'a') {
      1
    } else if (x == 'b') {
      2
    } else if (x == 'c') {
      3
      4
      5
    }
  ")
  expect_no_lint(if_else_one_branch_lines, max_lines2_linter)
  expect_no_lint(if_else_one_branch_lines, max_expr2_linter)

  # lint if _any_ branch is too complex
  switch_one_branch_lines <- trim_some("
    switch(x,
      a = {
        1
      },
      b = {
        2
      },
      c = {
        3
        4
        5
      }
    )
  ")
  expect_lint(switch_one_branch_lines, lint_msg, max_lines2_linter)
  expect_lint(switch_one_branch_lines, lint_msg, max_expr2_linter)
})

test_that("max_branch_lines= and max_branch_expressions= work for non-braced branches", {
  max_lines2_linter <- if_switch_linter(max_branch_lines = 2L)
  max_expr1_linter <- if_switch_linter(max_branch_expressions = 1L)
  lint_msg_if <- rex::rex("Prefer switch() statements over repeated if/else equality tests")
  lint_msg_switch <- rex::rex("Prefer repeated if/else statements over overly-complicated switch() statements.")

  # Multi-line non-braced branch in if-else (too complex for switch)
  if_multi_line_nobrace <- trim_some("
    if (x == 'a') foo(
      1,
      2
    ) else if (x == 'b') 2 else if (x == 'c') 3 else 4
  ")
  expect_no_lint(if_multi_line_nobrace, max_lines2_linter)

  # Multi-line non-braced branch in switch (too complex)
  switch_multi_line_nobrace <- trim_some("
    switch(x,
      a = foo(
        1,
        2
      ),
      b = 2,
      3
    )
  ")
  expect_lint(switch_multi_line_nobrace, lint_msg_switch, max_lines2_linter)

  # Multi-expression check on non-braced (always 1 expr, so should not exceed max_branch_expressions = 1)
  if_single_expr_nobrace <- "if (x == 'a') foo(1) else if (x == 'b') 2 else if (x == 'c') 3 else 4"
  expect_lint(if_single_expr_nobrace, lint_msg_if, max_expr1_linter)

  switch_single_expr_nobrace <- trim_some("
    switch(x,
      a = foo(1),
      b = 2,
      3
    )
  ")
  expect_no_lint(switch_single_expr_nobrace, max_expr1_linter)
})

# fuzzer enable: comment_injection

# fuzzer disable: comment_injection
test_that("semicolon_linter skips allowed usages", {
  linter <- semicolon_linter()

  expect_no_lint("", linter)
  expect_no_lint("a <- 1", linter)
  expect_no_lint("function() {a <- 1}", linter)
  expect_no_lint('a <- "foo;bar"', linter)
  expect_no_lint('function() {a <- "foo;bar"}', linter)
  expect_no_lint("a <- FALSE # ok; cool!", linter)
  expect_no_lint(
    trim_some("
      function() {
        a <- FALSE # ok; cool!
      }
    "),
    linter
  )
})

test_that("semicolon_linter handles trailing semicolons", {
  linter <- semicolon_linter()
  lint_msg <- rex::rex("Remove trailing semicolons.")

  expect_lint(
    "a <- 1;",
    list(lint_msg, line_number = 1L, column_number = 7L),
    linter
  )
  expect_lint(
    "function(){a <- 1;}",
    list(lint_msg, line_number = 1L, column_number = 18L),
    linter
  )
  expect_lint(
    trim_some("
      function() { a <- 1;
      }"),
    list(lint_msg, line_number = 1L, column_number = 20L),
    linter
  )
})

test_that("semicolon_linter handles compound semicolons", {
  linter <- semicolon_linter()
  lint_msg <- rex::rex("Replace compound semicolons by a newline.")

  expect_lint(
    "a <- 1;b <- 2",
    list(lint_msg, line_number = 1L, column_number = 7L),
    linter
  )
  expect_lint(
    "function() {a <- 1;b <- 2}",
    list(lint_msg, line_number = 1L, column_number = 19L),
    linter
  )
  expect_lint(
    trim_some("
      foo <-
         1 ; foo <- 1.23
    "),
    list(lint_msg, line_number = 2L, column_number = 6L),
    linter
  )
  expect_lint(
    trim_some("
      function() {
        foo <-
         1 ; foo <- 1.23
      }
    "),
    list(lint_msg, line_number = 3L, column_number = 6L),
    linter
  )
})

test_that("semicolon_linter handles multiple/mixed semicolons", {
  linter <- semicolon_linter()
  trail_msg <- rex::rex("Remove trailing semicolons.")
  comp_msg <- rex::rex("Replace compound semicolons by a newline.")

  expect_lint(
    trim_some("
      a <- 1 ; b <- 2;
      c <- 3;
    "),
    list(
      list(message = comp_msg, line_number = 1L, column_number = 8L),
      list(message = trail_msg, line_number = 1L, column_number = 16L),
      list(message = trail_msg, line_number = 2L, column_number = 7L)
    ),
    linter
  )
  expect_lint(
    trim_some("
      function() { a <- 1 ; b <- 2;
        c <- 3;}
    "),
    list(
      list(message = comp_msg, line_number = 1L, column_number = 21L),
      list(message = trail_msg, line_number = 1L, column_number = 29L),
      list(message = trail_msg, line_number = 2L, column_number = 9L)
    ),
    linter
  )
})


test_that("Compound semicolons only", {
  linter <- semicolon_linter(allow_trailing = TRUE)
  expect_no_lint("a <- 1;", linter)
  expect_no_lint("function(){a <- 1;}", linter)
  expect_no_lint(
    trim_some("
      function(){a <- 1;
      }
    "),
    linter
  )
})


test_that("Trailing semicolons only", {
  linter <- semicolon_linter(allow_compound = TRUE)
  expect_no_lint("a <- 1;b <- 2", linter)
  expect_no_lint("function() {a <- 1;b <- 2}", linter)
  expect_no_lint(
    trim_some("
      f <-
       1 ;f <- 1.23
    "),
    linter
  )
  expect_no_lint(
    trim_some("
      function(){
        f <-
          1 ;f <- 1.23
      }
    "),
    linter
  )
})


test_that("Compound semicolons only", {
  expect_error(
    semicolon_linter(allow_trailing = TRUE, allow_compound = TRUE),
    "At least one of `allow_compound` or `allow_trailing` must be `FALSE`",
    fixed = TRUE
  )
})
# fuzzer enable: comment_injection

test_that("equals_na_linter skips allowed usages", {
  linter <- equals_na_linter()

  expect_no_lint("blah", linter)
  expect_no_lint("  blah", linter)
  expect_no_lint("  blah", linter)
  expect_no_lint("x=NA", linter)
  expect_no_lint("x = NaN", linter)
  expect_no_lint("x = NA_real_", linter)
  expect_no_lint("is.na(x)", linter)
  expect_no_lint("is.nan(x)", linter)
  expect_no_lint("x[!is.na(x)]", linter)

  # equals_na_linter should ignore strings and comments
  expect_no_lint("is.na(x) # do not flag x == NA if inside a comment", linter)
  expect_no_lint("lint_msg <- 'do not flag x == NA if inside a string'", linter)

  # nested NAs are okay
  expect_no_lint("x==f(1, ignore = NA)", linter)

  # this should be covered by any_is_na_linter()
  expect_no_lint("NA %in% x", linter)
})

skip_if_not_installed("tibble")
patrick::with_parameters_test_that(
  "equals_na_linter blocks disallowed usages for all combinations of operators and types of NAs",
  expect_lint(
    paste("x", operation, type_na),
    rex::rex("Use is.na() instead of x ", operation, " NA"),
    equals_na_linter()
  ),
  .cases = tibble::tribble(
    ~.test_name,                 ~operation, ~type_na,
    "equality, logical NA",      "==",       "NA",
    "equality, integer NA",      "==",       "NA_integer_",
    "equality, real NA",         "==",       "NA_real_",
    "equality, complex NA",      "==",       "NA_complex_",
    "equality, character NA",    "==",       "NA_character_",
    "containment, logical NA",   "%in%",     "NA",
    "containment, integer NA",   "%in%",     "NA_integer_",
    "containment, real NA",      "%in%",     "NA_real_",
    "containment, complex NA",   "%in%",     "NA_complex_",
    "containment, character NA", "%in%",     "NA_character_",
    "inequality, logical NA",    "!=",       "NA",
    "inequality, integer NA",    "!=",       "NA_integer_",
    "inequality, real NA",       "!=",       "NA_real_",
    "inequality, complex NA",    "!=",       "NA_complex_",
    "inequality, character NA",  "!=",       "NA_character_"
  )
)

test_that("equals_na_linter blocks disallowed usages in edge cases", {
  linter <- equals_na_linter()
  lint_msg_part <- "Use is.na() instead of x "

  # missing spaces around operators
  expect_lint(
    "x==NA",
    list(message = rex::rex(lint_msg_part, "== NA"), line_number = 1L, column_number = 1L),
    linter
  )
  expect_lint(
    "x!=NA",
    list(message = rex::rex(lint_msg_part, "!= NA"), line_number = 1L, column_number = 1L),
    linter
  )

  # order doesn't matter
  expect_lint(
    "NA == x",
    list(message = rex::rex(lint_msg_part, "== NA"), line_number = 1L, column_number = 1L),
    linter
  )

  # correct line number for multiline code
  expect_lint(
    "x ==\nNA",
    list(line_number = 1L, column_number = 1L, ranges = list(c(1L, 4L))),
    linter
  )
})

test_that("linter returns correct linting", {
  linter <- undesirable_function_linter(fun = c(return = NA, log10 = "use log()"))
  msg_return <- rex::rex('Avoid undesirable function "return".', end)
  msg_log10 <- rex::rex('Avoid undesirable function "log10". As an alternative, use log().')

  expect_no_lint("x <- options()", linter)
  expect_no_lint("cat(\"Try to return\")", linter)
  expect_lint("lapply(x, log10)", list(message = msg_log10, line_number = 1L, column_number = 11L), linter)
  expect_lint("return()", list(message = msg_return, line_number = 1L, column_number = 1L), linter)
  expect_lint(
    trim_some("
      function(x) {
        print(options())
        y <- log10(x)
        return(y)
      }"),
    list(
      list(message = msg_log10, line_number = 3L, column_number = 8L),
      list(message = msg_return, line_number = 4L, column_number = 3L)
    ),
    linter
  )
  # regression test for #1050
  expect_no_lint("df$return <- 1", linter)
  expect_no_lint("df@return <- 1", linter)
})

test_that("it's possible to NOT lint symbols", {
  linter <- undesirable_function_linter(
    fun = c(dir = NA, log10 = "use log()"),
    symbol_is_undesirable = FALSE
  )
  expect_no_lint("dir <- 'path/to/a/directory'", linter)
  expect_no_lint("lapply(x, log10)", linter)
})

test_that("undesirable_function_linter doesn't lint library and require calls", {
  linter <- undesirable_function_linter(fun = c(foo = NA))
  expect_lint("test::foo()", "undesirable", linter)
  expect_no_lint("foo::test()", linter)
  expect_no_lint("library(foo)", linter)
  expect_no_lint("require(foo)", linter)

  linter <- undesirable_function_linter(fun = c(foo = NA, bar = NA))
  expect_no_lint("library(foo)", linter)

  linter <- undesirable_function_linter(fun = c(foo = NA, bar = NA), symbol_is_undesirable = FALSE)
  expect_no_lint("library(foo)", linter)
})

# regression test for #866
test_that("Line numbers are extracted correctly", {
  lines <- c(rep(letters, 10L), "tmp <- tempdir()")
  expect_lint(paste(lines, collapse = "\n"), "undesirable", undesirable_function_linter(c(tempdir = NA)))
})

test_that("invalid inputs fail correctly", {
  expect_error(
    undesirable_function_linter(fun = NULL),
    "`fun` must be a non-empty character vector",
    fixed = TRUE
  )
  expect_error(
    undesirable_function_linter(fun = character(0L)),
    "`fun` must be a non-empty character vector",
    fixed = TRUE
  )
  expect_error(
    undesirable_function_linter(c(NA, NA)),
    rex::rex("Unnamed elements of `fun` must not be missing", anything, "1", anything, "2")
  )

  expect_error(
    undesirable_function_linter(symbol_is_undesirable = 1.0),
    "symbol_is_undesirable` must be a logical",
    fixed = TRUE
  )
})

test_that("Default recommendations can be specified multiple ways", {
  linter_na <- undesirable_function_linter(c(foo = NA))
  linter_unnamed1 <- undesirable_function_linter("foo")
  linter_unnamed2 <- undesirable_function_linter(c("foo", "bar"))
  linter_mixed1 <- undesirable_function_linter(c("foo", bar = "no bar"))
  linter_mixed2 <- undesirable_function_linter(c("foo", bar = NA))

  lint_message <- rex::rex('Avoid undesirable function "foo"')

  lint_str <- "foo()"
  expect_lint(lint_str, lint_message, linter_na)
  expect_lint(lint_str, lint_message, linter_unnamed1)
  expect_lint(lint_str, lint_message, linter_unnamed2)
  expect_lint(lint_str, lint_message, linter_mixed1)
  expect_lint(lint_str, lint_message, linter_mixed2)
})

# fuzzer disable: assignment
test_that("function_return_linter skips allowed usages", {
  lines_simple <- trim_some("
    foo <- function(x) {
      x <- x + 1
      return(x)
    }
  ")
  expect_no_lint(lines_simple, function_return_linter())

  # arguably an expression as complicated as this should also be assigned,
  #   but for now that's out of the scope of this linter
  lines_subassignment <- trim_some("
    foo <- function(x) {
      return(x[, {
        col <- col + 1
        .(grp, col)
      }])
    }
  ")
  expect_no_lint(lines_subassignment, function_return_linter())
})

test_that("function_return_linter blocks simple disallowed usages", {
  linter <- function_return_linter()
  lint_msg <- rex::rex("Move the assignment outside of the return() clause")

  expect_lint(
    trim_some("
      foo <- function(x) {
        return(x <- x + 1)
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(x) {
        return(x <<- x + 1)
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(x) {
        return(x + 1 ->> x)
      }
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(x) {
        return(x + 1 -> x)
      }
    "),
    lint_msg,
    linter
  )

  side_effect_lines <- expect_lint(
    trim_some("
      e <- new.env()
      foo <- function(x) {
        return(e$val <- x + 1)
      }
    "),
    lint_msg,
    linter
  )
})

test_that("lints vectorize", {
  linter <- function_return_linter()
  lint_msg <- rex::rex("Move the assignment outside of the return() clause")

  expect_lint(
    trim_some("{
      function() {
        return(x <- 1)
      }
      function() {
        return(y <- 2)
      }
    }"),
    list(
      list(lint_msg, line_number = 3L),
      list(lint_msg, line_number = 6L)
    ),
    function_return_linter()
  )
})
# fuzzer enable: assignment

test_that("paste_linter skips allowed usages for sep=''", {
  linter <- paste_linter()

  expect_no_lint("paste('a', 'b', 'c')", linter)
  expect_no_lint("paste('a', 'b', 'c', sep = ',')", linter)
  expect_no_lint("paste('a', 'b', collapse = '')", linter)
  expect_no_lint("cat(paste('a', 'b'), sep = '')", linter)
  expect_no_lint("sep <- ''; paste('a', sep)", linter)
  expect_no_lint("paste(sep = ',', '', 'a')", linter)
  expect_no_lint("paste0('a', 'b', 'c')", linter)
  expect_no_lint("expression(2); paste('a', 'b', 'c', sep = ',')", linter)
  expect_no_lint("paste('a', 'b', expression(2), sep = ',')", linter)
  expect_no_lint("expression(paste('a', 'b'))", linter)
})

test_that("paste_linter blocks simple disallowed usages for sep=''", {
  linter <- paste_linter()
  paste_sep_msg <- rex::rex('paste0(...) is better than paste(..., sep = "").')
  expr_sep_msg <- rex::rex("inside expression(...), paste does not accept a 'sep' argument.")

  expect_lint("paste(sep = '', 'a', 'b')", paste_sep_msg, linter)
  expect_lint("paste('a', 'b', sep = '')", paste_sep_msg, linter)
  expect_lint("paste('a', 'b', expression(2), sep = '')", paste_sep_msg, linter)

  expect_lint("expression(paste('a', 'b', sep = ','))", expr_sep_msg, linter)
  expect_lint("expression(paste('a', 'b', sep = ''))", expr_sep_msg, linter)
  expect_lint("paste('a', 'b', expression(2), sep = '')", paste_sep_msg, linter)
  expect_lint("expression(italic(paste('a', sep = '')))", expr_sep_msg, linter)

  # Correct differentiation of lints in/out of expression()
  expect_lint(
    trim_some('{
      expression(paste(x, y, sep=""))
      paste(x, y, sep="")
      expression(paste(a, b, sep=""))
      paste(a, b, sep="")
      expression(c(paste(d, e, sep=""), paste(f, g), paste(h, i, sep="")))
    }'),
    list(
      list(expr_sep_msg, line_number = 2L),
      list(paste_sep_msg, line_number = 3L),
      list(expr_sep_msg, line_number = 4L),
      list(paste_sep_msg, line_number = 5L),
      list(expr_sep_msg, line_number = 6L, column_number = 16L),
      list(expr_sep_msg, line_number = 6L, column_number = 50L)
    ),
    linter
  )
})

test_that("paste_linter skips allowed usages for collapse=', '", {
  linter <- paste_linter()

  expect_no_lint("paste('a', 'b', 'c')", linter)
  expect_no_lint("paste(x, sep = ', ')", linter)
  expect_no_lint("paste(x, collapse = ',')", linter)
  expect_no_lint("paste(foo(x), collapse = '/')", linter)
  # harder to catch statically
  expect_no_lint("collapse <- ', '; paste(x, collapse = collapse)", linter)

  # paste(..., sep=sep, collapse=", ") is not a trivial swap to toString
  expect_no_lint("paste(x, y, sep = '.', collapse = ', ')", linter)
  # any call involving ...length() > 1 will implicitly use the default sep
  expect_no_lint("paste(x, y, collapse = ', ')", linter)
  expect_no_lint("paste0(x, y, collapse = ', ')", linter)

  expect_no_lint("toString(x)", linter)

  # string match of ", " is OK -- lint only _exact_ match
  expect_no_lint('paste(x, collapse = ", \n")', linter)
})

test_that("paste_linter blocks simple disallowed usages for collapse=', '", {
  expect_lint(
    "paste(collapse = ', ', x)",
    rex::rex('toString(.) is more expressive than paste(., collapse = ", ")'),
    paste_linter()
  )
})

test_that("paste_linter respects non-default arguments", {
  expect_no_lint("paste(sep = '', 'a', 'b')", paste_linter(allow_empty_sep = TRUE))
  expect_no_lint("paste('a', 'b', sep = '')", paste_linter(allow_empty_sep = TRUE))

  expect_no_lint("paste(collapse = ', ', x)", paste_linter(allow_to_string = TRUE))
})

test_that("paste_linter works for raw strings", {
  expect_no_lint("paste(a, b, sep = R'(xyz)')", paste_linter())
  expect_lint(
    'paste(a, b, sep = R"---[]---")',
    rex::rex('paste0(...) is better than paste(..., sep = "").'),
    paste_linter()
  )

  expect_no_lint("paste(x, collapse = R'(,,)')", paste_linter())
  expect_lint(
    'paste(c(a, b, c), collapse = R"-{, }-")',
    rex::rex('toString(.) is more expressive than paste(., collapse = ", ")'),
    paste_linter()
  )
})

test_that("paste_linter catches use of paste0 with sep=", {
  expect_lint(
    "paste0(x, y, sep = '')",
    rex::rex("sep= is not a formal argument to paste0();"),
    paste_linter()
  )
})

test_that("paste_linter skips allowed usages for strrep()", {
  linter <- paste_linter()

  expect_no_lint("paste(x, collapse = '')", linter)
  expect_no_lint("paste(rep('*', 10), collapse = '+')", linter)
  expect_no_lint("paste(rep(c('a', 'b'), 2), collapse = '')", linter)
  expect_no_lint("paste0(rep('a', 2), 'b', collapse = '')", linter)
  # no collapse
  expect_no_lint("paste(rep('*', 10))", linter)
  # combined before aggregating
  expect_no_lint("paste(rep('*', 10), rep('x', 10), collapse = '')", linter)
})

test_that("paste_linter blocks simple disallowed usages", {
  expect_lint(
    "paste(rep('#', width), collapse='')",
    rex::rex("strrep(x, times) is better than paste"),
    paste_linter()
  )
})

test_that("paste_linter skips allowed usages for file paths", {
  linter <- paste_linter()

  expect_no_lint("paste('a', 'b', 'c')", linter)
  expect_no_lint("paste('a', 'b', 'c', sep = ',')", linter)
  expect_no_lint("paste('a', 'b', collapse = '/')", linter)
  expect_no_lint("cat(paste('a', 'b'), sep = '/')", linter)
  expect_no_lint("sep <- '/'; paste('a', sep)", linter)
  expect_no_lint("paste(sep = ',', '/', 'a')", linter)

  # paste(..., sep='/', collapse=collapse) is not a trivial swap to file.path
  expect_no_lint("paste(x, y, sep = '/', collapse = ':')", linter)

  expect_no_lint("file.path('a', 'b', 'c')", linter)

  # testing the sep starts with / is not enough
  expect_no_lint("paste('a', 'b', sep = '//')", linter)
})

test_that("paste_linter blocks simple disallowed usages for file paths", {
  linter <- paste_linter()
  lint_msg <- rex::rex("Construct file paths with file.path(...) instead of")

  expect_lint("paste(sep = '/', 'a', 'b')", lint_msg, linter)
  expect_lint("paste('a', 'b', sep = '/')", lint_msg, linter)
})

test_that("paste_linter ignores non-path cases with paste0", {
  linter <- paste_linter()

  expect_no_lint("paste0(x, y)", linter)
  expect_no_lint("paste0('abc', 'def')", linter)
  expect_no_lint("paste0('/abc', 'def/')", linter)
  expect_no_lint("paste0(x, 'def/')", linter)
  expect_no_lint("paste0('/abc', y)", linter)
  expect_no_lint("paste0(foo(x), y)", linter)
  expect_no_lint("paste0(foo(x), 'def')", linter)

  # these might be a different lint (as.character instead, e.g.) but not here
  expect_no_lint("paste0(x)", linter)
  expect_no_lint("paste0('a')", linter)
  expect_no_lint("paste0('a', 1)", linter)
})

test_that("paste_linter detects paths built with '/' and paste0", {
  linter <- paste_linter()
  lint_msg <- rex::rex("Construct file paths with file.path(...) instead of")

  expect_lint("paste0(x, '/', y)", lint_msg, linter)
  expect_lint("paste0(x, '/', y, '/', z)", lint_msg, linter)
  expect_lint("paste0(x, '/abc/', 'def/', y)", lint_msg, linter)
  expect_lint("paste0(foo(x), '/abc/', 'def/', bar(y))", lint_msg, linter)
})

test_that("paste_linter skips initial/terminal '/' and repeated '/' for paths", {
  linter <- paste_linter()

  expect_no_lint("paste0('/', x)", linter)
  expect_no_lint("paste0(x, '/')", linter)
  expect_no_lint("paste0(x, '//hey/', y)", linter)
  expect_no_lint("paste0(x, '/hey//', y)", linter)
})

test_that("paste_linter doesn't skip all initial/terminal '/' for paths", {
  linter <- paste_linter()
  lint_msg <- rex::rex("Construct file paths with file.path(...) instead of")

  expect_lint('paste0("/abc/", "def")', lint_msg, linter)
  expect_lint('paste0("abc/", "def/")', lint_msg, linter)
})

test_that("multiple path lints are generated correctly", {
  linter <- paste_linter()

  expect_lint(
    trim_some("{
      paste(x, y, sep = '/')
      paste0(x, '/', y)
    }"),
    list(
      rex::rex('paste(..., sep = "/")'),
      rex::rex('paste0(x, "/", y, "/", z)')
    ),
    linter
  )

  # check vectorization of multiple paste0 file paths
  expect_lint(
    trim_some('{
      paste0(x, y)
      paste0("a/", x, y, "/b")
      paste0("a", "b")
      paste0("a", x)
      paste0(a, "x")
      paste0("a/", NA_character_, "/b")
      paste0("a/", "b")
      paste0("a/", "bcd//efg", "/h")
      paste0("/", a)
      paste0(a, "/")
    }'),
    list(message = rex::rex("Construct file paths with file.path(...)"), line_number = 8L),
    linter
  )
})

test_that("allow_file_path argument works", {
  expect_no_lint("paste(x, y, sep = '/')", paste_linter(allow_file_path = "always"))
})

test_that("URLs are ignored by default, linted optionally", {
  linter <- paste_linter()
  linter_url <- paste_linter(allow_file_path = "never")

  expect_no_lint("paste0('http://site.com/', x)", linter)
  expect_lint("paste0('http://site.com/', x)", rex::rex("Construct file paths with file.path(...)"), linter_url)
})

test_that("raw strings are detected in file path logic", {
  linter <- paste_linter()
  lint_msg <- rex::rex("Construct file paths with file.path(...) instead of ")

  expect_no_lint("paste0(x, R'{abc}', y)", linter)
  expect_lint("paste0(x, R'{/abc/}', y)", lint_msg, linter)
  expect_no_lint("paste(x, y, sep = R'{//}')", linter)
  expect_lint("paste(x, y, sep = R'{/}')", lint_msg, linter)
})

test_that("paste0(collapse=...) is caught", {
  linter <- paste_linter()
  lint_msg <- rex::rex("Use paste(), not paste0(), to collapse a character vector when sep= is not used.")

  expect_no_lint("paste(x, collapse = '')", linter)
  expect_no_lint("paste0(a, b, collapse = '')", linter)
  # pass-through can pass any number of arguments
  expect_no_lint("paste0(..., collapse = '')", linter)
  expect_lint("paste0(x, collapse = '')", lint_msg, linter)
  expect_lint("paste0(x, collapse = 'xxx')", lint_msg, linter)
  expect_lint("paste0(foo(x, y, z), collapse = '')", lint_msg, linter)
})

local({
  linter <- paste_linter()
  lint_msg <- rex::rex("Use paste(), not paste0(), to collapse a character vector when sep= is not used.")
  pipes <- pipes()

  patrick::with_parameters_test_that(
    "paste0(collapse=...) is caught in pipes",
    {
      expect_no_lint(sprintf('x %s paste0(y, collapse = "")', pipe), linter)
      expect_lint(sprintf('x %s paste0(collapse = "")', pipe), lint_msg, linter)
    },
    pipe = pipes,
    .test_name = pipes
  )
})

test_that("paste0(collapse=...) cases interacting with other rules are handled", {
  linter <- paste_linter()
  lint_msg <- rex::rex("Use paste(), not paste0(), to collapse a character vector when sep= is not used.")

  # multiple lints when collapse= happens to be ", "
  expect_lint(
    "paste0(foo(x), collapse = ', ')",
    list(rex::rex('toString(.) is more expressive than paste(., collapse = ", ")'), lint_msg),
    linter
  )
  expect_lint("paste0(foo(x), collapse = ', ')", lint_msg, paste_linter(allow_to_string = TRUE))

  expect_lint(
    "paste0(rep('*', 20L), collapse='')",
    list(rex::rex("strrep(x, times) is better than paste"), lint_msg),
    linter
  )

  # paste0(..., collapse=collapse) not directly mapped to file.path
  expect_lint("paste0(x, collapse = '/')", lint_msg, linter)
  expect_no_lint("paste0(x, y, collapse = '/')", linter)
})

test_that("paste_linter skips allowed deparse() usages", {
  linter <- paste_linter()

  expect_no_lint("deparse1(x)", linter)
  expect_no_lint("deparse(x)", linter)
  # no collapse=
  expect_no_lint("paste(deparse(x))", linter)
  # multiple positional args to paste
  expect_no_lint("paste('Error: ', deparse(x), collapse = ' ')", linter)
  # 'deparse' as a symbol, not a call
  expect_no_lint("paste(deparse, collapse = ' ')", linter)
  # deparse() supplies the separator, i.e. it is not the collapsed vector
  expect_no_lint("paste(x, collapse = deparse(sep))", linter)
  # collapse = NULL does not collapse, so it keeps deparse()'s multi-element output
  expect_no_lint("paste(deparse(x), collapse = NULL)", linter)
})

test_that("paste_linter recommends deparse1() over paste(deparse(), collapse=)", {
  linter <- paste_linter()
  lint_msg <- rex::rex("Use deparse1(x) instead of paste(deparse(x), collapse = ...).")

  expect_lint("paste(deparse(x), collapse = ' ')", lint_msg, linter)
  # nested inner call
  expect_lint("paste(deparse(substitute(x)), collapse = ' ')", lint_msg, linter)
  expect_lint("paste(deparse(x, width.cutoff = 500L), collapse = ' ')", lint_msg, linter)
  # arg order doesn't matter
  expect_lint("paste(collapse = ' ', deparse(x))", lint_msg, linter)
  # namespace-qualified
  expect_lint("base::paste(base::deparse(x), collapse = ' ')", lint_msg, linter)
})

test_that("deparse1 recommendation vectorizes and coexists with the paste0(collapse=) rule", {
  linter <- paste_linter()
  deparse1_msg <- rex::rex("Use deparse1(x) instead of paste(deparse(x), collapse = ...).")

  expect_lint(
    trim_some("{
      paste(deparse(x), collapse = ' ')
      paste(deparse(y), collapse = ' ')
    }"),
    list(
      list(deparse1_msg, line_number = 2L),
      list(deparse1_msg, line_number = 3L)
    ),
    linter
  )

  # paste0(deparse(x), collapse = '') also triggers the paste0()->paste() collapse rule;
  #   both are reported, consistent with the strrep/paste0-collapse interaction above.
  expect_lint(
    "paste0(deparse(x), collapse = '')",
    list(
      rex::rex("Use paste(), not paste0(), to collapse a character vector when sep= is not used."),
      deparse1_msg
    ),
    linter
  )
})

test_that("literal_coercion_linter skips allowed usages", {
  linter <- literal_coercion_linter()

  # naive xpath includes the "_f0" here as a literal
  expect_no_lint('as.numeric(x$"_f0")', linter)
  expect_no_lint('as.numeric(x@"_f0")', linter)
  # only examine the first method for as.<type> methods
  expect_no_lint("as.character(as.Date(x), '%Y%m%d')", linter)

  # we are as yet agnostic on whether to prefer literals over coerced vectors
  expect_no_lint("as.integer(c(1, 2, 3))", linter)
  # even more ambiguous for character vectors like here, where quotes are much
  #   more awkward to type than a sequence of numbers
  expect_no_lint("as.character(c(1, 2, 3))", linter)
  # not possible to declare raw literals
  expect_no_lint("as.raw(c(1, 2, 3))", linter)
  # also not taking a stand on as.complex(0) vs. 0 + 0i
  expect_no_lint("as.complex(0)", linter)
  # ditto for as.integer(1e6) vs. 1000000L
  expect_no_lint("as.integer(1e6)", linter)
  # ditto for as.numeric(1:3) vs. c(1, 2, 3)
  expect_no_lint("as.numeric(1:3)", linter)
})

test_that("literal_coercion_linter skips allowed rlang usages", {
  linter <- literal_coercion_linter()

  expect_no_lint("int(1, 2.0, 3)", linter)
  expect_no_lint("chr('e', 'ab', 'xyz')", linter)
  expect_no_lint("lgl(0, 1)", linter)
  expect_no_lint("lgl(0L, 1)", linter)
  expect_no_lint("dbl(1.2, 1e5, 3L, 2E4)", linter)
  # make sure using namespace (`rlang::`) doesn't create problems
  expect_no_lint("rlang::int(1, 2, 3)", linter)
  # even if scalar, carve out exceptions for the following
  expect_no_lint("int(1.0e6)", linter)
})

test_that("literal_coercion_linter skips quoted keyword arguments", {
  linter <- literal_coercion_linter()
  expect_no_lint("as.numeric(foo('a' = 1))", linter)
  expect_no_lint(
    trim_some("
      as.numeric(foo('a' # comment
      = 1))
    "),
    linter
  )
})

test_that("no warnings surfaced by running coercion", {
  linter <- literal_coercion_linter()

  expect_no_warning(
    expect_lint("as.integer('a')", "Use NA_integer_", linter)
  )

  expect_no_warning(
    expect_lint("as.integer(2147483648)", "Use NA_integer_", linter)
  )

  expect_no_warning(
    expect_lint(
      trim_some("
        as.double(
          NA # comment
        )
      "),
      "Use NA_real_",
      linter
    )
  )
})

skip_if_not_installed("tibble")
patrick::with_parameters_test_that(
  "literal_coercion_linter blocks simple disallowed usages",
  expect_lint(
    sprintf("as.%s(%s)", out_type, input),
    lint_msg,
    literal_coercion_linter()
  ),
  .cases = tibble::tribble(
    ~.test_name,     ~out_type,   ~input,   ~lint_msg,
    "lgl, from int", "logical",   "1L",     rex::rex("Use TRUE instead of as.logical(1L)"),
    "lgl, from num", "logical",   "1",      rex::rex("Use TRUE instead of as.logical(1)"),
    "lgl, from chr", "logical",   '"true"', rex::rex('Use TRUE instead of as.logical("true")'),
    "int, from num", "integer",   "1",      rex::rex("Use 1L instead of as.integer(1)"),
    "num, from num", "numeric",   "1",      rex::rex("Use 1 instead of as.numeric(1)"),
    "dbl, from num", "double",    "1",      rex::rex("Use 1 instead of as.double(1)"),
    "chr, from num", "character", "1",      rex::rex('Use "1" instead of as.character(1)'),
    "chr, from chr", "character", '"e"',    rex::rex('Use "e" instead of as.character("e")'),
    "chr, from chr", "character", '"E"',    rex::rex('Use "E" instead of as.character("E")'),
    # affirmatively lint as.<type>(NA) should be NA_<type>_
    "int, from NA",  "integer",   "NA",     rex::rex("Use NA_integer_ instead of as.integer(NA)"),
    "num, from NA",  "numeric",   "NA",     rex::rex("Use NA_real_ instead of as.numeric(NA)"),
    "dbl, from NA",  "double",    "NA",     rex::rex("Use NA_real_ instead of as.double(NA)"),
    "chr, from NA",  "character", "NA",     rex::rex("Use NA_character_ instead of as.character(NA)")
  )
)

skip_if_not_installed("rlang")
test_that("multiple lints return custom messages", {
  linter <- literal_coercion_linter()
  expect_lint(
    trim_some("{
      as.integer(1)
      lgl(1L)
    }"),
    list(
      list(rex::rex("Use 1L instead of as.integer(1)"), line_number = 2L),
      list(rex::rex("Use TRUE instead of lgl(1L)"), line_number = 3L)
    ),
    linter
  )

  # also ensure comment remove logic works across several lints
  expect_lint(
    trim_some("{
      as.integer( # comment
      1           # comment
      )           # comment
      lgl(        # comment
      1L          # comment
      )           # comment
    }"),
    list(
      list(rex::rex("Use 1L instead of as.integer(1)"), line_number = 2L),
      list(rex::rex("Use TRUE instead of lgl(1L)"), line_number = 5L)
    ),
    linter
  )
})

patrick::with_parameters_test_that(
  "literal_coercion_linter blocks rlang disallowed usages",
  expect_lint(
    sprintf("%s(%s)", out_type, input),
    lint_msg,
    literal_coercion_linter()
  ),
  # even if `as.character(1)` works, `chr(1)` doesn't, so no corresponding test case
  .cases = tibble::tribble(
    ~.test_name,      ~out_type,    ~input, ~lint_msg,
    "rlang::lgl",     "lgl",        "1L",   rex::rex("Use TRUE instead of lgl(1L)"),
    "rlang::lgl[ns]", "rlang::lgl", "1L",   rex::rex("Use TRUE instead of rlang::lgl(1L)"),
    "rlang::int",     "int",        "1.0",  rex::rex("Use 1L instead of int(1.0)"),
    "rlang::dbl",     "dbl",        "1L",   rex::rex("Use 1 instead of dbl(1L)"),
    "rlang::chr",     "chr",        '"e"',  rex::rex('Use "e" instead of chr("e")'),
    "rlang::chr",     "chr",        '"E"',  rex::rex('Use "E" instead of chr("E")')
  )
)

test_that("literal_coercion_linter blocks scalar rlang list2 construction", {
  expect_lint(
    "int(1, )",
    rex::rex("Use 1L instead of int(1,)"),
    literal_coercion_linter()
  )
})

test_that("expect_length_linter skips allowed usages", {
  linter <- expect_length_linter()

  expect_no_lint("expect_equal(nrow(x), 4L)", linter)
  # NB: also applies to tinytest, but it's sufficient to test testthat
  expect_no_lint("testthat::expect_equal(nrow(x), 4L)", linter)

  # only check the first argument. yoda tests in the second argument will be
  #   missed, but there are legitimate uses of length() in argument 2
  expect_no_lint("expect_equal(nrow(x), length(y))", linter)

  # expect_length() doesn't have info= or label= arguments
  expect_no_lint("expect_equal(length(x), n, info = 'x should have size n')", linter)
  expect_no_lint("expect_equal(length(x), n, label = 'x size')", linter)
  expect_no_lint("expect_equal(length(x), n, expected.label = 'target size')", linter)
})

test_that("expect_length_linter blocks simple disallowed usages", {
  linter <- expect_length_linter()
  lint_msg <- rex::rex("expect_length(x, n) is better than expect_equal(length(x), n)")

  expect_lint("expect_equal(length(x), 2L)", lint_msg, linter)
  expect_lint("testthat::expect_equal(length(DF), length(old))", lint_msg, linter)

  # yoda test cases
  expect_lint("expect_equal(2, length(x))", lint_msg, linter)
  expect_lint("expect_equal(2L, length(x))", lint_msg, linter)
})

test_that("expect_length_linter blocks expect_identical usage as well", {
  expect_lint(
    "expect_identical(length(x), 2L)",
    rex::rex("expect_length(x, n) is better than expect_identical(length(x), n)"),
    expect_length_linter()
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      expect_equal(length(x), n)
      expect_identical(length(x), n)
    }"),
    list(
      list("expect_equal", line_number = 2L),
      list("expect_identical", line_number = 3L)
    ),
    expect_length_linter()
  )
})

test_that("outer_negation_linter skips allowed usages", {
  linter <- outer_negation_linter()

  expect_no_lint("x <- any(y)", linter)
  expect_no_lint("y <- all(z)", linter)

  # extended usage of any is not covered
  expect_no_lint("any(!a & b)", linter)
  expect_no_lint("all(a | !b)", linter)

  expect_no_lint("any(a, b)", linter)
  expect_no_lint("all(b, c)", linter)
  expect_no_lint("any(!a, b)", linter)
  expect_no_lint("all(a, !b)", linter)
  expect_no_lint("any(a, !b, na.rm = TRUE)", linter)
  # ditto when na.rm is passed quoted
  expect_no_lint("any(a, !b, 'na.rm' = TRUE)", linter)
})

test_that("outer_negation_linter blocks simple disallowed usages", {
  linter <- outer_negation_linter()
  not_all_msg <- rex::rex("!all(x) is better than any(!x)")
  not_any_msg <- rex::rex("!any(x) is better than all(!x)")

  expect_lint("any(!x)", not_all_msg, linter)
  expect_lint("all(!foo(x))", not_any_msg, linter)
  # na.rm doesn't change the recommendation
  expect_lint("any(!x, na.rm = TRUE)", not_all_msg, linter)
  # also catch nested usage
  expect_lint("all(!(x + y))", not_any_msg, linter)
  # catch when all inputs are negated
  expect_lint("any(!x, !y)", not_all_msg, linter)
  expect_lint("all(!x, !y, na.rm = TRUE)", not_any_msg, linter)

  # adversarial comment
  expect_lint(
    trim_some("
      any(!x, na.rm = # comment
      TRUE)
    "),
    not_all_msg,
    linter
  )
})

test_that("outer_negation_linter doesn't trigger on empty calls", {
  linter <- outer_negation_linter()

  # minimal version of issue
  expect_no_lint("any()", linter)
  # closer to what was is practically relevant, as another regression test
  expect_no_lint("x %>% any()", linter)
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      any(!x)
      all(!y)
    }"),
    list(
      list(rex::rex("!all(x)"), line_number = 2L),
      list(rex::rex("!any(x)"), line_number = 3L)
    ),
    outer_negation_linter()
  )
})

test_that("redundant_ifelse_linter skips allowed usages", {
  linter <- redundant_ifelse_linter()

  expect_no_lint("ifelse(x > 5, 0, 2)", linter)
  expect_no_lint("ifelse(x > 5, TRUE, NA)", linter)
  expect_no_lint("ifelse(x > 5, FALSE, NA)", linter)
  expect_no_lint("ifelse(x > 5, TRUE, TRUE)", linter)

  expect_no_lint("ifelse(x > 5, 0L, 2L)", linter)
  expect_no_lint("ifelse(x > 5, 0L, 10L)", linter)
})

test_that("redundant_ifelse_linter blocks simple disallowed usages", {
  linter <- redundant_ifelse_linter()

  expect_lint(
    "ifelse(x > 5, TRUE, FALSE)",
    rex::rex("Just use the logical condition (or its negation) directly"),
    linter
  )
  expect_lint(
    "ifelse(x > 5, FALSE, TRUE)",
    rex::rex("Just use the logical condition (or its negation) directly"),
    linter
  )

  # other ifelse equivalents from common packages
  expect_lint(
    "if_else(x > 5, TRUE, FALSE)",
    rex::rex("Just use the logical condition (or its negation) directly"),
    linter
  )
  expect_lint(
    "fifelse(x > 5, FALSE, TRUE)",
    rex::rex("Just use the logical condition (or its negation) directly"),
    linter
  )
})

test_that("redundant_ifelse_linter blocks usages equivalent to as.numeric, optionally", {
  linter <- redundant_ifelse_linter()

  expect_lint(
    "ifelse(x > 5, 1L, 0L)",
    rex::rex("Prefer as.integer(x) to ifelse(x, 1L, 0L)"),
    linter
  )
  expect_lint(
    "ifelse(x > 5, 0L, 1L)",
    rex::rex("Prefer as.integer(!x) to ifelse(x, 0L, 1L)"),
    linter
  )

  expect_lint(
    "ifelse(x > 5, 1, 0)",
    rex::rex("Prefer as.numeric(x) to ifelse(x, 1, 0)"),
    linter
  )
  expect_lint(
    "ifelse(x > 5, 0, 1)",
    rex::rex("Prefer as.numeric(!x) to ifelse(x, 0, 1)"),
    linter
  )

  # mixing int and num
  expect_lint(
    "ifelse(x > 5, 0, 1L)",
    rex::rex("Prefer as.numeric(!x) to ifelse(x, 0, 1L)"),
    linter
  )
  expect_lint(
    "ifelse(x > 5, 0L, 1)",
    rex::rex("Prefer as.numeric(!x) to ifelse(x, 0L, 1)"),
    linter
  )
  expect_lint(
    "ifelse(x > 5, 1, 0L)",
    rex::rex("Prefer as.numeric(x) to ifelse(x, 1, 0L)"),
    linter
  )
  expect_lint(
    "ifelse(x > 5, 1L, 0)",
    rex::rex("Prefer as.numeric(x) to ifelse(x, 1L, 0)"),
    linter
  )

  # data.table/dplyr equivalents
  expect_lint(
    "dplyr::if_else(x > 5, 1L, 0L)",
    rex::rex("Prefer as.integer(x) to if_else(x, 1L, 0L)"),
    linter
  )
  expect_lint(
    "data.table::fifelse(x > 5, 0L, 1L)",
    rex::rex("Prefer as.integer(!x) to fifelse(x, 0L, 1L)"),
    linter
  )

  expect_lint(
    "if_else(x > 5, 1, 0)",
    rex::rex("Prefer as.numeric(x) to if_else(x, 1, 0)"),
    linter
  )
  expect_lint(
    "fifelse(x > 5, 0, 1)",
    rex::rex("Prefer as.numeric(!x) to fifelse(x, 0, 1)"),
    linter
  )
})

test_that("allow10 works as intended", {
  linter <- redundant_ifelse_linter(allow10 = TRUE)

  expect_no_lint("ifelse(x > 5, 1L, 0L)", linter)
  expect_no_lint("ifelse(x > 5, 0L, 1L)", linter)

  expect_no_lint("ifelse(x > 5, 1, 0)", linter)
  expect_no_lint("ifelse(x > 5, 0, 1)", linter)

  expect_no_lint("dplyr::if_else(x > 5, 1L, 0L)", linter)
  expect_no_lint("data.table::fifelse(x > 5, 0L, 1L)", linter)

  expect_no_lint("if_else(x > 5, 1, 0)", linter)
  expect_no_lint("fifelse(x > 5, 0, 1)", linter)
})

test_that("ifelse(missing = ) gives correct lints", {
  linter <- redundant_ifelse_linter()

  expect_lint("if_else(x > 5, TRUE, FALSE, NA)", rex::rex("Just use the logical condition"), linter)
  expect_no_lint("if_else(x > 5, TRUE, FALSE, TRUE)", linter)
  expect_no_lint("if_else(x > 5, TRUE, FALSE, 5L)", linter)

  expect_lint("if_else(x > 5, 1L, 0L, NA_integer_)", rex::rex("Prefer as.integer(x)"), linter)
  expect_no_lint("if_else(x > 5, 1L, 0L, 2L)", linter)
  expect_no_lint("if_else(x > 5, 1L, 0L, 5)", linter)

  expect_lint("if_else(x > 5, 1, 0, NA_real_)", rex::rex("Prefer as.numeric(x)"), linter)
  expect_no_lint("if_else(x > 5, 1, 0, 2)", linter)
  expect_no_lint("if_else(x > 5, 1, 0, '5')", linter)

  # TRUE/FALSE must be found in yes/no, not missing=
  expect_no_lint("if_else(x > 5, 'a', TRUE, FALSE)", linter)
  expect_no_lint("if_else(x > 5, 'a', 0L, 1L)", linter)
  expect_no_lint("if_else(x > 5, 'a', 1, 0)", linter)
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      ifelse(x > 0, TRUE, FALSE)
      fifelse(y == 0, 1, 0)
    }"),
    list(
      list("Just use the logical condition", line_number = 2L),
      list(rex::rex("refer as.numeric(x)"), line_number = 3L)
    ),
    redundant_ifelse_linter()
  )
})

test_that("expect_null_linter skips allowed usages", {
  linter <- expect_null_linter()

  # NB: this usage should be expect_false(), but this linter should
  #   nevertheless apply (e.g. for maintainers not using expect_not_linter)
  expect_no_lint("expect_true(!is.null(x))", linter)
  # NB: also applies to tinytest, but it's sufficient to test testthat
  expect_no_lint("testthat::expect_true(!is.null(x))", linter)

  # length-0 could be NULL, but could be integer() or list(), so let it pass
  expect_no_lint("expect_length(x, 0L)", linter)

  # no false positive for is.null() at the wrong positional argument
  expect_no_lint("expect_true(x, is.null(y))", linter)
})

test_that("expect_null_linter blocks simple disallowed usages", {
  linter <- expect_null_linter()

  expect_lint(
    "expect_equal(x, NULL)",
    rex::rex("expect_null(x) is better than expect_equal(x, NULL)"),
    linter
  )

  # expect_identical is treated the same as expect_equal
  expect_lint(
    "testthat::expect_identical(x, NULL)",
    rex::rex("expect_null(x) is better than expect_identical(x, NULL)"),
    linter
  )

  # reverse order lints the same
  expect_lint(
    "expect_equal(NULL, x)",
    rex::rex("expect_null(x) is better than expect_equal(x, NULL)"),
    linter
  )

  # different equivalent usage
  expect_lint(
    "expect_true(is.null(foo(x)))",
    rex::rex("expect_null(x) is better than expect_true(is.null(x))"),
    linter
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      expect_equal(x, NULL)
      expect_identical(x, NULL)
      expect_true(is.null(x))
    }"),
    list(
      list("expect_equal", line_number = 2L),
      list("expect_identical", line_number = 3L),
      list("expect_true", line_number = 4L)
    ),
    expect_null_linter()
  )
})

# fuzzer disable: comment_injection
test_that("returns the correct linting (with default parameters)", {
  linter <- commas_linter()
  msg_after <- rex::rex("Put a space after a comma.")
  msg_before <- rex::rex("Remove spaces before a comma.")

  expect_no_lint("blah", linter)
  expect_no_lint("fun(1, 1)", linter)
  expect_no_lint("fun(1,\n  1)", linter)
  expect_no_lint("fun(1,\n1)", linter)
  expect_no_lint("fun(1\n,\n1)", linter)
  expect_no_lint("fun(1\n  ,\n1)", linter)

  expect_lint("fun(1\n,1)", msg_after, linter)
  expect_lint("fun(1,1)", msg_after, linter)
  expect_lint("\nfun(1,1)", msg_after, linter)
  expect_lint("a(1,)", msg_after, linter)
  expect_lint("a[1,]", msg_after, linter)
  expect_lint("a[[1,]]", msg_after, linter)
  expect_lint(
    "fun(1 ,1)",
    list(
      list(msg_before, column_number = 6L),
      list(msg_after, column_number = 8L)
    ),
    linter
  )

  expect_no_lint('"fun(1 ,1)"', linter)
  expect_no_lint("a[1, , 2]", linter)
  expect_no_lint("a[1, , 2, , 3]", linter)

  expect_no_lint("switch(op, x = foo, y = bar)", linter)
  expect_no_lint("switch(op, x = , y = bar)", linter)
  expect_no_lint('switch(op, "x" = , y = bar)', linter)
  expect_no_lint("switch(op, x = ,\ny = bar)", linter)

  expect_lint("switch(op, x = foo , y = bar)", msg_before, linter)
  expect_lint("switch(op, x = foo , y = bar)", msg_before, linter)
  expect_lint("switch(op , x = foo, y = bar)", msg_before, linter)
  expect_lint("switch(op, x = foo, y = bar(a = 4 , b = 5))", msg_before, linter)
  expect_lint("fun(op, x = foo , y = switch(bar, a = 4, b = 5))", msg_before, linter)
  expect_lint(
    trim_some("
      switch(op ,
        x = foo,y = bar
      )
    "),
    list(
      list(msg_before, line_number = 1L),
      list(msg_after, line_number = 2L)
    ),
    linter
  )

  expect_lint(
    "fun(op    ,bar)",
    list(
      list(msg_before, column_number = 7L, ranges = list(c(7L, 10L))),
      list(msg_after, column_number = 12L, ranges = list(c(12L, 12L)))
    ),
    linter
  )
})

test_that("returns the correct linting (with 'allow_trailing' set)", {
  linter <- commas_linter(allow_trailing = TRUE)
  msg_after <- rex::rex("Put a space after a comma.")
  msg_before <- rex::rex("Remove spaces before a comma.")

  expect_no_lint("blah", linter)
  expect_no_lint("fun(1, 1)", linter)
  expect_no_lint("fun(1,\n  1)", linter)
  expect_no_lint("fun(1,\n1)", linter)
  expect_no_lint("fun(1\n,\n1)", linter)
  expect_no_lint("fun(1\n  ,\n1)", linter)
  expect_no_lint("a[1,]", linter)
  expect_no_lint("a(1,)", linter)

  expect_lint("fun(1\n,1)", msg_after, linter)
  expect_lint("fun(1,1)", msg_after, linter)
  expect_lint("\nfun(1,1)", msg_after, linter)
  expect_lint(
    "fun(1 ,1)",
    list(
      msg_before,
      msg_after
    ),
    linter
  )

  expect_no_lint('"fun(1 ,1)"', linter)
  expect_no_lint("a[1, , 2]", linter)
  expect_no_lint("a[1, , 2, , 3]", linter)
  expect_no_lint("a[[1,]]", linter)

  expect_no_lint("switch(op, x = foo, y = bar)", linter)
  expect_no_lint("switch(op, x = , y = bar)", linter)
  expect_no_lint('switch(op, "x" = , y = bar)', linter)
  expect_no_lint("switch(op, x = ,\ny = bar)", linter)

  expect_lint("switch(op, x = foo , y = bar)", msg_before, linter)
  expect_lint("switch(op, x = foo , y = bar)", msg_before, linter)
  expect_lint("switch(op , x = foo, y = bar)", msg_before, linter)
  expect_lint("switch(op, x = foo, y = bar(a = 4 , b = 5))", msg_before, linter)
  expect_lint("fun(op, x = foo , y = switch(bar, a = 4, b = 5))", msg_before, linter)

  expect_lint(
    "fun(op    ,bar)",
    list(
      list(msg_before, column_number = 7L, ranges = list(c(7L, 10L))),
      list(msg_after, column_number = 12L, ranges = list(c(12L, 12L)))
    ),
    linter
  )
})
# fuzzer enable: comment_injection

# fuzzer disable: comment_injection
test_that("spaces_inside_linter skips allowed usages", {
  linter <- spaces_inside_linter()

  expect_no_lint("blah", linter)
  expect_no_lint("print(blah)", linter)
  expect_no_lint("base::print(blah)", linter)
  expect_no_lint("a[, ]", linter)
  expect_no_lint("a[1]", linter)
  expect_no_lint("fun(\na[1]\n  )", linter)
  expect_no_lint("a(, )", linter)
  expect_no_lint("a(,)", linter)
  expect_no_lint("a(1)", linter)
  expect_no_lint('"a( 1 )"', linter)

  # trailing comments are OK (#636)
  expect_no_lint(
    trim_some("
      or( #code
        x, y
      )
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      fun(      # this is another comment
        a = 42, # because 42 is always the answer
        b = Inf
      )
    "),
    linter
  )
})

test_that("spaces_inside_linter blocks diallowed usages", {
  linter <- spaces_inside_linter()

  expect_lint(
    "a[1 ]",
    list(
      "Do not place spaces before square brackets",
      line_number = 1L,
      column_number = 4L,
      type = "style"
    ),
    linter
  )

  expect_lint(
    "a[[1 ]]",
    list(
      "Do not place spaces before square brackets",
      line_number = 1L,
      column_number = 5L,
      type = "style"
    ),
    linter
  )

  expect_lint(
    "\n\na[ 1]",
    list(
      "Do not place spaces after square brackets",
      line_number = 3L,
      column_number = 3L,
      type = "style"
    ),
    linter
  )

  expect_lint(
    "a[ 1 ]",
    list(
      list(
        "Do not place spaces after square brackets",
        line_number = 1L,
        column_number = 3L,
        type = "style"
      ),
      list(
        "Do not place spaces before square brackets",
        line_number = 1L,
        column_number = 5L,
        type = "style"
      )
    ),
    linter
  )

  expect_lint(
    "a(1 )",
    list(
      "Do not place spaces before parentheses",
      line_number = 1L,
      column_number = 4L,
      type = "style"
    ),
    linter
  )

  expect_lint(
    "a[[ 1]]",
    list(
      "Do not place spaces after square brackets",
      line_number = 1L,
      column_number = 4L,
      type = "style"
    ),
    linter
  )

  expect_lint(
    "a( 1)",
    list(
      "Do not place spaces after parentheses",
      line_number = 1L,
      column_number = 3L,
      type = "style"
    ),
    linter
  )

  expect_lint(
    "x[[ 1L ]]",
    list(
      list(
        "Do not place spaces after square brackets",
        line_number = 1L,
        column_number = 4L,
        type = "style"
      ),
      list(
        "Do not place spaces before square brackets",
        line_number = 1L,
        column_number = 7L,
        type = "style"
      )
    ),
    linter
  )

  expect_lint(
    "a( 1 )",
    list(
      list(
        "Do not place spaces after parentheses",
        line_number = 1L,
        column_number = 3L,
        type = "style"
      ),
      list(
        "Do not place spaces before parentheses",
        line_number = 1L,
        column_number = 5L,
        type = "style"
      )
    ),
    linter
  )

  # range covers all whitespace
  expect_lint(
    "a(  blah  )",
    list(
      list(
        "Do not place spaces after parentheses",
        line_number = 1L,
        column_number = 3L,
        ranges = list(c(3L, 4L)),
        type = "style"
      ),
      list(
        "Do not place spaces before parentheses",
        line_number = 1L,
        column_number = 9L,
        ranges = list(c(9L, 10L)),
        type = "style"
      )
    ),
    linter
  )
})

test_that("multi-line expressions have good markers", {
  expect_lint(
    trim_some("
      ( x |
        y )
    "),
    list(
      list("Do not place spaces after parentheses", line_number = 1L, ranges = list(c(2L, 2L))),
      list("Do not place spaces before parentheses", line_number = 2L, ranges = list(c(4L, 4L)))
    ),
    spaces_inside_linter()
  )
})

test_that("spaces_inside_linter blocks disallowed usages with a pipe", {
  linter <- spaces_inside_linter()

  expect_lint(
    "letters[1:3] %>% paste0( )",
    list(
      list(
        "Do not place spaces after parentheses",
        line_number = 1L,
        column_number = 25L,
        type = "style"
      ),
      list(
        "Do not place spaces before parentheses",
        line_number = 1L,
        column_number = 25L,
        type = "style"
      )
    ),
    linter
  )

  expect_lint(
    "letters[1:3] |> paste0( )",
    list(
      list(
        "Do not place spaces after parentheses",
        line_number = 1L,
        column_number = 24L,
        type = "style"
      ),
      list(
        "Do not place spaces before parentheses",
        line_number = 1L,
        column_number = 24L,
        type = "style"
      )
    ),
    linter
  )
})

test_that("terminal missing keyword arguments are OK", {
  expect_no_lint("alist(missing_arg = )", spaces_inside_linter())
})
# fuzzer enable: comment_injection

test_that("length_levels_linter skips allowed usages", {
  expect_no_lint("length(c(levels(x), 'a'))", length_levels_linter())
})

test_that("length_levels_linter blocks simple disallowed usages", {
  expect_lint(
    "2:length(levels(x))",
    rex::rex("nlevels(x) is better than length(levels(x))."),
    length_levels_linter()
  )
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("nlevels(x) is better than length(levels(x)).")

  expect_lint(
    trim_some("{
      length(levels(x))
      length(levels(y))
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    length_levels_linter()
  )
})

test_that("condition_message_linter skips allowed usages", {
  linter <- condition_message_linter()

  expect_no_lint("stop('a string', 'another')", linter)
  expect_no_lint("warning('a string', 'another')", linter)
  expect_no_lint("message('a string', 'another')", linter)
  # extracted calls likely don't obey base::stop() semantics
  expect_no_lint("ctx$stop(paste('a', 'b'))", linter)
  expect_no_lint("ctx@stop(paste('a', 'b'))", linter)

  # sprintf is OK -- gettextf() enforcement is left to other linters
  expect_no_lint("stop(sprintf('A %s!', 'string'))", linter)

  # get multiple sep= in one expression
  expect_no_lint(
    trim_some("
      tryCatch(
        foo(x),
        error = function(e) stop(paste(a, b, sep = '-')),
        warning = function(w) warning(paste(a, b, sep = '--')),
      )
    "),
    linter
  )
})

skip_if_not_installed("tibble")
patrick::with_parameters_test_that(
  "paste/paste0 allowed by condition_message_linter when using other seps and/or collapse",
  expect_no_lint(
    sprintf("%s(%s(x, %s = '%s'))", condition, fun, parameter, arg),
    condition_message_linter()
  ),
  .cases = tibble::tribble(
    ~.test_name,                           ~condition, ~fun,     ~parameter, ~arg,
    "stop, paste and collapse = ''",       "stop",     "paste",  "collapse",  "",
    "warning, paste and collapse = '\n'",  "warning",  "paste",  "collapse",  "\n",
    "message, paste and collapse = '|'",   "message",  "paste",  "collapse",  "|",
    "stop, paste0 and collapse = ''",      "stop",     "paste0", "collapse",  "",
    "warning, paste0 and collapse = '\n'", "warning",  "paste0", "collapse",  "\n",
    "message, paste0 and collapse = '|'",  "message",  "paste0", "collapse",  "|",
    "stop, paste and sep = '-'",           "stop",     "paste",  "sep",       "-",
    "warning, paste and sep = '\n'",       "warning",  "paste",  "sep",       "\n",
    "message, paste and sep = '|'",        "message",  "paste",  "sep",       "|"
  )
)

test_that("condition_message_linter blocks simple disallowed usages", {
  linter <- condition_message_linter()
  lint_msg_paste_stop <- rex::rex("Don't use paste to build stop strings.")

  expect_lint("stop(paste('a string', 'another'))", lint_msg_paste_stop, linter)

  expect_lint(
    "warning(paste0('a string ', 'another'))",
    rex::rex("Don't use paste0 to build warning strings."),
    linter
  )

  # `sep` argument allowed, but only if it is different from default
  expect_lint("stop(paste(x, sep = ' '))", lint_msg_paste_stop, linter)

  # not thrown off by named arguments
  expect_lint("stop(paste('a', 'b'), call. = FALSE)", lint_msg_paste_stop, linter)

  expect_lint(
    "warning(paste0('a', 'b'), immediate. = TRUE)",
    rex::rex("Don't use paste0 to build warning strings."),
    linter
  )

  expect_lint(
    trim_some("
      tryCatch(
        foo(x),
        error = function(e) stop(paste(a, b)),
        warning = function(w) warning(paste(a, b, sep = '--')),
      )
    "),
    lint_msg_paste_stop,
    linter
  )

  # one with no sep, one with linted sep
  expect_lint(
    trim_some("
      tryCatch(
        foo(x),
        error = function(e) stop(paste(a, b)),
        warning = function(w) warning(paste(a, b, sep = '')),
      )
    "),
    list(
      list(lint_msg_paste_stop, line_number = 3L),
      list(rex::rex("Don't use paste to build warning strings"), line_number = 4L)
    ),
    linter
  )
})

test_that("packageStartupMessage usages are also matched", {
  expect_lint(
    "packageStartupMessage(paste('a string', 'another'))",
    rex::rex("Don't use paste to build packageStartupMessage strings."),
    condition_message_linter()
  )

  expect_lint(
    "packageStartupMessage(paste0('a string ', 'another'))",
    rex::rex("Don't use paste0 to build packageStartupMessage strings."),
    condition_message_linter()
  )
})

test_that("raw strings are handled correctly", {
  linter <- condition_message_linter()
  lint_msg <- rex::rex("Don't use paste to build warning strings.")

  expect_lint("warning(paste(a, b, sep = R'( )'))", lint_msg, linter)
  expect_lint("warning(paste(a, b, sep = R'---[ ]---'))", lint_msg, linter)
})

test_that("message vectorization works", {
  expect_lint(
    trim_some("
      foo <- function() {
        warning(paste('uh oh!', 'spaghettios'))
        stop(paste0('look out ', 'below!'))
      }
    "),
    list(
      rex::rex("Don't use paste to build warning strings"),
      rex::rex("Don't use paste0 to build stop strings")
    ),
    condition_message_linter()
  )
})

# fuzzer disable: comment_injection
test_that("line_length_linter skips allowed usages", {
  linter <- line_length_linter(80L)

  expect_no_lint("blah", linter)
  expect_no_lint(strrep("x", 80L), linter)
})

test_that("line_length_linter blocks disallowed usages", {
  linter <- line_length_linter(80L)
  lint_msg <- rex::rex("Lines should not be more than 80 characters. This line is 81 characters.")

  expect_lint(
    strrep("x", 81L),
    list(
      message = lint_msg,
      column_number = 81L
    ),
    linter
  )

  expect_lint(
    paste(rep(strrep("x", 81L), 2L), collapse = "\n"),
    list(
      list(
        message = lint_msg,
        line_number = 1L,
        column_number = 81L
      ),
      list(
        message = lint_msg,
        line_number = 2L,
        column_number = 81L
      )
    ),
    linter
  )

  linter <- line_length_linter(20L)
  lint_msg <- rex::rex("Lines should not be more than 20 characters. This line is 22 characters.")
  expect_no_lint(strrep("a", 20L), linter)
  expect_lint(
    strrep("a", 22L),
    list(
      message = lint_msg,
      column_number = 21L
    ),
    linter
  )

  # Don't duplicate lints
  expect_length(
    lint(
      "x <- 2 # ------------\n",
      linters = linter,
      parse_settings = FALSE
    ),
    1L
  )
})

test_that("Multiple lints give custom messages", {
  expect_lint(
    trim_some("{
      abcdefg
      hijklmnop
    }"),
    list(
      list("9 characters", line_number = 2L),
      list("11 characters", line_number = 3L)
    ),
    line_length_linter(5L)
  )
})

test_that("string bodies can be ignored", {
  linter <- line_length_linter(10L, ignore_string_bodies = TRUE)
  lint_msg <- rex::rex("Lines should not be more than 10 characters. This line is 15 characters.")

  expect_no_lint(
    trim_some("
      1234567890
      str <- '
      123456789012345
      '
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      1234567890
      my_fun78('
      123456789012345
               '
      )
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      1234567890
      my_fun('90
      123456789012345
      123456789'
      )
    "),
    linter
  )

  expect_lint(
    trim_some("
      1234567890
      my_fun789('
      123456789012345
                '
      )
    "),
    list(
      list("11 characters", line_number = 2L),
      list("11 characters", line_number = 4L)
    ),
    linter
  )

  expect_lint(
    trim_some("
      1234567890
      my_fun('9012345
      1234567890
      123456789'
      )
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      1234567890
      my_fun('90
      1234567890
      12345678'); 234
    "),
    lint_msg,
    linter
  )

  expect_lint(
    "'1'; '2'; '345'",
    lint_msg,
    linter
  )

  expect_lint(
    "123456789012345",
    lint_msg,
    linter
  )

  expect_lint('"short" # 15!!!', lint_msg, linter)
  expect_lint('foo("a", long_)', lint_msg, linter)
})
# fuzzer enable: comment_injection

test_that("T_and_F_symbol_linter skips allowed usages", {
  linter <- T_and_F_symbol_linter()

  expect_no_lint("FALSE", linter)
  expect_no_lint("TRUE", linter)
  expect_no_lint('x <- "TRUE a vs FALSE b"', linter)
  expect_no_lint("y ~ T + F", linter)
  expect_no_lint("y ~ x + I(T^2)", linter)
  expect_no_lint("y ~ foo(T, F)", linter)
  expect_no_lint("T[1]", linter)
  expect_no_lint("T[[1]]", linter)
})

test_that("T_and_F_symbol_linter blocks disallowed usages", {
  linter <- T_and_F_symbol_linter()
  msg_true <- "Use TRUE instead of the symbol T."
  msg_false <- "Use FALSE instead of the symbol F."
  msg_variable_true <- "Don't use T as a variable name, as it can break code relying on T being TRUE."
  msg_variable_false <- "Don't use F as a variable name, as it can break code relying on F being FALSE."

  expect_lint("T", list(message = msg_true, line_number = 1L, column_number = 2L), linter)
  expect_lint("F", list(message = msg_false, line_number = 1L, column_number = 2L), linter)
  expect_lint("T = 42", list(message = msg_variable_true, line_number = 1L, column_number = 2L), linter)
  expect_lint("F = 42", list(message = msg_variable_false, line_number = 1L, column_number = 2L), linter)
  expect_lint(
    "for (i in 1:10) {x <- c(T, TRUE, F, FALSE)}",
    list(
      list(message = msg_true, line_number = 1L, column_number = 26L),
      list(message = msg_false, line_number = 1L, column_number = 35L)
    ),
    linter
  )

  expect_lint("DF$bool <- T", msg_true, linter)
  expect_lint("S4@bool <- T", msg_true, linter)
  expect_lint("sum(x, na.rm = T)", msg_true, linter)
  expect_lint("x[T]", msg_true, linter)
  expect_lint("x[, cols, drop = T]", msg_true, linter)
  expect_lint("y ~ foo(x, arg = T)", msg_true, linter)
  expect_lint(
    trim_some("
      y ~ foo(x, arg = #comment
      T)
    "),
    msg_true,
    linter
  )

  # Regression test for #657
  expect_lint(
    trim_some('
      x <- list(
        T = 42L,
        F = 21L
      )

      x$F <- 42L
      y@T <- 84L

      T <- "foo"
      F = "foo2"
      "foo3" -> T
    '),
    list(
      list(message = msg_variable_true, line_number = 9L),
      list(message = msg_variable_false, line_number = 10L),
      list(message = msg_variable_true, line_number = 11L)
    ),
    linter
  )
})

# styler: off
skip_if_not_installed("tibble")
local({
  # Note: cases indicated by "*" are whole numbers, but don't lint because the user has
  #   effectively declared "this is a double" much as adding '.0' is otherwise accepted.
  cases <- tibble::tribble(
    ~num_value_str,         ~lint_msg,
    "Inf",                  "",
    "NaN",                  "",
    "TRUE",                 "",
    "FALSE",                "",
    "NA",                   "",
    "NA_character_",        "",
    "2.000",                "",
    "2.",                   "",
    "2L",                   "",
    "2.0",                  "",
    "2.1",                  "",
    "2",                    "2L or 2.0",
    "1e3",                  "1000L or 1000.0",
    "1e3L",                 "",
    "1.0e3L",               "",
    "1.2e3",                "", # * ( = 1200)
    "1.2e-3",               "",
    "1e-3",                 "",
    "1e-33",                "",
    "1.2e0",                "",
    "0x1p+0",               "", # * ( = 1)
    "0x1.ecp+6L",           "",
    "0x1.ecp+6",            "", # * ( = 123)
    "0x1.ec66666666666p+6", "",
    "8i",                   "",
    "8.0i",                 ""
  )
  # for convenience of coercing these to string (since tribble doesn't support auto-conversion)
  int_max <- .Machine[["integer.max"]]  # largest number that R can represent as an integer
  cases_int_max <- tibble::tribble(
    ~num_value_str, ~lint_msg,
    -int_max - 1.0, "",
    -int_max,       sprintf("%1$dL or %1$d.0", -int_max),
    int_max,        sprintf("%1$dL or %1$d.0", int_max),
    int_max + 1.0,  ""
  )
  cases_int_max$num_value_str <- as.character(cases_int_max$num_value_str)
  cases <- rbind(cases, cases_int_max)

  linter <- implicit_integer_linter()
  patrick::with_parameters_test_that(
    "single numerical constants are properly identified ",
    expect_lint(num_value_str, if (nzchar(lint_msg)) lint_msg, linter),
    .cases = cases
  )
})
# styler: on

test_that("linter returns the correct linting", {
  linter <- implicit_integer_linter()
  lint_msg <- rex::rex("Use 1L or 1.0 to avoid implicit integers.")

  expect_no_lint("x <<- 1L", linter)
  expect_no_lint("1.0/-Inf -> y", linter)
  expect_lint(
    "y <- 1+i",
    list(message = lint_msg, line_number = 1L, column_number = 7L),
    linter
  )
  expect_lint(
    "z <- 1e5",
    list(message = rex::rex("100000L or 100000.0"), line_number = 1L, column_number = 9L),
    linter
  )
  expect_lint(
    "cat(1:n)",
    list(message = lint_msg, line_number = 1L, column_number = 6L),
    linter
  )
  expect_lint(
    "552^9",
    list(
      list(message = rex::rex("552L or 552.0"), line_number = 1L, column_number = 4L),
      list(message = rex::rex("9L or 9.0"), line_number = 1L, column_number = 6L)
    ),
    linter
  )
})

skip_if_not_installed("tibble")
local({
  .cases <- expand.grid(
    left = c("1", "-1", "1L", "-1L"),
    right = c("1", "-1", "1L", "-1L"),
    allow_colon = c(FALSE, TRUE),
    stringsAsFactors = FALSE
  )
  .cases <- within(.cases, {
    left_is_double <- !endsWith(left, "L")
    right_is_double <- !endsWith(right, "L")
    n_lints <- (!allow_colon) * (left_is_double + right_is_double)
    make_rex <- function(s) rex::rex("Use ", s, "L or ", s, ".0")
    left_replacement <- vapply(left, make_rex, character(1L))
    right_replacement <- vapply(right, make_rex, character(1L))
    rm(make_rex)
  })
  patrick::with_parameters_test_that(
    "Under allow_colon={allow_colon}, {left}:{right} throws {n_lints} lints",
    expect_lint(
      paste0(left, ":", right),
      if (n_lints > 0L) c(if (left_is_double) list(left_replacement), if (right_is_double) list(right_replacement)),
      implicit_integer_linter(allow_colon = allow_colon)
    ),
    .cases = .cases
  )
})

test_that("is_numeric_linter skips allowed usages involving ||", {
  linter <- is_numeric_linter()

  expect_no_lint("is.numeric(x) || is.integer(y)", linter)
  # x is used, but not identically
  expect_no_lint("is.numeric(x) || is.integer(foo(x))", linter)
  # not totally crazy, e.g. if input accepts a vector or a list
  expect_no_lint("is.numeric(x) || is.integer(x[[1]])", linter)
})

test_that("is_numeric_linter skips allowed usages involving %in%", {
  linter <- is_numeric_linter()

  # false positives for class(x) %in% c('integer', 'numeric') style
  expect_no_lint("class(x) %in% 1:10", linter)
  expect_no_lint("class(x) %in% 'numeric'", linter)
  expect_no_lint("class(x) %in% c('numeric', 'integer', 'factor')", linter)
  expect_no_lint("class(x) %in% c('numeric', 'integer', y)", linter)
})

test_that("is_numeric_linter blocks disallowed usages involving ||", {
  linter <- is_numeric_linter()
  lint_msg <- rex::rex("Use `is.numeric(x)` instead of the equivalent `is.numeric(x) || is.integer(x)`.")

  expect_lint("is.numeric(x) || is.integer(x)", lint_msg, linter)

  # order doesn't matter
  expect_lint("is.integer(x) || is.numeric(x)", lint_msg, linter)

  # identical expressions match too
  expect_lint( # nofuzz: dollar_at
    "is.integer(DT$x) || is.numeric(DT$x)",
    lint_msg,
    linter
  )

  # line breaks don't matter
  lines <- trim_some("
    if (
      is.integer(x)
      || is.numeric(x)
    ) TRUE
  ")
  expect_lint(lines, lint_msg, linter)

  # caught when nesting
  expect_lint("all(y > 5) && (is.integer(x) || is.numeric(x))", lint_msg, linter)

  # implicit nesting
  expect_lint("is.integer(x) || is.numeric(x) || is.logical(x)", lint_msg, linter)
})

test_that("is_numeric_linter blocks disallowed usages involving %in%", {
  linter <- is_numeric_linter()
  lint_msg <- rex::rex('Use is.numeric(x) instead of class(x) %in% c("integer", "numeric")')

  expect_lint("class(x) %in% c('integer', 'numeric')", lint_msg, linter)
  expect_lint('class(x) %in% c("numeric", "integer")', lint_msg, linter)
})

test_that("raw strings are handled properly when testing in class", {
  linter <- is_numeric_linter()
  lint_msg <- rex::rex('Use is.numeric(x) instead of class(x) %in% c("integer", "numeric")')

  expect_no_lint("class(x) %in% c(R'(numeric)', 'integer', 'factor')", linter)
  expect_no_lint("class(x) %in% c('numeric', R'--(integer)--', y)", linter)

  expect_lint("class(x) %in% c(R'(integer)', 'numeric')", lint_msg, linter)
  expect_lint('class(x) %in% c("numeric", R"--[integer]--")', lint_msg, linter)
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      is.numeric(x) || is.integer(x)
      class(x) %in% c('integer', 'numeric')
    }"),
    list(
      list(rex::rex("`is.numeric(x) || is.integer(x)`"), line_number = 2L),
      list(rex::rex('class(x) %in% c("integer", "numeric")'), line_number = 3L)
    ),
    is_numeric_linter()
  )
})

test_that("unnecessary_concatenation_linter skips allowed usages", {
  linter <- unnecessary_concatenation_linter()

  expect_no_lint("c(x)", linter)
  expect_no_lint("c(1, 2)", linter)
  expect_no_lint("c(x, recursive = TRUE)", linter)
  expect_no_lint("c(1, recursive = FALSE)", linter)
  expect_no_lint("lapply(1, c)", linter)
  expect_no_lint("c(a = 1)", linter)
  expect_no_lint("c('a' = 1)", linter)
})

test_that("unnecessary_concatenation_linter blocks disallowed usages", {
  linter <- unnecessary_concatenation_linter()
  msg_c <- rex::rex("Remove unnecessary c() of a constant.")
  msg_e <- rex::rex("Replace unnecessary c() by NULL or, whenever possible, vector()")

  expect_lint(
    "c()",
    list(message = msg_e, line_number = 1L, column_number = 1L),
    linter
  )
  expect_lint(
    "c(NULL)",
    list(message = msg_c, line_number = 1L, column_number = 1L),
    linter
  )
  expect_lint(
    "c(1)",
    list(message = msg_c, line_number = 1L, column_number = 1L),
    linter
  )
  expect_lint(
    "c (\n'a' )",
    list(message = msg_c, line_number = 1L, column_number = 1L),
    linter
  )
  expect_lint(
    "c(y, c('c('),\nc())",
    list(
      list(message = msg_c, line_number = 1L, column_number = 6L),
      list(message = msg_e, line_number = 2L, column_number = 1L)
    ),
    linter
  )
})

local({
  pipes <- pipes(exclude = "%$%")
  linter <- unnecessary_concatenation_linter()
  const_msg <- rex::rex("Remove unnecessary c() of a constant.")
  no_arg_msg <- rex::rex("Replace unnecessary c() by NULL or, whenever possible, vector()")

  patrick::with_parameters_test_that(
    "Correctly handles concatenation within magrittr pipes",
    {
      expect_no_lint(sprintf('"a" %s c("b")', pipe), linter)
      expect_lint(sprintf('"a" %s c()', pipe), const_msg, linter)
      expect_lint(sprintf('"a" %s list("b", c())', pipe), no_arg_msg, linter)
    },
    pipe = pipes,
    .test_name = names(pipes)
  )
})

test_that("logic survives adversarial comments", {
  expect_no_lint(
    trim_some('
      "a" %T>% # comment
        c("b")
    '),
    unnecessary_concatenation_linter()
  )
})

test_that("symbolic expressions are allowed, except by request", {
  linter <- unnecessary_concatenation_linter()
  linter_strict <- unnecessary_concatenation_linter(allow_single_expression = FALSE)
  lint_msg <- rex::rex("Remove unnecessary c() of a constant expression.")

  expect_no_lint("c(alpha / 2)", linter)
  expect_no_lint("c(paste0('.', 1:2))", linter)
  expect_no_lint("c(DF[cond > 1, col])", linter)

  # allow_single_expression = FALSE turns both into lints
  expect_lint("c(alpha / 2)", lint_msg, linter_strict)
  expect_lint("c(paste0('.', 1:2))", lint_msg, linter_strict)
  expect_lint("c(DF[cond > 1, col])", lint_msg, linter_strict)
})

test_that("sequences with : are linted whenever a constant is involved", {
  linter <- unnecessary_concatenation_linter()
  linter_strict <- unnecessary_concatenation_linter(allow_single_expression = FALSE)
  const_msg <- rex::rex("Remove unnecessary c() of a constant.")
  expr_msg <- rex::rex("Remove unnecessary c() of a constant expression.")

  expect_lint("c(1:10)", const_msg, linter)
  expect_lint("c(1:sum(x))", const_msg, linter)

  # this is slightly different if a,b are factors, in which case : does
  #   something like interaction
  expect_no_lint("c(a:b)", linter)
  expect_lint("c(a:b)", expr_msg, linter_strict)
  expect_no_lint("c(a:foo(b))", linter)
  expect_lint("c(a:foo(b))", expr_msg, linter_strict)
})

test_that("c(...) does not lint under !allow_single_expression", {
  expect_no_lint("c(...)", unnecessary_concatenation_linter(allow_single_expression = FALSE))
})

test_that("invalid allow_single_expression argument produce informative error messages", {
  expect_error(
    expect_no_lint("c()", unnecessary_concatenation_linter(allow_single_expression = 1.0)),
    rex::rex("`allow_single_expression` must be a single logical value.")
  )

  expect_error(
    expect_no_lint("c()", unnecessary_concatenation_linter(allow_single_expression = c(TRUE, FALSE))),
    rex::rex("`allow_single_expression` must be a single logical value.")
  )
})

test_that("one_call_pipe_linter skips allowed usages", {
  linter <- one_call_pipe_linter()

  # two pipe steps is OK
  expect_no_lint("x %>% foo() %>% bar()", linter)
  # call in first step --> OK
  expect_no_lint("foo(x) %>% bar()", linter)
  # both calls in second step --> OK
  expect_no_lint("x %>% foo(bar(.))", linter)

  # assignment pipe is exempted
  expect_no_lint("x %<>% as.character()", linter)
})

# fuzzer disable: pipe
test_that("one_call_pipe_linter blocks simple disallowed usages", {
  linter <- one_call_pipe_linter()
  lint_msg <- rex::rex("Avoid pipe %>% for expressions with only a single call.")

  expect_lint("x %>% foo()", lint_msg, linter)

  # new lines don't matter
  expect_lint("x %>%\n  foo()", lint_msg, linter)

  # nested case
  expect_lint("x %>% inner_join(y %>% filter(is_treatment))", lint_msg, linter)
})
# fuzzer enable: pipe

test_that("one_call_pipe_linter skips data.table chains", {
  linter <- one_call_pipe_linter()
  lint_msg <- rex::rex("Avoid pipe %>% for expressions with only a single call.")

  expect_no_lint("DT[x > 5, sum(y), by = keys] %>% .[, .SD[1], by = key1]", linter)

  # lint here: instead of a pipe, use DT[x > 5, sum(y), by = keys]
  expect_lint("DT %>% .[x > 5, sum(y), by = keys]", lint_msg, linter)

  # ditto for [[
  expect_no_lint("DT %>% rowSums() %>% .[[idx]]", linter)

  expect_lint("DT %>% .[[idx]]", lint_msg, linter)
})

test_that("one_call_pipe_linter treats all pipes equally", {
  linter <- one_call_pipe_linter()
  lint_msg_part <- " for expressions with only a single call."

  expect_no_lint("foo %>% bar() %$% col", linter)
  expect_lint("x %T>% foo()", rex::rex("%T>%", lint_msg_part), linter)
  expect_lint("x %$%\n  foo", rex::rex("%$%", lint_msg_part), linter)
  expect_no_lint('data %>% filter(type == "console") %$% obscured_id %>% unique()', linter)
})

test_that("multiple lints are generated correctly", { # nofuzz: pipe
  expect_lint(
    trim_some("{
      a %>% b()
      c %$% d
      e %T>%
        f()
    }"),
    list(
      list(rex::rex("pipe %>%"), line_number = 2L),
      list(rex::rex("pipe %$%"), line_number = 3L),
      list(rex::rex("pipe %T>%"), line_number = 4L)
    ),
    one_call_pipe_linter()
  )
})

test_that("Native pipes are handled as well", {
  linter <- one_call_pipe_linter()

  expect_lint( # nofuzz: pipe
    "x |> foo()",
    rex::rex("Avoid pipe |> for expressions with only a single call."),
    linter
  )

  # mixed pipes
  expect_no_lint("x |> foo() %>% bar()", linter)
  expect_no_lint("x %>% foo() |> bar()", linter)

  expect_lint( # nofuzz: pipe
    trim_some("{
      a %>% b()
      c |> d()
    }"),
    list(
      list(message = "pipe %>%"),
      list(message = "pipe |>")
    ),
    linter
  )
})

test_that("one_call_pipe_linter skips data.table chains with native pipe", {
  skip_unless_r(">= 4.3.0")

  linter <- one_call_pipe_linter()
  lint_msg <- rex::rex("Avoid pipe |> for expressions with only a single call.")

  expect_no_lint("DT[x > 5, sum(y), by = keys] |> _[, .SD[1], by = key1]", linter)

  # lint here: instead of a pipe, use DT[x > 5, sum(y), by = keys]
  expect_lint("DT |> _[x > 5, sum(y), by = keys]", lint_msg, linter)

  # ditto for [[
  expect_no_lint("DT |> rowSums() |> _[[idx]]", linter)

  expect_lint("DT |> _[[idx]]", lint_msg, linter)
})

test_that("repeat_linter works as expected", {
  linter <- repeat_linter()
  msg <- rex::rex("Use 'repeat' instead of 'while (TRUE)' for infinite loops.")

  expect_no_lint("repeat { }", linter)
  expect_no_lint("while (FALSE) { }", linter)
  expect_no_lint("while (i < 5) { }", linter)
  expect_no_lint("while (j < 5) TRUE", linter)
  expect_no_lint("while (TRUE && j < 5) { ... }", linter)

  expect_lint("while (TRUE) { }", msg, linter)
  expect_lint("for (i in 1:10) { while (TRUE) { if (i == 5) { break } } }", msg, linter)
  expect_lint("while (TRUE) { while (TRUE) { } }", list(msg, msg), linter)
  expect_lint(
    trim_some("{
      while (TRUE) {
      }
      while (TRUE) {
      }
    }"),
    list(
      list(msg, line_number = 2L, column_number = 3L, ranges = list(c(3L, 16L))),
      list(msg, line_number = 4L, column_number = 3L, ranges = list(c(3L, 16L)))
    ),
    linter
  )

  # fix for bad logic about range start/end
  expect_lint(
    trim_some("
                 while
      (TRUE) { }
    "),
    msg,
    linter
  )
})

test_that("yoda_test_linter skips allowed usages", {
  linter <- yoda_test_linter()

  expect_no_lint("expect_equal(x, 2)", linter)
  # namespace qualification doesn't matter
  expect_no_lint("testthat::expect_identical(x, 'a')", linter)
  # two variables can't be distinguished which is expected/actual (without
  #   playing quixotic games trying to parse that out from variable names)
  expect_no_lint("expect_equal(x, y)", linter)
})

test_that("yoda_test_linter blocks simple disallowed usages", {
  linter <- yoda_test_linter()
  lint_msg <- rex::rex("Compare objects in tests in the order 'actual', 'expected', not the reverse.")

  expect_lint("expect_equal(2, x)", lint_msg, linter)
  expect_lint("testthat::expect_identical('a', x)", lint_msg, linter)
  expect_lint("expect_setequal(2, x)", lint_msg, linter)
  # complex literals are slightly odd
  expect_lint("expect_equal(2 + 1i, x)", lint_msg, linter)
})

test_that("yoda_test_linter ignores strings in $ expressions", {
  linter <- yoda_test_linter()

  # the "key" here shows up at the same level of the parse tree as plain "key" normally would
  expect_no_lint('expect_equal(x$"key", 2)', linter)
  expect_no_lint('expect_equal(x@"key", 2)', linter)
})

# if we only inspect the first argument & ignore context, get false positives
local({
  pipes <- pipes(exclude = c("%<>%", "%$%"))
  linter <- yoda_test_linter()
  patrick::with_parameters_test_that(
    "yoda_test_linter ignores usage in pipelines",
    expect_no_lint(sprintf("foo() %s expect_identical(2)", pipe), linter),
    pipe = pipes,
    .test_name = names(pipes)
  )
})

test_that("yoda_test_linter throws a special message for placeholder tests", {
  expect_lint(
    "expect_equal(1, 1)",
    rex::rex("Avoid storing placeholder tests like expect_equal(1, 1)"),
    yoda_test_linter()
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      expect_equal(1, 1)
      expect_equal(2, foo(x))
    }"),
    list(
      list("Avoid storing placeholder tests", line_number = 2L),
      list("Compare objects in tests in the order 'actual', 'expected'", line_number = 3L)
    ),
    yoda_test_linter()
  )
})

# fuzzer disable: comment_injection
test_that("indentation linter flags unindented expressions", {
  linter <- indentation_linter(indent = 2L)

  expect_no_lint(
    trim_some("
      lapply(1:10, function(i) {
        i %% 2
      })
    "),
    linter
  )

  expect_lint(
    trim_some("
      lapply(1:10, function(i) {
       i %% 2  # indentation is only 1 character
      })
    "),
    "Indentation",
    linter
  )

  expect_lint(
    trim_some("
      lapply(1:10, function(i) {
       # indentation is only 1 character
        i %% 2
      })
    "),
    "Indentation",
    linter
  )

  # no double-block indents even if the indentation-starting tokens are immediately next to each other
  expect_lint(
    trim_some("
      local({
        # no lint
      })

      local({
          # must lint
      })
    "),
    list(line_number = 6L, message = "Indentation"),
    linter
  )

  expect_no_lint(
    trim_some("
      lapply(1:10, function(i) {
          i %% 2
      })
    "),
    indentation_linter(indent = 4L)
  )

  expect_lint(
    trim_some("
      lapply(1:10, function(i) {
        i %% 2  # indentation is only 2 characters
      })
    "),
    "Indentation",
    indentation_linter(indent = 4L)
  )

  # ugly code, but still correctly indented
  expect_no_lint(
    trim_some("
      list(
           1,
           2)
    "),
    linter
  )

  # comments do not trigger hanging indent rule
  expect_no_lint(
    trim_some("
      list( # comment
        ok
      )
    "),
    linter
  )

  # comments do not suppress block indents (#1751)
  expect_no_lint(
    trim_some("
      a <- # comment
        42L
    "),
    linter
  )

  # assignment triggers indent
  expect_no_lint(
    trim_some("
      a <-
        expr(
          42
        )
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      if (cond)
        code

      if (cond) code else code2

      if (cond) {
        code
      } else
        code

      if (cond) {
        code
      } else {
        code
      }
    "),
    linter
  )
})

test_that("indentation linter flags improper closing curly braces", {
  linter <- indentation_linter(indent = 2L)
  expect_no_lint(
    trim_some("
      lapply(1:10, function(i) {
        {
          i %% 2
        }
      })
    "),
    linter
  )

  expect_lint(
    trim_some("
      lapply(1:10, function(i) {
        i %% 2
        } # closing curly doesn't return to parent indentation
      )
    "),
    "Indentation",
    linter
  )
})

test_that("function argument indentation works in tidyverse-style", { # nofuzz: function_lambda
  linter <- indentation_linter()
  expect_no_lint(
    trim_some("
      function(a = 1L,
               b = 2L) {
        a + b
      }
    "),
    linter
  )

  # old double-indent style (#1754, #2830) lints under tidyverse style
  expect_lint(
    trim_some("
      function(
          a = 1L,
          b = 2L) {
        a + b
      }
    "),
    rex::rex("Indentation should be 2 spaces but is 4 spaces."),
    linter
  )

  expect_lint(
    trim_some("
      function(
            a = 1L,
            b = 2L) {
        a + b
      }
    "),
    rex::rex("Indentation should be 2 spaces but is 6 spaces."),
    linter
  )

  # Hanging is only allowed if there is an argument next to "("
  expect_lint(
    trim_some("
      function(
               a = 1L,
               b = 2L) {
        a + b
      }
    "),
    rex::rex("Indentation should be 2 spaces but is 9 spaces (or start argument on previous line)"),
    linter
  )

  expect_no_lint(
    trim_some("
      function(
        a = 1L,
        b = 2L
      ) {
        a + b
      }
    "),
    linter
  )

  # anchor is correctly found with assignments as well
  expect_no_lint( # nofuzz: assignment
    trim_some("
      test <- function(a = 1L,
                       b = 2L) {
        a + b
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      function(a = 1L,
         b = 2L) {
        a + b
      }
    "),
    "Hanging",
    linter
  )

  # This is a case for brace_linter
  expect_no_lint(
    trim_some("
      function(a = 1L,
               b = 2L)
      {
        a + b
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      function(
        a = function(
              x) {
          x
        }) {
        a()
      }
    "),
    list(rex::rex("Indentation should be 4 spaces but is 8 spaces."), line_number = 3L),
    linter
  )

  # proper single indentation of nested function definitions.
  expect_no_lint(
    trim_some("
      function(
        a = function(
          x
        ) {
          x
        }
      ) {
        a()
      }
    "),
    linter
  )
})

test_that("function argument indentation works in always-hanging-style", { # nofuzz: function_lambda
  linter <- indentation_linter(hanging_indent_style = "always")
  expect_no_lint(
    trim_some("
      function(a = 1L,
               b = 2L) {
        a + b
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      function(
          a = 1L,
          b = 2L) {
        a + b
      }
    "),
    "Hanging",
    linter
  )

  expect_no_lint(
    trim_some("
      function(
               a = 1L,
               b = 2L) {
        a + b
      }
    "),
    linter
  )

  # Block is only allowed if there is no argument next to ")"
  expect_lint(
    trim_some("
      function(
        a = 1L,
        b = 2L) {
        a + b
      }
    "),
    "Hanging",
    linter
  )

  expect_no_lint(
    trim_some("
      function(
        a = 1L,
        b = 2L
      ) {
        a + b
      }
    "),
    linter
  )

  # anchor is correctly found with assignments as well
  expect_no_lint( # nofuzz: assignment
    trim_some("
      test <- function(a = 1L,
                       b = 2L) {
        a + b
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      function(a = 1L,
         b = 2L) {
        a + b
      }
    "),
    "Hanging",
    linter
  )

  # This is a case for brace_linter
  expect_no_lint(
    trim_some("
      function(a = 1L,
               b = 2L)
      {
        a + b
      }
    "),
    linter
  )
})

test_that("indentation with operators works", {
  linter <- indentation_linter()
  expect_no_lint(
    trim_some("
      a %>%
        b()
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      (a + b + c) /
        (d + e + f) /
        (g + h + i)
    "),
    linter
  )

  expect_lint(
    trim_some("
      a %>%
          b()
    "),
    "Indentation",
    linter
  )

  expect_lint(
    trim_some("
      a +
       b()
    "),
    "Indentation",
    linter
  )

  expect_no_lint(
    trim_some("
      abc$
        def$
        ghi
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      abc@
        def@
        ghi
    "),
    linter
  )

  expect_lint(
    trim_some("
      first_step() +
                  second_step()
    "),
    list(rex::rex("Indentation should be 2 spaces but is 12 spaces."), line_number = 2L),
    linter
  )
})

test_that("indentation with bracket works", {
  linter <- indentation_linter()

  expect_no_lint(
    trim_some("
      dt[
        , col := 42L
      ][
        , ok
      ]

      bla[hanging,
          also_ok]
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      abc[[
        'elem'
      ]]

      def[[a,
           b]]
    "),
    linter
  )
})

test_that("indentation works with control flow statements", {
  linter <- indentation_linter()

  expect_no_lint(
    trim_some("
      if (TRUE) {
        do_something
      } else {
        do_other_thing
      }
    "),
    linter
  )

  expect_lint(
    trim_some("
      while (1 > 2) {
      do_something
      }
    "),
    "Indentation",
    linter
  )

  expect_lint(
    trim_some("
      if (FALSE) {
        do_something
        } else {
        do_other_thing
      }
    "),
    "Indentation",
    linter
  )
})

test_that("indentation lint messages are dynamic", {
  linter <- indentation_linter()

  expect_lint(
    trim_some("
      local({
          # should be 2
      })
    "),
    rex::rex("Indentation should be 2 spaces but is 4 spaces."),
    linter
  )

  expect_lint(
    trim_some("
      fun(
        3) # should be 4
    "),
    rex::rex("Hanging indent should be 4 spaces but is 2 spaces."),
    linter
  )
})

test_that("indentation within string constants is ignored", {
  linter <- indentation_linter()

  expect_no_lint(
    trim_some("
      x <- '
        an indented string
      '
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      x <- '
         an indented string with 3 spaces indentation
      '
    "),
    linter
  )

  # first line of a multi-line string can induce a lint
  expect_lint(
    trim_some("
      foo(
      '
        string
      ')
      bar('
        string2
      ')
      baz('
      string3
      ')
        x <- '
        string4
      '
    "),
    list(
      list(rex::rex("Hanging indent should be 4 spaces but is 0 spaces."), line_number = 2L),
      list(rex::rex("Indentation should be 0 spaces but is 2 spaces"), line_number = 11L)
    ),
    linter
  )
})

test_that("combined hanging and block indent works", {
  linter <- indentation_linter()

  expect_no_lint(
    trim_some("
      func(hang, and,
           block(
             combined
           ))
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      func(ha,
           func2(ab,
                 block(
                   indented
                 )))
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      func(func2(
        a = 42
      ))
    "),
    linter
  )

  # Adapted from cli R/ansi.R L231-234
  expect_no_lint(
    trim_some("
      stopifnot(is.character(style) && length(style) == 1 ||
                  is_rgb_matrix(style) && ncol(style) == 1,
                is.logical(bg) && length(bg) == 1,
                is.numeric(colors) && length(colors) == 1)
    "),
    linter
  )

  # Adapted from cli inst/scripts/up.R L26-37
  expect_no_lint(
    trim_some("
      http_head(url, ...)$
        then(function(res) {
          if (res$status_code < 300) {
            cli_alert_success()
          } else {
            cli_alert_danger()
          }
        })$
        catch(error = function(err) {
          e <- if (grepl('timed out', err$message)) 'timed out' else 'error'
          cli_alert_danger()
        })
    "),
    linter
  )

  # S4 equivalence
  expect_no_lint(
    trim_some("
      http_head(url, ...)@
        then(function(res) {
          if (res$status_code < 300) {
            cli_alert_success()
          } else {
            cli_alert_danger()
          }
        })@
        catch(error = function(err) {
          e <- if (grepl('timed out', err$message)) 'timed out' else 'error'
          cli_alert_danger()
        })
    "),
    linter
  )
})

test_that("hanging_indent_stlye works", {
  code_block_multi_line <- "map(x, f,\n  extra_arg = 42\n)"
  code_hanging_multi_line <- "map(x, f,\n    extra_arg = 42\n)"
  code_block_same_line <- "map(x, f,\n  extra_arg = 42)"
  code_hanging_same_line <- "map(x, f,\n    extra_arg = 42)"

  tidy_linter <- indentation_linter()
  hanging_linter <- indentation_linter(hanging_indent_style = "always")
  non_hanging_linter <- indentation_linter(hanging_indent_style = "never")

  expect_no_lint(code_block_multi_line, tidy_linter)
  expect_lint(code_block_multi_line, "Hanging indent", hanging_linter)
  expect_no_lint(code_block_multi_line, non_hanging_linter)

  expect_lint(code_hanging_multi_line, "Indent", tidy_linter)
  expect_no_lint(code_hanging_multi_line, hanging_linter)
  expect_lint(code_hanging_multi_line, "Indent", non_hanging_linter)

  expect_lint(code_block_same_line, "Hanging indent", tidy_linter)
  expect_lint(code_block_same_line, "Hanging indent", hanging_linter)
  expect_no_lint(code_block_same_line, non_hanging_linter)

  expect_no_lint(code_hanging_same_line, tidy_linter)
  expect_no_lint(code_hanging_same_line, hanging_linter)
  expect_lint(code_hanging_same_line, "Indent", non_hanging_linter)

  # regression test for #1898
  expect_no_lint(
    trim_some("
      outer_fun(inner_fun(x,
        one_indent = 42L
      ))
    "),
    tidy_linter
  )

  expect_no_lint(
    trim_some("
      outer_fun(inner_fun(x, # this is first arg
        one_indent = 42L # this is second arg
      ))
    "),
    tidy_linter
  )

  expect_no_lint(
    trim_some("
      outer_fun(inner_fun(
        x,
        one_indent = 42L
      ))
    "),
    tidy_linter
  )

  expect_no_lint(
    trim_some("
      outer_fun(
        inner_fun(
          x,
          one_indent = 42L
        )
      )
    "),
    tidy_linter
  )

  expect_lint(
    trim_some("
      outer_f(
        inner_g(x,
        y = 2)
      )
    "),
    list(rex::rex("Hanging indent should be 10 spaces but is 2 spaces."), line_number = 3L),
    tidy_linter
  )

  # don't suggest starting argument on previous line
  expect_lint(
    trim_some("
      outer_f(
        inner_g(x,
              y = 2)
      )
    "),
    list(rex::rex("Hanging indent should be 10 spaces but is 8 spaces."), line_number = 3L),
    tidy_linter
  )
})

test_that("previous token is respected when recommending to 'start argument on previous line'", {
  expect_lint(
    trim_some("
      result <- {
                 do_something()
      }
      abc[[
           'elem'
      ]]
      def[
          'key'
      ]
      first_step() +
                    second_step()
    "),
    list(
      list(rex::rex("should be 2 spaces but is 11 spaces."), line_number = 2L),
      list(rex::rex("should be 2 spaces but is 5 spaces (or start argument on previous line)."), line_number = 5L),
      list(rex::rex("should be 2 spaces but is 4 spaces (or start argument on previous line)."), line_number = 8L),
      list(rex::rex("should be 2 spaces but is 14 spaces."), line_number = 11L)
    ),
    indentation_linter()
  )
})

test_that("assignment_as_infix works", {
  # test function call restorator and LEFT_ASSIGN suppressor
  code_infix <- trim_some("
    ok_code <-
      var1 +
      f(
        var2 +
          var3
      ) +
      var4
  ")

  # test that innermost ancestor token decides the indentation
  code_infix_2 <- trim_some("
    lapply(x,
      function(e) {
        temp_var <-
          e +
          42
      }
    )
  ")

  # test brace restorator
  code_infix_3 <- trim_some("
    ok_code <-
      if (condition) {
        a +
          b
      } else {
        c +
          d
      } +
      e
  ")

  # test EQ_ASSIGN, EQ_SUB and EQ_FORMALS suppressors
  code_infix_4 <- trim_some("
    # EQ_ASSIGN
    ok_code =
      a +
      b

    # EQ_SUB
    f(
      a =
        b +
        c
    )

    # EQ_FORMALS
    f <- function(
      a =
        b +
        c
    ) {
      NULL
    }
  ")

  code_no_infix <- trim_some("
    ok_code <-
      var1 +
        f(
          var2 +
            var3
        ) +
        var4
  ")

  tidy_linter <- indentation_linter()
  no_infix_linter <- indentation_linter(assignment_as_infix = FALSE)

  expect_no_lint(code_infix, tidy_linter)
  expect_no_lint(code_infix_2, tidy_linter)
  expect_no_lint(code_infix_3, tidy_linter)
  expect_no_lint(code_infix_4, tidy_linter)
  expect_lint(code_no_infix, rex::rex("Indentation should be 2 spaces but is 4 spaces."), tidy_linter)

  expect_lint(code_infix, rex::rex("Indentation should be 4 spaces but is 2 spaces."), no_infix_linter)
  expect_lint(code_infix_2, rex::rex("Indentation should be 8 spaces but is 6 spaces."), no_infix_linter)
  expect_lint(code_infix_3, rex::rex("Indentation should be 4 spaces but is 2 spaces."), no_infix_linter)
  expect_lint(code_infix_4, list(
    list(line_number = 4L, rex::rex("Indentation should be 4 spaces but is 2 spaces.")),
    list(line_number = 10L, rex::rex("Indentation should be 6 spaces but is 4 spaces.")),
    list(line_number = 17L, rex::rex("Indentation should be 6 spaces but is 4 spaces."))
  ), no_infix_linter)
  expect_no_lint(code_no_infix, no_infix_linter)
})

test_that("consecutive same-level lints are suppressed", {
  bad_code <- trim_some("
    ok_code <- 42

    wrong_hanging <- fun(a, b, c,
                           d, e %>%
                             f())

    wrong_block <- function() {
        a + b
        c + d
        if (a == 24)
          boo
    }

    wrong_hanging_args <- function(a = 1, b = 2,
      c = 3, d = 4,
      e = 5, f = 6)
    {
      a + b + c + d + e + f
    }
  ")

  expect_lint(
    bad_code,
    list(
      list(line_number = 4L, message = "Hanging indent"),
      list(line_number = 8L, message = "Indentation"),
      list(line_number = 15L, message = "Hanging indent")
    ),
    indentation_linter()
  )
})

test_that("native pipe is supported", {
  linter <- indentation_linter()

  expect_no_lint(
    trim_some("
      a |>
        foo()
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      b <- a |>
        foo()
    "),
    linter
  )
})

test_that("it doesn't error on invalid code", {
  # Part of #1427
  expect_lint("function() {)", list(linter = "error", message = rex::rex("unexpected ')'")), indentation_linter())
})

test_that("function shorthand is handled", {
  linter <- indentation_linter()

  expect_no_lint(
    trim_some("
      lapply(1:10, \\(i) {
        i %% 2
      })
    "),
    linter
  )

  expect_lint(
    trim_some("
      lapply(1:10, \\(i) {
       i %% 2  # indentation is only 1 character
      })
    "),
    "Indentation",
    linter
  )

  expect_no_lint(
    trim_some(R"(
      \(
        a = 1L,
        b = 2L
      ) {
        a + b
      }
    )"),
    linter
  )
})

test_that("lint metadata works for 0-space case", {
  expect_lint(
    trim_some("
    if (TRUE) {
    FALSE
    }
    "),
    list(ranges = list(1L:2L)),
    indentation_linter()
  )
})

test_that("for loop gets correct linting", {
  linter <- indentation_linter()
  lint_msg <- rex::rex("Indentation should be 2 spaces")

  expect_no_lint(
    trim_some("
       for (i in 1:10)
         1
    "),
    linter
  )
  expect_lint(
    trim_some("
       for (i in 1:10)
          1
    "),
    lint_msg,
    linter
  )

  expect_no_lint(
    trim_some("
       for (i in 1:10) {
         1
       }
    "),
    linter
  )
  expect_lint(
    trim_some("
       for (i in 1:10) {
          1
       }
    "),
    lint_msg,
    linter
  )
})
# fuzzer enable: comment_injection

test_that("consecutive_mutate_linter skips allowed usages", {
  linter <- consecutive_mutate_linter()

  expect_no_lint("DF %>% mutate(x = 1)", linter)

  # intervening expression
  expect_no_lint("DF %>% mutate(x = 1) %>% filter(x > 2) %>% mutate(y = 2)", linter)

  # pipeline starts with mutate()
  expect_no_lint("mutate(DF, x = 1) %>% arrange(y) %>% mutate(z = 2)", linter)

  # new dplyr: .keep and .by arguments are ignored
  expect_no_lint("DF %>% mutate(a = 1) %>% mutate(a = a / sum(a), .by = b)", linter)
  expect_no_lint("DF %>% mutate(a = 1) %>% mutate(a = b, .keep = 'none')", linter)
  expect_no_lint("DF %>% mutate(a = a / sum(a), .by = b) %>% mutate(c = 1)", linter)
  expect_no_lint("DF %>% mutate(a = 1, .keep = 'none') %>% mutate(a = a + 1)", linter)
})

patrick::with_parameters_test_that(
  "consecutive_mutate_linter skips files loading SQL backends",
  {
    linter <- consecutive_mutate_linter(invalid_backends = backend)

    expect_no_lint(
      trim_some(glue::glue("
        library({backend})
        DF %>% mutate(a = a + 1) %>% mutate(b = a - 2)
      ")),
      linter
    )

    expect_no_lint(
      trim_some(glue::glue("
        require('{backend}')
        DF %>% mutate(a = a + 1) %>% mutate(b = a - 2)
      ")),
      linter
    )

    expect_no_lint(
      trim_some(glue("
        conn %>%
          tbl({backend}::sql('SELECT 1 AS x')) %>%
          mutate(a = x + 1) %>%
          mutate(b = a + 1)
      ")),
      linter
    )

    expect_no_lint(
      trim_some(glue("
        conn %>%
          tbl({backend}:::sql('SELECT 1 AS x')) %>%
          mutate(a = x + 1) %>%
          mutate(b = a + 1)
      ")),
      linter
    )

    expect_no_lint(
      trim_some(glue("
        #' @import {backend}
        NULL

        DF %>% mutate(a = a + 1) %>% mutate(b = a - 2)
      ")),
      linter
    )

    expect_no_lint(
      trim_some(glue("
        #' @importFrom {backend} sql
        NULL

        conn %>%
          tbl(sql('SELECT 1 AS x')) %>%
          mutate(a = x + 1) %>%
          mutate(b = a + 1)
      ")),
      linter
    )
  },
  .test_name = c("dbplyr", "custom.backend"),
  backend = c("dbplyr", "custom.backend")
)

test_that("consecutive_mutate_linter blocks simple disallowed usages", {
  linter <- consecutive_mutate_linter()
  lint_msg <- rex::rex("Unify consecutive calls to mutate().")

  # one test of inline usage
  expect_lint("DF %>% mutate(a = 1) %>% mutate(b = 2)", lint_msg, linter)

  expect_lint(
    trim_some("
      DF %>%
       mutate(a = 1) %>%
       mutate(b = 2)
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      DF %>%
       dplyr::mutate(a = 1) %>%
       dplyr::mutate(b = 2)
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      DF %>%
       mutate(a = 1) %>%
       # a comment on b
       mutate(b = 2)
    "),
    lint_msg,
    linter
  )

  # mutate to open pipeline followed by mutate
  expect_lint("mutate(DF, x = 1) %>% mutate(x = 2)", lint_msg, linter)
})

test_that("'parallel' calls are not linted", {
  linter <- consecutive_mutate_linter()

  expect_no_lint("foo(mutate(DF1, x = 1), mutate(DF2, y = 2))", linter)

  expect_no_lint("foo(DF1 %>% mutate(x = 1), DF2 %>% mutate(y = 2))", linter)

  expect_no_lint("DF1 %>% mutate(x = 1) %>% inner_join(DF2 %>% mutate(y = 2))", linter)
})

test_that("native pipe is linted", {
  linter <- consecutive_mutate_linter()
  lint_msg <- rex::rex("Unify consecutive calls to mutate().")

  expect_lint("DF |> mutate(a = 1) |> mutate(b = 2)", lint_msg, linter)
  # Ditto mixed pipes
  expect_lint("DF %>% mutate(a = 1) |> mutate(b = 2)", lint_msg, linter)
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Unify consecutive calls to mutate().")

  expect_lint(
    trim_some("
      DF %>%
        mutate(a = 1) %>%
        mutate(b = 2) %>%
        mutate(c = 3)
    "),
    list(
      list(lint_msg, line_number = 3L),
      list(lint_msg, line_number = 4L)
    ),
    consecutive_mutate_linter()
  )
})

test_that("class_equals_linter skips allowed usages", {
  linter <- class_equals_linter()

  expect_no_lint("class(x) <- 'character'", linter)
  expect_no_lint("class(x) = 'character'", linter)

  # proper way to test exact class
  expect_no_lint("identical(class(x), c('glue', 'character'))", linter)
  expect_no_lint("is_lm <- inherits(x, 'lm')", linter)
})

test_that("class_equals_linter blocks simple disallowed usages", {
  linter <- class_equals_linter()
  lint_msg <- rex::rex("Use inherits(x, 'class-name'), is.<class> for S3 classes, or is(x, 'S4Class') for S4 classes")

  expect_lint("if (class(x) == 'character') stop('no')", lint_msg, linter)
  expect_lint("is_regression <- class(x) == 'lm'", lint_msg, linter)
  expect_lint("is_regression <- 'lm' == class(x)", lint_msg, linter)
})

test_that("class_equals_linter blocks usage of %in% for checking class", {
  linter <- class_equals_linter()
  lint_msg <- rex::rex("Use inherits(x, 'class-name'), is.<class> for S3 classes, or is(x, 'S4Class') for S4 classes")

  expect_lint("if ('character' %in% class(x)) stop('no')", lint_msg, linter)
  expect_lint("if (class(x) %in% 'character') stop('no')", lint_msg, linter)
})

test_that("class_equals_linter blocks class(x) != 'klass'", {
  expect_lint(
    "if (class(x) != 'character') TRUE",
    rex::rex("Use inherits(x, 'class-name'), is.<class> for S3 classes, or is(x, 'S4Class') for S4 classes"),
    class_equals_linter()
  )
})

# as seen, e.g. in base R
test_that("class_equals_linter skips usage for subsetting", {
  linter <- class_equals_linter()

  expect_no_lint("class(x)[class(x) == 'foo']", linter)

  # but not further nesting
  expect_lint(
    "x[if (class(x) == 'foo') 1 else 2]",
    rex::rex("Use inherits(x, 'class-name'), is.<class> for S3 classes, or is(x, 'S4Class') for S4 classes"),
    linter
  )
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Use inherits(x, 'class-name'), is.<class> for S3 classes, or is(x, 'S4Class') for S4 classes")

  expect_lint(
    trim_some("{
      'character' %in% class(x)
      class(x) == 'character'
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    class_equals_linter()
  )
})

test_that("namespace_linter skips allowed usages", {
  linter <- namespace_linter()

  expect_no_lint("stats::sd", linter)
  expect_no_lint("stats::sd(c(1,2,3))", linter)
  expect_no_lint('"stats"::sd(c(1,2,3))', linter)
  expect_no_lint('stats::"sd"(c(1,2,3))', linter)
  expect_no_lint("stats::`sd`(c(1,2,3))", linter)

  expect_no_lint("datasets::mtcars", linter)
  expect_no_lint("stats:::print.formula", linter)
  expect_no_lint('"stats":::print.formula', linter)
})

test_that("namespace_linter respects check_exports and check_nonexports arguments", {
  expect_no_lint("stats::ssd(c(1,2,3))", namespace_linter(check_exports = FALSE))
  expect_no_lint("stats:::ssd(c(1,2,3))", namespace_linter(check_nonexports = FALSE))
  expect_no_lint("stats:::ssd(c(1,2,3))", namespace_linter(check_exports = FALSE, check_nonexports = FALSE))
})

test_that("namespace_linter can work with backticked symbols", {
  skip_if_not_installed("rlang")
  linter <- namespace_linter()

  expect_no_lint("rlang::`%||%`", linter)
  expect_no_lint("rlang::`%||%`()", linter)

  expect_no_lint("rlang::'%||%'", linter)
  expect_no_lint("rlang::'%||%'()", linter)
  expect_no_lint('rlang::"%||%"', linter)
  expect_no_lint('rlang::"%||%"()', linter)

  expect_lint("rlang::`%>%`", "'%>%' is not exported from {rlang}.", linter)
  expect_lint("rlang::'%>%'()", "'%>%' is not exported from {rlang}.", linter)
  expect_lint('rlang::"%>%"()', "'%>%' is not exported from {rlang}.", linter)
})

test_that("namespace_linter blocks disallowed usages", {
  linter <- namespace_linter()

  expect_lint(
    "statts::sd(c(1,2,3))",
    rex::rex("Package 'statts' is not installed."),
    linter
  )

  expect_lint(
    "stats::ssd(c(1,2,3))",
    rex::rex("'ssd' is not exported from {stats}"),
    linter
  )

  expect_lint(
    "stats:::sd(c(1,2,3))",
    rex::rex("Don't use `:::` to access sd, which is exported from stats."),
    linter
  )

  expect_lint(
    "statts:::sd(c(1,2,3))",
    rex::rex("Package 'statts' is not installed."),
    linter
  )

  expect_lint(
    "stats:::sdd(c(1,2,3))",
    rex::rex("'sdd' does not exist in {stats}"),
    linter
  )

  expect_lint(
    trim_some("
      stats::sd(c(1,2,3))
      stats::sdd(c(1,2,3))
    "),
    list(line = "stats::sdd(c(1,2,3))"),
    linter
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      statts::sd(c(1,2,3))
      stats::ssd(c(1,2,3))
      stats:::sd(c(1,2,3))
    }"),
    list(
      list(rex::rex("Package 'statts' is not installed."), line_number = 2L),
      list(rex::rex("'ssd' is not exported from {stats}"), line_number = 3L),
      list(rex::rex("Don't use `:::` to access sd"), line_number = 4L)
    ),
    namespace_linter()
  )
})

test_that("namespace_linter detects functions already imported in the NAMESPACE", { # nofuzz
  pkg_dir <- withr::local_tempdir("testpkg")
  dir.create(file.path(pkg_dir, "R"))
  write.dcf(
    list(Package = "testpkg", Version = "1.0.0"),
    file.path(pkg_dir, "DESCRIPTION")
  )
  writeLines(
    c("importFrom(stats, median)", "importFrom(utils, head)"),
    file.path(pkg_dir, "NAMESPACE")
  )

  test_file <- file.path(pkg_dir, "R", "test.R")
  writeLines(
    c(
      "stats::median(1:10)",
      "utils:::head(1:10)",
      "stats::sd(1:10) # not imported"
    ),
    test_file
  )

  expect_lint(
    file = test_file,
    checks = list(
      list("Don't use `::` to access median.*already imported", line_number = 1L),
      list("Don't use `:::` to access head.*already imported", line_number = 2L)
    ),
    linters = namespace_linter(check_nonexports = FALSE)
  )
})

test_that("namespace_linter works with backticked symbols", { # nofuzz
  skip_if_not_installed("rlang")

  pkg_dir <- withr::local_tempdir("testpkg_rlang")
  dir.create(file.path(pkg_dir, "R"))
  write.dcf(
    list(Package = "testpkg_rlang", Version = "1.0.0"),
    file.path(pkg_dir, "DESCRIPTION")
  )
  writeLines('importFrom("rlang", "%||%")', file.path(pkg_dir, "NAMESPACE"))

  test_file <- file.path(pkg_dir, "R", "test.R")
  writeLines("rlang::`%||%`", test_file)

  expect_lint(
    file = test_file,
    checks = rex::rex("Don't use `::` to access %||%, which is already imported from rlang."),
    linters = namespace_linter()
  )
})

# fuzzer disable: comment_injection
test_that("paren_body_linter returns correct lints", {
  linter <- paren_body_linter()
  lint_msg <- rex::rex("Put a space between a right parenthesis and a body expression.")

  # No space after the closing parenthesis prompts a lint
  expect_lint("function()test", lint_msg, linter)
  expect_lint("print('hello')\nx <- function(x)NULL\nprint('hello')", lint_msg, linter)
  expect_lint("if (TRUE)test", lint_msg, linter)
  expect_lint("while (TRUE)test", lint_msg, linter)
  expect_lint("for (i in seq_along(1))test", lint_msg, linter)

  # A space after the closing parenthesis does not prompt a lint
  expect_no_lint("function() test", linter)

  # Symbols after the closing parenthesis of a function call do not prompt a lint
  expect_no_lint("head(mtcars)$cyl", linter)

  # paren_body_linter returns the correct line number
  expect_lint(
    "print('hello')\nx <- function(x)NULL\nprint('hello')",
    list(line_number = 2L),
    linter
  )

  expect_lint(
    "function()test",
    list(
      line_number = 1L,
      column_number = 10L,
      type = "style",
      line = "function()test",
      ranges = list(c(10L, 10L))
    ),
    linter
  )

  expect_lint(
    "for (ii in 1:10)ii",
    list(column_number = 16L, ranges = list(c(16L, 16L))),
    linter
  )

  # paren_body_linter does not lint when the function body is defined on a new line
  expect_no_lint("function()\n  test", linter)

  # paren_body_linter does not lint comments
  expect_no_lint("#function()test", linter)

  # multiple lints on the same line
  expect_lint("function()if(TRUE)while(TRUE)test", list(lint_msg, lint_msg, lint_msg), linter)

  # No space after the closing parenthesis of an anonymous function prompts a lint
  expect_lint("\\()test", lint_msg, linter)
})

test_that("multi-line versions are caught", {
  linter <- paren_body_linter()
  lint_msg <- rex::rex("Put a space between a right parenthesis and a body expression.")

  expect_lint(
    trim_some("
      function(var
                  )x
    "),
    lint_msg,
    linter
  )
  expect_lint(
    trim_some("
      if (cond
              )x
    "),
    lint_msg,
    linter
  )
  expect_lint(
    trim_some("
      while (cond
                 )x
    "),
    lint_msg,
    linter
  )

  expect_lint(
    trim_some("
      \\(var
            )x
    "),
    lint_msg,
    linter
  )
})

test_that("function shorthand is handled", {
  linter <- paren_body_linter()
  lint_msg <- rex::rex("Put a space between a right parenthesis and a body expression.")

  expect_lint("\\()test", lint_msg, linter)
})
# fuzzer enable: comment_injection

test_that("function_left_parentheses_linter skips allowed usages", {
  linter <- function_left_parentheses_linter()

  expect_no_lint("blah", linter)
  expect_no_lint("print(blah)", linter)
  expect_no_lint('"print"(blah)', linter)
  expect_no_lint("base::print(blah)", linter)
  expect_no_lint('base::"print"(blah)', linter)
  expect_no_lint("base::print(blah, fun(1))", linter)
  expect_no_lint("blah <- function(blah) { }", linter) # nofuzz: comment_injection
  expect_no_lint("(1 + 1)", linter)
  expect_no_lint("( (1 + 1) )", linter)
  expect_no_lint("if (blah) { }", linter)
  expect_no_lint("for (i in j) { }", linter)
  expect_no_lint("1 * (1 + 1)", linter)
  expect_no_lint("!(1 == 1)", linter)
  expect_no_lint("(2 - 1):(3 - 1)", linter)
  expect_no_lint("c(1, 2, 3)[(2 - 1)]", linter)
  expect_no_lint("list(1, 2, 3)[[(2 - 1)]]", linter)
  expect_no_lint("range(10)[(2 - 1):(10 - 1)]", linter)
  expect_no_lint("function(){function(){}}()()", linter) # nofuzz: comment_injection
  expect_no_lint("c(function(){})[1]()", linter) # nofuzz: comment_injection
  expect_no_lint("function(x) (mean(x) + 3)", linter) # nofuzz: comment_injection
  expect_no_lint('"blah (1)"', linter)
})

test_that("function_left_parentheses_linter blocks disallowed usages", {
  linter <- function_left_parentheses_linter()
  fun_lint_msg <- rex::rex("Remove spaces before the left parenthesis in a function definition.")
  call_lint_msg <- rex::rex("Remove spaces before the left parenthesis in a function call.")

  expect_lint("blah (1)", call_lint_msg, linter)
  expect_lint("base::print (blah)", call_lint_msg, linter)
  expect_lint("base::print(blah, f (1))", call_lint_msg, linter)
  expect_lint("`+` (1, 1)", call_lint_msg, linter)
  expect_lint("test <- function (x) { }", fun_lint_msg, linter)

  expect_lint(
    "blah  (1)",
    list(message = call_lint_msg, column_number = 5L, ranges = list(c(5L, 6L))),
    linter
  )
  expect_lint(
    "test <- function  (x) { }",
    list(message = fun_lint_msg, column_number = 17L, ranges = list(c(17L, 18L))),
    linter
  )
})

test_that("multi-line cases are handled correctly", {
  linter <- function_left_parentheses_linter()
  call_lint_msg <- rex::rex("Left parenthesis should be on the same line as the function's symbol.")
  fun_lint_msg <- rex::rex("Left parenthesis should be on the same line as the 'function' symbol.")

  expect_lint(
    trim_some("
      foo <- function
      (
        param
      ) {
        param + 1
      }
    "),
    fun_lint_msg,
    linter
  )

  # edge case where '(' is in the right place on the wrong line
  expect_lint(
    trim_some("
      foo <- function
                     (
        param
      ) {
        param + 1
      }
    "),
    fun_lint_msg,
    linter
  )

  # ditto for function calls
  expect_lint(
    trim_some("
      if (
        y > sum
        (
          x
        )
      ) {
        TRUE
      }
    "),
    call_lint_msg,
    linter
  )
  expect_lint(
    trim_some("
      if (
        y > sum
               (
          x
        )
      ) {
        TRUE
      }
    "),
    call_lint_msg,
    linter
  )
})

test_that("multi-lint case works", {
  expect_lint(
    trim_some("
      if (
        y > sum
               (
          x
        ) &&
        z > mean
        (
          x
        ) &&
        a > sd (x) &&
        b > var  (x)
      ) {
        TRUE
      }
    "),
    list(
      list(
        message = "Left parenthesis should be on the same line as the function's symbol.",
        line_number = 2L,
        ranges = list(c(7L, 9L))
      ),
      list(
        message = "Left parenthesis should be on the same line as the function's symbol.",
        line_number = 6L,
        ranges = list(c(7L, 10L))
      ),
      list(
        message = "Remove spaces before the left parenthesis in a function call.",
        line_number = 10L,
        ranges = list(c(9L, 9L))
      ),
      list(
        message = "Remove spaces before the left parenthesis in a function call.",
        line_number = 11L,
        ranges = list(10L:11L)
      )
    ),
    function_left_parentheses_linter()
  )
})

test_that("it doesn't produce invalid lints", {
  # Part of #1427
  expect_no_warning(
    expect_lint(
      "function() {)",
      list(linter = "error"),
      function_left_parentheses_linter()
    )
  )
})

test_that("newline in character string doesn't trigger false positive (#1963)", {
  linter <- function_left_parentheses_linter()

  expect_no_lint('foo("\n")$bar()', linter)
  # also corrected the lint metadata for similar cases
  expect_lint(
    trim_some('
      (
        foo("
        ")$bar
        ()
      )
    '),
    # attach to 'b' in '$bar'
    list(line_number = 3L, column_number = 6L),
    linter
  )

  expect_lint(
    trim_some('
      (
        foo("
        ")@bar
        ()
      )
    '),
    # attach to 'b' in '@bar'
    list(line_number = 3L, column_number = 6L),
    linter
  )
})

test_that("shorthand functions are handled", { # nofuzz: comment_injection
  linter <- function_left_parentheses_linter()
  fun_lint_msg <- rex::rex("Remove spaces before the left parenthesis in a function definition.")

  expect_no_lint("blah <- \\(blah) { }", linter)
  expect_no_lint("\\(){\\(){}}()()", linter)
  expect_lint("test <- \\ (x) { }", fun_lint_msg, linter)
})

test_that("whitespace_linter skips allowed usages", {
  linter <- whitespace_linter()

  expect_no_lint("blah", linter)
  expect_no_lint("  blah", linter)
  expect_no_lint("  blah", linter)
  expect_no_lint("#\tblah", linter)
})

test_that("whitespace_linter skips allowed tab usages inside strings", {
  linter <- whitespace_linter()

  expect_no_lint(
    'lint_msg <- "dont flag tabs if\tthey are inside a string."',
    linter
  )

  expect_no_lint(
    'lint_msg <- "dont flag tabs if\n\tthey are inside multiline strings."',
    linter
  )
})

test_that("whitespace_linter blocks disallowed usages", {
  linter <- whitespace_linter()
  lint_msg <- rex::rex("Use spaces to indent, not tabs.")

  expect_lint(
    "\tblah",
    list(message = lint_msg, line_number = 1L, column_number = 1L, ranges = list(c(1L, 1L))),
    linter
  )

  expect_lint(
    "\n\t\t\tblah",
    list(message = lint_msg, line_number = 2L, column_number = 1L),
    linter
  )
})

test_that("whitespace_linter blocks disallowed usages with a pipe", {
  linter <- whitespace_linter()
  lint_msg <- rex::rex("Use spaces to indent, not tabs.")

  expect_lint(
    "a %>%\n\tb()",
    list(message = lint_msg, line_number = 2L, column_number = 1L, ranges = list(c(1L, 1L))),
    linter
  )

  expect_lint(
    "a |>\n\tb()",
    list(message = lint_msg, line_number = 2L, column_number = 1L, ranges = list(c(1L, 1L))),
    linter
  )
})

linter <- unnecessary_placeholder_linter()
pipes <- pipes(exclude = "|>")

patrick::with_parameters_test_that(
  "unnecessary_placeholder_linter skips allowed usages",
  {
    # . used in position other than first --> ok
    expect_no_lint(sprintf("x %s foo(y, .)", pipe), linter)
    # ditto for nested usage
    expect_no_lint(sprintf("x %s foo(y, bar(.))", pipe), linter)
    # . used twice --> ok
    expect_no_lint(sprintf("x %s foo(., .)", pipe), linter)
    # . used as a keyword argument --> ok
    expect_no_lint(sprintf("x %s foo(arg = .)", pipe), linter)
    # . used inside a scope --> ok
    expect_no_lint(sprintf("x %s { foo(arg = .) }", pipe), linter)
  },
  .test_name = names(pipes),
  pipe = pipes
)

patrick::with_parameters_test_that( # nofuzz: pipe
  "unnecessary_placeholder_linter blocks simple disallowed usages",
  {
    expect_lint(
      sprintf("x %s sum(.)", pipe),
      rex::rex("Don't use the placeholder (`.`) when it's not needed"),
      unnecessary_placeholder_linter()
    )

    expect_lint(
      sprintf("x %s y(.) %s sum()", pipe, pipe),
      rex::rex("Don't use the placeholder (`.`) when it's not needed"),
      unnecessary_placeholder_linter()
    )
  },
  .test_name = names(pipes),
  pipe = pipes
)

test_that("lints vectorize", { # nofuzz: pipe
  lint_msg <- rex::rex("Don't use the placeholder (`.`) when it's not needed")

  expect_lint(
    trim_some("{
      x %>% foo(.)
      y %T>% bar(.)
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    unnecessary_placeholder_linter()
  )
})

test_that("logic survives adversarial commenting", {
  expect_no_lint(
    trim_some("
      x %T>% foo(arg = # comment
      .)
    "),
    unnecessary_placeholder_linter()
  )
})

test_that("skips allowed usages", {
  linter <- length_test_linter()

  expect_no_lint("length(x) > 0", linter)
  expect_no_lint("length(DF[key == val, cols])", linter)
  expect_no_lint("nrow(x) > 0", linter)
  expect_no_lint("nrow(DF[key == val, cols])", linter)
  expect_no_lint("ncol(x) > 0", linter)
  expect_no_lint("ncol(DF[key == val, cols])", linter)
  expect_no_lint("NROW(x) > 0", linter)
  expect_no_lint("NROW(DF[key == val, cols])", linter)
  expect_no_lint("NCOL(x) > 0", linter)
  expect_no_lint("NCOL(DF[key == val, cols])", linter)
})

patrick::with_parameters_test_that(
  "blocks simple disallowed usages",
  {
    linter <- length_test_linter()
    lint_msg_stub <- sprintf("Checking the %s of a logical vector is likely a mistake. Did you mean ", fun)
    expect_lint(
      paste0(fun, "(x == 0)"),
      rex::rex(lint_msg_stub, "`", fun, "(x) == 0`?"),
      linter
    )
    expect_lint(
      paste0(fun, "(x == y)"),
      rex::rex(lint_msg_stub, "`", fun, "(x) == y`?"),
      linter
    )
    expect_lint(
      paste0(fun, "(x + y == 2)"),
      rex::rex(lint_msg_stub, "`", fun, "(x+y) == 2`?"),
      linter
    )
  },
  fun = c("length", "nrow", "ncol", "NROW", "NCOL")
)

test_that("adversarial comment is handled in lint message", {
  expect_lint(
    trim_some("
      length(x + #
      y == 2)
    "),
    rex::rex("Checking the length of a logical vector is likely a mistake. Did you mean `length(x+y) == 2`?"),
    length_test_linter()
  )
})

patrick::with_parameters_test_that(
  "all logical operators detected",
  expect_lint(
    sprintf("%s(x %s y)", fun, op),
    rex::rex("`", fun, "(x) ", op, " y`?"),
    length_test_linter()
  ),
  .cases = expand.grid(
    op = c("<", "<=", ">", ">=", "==", "!="),
    fun = c("length", "nrow", "ncol", "NROW", "NCOL")
  )
)

test_that("lints vectorize", {
  linter <- length_test_linter()

  expect_lint(
    trim_some("{
      length(x == y)
      length(y == z)
    }"),
    list(
      list(rex::rex("length(x) == y"), line_number = 2L),
      list(rex::rex("length(y) == z"), line_number = 3L)
    ),
    linter
  )

  expect_lint(
    trim_some("{
      length( # comment
      x       # comment
      ==      # comment
      y       # comment
      )       # comment
      length( # comment
      y       # comment
      ==      # comment
      z       # comment
      )
    }"),
    list(
      list(rex::rex("length(x) == y"), line_number = 2L),
      list(rex::rex("length(y) == z"), line_number = 7L)
    ),
    linter
  )
})

test_that("function_argument_linter skips allowed usages", {
  linter <- function_argument_linter()

  expect_no_lint("function(x, y = 1, z = 2) {}", linter)
  expect_no_lint("function(x, y, z = 1) {}", linter)

  # ... handled correctly
  expect_no_lint("function(x, y, z = 1, ...) {}", linter)
  expect_no_lint("function(x, y, ..., z = 1) {}", linter)
})

test_that("function_argument_linter blocks disallowed usages", {
  linter <- function_argument_linter()
  lint_msg <- rex::rex("Arguments without defaults should come before arguments with defaults.")

  expect_lint("function(x, y = 1, z) {}", list(message = lint_msg, column_number = 20L), linter)
  expect_lint("function(x, y = 1, z, w = 2) {}", list(message = lint_msg, column_number = 20L), linter)

  expect_lint("function(x, y, z = 1, ..., w) {}", list(message = lint_msg, column_number = 28L), linter)

  # Multi-line also works
  expect_lint(
    trim_some("
      function(x,
               y = 1,
               z) {
      }
    "),
    list(message = lint_msg, line_number = 3L),
    linter
  )
  expect_lint(
    trim_some("
      function(x,
               y # comment for distraction
               = 1,
               z,
               w = 2) {
      }
    "),
    list(message = lint_msg, line_number = 4L),
    linter
  )
  expect_no_lint(
    trim_some("
      function(x,
               y = 1,
               z = 2) {
      }
    "),
    linter
  )
})

test_that("function_argument_linter also lints lambda expressions", {
  linter <- function_argument_linter()
  lint_msg <- rex::rex("Arguments without defaults should come before arguments with defaults.")

  expect_lint("\\(x, y = 1, z) {}", list(message = lint_msg, column_number = 13L), linter)
  expect_lint("\\(x, y = 1, z, w = 2) {}", list(message = lint_msg, column_number = 13L), linter)
  expect_no_lint("\\(x, y = 1, z = 2) {}", linter)
  expect_no_lint("\\(x, y, z = 1) {}", linter)
})

test_that("Use of missing() is reported in the lint message", {
  linter <- function_argument_linter()
  simple_msg <- "Arguments without defaults should come before arguments with defaults."

  expect_lint(
    trim_some("
      function(x, y = 1, z) {
        if (missing(z)) {
          z <- 2
        }
      }
    "),
    rex::rex(simple_msg, anything, "missing()"),
    linter
  )

  expect_lint(
    trim_some("
      function(x, y = 1, z) {
        if (y > 1) {
          if (missing(z)) {
            z <- 2
          }
        }
      }
    "),
    rex::rex(simple_msg, anything, "missing()"),
    linter
  )

  # inline function
  expect_lint(
    "function(x = 2, y) if (missing(y)) x else y",
    rex::rex(simple_msg, anything, "missing()"),
    linter
  )

  # missing() used for a different argument
  expect_lint(
    trim_some("
      function(x, y = 1, z) {
        if (missing(x)) {
          z <- 2
        }
      }
    "),
    rex::rex(simple_msg, end),
    linter
  )

  # missing() used in the signature (not quite the same, and easier to spot)
  expect_lint(
    trim_some("
      function(x = 1, y, z = missing(y)) {
        x
      }
    "),
    rex::rex(simple_msg, end),
    linter
  )
})

test_that("multiple lints give correct message", {
  simple_msg <- "Arguments without defaults should come before arguments with defaults."

  expect_lint(
    trim_some("{
      function(x, y = 1, z) {
        x
      }
      function(x, y = 1, z) {
        if (missing(z)) {
          z <- 2
        }
      }
      function(x, y = 1, z, w) {
        if (missing(z)) {
          z <- 2
        }
      }
    }"),
    list(
      list(rex::rex(simple_msg, end), line_number = 2L),
      list(rex::rex(simple_msg, anything, "missing()"), line_number = 5L),
      list(rex::rex(simple_msg, anything, "missing()"), line_number = 10L, column_number = 22L),
      list(rex::rex(simple_msg, end), line_number = 10L, column_number = 25L)
    ),
    function_argument_linter()
  )
})

test_that("condition_call_linter skips allowed usages", {
  linter <- condition_call_linter()

  expect_no_lint("stop('test', call. = FALSE)", linter)
  expect_no_lint("stop(test)", linter)
  expect_no_lint("stop(errorCondition('test'))", linter)

  # works even with multiple arguments
  expect_no_lint("stop('this is a', 'test', call. = FALSE)", linter)

  linter <- condition_call_linter(display_call = TRUE)

  expect_no_lint("stop('test', call. = TRUE)", linter)
  expect_no_lint("stop(test)", linter)

  linter <- condition_call_linter(display_call = NA)

  expect_no_lint("stop('test', call. = TRUE)", linter)
  expect_no_lint("stop('test', call. = FALSE)", linter)
  expect_no_lint("stop(test)", linter)
})

patrick::with_parameters_test_that(
  "condition_call_linter blocks disallowed usages",
  {
    linter <- condition_call_linter()
    lint_message <- rex::rex(call_name, anything, "not to display the call")

    expect_lint(paste0(call_name, "('test')"), lint_message, linter)
    expect_lint(paste0(call_name, "('test', call. = TRUE)"), lint_message, linter)
    expect_lint(paste0(call_name, "(sprintf('test'))"), lint_message, linter)
    expect_lint(paste0(call_name, "(gettextf('test'))"), lint_message, linter)

    linter <- condition_call_linter(display_call = TRUE)
    lint_message <- rex::rex(call_name, anything, "to display the call")

    expect_lint(paste0(call_name, "('test', call. = FALSE)"), lint_message, linter)

    linter <- condition_call_linter(display_call = NA)
    lint_message <- rex::rex("explicit value", anything, call_name)

    expect_lint(paste0(call_name, "('test')"), lint_message, linter)
  },
  call_name = c("stop", "warning")
)

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      stop('e')
      warning('w')
    }"),
    list(
      list("stop", line_number = 2L),
      list("warning", line_number = 3L)
    ),
    condition_call_linter()
  )
})

test_that("inner_combine_linter lints a false positive-ish usage", {
  # By default as.POSIXct.character picks up the format to apply from
  #   the first element and, since it succeeds, applies that to the remaining
  #   timestamps. Whereas when run individually, it won't succeed until
  #   the correct format is matched for each input. Nevertheless, it is
  #   still preferable to vectorize the call, while being sure to use a
  #   consistent format for the inputs. In this case, the correct equivalent
  #   call is as.POSIXct(c("2021-01-01 00:00:00", "2021-01-01 01:00:00")).
  expect_lint(
    "c(as.POSIXct('2021-01-01'), as.POSIXct('2021-01-01 01:00:00'))",
    rex::rex("Combine inputs to vectorized functions first"),
    inner_combine_linter()
  )
})

local({
  vector_funs <- c(
    "as.Date", "as.POSIXct", "as.POSIXlt",
    "sin", "cos", "tan", "sinpi", "cospi", "tanpi", "asin", "acos", "atan",
    "log", "logb", "log2", "log10", "log1p", "exp", "expm1",
    "sqrt", "abs",
    "ymd", "ydm", "mdy", "myd", "dmy", "dym",
    "yq", "ym", "my",
    "ymd_hms", "ymd_hm", "ymd_h", "dmy_hms", "dmy_hm", "dmy_h",
    "mdy_hms", "mdy_hm", "mdy_h", "ydm_hms", "ydm_hm", "ydm_h",
    NULL
  )
  patrick::with_parameters_test_that(
    "inner_combine_linter blocks simple vectorized calls:",
    expect_lint(
      sprintf("c(%1$s(x), %1$s(y))", vector_fun),
      rex::rex("Combine inputs to vectorized functions first"),
      inner_combine_linter()
    ),
    .test_name = vector_funs,
    vector_fun = vector_funs
  )
})

patrick::with_parameters_test_that(
  "inner_combine_linter blocks as.Date with identical passed arguments:",
  expect_lint(
    sprintf("c(as.Date(x, %1$s), as.Date(y, %1$s))", arg),
    rex::rex("Combine inputs to vectorized functions first"),
    inner_combine_linter()
  ),
  .test_name = c("format", "origin", "tz", "tryFormats", "non-literal", "quoted arg name"),
  arg = c("format = '%F'", "origin = '1900-01-01'", "tz = 'Asia/Jakarta'", "tryFormats = '%F'", "tz = tz", "'tz' = tz")
)

patrick::with_parameters_test_that(
  "inner_combine_linter blocks as.POSIXct with identical passed arguments:",
  expect_lint(
    sprintf("c(as.POSIXct(x, %1$s), as.POSIXct(y, %1$s))", arg),
    rex::rex("Combine inputs to vectorized functions first"),
    inner_combine_linter()
  ),
  .test_name = c("format", "origin", "tz", "non-literal"),
  arg = c("format = '%F'", "origin = '1900-01-01'", "tz = 'UTC'", "tz = tz")
)

test_that("inner_combine_linter is order-agnostic for matching arguments", {
  expect_lint(
    "c(as.Date(x, format = f, tz = t), as.Date(y, tz = t, format = f))",
    rex::rex("Combine inputs to vectorized functions first"),
    inner_combine_linter()
  )
})

test_that("c() with ...length()=1 is OK", {
  expect_no_lint("c(exp())", inner_combine_linter())
})

skip_if_not_installed("tibble")
patrick::with_parameters_test_that(
  "inner_combine_linter skips allowed usages:",
  expect_no_lint(expr, inner_combine_linter()),
  .cases = tibble::tribble(
    ~.test_name,                      ~expr,
    "simple sin()",                   "x <- sin(1:10)",
    "mixed sin()+cos()",              "y <- c(sin(1:10), cos(2:20))",
    "present/absent vector function", "c(log(x), 0.5)",
    "absent/present vector function", "c(0.5, log(x))",
    "mismatched arg (Date)",          "c(as.Date(x, format = '%y'), as.Date(y, format = '%m'))",
    "present/absent arg (Date)",      "c(as.Date(x, format = '%y'), as.Date(y))",
    "absent/present arg (Date)",      "c(as.Date(x), as.Date(y, format = '%y'))",
    "matched value, not arg (Date)",  "c(as.Date(x, format = '%y'), as.Date(y, tz = '%y'))",
    "mismatched arg (POSIXct)",       "c(as.POSIXct(x, format = '%y'), as.POSIXct(y, format = '%m'))",
    "present/absent arg (POSIXct)",   "c(as.POSIXct(x, format = '%y'), as.POSIXct(y))",
    "mismatched arg (log)",           "c(log(x, base = 4), log(y, base = 5))",
    "present/absent arg (log)",       "c(log(x, base = 4), log(y))"
    # TODO(#2486): Reactivate these.
    # "unknown Date method argument",    "c(as.Date(x, zoo = zzz), as.Date(y, zoo = zzz))",
    # "known+unknown Date argument", "c(as.Date(x, format = '%y', zoo = zzz), as.Date(y, format = '%y', zoo = zzz))",
    # "unknown POSIXct method argument", "c(as.POSIXct(x, zoo = zzz), as.POSIXct(y, zoo = zzz))",
  )
)

test_that("lints vectorize", {
  lint_msg <- rex::rex("Combine inputs to vectorized functions first")

  expect_lint(
    trim_some("{
      c(sin(x), sin(y))
      c(log(x), log(y))
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    inner_combine_linter()
  )
})

with_content_to_parse <- function(content, code) {
  f <- withr::local_tempfile()
  local({
    con <- file(f, open = "w", encoding = "UTF-8")
    on.exit(close(con))
    writeLines(content, con)
  })
  source_expressions <- get_source_expressions(f)
  content_env <- new.env()
  content_env$pc <- lapply(source_expressions[["expressions"]], `[[`, "parsed_content")
  content_env$error <- source_expressions$error
  content_env$warning <- source_expressions$warning
  eval(substitute(code), envir = content_env)
}

test_that("tab positions have been corrected", {
  with_content_to_parse(
    "1\n\t",
    expect_length(pc, 2L)
  )

  with_content_to_parse(
    "TRUE",
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == "TRUE", c("col1", "col2")], use.names = FALSE), c(1L, 4L))
  )
  with_content_to_parse(
    "\tTRUE",
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == "TRUE", c("col1", "col2")], use.names = FALSE), c(2L, 5L))
  )

  with_content_to_parse(
    "\t\tTRUE",
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == "TRUE", c("col1", "col2")], use.names = FALSE), c(3L, 6L))
  )

  with_content_to_parse("x\t<-\tTRUE", {
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == "x", c("col1", "col2")], use.names = FALSE), c(1L, 1L))
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == "<-", c("col1", "col2")], use.names = FALSE), c(3L, 4L))
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == "TRUE", c("col1", "col2")], use.names = FALSE), c(6L, 9L))
  })

  with_content_to_parse("\tfunction\t(x)\t{\tprint(pc[\t,1])\t;\t}", {
    expect_identical(
      unlist(pc[[1L]][pc[[1L]][["text"]] == "function", c("col1", "col2")], use.names = FALSE),
      c(2L, 9L)
    )
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == "x", c("col1", "col2")], use.names = FALSE), c(12L, 12L))
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == "print", c("col1", "col2")], use.names = FALSE), c(17L, 21L))
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == ";", c("col1", "col2")], use.names = FALSE), c(32L, 32L))
    expect_identical(unlist(pc[[1L]][pc[[1L]][["text"]] == "}", c("col1", "col2")], use.names = FALSE), c(34L, 34L))
  })

  with_content_to_parse("# test tab\n\ns <- 'I have \\t a dog'\nrep(\ts, \t3)", {
    expect_identical(
      unlist(pc[[2L]][pc[[2L]][["text"]] == "'I have \\t a dog'", c("line1", "col1", "col2")], use.names = FALSE),
      c(3L, 6L, 22L)
    )
    expect_identical(
      unlist(pc[[3L]][pc[[3L]][["text"]] == "3", c("line1", "col1", "col2")], use.names = FALSE),
      c(4L, 10L, 10L)
    )
  })

  with_content_to_parse("function(){\nTRUE\n\t}", {
    expect_identical(
      unlist(pc[[1L]][1L, c("line1", "col1", "line2", "col2")], use.names = FALSE),
      c(1L, 1L, 3L, 2L),
      info = "expression that spans several lines"
    )
  })
})

test_that("Terminal newlines are detected correctly", {
  content <- "lm(y ~ x)"
  # NB: need to specify terminal newline explicitly with cat, not writeLines()
  tmp <- withr::local_tempfile(lines = content)
  tmp2 <- withr::local_tempfile()
  cat(content, file = tmp2)

  expect_true(get_source_expressions(tmp)$expressions[[2L]]$terminal_newline)
  expect_false(get_source_expressions(tmp2)$expressions[[2L]]$terminal_newline)
})

test_that("Multi-byte characters correct columns", {
  skip_if_not_utf8_locale()

  with_content_to_parse("`\U2020` <- 1", {
    # fix_column_numbers corrects the start of <-
    expect_identical(pc[[1L]]$col1[4L], pc[[1L]]$col1[2L] + 4L)
  })
})

test_that("Multi-byte character truncated by parser is ignored", {
  skip_if_not_utf8_locale()
  # \U2013 is the Unicode character 'en dash', which is
  # almost identical to a minus sign in monospaced fonts.
  content <- "y <- x \U2013 42"
  # message is like '<text>:1:8: unexpected invalid token\n1: ...'
  with_content_to_parse(content, {
    base_msg <- conditionMessage(tryCatch(str2lang(content), error = identity))
    # Just ensure that the captured message is a substring of the parser error, #2527
    expect_true(grepl(error$message, base_msg, fixed = TRUE, useBytes = TRUE))
    expect_identical(error$column_number, 8L)
  })
})

test_that("Can read non UTF-8 file", { # nofuzz. Ignore all encoding checks.
  withr::local_options(list(lintr.linter_file = tempfile()))
  proj_dir <- test_path("dummy_projects", "project")
  withr::local_dir(proj_dir)
  expect_no_lint(
    file = "cp1252.R",
    linters = list()
  )
})

test_that("Warns if encoding is misspecified, Pt. 1", {
  proj_dir <- test_path("dummy_projects", "project")
  withr::local_dir(proj_dir)

  lint_msg <- "Invalid multibyte character in parser. Is the encoding correct?"
  if (!isTRUE(l10n_info()[["UTF-8"]])) {
    # Prior to R 4.2.0, the Windows parser throws a different error message because the source code is converted to
    # native encoding.
    # This results in line 4 becoming <fc> <- 42 before the parser sees it.
    lint_msg <- "unexpected '<'"
  }

  expect_lint(
    file = "cp1252.R",
    parse_settings = FALSE,
    checks = list(rex::rex(lint_msg), linter = "error", line_number = 4L)
  )

  expect_lint(
    file = "cp1252_parseable.R",
    parse_settings = FALSE,
    checks = list(
      rex::rex("Invalid multibyte string. Is the encoding correct?"),
      linter = "error",
      line_number = 1L
    )
  )
})

test_that("Can extract line number from parser errors", {
  with_content_to_parse(
    trim_some('
      "ok"
      R"---a---"
    '),
    {
      expect_identical(error$message, "Malformed raw string literal.")
      expect_identical(error$line_number, 2L)
    }
  )

  with_content_to_parse(
    trim_some('
      ok
      ok
      "\\u{9999"
    '),
    {
      expect_identical(error$message, "Invalid \\u{xxxx} sequence.")
      expect_identical(error$line_number, 3L)
    }
  )

  with_content_to_parse(
    trim_some('
      ok
      ok
      "\\u{9999
    '),
    {
      # parser erroneously reports line 4
      expect_identical(error$message, "Invalid \\u{xxxx} sequence.")
      expect_identical(error$line_number, 3L)
    }
  )

  with_content_to_parse("function(a, a) {}", {
    expect_identical(error$message, "Repeated formal argument 'a'.")
    expect_identical(error$line_number, 1L)
  })
})

test_that("1- or 2-width octal expressions give the right STR_CONST values", {
  with_content_to_parse("'\\1'", expect_identical(pc[[1L]][1L, "text"], "'\\1'"))
  with_content_to_parse('"\\1"', expect_identical(pc[[1L]][1L, "text"], '"\\1"'))

  # multiple literals
  with_content_to_parse("'\\1'\n'\\2'", {
    expect_identical(pc[[1L]][1L, "text"], "'\\1'")
    expect_identical(pc[[2L]][1L, "text"], "'\\2'")
  })

  # multiple escapes
  with_content_to_parse("'\\1\\2'", expect_identical(pc[[1L]][1L, "text"], "'\\1\\2'"))

  # multi-line strings
  with_content_to_parse("'\n\\1\n'", expect_identical(pc[[1L]][1L, "text"], "'\n\\1\n'"))
  with_content_to_parse("a <- '\\1\n\\2'", expect_identical(pc[[1L]][5L, "text"], "'\\1\n\\2'"))

  # mixed-length strings
  with_content_to_parse("foo('\\1',\n  '\n\\2\n')", {
    expect_identical(pc[[1L]][5L, "text"], "'\\1'")
    expect_identical(pc[[1L]][8L, "text"], "'\n\\2\n'")
  })
})

test_that("returned data structure is complete", {
  lines <- c("line_1", "line_2", "line_3")
  temp_file <- withr::local_tempfile(lines = lines)

  lines_with_attr <- setNames(lines, seq_along(lines))
  attr(lines_with_attr, "terminal_newline") <- TRUE

  exprs <- get_source_expressions(temp_file)
  expect_named(exprs, c("expressions", "error", "warning", "lines"))
  expect_length(exprs$expressions, length(lines) + 1L)

  for (i in seq_along(lines)) {
    expr <- exprs$expressions[[i]]
    expect_named(expr, c(
      "filename", "line", "column", "lines", "parsed_content", "xml_parsed_content", "xml_find_function_calls",
      "content"
    ))
    expect_identical(expr$filename, temp_file)
    expect_identical(expr$line, i)
    expect_identical(expr$column, 1L)
    expect_identical(expr$lines, setNames(lines[i], i))
    expect_identical(nrow(expr$parsed_content), 2L)
    expect_true(xml2::xml_find_lgl(expr$xml_parsed_content, "count(//SYMBOL) > 0"))
    expect_identical(expr$content, lines[i])
  }
  full_expr <- exprs$expressions[[length(lines) + 1L]]
  expect_named(full_expr, c(
    "filename", "file_lines", "content", "full_parsed_content", "full_xml_parsed_content", "xml_find_function_calls",
    "terminal_newline"
  ))
  expect_identical(full_expr$filename, temp_file)
  expect_identical(full_expr$file_lines, lines_with_attr)
  expect_identical(full_expr$content, lines_with_attr)
  expect_identical(nrow(full_expr$full_parsed_content), 2L * length(lines))
  expect_identical(
    xml2::xml_find_num(full_expr$full_xml_parsed_content, "count(//SYMBOL)"),
    as.numeric(length(lines))
  )
  expect_true(full_expr$terminal_newline)

  expect_null(exprs$error)
  expect_identical(exprs$lines, lines_with_attr)
})

test_that("xml_find_function_calls works as intended", {
  lines <- c(
    "foo()",
    "bar()",
    "foo()",
    "s4Obj@baz()",
    "{ foo(); foo(); bar(); s4Obj@baz() }",
    NULL
  )
  temp_file <- withr::local_tempfile(lines = lines)

  exprs <- get_source_expressions(temp_file)

  expect_length(exprs$expressions[[1L]]$xml_find_function_calls("foo"), 1L)
  expect_length(exprs$expressions[[1L]]$xml_find_function_calls("bar"), 0L)
  expect_identical(
    exprs$expressions[[1L]]$xml_find_function_calls("foo"),
    xml2::xml_find_all(
      exprs$expressions[[1L]]$xml_parsed_content,
      "//SYMBOL_FUNCTION_CALL[text() = 'foo']/parent::expr"
    )
  )

  expect_length(exprs$expressions[[2L]]$xml_find_function_calls("foo"), 0L)
  expect_length(exprs$expressions[[2L]]$xml_find_function_calls("bar"), 1L)

  expect_length(exprs$expressions[[5L]]$xml_find_function_calls("foo"), 2L)
  expect_length(exprs$expressions[[5L]]$xml_find_function_calls("bar"), 1L)
  expect_length(exprs$expressions[[5L]]$xml_find_function_calls(c("foo", "bar")), 3L)

  # file-level source expression contains all function calls
  expect_length(exprs$expressions[[6L]]$xml_find_function_calls("foo"), 4L)
  expect_length(exprs$expressions[[6L]]$xml_find_function_calls("bar"), 2L)
  expect_length(exprs$expressions[[6L]]$xml_find_function_calls(c("foo", "bar")), 6L)

  # Also check order is retained:
  expect_identical(
    exprs$expressions[[6L]]$xml_find_function_calls(c("foo", "bar")),
    xml2::xml_find_all(exprs$expressions[[6L]]$full_xml_parsed_content, "//SYMBOL_FUNCTION_CALL/parent::expr")
  )

  # Check naming and full cache
  expect_identical(
    exprs$expressions[[6L]]$xml_find_function_calls(NULL),
    exprs$expressions[[6L]]$xml_find_function_calls(c("foo", "bar"))
  )
  expect_named(
    exprs$expressions[[5L]]$xml_find_function_calls(c("foo", "bar"), keep_names = TRUE),
    c("foo", "foo", "bar")
  )

  # include_s4_slots
  expect_identical(
    exprs$expressions[[6L]]$xml_find_function_calls(NULL, include_s4_slots = TRUE),
    exprs$expressions[[6L]]$xml_find_function_calls(c("foo", "bar", "baz"), include_s4_slots = TRUE)
  )
  expect_named(
    exprs$expressions[[5L]]$xml_find_function_calls(NULL, keep_names = TRUE, include_s4_slots = TRUE),
    c("foo", "foo", "bar", "baz")
  )
})

test_that("#1262: xml_parsed_content gets returned as missing even if there's no parsed_content", {
  tempfile <- withr::local_tempfile(lines = '"\\R"')

  source_expressions <- get_source_expressions(tempfile)
  expect_null(source_expressions$expressions[[1L]]$full_parsed_content)
  expect_identical(source_expressions$expressions[[1L]]$full_xml_parsed_content, xml2::xml_missing())
})

test_that("#743, #879, #1406: get_source_expressions works on R files matching a knitr pattern", {
  # from #743
  tempfile <- withr::local_tempfile(
    lines = trim_some('
      create_template <- function(x) {
        sprintf("
      ```{r code}
      foo <- function(x) x+%d
      foo(5)
      ```", x)
      }
    ')
  )
  source_expressions <- get_source_expressions(tempfile)
  expect_null(source_expressions$error)

  # from #879
  tempfile <- withr::local_tempfile(
    lines = trim_some('
      # `r print("7")`
      function() 2<=3
    ')
  )
  source_expressions <- get_source_expressions(tempfile)
  expect_null(source_expressions$error)

  # from #1406
  tempfile <- withr::local_tempfile()
  writeLines(c("x <- '", "```{r}", "'"), con = tempfile)
  source_expressions <- get_source_expressions(tempfile)
  expect_null(source_expressions$error)
})

test_that("Syntax errors in Rmd or qmd don't choke lintr", {
  tmp <- withr::local_tempfile(lines = c(
    "```{r}",
    "if (TRUE) {",
    "  1",
    # missing `}` here
    "if (TRUE) {",
    "}",
    "```"
  ))
  expect_silent(get_source_expressions(tmp))
})

test_that("Indented Rmd chunks don't cause spurious whitespace lints", {
  tmp <- withr::local_tempfile(lines = c(
    "* An enumeration item with code:",
    "",
    "  ```{r}",
    '  "properly indented"',
    "  ```",
    "",
    "# New section",
    "",
    "```{r unindented_chunk}",
    '  "improperly indented"',
    "```",
    "",
    "# Third section",
    "",
    "   ```{r staggered}",
    ' "leftmost code"',
    '  "further right"',
    '   "aligned with code gate"',
    "   ```"
  ))

  parsed_lines <- get_source_expressions(tmp)$lines
  expect_identical(parsed_lines[4L], '"properly indented"', ignore_attr = "names")
  expect_identical(parsed_lines[10L], '  "improperly indented"', ignore_attr = "names")
  expect_identical(parsed_lines[16L], '"leftmost code"', ignore_attr = "names")
  expect_identical(parsed_lines[17L], ' "further right"', ignore_attr = "names")
  expect_identical(parsed_lines[18L], '  "aligned with code gate"', ignore_attr = "names")
})

test_that("Reference chunks in Sweave/Rmd are ignored", {
  example_rnw <- system.file("Sweave", "example-1.Rnw", package = "utils")
  # ensure such a chunk continues to exist upstream
  expect_true(any(grepl("^\\s*<<[^>]*>>\\s*$", readLines(example_rnw))))
  expect_silent(lint(example_rnw))
})

# NB: this is just a cursory test for linters not to
#   fail on files where the XML content is xml_missing;
#   the main linter test files provide more thorough
#   evidence that things are working as intended.
bad_source <- withr::local_tempfile(lines = c("a <- 1L", "b <- 2L"))
expressions <- get_source_expressions(bad_source)$expressions

# "zap" the xml_parsed_content to be xml_missing -- this gets
#   around the issue of creating a file that fails to parse now,
#   but later fails in a different way -> xml not missing.
for (ii in seq_along(expressions)) {
  if ("xml_parsed_content" %in% names(expressions[[ii]])) {
    expressions[[ii]]$xml_parsed_content <- xml2::xml_missing()
  } else {
    expressions[[ii]]$full_xml_parsed_content <- xml2::xml_missing()
  }
}
param_df <- expand.grid(
  linter = available_linters(tags = NULL)$linter,
  expression_idx = seq_along(expressions),
  stringsAsFactors = FALSE
)
param_df$.test_name <- with(param_df, sprintf("%s on expression %d", linter, expression_idx))

patrick::with_parameters_test_that(
  "linters pass with xml_missing() content",
  {
    if (linter == "backport_linter") {
      # otherwise we test the trivial linter (#2339)
      linter <- backport_linter(r_version = "3.6.0")
    } else {
      if (linter == "cyclocomp_linter") {
        skip_if_not_installed("cyclocomp")
      }
      linter <- eval(call(linter))
    }
    expression <- expressions[[expression_idx]]
    is_valid_linter_level <-
      (is_linter_level(linter, "expression") && is_lint_level(expression, "expression")) ||
      (is_linter_level(linter, "file") && is_lint_level(expression, "file"))
    if (is_valid_linter_level) {
      expect_no_warning({
        lints <- linter(expression)
      })
      expect_length(lints, 0L)
    } else {
      # suppress "empty test" skips
      expect_true(TRUE)
    }
  },
  .test_name = param_df$.test_name,
  linter = param_df$linter,
  expression_idx = param_df$expression_idx
)

test_that("invalid function definition parser failure lints", {
  expect_lint(
    "function(a = 1, a = 1) NULL",
    rex::rex("Repeated formal argument 'a'."),
    linters = list()
  )
})

test_that("Disallowed embedded null gives parser failure lint", {
  expect_lint(
    "'\\0'",
    rex::rex("Nul character not allowed."),
    linters = list()
  )
})

test_that("parser warnings are captured in output", {
  with_content_to_parse("1e-3L", {
    expect_length(warning, 1L)
    expect_s3_class(warning, "lints")
  })
  with_content_to_parse("1e-3L; 1e-3L", {
    expect_length(warning, 2L)
  })
  with_content_to_parse("1e-3L; 1.0L; 1.1L", {
    expect_length(warning, 3L)
  })
  with_content_to_parse("1e-3L\n1.0L\n1.1L", {
    expect_length(warning, 3L)
  })
  with_content_to_parse("1e-3L\n1+1\n1.0L\n2+2\n1.1L", {
    expect_length(warning, 3L)
  })
  with_content_to_parse("1e-3L\nc(", {
    expect_length(warning, 1L)
    expect_length(error, 8L)
  })
})

test_that("parser warnings generate lints", {
  expect_lint(
    "1e-3L",
    "non-integer value 1e-3L",
    linters = list()
  )
  expect_lint(
    "1e-3L; 1e-3L",
    list(
      list("non-integer value 1e-3L", column_number = 1L),
      list("non-integer value 1e-3L", column_number = 8L)
    ),
    linters = list()
  )
  expect_lint(
    "1e-3L; 1.0L",
    list(
      list("non-integer value 1e-3L", column_number = 1L),
      list("integer literal 1\\.0L", column_number = 8L)
    ),
    linters = list()
  )
  expect_lint(
    trim_some("
      1e-3L
      1 + 1
      1.0L
      2 + 2
      1.1L
      3 + 3
      2.2L
      4 + 4
      2.0L
      5 + 5
      2e-3L
      # don't match strictly on regex, use parse tree
      # 3.0L
      # 3e-3L
      # 3.3L
      '4.0L'
      '4e-3L'
      '4.4L'
    "),
    list(
      list("non-integer value 1e-3L", line_number = 1L),
      list("integer literal 1\\.0L", line_number = 3L),
      list("integer literal 1.1L contains decimal", line_number = 5L),
      list("integer literal 2\\.2L contains decimal", line_number = 7L),
      list("integer literal 2\\.0L", line_number = 9L),
      list("non-integer value 2e-3L", line_number = 11L)
    ),
    linters = list()
  )
  # parser catches warning before erroring
  expect_lint(
    "1e-3L; c(",
    list(
      list("non-integer value 1e-3L", type = "warning"),
      list("unexpected end of input", type = "error")
    ),
    linters = list()
  )
})

test_that("get_source_expressions() handles unmarked UTF-8 lines correctly", {
  skip_if_not_utf8_locale()

  tf <- withr::local_tempfile()
  writeLines('x <- "\u00e4"', tf, useBytes = TRUE)
  text_native <- readLines(tf)

  src_exprs <- get_source_expressions(tf, lines = text_native)

  expect_length(src_exprs$expressions, 2L)
  expr <- src_exprs$expressions[[1L]]

  pd <- expr$parsed_content
  str_const <- pd[pd$token == "STR_CONST", ]
  expect_identical(str_const$text, '"\u00e4"')
  expect_identical(str_const$col1, 6L)
  expect_identical(str_const$col2, 8L)
})

test_that("commented_code_linter skips allowed usages", {
  linter <- commented_code_linter()

  expect_no_lint("blah", linter)
  expect_no_lint("#' blah <- 1", linter)
  expect_no_lint("a <- 1\n# comment without code", linter)
  expect_no_lint("a <- 1\n## whatever", linter)

  expect_no_lint("TRUE", linter)
  expect_no_lint("#' @examples", linter)
  expect_no_lint("#' foo(1) # list(1)", linter) # comment in roxygen block ignored
  expect_no_lint("1+1 # gives 2", linter)
  expect_no_lint("# Non-existent:", linter)
  expect_no_lint("# 1-a", linter) # "-" removed from code operators
  expect_no_lint('1+1  # for example cat("123")', linter)

  # regression test for #451
  expect_no_lint("c('#a#' = 1)", linter)
})

test_that("commented_code_linter blocks disallowed usages", {
  lint_msg <- rex::rex("Remove commented code.")
  linter <- commented_code_linter()

  expect_lint("# blah <- 1", lint_msg, linter)

  expect_lint(
    "bleurgh <- fun_call(1) # other_call()",
    list(message = lint_msg, column_number = 26L),
    linter
  )

  expect_lint(
    " #blah <- 1",
    list(message = lint_msg, column_number = 3L),
    linter
  )

  # Regression test for #742, line number and comment number don't match up
  expect_lint(
    trim_some("
    # non-code comment
    line_without_comment <- 42L
     #blah <- 1
    "),
    list(message = lint_msg, line_number = 3L, column_number = 3L),
    linter
  )

  expect_lint(
    trim_some("
      d <- t.test(
        x = dplyr::starwars$height,
        # var.equal = TRUE,
        conf.level = .99,
        mu = 175
      )
    "),
    list(message = lint_msg, line_number = 3L),
    linter
  )

  expect_lint(
    trim_some("
      d <- t.test(
        x = dplyr::starwars$height
        #, var.equal = TRUE
        , conf.level = .99
        , mu = 175
      )
    "),
    list(message = lint_msg, line_number = 3L),
    linter
  )

  expect_lint("1+1 # cat('123')", lint_msg, linter)
  expect_lint("#expect_ftype(1e-12 , t)", lint_msg, linter)
})

test_that("commented_code_linter can detect operators in comments and lint correctly", {
  linter <- commented_code_linter()
  lint_msg <- rex::rex("Remove commented code.")

  test_ops <- c(
    "+", "=", "==", "!=", "<=", ">=", "<-", "<<-", "<", ">", "->",
    "->>", "%%", "/", "^", "*", "**", "|", "||", "&", "&&", "%>%",
    "%anything%"
  )

  for (op in test_ops) {
    expect_no_lint(paste("i", op, "1", collapse = ""), linter)
    expect_no_lint(paste("# something like i", op, "1", collapse = ""), linter)
    expect_lint(paste("# i", op, "1", collapse = ""), lint_msg, linter)
  }
})

test_that("commented_code_linter can detect operators in comments and lint correctly", {
  expect_lint(
    "# 1:3 |> sum()",
    rex::rex("Remove commented code."),
    commented_code_linter()
  )
})

test_that("commented_code_linter can detect commented code ending with pipes", {
  linter <- commented_code_linter()
  lint_msg <- rex::rex("Remove commented code.")

  expect_lint("# f() %>%", lint_msg, linter)
  expect_lint("# f() |>", lint_msg, linter)
})

test_that("package_hooks_linter skips allowed usages of packageStartupMessage() & library.dynam()", {
  linter <- package_hooks_linter()

  # allowed in .onAttach, not .onLoad
  expect_no_lint(".onAttach <- function(lib, pkg) packageStartupMessage('hi')", linter)
  # allowed in .onLoad, not .onAttach
  expect_no_lint(".onLoad <- function(lib, pkg) library.dynam()", linter)
})

test_that("package_hooks_linter blocks simple disallowed usages of packageStartupMessage() & library.dynam()", {
  linter <- package_hooks_linter()

  # inline version
  expect_lint(
    ".onLoad <- function(lib, pkg) packageStartupMessage('hi')",
    rex::rex("Put packageStartupMessage() calls in .onAttach()"),
    linter
  )

  # multiline version
  expect_lint(
    trim_some("
      .onAttach <- function(libname, pkgname) {
        library.dynam()
      }
    "),
    rex::rex("Put library.dynam() calls in .onLoad, not .onAttach()."),
    linter
  )

  # found at deeper nesting too
  expect_lint(
    trim_some("
      .onLoad <- function(libname, pkgname) {
        foo(bar(baz(packageStartupMessage('hi'))))
      }
    "),
    rex::rex("Put packageStartupMessage() calls in .onAttach()"),
    linter
  )
})

test_that("package_hooks_linter blocks simple disallowed usages of other blocked messaging functions", {
  linter <- package_hooks_linter()

  # inline version
  expect_lint(
    ".onLoad <- function(lib, pkg) cat('hi')",
    rex::rex("Don't use cat() in .onLoad()"),
    linter
  )

  # multiline version
  expect_lint(
    trim_some("
      .onAttach <- function(libname, pkgname) {
        writeLines('hi')
      }
    "),
    rex::rex("Don't use writeLines() in .onAttach()"),
    linter
  )

  expect_lint(
    trim_some("
      .onLoad <- function(libname, pkgname) {
        print('hi')
      }
    "),
    rex::rex("Don't use print() in .onLoad()"),
    linter
  )

  # found at deeper nesting too
  expect_lint(
    trim_some("
      .onAttach <- function(libname, pkgname) {
        foo(bar(baz(message('hi'))))
      }
    "),
    rex::rex("Don't use message() in .onAttach()"),
    linter
  )
})

test_that("package_hooks_linter skips valid .onLoad() and .onAttach() arguments", {
  linter <- package_hooks_linter()

  expect_no_lint(".onAttach <- function(lib, pkg) { }", linter)
  expect_no_lint(".onLoad <- function(lib, pkg) { }", linter)

  # args only need to start with those characters
  expect_no_lint(".onAttach <- function(libname, pkgpath) { }", linter)
  expect_no_lint(".onLoad <- function(libXXXX, pkgYYYY) { }", linter)
})

test_that("package_hooks_linter blocks invalid .onLoad() / .onAttach() arguments", {
  linter <- package_hooks_linter()
  onload_msg <- rex::rex(".onLoad() should take two arguments")

  expect_lint(
    ".onAttach <- function(xxx, pkg) { }",
    rex::rex(".onAttach() should take two arguments"),
    linter
  )
  expect_lint(".onLoad <- function(lib, yyy) { }", onload_msg, linter)
  # only one lint if both are wrong
  expect_lint(".onLoad <- function(xxx, yyy) { }", onload_msg, linter)

  # exactly two arguments required.
  # NB: QC.R allows ... arguments to be passed, but disallow this flexibility in the linter.
  expect_lint(".onLoad <- function() { }", onload_msg, linter)
  expect_lint(".onLoad <- function(lib) { }", onload_msg, linter)
  expect_lint(".onLoad = function(lib) { }", onload_msg, linter)
  expect_lint(".onLoad <- function(lib, pkg, third) { }", onload_msg, linter)
  expect_lint(".onLoad = function(lib, pkg, third) { }", onload_msg, linter)
  expect_lint(".onLoad <- function(lib, ...) { }", onload_msg, linter)
})

test_that("package_hooks_linter skips valid namespace loading", {
  linter <- package_hooks_linter()

  expect_no_lint(".onAttach <- function(lib, pkg) { requireNamespace('foo') }", linter)
  expect_no_lint(".onLoad <- function(lib, pkg) {  requireNamespace('foo') }", linter)
})

test_that("package_hooks_linter blocks attaching namespaces", {
  linter <- package_hooks_linter()

  expect_lint(
    ".onAttach <- function(lib, pkg) { require(foo) }",
    rex::rex("Don't alter the search() path in .onAttach() by calling require()."),
    linter
  )
  expect_lint(
    ".onLoad <- function(lib, pkg) { library(foo) }",
    rex::rex("Don't alter the search() path in .onLoad() by calling library()."),
    linter
  )
  expect_lint(
    ".onLoad <- function(lib, pkg) { installed.packages() }",
    rex::rex("Don't slow down package load by running installed.packages() in .onLoad()."),
    linter
  )

  # find at further nesting too
  expect_lint(
    ".onAttach <- function(lib, pkg) { a(b(c(require(foo)))) }",
    rex::rex("Don't alter the search() path in .onAttach() by calling require()."),
    linter
  )
  expect_lint(
    ".onLoad <- function(lib, pkg) { d(e(f(library(foo)))) }",
    rex::rex("Don't alter the search() path in .onLoad() by calling library()."),
    linter
  )
  expect_lint(
    ".onLoad = function(lib, pkg) { d(e(f(library(foo)))) }",
    rex::rex("Don't alter the search() path in .onLoad() by calling library()."),
    linter
  )
  expect_lint(
    ".onLoad <- function(lib, pkg) { g(h(i(installed.packages()))) }",
    rex::rex("Don't slow down package load by running installed.packages() in .onLoad()."),
    linter
  )

  # also find when used as names
  expect_lint(
    ".onAttach <- function(lib, pkg) { sapply(c('a', 'b', 'c'), require, character.only = TRUE) }",
    rex::rex("Don't alter the search() path in .onAttach() by calling require()."),
    linter
  )
  expect_lint(
    ".onAttach <- function(lib, pkg) { lapply(c('a', 'b', 'c'), library, character.only = TRUE) }",
    rex::rex("Don't alter the search() path in .onAttach() by calling library()"),
    linter
  )
})

test_that("package_hooks_linter skips valid 'teardown' hooks", {
  linter <- package_hooks_linter()

  expect_no_lint(".onDetach <- function(lib) { }", linter)
  expect_no_lint(".onDetach <- function(libname) { }", linter)

  expect_no_lint(".Last.lib <- function(lib) { }", linter)
  expect_no_lint(".Last.lib <- function(libname) { }", linter)

  expect_no_lint(".onUnload <- function(lib) { }", linter)
  expect_no_lint(".onUnload <- function(libpath) { }", linter)
})

test_that("package_hooks_linter catches usage of library.dynam.unload()", {
  linter <- package_hooks_linter()

  expect_lint(
    ".onDetach <- function(lib) { library.dynam.unload() }",
    rex::rex("Use library.dynam.unload() calls in .onUnload(), not .onDetach()."),
    linter
  )
  expect_lint(
    ".Last.lib <- function(lib) { library.dynam.unload() }",
    rex::rex("Use library.dynam.unload() calls in .onUnload(), not .Last.lib()."),
    linter
  )
  # expected usage is in .onUnload
  expect_no_lint(".onUnload <- function(lib) { library.dynam.unload() }", linter)
})

test_that("package_hooks_linter detects bad argument names in 'teardown' hooks", {
  linter <- package_hooks_linter()
  lint_msg_part <- " should take one argument starting with 'lib'"

  expect_lint(
    ".onDetach <- function(xxx) { }",
    rex::rex(".onDetach()", lint_msg_part),
    linter
  )

  # assignment operator doesn't matter
  expect_lint(
    ".onDetach = function(xxx) { }",
    rex::rex(".onDetach()", lint_msg_part),
    linter
  )
  expect_lint(
    ".Last.lib <- function(yyy) { }",
    rex::rex(".Last.lib()", lint_msg_part),
    linter
  )
  expect_lint(
    ".onUnload <- function(pkg) { }",
    rex::rex(".onUnload()", lint_msg_part),
    linter
  )

  # exactly one argument required.
  # NB: QC.R allows ... arguments to be passed, but disallow this flexibility in the linter.
  expect_lint(
    ".onDetach <- function() { }",
    rex::rex(".onDetach()", lint_msg_part),
    linter
  )
  expect_lint(
    ".Last.lib <- function(lib, pkg) { }",
    rex::rex(".Last.lib()", lint_msg_part),
    linter
  )
  expect_lint(
    ".onDetach <- function(...) { }",
    rex::rex(".onDetach()", lint_msg_part),
    linter
  )
  expect_lint(
    ".onUnload <- function() { }",
    rex::rex(".onUnload()", lint_msg_part),
    linter
  )
  expect_lint(
    ".onUnload <- function(lib, pkg) { }",
    rex::rex(".onUnload()", lint_msg_part),
    linter
  )
  expect_lint(
    ".onUnload <- function(...) { }",
    rex::rex(".onUnload()", lint_msg_part),
    linter
  )
})

test_that("function shorthand is handled", {
  linter <- package_hooks_linter()

  expect_lint(
    ".onLoad <- \\(lib, pkg) packageStartupMessage('hi')",
    rex::rex("Put packageStartupMessage() calls in .onAttach()"),
    linter
  )
  expect_lint(
    ".onAttach <- \\(xxx, pkg) { }",
    rex::rex(".onAttach() should take two arguments"),
    linter
  )
  expect_lint(
    ".onAttach <- \\(lib, pkg) { require(foo) }",
    rex::rex("Don't alter the search() path in .onAttach() by calling require()."),
    linter
  )
  expect_lint(
    ".onDetach <- \\(lib) { library.dynam.unload() }",
    rex::rex("Use library.dynam.unload() calls in .onUnload(), not .onDetach()."),
    linter
  )
  expect_lint(
    ".onDetach <- \\(xxx) { }",
    rex::rex(".onDetach() should take one argument starting with 'lib'."),
    linter
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      .onLoad <- function(xxx, yyy) { }
      .onAttach <- function(aaa, bbb) { }
    }"),
    list(
      list(".onLoad", line_number = 2L),
      list(".onAttach", line_number = 3L)
    ),
    package_hooks_linter()
  )
})

test_that("sort_linter skips allowed usages", {
  linter <- sort_linter()

  expect_no_lint("order(y)", linter)

  expect_no_lint("y[order(x)]", linter)

  # If another function is intercalated, don't fail
  expect_no_lint("x[c(order(x))]", linter)

  expect_no_lint("x[order(y, x)]", linter)
  expect_no_lint("x[order(x, y)]", linter)
  # pretty sure this never makes sense, but test anyway
  expect_no_lint("x[order(y, na.last = x)]", linter)

  # is.unsorted false positives
  expect_no_lint("identical(sort(x), y)", linter)
  expect_no_lint("identical(sort(foo(x)), x)", linter)
})


test_that("sort_linter blocks simple disallowed usages for x[order(x)] cases", {
  linter <- sort_linter()
  lint_message <- rex::rex("sort(", anything, ") is better than")

  expect_lint("x[order(x)]", lint_message, linter)

  # Works with extra args in order()
  expect_lint("x[order(x, decreasing = TRUE)]", lint_message, linter)

  # ...even in disorder
  expect_lint("x[order(decreasing = TRUE, x)]", lint_message, linter)
})

test_that("sort_linter produces customized warning message", {
  linter <- sort_linter()

  expect_lint(
    "y[order(y)]",
    rex::rex("sort(y, na.last = TRUE) is better than y[order(y)]."),
    linter
  )

  # We capture the correct variable symbol
  expect_lint(
    "y + x[order(x)]",
    rex::rex("sort(x, na.last = TRUE) is better than x[order(x)]."),
    linter
  )

  # Default na.last = TRUE is overwritten if na.last is already provided
  expect_lint(
    "x[order(x, na.last = FALSE)]",
    rex::rex("sort(x, na.last = FALSE) is better than x[order(x, na.last = FALSE)]."),
    linter
  )

  expect_lint(
    "x[order(x, decreasing = FALSE)]",
    rex::rex("sort(x, decreasing = FALSE, na.last = TRUE) is better than x[order(x, decreasing = FALSE)]."),
    linter
  )

  expect_lint(
    "f()[order(f())]",
    rex::rex("sort(f(), na.last = TRUE) is better than f()[order(f())]"),
    linter
  )

  # comment torture
  expect_lint(
    trim_some("
      x[        # comment
        order(  # comment
        x       # comment
        ,       # comment
        na.last # comment
        =       # comment
        FALSE   # comment
        )       # comment
      ]
    "),
    rex::rex("sort(x, na.last = FALSE)"),
    linter
  )

  expect_lint(
    trim_some("
      f( # comment
      )[order(f())]
    "),
    rex::rex("sort(f(), na.last = TRUE) is better than f()[order(f())]"),
    linter
  )
})

test_that("sort_linter works with multiple lints in a single expression", {
  linter <- sort_linter()

  expect_lint(
    "c(
      x[order(x)],
      y[order(y, decreasing = TRUE, na.last = FALSE)]
    )",
    list(
      rex::rex("sort(x, na.last = TRUE) is better than x[order(x)]."),
      rex::rex(
        "sort(y, decreasing = TRUE, na.last = FALSE)",
        anything,
        "y[order(y, decreasing = TRUE, na.last = FALSE)]."
      )
    ),
    linter
  )
})

test_that("sort_linter skips usages calling sort arguments", {
  linter <- sort_linter()

  # any arguments to sort --> not compatible
  expect_no_lint("sort(x, decreasing = TRUE) == x", linter)
  expect_no_lint("sort(x, na.last = TRUE) != x", linter)
  expect_no_lint("sort(x, method_arg = TRUE) == x", linter)
})

test_that("sort_linter skips when inputs don't match", {
  linter <- sort_linter()

  expect_no_lint("sort(x) == y", linter)
  expect_no_lint("sort(x) == foo(x)", linter)
  expect_no_lint("sort(foo(x)) == x", linter)
})

test_that("sort_linter blocks simple disallowed usages for is.sorted cases", {
  linter <- sort_linter()
  unsorted_msg <- rex::rex("Use is.unsorted(x) to test the unsortedness of a vector.")
  sorted_msg <- rex::rex("Use !is.unsorted(x) to test the sortedness of a vector.")
  rev_sort_msg <- rex::rex("Use sort(x, decreasing = TRUE) instead of rev(sort(x)).")

  expect_lint("sort(x) == x", sorted_msg, linter)
  expect_lint("identical(x, sort(x))", sorted_msg, linter)
  expect_lint("rev(sort(x))", rev_sort_msg, linter)

  # argument order doesn't matter
  expect_lint("x == sort(x)", sorted_msg, linter)
  expect_lint("identical(sort(x), x)", sorted_msg, linter)

  # inverted version
  expect_lint("sort(x) != x", unsorted_msg, linter)
  expect_lint("!identical(x, sort(x))", unsorted_msg, linter)

  # expression matching
  expect_lint("sort(foo(x)) == foo(x)", sorted_msg, linter)
  expect_lint("identical(foo(x), sort(foo(x)))", sorted_msg, linter)
  expect_lint("identical(sort(foo(x)), foo(x))", sorted_msg, linter)
  expect_lint("rev(sort(foo(x)))", rev_sort_msg, linter)

  expect_lint(
    trim_some("
      sort(foo(x # comment
      )) == foo(x)
    "),
    sorted_msg,
    linter
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      x == sort(x)
      y[order(y)]
    }"),
    list(
      list(rex::rex("is.unsorted(x)"), line_number = 2L),
      list(rex::rex("sort(y"), line_number = 3L)
    ),
    sort_linter()
  )
})

test_that("duplicate_argument_linter doesn't block allowed usages", {
  linter <- duplicate_argument_linter()

  expect_no_lint("fun(arg = 1)", linter)
  expect_no_lint("fun('arg' = 1)", linter)
  expect_no_lint("fun(`arg` = 1)", linter)
  expect_no_lint("'fun'(arg = 1)", linter)
  expect_no_lint("(function(x, y) x + y)(x = 1)", linter)
  expect_no_lint("dt[i = 1]", linter)
})

test_that("duplicate_argument_linter blocks disallowed usages", {
  linter <- duplicate_argument_linter()
  lint_msg <- rex::rex("Avoid duplicate arguments in function calls.")

  expect_lint("fun(arg = 1, arg = 2)", lint_msg, linter)
  expect_lint("fun(arg = 1, 'arg' = 2)", lint_msg, linter)
  expect_lint("fun(arg = 1, `arg` = 2)", lint_msg, linter)
  expect_lint("'fun'(arg = 1, arg = 2)", lint_msg, linter)
  expect_lint("(function(x, y) x + y)(x = 1, x = 2)", lint_msg, linter)
  expect_lint("dt[i = 1, i = 2]", lint_msg, linter)

  expect_lint(
    trim_some("
      list(
        var = 1,
        var = 2
      )
    "),
    lint_msg,
    linter
  )
})

test_that("duplicate_argument_linter respects except argument", {
  linter_list <- duplicate_argument_linter(except = "list")
  linter_all <- duplicate_argument_linter(except = character())

  expect_no_lint(
    trim_some("
      list(
        var = 1,
        var = 2
      )
    "),
    linter_list
  )

  expect_no_lint(
    trim_some("
      (function(x, y) x + y)(x = 1)
      list(var = 1, var = 2)
    "),
    linter_list
  )

  expect_lint(
    "fun(`
` = 1, `
` = 2)",
    rex::rex("Avoid duplicate arguments in function calls."),
    linter_all
  )

  expect_lint(
    "function(arg = 1, arg = 1) {}",
    rex::rex("Repeated formal argument 'arg'."),
    linter_all
  )
})

test_that("doesn't lint duplicated arguments in allowed functions", {
  linter <- duplicate_argument_linter()

  expect_no_lint(
    trim_some("
      x %>%
       dplyr::mutate(
         col = a + b,
         col = col + d
       )
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      x %>%
       dplyr::transmute(
         col = a + b,
         col = col / 2.5
       )
    "),
    linter
  )

  expect_no_lint(
    trim_some("
      x |>
      dplyr::mutate(
        col = col |> str_replace('t', '') |> str_replace('\\\\s+$', 'xxx')
      )
    "),
    linter
  )
})

test_that("interceding comments don't trip up logic", {
  linter <- duplicate_argument_linter()
  lint_msg <- rex::rex("Avoid duplicate arguments")

  # comment before the EQ_SUB
  # actually this case "just works" even before #2402 since
  #   get_r_string() returns NA for both argument names
  expect_lint(
    trim_some("
      fun(
        arg # xxx
        = 1,
        arg # yyy
        = 2
      )
    "),
    list(lint_msg, line_number = 4L),
    linter
  )

  expect_lint(
    trim_some("
      fun(
        arg # xxx
        = 1,
        arg = 2
      )
    "),
    list(lint_msg, line_number = 4L),
    linter
  )

  expect_lint(
    trim_some("
      fun(
        arg = 1,
        arg # yyy
        = 2
      )
    "),
    list(lint_msg, line_number = 3L),
    linter
  )

  # comment after the EQ_SUB
  expect_lint(
    trim_some("
      fun(
        arg = # xxx
        1,
        arg = # yyy
        2
      )
    "),
    list(lint_msg, line_number = 4L),
    linter
  )

  expect_lint(
    trim_some("
      fun(
        arg = # xxx
        1,
        arg = 2
      )
    "),
    list(lint_msg, line_number = 4L),
    linter
  )

  expect_lint(
    trim_some("
      fun(
        arg = 1,
        arg = # yyy
        2
      )
    "),
    list(lint_msg, line_number = 3L),
    linter
  )

  # comment after the arg value
  expect_lint(
    trim_some("
      fun(
        arg = 1 # xxx
        ,
        arg = 2 # yyy
      )
    "),
    list(lint_msg, line_number = 4L),
    linter
  )

  expect_lint(
    trim_some("
      fun(
        arg = 1 # xxx
        ,
        arg = 2
      )
    "),
    list(lint_msg, line_number = 4L),
    linter
  )

  expect_lint(
    trim_some("
      fun(
        arg = 1,
        arg = 2 # yyy
      )
    "),
    list(lint_msg, line_number = 3L),
    linter
  )

  # comment after the OP-COMMA
  expect_lint(
    trim_some("
      fun(
        arg = 1, # xxx
        arg = 2 # yyy
      )
    "),
    list(lint_msg, line_number = 3L),
    linter
  )

  expect_lint(
    trim_some("
      fun(
        arg = 1, # xxx
        arg = 2
      )
    "),
    list(lint_msg, line_number = 3L),
    linter
  )
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Avoid duplicate arguments")

  expect_lint(
    trim_some("{
      c(a = 1, a = 2, a = 3)
      list(b = 1, b = 2, b = 3)
    }"),
    list(
      list(lint_msg, line_number = 2L, column_number = 12L),
      list(lint_msg, line_number = 2L, column_number = 19L),
      list(lint_msg, line_number = 3L, column_number = 15L),
      list(lint_msg, line_number = 3L, column_number = 22L)
    ),
    duplicate_argument_linter()
  )
})

test_that("rep_len_linter skips allowed usages", {
  linter <- rep_len_linter()

  # only catch length.out usages
  expect_no_lint("rep(x, y)", linter)
  expect_no_lint("rep(1:10, 2)", linter)
  expect_no_lint("rep(1:10, 10:1)", linter)

  # usage of each is not compatible with rep_len; see ?rep.
  expect_no_lint("rep(x, each = 4, length.out = 50)", linter)
  # each is implicitly the 4th positional argument. a very strange usage
  #   (because length.out is ignored), but doesn't hurt to catch it
  expect_no_lint("rep(a, b, length.out = c, d)", linter)
  # ditto for implicit length.out=
  expect_no_lint("rep(a, b, c, d)", linter)
})

test_that("rep_len_linter blocks simple disallowed usages", {
  linter <- rep_len_linter()
  lint_msg <- rex::rex("Use rep_len(x, n) instead of rep(x, length.out = n).")

  # only catch length.out usages
  expect_lint("rep(x, length.out = 4L)", lint_msg, linter)

  # implicit times= argument; length.out has priority over times= (see ?rep),
  #   so we still lint since it's as if times= is not supplied.
  # (notice here that the base behavior is odd -- one might expect output like
  #   head(rep(1:10, 10:1), 50), but instead we get rep(1:10, length.out = 50))
  expect_lint("rep(1:10, 10:1, length.out = 50)", lint_msg, linter)

  # ditto for explicit times= argument
  expect_lint("rep(1:10, times = 10:1, length.out = 50)", lint_msg, linter)

  # implicit usage in third argument
  expect_lint("rep(1:10, 10:1, 50)", lint_msg, linter)
})

test_that("vectorized lints work", {
  lint_msg <- rex::rex("Use rep_len(x, n) instead of rep(x, length.out = n).")

  expect_lint(
    trim_some("{
      rep(x, y)
      rep(1:10, length.out = 50)
      rep(x, each = 4, length.out = 50)
      rep(x, length.out = 50)
    }"),
    list(
      list(lint_msg, line_number = 3L),
      list(lint_msg, line_number = 5L)
    ),
    rep_len_linter()
  )
})

# fuzzer disable: comment_injection
test_that("returns the correct linting", {
  ops <- c(
    "+",
    "-",
    "~",
    "=",
    "==",
    "!=",
    "<=",
    ">=",
    "<-",
    ":=",
    "<<-",
    "<",
    ">",
    "->",
    "->>",
    "%%",
    "/",
    "*",
    "|",
    "||",
    "&",
    "&&",
    "%>%",
    "%Anything%",
    "%+%",
    NULL
  )

  linter <- infix_spaces_linter()
  lint_msg <- rex::rex("Put spaces around all infix operators.")

  expect_no_lint("blah", linter)

  for (op in ops) {
    expect_no_lint(paste0("1 ", op, " 2"), linter)
    expect_no_lint(paste0("1 ", op, "\n2"), linter)
    expect_no_lint(paste0("1 ", op, "\n 2"), linter)

    expect_lint(paste0("1", op, "2"), lint_msg, linter)

    # unary plus and minus can have no space before them
    if (!op %in% ops[1L:2L]) {
      expect_lint(paste0("1 ", op, "2"), lint_msg, linter)
    }

    expect_lint(paste0("1", op, " 2"), lint_msg, linter)
  }

  expect_no_lint("b <- 2E+4", linter)
  expect_no_lint("a <- 1e-3", linter)
  expect_no_lint("a[-1]", linter)
  expect_no_lint("a[-1 + 1]", linter)
  expect_no_lint("a[1 + -1]", linter)

  expect_lint("fun(a=1)", lint_msg, linter)
})

test_that("The three `=` are all linted", {
  linter <- infix_spaces_linter()
  lint_msg <- rex::rex("Put spaces around all infix operators.")

  # EQ_ASSIGN in the parse data
  expect_lint("a=1", lint_msg, linter)
  # EQ_FORMALS in the parse data
  expect_lint("foo <- function(x=1) {}", lint_msg, linter)
  # EQ_SUB in the parse data
  expect_lint("foo(x=1)", lint_msg, linter)
})

test_that("exclude_operators works", {
  lint_msg <- rex::rex("Put spaces around all infix operators.")

  expect_no_lint("a+b", infix_spaces_linter(exclude_operators = "+"))
  expect_no_lint(
    trim_some("
      a+b
      a-b
    "),
    infix_spaces_linter(exclude_operators = c("+", "-"))
  )

  # operators match on text, not hidden node
  expect_lint("a<<-1", lint_msg, infix_spaces_linter(exclude_operators = "<-"))
  expect_no_lint("a<<-1", infix_spaces_linter(exclude_operators = "<<-"))
  expect_lint("a:=1", lint_msg, infix_spaces_linter(exclude_operators = "<-"))
  expect_no_lint("a:=1", infix_spaces_linter(exclude_operators = ":="))
  expect_lint("a->>1", lint_msg, infix_spaces_linter(exclude_operators = "->"))
  expect_no_lint("a->>1", infix_spaces_linter(exclude_operators = "->>"))
  expect_no_lint("a%any%1", infix_spaces_linter(exclude_operators = "%%"))
  expect_no_lint("function(a=1) { }", infix_spaces_linter(exclude_operators = "="))
  expect_no_lint("foo(a=1)", infix_spaces_linter(exclude_operators = "="))
})

# more tests specifically for assignment
test_that("assignment cases return the correct linting", {
  linter <- infix_spaces_linter()
  lint_msg <- rex::rex("Put spaces around all infix operators.")

  expect_no_lint("fun(blah =  1)", linter)

  expect_no_lint("blah <- 1", linter)
  expect_no_lint("blah = 1", linter)

  expect_no_lint("\"my  =  variable\" <- 42.0", linter)

  expect_no_lint("if (0 <  1) x <- 42L", linter)
  expect_no_lint(
    trim_some("
    if (0 < 1) {
      x <- 42L
    }"),
    linter
  )
  expect_no_lint("my = bad = variable = name <- 2.0", linter)

  expect_lint("blah<-  1", lint_msg, linter)
  expect_lint("blah  <-1", lint_msg, linter)
  expect_lint("blah=  1", lint_msg, linter)
  expect_lint("blah  =1", lint_msg, linter)
})

test_that("infix_spaces_linter can allow >1 spaces optionally", {
  linter <- infix_spaces_linter(allow_multiple_spaces = FALSE)
  lint_msg <- rex::rex("Put exactly one space on each side of infix operators.")

  expect_lint("x  ~  1", lint_msg, linter)
  expect_lint("x  - 1", lint_msg, linter)
  expect_lint("x /  1", lint_msg, linter)
})

test_that("exception for box::use()", {
  linter <- infix_spaces_linter()

  expect_no_lint("box::use(a/b)", linter)
  expect_no_lint("box::use(./a/b)", linter)
  expect_no_lint(
    trim_some("
      box::use(
        a,
        a/b,
        ../a,
        alias = a/b/c[xyz = abc, ...],
      )
    "),
    linter
  )
})

test_that("multi-line, multi-expression case is caught", {
  expect_lint(
    trim_some("
      x +
        y+
        z
    "),
    rex::rex("Put spaces around all infix operators."),
    infix_spaces_linter()
  )
})

test_that("Rules around missing arguments are respected", {
  linter <- infix_spaces_linter()
  lint_msg <- rex::rex("Put spaces around all infix operators.")

  expect_no_lint("switch(a = , b = 2)", linter)
  expect_no_lint("alist(missing_arg = )", linter)

  expect_lint("switch(a =, b = 2)", lint_msg, linter)
  expect_lint("alist(missing_arg =)", lint_msg, linter)
})

test_that("native pipe is supported", {
  linter <- infix_spaces_linter()

  expect_no_lint("a |> foo()", linter)
  expect_lint("a|>foo()", rex::rex("Put spaces around all infix operators."), linter)
})

test_that("mixed unary & binary operators aren't mis-lint", {
  expect_lint(
    "-1-1",
    list(
      message = rex::rex("Put spaces around all infix operators."),
      column_number = 3L
    ),
    infix_spaces_linter()
  )
})

test_that("parse tags are accepted by exclude_operators", { # nofuzz: assignment
  expect_no_lint("sum(x, na.rm=TRUE)", infix_spaces_linter(exclude_operators = "EQ_SUB"))
  expect_no_lint("function(x, na.rm=TRUE) { }", infix_spaces_linter(exclude_operators = "EQ_FORMALS"))
  expect_no_lint("x=1", infix_spaces_linter(exclude_operators = "EQ_ASSIGN"))

  # uses parse_tag
  expect_no_lint("1+1", infix_spaces_linter(exclude_operators = "'+'"))

  # mixing
  text <- "x=function(a=foo(bar=1)) { }"
  col_assign <- list(column_number = 2L)
  col_formals <- list(column_number = 13L)
  col_sub <- list(column_number = 21L)
  expect_no_lint(text, infix_spaces_linter(exclude_operators = c("EQ_SUB", "EQ_FORMALS", "EQ_ASSIGN")))
  expect_lint(text, col_assign, infix_spaces_linter(exclude_operators = c("EQ_SUB", "EQ_FORMALS")))
  expect_lint(text, col_formals, infix_spaces_linter(exclude_operators = c("EQ_SUB", "EQ_ASSIGN")))
  expect_lint(text, col_sub, infix_spaces_linter(exclude_operators = c("EQ_FORMALS", "EQ_ASSIGN")))
  expect_lint(text, list(col_assign, col_formals), infix_spaces_linter(exclude_operators = "EQ_SUB"))
  expect_lint(text, list(col_assign, col_sub), infix_spaces_linter(exclude_operators = "EQ_FORMALS"))
  expect_lint(text, list(col_formals, col_sub), infix_spaces_linter(exclude_operators = "EQ_ASSIGN"))
})

test_that("lints vectorize", { # nofuzz: assignment
  lint_msg <- rex::rex("Put spaces around all infix operators.")

  expect_lint(
    trim_some("{
      a<-1
      1/2
      b<-c<-2
      d+e+f+g/3
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L),
      list(lint_msg, line_number = 4L, column_number = 4L),
      list(lint_msg, line_number = 4L, column_number = 7L),
      list(lint_msg, line_number = 5L, column_number = 4L),
      list(lint_msg, line_number = 5L, column_number = 6L),
      list(lint_msg, line_number = 5L, column_number = 8L),
      list(lint_msg, line_number = 5L, column_number = 10L)
    ),
    infix_spaces_linter()
  )
})
# fuzzer enable: comment_injection

test_that("list2df_linter skips allowed usages", {
  linter <- list2df_linter()

  expect_no_lint("cbind.data.frame(x, x)", linter)
  expect_no_lint("do.call(mean, x)", linter)
  expect_no_lint("do.call('c', x)", linter)

  # Other cbind methods
  expect_no_lint("do.call(cbind, x)", linter)

  # Anonymous function
  expect_no_lint("do.call(function(x) x, l)", linter)
})

test_that("list2df_linter blocks simple disallowed usages", {
  linter <- list2df_linter()
  lint_message <- rex::rex("use `data.frame(lst)`")

  expect_lint("do.call(cbind.data.frame, x)", lint_message, linter)
  expect_lint("do.call('cbind.data.frame', x)", lint_message, linter)
})

test_that("lints vectorize", {
  lint_message <- rex::rex("use `data.frame(lst)`")

  expect_lint(
    trim_some("{
      cbind(a, b)
      do.call(cbind.data.frame, x)
      do.call(function(x) x, l)
      do.call('cbind.data.frame', y)
    }"),
    list(
      list(lint_message, line_number = 3L, column_number = 3L),
      list(lint_message, line_number = 5L, column_number = 3L)
    ),
    list2df_linter()
  )
})

test_that("terminal_close_linter skips allowed cases", {
  linter <- terminal_close_linter()

  lines <- trim_some("
    foo <- function(bar) {
      tmp <- tempfile()
      on.exit(close(tmp))
      writeLines(bar, tmp)
      return(invisible())
    }
  ")
  expect_no_lint(lines, linter)

  lines <- trim_some("
    foo <- function(bar) {
      close <- bar + 1
      return(close)
    }
  ")
  expect_no_lint(lines, linter)

  lines <- trim_some("
    foo <- function(bar) {
      close <- bar + 1
      close
    }
  ")
  expect_no_lint(lines, linter)

  lines <- trim_some("
    foo <- \\(bar) {
      close <- bar + 1
      return(close)
    }
  ")
  expect_no_lint(lines, linter)
})

test_that("terminal_close_linter blocks simple cases", {
  linter <- terminal_close_linter()
  lint_msg <- rex::rex("Use on.exit(close(x)) to close connections")

  expect_lint(
    trim_some("
      foo <- function(bar) {
        tmp <- tempfile()
        writeLines(bar, tmp)
        return(close(tmp))
      }
    "),
    list(lint_msg, line_number = 4L, column_number = 3L),
    linter
  )

  expect_lint(
    trim_some("
      foo <- function(bar) {
        tmp <- tempfile()
        writeLines(bar, tmp)
        close(tmp)
      }
    "),
    list(lint_msg, line_number = 4L, column_number = 3L),
    linter
  )

  # When multiple terminations happen, only lint the one
  expect_lint(
    trim_some("
      foo <- function(bar) {
        tmp1 <- tempfile()
        tmp2 <- tempfile()
        writeLines(bar, tmp1)
        writeLines(bar, tmp2)
        close(tmp1)
        close(tmp2)
      }
    "),
    list(lint_msg, line_number = 7L, column_number = 3L),
    linter
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      foo <- function() {
        tmp <- file(tempfile())
        writeLines(letters, tmp)
        close(tmp)
      }
      bar <- \\() {
        tmp <- file(tempfile())
        writeLines(letters, tmp)
        close(tmp)
      }
    }"),
    list(
      list("close connections", line_number = 5L),
      list("close connections", line_number = 10L)
    ),
    terminal_close_linter()
  )
})

test_that("expect_type_linter skips allowed usages", {
  linter <- expect_type_linter()

  # expect_type doesn't have an inverted version
  expect_no_lint("expect_true(!is.numeric(x))", linter)
  # NB: also applies to tinytest, but it's sufficient to test testthat
  expect_no_lint("testthat::expect_true(!is.numeric(x))", linter)

  # other is.<x> calls are not suitable for expect_type in particular
  expect_no_lint("expect_true(is.data.frame(x))", linter)

  # expect_type(x, ...) cannot be cleanly used here:
  expect_no_lint("expect_true(typeof(x) %in% c('builtin', 'closure'))", linter)

  # expect_type() doesn't have info= or label= arguments
  expect_no_lint("expect_equal(typeof(x), t, info = 'x should have type t')", linter)
  expect_no_lint("expect_equal(typeof(x), t, label = 'x type')", linter)
  expect_no_lint("expect_equal(typeof(x), t, expected.label = 'type')", linter)
  expect_no_lint("expect_true(is.double(x), info = 'x should be double')", linter)
})

test_that("expect_type_linter blocks simple disallowed usages", {
  linter <- expect_type_linter()

  expect_lint(
    "expect_equal(typeof(x), 'double')",
    rex::rex("expect_type(x, t) is better than expect_equal(typeof(x), t)"),
    linter
  )

  # expect_identical is treated the same as expect_equal
  expect_lint(
    "testthat::expect_identical(typeof(x), 'language')",
    rex::rex("expect_type(x, t) is better than expect_identical(typeof(x), t)"),
    linter
  )

  # different equivalent usage
  expect_lint(
    "expect_true(is.complex(foo(x)))",
    rex::rex("expect_type(x, t) is better than expect_true(is.<t>(x))"),
    linter
  )

  # yoda test with clear expect_type replacement
  expect_lint(
    "expect_equal('integer', typeof(x))",
    rex::rex("expect_type(x, t) is better than expect_equal(typeof(x), t)"),
    linter
  )
})


local({
  # test for lint errors appropriately raised for all is.<type> calls
  is_types <- c(
    "raw", "logical", "integer", "double", "complex", "character", "list",
    "numeric", "function", "primitive", "environment", "pairlist", "promise",
    "language", "call", "name", "symbol", "expression"
  )
  patrick::with_parameters_test_that(
    "expect_type_linter catches expect_true(is.<base type>)",
    expect_lint(
      sprintf("expect_true(is.%s(x))", is_type),
      rex::rex("expect_type(x, t) is better than expect_true(is.<t>(x))"),
      expect_type_linter()
    ),
    .test_name = is_types,
    is_type = is_types
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      expect_true(is.integer(x))
      expect_equal(typeof(x), 'double')
    }"),
    list(
      list("expect_true", line_number = 2L),
      list("expect_equal", line_number = 3L)
    ),
    expect_type_linter()
  )
})

test_that("matrix_apply_linter skips allowed usages", {
  linter <- matrix_apply_linter()

  expect_no_lint("apply(x, 1, prod)", linter)

  expect_no_lint("apply(x, 1, function(i) sum(i[i > 0]))", linter)

  # sum as FUN argument
  expect_no_lint("apply(x, 1, f, sum)", linter)

  # mean() with named arguments other than na.rm is skipped because they are not
  # implemented in colMeans() or rowMeans()
  expect_no_lint("apply(x, 1, mean, trim = 0.2)", linter)
})

test_that("matrix_apply_linter is not implemented for complex MARGIN values", {
  linter <- matrix_apply_linter()

  # Could be implemented at some point
  expect_no_lint("apply(x, seq(2, 4), sum)", linter)

  # No equivalent
  expect_no_lint("apply(x, c(2, 4), sum)", linter)

  # Beyond the scope of static analysis
  expect_no_lint("apply(x, m, sum)", linter)

  expect_no_lint("apply(x, 1 + 2:4, sum)", linter)
})


test_that("matrix_apply_linter simple disallowed usages", {
  linter <- matrix_apply_linter()

  lint_message <- rex::rex("rowSums(x)")
  expect_lint("apply(x, 1, sum)", lint_message, linter)
  expect_lint("apply(x, MARGIN = 1, FUN = sum)", lint_message, linter)
  expect_lint("apply(x, 1L, sum)", lint_message, linter)
  expect_lint("apply(x, 1:4, sum)", rex::rex("rowSums(x, dims = 4)"), linter)
  expect_lint("apply(x, 2, sum)", rex::rex("rowSums(colSums(x))"), linter)
  expect_lint("apply(x, 2:4, sum)", rex::rex("rowSums(colSums(x), dims = 3)"), linter)

  lint_message <- rex::rex("rowMeans")
  expect_lint("apply(x, 1, mean)", lint_message, linter)
  expect_lint("apply(x, MARGIN = 1, FUN = mean)", lint_message, linter)

  # Works with extra args in mean()
  expect_lint("apply(x, 1, mean, na.rm = TRUE)", lint_message, linter)

  lint_message <- rex::rex("colMeans")
  expect_lint("apply(x, 2, mean)", lint_message, linter)
  expect_lint("apply(x, 2:4, mean)", lint_message, linter)

  # adversarial comments
  expect_lint(
    trim_some("
      apply(x, 2, #comment
      mean)
    "),
    lint_message,
    linter
  )
})

test_that("matrix_apply_linter recommendation includes na.rm if present in original call", {
  linter <- matrix_apply_linter()

  lint_message <- rex::rex("na.rm = TRUE")
  expect_lint("apply(x, 1, sum, na.rm = TRUE)", lint_message, linter)
  expect_lint("apply(x, 2, sum, na.rm = TRUE)", lint_message, linter)
  expect_lint("apply(x, 1, mean, na.rm = TRUE)", lint_message, linter)
  expect_lint("apply(x, 2, mean, na.rm = TRUE)", lint_message, linter)

  lint_message <- rex::rex("rowSums(x)")
  expect_lint("apply(x, 1, sum)", lint_message, linter)

  lint_message <- rex::rex("na.rm = foo")
  expect_lint("apply(x, 1, sum, na.rm = foo)", lint_message, linter)
})

test_that("matrix_apply_linter works with multiple lints in a single expression", {
  linter <- matrix_apply_linter()

  expect_lint(
    trim_some("{
      apply(x, 1, sum)
      apply(y, 2:4, mean, na.rm = TRUE)
    }"),
    list(
      list(rex::rex("rowSums(x)"), line_number = 2L),
      list(rex::rex("rowMeans(colMeans(y, na.rm = TRUE), dims = 3)"), line_number = 3L)
    ),
    linter
  )
})

test_that("basic usage of make_linter_from_xpath works", {
  linter <- make_linter_from_xpath("//NUM_CONST", "Number")
  expect_type(linter, "closure")
  expect_lint("1", list("Number", type = "warning"), linter())

  expect_lint("'a'", "Letter", make_linter_from_xpath("//STR_CONST", "Letter")())
  expect_lint("'a'", "Letter", make_linter_from_xpath("//STR_CONST", "Letter", level = "file")())
  expect_lint("'a'", list("Letter", type = "style"), make_linter_from_xpath("//STR_CONST", "Letter", type = "style")())
})

test_that("input validation works for make_linter_from_xpath", {
  expect_error(
    make_linter_from_xpath("//NUM_CONST", "Number", type = "x"),
    'one of "warning", "style", "error"',
    fixed = TRUE
  )

  expect_error(
    make_linter_from_xpath("//NUM_CONST", "Number", level = "x"),
    'one of "expression", "file"',
    fixed = TRUE
  )

  err_msg <- "`xpath` must be a character string"
  expect_error(make_linter_from_xpath(FALSE), err_msg, fixed = TRUE)
  expect_error(make_linter_from_xpath(letters), err_msg, fixed = TRUE)
  expect_error(make_linter_from_xpath(NA_character_), err_msg, fixed = TRUE)
  expect_error(make_linter_from_xpath(character()), err_msg, fixed = TRUE)

  expect_error(
    make_linter_from_xpath(""),
    "`lint_message` is required",
    fixed = TRUE
  )
})

test_that("basic usage of make_linter_from_function_xpath works", {
  linter <- make_linter_from_function_xpath(
    "sum",
    "following-sibling::SYMBOL_SUB[text() = 'na.rm']",
    "na.rm"
  )
  expect_type(linter, "closure")
  expect_no_lint("sum()", linter())
  expect_lint("sum(na.rm=TRUE)", "na.rm", linter())
})

test_that("input validation works for make_linter_from_function_xpath", {
  expect_error(
    make_linter_from_function_xpath(1L),
    "`function_names` must be a character vector",
    fixed = TRUE
  )
  expect_error(
    make_linter_from_function_xpath(character()),
    "`function_names` must be a character vector",
    fixed = TRUE
  )

  expect_error(
    make_linter_from_function_xpath("sum", 1L),
    "`xpath` must be a character string",
    fixed = TRUE
  )
  expect_error(
    make_linter_from_function_xpath("sum", character()),
    "`xpath` must be a character string",
    fixed = TRUE
  )
  expect_error(
    make_linter_from_function_xpath("sum", letters),
    "`xpath` must be a character string",
    fixed = TRUE
  )
  expect_error(
    make_linter_from_function_xpath("sum", NA_character_),
    "`xpath` must be a character string",
    fixed = TRUE
  )

  expect_error(
    make_linter_from_function_xpath("sum", "XP"),
    "`lint_message` is required",
    fixed = TRUE
  )
})

test_that("list_comparison_linter skips allowed usages", {
  expect_no_lint("sapply(x, sum) > 10", list_comparison_linter())
})

local({
  linter <- list_comparison_linter()
  lint_msg <- rex::rex("a list(), is being coerced for comparison")

  cases <- expand.grid(
    list_mapper = c("lapply", "map", "Map", ".mapply"),
    comparator = c("==", "!=", ">=", "<=", ">", "<")
  )
  cases$.test_name <- with(cases, paste(list_mapper, comparator))
  patrick::with_parameters_test_that(
    "list_comparison_linter blocks simple disallowed usages",
    expect_lint(sprintf("%s(x, sum) %s 10", list_mapper, comparator), lint_msg, linter),
    .cases = cases
  )
})

test_that("list_comparison_linter vectorizes", {
  expect_lint(
    trim_some("{
      sapply(x, sum) > 10
      .mapply(`+`, list(1:10, 1:10), NULL) == 2
      lapply(x, sum) < 5
    }"),
    list(
      list(rex::rex(".mapply()", anything, "`==`"), line_number = 3L),
      list(rex::rex("lapply()", anything, "`<`"), line_number = 4L)
    ),
    list_comparison_linter()
  )
})

test_that("print_linter skips allowed usages", {
  linter <- print_linter()

  expect_no_lint("print(x)", linter)
  expect_no_lint("print(foo(x))", linter)
})

test_that("print_linter blocks disallowed usages", {
  linter <- print_linter()
  lint_msg <-
    rex::rex("Use cat() instead of print() logging messages. Use message() in cases calling for a signalled condition.")

  expect_lint('print("hi")', lint_msg, linter)

  # basic known-character functions
  expect_lint('print(paste(x, "b", y))', lint_msg, linter)
  expect_lint('print(paste0(x, "c", y))', lint_msg, linter)
  expect_lint('print(sprintf("a %s", x))', lint_msg, linter)

  # vectorization, metadata
  expect_lint(
    trim_some("{
      print('a')
      print(paste('x', y))
      print(z)
      print(sprintf('%s', b))
    }"),
    list(
      list(lint_msg, line_number = 2L, column_number = 3L),
      list(lint_msg, line_number = 3L, column_number = 3L),
      list(lint_msg, line_number = 5L, column_number = 3L)
    ),
    linter
  )
})

test_that("comparison_negation_linter skips allowed usages", {
  linter <- comparison_negation_linter()

  # doesn't apply to joint statements
  expect_no_lint("!(x == y | y == z)", linter)
  # don't force de Morgan's laws
  expect_no_lint("!(x & y)", linter)

  # naive xpath will include !foo(x) cases
  expect_no_lint("!any(x > y)", linter)
  # ditto for tidyeval cases
  expect_no_lint("!!target == 1 ~ 'target'", linter)
  # ditto for !x[f == g]
  expect_no_lint("!passes.test[stage == 1]", linter)
})

local({
  linter <- comparison_negation_linter()

  comparators <- c("==", "!=", ">=", ">", "<=", "<")
  inverses <- c("!=", "==", "<", "<=", ">", ">=")
  patrick::with_parameters_test_that(
    "comparison_negation_linter blocks simple disallowed usages",
    {
      expect_lint(
        sprintf("!(x %s y)", comparator),
        rex::rex(sprintf("Use x %s y, not !(x %s y)", inverse, comparator)),
        linter
      )
      expect_lint(
        sprintf("!x %s y", comparator),
        rex::rex(sprintf("Use x %s y, not !(x %s y)", inverse, comparator)),
        linter
      )
    },
    .test_name = comparators,
    comparator = comparators,
    inverse = inverses
  )
})

test_that("comparison_negation_linter catches plain ! (no parens) + call", {
  # Earlier logic would find the wrong node due to expr in length()
  expect_lint(
    "!length(x) > 0",
    rex::rex("Use x <= y, not !(x > y)"),
    comparison_negation_linter()
  )
})

test_that("Lints vectorize", {
  expect_lint(
    trim_some("{
      !(x > y)
      !x == y
    }"),
    list(
      list(rex::rex("Use x <= y, not !(x > y)."), line_number = 2L),
      list(rex::rex("Use x != y, not !(x == y)."), line_number = 3L)
    ),
    comparison_negation_linter()
  )
})

test_that("logic survives adversarial comments", {
  expect_lint(
    trim_some("
      !(x #
      > y)
    "),
    rex::rex("Use x <= y, not !(x > y)"),
    comparison_negation_linter()
  )
})

patrick::with_parameters_test_that(
  "lints correctly",
  {
    linter <- routine_registration_linter()
    expect_no_lint(sprintf("%s(ROUTINE, 1)", caller), linter)
    expect_lint(
      sprintf("%s('ROUTINE', PACKAGE = 'foo')", caller),
      "Register your native code routines with useDynLib",
      linter
    )
  },
  .test_name = c(".C", ".Call", ".External", ".Fortran"),
  caller = c(".C", ".Call", ".External", ".Fortran")
)

test_that("lints vectorize", {
  lint_msg <- "Register your native code routines with useDynLib"

  expect_lint(
    trim_some("{
      .C('ROUTINE', PACKAGE = 'foo')
      .External('POUTINE', PACKAGE = 'bar')
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    routine_registration_linter()
  )
})

test_that("nzchar_linter skips allowed usages", {
  linter <- nzchar_linter()

  expect_no_lint("if (any(nzchar(x))) TRUE", linter)

  expect_no_lint("letters == 'a'", linter)

  expect_no_lint("which(nchar(x) == 4)", linter)
  expect_no_lint("which(nchar(x) != 2)", linter)
})

test_that("nzchar_linter skips as appropriate for other nchar args", {
  linter <- nzchar_linter()

  # using type="width" can lead to 0-width strings that are counted as
  #   nzchar, c.f. nchar("\u200b", type="width"), so don't lint this.
  # type="bytes" should be >= the value for the default (type="chars")
  expect_no_lint("nchar(x, type='width') == 0L", linter)

  # nchar(x) with invalid multibyte strings -->
  #   error, while nzchar(x) returns TRUE for those entries.
  # nchar(x, allowNA=TRUE) with invalid multibyte strings -->
  #   NA in each element with an invalid entry, while nzchar returns TRUE.
  expect_no_lint("nchar(x, allowNA=TRUE) == 0L", linter)

  # nzchar also has keepNA argument so a drop-in switch is easy
  expect_lint(
    "nchar(x, keepNA=TRUE) == 0",
    rex::rex("Use !nzchar(x) instead of nchar(x) == 0"),
    linter
  )
})

test_that("nzchar_linter blocks simple disallowed usages", {
  linter <- nzchar_linter()
  lint_msg <- rex::rex("Use !nzchar(x) instead of nchar(x) == 0")

  expect_lint("which(x == '')", rex::rex('Use !nzchar(x) instead of x == ""'), linter)
  expect_lint("any(nchar(x) >= 0)", rex::rex("nchar(x) >= 0 is always true, maybe you want nzchar(x)?"), linter)
  expect_lint("all(nchar(x) == 0L)", lint_msg, linter)
  expect_lint("sum(0.0 < nchar(x))", rex::rex("Use nzchar(x) instead of nchar(x) > 0"), linter)

  # adversarial comment
  expect_lint(
    trim_some("
      all(nchar(x) #comment
      == 0L)
    "),
    lint_msg,
    linter
  )
})

test_that("nzchar_linter skips comparison to '' in if/while statements", {
  linter <- nzchar_linter()
  lint_msg_quote <- rex::rex('Use !nzchar(x) instead of x == ""')
  lint_msg_nchar <- rex::rex("Use nzchar(x) instead of nchar(x) > 0")

  # still lint nchar() comparisons
  expect_lint("if (nchar(x) > 0) TRUE", lint_msg_nchar, linter)
  expect_no_lint('if (x == "") TRUE', linter)
  expect_no_lint('while (x == "") TRUE', linter)

  # nested versions, a la nesting issues with vector_logic_linter
  expect_no_lint('if (TRUE || (x == "" && FALSE)) TRUE', linter)
  expect_no_lint('if (TRUE && x == "" && FALSE) TRUE', linter)
  expect_lint('if (any(x == "")) TRUE', lint_msg_quote, linter)
  expect_lint('if (TRUE || any(x == "" | FALSE)) TRUE', lint_msg_quote, linter)
  expect_no_lint('foo(if (x == "") y else z)', linter)
})

test_that("multiple lints are generated correctly", {
  expect_lint(
    trim_some("{
      a == ''
      '' < b
      nchar(c) != 0
      0.0 > nchar(d)
    }"),
    list(
      list(rex::rex('Use !nzchar(x) instead of x == ""'), line_number = 2L),
      list(rex::rex('Use nzchar(x) instead of x > ""'), line_number = 3L),
      list(rex::rex("Use nzchar(x) instead of nchar(x) != 0."), line_number = 4L),
      list(rex::rex("nchar(x) < 0 is always false, maybe you want !nzchar(x)?"), line_number = 5L)
    ),
    nzchar_linter()
  )
})

test_that("nested_ifelse_linter skips allowed usages", {
  linter <- nested_ifelse_linter()

  expect_no_lint("if (TRUE) 1 else 2", linter)
  expect_no_lint("if (TRUE) 1 else if (TRUE) 2 else 3", linter)

  expect_no_lint("ifelse(runif(10) > .2, 4, 6)", linter)

  # don't block suggested alternatives
  expect_no_lint("fcase(l1, v1, l2, v2)", linter)
  expect_no_lint("case_when(l1 ~ v1, l2 ~ v2)", linter)
})

test_that("nested_ifelse_linter blocks simple disallowed usages", {
  expect_lint(
    "ifelse(l1, v1, ifelse(l2, v2, v3))",
    rex::rex("Don't use nested ifelse() calls"),
    nested_ifelse_linter()
  )

  expect_lint(
    "ifelse(l1, ifelse(l2, v1, v2), v3)",
    rex::rex("Don't use nested ifelse() calls"),
    nested_ifelse_linter()
  )
})

test_that("nested_ifelse_linter also catches dplyr::if_else", {
  expect_lint(
    "if_else(l1, v1, if_else(l2, v2, v3))",
    rex::rex("Don't use nested if_else() calls"),
    nested_ifelse_linter()
  )

  expect_lint(
    "dplyr::if_else(l1, dplyr::if_else(l2, v1, v2), v3)",
    rex::rex("Don't use nested if_else() calls"),
    nested_ifelse_linter()
  )
})

test_that("nested_ifelse_linter also catches data.table::fifelse", {
  expect_lint(
    "fifelse(l1, v1, fifelse(l2, v2, v3))",
    rex::rex("Don't use nested fifelse() calls"),
    nested_ifelse_linter()
  )

  expect_lint(
    "data.table::fifelse(l1, v1, data.table::fifelse(l2, v2, v3))",
    rex::rex("Don't use nested fifelse() calls"),
    nested_ifelse_linter()
  )

  # not sure why anyone would do this, but the readability still argument holds
  expect_lint(
    "data.table::fifelse(l1, dplyr::if_else(l2, v1, v2), v3)",
    rex::rex("Don't use nested if_else() calls"),
    nested_ifelse_linter()
  )
})

test_that("lints vectorize", {
  expect_lint(
    trim_some("{
      ifelse(x < 0, ifelse(x == 0, 0, 1), -1)
      fifelse(y < 0, fifelse(y == 0, 0, 1), -1)
    }"),
    list(
      list("ifelse", line_number = 2L),
      list("fifelse", line_number = 3L)
    ),
    nested_ifelse_linter()
  )
})

#' Block usage of nested `ifelse()` calls
#'
#' Calling [ifelse()] in nested calls is problematic for two main reasons:
#'   1. It can be hard to read -- mapping the code to the expected output
#'      for such code can be a messy task/require a lot of mental bandwidth,
#'      especially for code that nests more than once
#'   2. It is inefficient -- `ifelse()` can evaluate _all_ of its arguments at
#'      both yes and no (see <https://stackoverflow.com/q/16275149>); this issue
#'      is exacerbated for nested calls
#'
#' Users can instead rely on a more readable alternative modeled after SQL
#'   CASE WHEN statements.
#'
#' Let's say this is our original code:
#'
#' ```r
#' ifelse(
#'   x == "a",
#'   2L,
#'   ifelse(x == "b", 3L, 1L)
#' )
#' ```
#'
#' Here are a few ways to avoid nesting and make the code more readable:
#'
#'   - Use `data.table::fcase()`
#'
#'     ```r
#'     data.table::fcase(
#'       x == "a", 2L,
#'       x == "b", 3L,
#'       default = 1L
#'     )
#'     ```
#'
#'   - Use `dplyr::case_match()`
#'
#'     ```r
#'     dplyr::case_match(
#'       x,
#'       "a" ~ 2L,
#'       "b" ~ 3L,
#'       .default = 1L
#'     )
#'     ```
#'
#'   - Use a look-up-and-merge approach (build a mapping table between values
#'     and outputs and merge this to the input)
#'
#'     ```r
#'     default <- 1L
#'     values <- data.frame(
#'       a = 2L,
#'       b = 3L
#'     )
#'     found_value <- values[[x]]
#'     ifelse(is.null(found_value), default, found_value)
#'     ```
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'ifelse(x == "a", 1L, ifelse(x == "b", 2L, 3L))',
#'   linters = nested_ifelse_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'dplyr::case_when(x == "a" ~ 1L, x == "b" ~ 2L, TRUE ~ 3L)',
#'   linters = nested_ifelse_linter()
#' )
#'
#' lint(
#'   text = 'data.table::fcase(x == "a", 1L, x == "b", 2L, default = 3L)',
#'   linters = nested_ifelse_linter()
#' )
#'
#' @evalRd rd_tags("nested_ifelse_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
nested_ifelse_linter <- function() {
  # NB: land on the nested (inner) call, not the outer call, and throw a lint with the inner call's name
  xpath <- glue("
  following-sibling::expr[
    expr[1]/SYMBOL_FUNCTION_CALL[ {xp_text_in_table(ifelse_funs)} ]
  ]")

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(ifelse_funs)
    bad_expr <- xml_find_all_(xml_calls, xpath)

    matched_call <- xp_call_name(bad_expr)
    lint_message <- paste(
      sprintf("Don't use nested %s() calls;", matched_call),
      "instead, try (1) data.table::fcase; (2) dplyr::case_when; or (3) using a lookup table."
    )
    xml_nodes_to_lints(bad_expr, source_expression, lint_message, type = "warning")
  })
}

# utils for working with xpaths

# like `text() %in% table`, translated to XPath 1.0
xp_text_in_table <- function(table) {
  if (length(table) == 0L) {
    return("true")
  }
  # xpath doesn't seem to have a standard way of escaping quotes, so attempt
  #   to use "" whenever the string has ' (not a perfect solution). info on
  #   escaping from https://stackoverflow.com/questions/14822153
  single_quoted <- grepl("'", table, fixed = TRUE)
  table[single_quoted] <- sQuote(table[single_quoted], '"')
  table[!single_quoted] <- sQuote(table[!single_quoted], "'")
  paste0("text() = ", table, collapse = " or ")
}

paren_wrap <- function(..., sep) {
  sep <- paste(")", sep, "(")
  dots <- list(...)
  if (length(dots) == 1L && length(dots[[1L]]) > 1L) {
    inner <- paste(dots[[1L]], collapse = sep)
  } else {
    inner <- paste(..., sep = sep)
  }
  paste0("(", inner, ")")
}

#' Safer wrapper for paste(..., sep = " and ")
#'
#' The intent is to use this for readability when combining XPath conditions so
#'   the `...` elements should be in that language, but this is not enforced.
#'
#' Inputs are also wrapped in `()` so as not to risk mixing operator precedence.
#'
#' @param ... Series of conditions
#' @noRd
xp_and <- function(...) paren_wrap(..., sep = "and")

#' Safer wrapper for paste(..., sep = " or ")
#'
#' The intent is to use this for readability when combining XPath conditions so
#'   the `...` elements should be in that language, but this is not enforced.
#'
#' Inputs are also wrapped in `()` so as not to risk mixing operator precedence.
#'
#' @param ... Series of conditions
#' @noRd
xp_or <- function(...) paren_wrap(..., sep = "or")

#' Get the name of the function matched by an XPath
#'
#' Often, it is more helpful to tailor the `message` of a lint to record
#'   which function was matched by the lint logic. This function encapsulates
#'   the logic to pull out the matched call in common situations.
#'
#' @param expr An `xml_node` or `xml_nodeset`, e.g. from [xml2::xml_find_all()].
#' @param depth Integer, default `1L`. How deep in the AST represented by `expr`
#'   should we look to find the call? By default, we assume `expr` is matched
#'   to an `<expr>` node under which the corresponding `<SYMBOL_FUNCTION_CALL>`
#'   node is found directly. `depth = 0L` means `expr` is matched directly
#'   to the `SYMBOL_FUNCTION_CALL`; `depth > 1L` means `depth` total `<expr>`
#'   nodes must be traversed before finding the call.
#' @param condition An additional (XPath condition on the `SYMBOL_FUNCTION_CALL`
#'   required for a match. The default (`NULL`) is no condition. See examples.
#'
#' @examples
#' xml_from_code <- function(str) {
#'   xml2::read_xml(xmlparsedata::xml_parse_data(parse(text = str, keep.source = TRUE)))
#' }
#' xml <- xml_from_code("sum(1:10)")
#' xp_call_name(xml, depth = 2L)
#'
#' xp_call_name(xml2::xml_find_first(xml, "expr"))
#'
#' xml <- xml_from_code(c("sum(1:10)", "sd(1:10)"))
#' xp_call_name(xml, depth = 2L, condition = "text() = 'sum'")
#'
#' @export
xp_call_name <- function(expr, depth = 1L, condition = NULL) {
  if (!is.numeric(depth) || depth < 0L) {
    cli_abort("{.arg depth} must be a non-negative number, not {.obj_type_friendly {depth}}.")
  }
  if (!is.null(condition) && !is.character(condition)) {
    cli_abort("{.arg condition} must be a character vector or NULL, not {.obj_type_friendly {condition}}.")
  }
  is_valid_expr <- is_node(expr) || is_nodeset(expr)
  if (!is_valid_expr) {
    cli_abort(c(
      i = "{.arg expr} must be an {.cls xml_nodeset} or an {.cls xml_node}.",
      x = "Instead, it is {.obj_type_friendly {expr}}."
    ))
  }

  if (is.null(condition)) {
    node <- "SYMBOL_FUNCTION_CALL"
  } else {
    node <- sprintf("SYMBOL_FUNCTION_CALL[%s]", condition)
  }

  xpath <- paste0("string(", strrep("expr/", depth), node, ")")

  xml_find_chr_(expr, xpath)
}

xp_find_location <- function(xml, xpath) {
  if (identical(xpath, "number(./@col1)")) {
    as.integer(xml_attr_(xml, "col1"))
  } else if (identical(xpath, "number(./@col2)")) {
    as.integer(xml_attr_(xml, "col2"))
  } else {
    as.integer(xml_find_num_(xml, xpath))
  }
}

#' Strip XPath 2.0-style comments from an XPath
#'
#' `{xml2}` uses XPath 1.0, which has no support for comments. But comments are
#'   useful in a codebase with as many XPaths as we maintain, so we fudge our
#'   way to XPath 2.0-ish support by writing this simple function to remove comments.
#'
#' @noRd
xpath_comment_re <- rex(
  "(:",
  zero_or_more(not(":)")),
  ":)"
)
xp_strip_comments <- function(xpath) re_substitutes(xpath, xpath_comment_re, "", global = TRUE)

#' Combine two or more nodesets to a single nodeset
#'
#' Useful for calling `{xml2}` functions on a combined set of nodes obtained using different XPath searches.
#'
#' @noRd
# TODO(r-lib/xml2#433): remove this and just use c()
combine_nodesets <- function(...) {
  res <- c(...)
  class(res) <- "xml_nodeset"
  res
}

test_that("xp_call_name works", {
  xml_from_code <- function(str) {
    xml2::read_xml(xmlparsedata::xml_parse_data(parse(text = str, keep.source = TRUE)))
  }
  xml <- xml_from_code("sum(1:10)")
  expect_identical(xp_call_name(xml, depth = 2L), "sum")

  expect_identical(xp_call_name(xml2::xml_find_first(xml, "expr")), "sum")

  xml <- xml_from_code(c("sum(1:10)", "sd(1:10)"))
  expect_identical(xp_call_name(xml, depth = 2L, condition = "text() = 'sum'"), "sum")
})

test_that("xp_call_name input validation works", {
  expect_error(xp_call_name(2L), "`expr` must be an <xml_nodeset>", fixed = TRUE)

  xml <- xml2::read_xml("<a></a>")
  expect_error(xp_call_name(xml, depth = -1L), "`depth` must be a non-negative number", fixed = TRUE)
  expect_error(xp_call_name(xml, depth = "1"), "`depth` must be a non-negative number", fixed = TRUE)
  expect_error(xp_call_name(xml, condition = 1L), "`condition` must be a character vector or NULL", fixed = TRUE)
})

#' Require usage of `expect_s3_class()`
#'
#' [testthat::expect_s3_class()] exists specifically for testing the class
#'   of S3 objects. [testthat::expect_equal()], [testthat::expect_identical()],
#'   and [testthat::expect_true()] can also be used for such tests,
#'   but it is better to use the tailored function instead.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'expect_equal(class(x), "data.frame")',
#'   linters = expect_s3_class_linter()
#' )
#'
#' lint(
#'   text = 'expect_equal(class(x), "numeric")',
#'   linters = expect_s3_class_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'expect_s3_class(x, "data.frame")',
#'   linters = expect_s3_class_linter()
#' )
#'
#' lint(
#'   text = 'expect_type(x, "double")',
#'   linters = expect_s3_class_linter()
#' )
#'
#' @evalRd rd_tags("expect_s3_class_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - [expect_s4_class_linter()]
#' @export
expect_s3_class_linter <- function() {
  # (1) expect_{equal,identical}(class(x), C)
  # (2) expect_true(is.<class>(x)) and expect_true(inherits(x, C))
  expect_equal_identical_xpath <- "
  following-sibling::expr[
    expr[1][SYMBOL_FUNCTION_CALL[text() = 'class']]
    and (position() = 1 or preceding-sibling::expr[STR_CONST])
  ]
    /parent::expr[not(SYMBOL_SUB[text() = 'info' or text() = 'label' or text() = 'expected.label'])]
  "
  # NB: there is no easy way to make an exhaustive list of places where an
  #   is.<x> call can be replaced by expect_s3_class(); this list was manually
  #   populated from the default R packages by inspection. For example,
  #   is.matrix(x) cannot be replaced by expect_s3_class(x, "matrix") because
  #   it is not actually an S3 class (is.object(x) is not TRUE).
  #   Further, there are functions named is.<x> that have nothing to do with
  #   object type, e.g. is.finite(), is.nan(), or is.R().
  is_s3_class_calls <- paste0("is.", c(
    # base
    "data.frame", "factor", "numeric_version", "ordered", "package_version", "qr", "table",
    #      utils grDevices     tcltk    tcltk    grid    grid
    "relistable", "raster", "tclObj", "tkwin", "grob", "unit",
    # stats
    "mts", "stepfun", "ts", "tskernel"
  ))
  is_class_call <- xp_text_in_table(c(is_s3_class_calls, "inherits"))
  expect_true_xpath <- glue("
  following-sibling::expr[1][expr[1][SYMBOL_FUNCTION_CALL[ {is_class_call} ]]]
    /parent::expr[not(SYMBOL_SUB[text() = 'info' or text() = 'label'])]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    expect_equal_identical_calls <- source_expression$xml_find_function_calls(c("expect_equal", "expect_identical"))
    expect_true_calls <- source_expression$xml_find_function_calls("expect_true")

    bad_expr <- combine_nodesets(
      xml_find_all_(expect_equal_identical_calls, expect_equal_identical_xpath),
      xml_find_all_(expect_true_calls, expect_true_xpath)
    )
    matched_function <- xp_call_name(bad_expr)
    msg <- ifelse(
      matched_function %in% c("expect_equal", "expect_identical"),
      sprintf("expect_s3_class(x, k) is better than %s(class(x), k).", matched_function),
      "expect_s3_class(x, k) is better than expect_true(is.<k>(x)) or expect_true(inherits(x, k))."
    )
    xml_nodes_to_lints(
      bad_expr,
      source_expression,
      lint_message = paste(msg, "Note also expect_s4_class() available for testing S4 objects."),
      type = "warning"
    )
  })
}

#' Require usage of `expect_true(x)` over `expect_equal(x, TRUE)`
#'
#' [testthat::expect_true()] and [testthat::expect_false()] exist specifically
#'   for testing the `TRUE`/`FALSE` value of an object.
#'   [testthat::expect_equal()] and [testthat::expect_identical()] can also be
#'   used for such tests, but it is better to use the tailored function instead.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "expect_equal(x, TRUE)",
#'   linters = expect_true_false_linter()
#' )
#'
#' lint(
#'   text = "expect_equal(x, FALSE)",
#'   linters = expect_true_false_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "expect_true(x)",
#'   linters = expect_true_false_linter()
#' )
#'
#' lint(
#'   text = "expect_false(x)",
#'   linters = expect_true_false_linter()
#' )
#'
#' @evalRd rd_tags("expect_true_false_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
expect_true_false_linter <- function() {
  pipe_cond <- glue("parent::expr/parent::expr[PIPE or SPECIAL[{xp_text_in_table(magrittr_pipes)}]]")
  xpath <- glue("
  following-sibling::expr[
    NUM_CONST[text() = 'TRUE' or text() = 'FALSE']
    and position() <= 2 - count({pipe_cond})
  ]
    /parent::expr
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(c("expect_equal", "expect_identical"))
    bad_expr <- xml_find_all_(xml_calls, xpath)

    # NB: use expr/$node, not expr[$node], to exclude other things (especially ns:: parts of the call)
    call_name <- xp_call_name(bad_expr, condition = "starts-with(text(), 'expect_')")
    truth_value <- xml_find_chr_(bad_expr, "string(expr/NUM_CONST[text() = 'TRUE' or text() = 'FALSE'])")
    lint_message <- sprintf(
      "expect_%s(x) is better than %s(x, %s)",
      tolower(truth_value), call_name, truth_value
    )

    xml_nodes_to_lints(bad_expr, source_expression, lint_message, type = "warning")
  })
}

#' Block obvious "yoda tests"
#'
#' Yoda tests use `(expected, actual)` instead of the more common `(actual, expected)`.
#' This is not always possible to detect statically; this linter focuses on
#'   the simple case of testing an expression against a literal value, e.g.
#'   `(1L, foo(x))` should be `(foo(x), 1L)`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "expect_equal(2, x)",
#'   linters = yoda_test_linter()
#' )
#'
#' lint(
#'   text = 'expect_identical("a", x)',
#'   linters = yoda_test_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "expect_equal(x, 2)",
#'   linters = yoda_test_linter()
#' )
#'
#' lint(
#'   text = 'expect_identical(x, "a")',
#'   linters = yoda_test_linter()
#' )
#'
#' @evalRd rd_tags("yoda_test_linter")
#' @seealso
#'   [linters] for a complete list of linters available in lintr.
#'   <https://en.wikipedia.org/wiki/Yoda_conditions>
#' @export
yoda_test_linter <- function() {
  # catch the following types of literal in the first argument:
  #   (1) numeric literal (e.g. TRUE, 1L, 1.0, NA) [NUM_CONST]
  #   (2) string literal (e.g. 'str' or "str")     [STR_CONST]
  #       (but _not_ x$"key", #1067)
  #   (3) arithmetic literal (e.g. 1+1 or 0+1i)    [OP-PLUS or OP-MINUS...]
  # TODO(#963): fully generalize this & re-use elsewhere
  const_condition <- "
    NUM_CONST
    or (STR_CONST and not(OP-DOLLAR or OP-AT))
    or ((OP-PLUS or OP-MINUS) and count(expr[NUM_CONST]) = 2)
  "
  pipes <- setdiff(magrittr_pipes, c("%$%", "%<>%"))
  xpath <- glue("
  following-sibling::expr[1][ {const_condition} ]
    /parent::expr[not(preceding-sibling::*[self::PIPE or self::SPECIAL[{ xp_text_in_table(pipes) }]])]
  ")

  second_const_xpath <- glue("expr[position() = 3 and ({const_condition})]")

  Linter(linter_level = "expression", function(source_expression) {
    bad_expr <- xml_find_all_(
      source_expression$xml_find_function_calls(c("expect_equal", "expect_identical", "expect_setequal")),
      xpath
    )

    matched_call <- xp_call_name(bad_expr)
    second_const <- xml_find_first_(bad_expr, second_const_xpath)
    lint_message <- ifelse(
      is.na(second_const),
      paste(
        "Compare objects in tests in the order 'actual', 'expected', not the reverse.",
        sprintf("For example, do %1$s(foo(x), 2L) instead of %1$s(2L, foo(x)).", matched_call)
      ),
      sprintf("Avoid storing placeholder tests like %s(1, 1)", matched_call)
    )

    xml_nodes_to_lints(bad_expr, source_expression, lint_message, type = "warning")
  })
}

#' Force `&&` conditions to be written separately where appropriate
#'
#' For readability of test outputs, testing only one thing per call to
#'   [testthat::expect_true()] is preferable, i.e.,
#'   `expect_true(A); expect_true(B)` is better than `expect_true(A && B)`, and
#'   `expect_false(A); expect_false(B)` is better than `expect_false(A || B)`.
#'
#' Similar reasoning applies to `&&` usage inside [base::stopifnot()] and `assertthat::assert_that()` calls.
#'
#' Relatedly, `dplyr::filter(DF, A & B)` is the same as `dplyr::filter(DF, A, B)`, but the latter will be more readable
#'   / easier to format for long conditions. Note that this linter assumes usages of `filter()` are `dplyr::filter()`;
#'   if you're using another function named `filter()`, e.g. [stats::filter()], please namespace-qualify it to avoid
#'   false positives. You can omit linting `filter()` expressions altogether via `allow_filter = TRUE`.
#'
#' @param allow_named_stopifnot Logical, `TRUE` by default. If `FALSE`, "named" calls to `stopifnot()` are also linted.
#' @param allow_filter Character naming the method for linting calls to `filter()`. The default, `"never"`, means
#'   `filter()` and `dplyr::filter()` calls are linted; `"not_dplyr"` means only `dplyr::filter()` calls are linted;
#'   and `"always"` means no calls to `filter()` are linted. Calls like `stats::filter()` are never linted.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "expect_true(x && y)",
#'   linters = conjunct_test_linter()
#' )
#'
#' lint(
#'   text = "expect_false(x || (y && z))",
#'   linters = conjunct_test_linter()
#' )
#'
#' lint(
#'   text = "stopifnot('x must be a logical scalar' = length(x) == 1 && is.logical(x) && !is.na(x))",
#'   linters = conjunct_test_linter(allow_named_stopifnot = FALSE)
#' )
#'
#' lint(
#'   text = "dplyr::filter(mtcars, mpg > 20 & vs == 0)",
#'   linters = conjunct_test_linter()
#' )
#'
#' lint(
#'   text = "filter(mtcars, mpg > 20 & vs == 0)",
#'   linters = conjunct_test_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "expect_true(x || (y && z))",
#'   linters = conjunct_test_linter()
#' )
#'
#' lint(
#'   text = 'stopifnot("x must be a logical scalar" = length(x) == 1 && is.logical(x) && !is.na(x))',
#'   linters = conjunct_test_linter(allow_named_stopifnot = TRUE)
#' )
#'
#' lint(
#'   text = "dplyr::filter(mtcars, mpg > 20 & vs == 0)",
#'   linters = conjunct_test_linter(allow_filter = "always")
#' )
#'
#' lint(
#'   text = "filter(mtcars, mpg > 20 & vs == 0)",
#'   linters = conjunct_test_linter(allow_filter = "not_dplyr")
#' )
#'
#' lint(
#'   text = "stats::filter(mtcars$cyl, mtcars$mpg > 20 & mtcars$vs == 0)",
#'   linters = conjunct_test_linter()
#' )
#'
#' @evalRd rd_tags("conjunct_test_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
conjunct_test_linter <- function(allow_named_stopifnot = TRUE,
                                 allow_filter = c("never", "not_dplyr", "always")) {
  allow_filter <- match.arg(allow_filter)

  expect_true_assert_that_xpath <- "
  following-sibling::expr[1][AND2]
    /parent::expr
  "
  named_stopifnot_condition <-
    if (allow_named_stopifnot) "and not(preceding-sibling::*[not(self::COMMENT)][1][self::EQ_SUB])" else ""
  stopifnot_xpath <- glue("
  following-sibling::expr[1][AND2 {named_stopifnot_condition}]
    /parent::expr
  ")
  expect_false_xpath <- "
  following-sibling::expr[1][OR2]
    /parent::expr
  "

  filter_ns_cond <- switch(allow_filter,
    never = "not(SYMBOL_PACKAGE[text() != 'dplyr'])",
    not_dplyr = "SYMBOL_PACKAGE[text() = 'dplyr']",
    always = "true"
  )
  filter_xpath <- glue("
  self::*[{ filter_ns_cond }]
    /parent::expr
    /expr[AND]
  ")

  Linter(linter_level = "file", function(source_expression) {
    # need the full file to also catch usages at the top level
    expect_true_assert_that_calls <- source_expression$xml_find_function_calls(c("expect_true", "assert_that"))
    stopifnot_calls <- source_expression$xml_find_function_calls("stopifnot")
    expect_false_calls <- source_expression$xml_find_function_calls("expect_false")
    test_expr <- combine_nodesets(
      xml_find_all_(expect_true_assert_that_calls, expect_true_assert_that_xpath),
      xml_find_all_(stopifnot_calls, stopifnot_xpath),
      xml_find_all_(expect_false_calls, expect_false_xpath)
    )

    matched_fun <- xp_call_name(test_expr)
    operator <- xml_find_chr_(test_expr, "string(expr/*[self::AND2 or self::OR2])")
    replacement_fmt <- ifelse(
      matched_fun %in% c("expect_true", "expect_false"),
      "Write multiple expectations like %1$s(A) and %1$s(B)",
      "Write multiple conditions like %s(A, B)"
    )
    lint_message <- paste(
      # as.character() needed for 0-lint case where ifelse(logical(0)) returns logical(0)
      sprintf(as.character(replacement_fmt), matched_fun),
      sprintf("instead of %s(A %s B).", matched_fun, operator),
      "The latter will produce better error messages in the case of failure."
    )
    lints <- xml_nodes_to_lints(
      test_expr,
      source_expression = source_expression,
      lint_message = lint_message,
      type = "warning"
    )

    if (allow_filter != "always") {
      xml_calls <- source_expression$xml_find_function_calls("filter")
      filter_expr <- xml_find_all_(xml_calls, filter_xpath)

      filter_lints <- xml_nodes_to_lints(
        filter_expr,
        source_expression = source_expression,
        lint_message = "Use dplyr::filter(DF, A, B) instead of dplyr::filter(DF, A & B).",
        type = "warning"
      )

      lints <- c(lints, filter_lints)
    }

    lints
  })
}

#' Require usage of `expect_null` for checking `NULL`
#'
#' Require usage of `expect_null(x)` over `expect_equal(x, NULL)` and similar
#' usages.
#'
#' [testthat::expect_null()] exists specifically for testing for `NULL` objects.
#' [testthat::expect_equal()], [testthat::expect_identical()], and
#' [testthat::expect_true()] can also be used for such tests,
#' but it is better to use the tailored function instead.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "expect_equal(x, NULL)",
#'   linters = expect_null_linter()
#' )
#'
#' lint(
#'   text = "expect_identical(x, NULL)",
#'   linters = expect_null_linter()
#' )
#'
#' lint(
#'   text = "expect_true(is.null(x))",
#'   linters = expect_null_linter()
#' )
#'
#'
#' # okay
#' lint(
#'   text = "expect_null(x)",
#'   linters = expect_null_linter()
#' )
#'
#' @evalRd rd_tags("expect_null_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
expect_null_linter <- function() {
  # two cases two match:
  #  (1) expect_{equal,identical}(x, NULL) (or NULL, x)
  #  (2) expect_true(is.null(x))
  expect_equal_identical_xpath <- "
  following-sibling::expr[position() <= 2 and NULL_CONST]
    /parent::expr
  "
  expect_true_xpath <- "
  following-sibling::expr[1][expr[1]/SYMBOL_FUNCTION_CALL[text() = 'is.null']]
    /parent::expr
  "

  Linter(linter_level = "expression", function(source_expression) {
    expect_equal_identical_calls <- source_expression$xml_find_function_calls(c("expect_equal", "expect_identical"))
    expect_true_calls <- source_expression$xml_find_function_calls("expect_true")

    bad_expr <- combine_nodesets(
      xml_find_all_(expect_equal_identical_calls, expect_equal_identical_xpath),
      xml_find_all_(expect_true_calls, expect_true_xpath)
    )

    matched_function <- xp_call_name(bad_expr)
    msg <- ifelse(
      matched_function %in% c("expect_equal", "expect_identical"),
      sprintf("expect_null(x) is better than %s(x, NULL)", matched_function),
      "expect_null(x) is better than expect_true(is.null(x))"
    )
    xml_nodes_to_lints(
      bad_expr,
      source_expression,
      lint_message = msg,
      type = "warning"
    )
  })
}

#' Require usage of `expect_type(x, type)` over `expect_equal(typeof(x), type)`
#'
#' [testthat::expect_type()] exists specifically for testing the storage type
#'   of objects. [testthat::expect_equal()], [testthat::expect_identical()], and
#'   [testthat::expect_true()] can also be used for such tests,
#'   but it is better to use the tailored function instead.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'expect_equal(typeof(x), "double")',
#'   linters = expect_type_linter()
#' )
#'
#' lint(
#'   text = 'expect_identical(typeof(x), "double")',
#'   linters = expect_type_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'expect_type(x, "double")',
#'   linters = expect_type_linter()
#' )
#'
#' @evalRd rd_tags("expect_type_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
expect_type_linter <- function() {
  # NB: the full list of values that can arise from `typeof(x)` is available
  #   in ?typeof (or, slightly more robustly, in the R source: src/main/util.c.
  #   Not all of them are available in is.<type> form, e.g. 'any' or
  #   'special'. 'builtin' and 'closure' are special cases, corresponding to
  #   is.primitive and is.function (essentially).
  base_types <- c(
    "raw", "logical", "integer", "double", "complex", "character", "list",
    "numeric", "function", "primitive", "environment", "pairlist", "promise",
    # Per ?is.language, it's the same as is.call || is.name || is.expression.
    #   so by blocking it, we're forcing more precise tests of one of
    #   those directly ("language", "symbol", and "expression", resp.)
    # NB: is.name and is.symbol are identical.
    "language", "call", "name", "symbol", "expression"
  )
  base_type_tests <- xp_text_in_table(paste0("is.", base_types))
  expect_equal_identical_xpath <- "
  following-sibling::expr[
    expr[1][SYMBOL_FUNCTION_CALL[text() = 'typeof']]
    and (position() = 1 or preceding-sibling::expr[STR_CONST])
  ]
    /parent::expr[not(SYMBOL_SUB[text() = 'info' or text() = 'label' or text() = 'expected.label'])]
  "
  expect_true_xpath <- glue("
  following-sibling::expr[1][expr[1][SYMBOL_FUNCTION_CALL[ {base_type_tests} ]]]
    /parent::expr[not(SYMBOL_SUB[text() = 'info' or text() = 'label'])]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    expect_equal_identical_calls <- source_expression$xml_find_function_calls(c("expect_equal", "expect_identical"))
    expect_true_calls <- source_expression$xml_find_function_calls("expect_true")
    bad_expr <- combine_nodesets(
      xml_find_all_(expect_equal_identical_calls, expect_equal_identical_xpath),
      xml_find_all_(expect_true_calls, expect_true_xpath)
    )
    matched_function <- xp_call_name(bad_expr)
    msg <- ifelse(
      matched_function %in% c("expect_equal", "expect_identical"),
      sprintf("expect_type(x, t) is better than %s(typeof(x), t)", matched_function),
      "expect_type(x, t) is better than expect_true(is.<t>(x))"
    )
    xml_nodes_to_lints(
      bad_expr,
      source_expression,
      lint_message = msg,
      type = "warning"
    )
  })
}

#' Require usage of grep over which(grepl(.))
#'
#' `which(grepl(pattern, x))` is the same as `grep(pattern, x)`, but harder
#'   to read and requires two passes over the vector.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "which(grepl('^a', x))",
#'   linters = which_grepl_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "which(grepl('^a', x) | grepl('^b', x))",
#'   linters = which_grepl_linter()
#' )
#'
#' @evalRd rd_tags("which_grepl_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
which_grepl_linter <- make_linter_from_function_xpath(
  function_names = "grepl",
  xpath = "
  parent::expr
    /parent::expr[expr/SYMBOL_FUNCTION_CALL[text() = 'which']]
  ",
  lint_message = "grep(pattern, x) is better than which(grepl(pattern, x))."
)

#' Function left parentheses linter
#'
#' Check that all left parentheses in a function call do not have spaces before them
#'  (e.g. `mean  (1:3)`). Although this is syntactically valid, it makes the code
#'  difficult to read.
#'
#' Exceptions are made for control flow functions (`if`, `for`, etc.).
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "mean (x)",
#'   linters = function_left_parentheses_linter()
#' )
#'
#' lint(
#'   text = "stats::sd(c (x, y, z))",
#'   linters = function_left_parentheses_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "mean(x)",
#'   linters = function_left_parentheses_linter()
#' )
#'
#' lint(
#'   text = "stats::sd(c(x, y, z))",
#'   linters = function_left_parentheses_linter()
#' )
#'
#' lint(
#'   text = "foo <- function(x) (x + 1)",
#'   linters = function_left_parentheses_linter()
#' )
#'
#' @evalRd rd_tags("function_left_parentheses_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#parentheses>
#' - [spaces_left_parentheses_linter()]
#' @export
function_left_parentheses_linter <- function() { # nolint: object_length.
  # NB: attach to SYMBOL_FUNCTION_CALL instead of SYMBOL_FUNCTION_CALL/parent::expr because
  #   the latter might be associated with a different line, e.g. in the case of a
  #   complicated call to an "extracted" function (see #1963). This mistake was made earlier
  #   because it allows the xpath to be the same for both FUNCTION and SYMBOL_FUNCTION_CALL.
  #   Further, write 4 separate XPaths because the 'range_end_xpath' differs for these two nodes.
  bad_line_fun_xpath <- "(//FUNCTION | //OP-LAMBDA)[@line1 != following-sibling::OP-LEFT-PAREN/@line1]"
  bad_line_call_xpath <-
    "(//SYMBOL_FUNCTION_CALL | //SLOT)[@line1 != parent::expr/following-sibling::OP-LEFT-PAREN/@line1]"
  bad_col_fun_xpath <- "(//FUNCTION | //OP-LAMBDA)[
    @line1 = following-sibling::OP-LEFT-PAREN/@line1
    and @col2 != following-sibling::OP-LEFT-PAREN/@col1 - 1
  ]"
  bad_col_call_xpath <- "//SYMBOL_FUNCTION_CALL[
    @line1 = parent::expr/following-sibling::OP-LEFT-PAREN/@line1
    and @col2 != parent::expr/following-sibling::OP-LEFT-PAREN/@col1 - 1
  ]"

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_line_fun_exprs <- xml_find_all_(xml, bad_line_fun_xpath)
    bad_line_fun_lints <- xml_nodes_to_lints(
      bad_line_fun_exprs,
      source_expression = source_expression,
      lint_message = "Left parenthesis should be on the same line as the 'function' symbol."
    )

    bad_line_call_exprs <- xml_find_all_(xml, bad_line_call_xpath)
    bad_line_call_lints <- xml_nodes_to_lints(
      bad_line_call_exprs,
      source_expression = source_expression,
      lint_message = "Left parenthesis should be on the same line as the function's symbol."
    )

    bad_col_fun_exprs <- xml_find_all_(xml, bad_col_fun_xpath)
    bad_col_fun_lints <- xml_nodes_to_lints(
      bad_col_fun_exprs,
      source_expression = source_expression,
      lint_message = "Remove spaces before the left parenthesis in a function definition.",
      range_start_xpath = "number(./@col2 + 1)", # start after function
      range_end_xpath = "number(./following-sibling::OP-LEFT-PAREN/@col1 - 1)" # end before (
    )

    bad_col_call_exprs <- xml_find_all_(xml, bad_col_call_xpath)
    bad_col_call_lints <- xml_nodes_to_lints(
      bad_col_call_exprs,
      source_expression = source_expression,
      lint_message = "Remove spaces before the left parenthesis in a function call.",
      range_start_xpath = "number(./@col2 + 1)", # start after call name
      range_end_xpath = "number(./parent::expr/following-sibling::OP-LEFT-PAREN/@col1 - 1)" # end before (
    )

    c(bad_line_fun_lints, bad_line_call_lints, bad_col_fun_lints, bad_col_call_lints)
  })
}

#' Read lintr settings
#'
#' Lintr searches for settings for a given source file in the following order:
#'  1. options defined as `linter.setting`.
#'  2. `linter_file` in the same directory
#'  3. `linter_file` in the project directory
#'  4. `linter_file` in the user home directory
#'  5. [default_settings()]
#'
#' The default linter_file name is `.lintr` but it can be changed with option `lintr.linter_file`
#'   or the environment variable `R_LINTR_LINTER_FILE`
#' This file is a DCF file, see [base::read.dcf()] for details.
#' Here is an example of a `.lintr` file:
#'
#'  ```
#'  linters: linters_with_defaults(
#'      any_duplicated_linter(),
#'      any_is_na_linter(),
#'      backport_linter("oldrel-4", except = c("R_user_dir", "str2lang")),
#'      line_length_linter(120L),
#'      missing_argument_linter(),
#'      unnecessary_concatenation_linter(allow_single_expression = FALSE),
#'      yoda_test_linter()
#'    )
#'  exclusions: list(
#'      "inst/doc/creating_linters.R" = 1,
#'      "inst/example/bad.R",
#'      "tests/testthat/default_linter_testcode.R",
#'      "tests/testthat/dummy_packages"
#'    )
#'  ```
#'
#' Experimentally, we also support keeping the config in a plain R file. By default we look for
#'   a file named `.lintr.R` (in the same directories where we search for `.lintr`).
#' We are still deciding the future of config support in lintr, so user feedback is welcome.
#'   The advantage of R is that it maps more closely to how the configs are actually stored,
#'   whereas the DCF approach requires somewhat awkward formatting of parseable R code within
#'   valid DCF key-value pairs. The main disadvantage of the R file is it might be _too_ flexible,
#'   with users tempted to write configs with side effects causing hard-to-detect bugs or
# "   otherwise "abusing" the ability to evaluate generic R code. Other recursive key-value stores
#'   like YAML could work, but require new dependencies and are harder to parse
#'   both programmatically and visually.
#' Here is an example of a `.lintr.R` file:
#'
#'  ```r
#'  linters <- linters_with_defaults(
#'      any_duplicated_linter(),
#'      any_is_na_linter(),
#'      backport_linter("oldrel-4", except = c("R_user_dir", "str2lang")),
#'      line_length_linter(120L),
#'      missing_argument_linter(),
#'      unnecessary_concatenation_linter(allow_single_expression = FALSE),
#'      yoda_test_linter()
#'    )
#'  exclusions <- list(
#'      "inst/doc/creating_linters.R" = 1,
#'      "inst/example/bad.R",
#'      "tests/testthat/default_linter_testcode.R",
#'      "tests/testthat/dummy_packages"
#'    )
#'  ```
#'
#' @param filename Source file to be linted.
#' @param call Passed to malformed to ensure linear trace.
read_settings <- function(filename, call = parent.frame()) {
  reset_settings()

  # doing lint(text=) should read settings from the current directory, #2847
  location <- if (missing(filename)) "." else filename
  config_file <- find_config(location)
  default_encoding <- find_default_encoding(location)
  if (!is.null(default_encoding)) {
    # Locally override the default for encoding if we found a smart default
    default_settings[["encoding"]] <- default_encoding
  }

  config <- read_config_file(config_file, call = call)
  validate_config_file(config, config_file, default_settings)

  for (setting in names(default_settings)) {
    value <- get_setting(setting, config, default_settings)
    if (setting == "exclusions") {
      if (!is.null(config_file)) {
        root <- dirname(config_file)
      } else {
        root <- getwd()
      }
      value <- normalize_exclusions(value, root = root)
    }

    settings[[setting]] <- value
  }
}

#' @param call Passed to malformed to ensure linear trace.
#' @noRd
read_config_file <- function(config_file, call = parent.frame()) {
  if (is.null(config_file)) {
    return(NULL)
  }

  # clickable link for eventual error messages.
  malformed_file <- link_config_file(config_file) # nolint: object_usage_linter. TODO(#2252).
  config <- new.env()
  if (endsWith(config_file, ".R")) {
    load_config <- function(file) sys.source(file, config, keep.source = FALSE, keep.parse.data = FALSE)
    malformed <- function(e) {
      cli_abort(
        "Malformed config file ({malformed_file}), ensure it is valid R syntax.",
        parent = e,
        call = call
      )
    }
  } else {
    load_config <- function(file) {
      dcf_values <- read.dcf(file, all = TRUE)
      for (setting in names(dcf_values)) {
        parsed_setting <- withCallingHandlers(
          str2lang(dcf_values[[setting]]),
          error = function(e) {
            cli_abort(
              "Malformed config setting {.field {setting}}:",
              parent = e
            )
          }
        )
        # https://adv-r.hadley.nz/conditions.html
        setting_value <- withCallingHandlers(
          tryCatch(
            eval(parsed_setting),
            error = function(e) {
              cli_abort(
                "Error from config setting {.code {setting}}.",
                parent = e
              )
            }
          ),
          warning = function(w) {
            cli_warn(
              "Warning from config setting {.code {setting}}.",
              parent = w
            )
            invokeRestart("muffleWarning")
          }
        )
        assign(setting, setting_value, envir = config)
      }
    }
    malformed <- function(e) {
      cli_abort(
        "Malformed config file ({malformed_file}):",
        parent = e,
        call = call
      )
    }
  }
  # https://adv-r.hadley.nz/conditions.html
  withCallingHandlers(
    tryCatch(
      load_config(config_file),
      error = malformed
    ),
    warning = function(w) {
      cli_warn(
        "Warning encountered while loading config:",
        parent = w
      )
      invokeRestart("muffleWarning")
    }
  )
  config
}

validate_config_file <- function(config, config_file, defaults) {
  matched <- names(config) %in% names(defaults)
  if (!all(matched)) {
    unused_settings <- names(config)[!matched] # nolint: object_usage_linter. TODO(#2252).
    config_link <- link_config_file(config_file) # nolint: object_usage_linter. TODO(#2252).
    cli_warn("Found unused settings in config file ({config_link}): {.field unused_settings}")
  }

  validate_regex(config,
    c("exclude", "exclude_next", "exclude_start", "exclude_end", "exclude_linter", "exclude_linter_sep")
  )
  validate_character_string(config, c("encoding", "cache_directory"))
  validate_true_false(config, "error_on_lint")
  validate_linters(config$linters)
  validate_exclusions(config$exclusions)
}

is_character_string <- function(x) is.character(x) && length(x) == 1L && !is.na(x)
# perl=TRUE matches rex::re_matches()
is_valid_regex <- function(str) !inherits(tryCatch(grepl(str, "", perl = TRUE), condition = identity), "condition")
is_single_regex <- function(x) is_character_string(x) && is_valid_regex(x)
is_true_false <- function(x) is.logical(x) && length(x) == 1L && !is.na(x)

validate_keys <- function(config, keys, test, what) {
  for (key in keys) {
    val <- config[[key]]
    if (is.null(val)) {
      next
    }
    if (!test(val)) {
      cli_abort(c(
        i = "Setting {.code {key}} should be {.strong {what}}.",
        x = "Instead, it is {.field {val}}."
      ))
    }
  }
}

validate_regex <- function(config, keys) {
  validate_keys(config, keys, is_single_regex, "a single regular expression")
}

validate_character_string <- function(config, keys) {
  validate_keys(config, keys, is_character_string, "a character string")
}

validate_true_false <- function(config, keys) {
  validate_keys(config, keys, is_true_false, "TRUE or FALSE")
}

validate_linters <- function(linters) {
  if (is.null(linters)) {
    return(invisible())
  }

  is_linters <- vapply(linters, is_linter, logical(1L))
  if (!all(is_linters)) {
    non_linters <- which(!is_linters) # nolint: object_usage_linter. TODO(#2252).
    cli_abort(c(
      i = "Setting {.arg linters} should be a list of linters.",
      x = "Found non-linters at elements: {.str {non_linters}}."
    ))
  }
}

validate_exclusions <- function(exclusions) {
  if (is.null(exclusions)) {
    return(invisible())
  }

  exclusion_names <- names2(exclusions)
  has_names <- nzchar(exclusion_names)
  unnamed_is_string <-
    vapply(exclusions[!has_names], \(x) is.character(x) && length(x) == 1L && !is.na(x), logical(1L))
  if (!all(unnamed_is_string)) {
    problematic_entries <- which(!has_names)[!unnamed_is_string] # nolint: object_usage_linter. TODO(#2252).
    cli_abort(c(
      i = "Unnamed entries of setting {.arg exclusions} should be strings naming files or directories.",
      x = "Check exclusions: {.str {problematic_entries}}."
    ))
  }
  for (ii in which(has_names)) validate_named_exclusion(exclusions, ii)
}

validate_named_exclusion <- function(exclusions, idx) {
  entry <- exclusions[[idx]]
  if (is.list(entry)) {
    valid_entry <- vapply(entry, \(x) is.numeric(x) && !anyNA(x), logical(1L))
  } else {
    valid_entry <- is.numeric(entry) && !anyNA(entry)
  }
  if (!all(valid_entry)) {
    cli_abort(c(
      i = "Named entries of setting {.arg exclusions} should designate line numbers for exclusion.",
      x = "Check exclusions: {idx}."
    ))
  }
}

lintr_option <- function(setting, default = NULL) getOption(paste0("lintr.", setting), default)

get_setting <- function(setting, config, defaults) {
  lintr_option(setting) %||% config[[setting]] %||% defaults[[setting]]
}

reset_settings <- function() list2env(default_settings, envir = settings)

find_default_encoding <- function(filename) {
  if (is.null(filename)) {
    return(NULL)
  }

  filename |>
    find_package(allow_rproj = TRUE) |>
    find_rproj_at() |>
    get_encoding_from_dcf()
}

get_encoding_from_dcf <- function(file) {
  if (is.null(file)) {
    return(NULL)
  }

  encodings <- tryCatch(
    unname(drop(read.dcf(file, "Encoding"))),
    error = \(e) NULL,
    warning = \(e) NULL
  )

  encodings <- encodings[!is.na(encodings)]
  if (length(encodings) > 0L) {
    return(encodings[1L])
  }

  NULL
}

link_config_file <- function(path) {
  cli::style_hyperlink(
    cli::col_blue(basename(path)),
    paste0("file://", path)
  )
}

#' Sequence linter
#'
#' This linter checks for `1:length(...)`, `1:nrow(...)`, `1:ncol(...)`,
#' `1:NROW(...)` and `1:NCOL(...)` expressions in base-R, or their usage in
#' conjunction with `seq()` (e.g., `seq(length(...))`, `seq(nrow(...))`, etc.).
#'
#' Additionally, it checks for `1:n()` (from `{dplyr}`) and `1:.N` (from `{data.table}`).
#'
#' These often cause bugs when the right-hand side is zero.
#' Instead, it is safer to use [base::seq_len()] (to create a sequence of a specified *length*) or
#'   [base::seq_along()] (to create a sequence *along* an object).
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "seq(length(x))",
#'   linters = seq_linter()
#' )
#'
#' lint(
#'   text = "1:nrow(x)",
#'   linters = seq_linter()
#' )
#'
#' lint(
#'   text = "dplyr::mutate(x, .id = 1:n())",
#'   linters = seq_linter()
#' )
#'
#' lint(
#'   text = "seq_len(length(x))",
#'   linters = seq_linter()
#' )
#'
#' lint(
#'   text = "unlist(lapply(x, seq_len))",
#'   linters = seq_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "seq_along(x)",
#'   linters = seq_linter()
#' )
#'
#' lint(
#'   text = "seq_len(nrow(x))",
#'   linters = seq_linter()
#' )
#'
#' lint(
#'   text = "dplyr::mutate(x, .id = seq_len(n()))",
#'   linters = seq_linter()
#' )
#'
#' lint(
#'   text = "seq_along(x)",
#'   linters = seq_linter()
#' )
#'
#' lint(
#'   text = "sequence(x)",
#'   linters = seq_linter()
#' )
#'
#' @evalRd rd_tags("seq_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
seq_linter <- function() {
  bad_funcs <- xp_text_in_table(c("length", "n", "nrow", "ncol", "NROW", "NCOL", "dim"))

  # Exact `xpath` depends on whether bad function was used in conjunction with `seq()`
  seq_xpath <- glue("
  following-sibling::expr[1][expr/SYMBOL_FUNCTION_CALL[ {bad_funcs} ]]
    /parent::expr[count(expr) = 2]
  ")
  # `.N` from {data.table} is special since it's not a function but a symbol
  colon_xpath <- glue("
  //OP-COLON
    /parent::expr[
      expr[NUM_CONST[text() = '1' or text() = '1L']]
      and (
        expr[expr[(expr|self::*)[SYMBOL_FUNCTION_CALL[ {bad_funcs} ]]]]
        or expr[SYMBOL = '.N']
      )
    ]
  ")

  seq_len_xpath <- "
    parent::expr[expr/expr/SYMBOL_FUNCTION_CALL[text() = 'length']]
  "

  map_funcs <- c("sapply", "lapply", "map")
  seq_funcs <- xp_text_in_table(c("seq_len", "seq"))
  # count(expr) = 3 because we only want seq() calls without extra arguments
  sequence_xpath <- glue("
    parent::expr[
      count(expr) = 3
      and expr/SYMBOL[ {seq_funcs} ]
      and preceding-sibling::expr/SYMBOL_FUNCTION_CALL[text() = 'unlist']
    ]
  ")

  ## The actual order of the nodes is document order
  ## In practice we need to handle length(x):1
  get_fun <- function(expr, n) {
    funcall <- xml_find_chr_(expr, sprintf("string(./expr[%d])", n))

    # `dplyr::n()` is special because it has no arguments, so the lint message
    # should mention `n()`, and not `n(...)`
    if (identical(funcall, "n()")) {
      return(funcall)
    }

    fun <- gsub("\\(.*\\)", "(...)", trimws(funcall))
    bad_fun <- fun %in% bad_funcs
    fun[bad_fun] <- paste0(fun[bad_fun], "(...)")
    fun
  }

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content
    seq_calls <- source_expression$xml_find_function_calls("seq")

    seq_expr <- combine_nodesets(
      xml_find_all_(seq_calls, seq_xpath),
      xml_find_all_(xml, colon_xpath)
    )
    seq_expr <- strip_comments_from_subtree(seq_expr)

    dot_expr1 <- get_fun(seq_expr, 1L)
    dot_expr2 <- get_fun(seq_expr, 2L)
    seq_along_idx <- grepl("length(", dot_expr1, fixed = TRUE) | grepl("length(", dot_expr2, fixed = TRUE)
    rev_idx <- startsWith(dot_expr2, "1")

    replacement <- rep("seq_along(...)", length(seq_expr))
    replacement[!seq_along_idx] <- paste0("seq_len(", ifelse(rev_idx, dot_expr1, dot_expr2)[!seq_along_idx], ")")
    replacement[rev_idx] <- paste0("rev(", replacement[rev_idx], ")")

    lint_message <- ifelse(
      grepl("seq", dot_expr1, fixed = TRUE),
      sprintf(
        "Use %s instead of %s(%s), which is likely to be wrong in the empty edge case.",
        replacement, dot_expr1, dot_expr2
      ),
      sprintf(
        "Use %s instead of %s:%s, which is likely to be wrong in the empty edge case.",
        replacement, dot_expr1, dot_expr2
      )
    )

    seq_lints <- xml_nodes_to_lints(seq_expr, source_expression, lint_message, type = "warning")

    seq_len_calls <- source_expression$xml_find_function_calls("seq_len")
    seq_len_expr <- xml_find_all_(seq_len_calls, seq_len_xpath)
    seq_len_lints <- xml_nodes_to_lints(
      seq_len_expr,
      source_expression,
      "Use seq_along(x) instead of seq_len(length(x)).",
      type = "warning"
    )

    xml_map_calls <- source_expression$xml_find_function_calls(map_funcs)
    potential_sequence_calls <- xml_find_all_(xml_map_calls, sequence_xpath)
    sequence_lints <- xml_nodes_to_lints(
      potential_sequence_calls,
      source_expression,
      "Use sequence() to generate a concatenated sequence of seq_len().",
      type = "warning"
    )

    c(seq_lints, seq_len_lints, sequence_lints)
  })
}

#' Require usage of `anyNA(x)` over `any(is.na(x))`
#'
#' [base::anyNA()] exists as a replacement for `any(is.na(x))` which is more efficient
#'   for simple objects, and is at worst equally efficient.
#'   Therefore, it should be used in all situations instead of the latter.
#'
#' This also lints `NA %in% x`, which is also more readably expressed as `anyNA(x)`. In
#'   this case, note a subtle difference of behavior, namely that `is.na(NaN)` is `TRUE`
#'   while `NA %in% NaN` is `FALSE`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "any(is.na(x), na.rm = TRUE)",
#'   linters = any_is_na_linter()
#' )
#'
#' lint(
#'   text = "any(is.na(foo(x)))",
#'   linters = any_is_na_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "anyNA(x)",
#'   linters = any_is_na_linter()
#' )
#'
#' lint(
#'   text = "anyNA(foo(x))",
#'   linters = any_is_na_linter()
#' )
#'
#' lint(
#'   text = "any(!is.na(x), na.rm = TRUE)",
#'   linters = any_is_na_linter()
#' )
#'
#' @evalRd rd_tags("any_is_na_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
any_is_na_linter <- function() {
  any_xpath <- "
  following-sibling::expr[1][expr[1][SYMBOL_FUNCTION_CALL[text() = 'is.na']]]
    /parent::expr[
      count(expr) = 2
      or (count(expr) = 3 and SYMBOL_SUB[text() = 'na.rm'])
    ]
  "

  in_xpath <- "//SPECIAL[text() = '%in%']/preceding-sibling::expr[NUM_CONST[starts-with(text(), 'NA')]]"

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content
    xml_calls <- source_expression$xml_find_function_calls("any")

    any_expr <- xml_find_all_(xml_calls, any_xpath)
    any_lints <- xml_nodes_to_lints(
      any_expr,
      source_expression = source_expression,
      lint_message = "anyNA(x) is better than any(is.na(x)).",
      type = "warning"
    )

    in_expr <- xml_find_all_(xml, in_xpath)
    in_lints <- xml_nodes_to_lints(
      in_expr,
      source_expression = source_expression,
      lint_message = "anyNA(x) is better than NA %in% x.",
      type = "warning"
    )

    c(any_lints, in_lints)
  })
}

#' Package hooks linter
#'
#' Check various common "gotchas" in [.onLoad()], [.onAttach()], [.Last.lib()], [.onDetach()], and [.onUnload()]
#'   namespace hooks that will cause `R CMD check` issues. See Writing R Extensions for details.
#'
#' 1. `.onLoad()` shouldn't call [cat()], [message()], [print()], [writeLines()], [packageStartupMessage()],
#'    [require()], [library()], or [installed.packages()].
#' 2. `.onAttach()` shouldn't call `cat()`, `message()`, `print()`, `writeLines()`, [library.dynam()],
#'    `require()`, `library()`, or `installed.packages()`.
#' 3. `.Last.lib()` and `.onDetach()` shouldn't call [library.dynam.unload()].
#' 4. `.onLoad()` and `.onAttach()` should take two arguments, with names matching `^lib` and `^pkg`;
#'    `.Last.lib()`, `.onDetach()`, and `.onUnload()` should take one argument with name matching `^lib`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = ".onLoad <- function(lib, ...) { }",
#'   linters = package_hooks_linter()
#' )
#'
#' lint(
#'   text = ".onAttach <- function(lib, pkg) { require(foo) }",
#'   linters = package_hooks_linter()
#' )
#'
#' lint(
#'   text = ".onDetach <- function(pkg) { }",
#'   linters = package_hooks_linter()
#' )
#'
#' lint(
#'   text = ".onUnload <- function() { }",
#'   linters = package_hooks_linter()
#' )
#'
#' # okay
#' lint(
#'   text = ".onLoad <- function(lib, pkg) { }",
#'   linters = package_hooks_linter()
#' )
#'
#' lint(
#'   text = '.onAttach <- function(lib, pkg) { loadNamespace("foo") }',
#'   linters = package_hooks_linter()
#' )
#'
#' lint(
#'   text = ".onDetach <- function(lib) { }",
#'   linters = package_hooks_linter()
#' )
#'
#' lint(
#'   text = ".onUnload <- function(libpath) { }",
#'   linters = package_hooks_linter()
#' )
#'
#' @evalRd rd_tags("package_hooks_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
package_hooks_linter <- function() {
  bad_msg_calls <- c("cat", "message", "print", "writeLines")
  bad_calls <- list(
    .onLoad = c(bad_msg_calls, "packageStartupMessage"),
    .onAttach = c(bad_msg_calls, "library.dynam")
  )
  bad_msg_call_xpath_fmt <- "
  (//FUNCTION | //OP-LAMBDA)
    /parent::expr[preceding-sibling::expr/SYMBOL[text() = '%s']]
    //SYMBOL_FUNCTION_CALL[%s]
  "
  bad_call_xpaths <- vapply(
    seq_along(bad_calls),
    \(ii) sprintf(bad_msg_call_xpath_fmt, names(bad_calls)[ii], xp_text_in_table(bad_calls[[ii]])),
    character(1L)
  )
  names(bad_call_xpaths) <- names(bad_calls)

  make_bad_call_lint_message <- function(expr, hook) {
    call_name <- xml_text(expr)
    lint_message <- sprintf("Don't use %s() in %s().", call_name, hook)
    lint_message[call_name == "packageStartupMessage"] <-
      "Put packageStartupMessage() calls in .onAttach(), not .onLoad()."
    lint_message[call_name == "library.dynam"] <-
      "Put library.dynam() calls in .onLoad, not .onAttach()."
    lint_message
  }

  # lints here will hit the function <expr>,
  #   this path returns to the corresponding namespace hook's name
  ns_calls <- xp_text_in_table(c(".onLoad", ".onAttach", ".onDetach", ".onUnload", ".Last.lib"))

  # usually any given package will have one or maybe two files defining namespace hooks.
  #   given the number of checks, then, it's prudent to check first if any such hook is defined,
  #   exiting early if not.
  any_hook_xpath <- glue("(//FUNCTION | //OP-LAMBDA)/parent::expr/preceding-sibling::expr/SYMBOL[{ns_calls}]")

  # * for '=' assignment
  hook_xpath <- sprintf("string(./ancestor::*/expr/SYMBOL[%s])", ns_calls)

  load_arg_name_xpath <- "
  (//FUNCTION | //OP-LAMBDA)
    /parent::expr[
      preceding-sibling::expr/SYMBOL[text() = '.onAttach' or text() = '.onLoad']
      and (
        count(SYMBOL_FORMALS) != 2
        or SYMBOL_FORMALS[
          (position() = 1 and not(starts-with(text(), 'lib')))
          or (position() = 2 and not(starts-with(text(), 'pkg')))
        ]
      )
    ]
  "

  library_require_xpath <- "
  (//FUNCTION | //OP-LAMBDA)
    /parent::expr[preceding-sibling::expr/SYMBOL[text() = '.onAttach' or text() = '.onLoad']]
    //*[1][
      (self::SYMBOL or self::SYMBOL_FUNCTION_CALL)
      and (text() = 'require' or text() = 'library' or text() = 'installed.packages')
    ]
  "

  bad_unload_call_xpath <- "
  (//FUNCTION | //OP-LAMBDA)
    /parent::expr[preceding-sibling::expr/SYMBOL[text() = '.Last.lib' or text() = '.onDetach']]
    //SYMBOL_FUNCTION_CALL[text() = 'library.dynam.unload']
  "

  unload_arg_name_xpath <- "
  (//FUNCTION | //OP-LAMBDA)
    /parent::expr[
      preceding-sibling::expr/SYMBOL[text() = '.onDetach' or text() = '.onUnload' or text() = '.Last.lib']
      and (
        count(SYMBOL_FORMALS) != 1
        or SYMBOL_FORMALS[not(starts-with(text(), 'lib'))]
      )
    ]
  "

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    any_hook <- xml_find_first_(xml, any_hook_xpath)
    if (is.na(any_hook)) {
      return(list())
    }

    # inherits: source_expression, bad_call_xpaths
    bad_msg_call_lints <- function(xml, hook) {
      bad_expr <- xml_find_all_(xml, bad_call_xpaths[[hook]])
      lint_message <- make_bad_call_lint_message(bad_expr, hook)
      xml_nodes_to_lints(bad_expr, source_expression, lint_message, type = "warning")
    }

    # (1) improper messaging calls shouldn't be used inside .onLoad()/.onAttach()
    onload_bad_msg_call_lints <- bad_msg_call_lints(xml, ".onLoad")
    onattach_bad_msg_call_lints <- bad_msg_call_lints(xml, ".onAttach")

    # (2) .onLoad() and .onAttach() should take two arguments, with names matching ^lib and ^pkg
    load_arg_name_expr <- xml_find_all_(xml, load_arg_name_xpath)

    load_arg_name_message <- sprintf(
      "%s() should take two arguments, with the first starting with 'lib' and the second starting with 'pkg'.",
      xml_find_chr_(load_arg_name_expr, hook_xpath)
    )
    load_arg_name_lints <-
      xml_nodes_to_lints(load_arg_name_expr, source_expression, load_arg_name_message, type = "warning")

    # (3) .onLoad() and .onAttach() shouldn't call require(), library(), or installed.packages()
    # NB: base only checks the SYMBOL_FUNCTION_CALL version, not SYMBOL.
    library_require_expr <- xml_find_all_(xml, library_require_xpath)

    library_require_bad_call <- xml_text(library_require_expr)
    library_require_hook <- xml_find_chr_(library_require_expr, hook_xpath)
    library_require_message <- character(length(library_require_bad_call))
    is_installed_packages <- library_require_bad_call == "installed.packages"
    library_require_message[is_installed_packages] <-
      sprintf("Don't slow down package load by running installed.packages() in %s().", library_require_hook)
    library_require_message[!is_installed_packages] <-
      sprintf("Don't alter the search() path in %s() by calling %s().", library_require_hook, library_require_bad_call)
    library_require_lints <-
      xml_nodes_to_lints(library_require_expr, source_expression, library_require_message, type = "warning")

    # (4) .Last.lib() and .onDetach() shouldn't call library.dynam.unload()
    bad_unload_call_expr <- xml_find_all_(xml, bad_unload_call_xpath)

    bad_unload_call_message <- sprintf(
      "Use library.dynam.unload() calls in .onUnload(), not %s().",
      xml_find_chr_(bad_unload_call_expr, hook_xpath)
    )
    bad_unload_call_lints <-
      xml_nodes_to_lints(bad_unload_call_expr, source_expression, bad_unload_call_message, type = "warning")

    # (5) .Last.lib(), .onDetach(), and .onUnload() should take one argument with name matching ^lib
    unload_arg_name_expr <- xml_find_all_(xml, unload_arg_name_xpath)

    unload_arg_name_message <- sprintf(
      "%s() should take one argument starting with 'lib'.",
      xml_find_chr_(unload_arg_name_expr, hook_xpath)
    )
    unload_arg_name_lints <-
      xml_nodes_to_lints(unload_arg_name_expr, source_expression, unload_arg_name_message, type = "warning")

    c(
      onload_bad_msg_call_lints,
      onattach_bad_msg_call_lints,
      load_arg_name_lints,
      library_require_lints,
      bad_unload_call_lints,
      unload_arg_name_lints
    )
  })
}

# utils for working with XML

#' str2lang, but for xml children.
#'
#' [xml2::xml_text()] is deceptively close to obviating this helper, but it collapses
#'   text across lines. R is _mostly_ whitespace-agnostic, so this only matters in some edge cases,
#'   in particular when there are comments within an expression (`<expr>` node). See #1919.
#'
#' @noRd
xml2lang <- function(x) {
  x_strip_comments <- xml_find_all_(x, ".//*[not(self::COMMENT or self::expr)]")
  str2lang(paste(xml_text(x_strip_comments), collapse = " "))
}

# TODO(r-lib/xml2#341): Use xml_clone() instead.
clone_xml_ <- function(x) {
  tmp_doc <- tempfile()
  on.exit(unlink(tmp_doc))

  doc <- xml2::xml_new_root("root")
  for (ii in seq_along(x)) {
    xml2::write_xml(x[[ii]], tmp_doc)
    xml2::xml_add_child(doc, xml2::read_xml(tmp_doc))
  }
  xml_find_all_(doc, "*")
}

# caveat: whether this is a copy or not is inconsistent. assume the output is read-only!
strip_comments_from_subtree <- function(expr) {
  if (all(is.na(xml_find_first_(expr, ".//COMMENT")))) {
    return(expr)
  }
  expr <- clone_xml_(expr)
  for (comment in xml_find_all_(expr, ".//COMMENT")) xml2::xml_remove(comment)
  expr
}

safe_parse_to_xml <- function(parsed_content) {
  if (is.null(parsed_content)) {
    return(xml_missing())
  }
  tryCatch(
    xml2::read_xml(xmlparsedata::xml_parse_data(parsed_content)),
    # use xml_missing so that code doesn't always need to condition on XML existing
    error = \(e) xml_missing()
  )
}

is_node <- function(xml) inherits(xml, "xml_node")
is_nodeset <- function(xml) inherits(xml, "xml_nodeset")
is_nodeset_like <- function(xml) {
  is_nodeset(xml) ||
    (is.list(xml) && all(vapply(xml, is_node, logical(1L))))
}

# TODO(r-lib/xml2#327): Remove this workaround if upstream bottleneck is resolved.
# nolint start: undesirable_function_name_linter.
# placeholder xml_ns() object to skip this call on xml2 invocations
empty_ns <- character()
names(empty_ns) <- character()
class(empty_ns) <- "xml_namespace"

xml_find_all_ <- function(x, xpath, ns = empty_ns, ...) {
  xml_find_all(x, xpath, ns = ns, ...)
}

xml_find_first_ <- function(x, xpath, ns = empty_ns) {
  xml_find_first(x, xpath, ns = ns)
}

xml_find_chr_ <- function(x, xpath, ns = empty_ns) {
  xml_find_chr(x, xpath, ns = ns)
}

xml_find_num_ <- function(x, xpath, ns = empty_ns) {
  xml_find_num(x, xpath, ns = ns)
}

xml_find_lgl_ <- function(x, xpath, ns = empty_ns) {
  xml_find_lgl(x, xpath, ns = ns)
}

xml_attr_ <- function(x, attr, ns = empty_ns, default = NA_character_) {
  xml_attr(x, attr, ns = ns, default = default)
}

xml_attrs_ <- function(x, ns = empty_ns) {
  xml_attrs(x, ns = ns)
}

xml_child_ <- function(x, search = 1L, ns = empty_ns) {
  xml_child(x, search = search, ns = ns)
}

xml_name_ <- function(x, ns = empty_ns) {
  xml_name(x, ns = ns)
}
# nolint end: undesirable_function_name_linter.

#' Require usage of `expect_gt(x, y)` over `expect_true(x > y)` (and similar)
#'
#' [testthat::expect_gt()], [testthat::expect_gte()], [testthat::expect_lt()],
#'   [testthat::expect_lte()], and [testthat::expect_equal()] exist specifically
#'   for testing comparisons between two objects. [testthat::expect_true()] can
#'   also be used for such tests, but it is better to use the tailored function
#'   instead.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "expect_true(x > y)",
#'   linters = expect_comparison_linter()
#' )
#'
#' lint(
#'   text = "expect_true(x <= y)",
#'   linters = expect_comparison_linter()
#' )
#'
#' lint(
#'   text = "expect_true(x == (y == 2))",
#'   linters = expect_comparison_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "expect_gt(x, y)",
#'   linters = expect_comparison_linter()
#' )
#'
#' lint(
#'   text = "expect_lte(x, y)",
#'   linters = expect_comparison_linter()
#' )
#'
#' lint(
#'   text = "expect_identical(x, y == 2)",
#'   linters = expect_comparison_linter()
#' )
#'
#' lint(
#'   text = "expect_true(x < y | x > y^2)",
#'   linters = expect_comparison_linter()
#' )
#'
#' @evalRd rd_tags("expect_comparison_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
expect_comparison_linter <- function() {
  # != doesn't have a clean replacement
  comparator_nodes <- setdiff(infix_metadata$xml_tag[infix_metadata$comparator], "NE")
  xpath <- glue("
  following-sibling::expr[1][ {xp_or(comparator_nodes)} ]
    /parent::expr[not(SYMBOL_SUB[text() = 'info'])]
  ")

  comparator_expectation_map <- c(
    `>` = "expect_gt", `>=` = "expect_gte",
    `<` = "expect_lt", `<=` = "expect_lte",
    `==` = "expect_identical"
  )

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls("expect_true")
    bad_expr <- xml_find_all_(xml_calls, xpath)

    comparator <- xml_find_chr_(bad_expr, "string(expr[2]/*[not(self::COMMENT)][2])")
    expectation <- comparator_expectation_map[comparator]
    lint_message <- sprintf("%s(x, y) is better than expect_true(x %s y).", expectation, comparator)
    xml_nodes_to_lints(bad_expr, source_expression, lint_message = lint_message, type = "warning")
  })
}

#' Block instances of unnecessary nesting
#'
#' Excessive nesting harms readability. Use helper functions or early returns
#'   to reduce nesting wherever possible.
#'
#' @param allow_assignment Logical, default `TRUE`, in which case
#'   braced expressions consisting only of a single assignment are skipped.
#'   if `FALSE`, all braced expressions with only one child expression are linted.
#'   The `TRUE` case facilitates interaction with [implicit_assignment_linter()]
#'   for certain cases where an implicit assignment is necessary, so a braced
#'   assignment is used to further distinguish the assignment. See examples.
#' @param allow_functions Character vector of functions which always allow
#'   one-child braced expressions. `testthat::test_that()` is always allowed because
#'   testthat requires a braced expression in its `code` argument. The other defaults
#'   similarly compute on expressions in a way which is worth highlighting by
#'   em-bracing them, even if there is only one expression, while [switch()] is allowed
#'   for its use as a control flow analogous to `if`/`else`.]
#' @param branch_exit_calls Character vector of functions which are considered
#'   as "exiting" a branch for the purpose of recommending removing nesting in a branch
#'   _lacking_ an exit call when the other branch terminates with one. Calls which
#'   always interrupt or quit the current call or R session,
#'   e.g. [stop()] and [q()], are always included.
#' @examples
#' # will produce lints
#' code <- "if (A) {\n  stop('A is bad!')\n} else {\n  do_good()\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' code <- "tryCatch(\n  {\n    foo()\n  },\n  error = identity\n)"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' code <- "expect_warning(\n  {\n    x <- foo()\n  },\n  'warned'\n)"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter(allow_assignment = FALSE)
#' )
#'
#' code <- "if (x) { \n  if (y) { \n   return(1L) \n  } \n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' lint(
#'   text = "my_quote({x})",
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' code <- paste(
#'   "if (A) {",
#'   "  stop('A is bad because a.')",
#'   "} else {",
#'   "  warning('!A requires caution.')",
#'   "}",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' # okay
#' code <- "if (A) {\n  stop('A is bad because a.')\n} else {\n  stop('!A is bad too.')\n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' code <- "capture.output({\n  foo()\n})"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' code <- "expect_warning(\n  {\n    x <- foo()\n  },\n  'warned'\n)"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' code <- "if (x && y) { \n  return(1L) \n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' code <- "if (x) { \n  y <- x + 1L\n  if (y) { \n   return(1L) \n  } \n}"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter()
#' )
#'
#' lint(
#'   text = "my_quote({x})",
#'   linters = unnecessary_nesting_linter(allow_functions = "my_quote")
#' )
#'
#' code <- paste(
#'   "if (A) {",
#'   "  stop('A is bad because a.')",
#'   "} else {",
#'   "  warning('!A requires caution.')",
#'   "}",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = unnecessary_nesting_linter(branch_exit_calls = c("stop", "warning"))
#' )
#'
#' @evalRd rd_tags("unnecessary_nesting_linter")
#' @seealso
#'  - [cyclocomp_linter()] for another linter that penalizes overly complex code.
#'  - [linters] for a complete list of linters available in lintr.
#' @export
unnecessary_nesting_linter <- function(
  allow_assignment = TRUE,
  allow_functions = c(
    "switch",
    "try", "tryCatch", "withCallingHandlers",
    "quote", "expression", "bquote", "substitute",
    "with_parameters_test_that",
    "reactive", "observe", "observeEvent",
    "renderCachedPlot", "renderDataTable", "renderImage", "renderPlot",
    "renderPrint", "renderTable", "renderText", "renderUI"
  ),
  branch_exit_calls = character()
) {
  default_branch_exit_calls <- c("stop", "return", "abort", "quit", "q")
  branch_exit_calls <- union(default_branch_exit_calls, branch_exit_calls)

  exit_call_expr <- glue("
    expr[SYMBOL_FUNCTION_CALL[{xp_text_in_table(branch_exit_calls)}]]
  ")
  # block IF here for cases where a nested if/else is entirely within
  #   one of the branches.
  no_exit_call_expr <- glue("
  expr[
    OP-LEFT-BRACE
    and expr[
      position() = last()
      and not(IF)
      and not(expr[SYMBOL_FUNCTION_CALL[{xp_text_in_table(branch_exit_calls)}]])
    ]
  ]")
  # condition for ELSE should be redundant, but include for robustness
  # condition on parent::expr[IF] ensures we're at the first `if` of a sequence of if/else statements
  # condition on expr uses following-sibling or preceding-sibling to ensure
  #   that the other expr falls on a different branch (earlier used separate
  #   conditions on two expr[], but seems to allow any branch with >1 statement
  #   to lead to a lint.
  # use position() = last() to ignore any expr but the last one in any branch.
  if_else_exit_xpath <- glue("
  //IF/parent::expr[
    ELSE
    and not(parent::expr[IF])
    and expr[
      OP-LEFT-BRACE
      and expr[position() = last() and {exit_call_expr}]
      and (
        following-sibling::{no_exit_call_expr}
        or preceding-sibling::{no_exit_call_expr}
      )
    ]
  ]
  ")

  used_exit_call_xpath <- glue("expr/expr[position() = last()]/{exit_call_expr}")

  assignment_cond <- if (allow_assignment) "*[LEFT_ASSIGN or RIGHT_ASSIGN or EQ_ASSIGN]" else "false"

  # several carve-outs of common cases where single-expression braces are OK
  #   - control flow statements: if, for, while, repeat
  #       + include foreach() as a common package-based for loop extension
  #   - function definitions
  #       + includes purrr-like anonymous functions as ~ {...}
  #   - rlang's double-brace expressions like {{ var }}
  #       + NB: both braces would trigger here, so we must exclude both of them
  #   - any expression starting like `({` or `[{` or ending like `})` or `}]`
  #       + note that nesting is not improved by "fixing" such cases, and could also be worsened
  #       + motivated by the most common cases:
  #          * test_that("test", { expr })
  #          * with(x, { expr }) / within(x, { expr })
  #          * suppressWarnings({ expr })
  #          * DataTable[, { expr }]
  #          * DataTable[, col := { expr }] <- requires carve-out for `:=`
  # 'count(*) - ...' used to be just 'count(expr)', but that misses '=' assigments (exprlist).
  unnecessary_brace_xpath <- glue("
  //OP-LEFT-BRACE
    /parent::expr[
      count(*) - count(COMMENT) - count(OP-LEFT-BRACE) - count(OP-RIGHT-BRACE) = 1
      and not(preceding-sibling::*[
        self::FUNCTION
        or self::OP-LAMBDA
        or self::FOR
        or self::IF
        or self::WHILE
        or self::REPEAT
        or self::expr/SYMBOL_FUNCTION_CALL[{ xp_text_in_table(c('test_that', allow_functions)) }]
        or self::expr/expr/SYMBOL_FUNCTION_CALL[text() = 'foreach']
        or self::OP-TILDE
        or self::LEFT_ASSIGN[text() = ':=']
      ])
      and not(expr/OP-LEFT-BRACE)
      and not(preceding-sibling::OP-LEFT-BRACE)
      and not(
        OP-LEFT-BRACE/@end - 1 = preceding-sibling::*[1][self::OP-LEFT-PAREN or self::OP-LEFT-BRACKET]/@end
        or OP-RIGHT-BRACE/@end + 1 = following-sibling::*[1][self::OP-RIGHT-PAREN or self::OP-RIGHT-BRACKET]/@end
      )
      and not({assignment_cond})
    ]
  ")

  unnecessary_nested_if_xpath <- paste0(
    "//IF/parent::expr[not(ELSE)]/OP-RIGHT-PAREN/",
    c(
      # catch if (cond) if (other_cond) { ... }
      "following-sibling::expr[IF and not(ELSE)]",
      # catch if (cond) { if (other_cond) { ... } }
      #   count(*): only OP-LEFT-BRACE, one <expr>, and OP-RIGHT-BRACE.
      #             Note that third node could be <expr_or_assign_or_help>.
      "following-sibling::expr[OP-LEFT-BRACE and count(*) - count(COMMENT) = 3]/expr[IF and not(ELSE)]"
    ),
    collapse = " | "
  )
  # "un-walk" from the unnecessary IF to the IF with which it should be combined
  corresponding_if_xpath <- "preceding-sibling::IF | parent::expr/preceding-sibling::IF"

  unnecessary_else_brace_xpath <- "
  //IF/parent::expr[parent::expr[
    preceding-sibling::ELSE and count(*) - count(COMMENT) - count(OP-LEFT-BRACE) - count(OP-RIGHT-BRACE) = 1
  ]]"

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    if_else_exit_expr <- xml_find_all_(xml, if_else_exit_xpath)
    used_exit_call <- get_r_string(if_else_exit_expr, used_exit_call_xpath)
    if_else_exit_lints <- xml_nodes_to_lints(
      if_else_exit_expr,
      source_expression = source_expression,
      lint_message = paste0(
        "Reduce the nesting of this if/else statement by unnesting the ",
        "portion without an exit clause, i.e., ", used_exit_call, "()."
      ),
      type = "warning"
    )

    unnecessary_brace_expr <- xml_find_all_(xml, unnecessary_brace_xpath)
    unnecessary_brace_lints <- xml_nodes_to_lints(
      unnecessary_brace_expr,
      source_expression = source_expression,
      lint_message = "Reduce the nesting of this statement by removing the braces {}.",
      type = "warning"
    )

    unnecessary_nested_if_expr <- xml_find_all_(xml, unnecessary_nested_if_xpath)
    corresponding_brace <- xml_find_first_(unnecessary_nested_if_expr, corresponding_if_xpath)
    corresponding_line <- xml_attr_(corresponding_brace, "line1")
    corresponding_column <- xml_attr_(corresponding_brace, "col1")
    unnecessary_nested_if_lints <- xml_nodes_to_lints(
      unnecessary_nested_if_expr,
      source_expression = source_expression,
      lint_message = paste0(
        "Combine this `if` statement with the one found at line ",
        corresponding_line, ", column ", corresponding_column, " to reduce nesting."
      ),
      type = "warning"
    )

    unnecessary_else_brace_expr <- xml_find_all_(xml, unnecessary_else_brace_xpath)
    unnecessary_else_brace_lints <- xml_nodes_to_lints(
      unnecessary_else_brace_expr,
      source_expression = source_expression,
      lint_message = "Simplify this condition by using 'else if' instead of 'else { if.",
      type = "warning"
    )

    c(if_else_exit_lints, unnecessary_brace_lints, unnecessary_nested_if_lints, unnecessary_else_brace_lints)
  })
}

#' Clear the lintr cache
#'
#' @param file filename whose cache to clear. If you pass `NULL`, it will delete all of the caches.
#' @param path directory to store caches. Reads option 'lintr.cache_directory' as the default.
#' @return 0 for success, 1 for failure, invisibly.
#' @export
clear_cache <- function(file = NULL, path = NULL) {
  if (is.null(path)) {
    # Only retrieve settings if `path` isn't specified.
    # Otherwise, other settings may inadvertently be loaded, such as exclusions.
    read_settings(file)
    path <- settings$cache_directory
  }

  if (!is.null(file)) {
    path <- get_cache_file_path(file, path)
  }

  unlink(path, recursive = TRUE)
}

define_cache_path <- function(cache) {
  if (isTRUE(cache)) {
    settings$cache_directory
  } else if (is.character(cache)) {
    cache
  } else {
    character()
  }
}

define_cache_key <- function(filename, inline_data, lines) {
  if (inline_data) list(content = get_content(lines), TRUE) else filename
}


get_cache_file_path <- function(file, path) {
  # this assumes that a normalized absolute file path was given
  file.path(path, digest::digest(file, algo = "sha1"))
}

load_cache <- function(file, path = NULL) {
  if (is.null(path) || length(path) == 0L) {
    return()
  }
  env <- new.env(parent = emptyenv())

  file <- get_cache_file_path(file, path)
  if (file.exists(file)) {
    tryCatch(
      load(file = file, envir = env),
      warning = function(w) {
        invokeRestart("muffleWarning")
      },
      error = function(e) {
        cli_warn(
          "Could not load cache file {.file {file}}:",
          parent = e
        )
      }
    )
  } # else nothing to do for source file that has no cache

  env
}

save_cache <- function(cache, file, path = NULL) {
  if (is.null(cache)) {
    return()
  }
  if (!file.exists(path)) {
    dir.create(path, recursive = TRUE)
  }

  save(file = get_cache_file_path(file, path), envir = cache, list = ls(envir = cache))
}

cache_file <- function(cache, filename, linters, lints) {
  if (is.null(cache)) {
    return()
  }
  assign(
    envir = cache,
    x = digest_content(linters, filename),
    value = lints,
    inherits = FALSE
  )
}

retrieve_file <- function(cache, filename, linters) {
  if (is.null(cache)) {
    return(NULL)
  }
  mget(
    envir = cache,
    x = digest_content(linters, filename),
    mode = "list",
    inherits = FALSE,
    ifnotfound = list(NULL)
  )[[1L]]
}

cache_lint <- function(cache, expr, linter, lints) {
  if (is.null(cache)) {
    return()
  }
  assign(
    envir = cache,
    x = digest_content(linter, expr),
    value = lints,
    inherits = FALSE
  )
}

retrieve_lint <- function(cache, expr, linter, lines) {
  lints <- get(
    envir = cache,
    x = digest_content(linter, expr),
    mode = "list",
    inherits = FALSE
  )
  for (i in seq_along(lints)) {
    new_line_number <- find_new_line(
      lints[[i]]$line_number,
      unname(lints[[i]]$line),
      lines
    )
    if (is.na(new_line_number)) {
      return(NULL)
    }
    lints[[i]]$line_number <- new_line_number
  }
  cache_lint(cache, expr, linter, lints)
  lints
}

has_lint <- function(cache, expr, linter) {
  if (is.null(cache)) {
    return(FALSE)
  }
  exists(
    envir = cache,
    x = digest_content(linter, expr),
    mode = "list",
    inherits = FALSE
  )
}

digest_content <- function(linters, obj) {
  content <- if (is.list(obj)) {
    # assume an expression (global expression if obj$parsed_content is lacking)
    list(linters, obj$content, is.null(obj$parsed_content))
  } else {
    # assume a filename
    list(linters, readLines(obj))
  }
  digest::digest(content, algo = "sha1")
}

find_new_line <- function(line_number, line, lines) {
  if (lines[line_number] %==% line) {
    return(line_number)
  }

  width <- 1L

  while (width <= length(lines)) {
    low <- line_number - width
    if (low > 0L && lines[low] %==% line) {
      return(low)
    }
    high <- line_number + width
    if (high <= length(lines) && lines[high] %==% line) {
      return(high)
    }
    width <- width + 1L
  }
  NA
}

#' Character string quote linter
#'
#' Check that the desired quote delimiter is used for string constants.
#'
#' @param delimiter Which quote delimiter to accept. Defaults to the tidyverse
#'   default of `"` (double-quoted strings).
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "c('a', 'b')",
#'   linters = quotes_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'c("a", "b")',
#'   linters = quotes_linter()
#' )
#'
#' code_lines <- "paste0(x, '\"this is fine\"')"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = quotes_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "c('a', 'b')",
#'   linters = quotes_linter(delimiter = "'")
#' )
#'
#' @evalRd rd_tags("quotes_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#character-vectors>
#' @export
quotes_linter <- function(delimiter = c('"', "'")) {
  delimiter <- match.arg(delimiter)
  if (delimiter == '"') {
    quote_regex <- rex(
      start,
      zero_or_one(character_class("rR")),
      single_quote,
      any_non_double_quotes,
      single_quote,
      end
    )
    lint_message <- "Only use double-quotes."
  } else {
    quote_regex <- rex(
      start,
      zero_or_one(character_class("rR")),
      double_quote,
      any_non_single_quotes,
      double_quote,
      end
    )
    lint_message <- "Only use single-quotes."
  }

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    string_exprs <- xml_find_all_(xml, "//STR_CONST")
    is_bad <- re_matches(xml_text(string_exprs), quote_regex)

    xml_nodes_to_lints(string_exprs[is_bad], source_expression, lint_message)
  })
}

#' Require usage of switch() over repeated if/else blocks
#'
#' [switch()] statements in R are used to delegate behavior based
#'   on the value of some input scalar string, e.g.
#'   `switch(x, a = 1, b = 3, c = 7, d = 8)` will be one of
#'   `1`, `3`, `7`, or `8`, depending on the value of `x`.
#'
#' This can also be accomplished by repeated `if`/`else` statements like
#'   so: `if (x == "a") 1 else if (x == "b") 2 else if (x == "c") 7 else 8`
#'   (implicitly, the last `else` assumes x only takes 4 possible values),
#'   but this is more cluttered and slower (note that `switch()` takes the same
#'   time to evaluate regardless of the value of `x`, and is faster even
#'   when `x` takes the first value (here `a`), and that the `if`/`else`
#'   approach is roughly linear in the number of conditions that need to
#'   be evaluated, here up to 3 times).
#'
#' @param max_branch_lines,max_branch_expressions Integer, default 0 indicates "no maximum".
#'   If set any `if`/`else if`/.../`else` chain where any branch occupies more than
#'   this number of lines (resp. expressions) will not be linted. The conjugate
#'   applies to `switch()` statements -- if these parameters are set, any `switch()`
#'   statement with any overly-complicated branches will be linted. See examples.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "if (x == 'a') 1 else if (x == 'b') 2 else 3",
#'   linters = if_switch_linter()
#' )
#'
#' code <- paste(
#'   "if (x == 'a') {",
#'   "  1",
#'   "} else if (x == 'b') {",
#'   "  2",
#'   "} else if (x == 'c') {",
#'   "  y <- x",
#'   "  z <- sqrt(match(y, letters))",
#'   "  z",
#'   "}",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = if_switch_linter()
#' )
#'
#' code <- paste(
#'   "if (x == 'a') {",
#'   "  1",
#'   "} else if (x == 'b') {",
#'   "  2",
#'   "} else if (x == 'c') {",
#'   "  y <- x",
#'   "  z <- sqrt(",
#'   "    match(y, letters)",
#'   "  )",
#'   "  z",
#'   "}",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = if_switch_linter()
#' )
#'
#' code <- paste(
#'   "switch(x,",
#'   "  a = {",
#'   "    1",
#'   "    2",
#'   "    3",
#'   "  },",
#'   "  b = {",
#'   "    1",
#'   "    2",
#'   "  }",
#'   ")",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = if_switch_linter(max_branch_lines = 2L)
#' )
#'
#' # okay
#' lint(
#'   text = "switch(x, a = 1, b = 2, 3)",
#'   linters = if_switch_linter()
#' )
#'
#' # switch() version not as clear
#' lint(
#'   text = "if (x == 'a') 1 else if (x == 'b' & y == 2) 2 else 3",
#'   linters = if_switch_linter()
#' )
#'
#' code <- paste(
#'   "if (x == 'a') {",
#'   "  1",
#'   "} else if (x == 'b') {",
#'   "  2",
#'   "} else if (x == 'c') {",
#'   "  y <- x",
#'   "  z <- sqrt(match(y, letters))",
#'   "  z",
#'   "}",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = if_switch_linter(max_branch_lines = 2L)
#' )
#'
#' code <- paste(
#'   "if (x == 'a') {",
#'   "  1",
#'   "} else if (x == 'b') {",
#'   "  2",
#'   "} else if (x == 'c') {",
#'   "  y <- x",
#'   "  z <- sqrt(",
#'   "    match(y, letters)",
#'   "  )",
#'   "  z",
#'   "}",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = if_switch_linter(max_branch_expressions = 2L)
#' )
#'
#' code <- paste(
#'   "switch(x,",
#'   "  a = {",
#'   "    1",
#'   "    2",
#'   "    3",
#'   "  },",
#'   "  b = {",
#'   "    1",
#'   "    2",
#'   "  }",
#'   ")",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = if_switch_linter(max_branch_lines = 3L)
#' )
#'
#' @evalRd rd_tags("if_switch_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
if_switch_linter <- function(max_branch_lines = 0L, max_branch_expressions = 0L) {
  equal_str_cond <- "expr[1][EQ and expr/STR_CONST[string-length(text()) > 2]]"

  if (max_branch_lines > 0L || max_branch_expressions > 0L) {
    complexity_cond <- xp_or(c(
      if (max_branch_lines > 0L) {
        c(
          paste("OP-LEFT-BRACE and (OP-RIGHT-BRACE/@line2 - OP-LEFT-BRACE/@line1 > 1 +", max_branch_lines, ")"),
          paste("not(OP-LEFT-BRACE) and (@line2 - @line1 >=", max_branch_lines, ")")
        )
      },
      if (max_branch_expressions > 0L) paste("OP-LEFT-BRACE and count(expr) >", max_branch_expressions)
    ))
    branch_expr_cond <- xp_and(c(
      xp_or(
        # if (x) { <this expr> } ...
        xp_and("preceding-sibling::IF", "position() = 2"),
        # if (x) { ... } else { <this expr> }
        xp_and("preceding-sibling::ELSE", "not(IF)")
      ),
      complexity_cond
    ))
    max_lines_cond <- glue(".//expr[{branch_expr_cond}]")

    switch_xpath <- glue("
    parent::expr[expr[
      position() > 2
      and {complexity_cond}
    ]]")
  } else {
    max_lines_cond <- "false"

    switch_xpath <- NULL
  }

  if_xpath <- glue("
  //IF
    /parent::expr[
      not(preceding-sibling::IF)
      and {equal_str_cond}
      and ELSE/following-sibling::expr[
        IF
        and {equal_str_cond}
        and ELSE/following-sibling::expr[IF and {equal_str_cond}]
      ]
      and not({ max_lines_cond })
    ]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, if_xpath)

    bad_expr_clean <- strip_comments_from_subtree(bad_expr)
    expr_all_equal <- vapply(bad_expr_clean, if_else_chain_expr_is_unique, logical(1L))
    bad_expr <- bad_expr[expr_all_equal]

    # Exclude empty strings, which can't be used as switch() case names
    nonempty <- vapply(bad_expr, function(expr) {
      str_nodes <- if_else_chain_strings(expr)
      all(vapply(str_nodes, function(n) nzchar(get_r_string(n)), logical(1L)))
    }, logical(1L))
    bad_expr <- bad_expr[nonempty]

    lints <- xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste(
        "Prefer switch() statements over repeated if/else equality tests,",
        "e.g., switch(x, a = 1, b = 2) over",
        'if (x == "a") 1 else if (x == "b") 2.'
      ),
      type = "warning"
    )

    if (!is.null(switch_xpath)) {
      xml_calls <- source_expression$xml_find_function_calls("switch")
      switch_expr <- xml_find_all_(xml_calls, switch_xpath)

      lints <- c(lints, xml_nodes_to_lints(
        switch_expr,
        source_expression = source_expression,
        lint_message = "Prefer repeated if/else statements over overly-complicated switch() statements.",
        type = "warning"
      ))
    }

    lints
  })
}

# Extract STR_CONST nodes from equality conditions in an if/else if chain
if_else_chain_strings <- function(expr) {
  str_nodes <- list()
  first <- xml_find_first_(expr, "IF/following-sibling::expr[1][EQ]/expr/STR_CONST")
  if (!is.na(first)) str_nodes <- c(str_nodes, list(first))
  current <- expr
  repeat {
    else_if <- xml_find_first_(current, "ELSE/following-sibling::expr[IF]")
    if (is.na(else_if)) break
    current <- else_if
    str_node <- xml_find_first_(current, "IF/following-sibling::expr[1][EQ]/expr/STR_CONST")
    if (!is.na(str_node)) str_nodes <- c(str_nodes, list(str_node))
  }
  str_nodes
}

# Check that equality conditions in an if/else if chain use the same expression
if_else_chain_expr_is_unique <- function(expr) {
  expr_nodes <- character()
  first <- xml_find_first_(expr, "IF/following-sibling::expr[1][EQ]/expr[not(STR_CONST)]")
  if (!is.na(first)) expr_nodes <- c(expr_nodes, xml_text(first))
  current <- expr
  repeat {
    else_if <- xml_find_first_(current, "ELSE/following-sibling::expr[IF]")
    if (is.na(else_if)) break
    current <- else_if
    expr_node <- xml_find_first_(current, "IF/following-sibling::expr[1][EQ]/expr[not(STR_CONST)]")
    if (!is.na(expr_node)) expr_nodes <- c(expr_nodes, xml_text(expr_node))
  }
  length(unique(expr_nodes)) == 1L
}

#' Require usage of nlevels over length(levels(.))
#'
#' `length(levels(x))` is the same as `nlevels(x)`, but harder to read.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "length(levels(x))",
#'   linters = length_levels_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "length(c(levels(x), levels(y)))",
#'   linters = length_levels_linter()
#' )
#'
#' @evalRd rd_tags("length_levels_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
length_levels_linter <- make_linter_from_function_xpath(
  function_names = "levels",
  xpath = "
  parent::expr
    /parent::expr[expr/SYMBOL_FUNCTION_CALL[text() = 'length']]
  ",
  lint_message = "nlevels(x) is better than length(levels(x))."
)

# here are some extraneous variables that are not part of the config directly

non_default_linter <- any_duplicated_linter()
attr(non_default_linter, "name") <- "any_duplicated_linter"
excluded_files <- "tests/testthat.R"

linters <- linters_with_defaults(
  non_default_linter,
  assignment_linter = NULL
)
exclude <- "# NOLINT"
exclusions <- list(excluded_files)

#' Require usage of a leading zero in all fractional numerics
#'
#' While .1 and 0.1 mean the same thing, the latter is easier to read due
#' to the small size of the '.' glyph.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x <- .1",
#'   linters = numeric_leading_zero_linter()
#' )
#'
#' lint(
#'   text = "x <- -.1",
#'   linters = numeric_leading_zero_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x <- 0.1",
#'   linters = numeric_leading_zero_linter()
#' )
#'
#' lint(
#'   text = "x <- -0.1",
#'   linters = numeric_leading_zero_linter()
#' )
#'
#' @evalRd rd_tags("numeric_leading_zero_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
numeric_leading_zero_linter <- make_linter_from_xpath(
  # NB:
  #  1. negative constants are split to two components:
  #    OP-MINUS, NUM_CONST
  #  2. complex constants are split to three components:
  #    NUM_CONST, OP-PLUS/OP-MINUS, NUM_CONST
  xpath = "//NUM_CONST[starts-with(text(), '.')]",
  lint_message = "Include the leading zero for fractional numeric constants."
)

#' Encourage usage of the null coalescing operator `%||%`
#'
#' The `x %||% y` is equivalent to
#'   `if (is.null(x)) y else x`, but more expressive.
#'   It is exported by R since 4.4.0, and equivalents
#'   have been available in other tidyverse packages
#'   for much longer, e.g. 2008 for ggplot2.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "if (is.null(x)) y else x",
#'   linters = coalesce_linter()
#' )
#'
#' lint(
#'   text = "if (!is.null(x)) x else y",
#'   linters = coalesce_linter()
#' )
#'
#' lint(
#'   text = "if (is.null(x[1])) x[2] else x[1]",
#'   linters = coalesce_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x %||% y",
#'   linters = coalesce_linter()
#' )
#'
#' lint(
#'   text = "x %||% y",
#'   linters = coalesce_linter()
#' )
#'
#' lint(
#'   text = "x[1] %||% x[2]",
#'   linters = coalesce_linter()
#' )
#'
#'
#' @evalRd rd_tags("coalesce_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
coalesce_linter <- function() {
  braced_expr_cond <- "expr[1][OP-LEFT-BRACE and count(*) = 3]/expr"
  xpath <- glue("
  expr[expr[
    preceding-sibling::IF
    and (
      expr[2] = following-sibling::ELSE/following-sibling::expr
      or expr[2] = following-sibling::ELSE/following-sibling::{braced_expr_cond}
      or expr[2][LEFT_ASSIGN]/expr[1] = following-sibling::ELSE/following-sibling::expr
      or expr[2][LEFT_ASSIGN]/expr[1] = following-sibling::ELSE/following-sibling::{braced_expr_cond}
    )
  ]]
  |
  self::*[expr[
    preceding-sibling::IF
    and following-sibling::ELSE
    and OP-EXCLAMATION
    and (
      expr/expr[2] = following-sibling::expr[1]
      or expr/expr[2] = following-sibling::{braced_expr_cond}
      or expr/expr[2][LEFT_ASSIGN]/expr[1] = following-sibling::expr[1]
      or expr/expr[2][LEFT_ASSIGN]/expr[1] = following-sibling::{braced_expr_cond}
    )
  ]]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    null_calls <- xml_find_all_(
      source_expression$xml_find_function_calls("is.null"),
      "parent::*/parent::*/parent::*"
    )
    null_calls <- strip_comments_from_subtree(null_calls)
    bad_expr <- xml_find_all_(null_calls, xpath)
    is_negation <- !is.na(xml_find_first_(bad_expr, "expr/OP-EXCLAMATION"))
    observed <- ifelse(is_negation, "if (!is.null(x)) x else y", "if (is.null(x)) y else x")

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste0("Use x %||% y instead of ", observed, "."),
      type = "warning"
    )
  })
}

#' Function argument linter
#'
#' @description
#' Check that arguments with defaults come last in all function declarations,
#' as per the tidyverse design guide.
#'
#' Changing the argument order can be a breaking change. An alternative to changing the argument order
#' is to instead set the default for such arguments to `NULL`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "function(y = 1, z = 2, x) {}",
#'   linters = function_argument_linter()
#' )
#'
#' lint(
#'   text = "function(x, y, z = 1, ..., w) {}",
#'   linters = function_argument_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "function(x, y = 1, z = 2) {}",
#'   linters = function_argument_linter()
#' )
#'
#' lint(
#'   text = "function(x, y, w, z = 1, ...) {}",
#'   linters = function_argument_linter()
#' )
#'
#' lint(
#'   text = "function(y = 1, z = 2, x = NULL) {}",
#'   linters = function_argument_linter()
#' )
#'
#' lint(
#'   text = "function(x, y, z = 1, ..., w = NULL) {}",
#'   linters = function_argument_linter()
#' )
#'
#' @evalRd rd_tags("function_argument_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://design.tidyverse.org/required-no-defaults.html>
#' @export
function_argument_linter <- function() {
  xpath <- "
  (//FUNCTION | //OP-LAMBDA)
    /following-sibling::EQ_FORMALS[1]
    /following-sibling::SYMBOL_FORMALS[
      text() != '...'
      and not(following-sibling::*[not(self::COMMENT)][1][self::EQ_FORMALS])
    ]
  "

  used_in_missing_xpath <- "
    text() = following-sibling::expr[last()]//expr[expr/SYMBOL_FUNCTION_CALL[text() = 'missing']]/expr[2]/SYMBOL/text()
  "

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)

    uses_missing <- xml_find_lgl_(bad_expr, used_in_missing_xpath)

    missing_note <-
      ifelse(uses_missing, " Consider setting the default to NULL and using is.null() instead of using missing()", "")

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste0("Arguments without defaults should come before arguments with defaults.", missing_note),
      type = "style"
    )
  })
}

# TODO(R>=4.4.0): remove this.
# NB: {backports} approach doesn't work since we need this before .onLoad() in names2()
if (!exists("%||%", "package:base")) {
  `%||%` <- function(x, y) if (is.null(x)) y else x # nolint: coalesce_linter.
}

`%==%` <- function(x, y) {
  identical(x, y)
}

`%:::%` <- function(p, f) {
  get(f, envir = asNamespace(p))
}

flatten_lints <- function(x) {
  x <- flatten_list(x, class = "lint")
  class(x) <- "lints"
  x
}

# any function using unlist or c was dropping the classnames,
# so need to brute force copy the objects
flatten_list <- function(x, class) {
  if (length(x) == 0L) {
    return(list())
  }
  if (inherits(x, class)) {
    return(list(x))
  }
  if (is.list(x)) {
    is_flat <- TRUE
    for (i in seq_along(x)) {
      if (!inherits(x[[i]], class)) {
        is_flat <- FALSE
        break
      }
    }
    if (is_flat) {
      if (!is.null(names(x))) {
        names(x) <- NULL
      }
      return(x)
    }
  }

  outer_env <- new.env(parent = emptyenv())
  outer_env$res <- list()
  outer_env$itr <- 1L
  assign_item <- function(x) {
    if (inherits(x, class)) {
      outer_env$res[[outer_env$itr]] <- x
      outer_env$itr <- outer_env$itr + 1L
    } else if (is.list(x)) {
      lapply(x, assign_item)
    }
  }
  assign_item(x)
  outer_env$res
}

fix_names <- function(x, default) {
  nms <- names(x)

  if (is.null(nms)) {
    nms <- default
  } else {
    nms[!nzchar(nms)] <- default
  }
  names(x) <- nms
  x
}

linter_auto_name <- function(which = -3L) {
  sys_call <- sys.call(which = which)
  nm <- deparse1(sys_call)
  regex <- rex(start, one_or_more(alnum %or% "." %or% "_" %or% ":"))
  if (re_matches(nm, regex)) {
    match_data <- re_matches(nm, regex, locations = TRUE)
    nm <- nm |>
      substr(start = 1L, stop = match_data[1L, "end"]) |>
      re_substitutes(rex(start, alnums, "::"), "")
  }
  nm
}

auto_names <- function(x) {
  nms <- names2(x)
  empty <- !nzchar(nms, keepNA = TRUE)
  if (!any(empty)) {
    return(nms)
  }

  default_name <- function(x) {
    if (is_linter(x)) {
      attr(x, "name", exact = TRUE)
    } else {
      deparse1(x)
    }
  }
  defaults <- vapply(x[empty], default_name, character(1L), USE.NAMES = FALSE)

  nms[empty] <- defaults
  nms
}

# The following functions is from dplyr
names2 <- function(x) {
  names(x) %||% rep("", length(x))
}

#' @param needs_braces Logical, default FALSE. If it's possible or known that
#'   `lines` cannot be parsed as a standalone expression, only a "child"
#'   expression, we insert braces `{}` around it to ensure that it parses.
#'   There is only one known case for this, namely, when finding code with
#'   shorthand lambda `\(` where `\` and `(` are separated by a comment;
#'   see `object_usage_linter()` for more details.
#' @noRd
get_content <- function(lines, info, needs_braces = FALSE) {
  lines[is.na(lines)] <- ""

  if (!missing(info)) {
    # put in data.frame-like format
    if (is_node(info)) {
      info <- lapply(xml_attrs_(info), as.integer)
    }

    lines <- lines[seq(info$line1, info$line2)]
    lines[length(lines)] <- substr(lines[length(lines)], 1L, info$col2)
    lines[1L] <- substr(lines[1L], info$col1, nchar(lines[1L]))
  }
  if (needs_braces) lines <- c("{", lines, "}")
  paste(lines, collapse = "\n")
}

try_silently <- function(expr) {
  suppressWarnings(
    suppressMessages(
      try(expr, silent = TRUE)
    )
  )
}

# interface to work like options() or setwd() -- returns the old value for convenience
set_lang <- function(new_lang) {
  old_lang <- Sys.getenv("LANGUAGE", unset = NA)
  Sys.setenv(LANGUAGE = new_lang) # nolint: undesirable_function_name. Avoiding {withr} dep in pkg.
  old_lang
}
# handle the logic of either unsetting if it was previously unset, or resetting
reset_lang <- function(old_lang) {
  if (is.na(old_lang)) {
    Sys.unsetenv("LANGUAGE")
  } else {
    Sys.setenv(LANGUAGE = old_lang) # nolint: undesirable_function_name. Avoiding {withr} dep in pkg.
  }
}

#' Create a `linter` closure
#'
#' @param fun A function that takes a source file and returns `lint` objects.
#' @param name Default name of the Linter.
#' Lints produced by the linter will be labelled with `name` by default.
#' @param linter_level Which level of expression is the linter working with?
#'   `"expression"` means an individual expression in `xml_parsed_content`, while `"file"` means all expressions
#'   in the current file are available in `full_xml_parsed_content`.
#'   `NA` means the linter will be run with both, expression-level and file-level source expressions.
#'
#' @return The same function with its class set to 'linter'.
#' @export
Linter <- function(fun, name = linter_auto_name(), linter_level = c(NA_character_, "file", "expression")) { # nolint: object_name, line_length.
  if (!is.function(fun) || length(formals(args(fun))) != 1L) {
    cli_abort("{.arg fun} must be a function taking exactly one argument.")
  }
  linter_level <- match.arg(linter_level)
  force(name)
  class(fun) <- c("linter", "function")
  attr(fun, "name") <- name
  attr(fun, "linter_level") <- linter_level
  fun
}

ensure_utf8 <- function(lines, encoding = NULL) {
  if (is.null(encoding) || is.na(encoding)) {
    encoding <- ""
  }

  if (encoding == "" && l10n_info()[["UTF-8"]]) {
    encoding <- "UTF-8"
  }

  if (encoding == "UTF-8") {
    Encoding(lines) <- "UTF-8"
    return(lines)
  }

  is_utf8 <- Encoding(lines) == "UTF-8"
  if (any(is_utf8)) {
    return(lines)
  }

  lines_conv <- iconv(lines, from = encoding, to = "UTF-8")
  lines[!is.na(lines_conv)] <- lines_conv[!is.na(lines_conv)]
  Encoding(lines) <- "UTF-8"
  lines
}

read_lines <- function(file, ...) {
  outer_env <- new.env(parent = emptyenv())
  outer_env$terminal_newline <- TRUE
  lines <- withCallingHandlers(
    readLines(file, warn = TRUE, ...),
    warning = function(w) {
      if (grepl("incomplete final line found on", conditionMessage(w), fixed = TRUE)) {
        outer_env$terminal_newline <- FALSE
        invokeRestart("muffleWarning")
      }
    }
  )
  attr(lines, "terminal_newline") <- outer_env$terminal_newline
  lines
}

# see issue #923, PR #2455 -- some locales ignore _ when running sort(), others don't.
#   We want to consistently treat "_" < "n" = "N"; C locale does this, which 'radix' uses.
platform_independent_order <- function(x) order(tolower(x), method = "radix")
platform_independent_sort <- function(x) x[platform_independent_order(x)]

#' re_matches with type-stable logical output
#' TODO(r-lib/rex#94): Use re_matches() option directly & deprecate this.
#' @noRd
re_matches_logical <- function(x, regex, ...) {
  res <- re_matches(x, regex, ...)
  if (is.data.frame(res)) {
    res <- complete.cases(res)
  }
  res
}

#' re_matches with type-stable locations output for the overall match
#' @noRd
re_matches_locations <- function(x, regex, ...) {
  m <- regexpr(regex, x, perl = TRUE, ...)
  match_start <- as.vector(m)
  match_end <- match_start + attr(m, "match.length") - 1L
  matched <- match_start != -1L
  match_start[!matched] <- NA_integer_
  match_end[!matched] <- NA_integer_
  data.frame(start = match_start, end = match_end)
}

#' Extract text from `STR_CONST` nodes
#'
#' Convert `STR_CONST` `text()` values into R strings. This is useful to account for arbitrary
#'  character literals, e.g. `R"------[hello]------"`, which is parsed in R as `"hello"`.
#'  It is quite cumbersome to write XPaths allowing for strings like this, so whenever your
#'  linter logic requires testing a `STR_CONST` node's value, use this function.
#' NB: this is also properly vectorized on `s`, and accepts a variety of inputs. Empty inputs
#'  will become `NA` outputs, which helps ensure that `length(get_r_string(s)) == length(s)`.
#'
#' @param s An input string or strings. If `s` is an `xml_node` or `xml_nodeset` and `xpath` is `NULL`,
#'   extract its string value with [xml2::xml_text()]. If `s` is an `xml_node` or `xml_nodeset`
#'   and `xpath` is specified, it is extracted with [xml2::xml_find_chr()].
#' @param xpath An XPath, passed on to [xml2::xml_find_chr()] after wrapping with `string()`.
#'
#' @examples
#' tmp <- tempfile()
#' writeLines("c('a', 'b')", tmp)
#' expr_as_xml <- get_source_expressions(tmp)$expressions[[1L]]$xml_parsed_content
#' writeLines(as.character(expr_as_xml))
#' get_r_string(expr_as_xml, "expr[2]")
#' get_r_string(expr_as_xml, "expr[3]")
#' unlink(tmp)
#'
#' # more importantly, extract raw strings correctly
#' tmp_raw <- tempfile()
#' writeLines("c(R'(a\\b)', R'--[a\\\"\'\"\\b]--')", tmp_raw)
#' expr_as_xml_raw <- get_source_expressions(tmp_raw)$expressions[[1L]]$xml_parsed_content
#' writeLines(as.character(expr_as_xml_raw))
#' get_r_string(expr_as_xml_raw, "expr[2]")
#' get_r_string(expr_as_xml_raw, "expr[3]")
#' unlink(tmp_raw)
#'
#' @export
get_r_string <- function(s, xpath = NULL) {
  if (length(s) == 0L) {
    return(character())
  }
  if (is_node(s) || is_nodeset(s)) {
    if (is.null(xpath)) {
      s <- xml_text(s)
    } else {
      s <- xml_find_chr_(s, sprintf("string(%s)", xpath))
    }
  }
  r_string_from_parse_text(s)
}

# parse() skips "" elements --> offsets the length of the output,
#   but NA in --> NA out
r_string_from_parse_text <- function(s) {
  is.na(s) <- !nzchar(s)
  out <- as.character(parse(text = s, keep.source = FALSE))
  is.na(out) <- is.na(s)
  out
}

is_linter <- function(x) inherits(x, "linter")
is_lint <- function(x) inherits(x, "lint")

is_error <- function(x) inherits(x, "error")

is_tainted <- function(lines) {
  is_error(tryCatch(nchar(lines), error = identity))
}

#' Check that the entries in ... are valid
#'
#' @param dot_names Supplied names, from [...names()].
#' @param ref_calls Functions consuming these `...` (character).
#' @param ref_help Help page to refer users hitting an error to.
#' @noRd
check_dots <- function(dot_names, ref_calls, ref_help = as.character(sys.call(-1L)[[1L]])) {
  valid_args <- unlist(lapply(ref_calls, \(f) names(formals(f))))
  # TODO(#2502): needn't check is.na() after R 4.2.0
  is_valid <- is.na(dot_names) | !nzchar(dot_names) | dot_names %in% valid_args
  if (all(is_valid)) {
    return(invisible())
  }
  invalid_args <- dot_names[!is_valid] # nolint: object_usage_linter. TODO(#2252).
  cli_abort(c(
    x = "Found unknown arguments in `...`: {.arg {invalid_args}}.",
    i = "Check for typos and see ?{ref_help} for valid arguments."
  ))
}

cli_abort_internal <- function(...) {
  cli_abort(..., .internal = TRUE) # nocov
}

#' Prevent `ifelse()` from being used to produce `TRUE`/`FALSE` or `1`/`0`
#'
#' Expressions like `ifelse(x, TRUE, FALSE)` and `ifelse(x, FALSE, TRUE)` are
#'   redundant; just `x` or `!x` suffice in R code where logical vectors are a
#'   core data structure. `ifelse(x, 1, 0)` is also `as.numeric(x)`, but even
#'   this should be needed only rarely.
#'
#' @evalRd rd_tags("redundant_ifelse_linter")
#' @param allow10 Logical, default `FALSE`. If `TRUE`, usage like
#'   `ifelse(x, 1, 0)` is allowed, i.e., only usage like
#'   `ifelse(x, TRUE, FALSE)` is linted.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "ifelse(x >= 2.5, TRUE, FALSE)",
#'   linters = redundant_ifelse_linter()
#' )
#'
#' lint(
#'   text = "ifelse(x < 2.5, 1L, 0L)",
#'   linters = redundant_ifelse_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x >= 2.5",
#'   linters = redundant_ifelse_linter()
#' )
#'
#' # Note that this is just to show the strict equivalent of the example above;
#' # converting to integer is often unnecessary and the logical vector itself
#' # should suffice.
#' lint(
#'   text = "as.integer(x < 2.5)",
#'   linters = redundant_ifelse_linter()
#' )
#'
#' lint(
#'   text = "ifelse(x < 2.5, 1L, 0L)",
#'   linters = redundant_ifelse_linter(allow10 = TRUE)
#' )
#'
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
redundant_ifelse_linter <- function(allow10 = FALSE) {
  tf_xpath <- glue("
  parent::expr[
    expr[position() <= 4 and NUM_CONST[text() = 'TRUE']]
    and expr[position() <= 4 and NUM_CONST[text() = 'FALSE']]
    and (
      count(expr) = 4
      or expr[5]/NUM_CONST[text() = 'NA']
    )
  ]")

  num_xpath <- glue("
  parent::expr[
    expr[position() <= 4 and NUM_CONST[text() = '1' or text() = '1L']]
    and expr[position() <= 4 and NUM_CONST[text() = '0' or text() = '0L']]
    and (
      count(expr) = 4
      or expr[5]/NUM_CONST[text() = 'NA' or text() = 'NA_integer_' or text() = 'NA_real_']
    )
  ]")

  Linter(linter_level = "expression", function(source_expression) {
    xml_targets <- source_expression$xml_find_function_calls(ifelse_funs)

    lints <- list()

    tf_expr <- xml_find_all_(xml_targets, tf_xpath)
    matched_call <- xp_call_name(tf_expr)
    # [1] call; [2] logical condition
    first_arg <- xml_find_chr_(tf_expr, "string(expr[3]/NUM_CONST)")
    second_arg <- xml_find_chr_(tf_expr, "string(expr[4]/NUM_CONST)")
    tf_message <- sprintf(
      "Just use the logical condition (or its negation) directly instead of calling %s(x, %s, %s)",
      matched_call, first_arg, second_arg
    )
    lints <- c(lints, xml_nodes_to_lints(tf_expr, source_expression, tf_message, type = "warning"))

    if (!allow10) {
      num_expr <- xml_find_all_(xml_targets, num_xpath)
      matched_call <- xp_call_name(num_expr)
      # [1] call; [2] logical condition
      first_arg <- xml_find_chr_(num_expr, "string(expr[3]/NUM_CONST)")
      second_arg <- xml_find_chr_(num_expr, "string(expr[4]/NUM_CONST)")
      is_numeric_01 <- first_arg %in% c("0", "1") | second_arg %in% c("0", "1")
      coercion_function <- ifelse(is_numeric_01, "as.numeric", "as.integer")
      is_negated <- first_arg %in% c("0", "0L")
      replacement_argument <- ifelse(is_negated, "!x", "x")
      lint_message <- paste(
        sprintf(
          "Prefer %s(%s) to %s(x, %s, %s) if really needed.",
          coercion_function, replacement_argument, matched_call, first_arg, second_arg
        )
      )
      lints <- c(lints, xml_nodes_to_lints(num_expr, source_expression, lint_message, type = "warning"))
    }

    lints
  })
}

#' Block usage of `nrow(subset(x, .))`
#'
#' Using `nrow(subset(x, condition))` to count the instances where `condition`
#'   applies inefficiently requires doing a full subset of `x` just to
#'   count the number of rows in the resulting subset.
#' There are a number of equivalent expressions that don't require the full
#'   subset, e.g. `with(x, sum(condition))` (or, more generically,
#'   `with(x, sum(condition, na.rm = TRUE))`).
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "nrow(subset(x, is_treatment))",
#'   linters = nrow_subset_linter()
#' )
#'
#' lint(
#'   text = "nrow(filter(x, is_treatment))",
#'   linters = nrow_subset_linter()
#' )
#'
#' lint(
#'   text = "x %>% filter(x, is_treatment) %>% nrow()",
#'   linters = nrow_subset_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "with(x, sum(is_treatment, na.rm = TRUE))",
#'   linters = nrow_subset_linter()
#' )
#'
#' @evalRd rd_tags("nrow_subset_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @include shared_constants.R
#' @export
nrow_subset_linter <- make_linter_from_function_xpath(
  function_names = c("subset", "filter"),
  xpath = glue("
  parent::expr
    /parent::expr[
      expr/SYMBOL_FUNCTION_CALL[text() = 'nrow']
      or (self::expr | parent::expr)[
        (PIPE or SPECIAL[{ xp_text_in_table(setdiff(magrittr_pipes, c('%$%', '%<>%'))) }])
        and expr/expr/SYMBOL_FUNCTION_CALL[text() = 'nrow']
      ]
    ]
  "),
  lint_message = paste(
    "Use arithmetic to count the number of rows satisfying a condition,",
    "rather than fully subsetting the data.frame and counting the resulting rows.",
    "For example, replace nrow(subset(x, is_treatment)) with sum(x$is_treatment).",
    "NB: use na.rm = TRUE if `is_treatment` has missing values."
  )
)

#' Spaces inside linter
#'
#' Check that parentheses and square brackets do not have spaces directly
#'   inside them, i.e., directly following an opening delimiter or directly
#'   preceding a closing delimiter.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "c( TRUE, FALSE )",
#'   linters = spaces_inside_linter()
#' )
#'
#' lint(
#'   text = "x[ 1L ]",
#'   linters = spaces_inside_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "c(TRUE, FALSE)",
#'   linters = spaces_inside_linter()
#' )
#'
#' lint(
#'   text = "x[1L]",
#'   linters = spaces_inside_linter()
#' )
#'
#' @evalRd rd_tags("spaces_inside_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#parentheses>
#' @export
spaces_inside_linter <- function() {
  left_xpath_condition <- "
    not(following-sibling::*[1][self::COMMENT])
    and @end != following-sibling::*[1]/@start - 1
    and @line1 = following-sibling::*[1]/@line1
  "
  left_xpath <- glue("
  //OP-LEFT-BRACKET[{left_xpath_condition}]
  | //LBB[{left_xpath_condition}]
  | //OP-LEFT-PAREN[{left_xpath_condition}]")

  right_xpath_condition <- "
    not(preceding-sibling::*[1][self::OP-COMMA])
    and @start != preceding-sibling::*[1]/@end + 1
    and @line1 = preceding-sibling::*[1]/@line2
  "
  right_xpath <- glue("
  //OP-RIGHT-BRACKET[{right_xpath_condition}]
  | //OP-RIGHT-PAREN[{right_xpath_condition} and not(preceding-sibling::*[1][self::EQ_SUB])]")

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    left_expr <- xml_find_all_(xml, left_xpath)
    left_msg <- ifelse(
      xml_text(left_expr) %in% c("[", "[["),
      "Do not place spaces after square brackets.",
      "Do not place spaces after parentheses."
    )

    left_lints <- xml_nodes_to_lints(
      left_expr,
      source_expression = source_expression,
      lint_message = left_msg,
      range_start_xpath = "number(./@col2 + 1)", # start after ( or [
      range_end_xpath = "number(./following-sibling::*[1]/@col1 - 1)" # end before following expr
    )

    right_expr <- xml_find_all_(xml, right_xpath)
    right_msg <- ifelse(
      xml_text(right_expr) == "]",
      "Do not place spaces before square brackets.",
      "Do not place spaces before parentheses."
    )

    right_lints <- xml_nodes_to_lints(
      right_expr,
      source_expression = source_expression,
      lint_message = right_msg,
      range_start_xpath = "number(./preceding-sibling::*[1]/@col2 + 1)", # start after preceding expression
      range_end_xpath = "number(./@col1 - 1)" # end before ) or ]
    )

    c(left_lints, right_lints)
  })
}

#' @export
format.lint <- function(x, ..., width = getOption("lintr.format_width")) {
  color <- switch(x$type,
    warning = cli::col_magenta,
    error = cli::col_red,
    style = cli::col_blue,
    cli::style_bold
  )
  emph <- cli::style_bold

  line_ref <- build_line_ref(x)
  annotated_msg <- paste0(
    emph(line_ref, ": "),
    color(x$type, ": ", sep = ""),
    "[", x$linter, "] ",
    emph(x$message)
  )

  if (!is.null(width)) {
    annotated_msg <- paste(strwrap(annotated_msg, exdent = 4L, width = width), collapse = "\n")
  }

  paste0(
    annotated_msg, "\n",
    # swap tabs for spaces for #528 (sorry Richard Hendricks)
    chartr("\t", " ", x$line), "\n",
    highlight_string(x$message, x$column_number, x$ranges),
    "\n"
  )
}

build_line_ref <- function(x) {
  line_ref <- paste0(
    x$filename, ":",
    as.character(x$line_number), ":",
    as.character(x$column_number)
  )

  if (!cli::ansi_has_hyperlink_support()) {
    return(line_ref)
  }
  cli::format_inline("{.path {line_ref}}")
}

#' @export
print.lint <- function(x, ...) {
  cat(format(x, ...))
  invisible(x)
}

#' @export
format.lints <- function(x, ..., width = getOption("lintr.format_width")) {
  paste(vapply(x, format, character(1L), width = width), collapse = "\n")
}

#' @export
print.lints <- function(x, ...) {
  use_rstudio_source_markers <- lintr_option("rstudio_source_markers", TRUE) &&
    requireNamespace("rstudioapi", quietly = TRUE) &&
    rstudioapi::hasFun("sourceMarkers")

  github_annotation_project_dir <- lintr_option("github_annotation_project_dir", "")

  if (length(x) > 0L) {
    inline_data <- x[[1L]][["filename"]] == "<text>"
    if (!inline_data && use_rstudio_source_markers) {
      rstudio_source_markers(x)
    } else if (in_github_actions() && !in_pkgdown()) {
      github_actions_log_lints(x, project_dir = github_annotation_project_dir)
    } else {
      lapply(x, print, ...)
    }

    if (isTRUE(settings$error_on_lint)) {
      quit("no", 31L, FALSE) # nocov
    }
  } else {
    # Empty lints
    cli_inform(c(i = "No lints found."))
    if (use_rstudio_source_markers) {
      rstudio_source_markers(x) # clear RStudio source markers
    }
  }

  invisible(x)
}

#' @export
names.lints <- function(x, ...) {
  vapply(x, `[[`, character(1L), "filename")
}

#' @export
split.lints <- function(x, f = NULL, ...) {
  if (is.null(f)) f <- names(x)
  splt <- split.default(x, f)
  for (i in names(splt)) class(splt[[i]]) <- "lints"
  splt
}

#' @export
as.data.frame.lints <- function(x, row.names = NULL, optional = FALSE, ...) { # nolint: object_name. (row.names, #764)
  data.frame(
    filename = vapply(x, `[[`, character(1L), "filename"),
    line_number = vapply(x, `[[`, integer(1L), "line_number"),
    column_number = vapply(x, `[[`, integer(1L), "column_number"),
    type = vapply(x, `[[`, character(1L), "type"),
    message = vapply(x, `[[`, character(1L), "message"),
    line = vapply(x, `[[`, character(1L), "line"),
    linter = vapply(x, `[[`, character(1L), "linter")
  )
}

#' @exportS3Method tibble::as_tibble
as_tibble.lints <- function(x, ..., # nolint: object_name_linter.
                            .rows = NULL,
                            .name_repair = c("check_unique", "unique", "universal", "minimal"),
                            rownames = NULL) {
  if (!requireNamespace("tibble", quietly = TRUE)) {
    cli_abort("Package {.pkg tibble} is required to convert lints to a tibble.")
  }
  tibble::as_tibble(as.data.frame(x), ..., .rows = .rows, .name_repair = .name_repair, rownames = rownames)
}

#' @exportS3Method data.table::as.data.table
as.data.table.lints <- function(x, keep.rownames = FALSE, ...) { # nolint: object_name_linter.
  if (!requireNamespace("data.table", quietly = TRUE)) {
    cli_abort("Package {.pkg data.table} is required to convert lints to a data.table.")
  }
  data.table::setDT(as.data.frame(x), keep.rownames = keep.rownames, ...)
}

#' @export
`[.lints` <- function(x, ...) {
  attrs <- attributes(x)
  x <- unclass(x)
  x <- x[...]
  attributes(x) <- attrs
  x
}

#' @export
summary.lints <- function(object, ...) {
  filenames <- vapply(object, `[[`, character(1L), "filename")
  types <- factor(vapply(object, `[[`, character(1L), "type"),
    levels = c("style", "warning", "error")
  )
  tbl <- table(filenames, types)
  filenames <- rownames(tbl)
  res <- as.data.frame.matrix(tbl, row.names = NULL)
  res$filenames <- filenames %||% character()
  nms <- colnames(res)
  res[order(res$filenames), c("filenames", nms[nms != "filenames"])]
}

#' Check for common mistakes around sorting vectors
#'
#' This linter checks for some common mistakes when using [order()] or [sort()].
#'
#' First, it requires usage of `sort()` over `.[order(.)]`.
#'
#' [sort()] is the dedicated option to sort a list or vector. It is more legible
#' and around twice as fast as `.[order(.)]`, with the gap in performance
#' growing with the vector size.
#'
#' Second, it requires usage of [is.unsorted()] over equivalents using `sort()`.
#'
#' The base function `is.unsorted()` exists to test the sortedness of a vector.
#'   Prefer it to inefficient and less-readable equivalents like
#'   `x != sort(x)`. The same goes for checking `x == sort(x)` -- use
#'   `!is.unsorted(x)` instead.
#'
#' Moreover, use of `x == sort(x)` can be risky because [sort()] drops missing
#'   elements by default, meaning `==` might end up trying to compare vectors
#'   of differing lengths.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x[order(x)]",
#'   linters = sort_linter()
#' )
#'
#' lint(
#'   text = "x[order(x, decreasing = TRUE)]",
#'   linters = sort_linter()
#' )
#'
#' lint(
#'   text = "sort(x) == x",
#'   linters = sort_linter()
#' )
#'
#' lint(
#'   text = "rev(sort(x))",
#'   linters = sort_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x[sample(order(x))]",
#'   linters = sort_linter()
#' )
#'
#' lint(
#'   text = "y[order(x)]",
#'   linters = sort_linter()
#' )
#'
#' lint(
#'   text = "sort(x, decreasing = TRUE) == x",
#'   linters = sort_linter()
#' )
#'
#' lint(
#'   text = "sort(x, decreasing = TRUE)",
#'   linters = sort_linter()
#' )
#'
#' # If you are sorting several objects based on the order of one of them, such
#' # as:
#' x <- sample(1:26)
#' y <- letters
#' newx <- x[order(x)]
#' newy <- y[order(x)]
#' # This will be flagged by the linter. However, in this very specific case,
#' # it would be clearer and more efficient to run order() once and assign it
#' # to an object, rather than mix and match order() and sort()
#' index <- order(x)
#' newx <- x[index]
#' newy <- y[index]
#'
#' @evalRd rd_tags("sort_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
sort_linter <- function() {
  # NB: assumes COMMENTs stripped
  non_keyword_arg <- "expr[position() > 1 and not(preceding-sibling::*[1][self::EQ_SUB])]"
  order_xpath <- glue("
  self::expr[
    expr[1] = expr/{non_keyword_arg}
  ]
    /OP-LEFT-BRACKET
    /following-sibling::expr[1][
      count({non_keyword_arg}) = 1
    ]
  ")

  sorted_xpath <- "
  expr[
    (EQ or NE)
    and expr/expr = expr
    and not(expr/EQ_SUB)
  ]"
  sorted_identical_xpath <- "
  expr[
    expr/SYMBOL_FUNCTION_CALL[text() = 'identical']
    and expr/expr = expr
  ]
  "

  arguments_xpath <-
    ".//SYMBOL_SUB[text() = 'method' or text() = 'decreasing' or text() = 'na.last']"

  arg_values_xpath <- glue("{arguments_xpath}/following-sibling::expr[1]")

  Linter(linter_level = "expression", function(source_expression) {
    order_calls <- source_expression$xml_find_function_calls("order") |>
      xml_find_all_("parent::*/parent::*") |>
      strip_comments_from_subtree()

    order_expr <- xml_find_all_(order_calls, order_xpath)

    variable <- xml_text(xml_find_first_(
      order_expr,
      ".//SYMBOL_FUNCTION_CALL[text() = 'order']/parent::expr[1]/following-sibling::expr[1]"
    ))

    orig_call <- sprintf("%s[%s]", variable, get_r_string(order_expr))

    # Reconstruct new argument call for each expression separately
    arguments <- vapply(order_expr, function(e) {
      arg_names <- xml_text(xml_find_all_(e, arguments_xpath))
      arg_values <- xml_text(xml_find_all_(e, arg_values_xpath))
      if (!"na.last" %in% arg_names) {
        arg_names <- c(arg_names, "na.last")
        arg_values <- c(arg_values, "TRUE")
      }
      paste(arg_names, "=", arg_values, collapse = ", ")
    }, character(1L))

    new_call <- sprintf("sort(%s, %s)", variable, arguments)

    order_lints <- xml_nodes_to_lints(
      order_expr,
      source_expression = source_expression,
      lint_message = paste0(
        new_call, " is better than ", orig_call, ". ",
        "Note that it's always preferable to save the output of order() for the same variable ",
        "as a local variable than to re-compute it."
      ),
      type = "warning"
    )

    sort_calls <- source_expression$xml_find_function_calls("sort") |>
      xml_find_all_("parent::*/parent::*/parent::*") |>
      strip_comments_from_subtree()
    sorted_expr <- xml_find_all_(sort_calls, sorted_xpath)

    sorted_op <- xml_text(xml_find_first_(sorted_expr, "*[2]"))
    lint_message <- ifelse(
      sorted_op == "==",
      "Use !is.unsorted(x) to test the sortedness of a vector.",
      "Use is.unsorted(x) to test the unsortedness of a vector."
    )

    sorted_identical_expr <- xml_find_all_(sort_calls, sorted_identical_xpath)
    is_negated <- !is.na(
      xml_find_first_(sorted_identical_expr, "preceding-sibling::OP-EXCLAMATION")
    )

    lint_message <- c(
      lint_message,
      ifelse(
        is_negated,
        "Use is.unsorted(x) to test the unsortedness of a vector.",
        "Use !is.unsorted(x) to test the sortedness of a vector."
      )
    )

    sorted_lints <- xml_nodes_to_lints(
      combine_nodesets(sorted_expr, sorted_identical_expr),
      source_expression = source_expression,
      lint_message = lint_message,
      type = "warning"
    )

    rev_sort_calls <- xml_find_all_(sort_calls, "expr[1][expr[1]/SYMBOL_FUNCTION_CALL[text() = 'rev']]")
    rev_sort_lints <- xml_nodes_to_lints(
      rev_sort_calls,
      source_expression = source_expression,
      lint_message =
        "Use sort(x, decreasing = TRUE) instead of rev(sort(x)). If present, `na.last` value needs to be flipped",
      type = "warning"
    )

    c(order_lints, sorted_lints, rev_sort_lints)
  })
}

#' `T` and `F` symbol linter
#'
#' Although they can be synonyms, avoid the symbols `T` and `F`, and use `TRUE` and `FALSE`, respectively, instead.
#' `T` and `F` are not reserved keywords and can be assigned to any other values.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x <- T; y <- F",
#'   linters = T_and_F_symbol_linter()
#' )
#'
#' lint(
#'   text = "T = 1.2; F = 2.4",
#'   linters = T_and_F_symbol_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x <- c(TRUE, FALSE)",
#'   linters = T_and_F_symbol_linter()
#' )
#'
#' lint(
#'   text = "t = 1.2; f = 2.4",
#'   linters = T_and_F_symbol_linter()
#' )
#'
#' @evalRd rd_tags("T_and_F_symbol_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#logical-vectors>
#' @export
T_and_F_symbol_linter <- function() { # nolint: object_name.
  symbol_xpath <- "//SYMBOL[
    (text() = 'T' or text() = 'F')
    and not(parent::expr[OP-DOLLAR or OP-AT or following-sibling::OP-LEFT-BRACKET or following-sibling::LBB])
    and (
      not(ancestor::expr[OP-TILDE])
      or parent::expr/preceding-sibling::*[not(self::COMMENT)][1][self::EQ_SUB]
    )
  ]"
  assignment_xpath <-
    "parent::expr[following-sibling::LEFT_ASSIGN or preceding-sibling::RIGHT_ASSIGN or following-sibling::EQ_ASSIGN]"

  usage_xpath <- glue("{symbol_xpath}[not({assignment_xpath})]")
  assignment_xpath <- glue("{symbol_xpath}[{assignment_xpath}]")

  replacement_map <- c(T = "TRUE", F = "FALSE")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_usage <- xml_find_all_(xml, usage_xpath)
    bad_assignment <- xml_find_all_(xml, assignment_xpath)

    make_lints <- function(expr, fmt) {
      symbol <- xml_text(expr)
      lint_message <- sprintf(fmt, replacement_map[symbol], symbol)
      xml_nodes_to_lints(
        xml = expr,
        source_expression = source_expression,
        lint_message = lint_message,
        type = "style",
        column_number_xpath = "number(./@col2 + 1)", # mark at end
        range_end_xpath = "number(./@col2 + 1)" # end after T/F for easy fixing
      )
    }

    c(
      make_lints(bad_usage, "Use %s instead of the symbol %s."),
      make_lints(bad_assignment, "Don't use %2$s as a variable name, as it can break code relying on %2$s being %1$s.")
    )
  })
}

rx_non_active_char <- rex(none_of("^${(.*+?|[\\"))
rx_static_escape <- local({
  rx_char_escape <- rex(or(
    group("\\", none_of(alnum)),
    group("\\x", between(xdigit, 1L, 2L)),
    group("\\", between("0":"7", 1L, 3L)),
    group("\\u{", between(xdigit, 1L, 4L), "}"),
    group("\\u", between(xdigit, 1L, 4L)),
    group("\\U{", between(xdigit, 1L, 8L), "}"),
    group("\\U", between(xdigit, 1L, 8L))
  ))
  rx_trivial_char_group <- rex(
    "[",
    or(
      any,
      group("\\", none_of("dswDSW")), # character classes, e.g. \d are enabled in [] too if perl = TRUE
      rx_char_escape
    ),
    "]"
  )
  rex(or(
    capture(rx_char_escape, name = "char_escape"),
    capture(rx_trivial_char_group, name = "trivial_char_group")
  ))
})

rx_static_token <- local({
  rex(or(
    rx_non_active_char,
    rx_static_escape
  ))
})

rx_unescaped_regex <- paste0("(?s)", rex(start, zero_or_more(rx_non_active_char), end))
rx_static_regex <- paste0("(?s)", rex(start, zero_or_more(rx_static_token), end))
rx_first_static_token <- paste0("(?s)", rex(start, zero_or_more(rx_non_active_char), rx_static_escape))
rx_escapable_tokens <- "^${}().*+?|[]\\<>=:;/_-!@#%&,~"

#' Determine whether a regex pattern actually uses regex patterns
#'
#' Note that is applies to the strings that are found on the XML parse tree,
#'   _not_ plain strings. This is important for backslash escaping, which
#'   happens at different layers of escaping than one might expect. So testing
#'   this function is best done through testing the expected results of a lint
#'   on a given file, rather than passing strings to this function, which can
#'   be confusing.
#'
#' @param str A character vector.
#' @return A logical vector, `TRUE` wherever `str` could be replaced by a
#'   string with `fixed = TRUE`.
#' @noRd
is_not_regex <- function(str, allow_unescaped = FALSE) {
  # need to add single-line option to allow literal newlines
  if (allow_unescaped) {
    !grepl(rx_unescaped_regex, str, perl = TRUE)
  } else {
    grepl(rx_static_regex, str, perl = TRUE)
  }
}

#' Compute a fixed string equivalent to a static regular expression
#'
#' @param static_regex A regex for which `is_not_regex()` returns `TRUE`
#' @return A string such that `grepl(static_regex, x)` is equivalent to
#' `grepl(get_fixed_string(static_regex), x, fixed = TRUE)`
#'
#' @noRd
get_fixed_string <- function(static_regex) {
  if (length(static_regex) == 0L) {
    return(character())
  } else if (length(static_regex) > 1L) {
    return(vapply(static_regex, get_fixed_string, character(1L)))
  }
  fixed_string <- ""
  current_match <- regexpr(rx_first_static_token, static_regex, perl = TRUE)
  while (current_match != -1L) {
    token_type <- attr(current_match, "capture.names")[attr(current_match, "capture.start") > 0L]
    token_start <- max(attr(current_match, "capture.start"))
    if (token_start > 1L) {
      fixed_string <- paste0(fixed_string, substr(static_regex, 1L, token_start - 1L))
    }
    consume_to <- attr(current_match, "match.length")
    token_content <- substr(static_regex, token_start, consume_to)
    fixed_string <- paste0(fixed_string, get_token_replacement(token_content, token_type))
    static_regex <- substr(static_regex, start = consume_to + 1L, stop = nchar(static_regex))
    current_match <- regexpr(rx_first_static_token, static_regex, perl = TRUE)
  }
  paste0(fixed_string, static_regex)
}

#' Get a fixed string equivalent to a regular expression token
#'
#' This handles two cases: converting a "trivial" character group like `[$]` to `$`,
#'   and converting an escaped character like `"\\$"` to `$`. Splitting a full expression
#'   into tokens is handled by `get_fixed_string()`.
#'
#' @noRd
get_token_replacement <- function(token_content, token_type) {
  if (token_type == "trivial_char_group") { # otherwise, char_escape
    token_content <- substr(token_content, start = 2L, stop = nchar(token_content) - 1L)
    if (startsWith(token_content, "\\")) { # escape within trivial char group
      get_token_replacement(token_content, "char_escape")
    } else {
      token_content
    }
  } else if (re_matches(token_content, rex("\\", one_of(rx_escapable_tokens)))) {
    substr(token_content, start = 2L, stop = nchar(token_content))
  } else {
    eval(parse(text = paste0('"', token_content, '"')))
  }
}

# some metadata about infix operators on the R parse tree.
#   xml_tag gives the XML tag as returned by xmlparsedata::xml_parse_data().
#   r_string gives the operator as you would write it in R code.

# styler: off
infix_metadata <- data.frame(matrix(byrow = TRUE, ncol = 2L, c(
  "OP-PLUS",         "+",
  "OP-MINUS",        "-",
  "OP-TILDE",        "~",
  "GT",              ">",
  "GE",              ">=",
  "LT",              "<",
  "LE",              "<=",
  "EQ",              "==",
  "NE",              "!=",
  "AND",              "&",
  "OR",              "|",
  "AND2",            "&&",
  "OR2",             "||",
  "LEFT_ASSIGN",     "<-",
  "LEFT_ASSIGN",     ":=",
  "LEFT_ASSIGN",     "<<-",
  "RIGHT_ASSIGN",    "->",
  "RIGHT_ASSIGN",    "->>",
  "EQ_ASSIGN",       "=",
  "EQ_SUB",          "=",   # in calls: foo(x = 1)
  "EQ_FORMALS",      "=",   # in definitions: function(x = 1)
  "PIPE",            "|>",
  "SPECIAL",         "%%",
  "OP-SLASH",        "/",
  "OP-STAR",         "*",
  "OP-COMMA",        ",",
  "OP-CARET",        "^",
  "OP-CARET",        "**",
  "OP-AT",           "@",
  "OP-EXCLAMATION",  "!",
  "OP-COLON",        ":",
  "NS_GET",          "::",
  "NS_GET_INT",      ":::",
  "OP-LEFT-BRACE",   "{",
  "OP-LEFT-BRACKET", "[",
  "LBB",             "[[",
  "OP-LEFT-PAREN",   "(",
  "OP-QUESTION",     "?",
  "OP-DOLLAR",       "$",
  NULL
)))
# styler: on

names(infix_metadata) <- c("xml_tag", "string_value")
# utils::getParseData()'s designation for the tokens wouldn't be valid as XML tags
infix_metadata$parse_tag <- ifelse(
  startsWith(infix_metadata$xml_tag, "OP-"),
  sQuote(infix_metadata$string_value, "'"),
  infix_metadata$xml_tag
)
# treated separately because spacing rules are different for unary operators
infix_metadata$unary <- infix_metadata$xml_tag %in% c("OP-PLUS", "OP-MINUS", "OP-TILDE", "OP-EXCLAMATION")
# high-precedence operators are ignored by this linter; see
#   https://style.tidyverse.org/syntax.html#infix-operators
infix_metadata$low_precedence <- infix_metadata$string_value %in% c(
  "+", "-", "~", ">", ">=", "<", "<=", "==", "!=", "&", "&&", "|", "||",
  "<-", ":=", "<<-", "->", "->>", "=", "%%", "/", "*", "|>"
)
# comparators come up in several lints
infix_metadata$comparator <- infix_metadata$string_value %in% c("<", "<=", ">", ">=", "==", "!=")

# these XML nodes require checking the text() to disambiguate multiple operators using the same tag
infix_metadata$ambiguous_tag <- infix_metadata$xml_tag %in% infix_metadata$xml_tag[duplicated(infix_metadata$xml_tag)]
infix_metadata$xml_tag_exact <- infix_metadata$xml_tag
infix_metadata$xml_tag_exact[infix_metadata$ambiguous_tag] <- sprintf(
  "%s[text() = '%s']",
  infix_metadata$xml_tag_exact[infix_metadata$ambiguous_tag],
  infix_metadata$string_value[infix_metadata$ambiguous_tag]
)

# functions equivalent to base::ifelse() for linting purposes
ifelse_funs <- c("ifelse", "if_else", "fifelse")

object_name_xpath <- local({
  # search ancestor:: axis for assignments of symbols for
  #   cases like a$b$c. We only try to lint 'a' since 'b'
  #   and 'c' might be beyond the user's control to name.
  #   the tree structure for 'a$b$c <- 1' has 'a'
  #   at the 'bottom' of the <expr> list; it is distinguished
  #   from 'b' and 'c' by not having '$' as a sibling.
  # search parent:: axis for assignments of strings because
  #   the complicated nested assignment available for symbols
  #   is not possible for strings, though we do still have to
  #   be aware of cases like 'a$"b" <- 1'.
  xp_assignment_target_fmt <- "
    not(parent::expr[OP-DOLLAR or OP-AT])
    and %1$s::expr[
      following-sibling::LEFT_ASSIGN%2$s
      or preceding-sibling::RIGHT_ASSIGN
      or following-sibling::EQ_ASSIGN
    ]
    and not(%1$s::expr[
     preceding-sibling::OP-LEFT-BRACKET
     or preceding-sibling::LBB
    ])
  "

  # strings on LHS of := are only checked if they look like data.table usage DT[, "a" := ...]
  dt_walrus_cond <- "[
    text() != ':='
    or parent::expr/preceding-sibling::OP-LEFT-BRACKET
  ]"

  # either an argument supplied positionally, i.e., not like 'arg = val', or the call <expr>
  not_kwarg_cond <- "not(preceding-sibling::*[not(self::COMMENT)][1][self::EQ_SUB])"

  glue(xp_strip_comments("
  //SYMBOL[ {sprintf(xp_assignment_target_fmt, 'ancestor', '')} ]
  |  //STR_CONST[
      ({sprintf(xp_assignment_target_fmt, 'parent', dt_walrus_cond)})
      or parent::expr
        /preceding-sibling::expr[1]
        /SYMBOL_FUNCTION_CALL[text() = 'setGeneric']
      (: x= argument is the first positional argument, if not given as x= :)
      or parent::expr[
        (
          ({not_kwarg_cond})
          and count(preceding-sibling::expr[{not_kwarg_cond}]) = 1
        )
        or preceding-sibling::SYMBOL_SUB[1][text() = 'x']
      ]
        /preceding-sibling::expr[last()]
        /SYMBOL_FUNCTION_CALL[text() = 'assign']
     ]
  |  //SYMBOL_FORMALS
  "))
})

# Remove quotes or other things from names
strip_names <- function(x) {
  x <- re_substitutes(x, rex(start, some_of(quote, "`", "%")), "")
  x <- re_substitutes(x, rex(some_of(quote, "`", "<", "-", "%"), end), "")
  x
}

#' Pull out symbols used in glue strings under the current sub-tree
#'
#' Required by any linter (e.g. [object_usage_linter()] / [unused_import_linter()])
#'   that lints based on whether certain symbols are present, to ensure any
#'   symbols only used inside glue strings are also visible to the linter.
#'
#' @param expr An XML AST
#' @param interpret_glue Logical, if `FALSE` return nothing.
#' @return A character vector of symbols (variables, infix operators, and
#'   function calls) found in glue calls under `expr`.
#' @noRd
extract_glued_symbols <- function(expr, interpret_glue = TRUE) {
  if (!isTRUE(interpret_glue)) {
    return(character())
  }
  # TODO(#2448): support more glue functions
  # NB: position() > 1 because position=1 is <expr><SYMBOL_FUNCTION_CALL>
  glue_call_xpath <- "
    descendant::SYMBOL_FUNCTION_CALL[text() = 'glue']
      /parent::expr
      /parent::expr[
        not(SYMBOL_SUB[text() = '.envir' or text() = '.transform'])
        and not(expr[position() > 1 and not(STR_CONST)])
      ]
  "
  glue_calls <- xml_find_all_(expr, glue_call_xpath)

  glued_symbols <- new.env(parent = emptyenv())
  for (glue_call in glue_calls) {
    parsed_call <- xml2lang(glue_call)
    parsed_call[[".envir"]] <- glued_symbols
    parsed_call[[".transformer"]] <- glue_symbol_extractor
    # #1459: syntax errors in glue'd code are ignored with warning, rather than crashing lint
    tryCatch(eval(parsed_call), error = glue_parse_failure_warning)
  }
  names(glued_symbols)
}

glue_parse_failure_warning <- function(cond) {
  cli_warn(c(
    x = "Evaluating glue expression while testing for local variable usage failed: {conditionMessage(cond)}",
    i = "Please ensure correct glue syntax, e.g., matched delimiters."
  ))
  NULL
}
glue_symbol_extractor <- function(text, envir, data) {
  symbols <- tryCatch(
    all.vars(parse(text = text), functions = TRUE),
    error = \(...) NULL,
    warning = \(...) NULL
  )
  for (sym in symbols) {
    assign(sym, NULL, envir = envir)
  }
  ""
}

magrittr_pipes <- c("%>%", "%!>%", "%T>%", "%$%", "%<>%")

purrr_mappers <- c(
  "map", "walk",
  "map_raw", "map_lgl", "map_int", "map_dbl", "map_chr", "map_vec",
  "map_df", "map_dfr", "map_dfc"
)

# see ?".onLoad", ?Startup, and ?quit.
#   All of .onLoad, .onAttach, and .onUnload are used in base packages,
#   and should be caught in is_base_function; they're included here for completeness / stability
#   (they don't strictly _have_ to be defined in base, so could in principle be removed).
#   .Last.sys and .First.sys are part of base itself, so aren't included here.
special_funs <- c(
  ".onLoad",
  ".onAttach",
  ".onUnload",
  ".onDetach",
  ".Last.lib",
  ".First",
  ".Last"
)

is_special_function <- function(x) {
  x %in% special_funs
}

# content is the file content from readLines
extract_r_source <- function(filename, lines, error = identity) {
  pattern <- get_knitr_pattern(filename, lines)
  if (is.null(pattern$chunk.begin) || is.null(pattern$chunk.end)) {
    return(lines)
  }

  # mask non-source lines by NA, but keep total line count identical so the line number for EOF is correct, see #1400
  output_env <- new.env(parent = emptyenv())
  output_env$output <- rep.int(NA_character_, length(lines))

  chunks <- tryCatch(get_chunk_positions(pattern = pattern, lines = lines), error = error)
  if (is_error(chunks) || is_lint(chunks)) {
    assign("e", chunks, envir = parent.frame())
    # error, so return empty code
    return(output_env$output)
  }

  # no chunks found, so just return the lines
  if (length(chunks[["starts"]]) == 0L || length(chunks[["ends"]]) == 0L) {
    return(output_env$output)
  }

  Map(
    function(start, end, indent) {
      line_seq <- seq(start + 1L, end - 1L)
      chunk_code <- lines[line_seq]
      output_env$output[line_seq] <- if (indent > 0L) substr(chunk_code, indent + 1L, nchar(chunk_code)) else chunk_code
    },
    chunks[["starts"]],
    chunks[["ends"]],
    chunks[["indents"]]
  )
  # drop <<chunk>> references, too
  is.na(output_env$output) <- grep(pattern$ref.chunk, output_env$output)
  replace_prefix(output_env$output, pattern$chunk.code)
}

get_knitr_pattern <- function(filename, lines) {
  # Early return if the source code is parseable as plain R code.
  # Otherwise, R code containing a line which matches any knitr pattern will be treated as a knitr file.
  # See #1406 for details.
  if (parsable(lines)) {
    return(NULL)
  }
  # suppressWarnings for #1920. TODO(michaelchirico): this is a bit sloppy -- we ignore
  #   warnings here because encoding issues are caught later and that code path handles them
  #   correctly by converting to a lint. It would require some refactoring to get that
  #   right here as well, but it would avoid the duplication.
  pattern <- withCallingHandlers(
    ("knitr" %:::% "detect_pattern")(lines, tolower(xfun::file_ext(filename))),
    warning = function(cond) {
      if (!grepl("invalid UTF-8", conditionMessage(cond), fixed = TRUE)) {
        cli_warn(cond) # nocov. No known way to reach here.
      }
      invokeRestart("muffleWarning")
    }
  )
  if (!is.null(pattern)) {
    knitr::all_patterns[[pattern]]
  } else {
    NULL
  }
}

get_chunk_positions <- function(pattern, lines) {
  starts <- filter_chunk_start_positions(
    starts = grep(pattern$chunk.begin, lines, perl = TRUE),
    lines = lines
  )
  ends <- filter_chunk_end_positions(
    starts = starts,
    ends = grep(pattern$chunk.end, lines, perl = TRUE)
  )
  # only keep those blocks that contain at least one line of code
  nonempty_keep <- which(ends - starts > 1L)

  starts <- starts[nonempty_keep]
  ends <- ends[nonempty_keep]

  eval_keep <- vapply(
    seq_along(starts),
    \(ii) is_eval_chunk(starts[ii], ends[ii], lines, pattern),
    logical(1L)
  )
  starts <- starts[eval_keep]
  ends <- ends[eval_keep]

  # Check indent on all lines in the chunk to allow for staggered indentation within a chunk;
  #   set the initial column to the leftmost one within each chunk (including the start+end gates). See tests.
  # use 'ws_re' to make clear that we're matching knitr's definition of initial whitespace.
  ws_re <- sub("```.*", "", pattern$chunk.begin)
  extract_min_chunk_indent <- function(start, end) {
    indents <- attr(regexpr(ws_re, lines[start:end], perl = TRUE), "match.length")
    min(indents)
  }
  # NB: min() guarantees length(indents) == length(starts)
  indents <- unlist(Map(extract_min_chunk_indent, starts, ends))
  list(starts = starts, ends = ends, indents = indents)
}

filter_chunk_start_positions <- function(starts, lines) {
  # keep blocks that don't set a knitr engine (and so contain evaluated R code)
  drop_idx <- defines_knitr_engine(lines[starts])
  starts[!drop_idx]
}

filter_chunk_end_positions <- function(starts, ends) {
  # In a valid file, possibly with plain-code-blocks,
  # - there should be at least as many ends as starts
  # In Rmarkdown and Quarto, unevaluated blocks may open & close with the same ``` pattern
  # that defines the end-pattern for an evaluated block

  # This returns the first end-position that succeeds each start-position
  # starts (1, 3, 5, 7,        11)  --> (1, 3, 5, 7, 11)
  # ends   (2, 4, 6, 8, 9, 10, 12)  --> (2, 4, 6, 8, 12) # return this
  length_difference <- length(ends) - length(starts)
  if (length_difference == 0L && all(ends > starts)) {
    return(ends)
  }

  positions <- sort(c(starts = starts, ends = ends))
  code_start_indexes <- grep("starts", names(positions), fixed = TRUE)

  code_ends <- positions[pmin(1L + code_start_indexes, length(positions))]

  bad_end_indexes <- grep("starts", names(code_ends), fixed = TRUE)
  if (length(bad_end_indexes) > 0L) {
    bad_start_positions <- positions[code_start_indexes[bad_end_indexes]]
    # This error message is formatted like a parse error; don't use {cli}
    stop(sprintf( # nolint: undesirable_function_call_linter.
      "<rmd>:%1$d:1: Missing chunk end for chunk (maybe starting at line %1$d).\n",
      bad_start_positions[1L]
    ), call. = FALSE)
  }

  code_ends
}

defines_knitr_engine <- function(start_lines) {
  # Other packages defining custom engines should have them loaded and thus visible
  #   via knitr_engines$get() below. It seems the simplest way to accomplish this is
  #   for those packages to set some code in their .onLoad() hook, but that's not
  #   always done (nor quite recommended as a "best practice" by knitr).
  #   See the discussion on #1552.
  # TODO(#1617): explore running loadNamespace() automatically.
  engines <- names(knitr::knit_engines$get())

  # {some_engine}, {some_engine label, ...} or {some_engine, ...}
  bare_engine_pattern <- rex(
    "{", or(engines), one_of("}", " ", ",")
  )
  # {... engine = "some_engine" ...}
  explicit_engine_pattern <- rex(
    boundary, "engine", any_spaces, "="
  )

  re_matches(start_lines, explicit_engine_pattern) |
    re_matches(start_lines, bare_engine_pattern)
}

is_eval_chunk <- function(start, end, lines, pattern) {
  header <- lines[start]
  # essentially knitr:::extract_params_src
  params_src <- trimws(gsub(pattern$chunk.begin, "\\1", header))
  header_params <- tryCatch(suppressMessages(xfun::csv_options(params_src)), error = \(e) NULL)

  code <- lines[(start + 1L):(end - 1L)]
  body_params <- tryCatch(
    suppressMessages(suppressWarnings(xfun::divide_chunk("r", code)))$options,
    error = \(e) NULL
  )

  eval_value <- body_params$eval %||% header_params$eval
  # nolint next: T_and_F_symbol_linter.
  if (identical(eval_value, quote(F))) {
    return(FALSE)
  }
  !isFALSE(eval_value)
}

replace_prefix <- function(lines, prefix_pattern) {
  if (is.null(prefix_pattern)) {
    return(lines)
  }

  m <- gregexpr(prefix_pattern, lines)
  non_na <- !is.na(m)

  prefix_lengths <- lapply(regmatches(lines[non_na], m[non_na]), nchar)
  regmatches(lines[non_na], m[non_na]) <- lapply(prefix_lengths, strrep, x = " ")

  lines
}

#' Require consecutive calls to mutate() to be combined when possible
#'
#' `dplyr::mutate()` accepts any number of columns, so sequences like
#'   `DF %>% dplyr::mutate(..1) %>% dplyr::mutate(..2)` are redundant --
#'   they can always be expressed with a single call to `dplyr::mutate()`.
#'
#' An exception is for some SQL back-ends, where the translation logic may not be
#'   as sophisticated as that in the default `dplyr`, for example in
#'   `DF %>% mutate(a = a + 1) %>% mutate(b = a - 2)`.
#'
#' @param invalid_backends Character vector of packages providing dplyr backends
#'   which may not be compatible with combining `mutate()` calls in all cases.
#'   Defaults to `"dbplyr"` since not all SQL backends can handle re-using
#'   a variable defined in the same `mutate()` expression.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x %>% mutate(a = 1) %>% mutate(b = 2)",
#'   linters = consecutive_mutate_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x %>% mutate(a = 1, b = 2)",
#'   linters = consecutive_mutate_linter()
#' )
#'
#' code <- "library(dbplyr)\nx %>% mutate(a = 1) %>% mutate(a = a + 1)"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = consecutive_mutate_linter()
#' )
#'
#' @evalRd rd_tags("consecutive_mutate_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
consecutive_mutate_linter <- function(invalid_backends = "dbplyr") {
  attach_pkg_xpath <- "
  following-sibling::expr
    /*[self::SYMBOL or self::STR_CONST]
  "

  namespace_xpath <- glue("
  //SYMBOL_PACKAGE[{ xp_text_in_table(invalid_backends) }]
  |
  //COMMENT[
    contains(text(), '@import')
    and (
      {xp_or(sprintf(\"contains(text(), '%s')\", invalid_backends))}
    )
  ]
  ")

  # match on the expr, not the SYMBOL_FUNCTION_CALL, to ensure
  #   namespace-qualified calls only match if the namespaces do.
  # expr[2] needed in expr[1][expr[2]] to skip matches on pipelines
  #   starting like mutate(DF, ...) %>% foo() %>% mutate().
  # similarly, expr[1][expr[call='mutate']] covers pipelines
  #   starting like mutate(DF, ...) %>% mutate(...)
  mutate_cond <- xp_and(
    "expr/SYMBOL_FUNCTION_CALL[text() = 'mutate']",
    "not(SYMBOL_SUB[text() = '.keep' or text() = '.by'])"
  )
  xpath <- glue("
  (//PIPE | //SPECIAL[{ xp_text_in_table(magrittr_pipes) }])
    /preceding-sibling::expr[expr[2][{ mutate_cond }] or ({ mutate_cond })]
    /following-sibling::expr[{ mutate_cond }]
  ")

  Linter(linter_level = "file", function(source_expression) {
    # need the full file to also catch usages at the top level
    xml <- source_expression$full_xml_parsed_content

    attach_str <- get_r_string(xml_find_all_(
      source_expression$xml_find_function_calls(c("library", "require")),
      attach_pkg_xpath
    ))
    if (any(invalid_backends %in% attach_str)) {
      return(list())
    }

    namespace_expr <- xml_find_first_(xml, namespace_xpath)
    if (!is.na(namespace_expr)) {
      return(list())
    }

    bad_expr <- xml_find_all_(xml, xpath)

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = "Unify consecutive calls to mutate().",
      type = "warning"
    )
  })
}

#' Duplicate argument linter
#'
#' Check for duplicate arguments in function calls. Some cases are run-time errors
#' (e.g. `mean(x = 1:5, x = 2:3)`), otherwise this linter is used to discourage
#'  explicitly providing duplicate names to objects (e.g. `c(a = 1, a = 2)`).
#'  Duplicate-named objects are hard to work with programmatically and
#'  should typically be avoided.
#'
#' @param except A character vector of function names as exceptions. Defaults to
#'   functions that allow sequential updates to variables, currently `dplyr::mutate()`
#'   and `dplyr::transmute()`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "list(x = 1, x = 2)",
#'   linters = duplicate_argument_linter()
#' )
#'
#' lint(
#'   text = "fun(arg = 1, arg = 2)",
#'   linters = duplicate_argument_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "list(x = 1, x = 2)",
#'   linters = duplicate_argument_linter(except = "list")
#' )
#'
#' lint(
#'   text = "df %>% dplyr::mutate(x = a + b, x = x + d)",
#'   linters = duplicate_argument_linter()
#' )
#'
#' @evalRd rd_tags("duplicate_argument_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
duplicate_argument_linter <- function(except = c("mutate", "transmute")) {
  # NB: approach checking for duplicates in XPath is hard because of
  #   quoted names, e.g. foo(a = 1, `a` = 2), so compute duplicates in R
  xpath_call_with_args <- glue("
    //EQ_SUB[not(
      preceding-sibling::expr/SYMBOL_FUNCTION_CALL[{ xp_text_in_table(except) }]
    )]
      /parent::expr[count(EQ_SUB) > 1]
  ")
  xpath_arg_name <- "./EQ_SUB/preceding-sibling::*[not(self::COMMENT)][1]"

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    call_expr <- xml_find_all_(xml, xpath_call_with_args)

    bad_expr <- lapply(
      call_expr,
      function(expr) {
        arg_expr <- xml_find_all_(expr, xpath_arg_name)
        arg_expr[duplicated(get_r_string(arg_expr))]
      }
    )

    xml_nodes_to_lints(
      unlist(bad_expr, recursive = FALSE),
      source_expression = source_expression,
      lint_message = "Avoid duplicate arguments in function calls.",
      type = "warning"
    )
  })
}

#' Build the `xml_find_function_calls()` helper for a source expression
#'
#' @param xml The XML parse tree as an XML object (`xml_parsed_content` or `full_xml_parsed_content`)
#'
#' @return A fast function to query the common XPath expression
#'   `xml_find_all_(xml, glue::glue("//SYMBOL_FUNCTION_CALL[text() = '{function_names[1]}' or ...]/parent::expr"))`,
#'   or, using the internal function `xp_text_in_table()`,
#'   `xml_find_all_(xml, glue::glue("//SYMBOL_FUNCTION_CALL[{ xp_text_in_table(function_names) }]/parent::expr"))`,
#'   i.e., the `parent::expr` of the `SYMBOL_FUNCTION_CALL` node corresponding to given function names.
#'
#' @noRd
build_xml_find_function_calls <- function(xml) {
  function_call_cache <- xml_find_all_(xml, "//SYMBOL_FUNCTION_CALL/parent::*")
  names(function_call_cache) <- get_r_string(function_call_cache, "SYMBOL_FUNCTION_CALL")

  s4_slot_cache <- xml_find_all_(xml, "//SLOT/parent::expr[following-sibling::OP-LEFT-PAREN]")
  names(s4_slot_cache) <- get_r_string(s4_slot_cache, "SLOT")

  function(function_names, keep_names = FALSE, include_s4_slots = FALSE) {
    if (is.null(function_names)) {
      if (include_s4_slots) {
        res <- combine_nodesets(function_call_cache, s4_slot_cache)
      } else {
        res <- function_call_cache
      }
    } else {
      include_function_idx <- names(function_call_cache) %in% function_names
      if (include_s4_slots) {
        res <- combine_nodesets(
          function_call_cache[include_function_idx],
          s4_slot_cache[names(s4_slot_cache) %in% function_names]
        )
      } else {
        res <- function_call_cache[include_function_idx]
      }
    }
    if (keep_names) res else unname(res)
  }
}

#' Block unreachable code and comments following return statements
#'
#' Code after e.g. a [return()] or [stop()]
#'   or in deterministically false conditional loops like `if (FALSE)` can't be reached;
#'   typically this is vestigial code left after refactoring or sandboxing code, which
#'   is fine for exploration, but shouldn't ultimately be checked in. Comments
#'   meant for posterity should be placed *before* the final `return()`.
#'
#' @param allow_comment_regex Character vector of regular expressions which identify
#'   comments to exclude when finding unreachable terminal comments. By default, this
#'   includes the default "skip region" end marker for `{covr}`
#'   (option "covr.exclude_end", or `"# nocov end"` if unset).
#'   The end marker for `{lintr}` (`settings$exclude_end`) is always included.
#'   Note that the regexes should include the initial comment character `#`.
#'
#' @examples
#' # will produce lints
#' code_lines <- "f <- function() {\n  return(1 + 1)\n  2 + 2\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unreachable_code_linter()
#' )
#'
#' code_lines <- "if (FALSE) {\n 2 + 2\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unreachable_code_linter()
#' )
#'
#' code_lines <- "while (FALSE) {\n 2 + 2\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unreachable_code_linter()
#' )
#'
#' code_lines <- "f <- function() {\n  return(1)\n  # end skip\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unreachable_code_linter()
#' )
#'
#' # okay
#' code_lines <- "f <- function() {\n  return(1 + 1)\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unreachable_code_linter()
#' )
#'
#' code_lines <- "if (foo) {\n 2 + 2\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unreachable_code_linter()
#' )
#'
#' code_lines <- "while (foo) {\n 2 + 2\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unreachable_code_linter()
#' )
#'
#' code_lines <- "f <- function() {\n  return(1)\n  # end skip\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = unreachable_code_linter(allow_comment_regex = "# end skip")
#' )
#'
#' @evalRd rd_tags("unreachable_code_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
unreachable_code_linter <- function(allow_comment_regex = getOption("covr.exclude_end", "# nocov end")) {
  expr_after_control <- "
    (//REPEAT | //ELSE | //FOR)/following-sibling::expr[1]
    | (//IF | //WHILE)/following-sibling::expr[2]
  "

  # normal case: expression after terminal call is on the next line
  unreachable_expr_cond_ws <- "
  following-sibling::*[
    not(self::OP-RIGHT-BRACE or self::OP-SEMICOLON or self::ELSE or preceding-sibling::ELSE)
    and (not(self::COMMENT) or @line2 > preceding-sibling::*[not(self::COMMENT)][1]/@line2)
  ][1]"
  # robustness case: expression after terminal call is on the same line (after a semicolon),
  #   wherefore the condition is a bit different due to <exprlist> nodes
  unreachable_expr_cond_sc <- "
  parent::exprlist[OP-SEMICOLON]
    /following-sibling::*[
      not(self::OP-RIGHT-BRACE)
      and (not(self::COMMENT) or @line1 > preceding-sibling::exprlist/expr/@line2)
    ][1]
  "

  terminal_fun_expr <-
    "(//FUNCTION | //OP-LAMBDA)/following-sibling::expr[OP-LEFT-BRACE][last()]"

  # NB: use not(OP-DOLLAR) to prevent matching process$stop(), #1051
  terminal_call_cond <- "expr[1][
    not(OP-DOLLAR or OP-AT)
    and SYMBOL_FUNCTION_CALL[text() = 'return' or text() = 'stop']
  ]"

  xpath_after_terminal_node <- glue("
    ({expr_after_control} | {terminal_fun_expr})//expr[{terminal_call_cond}]/{unreachable_expr_cond_ws}
    | ({expr_after_control} | {terminal_fun_expr})//expr[{terminal_call_cond}]/{unreachable_expr_cond_sc}
    | ({expr_after_control})//expr[NEXT or BREAK]/{unreachable_expr_cond_ws}
    | ({expr_after_control})//expr[NEXT or BREAK]/{unreachable_expr_cond_sc}
  ")

  xpath_terminal_node <- "
    (preceding-sibling::exprlist//expr | preceding-sibling::expr)
      /*[
        self::expr/SYMBOL_FUNCTION_CALL[text() = 'return' or text() = 'stop']
        or self::NEXT
        or self::BREAK
      ]
  "

  xpath_if_while <- "
    (//IF | //WHILE)[following-sibling::expr[1]/NUM_CONST[text() = 'FALSE']]
      /parent::expr
  "

  xpath_else <- "
  //IF[following-sibling::expr[1]/NUM_CONST[text() = 'TRUE']]
    /following-sibling::ELSE/following-sibling::expr[1]
  "

  handle_inline_conditions <- function(expr) {
    expr <- lapply(
      expr,
      function(x) {
        if (xml_name_(xml_child_(x)) == "OP-LEFT-BRACE") {
          xml_find_first_(x, "expr")
        } else {
          x
        }
      }
    )
    expr[vapply(expr, xml2::xml_length, integer(1L)) != 0L]
  }

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    # run here because 'settings$exclude_end' may not be set correctly at "compile time".
    # also build with '|', not rex::rex(or(.)), the latter which will double-escape the regex.
    allow_comment_regex <- paste(union(allow_comment_regex, settings$exclude_end), collapse = "|")

    expr_after_terminal_node <- xml_find_all_(xml, xpath_after_terminal_node)

    is_valid_comment <- xml_name_(expr_after_terminal_node) == "COMMENT" &
      re_matches_logical(xml_text(expr_after_terminal_node), allow_comment_regex)

    expr_after_terminal_node <- expr_after_terminal_node[!is_valid_comment]
    terminal_node <- xml_text(xml_find_first_(expr_after_terminal_node, xpath_terminal_node))
    terminal_node <-
      ifelse(terminal_node %in% c("return", "stop"), paste0(terminal_node, "()"), paste0("`", terminal_node, "`"))

    lints_after_terminal_node <- xml_nodes_to_lints(
      expr_after_terminal_node,
      source_expression = source_expression,
      lint_message = sprintf("Remove code and comments coming after %s.", terminal_node),
      type = "warning"
    )

    expr_if_while <- handle_inline_conditions(xml_find_all_(xml, xpath_if_while))

    lints_if_while <- xml_nodes_to_lints(
      expr_if_while,
      source_expression = source_expression,
      lint_message = "Remove code inside a conditional loop with a deterministically false condition.",
      type = "warning"
    )

    expr_else <- handle_inline_conditions(xml_find_all_(xml, xpath_else))

    lints_else <- xml_nodes_to_lints(
      expr_else,
      source_expression = source_expression,
      lint_message = "Remove code inside an else block after a deterministically true condition.",
      type = "warning"
    )

    c(lints_after_terminal_node, lints_if_while, lints_else)
  })
}

#' Block usage of `==`, `!=` on logical vectors
#'
#' Testing `x == TRUE` is redundant if `x` is a logical vector. Wherever this is
#'   used to improve readability, the solution should instead be to improve the
#'   naming of the object to better indicate that its contents are logical. This
#'   can be done using prefixes (is, has, can, etc.). For example, `is_child`,
#'   `has_parent_supervision`, `can_watch_horror_movie` clarify their logical
#'   nature, while `child`, `parent_supervision`, `watch_horror_movie` don't.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "if (any(x == TRUE)) 1",
#'   linters = redundant_equals_linter()
#' )
#'
#' lint(
#'   text = "if (any(x != FALSE)) 0",
#'   linters = redundant_equals_linter()
#' )
#'
#' lint(
#'   text = "dt[is_tall == FALSE, y]",
#'   linters = redundant_equals_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "if (any(x)) 1",
#'   linters = redundant_equals_linter()
#' )
#'
#' lint(
#'   text = "if (!all(x)) 0",
#'   linters = redundant_equals_linter()
#' )
#'
#' # in `{data.table}` semantics, `dt[x]` is a join, `dt[(x)]` is a subset
#' lint(
#'   text = "dt[(!is_tall), y]",
#'   linters = redundant_equals_linter()
#' )
#'
#' @evalRd rd_tags("redundant_equals_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - [outer_negation_linter()]
#' @export
redundant_equals_linter <- function() {
  xpath <- "
  (//EQ | //NE)
    /parent::expr
    /expr[NUM_CONST[text() = 'TRUE' or text() = 'FALSE']]
    /parent::expr
  "

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)
    op <- xml_text(xml_find_first_(bad_expr, "*[not(self::COMMENT)][2]"))

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste(
        "Using", op, "on a logical vector is redundant.",
        "Well-named logical vectors can be used directly in filtering.",
        "For data.table's `i` argument, wrap the column name in (), like `DT[(is_treatment)]`."
      ),
      type = "warning"
    )
  })
}

#' Block assignment of `{}`
#'
#' Assignment of `{}` is the same as assignment of `NULL`; use the latter
#'   for clarity. Closely related: [unnecessary_concatenation_linter()].
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x <- {}",
#'   linters = empty_assignment_linter()
#' )
#'
#' writeLines("x = {\n}")
#' lint(
#'   text = "x = {\n}",
#'   linters = empty_assignment_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x <- { 3 + 4 }",
#'   linters = empty_assignment_linter()
#' )
#'
#' lint(
#'   text = "x <- NULL",
#'   linters = empty_assignment_linter()
#' )
#'
#' @evalRd rd_tags("empty_assignment_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
empty_assignment_linter <- make_linter_from_xpath(
  # for some reason, the parent in the `=` case is <equal_assign>, not <expr>, hence parent::expr
  xpath = "
  //OP-LEFT-BRACE[following-sibling::*[not(self::COMMENT)][1][self::OP-RIGHT-BRACE]]
    /parent::expr[
      preceding-sibling::LEFT_ASSIGN
      or preceding-sibling::EQ_ASSIGN
      or following-sibling::RIGHT_ASSIGN
    ]
    /parent::*
  ",
  lint_message = "Assign NULL explicitly or, whenever possible, allocate the empty object with the right type and size."
)

#' Cyclomatic complexity linter
#'
#' Check for overly complicated expressions. See `cyclocomp()` function from `{cyclocomp}`.
#'
#' @param complexity_limit Maximum cyclomatic complexity, default `15`. Expressions more complex
#' than this are linted.
#'
#' @examplesIf requireNamespace("cyclocomp", quietly = TRUE)
#' # will produce lints
#' lint(
#'   text = "if (TRUE) 1 else 2",
#'   linters = cyclocomp_linter(complexity_limit = 1L)
#' )
#'
#' # okay
#' lint(
#'   text = "if (TRUE) 1 else 2",
#'   linters = cyclocomp_linter(complexity_limit = 2L)
#' )
#'
#' @evalRd rd_tags("cyclocomp_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
cyclocomp_linter <- function(complexity_limit = 15L) {
  # nocov start
  if (!requireNamespace("cyclocomp", quietly = TRUE)) {
    cli_warn(c(
      "Cyclocomp complexity is computed using {.fn cyclocomp::cyclocomp}.",
      i = "Please install the needed {.pkg cyclocomp} package."
    ))
    return(Linter(\(.) cli_abort("cyclocomp_linter is disabled due to lack of the {.pkg cyclocomp} package")))
  }
  # nocov end
  Linter(linter_level = "expression", function(source_expression) {
    complexity <- try_silently(
      cyclocomp::cyclocomp(parse(text = source_expression$content, keep.source = FALSE))
    )
    if (inherits(complexity, "try-error") || complexity <= complexity_limit) {
      return(list())
    }
    col1 <- source_expression[["column"]][1L]
    Lint(
      filename = source_expression[["filename"]],
      line_number = source_expression[["line"]][1L],
      column_number = source_expression[["column"]][1L],
      type = "style",
      message = sprintf(
        "Reduce the cyclomatic complexity of this expression from %d to at most %d. %s",
        complexity, complexity_limit,
        "Consider replacing high-complexity sections like loops and branches with helper functions."
      ),
      ranges = list(rep(col1, 2L)),
      line = source_expression$lines[1L]
    )
  })
}

#' Deprecated functions
#'
#' Functions that have been deprecated and replaced by newer ones. They will be removed in an
#' upcoming version of \pkg{lintr} and should thus not be used anymore.
#' @noRd
NULL

lintr_deprecated <- function(what, alternative = NULL, version = NULL,
                             type = "Function", signal = c("warning", "stop")) {
  signal <- match.arg(signal)
  signal <- match.fun(signal)
  msg <- c(
    c(type, " ", what, " was deprecated"),
    if (length(version) > 0L) c(" in lintr version ", version),
    ". ",
    if (length(alternative) > 0L) c("Use ", alternative, " instead.")
  )
  msg <- paste(msg, collapse = "")
  signal(msg, call. = FALSE, domain = NA)
}

#' Identify unregistered native routines
#'
#' It is preferable to register routines for efficiency and safety.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = '.Call("cpp_routine", PACKAGE = "mypkg")',
#'   linters = routine_registration_linter()
#' )
#'
#' lint(
#'   text = '.Fortran("f_routine", PACKAGE = "mypkg")',
#'   linters = routine_registration_linter()
#' )
#'
#' # okay
#' lint(
#'   text = ".Call(cpp_routine)",
#'   linters = routine_registration_linter()
#' )
#'
#' lint(
#'   text = ".Fortran(f_routine)",
#'   linters = routine_registration_linter()
#' )
#'
#' @evalRd rd_tags("routine_registration_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Registering-native-routines>
#'
#' @export
routine_registration_linter <- make_linter_from_function_xpath(
  function_names = c(".C", ".Call", ".Fortran", ".External"),
  xpath = "
  following-sibling::expr[1]/STR_CONST
    /parent::expr
  ",
  lint_message = "Register your native code routines with useDynLib and R_registerRoutines()."
)

#' Get parsed IDs by token
#'
#' Gets the source IDs (row indices) corresponding to given token.
#'
#' @param source_expression A list of source expressions, the result of a call to [get_source_expressions()],
#'   for the desired filename.
#' @param value Character. String corresponding to the token to search for.
#' For example:
#'
#'  * "SYMBOL"
#'  * "FUNCTION"
#'  * "EQ_FORMALS"
#'  * "$"
#'  * "("
#'
#' @param fun For additional flexibility, a function to search for in
#' the `token` column of `parsed_content`. Typically `==` or `%in%`.
#'
#' @examples
#' tmp <- tempfile()
#' writeLines(c("x <- 1", "y <- x + 1"), tmp)
#' source_exprs <- get_source_expressions(tmp)
#' ids_with_token(source_exprs$expressions[[1L]], value = "SYMBOL")
#' with_id(source_exprs$expressions[[1L]], 2L)
#' unlink(tmp)
#'
#' @return `ids_with_token`: The indices of the `parsed_content` data frame
#' entry of the list of source expressions. Indices correspond to the
#' *rows* where `fun` evaluates to `TRUE` for the `value` in the *token* column.
#' @export
ids_with_token <- function(source_expression, value, fun = `==`) {
  if (!is_lint_level(source_expression, "expression")) {
    return(integer())
  }
  loc <- which(fun(source_expression$parsed_content$token, value))
  if (loc %==% integer()) {
    return(integer())
  }
  loc
}

#' Require usage of `fixed=TRUE` in regular expressions where appropriate
#'
#' Invoking a regular expression engine is overkill for cases when the search
#'   pattern only involves static patterns.
#'
#' NB: for `stringr` functions, that means wrapping the pattern in `stringr::fixed()`.
#'
#' NB: this linter is likely not able to distinguish every possible case when
#'   a fixed regular expression is preferable, rather it seeks to identify
#'   likely cases. It should _never_ report false positives, however; please
#'   report false positives as an error.
#'
#' @param allow_unescaped Logical, default `FALSE`. If `TRUE`, only patterns that
#'   require regex escapes (e.g. `"\\$"` or `"[$]"`) will be linted. See examples.
#' @examples
#' # will produce lints
#' code_lines <- 'gsub("\\\\.", "", x)'
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = fixed_regex_linter()
#' )
#'
#' lint(
#'   text = 'grepl("a[*]b", x)',
#'   linters = fixed_regex_linter()
#' )
#'
#' lint(
#'   text = 'grepl("a[*]b", x)',
#'   linters = fixed_regex_linter(allow_unescaped = TRUE)
#' )
#'
#' code_lines <- 'stringr::str_subset(x, "\\\\$")'
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = fixed_regex_linter()
#' )
#'
#' lint(
#'   text = 'grepl("Munich", address)',
#'   linters = fixed_regex_linter()
#' )
#'
#' # okay
#' code_lines <- 'gsub("\\\\.", "", x, fixed = TRUE)'
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = fixed_regex_linter()
#' )
#'
#' lint(
#'   text = 'grepl("a*b", x, fixed = TRUE)',
#'   linters = fixed_regex_linter()
#' )
#'
#' lint(
#'   text = 'stringr::str_subset(x, stringr::fixed("$"))',
#'   linters = fixed_regex_linter()
#' )
#'
#' lint(
#'   text = 'grepl("Munich", address, fixed = TRUE)',
#'   linters = fixed_regex_linter()
#' )
#'
#' lint(
#'   text = 'grepl("Munich", address)',
#'   linters = fixed_regex_linter(allow_unescaped = TRUE)
#' )
#'
#' @evalRd rd_tags("fixed_regex_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
fixed_regex_linter <- function(allow_unescaped = FALSE) {
  # regular expression pattern is the first argument
  pos_1_regex_funs <- c(
    "grep", "grepv", "gsub", "sub", "regexec", "grepl", "regexpr", "gregexpr"
  )

  # regular expression pattern is the second argument
  pos_2_regex_funs <- c(
    # base functions.
    "strsplit",
    # data.table functions.
    "tstrsplit",
    # stringr functions.
    #   even though the user action is different
    #   (setting fixed=TRUE vs. wrapping stringr::fixed()),
    #   detection of the lint is the same
    "str_count", "str_detect", "str_ends", "str_extract", "str_extract_all",
    "str_locate", "str_locate_all", "str_match", "str_match_all",
    "str_remove", "str_remove_all", "str_replace", "str_replace_all",
    "str_split", "str_starts", "str_subset",
    "str_view", "str_view_all", "str_which"
  )

  pipes <- setdiff(magrittr_pipes, c("%$%", "%T>%"))
  in_pipe_cond <- glue("
    parent::expr/preceding-sibling::SPECIAL[{ xp_text_in_table(pipes) }]
    | parent::expr/preceding-sibling::PIPE
  ")

  # NB: strsplit doesn't have an ignore.case argument
  # NB: we intentionally exclude cases like gsub(x, c("a" = "b")), where "b" is fixed
  pos_1_xpath <- glue("
  self::*[
    not(following-sibling::SYMBOL_SUB[
      (text() = 'fixed' or text() = 'ignore.case')
      and following-sibling::expr[1][NUM_CONST[text() = 'TRUE'] or SYMBOL[text() = 'T']]
    ])
  ]
    /following-sibling::expr[
      (
        position() = 1
        and STR_CONST
        and not(EQ_SUB)
        and not({ in_pipe_cond })
      ) or (
        STR_CONST
        and preceding-sibling::*[not(self::COMMENT)][2][self::SYMBOL_SUB/text() = 'pattern']
      )
    ]
  ")
  pos_2_xpath <- glue("
  self::*[
    not(following-sibling::SYMBOL_SUB[
      text() = 'fixed'
      and following-sibling::expr[1][NUM_CONST[text() = 'TRUE'] or SYMBOL[text() = 'T']]
    ])
  ]
    /following-sibling::expr[
      position() = 2 - count({ in_pipe_cond })
      and STR_CONST
      and not(EQ_SUB)
    ]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    pos_1_calls <- source_expression$xml_find_function_calls(pos_1_regex_funs)
    pos_2_calls <- source_expression$xml_find_function_calls(pos_2_regex_funs)
    patterns <- combine_nodesets(
      xml_find_all_(pos_1_calls, pos_1_xpath),
      xml_find_all_(pos_2_calls, pos_2_xpath)
    )
    pattern_strings <- get_r_string(patterns)

    is_static <- is_not_regex(pattern_strings, allow_unescaped)
    patterns <- patterns[is_static]
    pattern_strings <- pattern_strings[is_static]

    fixed_equivalent <- encodeString(get_fixed_string(pattern_strings), quote = '"', justify = "none")
    call_name <- xml_find_chr_(patterns, "string(preceding-sibling::expr[last()]/SYMBOL_FUNCTION_CALL)")

    is_stringr <- startsWith(call_name, "str_")
    replacement_suggestion <- ifelse(
      is_stringr,
      sprintf("stringr::fixed(%s) as the pattern", fixed_equivalent),
      sprintf("%s with fixed = TRUE", fixed_equivalent)
    )
    msg <- paste(
      "Use", replacement_suggestion, "here. This regular expression is static, i.e.,",
      "its matches can be expressed as a fixed substring expression, which is faster to compute."
    )

    xml_nodes_to_lints(
      patterns,
      source_expression = source_expression,
      lint_message = msg,
      type = "warning"
    )
  })
}

#' Pipe continuation linter
#'
#' Check that each step in a pipeline is on a new line, or the entire pipe fits on one line.
#'
#' @evalRd rd_tags("pipe_continuation_linter")
#'
#' @examples
#' # will produce lints
#' code_lines <- "1:3 %>%\n mean() %>% as.character()"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = pipe_continuation_linter()
#' )
#'
#' code_lines <- "1:3 |> mean() |>\n as.character()"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = pipe_continuation_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "1:3 %>% mean() %>% as.character()",
#'   linters = pipe_continuation_linter()
#' )
#'
#' code_lines <- "1:3 %>%\n mean() %>%\n as.character()"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = pipe_continuation_linter()
#' )
#'
#' lint(
#'   text = "1:3 |> mean() |> as.character()",
#'   linters = pipe_continuation_linter()
#' )
#'
#' code_lines <- "1:3 |>\n mean() |>\n as.character()"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = pipe_continuation_linter()
#' )
#'
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/pipes.html#long-lines-2>
#' @export
pipe_continuation_linter <- function() {
  # Where a single-line pipeline is nested inside a larger expression
  #   e.g. inside a function definition), the outer expression can span multiple lines
  #   without throwing a lint.
  pipe_node <- glue("SPECIAL[{ xp_text_in_table(magrittr_pipes) }]")
  preceding_pipe <- glue("preceding-sibling::expr[1]/descendant::*[self::{pipe_node} or self::PIPE]")
  xpath <- glue("
  (//PIPE | //{pipe_node})[
    parent::expr[@line1 < @line2]
    and {preceding_pipe}
    and (
      preceding-sibling::expr[1]/descendant-or-self::expr/@line2
      = following-sibling::expr[1]/descendant-or-self::expr/@line1
      or @line1 = {preceding_pipe}/@line1
    )
  ]
  ")

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    pipe_exprs <- xml_find_all_(xml, xpath)
    pipe_text <- xml_text(pipe_exprs)

    xml_nodes_to_lints(
      pipe_exprs,
      source_expression = source_expression,
      lint_message = sprintf(
        "Put a space before `%s` and a new line after it, unless the full pipeline fits on one line.",
        pipe_text
      ),
      type = "style"
    )
  })
}

# nocov start
addin_lint <- function() {
  if (!requireNamespace("rstudioapi", quietly = TRUE)) {
    cli_abort("{.pkg rstudioapi} is required for add-ins.")
  }
  filename <- rstudioapi::getSourceEditorContext()
  if (filename$path == "") {
    cli_warn("Current source has no path. Please save before continuing.")
    return(flatten_lints(list()))
  }

  lint(filename$path)
}

addin_lint_package <- function() {
  if (!requireNamespace("rstudioapi", quietly = TRUE)) {
    cli_abort("{.pkg rstudioapi} is required for add-ins.")
  }
  project <- rstudioapi::getActiveProject()
  if (is.null(project)) {
    cli_inform("No project found, passing current directory.")
    project_path <- getwd()
  } else {
    project_path <- project
  }

  lint_package(project_path)
}
# nocov end

#' Block usage of `ifelse()` where `pmin()` or `pmax()` is more appropriate
#'
#' `ifelse(x > M, M, x)` is the same as `pmin(x, M)`, but harder
#'   to read and requires several passes over the vector.
#'
#' The same goes for other similar ways to censor a vector, e.g.
#'   `ifelse(x <= M, x, M)` is `pmin(x, M)`,
#'   `ifelse(x < m, m, x)` is `pmax(x, m)`, and
#'   `ifelse(x >= m, x, m)` is `pmax(x, m)`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "ifelse(5:1 < pi, 5:1, pi)",
#'   linters = ifelse_censor_linter()
#' )
#'
#' lint(
#'   text = "ifelse(x > 0, x, 0)",
#'   linters = ifelse_censor_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "pmin(5:1, pi)",
#'   linters = ifelse_censor_linter()
#' )
#'
#' lint(
#'   text = "pmax(x, 0)",
#'   linters = ifelse_censor_linter()
#' )
#'
#' @evalRd rd_tags("ifelse_censor_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
ifelse_censor_linter <- function() {
  xpath <- glue("
  self::*[expr[
    (LT or GT or LE or GE)
    and expr[1] = following-sibling::expr
    and expr[2] = following-sibling::expr
  ]]")

  Linter(linter_level = "expression", function(source_expression) {
    # nolint next: undesirable_function_name_linter.
    ifelse_calls <- xml_parent(source_expression$xml_find_function_calls(ifelse_funs))
    ifelse_calls <- strip_comments_from_subtree(ifelse_calls)
    bad_expr <- xml_find_all_(ifelse_calls, xpath)

    matched_call <- xp_call_name(bad_expr)
    operator <- xml_find_chr_(bad_expr, "string(expr[2]/*[not(self::COMMENT)][2])")
    match_first <- !is.na(xml_find_first_(bad_expr, "expr[2][expr[1] = following-sibling::expr[1]]"))
    optimizer <- ifelse((operator %in% c("<", "<=")) == match_first, "pmin", "pmax")
    first_var <- rep_len("x", length(match_first))
    second_var <- rep_len("y", length(match_first))
    first_var[!match_first] <- "y"
    second_var[!match_first] <- "x"

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = sprintf(
        "%s(x, y) is preferable to %s(x %s y, %s, %s).",
        optimizer, matched_call, operator, first_var, second_var
      ),
      type = "warning"
    )
  })
}

#' Lint a file, directory, or package
#'
#' * `lint()` lints a single file.
#' * `lint_dir()` lints all files in a directory.
#' * `lint_package()` lints all likely locations for R files in a package, i.e.
#'   `R/`, `tests/`, `inst/`, `vignettes/`, `data-raw/`, `demo/`, and `exec/`.
#'
#' Read `vignette("lintr")` to learn how to configure which linters are run
#' by default.
#' Note that if files contain unparseable encoding problems, only the encoding problem will be linted to avoid
#' unintelligible error messages from other linters.
#'
#' @param filename Either the filename for a file to lint, or a character string of inline R code for linting.
#'   The latter (inline data) applies whenever `filename` has a newline character (\\n).
#' @param linters A named list of linter functions to apply. See [linters] for a full list of default and available
#'   linters.
#' @param ... Provide additional arguments to be passed to:
#'   - [exclude()] (in case of `lint()`; e.g. `lints` or `exclusions`)
#'   - [lint()] (in case of `lint_dir()` and `lint_package()`; e.g. `linters` or `cache`)
#' @param cache When logical, toggle caching of lint results. If passed a character string, store the cache in this
#'   directory.
#' @param parse_settings Logical. Whether to try and parse the [settings][read_settings]. Otherwise, the
#'   [default_settings()] are used. `TRUE` by default when linting files, as opposed to `text=`.
#' @param text Optional argument for supplying a string or lines directly, e.g. if the file is already in memory or
#'   linting is being done ad hoc.
#'
#' @return An object of class `c("lints", "list")`, each element of which is a `"list"` object.
#'
#' @examples
#' # linting inline-code
#' lint("a = 123\n")
#' lint(text = "a = 123")
#'
#' # linting a file
#' f <- tempfile()
#' writeLines("a=1", f)
#' lint(f)
#' unlink(f)
#'
#' @export
lint <- function(filename, linters = NULL, ..., cache = FALSE, parse_settings = !inline_data, text = NULL) {
  check_dots(...names(), c("exclude", "parse_exclusions"))

  needs_tempfile <- missing(filename) || re_matches(filename, rex(newline))
  inline_data <- !is.null(text) || needs_tempfile

  if (parse_settings) {
    read_settings(filename)
    on.exit(reset_settings(), add = TRUE)
  }

  lines <- get_lines(filename, text)

  if (needs_tempfile) {
    filename <- tempfile()
    con <- file(filename, open = "w", encoding = settings$encoding)
    on.exit(unlink(filename), add = TRUE)
    writeLines(text = lines, con = con, sep = "\n")
    close(con)
  }

  filename <- normalize_path(filename, mustWork = !inline_data) # to ensure a unique file in cache
  source_expressions <- get_source_expressions(filename, lines)

  linters <- define_linters(linters)
  linters <- Map(validate_linter_object, linters, names(linters))

  cache_path <- define_cache_path(cache)

  lint_cache <- load_cache(filename, cache_path)
  lint_obj <- define_cache_key(filename, inline_data, lines)
  lints <- retrieve_file(lint_cache, lint_obj, linters)
  if (!is.null(lints)) {
    return(exclude(lints, lines = lines, linter_names = names(linters), ...))
  }

  lints <- lint_impl_(linters, lint_cache, filename, source_expressions)

  lints <- lints |>
    maybe_append_condition_lints(source_expressions, lint_cache, filename) |>
    flatten_lints() |>
    reorder_lints()
  class(lints) <- c("lints", "list")

  cache_file(lint_cache, filename, linters, lints)
  save_cache(lint_cache, filename, cache_path)

  res <- exclude(lints, lines = lines, linter_names = names(linters), ...)

  # simplify filename if inline
  zap_temp_filename(res, needs_tempfile)
}

lint_impl_ <- function(linters, lint_cache, filename, source_expressions) {
  if (is_tainted(source_expressions$lines)) {
    return(list())
  }

  file_linter_names <- names(linters)[vapply(linters, is_linter_level, logical(1L), "file")]
  expression_linter_names <- names(linters)[vapply(linters, is_linter_level, logical(1L), "expression")]

  lints <- list()
  for (expr in source_expressions$expressions) {
    for (linter in necessary_linters(expr, expression_linter_names, file_linter_names)) {
      # use withCallingHandlers for friendlier failures on unexpected linter errors
      lints[[length(lints) + 1L]] <- withCallingHandlers(
        get_lints(expr, linter, linters[[linter]], lint_cache, source_expressions$lines),
        error = function(cond) {
          cli_abort(
            "Linter {.fn linter} failed in {.file {filename}}:",
            parent = cond
          )
        }
      )
    }
  }

  lints
}

#' @param path For the base directory of the project (for `lint_dir()`) or
#'   package (for `lint_package()`).
#' @param relative_path if `TRUE`, file paths are printed using their path relative to the base directory.
#'   If `FALSE`, use the full absolute path.
#' @param exclusions exclusions for [exclude()], relative to the package path.
#' @param pattern pattern for files, by default it will take files with any of the extensions
#' .R, .Rmd, .qmd, .Rnw, .Rhtml, .Rrst, .Rtex, .Rtxt allowing for lowercase r (.r, ...).
#' @param show_progress Logical controlling whether to show linting progress with
#'   [cli::cli_progress_along()]. The default behavior is to show progress in [interactive()] sessions
#'   not running a testthat suite.
#'
#' @examples
#' if (FALSE) {
#'   lint_dir()
#'
#'   lint_dir(
#'     linters = list(semicolon_linter()),
#'     exclusions = list(
#'       "inst/doc/creating_linters.R" = 1,
#'       "inst/example/bad.R",
#'       "renv"
#'     )
#'   )
#' }
#' @export
#' @rdname lint
lint_dir <- function(path = ".", ...,
                     relative_path = TRUE,
                     exclusions = list("renv", "packrat"),
                     # TODO(r-lib/rex#85): Re-write in case-sensitive rex()
                     pattern = "(?i)[.](r|rmd|qmd|rnw|rhtml|rrst|rtex|rtxt)$",
                     parse_settings = TRUE,
                     show_progress = NULL) {
  check_dots(...names(), c("lint", "exclude", "parse_exclusions"))

  if (isTRUE(parse_settings)) {
    read_settings(path)
    on.exit(reset_settings(), add = TRUE)

    exclusions <- c(exclusions, settings$exclusions)
  }

  if (is.null(show_progress)) show_progress <- interactive() && !identical(Sys.getenv("TESTTHAT"), "true")

  exclusions <- normalize_exclusions(
    exclusions,
    root = path,
    pattern = pattern
  )

  # normalize_path ensures names(exclusions) and files have the same names for the same files.
  # It also ensures all paths have forward slash
  # Otherwise on windows, files might incorrectly not be excluded in to_exclude
  files <- normalize_path(dir(
    path,
    pattern = pattern,
    recursive = TRUE,
    full.names = TRUE
  ))

  # Remove fully ignored files to avoid reading & parsing
  files <- drop_excluded(files, exclusions)

  if (length(files) == 0L) {
    lints <- list()
    class(lints) <- "lints"
    return(lints)
  }

  if (isTRUE(show_progress)) {
    # nocov start. Only displays after >=2 seconds;
    #   we get ample testing interactively, so don't slow down tests just for coverage.
    lints <- lapply(
      # NB: This cli API is experimental (https://github.com/r-lib/cli/issues/709)
      cli::cli_progress_along(files, name = "Running linters"),
      function(idx) {
        lint(files[idx], ..., parse_settings = FALSE, exclusions = exclusions)
      }
    )
    # nocov end
  } else {
    lints <- lapply(
      files,
      function(file) { # nolint: unnecessary_lambda_linter.
        lint(file, ..., parse_settings = FALSE, exclusions = exclusions)
      }
    )
  }

  lints <- flatten_lints(lints)
  lints <- reorder_lints(lints)

  if (relative_path) {
    path <- normalize_path(path, mustWork = FALSE)
    lints[] <- lapply(
      lints,
      function(x) {
        x$filename <- re_substitutes(x$filename, rex(path, one_of("/", "\\")), "")
        x
      }
    )
    attr(lints, "path") <- path
  }

  class(lints) <- "lints"

  lints
}

drop_excluded <- function(files, exclusions) {
  to_exclude <- vapply(
    files,
    \(file) file %in% names(exclusions) && is_excluded_file(exclusions[[file]]),
    logical(1L)
  )
  files[!to_exclude]
}

#' @examples
#' if (FALSE) {
#'   lint_package()
#'
#'   lint_package(
#'     linters = linters_with_defaults(semicolon_linter = semicolon_linter()),
#'     exclusions = list("inst/doc/creating_linters.R" = 1, "inst/example/bad.R")
#'   )
#' }
#' @export
#' @rdname lint
lint_package <- function(path = ".", ...,
                         relative_path = TRUE,
                         exclusions = list("R/RcppExports.R"),
                         parse_settings = TRUE,
                         show_progress = NULL) {
  if (length(path) > 1L) {
    cli_abort(c(
      x = "Only linting one package at a time is supported.",
      i = "Instead, {.val {length(path)}} package paths were provided."
    ))
  }
  pkg_path <- find_package(path)

  if (is.null(pkg_path)) {
    cli_warn(c(
      i = "Didn't find any R package searching upwards from {.file {normalize_path(path)}}"
    ))
    return(NULL)
  }

  if (parse_settings) {
    read_settings(pkg_path)
    on.exit(reset_settings(), add = TRUE)
  }

  exclusions <- normalize_exclusions(
    c(exclusions, settings$exclusions),
    root = pkg_path
  )

  r_directories <- file.path(pkg_path, c("R", "tests", "inst", "vignettes", "data-raw", "demo", "exec"))
  lints <- lint_dir(r_directories,
    relative_path = FALSE,
    exclusions = exclusions,
    parse_settings = FALSE,
    show_progress = show_progress,
    ...
  )

  if (isTRUE(relative_path)) {
    path <- normalize_path(pkg_path, mustWork = FALSE)
    lints[] <- lapply(
      lints,
      function(x) {
        x$filename <- re_substitutes(x$filename, rex(path, one_of("/", "\\")), "")
        x
      }
    )
    attr(lints, "path") <- path
  }

  lints
}

#' Run a linter on a source expression, optionally using a cache
#'
#' @param expr A source expression.
#' @param linter Name of the linter.
#' @param linter_fun Closure of the linter.
#' @param lint_cache Cache environment, or `NULL` if caching is disabled.
#'
#' @return A list of lints generated by the linter on `expr`.
#'
#' @noRd
get_lints <- function(expr, linter, linter_fun, lint_cache, lines) {
  expr_lints <- NULL
  if (has_lint(lint_cache, expr, linter)) {
    # retrieve_lint() might return NULL if missing line number is encountered.
    # It could be caused by nolint comments.
    expr_lints <- retrieve_lint(lint_cache, expr, linter, lines)
  }

  if (is.null(expr_lints)) {
    expr_lints <- flatten_lints(linter_fun(expr))

    for (i in seq_along(expr_lints)) {
      expr_lints[[i]]$linter <- linter
    }

    cache_lint(lint_cache, expr, linter, expr_lints)
  }
  expr_lints
}

define_linters <- function(linters = NULL) {
  if (is.null(linters)) {
    linters <- settings$linters
    names(linters) <- auto_names(linters)
  } else if (is_linter(linters)) {
    linters <- list(linters)
    names(linters) <- attr(linters[[1L]], "name", exact = TRUE)
  } else if (is.list(linters)) {
    names(linters) <- auto_names(linters)
  } else {
    name <- deparse(substitute(linters))
    linters <- list(linters)
    names(linters) <- name
  }
  linters
}

validate_linter_object <- function(linter, name) {
  if (is_linter(linter)) {
    return(linter)
  }
  cli_abort(c(
    i = "Expected {.fn {name}} to be a function of class {.cls linter}.",
    x = "Instead, it is {.obj_type_friendly {linter}}."
  ))
}

reorder_lints <- function(lints) {
  files <- vapply(lints, `[[`, character(1L), "filename")
  lines <- vapply(lints, `[[`, integer(1L), "line_number")
  columns <- vapply(lints, `[[`, integer(1L), "column_number")
  lints[order(
    files,
    lines,
    columns
  )]
}


#' Create a `lint` object
#' @param filename path to the source file that was linted.
#' @param line_number line number where the lint occurred.
#' @param column_number column number where the lint occurred.
#' @param type type of lint.
#' @param message message used to describe the lint error
#' @param line code source where the lint occurred
#' @param ranges a list of ranges on the line that should be emphasized.
#' @return an object of class `c("lint", "list")`.
#' @name lint-s3
#' @export
Lint <- function(filename, line_number = 1L, column_number = 1L, # nolint: object_name.
                 type = c("style", "warning", "error"),
                 message = "", line = "", ranges = NULL) {
  validate_lint_object(message, line, line_number, column_number, ranges)

  type <- match.arg(type)

  obj <- list(
    filename = filename,
    line_number = as.integer(line_number),
    column_number = as.integer(column_number),
    type = type,
    message = message,
    line = line,
    ranges = ranges,
    linter = NA_character_
  )
  class(obj) <- c("lint", "list")
  obj
}

# nolint next: cyclocomp_linter. Sequence of checks + early returns.
validate_lint_object <- function(message, line, line_number, column_number, ranges) {
  if (length(message) != 1L || !is.character(message)) {
    cli_abort("{.arg message} must be a character string")
  }
  if (is.object(message)) {
    cli_warn(c(
      "{.arg message} must be a simple string, but has class {.cls {class(message)}}",
      i = "This will be an error in the next lintr release"
    ))
    message <- unclass(message)
  }
  if (length(line) != 1L || !is.character(line)) {
    cli_abort("{.arg line} must be a character string.")
  }
  max_col <- max(nchar(line) + 1L, 1L, na.rm = TRUE)
  if (!is_number(column_number) || column_number < 0L || column_number > max_col) {
    cli_abort("
      {.arg column_number} must be an integer between {.val {0}} and {.val {max_col}} ({.code nchar(line) + 1}),
      not {.obj_type_friendly {column_number}}.
    ")
  }
  if (!is_number(line_number) || line_number < 1L) {
    cli_abort("{.arg line_number} must be a positive integer, not {.obj_type_friendly {line_number}}.")
  }
  check_ranges(ranges, max_col)
  invisible()
}

is_number <- function(number, n = 1L) {
  length(number) == n && is.numeric(number) && !anyNA(number)
}

is_valid_range <- function(range, max_col) {
  0L <= range[[1L]] &&
    range[[1L]] <= range[[2L]] &&
    range[[2L]] <= max_col
}

check_ranges <- function(ranges, max_col, call = parent.frame()) {
  if (is.null(ranges)) {
    return()
  }
  if (!is.list(ranges)) {
    cli_abort(
      "{.arg ranges} must be {.code NULL} or a list, not {.obj_type_friendly {ranges}}.",
      call = call
    )
  }

  for (range in ranges) {
    if (!is_number(range, 2L)) {
      cli_abort(
        "{.arg ranges} must only contain integer vectors of length 2 without {.code NA}s.",
        call = call
      )
    } else if (!is_valid_range(range, max_col)) {
      cli_abort(
        "{.arg ranges} must satisfy {.val {0}} <= range[1L] <= range[2L] <= {.val {max_col}} (nchar(line) + 1).",
        call = call
      )
    }
  }
}

rstudio_source_markers <- function(lints) {
  if (!requireNamespace("rstudioapi", quietly = TRUE)) {
    cli_abort("{.pkg rstudioapi} is required for {.fn rstudio_source_markers}.") # nocov
  }

  # package path will be NULL unless it is a relative path
  package_path <- attr(lints, "path")

  # generate the markers
  markers <- lapply(lints, function(x) {
    filename <- if (!is.null(package_path)) {
      file.path(package_path, x$filename)
    } else {
      x$filename
    }

    marker <- list()
    marker$type <- x$type
    marker$file <- filename
    marker$line <- x$line_number
    marker$column <- x$column_number
    marker$message <- paste0("[", x$linter, "] ", x$message)
    marker
  })

  # request source markers
  out <- rstudioapi::callFun(
    "sourceMarkers",
    name = "lintr",
    markers = markers,
    basePath = package_path,
    autoSelect = "first"
  )

  # workaround to avoid focusing an empty Markers pane
  # when possible, better solution is to delete the "lintr" source marker list
  # https://github.com/rstudio/rstudioapi/issues/209
  if (length(lints) == 0L) {
    Sys.sleep(0.1)
    rstudioapi::executeCommand("activateConsole")
  }

  out
}

#' Checkstyle Report for lint results
#'
#' Generate a report of the linting results using the [Checkstyle](https://checkstyle.sourceforge.io) XML format.
#'
#' @param lints the linting results.
#' @param filename the name of the output report
#' @export
checkstyle_output <- function(lints, filename = "lintr_results.xml") {
  # package path will be NULL unless it is a relative path
  package_path <- attr(lints, "path")

  # setup file
  d <- xml2::xml_new_document()
  n <- xml2::xml_add_child(d, "checkstyle", version = paste0("lintr-", utils::packageVersion("lintr")))

  # output the style markers to the file
  lapply(split(lints, names(lints)), function(lints_per_file) {
    filename <- if (!is.null(package_path)) {
      file.path(package_path, lints_per_file[[1L]]$filename)
    } else {
      lints_per_file[[1L]]$filename
    }
    f <- xml2::xml_add_child(n, "file", name = filename)

    lapply(lints_per_file, function(x) {
      xml2::xml_add_child(
        f, "error",
        line = as.character(x$line_number),
        column = as.character(x$column_number),
        severity = switch(x$type,
          style = "info",
          x$type
        ),
        message = x$message
      )
    })
  })

  xml2::write_xml(d, filename)
}

#' SARIF Report for lint results
#'
#' Generate a report of the linting results using the [SARIF](https://sarifweb.azurewebsites.net/) format.
#'
#' @param lints the linting results.
#' @param filename the name of the output report
#' @export
sarif_output <- function(lints, filename = "lintr_results.sarif") {
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    cli_abort("{.pkg jsonlite} is required to produce SARIF reports. Please install to continue.") # nocov
  }

  # package path will be `NULL` unless it is a relative path
  package_path <- attr(lints, "path")

  if (is.null(package_path)) {
    cli_abort("Package path needs to be a relative path.")
  }

  # setup template
  sarif <- jsonlite::fromJSON(
    system.file("extdata", "sarif-template.json", package = "lintr"),
    simplifyVector = TRUE,
    simplifyDataFrame = FALSE,
    simplifyMatrix = FALSE
  )

  # assign values
  sarif$runs[[1L]]$results <- NULL
  sarif$runs[[1L]]$tool$driver$rules <- NULL
  sarif$runs[[1L]]$tool$driver$version <- as.character(utils::packageVersion("lintr"))
  sarif$runs[[1L]]$originalUriBaseIds$ROOTPATH$uri <- ""
  rule_index_exists <- FALSE
  root_path_uri <- gsub("\\", "/", package_path, fixed = TRUE)

  if (startsWith(root_path_uri, "/")) {
    root_path_uri <- paste0("file://", root_path_uri)
  } else {
    root_path_uri <- paste0("file:///", root_path_uri) # nocov
  }

  if (!endsWith(root_path_uri, "/")) {
    root_path_uri <- paste0(root_path_uri, "/")
  }

  sarif$runs[[1L]]$originalUriBaseIds$ROOTPATH$uri <- root_path_uri

  # loop and assign result values
  for (lint in lints) {
    one_result <- list()

    if (is.null(sarif$runs[[1L]]$tool$driver$rules)) {
      rule_index_exists <- 0L
    } else {
      rule_index_exists <-
        which(vapply(
          sarif$runs[[1L]]$tool$driver$rules,
          \(x) x$id == lint$linter,
          logical(1L)
        ))
      if (length(rule_index_exists) == 0L || is.na(rule_index_exists[1L])) {
        rule_index_exists <- 0L
      }
    }

    if (rule_index_exists == 0L) {
      new_rule <- list(
        id = lint$linter,
        fullDescription = list(text = lint$message),
        defaultConfiguration = list(
          level = switch(lint$type, style = "note", lint$type)
        )
      )
      sarif$runs[[1L]]$tool$driver$rules <- append(sarif$runs[[1L]]$tool$driver$rules, list(new_rule))
      rule_index <- length(sarif$runs[[1L]]$tool$driver$rules) - 1L
    } else {
      rule_index <- rule_index_exists - 1L
    }

    one_result <- append(one_result, c(ruleId = lint$linter))
    one_result <- append(one_result, c(ruleIndex = rule_index))
    one_result <- append(one_result, list(message = list(text = lint$message)))
    one_location <- list(physicalLocation = list(
      artifactLocation = list(
        uri = gsub("\\", "/", lint$filename, fixed = TRUE),
        uriBaseId = "ROOTPATH"
      ),
      region = list(
        startLine = lint$line_number,
        startColumn = lint$column_number,
        snippet = list(text = lint$line)
      )
    ))
    one_result <- append(one_result, c(locations = list(list(one_location))))
    sarif$runs[[1L]]$results <- append(sarif$runs[[1L]]$results, list(one_result))
  }

  # if lints is empty, add empty results list
  if (length(lints) == 0L) {
    sarif$runs[[1L]]$results <- list()
  }

  write(jsonlite::toJSON(sarif, pretty = TRUE, auto_unbox = TRUE), filename)
}

#' GitLab Report for lint results
#'
#' Generate a report of the linting results using the
#' [GitLab](https://docs.gitlab.com/ci/testing/code_quality/#code-quality-report-format) format.
#'
#' @details
#' lintr only supports three severity types ("style", "warning", and "error")
#' while the GitLab format supports five ("info", "minor", "major", "critical",
#' and "blocker"). The types "style", "warning", and "error" are mapped to
#' the GitLab types "info", "major", and "blocker", respectively. The GitLab
#' types "minor" and "critical" are ignored.
#'
#' @param lints The linting results
#' @param filename The file name of the output report
#' @export
gitlab_output <- function(lints, filename = "lintr_results.json") {
  if (!inherits(lints, "lints")) {
    cli_abort("{.arg lints} must be a {.cls lints} object, not {.obj_type_friendly {lints}}.")
  }
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    cli_abort("{.pkg jsonlite} is required to produce Gitlab reports. Please install to continue.") # nocov
  }

  # Format
  # (copied from https://docs.gitlab.com/ci/testing/code_quality/#code-quality-report-format)
  # ==============================
  # description          String   A human-readable description of the code quality violation.
  # check_name           String   A unique name representing the check, or rule, associated with this violation.
  # fingerprint          String   A unique fingerprint to identify this specific code quality violation, such as a hash
  #                               of its contents.
  # location.path        String   The file containing the code quality violation, expressed as a relative path in the
  #                               repository. Do not prefix with ./.
  # location.lines.begin
  # -or-                 Integer  The line on which the code quality violation occurred.
  # location.positions.begin.line
  #
  # severity             String   The severity of the violation, can be one of info, minor, major, critical, or blocker.

  # Gitlab format as R data structure
  res <-
    lapply(
      lints,
      function(lint) {
        with(
          lint,
          list(
            description = message,
            check_name = linter,
            fingerprint =  digest::digest(line, algo = "sha1"),
            location = list(
              path = filename,
              lines = list(
                begin = line_number
              )
            ),
            severity = switch(type,
              style = "info",
              error = "blocker",
              warning = "major",
              "info"
            )
          )
        )
      }
    )

  write(jsonlite::toJSON(res, pretty = TRUE, auto_unbox = TRUE), filename)
}

highlight_string <- function(message, column_number = NULL, ranges = NULL) {
  maximum <- max(column_number, unlist(ranges))

  line <- fill_with(" ", maximum)

  for (range in ranges) {
    substr(line, range[1L], range[2L]) <- fill_with("~", range[2L] - range[1L] + 1L)
  }

  substr(line, column_number, column_number + 1L) <- "^"

  line
}

fill_with <- function(character = " ", length = 1L) {
  paste(collapse = "", rep.int(character, length))
}

maybe_append_condition_lints <- function(lints, source_expression, lint_cache, filename) {
  error <- source_expression$error
  if (is_lint(error)) {
    error$linter <- "error"
    lints[[length(lints) + 1L]] <- error

    if (!is.null(lint_cache)) {
      cache_lint(lint_cache, list(filename = filename, content = ""), "error", error)
    }
  }
  for (l in source_expression$warning) {
    l$linter <- "parser_warning_linter"
    lints[[length(lints) + 1L]] <- l

    if (!is.null(lint_cache)) {
      cache_lint(lint_cache, list(filename = filename, content = ""), "parser_warning_linter", l)
    }
  }
  lints
}

get_lines <- function(filename, text) {
  encoding <- NULL
  if (!is.null(text)) {
    lines <- strsplit(paste(text, collapse = "\n"), "\n", fixed = TRUE)[[1L]]
  } else if (re_matches(filename, rex(newline))) {
    lines <- strsplit(gsub("\n$", "", filename), "\n", fixed = TRUE)[[1L]]
  } else {
    lines <- read_lines(filename)
    encoding <- settings$encoding
  }

  ensure_utf8(lines, encoding)
}

zap_temp_filename <- function(res, needs_tempfile) {
  if (needs_tempfile) {
    for (i in seq_along(res)) {
      res[[i]][["filename"]] <- "<text>"
    }
  }
  res
}

necessary_linters <- function(expr, expression_linter_names, file_linter_names) {
  if (is_lint_level(expr, "expression")) {
    expression_linter_names
  } else {
    file_linter_names
  }
}

test_that("read_config_file() warns if the config file does not end in a newline", {
  .lintr <- withr::local_tempfile()
  withr::local_options(lintr.linter_file = .lintr)
  withr::local_dir(withr::local_tempdir())

  # cat() not writeLines() to ensure no trailing \n
  cat("linters: linters_with_defaults(brace_linter = NULL)", file = .lintr)
  writeLines("a <- 1", "aaa.R")

  # on R 4.7.0+, read.dcf() uses warn=FALSE in internal readLines(), so we
  #   no longer catch a warning here as part of the fix for old #160.
  if (inherits(tryCatch(read.dcf(.lintr, all = TRUE), condition = identity), "warning")) {
    expect_warning(lint_dir(), "Warning encountered while loading config", fixed = TRUE)
  } else {
    expect_silent(lint_dir())
  }
})

test_that("it gives informative errors if the config file contains errors", {
  .lintr <- withr::local_tempfile(lines = c(
    "linters: linters_with_defaults(",
    "    brace_linter = NULL,",
    "  )"
  ))
  withr::local_options(lintr.linter_file = .lintr)
  withr::local_dir(withr::local_tempdir())

  writeLines("a <- 1", "aaa.R")
  expect_error(lint_dir(), "Error from config setting `linters`", fixed = TRUE)
})

# fixing #774
test_that("linters_with_defaults doesn't break on very long input", {
  expect_named(
    linters_with_defaults(
      defaults = list(),
      lintr::undesirable_function_linter(c(
        detach = paste(
          "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
          "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
          "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
          "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
          "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
          "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
          "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
        )
      ))
    ),
    "undesirable_function_linter"
  )
})

test_that("validate_config_file() detects improperly-formed settings", {
  .lintr <- withr::local_tempfile()
  withr::local_options(lintr.linter_file = .lintr)
  withr::local_dir(withr::local_tempdir())

  writeLines("asdf: 1", .lintr)
  expect_warning(lint_dir(), "Found unused settings in config", fixed = TRUE)

  writeLines("a=1", "aaa.R")
  writeLines(c('exclusions: list("aaa.R")', "asdf: 1"), .lintr)
  expect_warning(lint_dir(), "Found unused settings in config", fixed = TRUE)

  encoding_error_msg <- "Setting `encoding` should be a character string"

  writeLines("encoding: FALSE", .lintr)
  expect_error(lint_dir(), encoding_error_msg, fixed = TRUE)

  writeLines("encoding: NA_character_", .lintr)
  expect_error(lint_dir(), encoding_error_msg, fixed = TRUE)

  writeLines('encoding: c("a", "b")', .lintr)
  expect_error(lint_dir(), encoding_error_msg, fixed = TRUE)

  exclude_error_msg <- "Setting `exclude` should be a single regular expression"

  writeLines("exclude: FALSE", .lintr)
  expect_error(lint_dir(), exclude_error_msg, fixed = TRUE)

  writeLines(c('exclusions: list("aaa.R")', "exclude: FALSE"), .lintr)
  expect_error(lint_dir(), exclude_error_msg, fixed = TRUE)

  writeLines('exclude: "("', .lintr)
  expect_error(lint_dir(), exclude_error_msg, fixed = TRUE)

  writeLines("linters: list(1)", .lintr)
  expect_error(lint_dir(), "Setting `linters` should be a list of linters", fixed = TRUE)

  writeLines("linters: list(assignment_linter(), 1)", .lintr)
  expect_error(lint_dir(), "Setting `linters` should be a list of linters", fixed = TRUE)

  name_exclusion_error_msg <- "Unnamed entries of setting `exclusions` should be strings"

  writeLines("exclusions: list(1L)", .lintr)
  expect_error(lint_dir(), name_exclusion_error_msg, fixed = TRUE)

  writeLines('exclusions: list("aaa.R", 1L)', .lintr)
  expect_error(lint_dir(), name_exclusion_error_msg, fixed = TRUE)

  writeLines("exclusions: list(letters)", .lintr)
  expect_error(lint_dir(), name_exclusion_error_msg, fixed = TRUE)

  writeLines("exclusions: list(NA_character_)", .lintr)
  expect_error(lint_dir(), name_exclusion_error_msg, fixed = TRUE)

  exclusion_error_msg <- "Named entries of setting `exclusions` should designate line numbers"

  writeLines('exclusions: list(aaa.R = "abc")', .lintr)
  expect_error(lint_dir(), exclusion_error_msg, fixed = TRUE)

  writeLines("exclusions: list(aaa.R = NA_integer_)", .lintr)
  expect_error(lint_dir(), exclusion_error_msg, fixed = TRUE)

  writeLines('exclusions: list(aaa.R = list("abc"))', .lintr)
  expect_error(lint_dir(), exclusion_error_msg, fixed = TRUE)

  writeLines("exclusions: list(aaa.R = list(NA_integer_))", .lintr)
  expect_error(lint_dir(), exclusion_error_msg, fixed = TRUE)

  writeLines('exclusions: list(aaa.R = list(assignment_linter = "abc"))', .lintr)
  expect_error(lint_dir(), exclusion_error_msg, fixed = TRUE)

  writeLines("exclusions: list(aaa.R = list(assignment_linter = NA_integer_))", .lintr)
  expect_error(lint_dir(), exclusion_error_msg, fixed = TRUE)
})

test_that("exclusions can be a character vector", {
  withr::local_dir(withr::local_tempdir())
  # exclusions are relative to dirname(.lintr), so must create it here
  .lintr <- withr::local_tempfile(tmpdir = getwd())
  withr::local_options(lintr.linter_file = .lintr)

  writeLines('exclusions: "aaa.R"', .lintr)
  writeLines("a<-1", "aaa.R")
  writeLines("b<-1", "bbb.R")
  expect_length(lint_dir(linters = infix_spaces_linter()), 1L)

  writeLines('exclusions: c("aaa.R", "bbb.R")', .lintr)
  expect_length(lint_dir(linters = infix_spaces_linter()), 0L)
})

test_that("lines Inf means 'all lines'", {
  withr::local_dir(withr::local_tempdir())
  # exclusions are relative to dirname(.lintr), so must create it here
  .lintr <- withr::local_tempfile(tmpdir = getwd())
  withr::local_options(lintr.linter_file = .lintr)

  writeLines("exclusions: list(aaa.R = Inf)", .lintr)
  writeLines("a<-1", "aaa.R")
  expect_length(lint_dir(linters = infix_spaces_linter()), 0L)

  writeLines("exclusions: list(aaa.R = list(infix_spaces_linter = Inf))", .lintr)
  # exclude infix_spaces_linter, include assignment_linter()
  writeLines("a=1", "aaa.R")
  expect_length(lint_dir(linters = list(assignment_linter(), infix_spaces_linter())), 1L)
})

test_that("read_config_file() bubbles up warnings helpfully, without erroring (#2253)", {
  .lintr <- withr::local_tempfile(lines = 'linters: list(backport_linter("2.0.0"))')
  withr::local_options(lintr.linter_file = .lintr)
  withr::local_dir(withr::local_tempdir())

  writeLines("a <- 1", "aaa.R")
  expect_warning(
    lint_dir(),
    'Depending on an R version older than "3.0.0" is not recommended'
  )
})

test_that("perl-only regular expressions are accepted in config", {
  .lintr <- withr::local_tempfile(lines = 'exclude: "# (?<=a)b"')
  withr::local_options(lintr.linter_file = .lintr)
  withr::local_dir(withr::local_tempdir())

  writeLines("a <- 1", "aaa.R")
  expect_silent(lint("aaa.R"))
})

test_that("settings can be put in a sub-directory", {
  withr::local_dir(withr::local_tempdir())

  dir.create(".settings")
  .lintr <- ".settings/.lintr.R"
  writeLines("linters <- list(line_length_linter(10L))", .lintr)

  dir.create("R")
  writeLines("abcdefghijklmnopqrstuvwxyz=1", "R/a.R")

  writeLines(c("Package: foo", "Version: 0.1"), "DESCRIPTION")

  withr::local_options(lintr.linter_file = .lintr)
  expect_length(lint_package(), 1L)
})

test_that("`sarif_output` produces expected error", {
  skip_if_not_installed("jsonlite")

  l <- lint(text = "x = 1", linters = assignment_linter())
  expect_error(sarif_output(l), "Package path needs to be a relative path", fixed = TRUE)
})

test_that("`sarif_output` writes expected files", {
  skip_if_not_installed("jsonlite")

  l <- lint_package(
    test_path("dummy_packages", "missing_dep"),
    linters = object_length_linter(),
    parse_settings = FALSE
  )

  withr::with_tempdir({
    sarif_output(l)
    expect_true(file.exists("lintr_results.sarif"))
  })

  withr::with_tempdir({
    sarif_output(l, filename = "myfile.sarif")
    expect_true(file.exists("myfile.sarif"))
  })
})

test_that("`sarif_output` produces valid files", {
  skip_if_not_installed("jsonlite")

  l <- lint_package(
    test_path("dummy_packages", "clean"),
    linters = default_linters,
    parse_settings = FALSE
  )

  withr::with_tempdir({
    sarif <- sarif_output(l)
    sarif <- jsonlite::fromJSON(
      "lintr_results.sarif",
      simplifyVector = TRUE,
      simplifyDataFrame = FALSE,
      simplifyMatrix = FALSE
    )

    expect_false(is.null(sarif$runs))
    expect_false(is.null(sarif$runs[[1L]]$results))
  })
})

test_that("as.data.frame.lints", {
  l1 <- Lint(
    "dummy.R",
    line_number = 1L,
    type = "style",
    message = "",
    line = ""
  )

  # A minimum lint
  expect_s3_class(l1, "lint")
  expect_type(l1, "list")

  # A larger lint
  l2 <- Lint(
    "dummy.R",
    line_number = 2L,
    column_number = 6L,
    type = "error",
    message = "Under no circumstances is the use of foobar allowed.",
    line = "a <- 1",
    ranges = list(c(1L, 2L), c(6L, 7L))
  )
  expect_s3_class(l2, "lint")

  expect_error(
    Lint("dummy.R", linter = "deprecated"),
    regexp = "deprecated",
    fixed = TRUE
  )

  # Convert lints to data.frame
  lints <- list(l1, l2)
  class(lints) <- "lints"
  df <- as.data.frame(lints)
  expect_s3_class(df, "data.frame")

  exp <- data.frame(
    filename = "dummy.R",
    line_number = c(1L, 2L),
    column_number = c(1L, 6L),
    type = c("style", "error"),
    message = c("", "Under no circumstances is the use of foobar allowed."),
    line = c("", "a <- 1"),
    linter = NA_character_ # These are assigned in lint() now.
  )

  expect_identical(df, exp)
})

test_that("summary.lints() works (no lints)", {
  no_lints <- lint("x <- 1\n", linters = assignment_linter())
  no_lint_summary <- summary(no_lints)
  expect_s3_class(no_lint_summary, "data.frame")
  expect_identical(nrow(no_lint_summary), 0L)
})

test_that("summary.lints() works (lints found)", {
  has_lints <- lint("x = 1\n", linters = assignment_linter())
  has_lint_summary <- summary(has_lints)

  expect_s3_class(has_lint_summary, "data.frame")
  expect_identical(nrow(has_lint_summary), 1L)
  expect_gt(has_lint_summary$style, 0L)
  expect_identical(has_lint_summary$warning, 0L)
  expect_identical(has_lint_summary$error, 0L)
})

test_that("print.lint works", {
  # don't treat \t as width-1, #528
  l <- Lint(
    filename = "tmp", line_number = 1L, column_number = 3L,
    type = "warning", message = "this is a lint",
    line = c(`1` = "\t\t1:length(x)"), ranges = list(c(3L, 3L))
  )
  expect_output(print(l), "  1:length(x)", fixed = TRUE)
})

test_that("print.lint works with empty lints", {
  withr::local_options(list(lintr.rstudio_source_markers = FALSE))
  l <- lint(text = "1L")

  expect_message(print(l), "No lints found", fixed = TRUE)
})

test_that("print.lint works for inline data, even in RStudio", {
  l <- lint("x = 1\n")

  # Make sure lints print to console.
  # The full output is:
  #
  # > <text>:1:3: style: Use <-, not =, for assignment.
  # > x = 1
  # >   ^

  withr::with_options(
    list(lintr.rstudio_source_markers = FALSE),
    expect_output(print(l), "not =")
  )

  skip_if_not_installed("rstudioapi")
  local_mocked_bindings(
    hasFun = \(...) FALSE,
    .package = "rstudioapi"
  )
  withr::with_options(
    list(lintr.rstudio_source_markers = TRUE),
    expect_output(print(l), "not =")
  )
})

test_that("print.lints works", {
  withr::local_options(lintr.rstudio_source_markers = FALSE)

  tmp <- withr::local_tempfile()
  stopifnot(file.create(tmp))
  expect_message(expect_invisible(print(lint(tmp))), "No lints found")
})

test_that("split.lint works as intended", {
  tmp <- withr::local_tempfile(lines = c("1:nrow(x)", "1:ncol(x)"))
  l <- lint(tmp, seq_linter())
  expect_true(all(vapply(split(l), inherits, logical(1L), "lints")))
})

test_that("within.list is dispatched", {
  l <- lint(text = "a=1\nb=2", linters = infix_spaces_linter())
  expect_silent({
    # nolint next: implicit_assignment_linter. Workaround is pretty horrid.
    l <- lapply(l, within, line_number <- line_number + 1L)
  })
  expect_identical(vapply(l, `[[`, integer(1L), "line_number"), 2L:3L)
})

test_that("as_tibble.list is _not_ dispatched directly", {
  skip_if_not_installed("tibble")

  lints <- lint(text = "a = 1", linters = assignment_linter())
  expect_identical(nrow(tibble::as_tibble(lints)), 1L)

  as_tibble <- tibble::as_tibble
  with_mocked_bindings(
    requireNamespace = \(pkg, ...) pkg != "tibble" && base::requireNamespace(pkg, ...),
    expect_error(as_tibble(lints), "tibble is required to convert lints", fixed = TRUE)
  )
})

test_that("as.data.table.list is _not_ dispatched directly", {
  skip_if_not_installed("data.table")

  lints <- lint(text = "a = 1", linters = assignment_linter())
  expect_identical(nrow(data.table::as.data.table(lints)), 1L)

  # nolint next: object_name_linter. Retain data.table naming style for clarity.
  as.data.table <- data.table::as.data.table
  with_mocked_bindings(
    requireNamespace = \(pkg, ...) pkg != "data.table" && base::requireNamespace(pkg, ...),
    expect_error(as.data.table(lints), "data.table is required to convert lints", fixed = TRUE)
  )
})

local({
  # avoid impact of CLI mark-up on strwrap output.
  #   (testthat, or cli, already do so, but force it explicitly here for emphasis)
  withr::local_options(c(cli.num_colors = 0L))
  # force "default" print method even on GHA
  withr::local_envvar(c(GITHUB_ACTIONS = NA))

  test_linter <- make_linter_from_xpath("*[1]", lint_message = "The quick brown fox jumps over the lazy dog.")

  lints <- lint(text = "a", linters = test_linter())
  lint <- lints[[1L]]

  widths <- c(10L, 20L, 40L, 80L)
  test_names <- paste0(": width = ", widths)

  patrick::with_parameters_test_that(
    "print.lint, print.lints support optional message wrapping",
    {
      expect_snapshot(print(lints, width = width))

      withr::with_options(c(lintr.format_width = width), {
        expect_snapshot(print(lints))
      })
    },
    .test_name = test_names,
    width = widths
  )

  wrapped_strings <- c(
    "[test_linter]\n    The\n    quick\n    brown\n    fox\n    jumps\n    over\n    the\n    lazy\n    dog.",
    "[test_linter]\n    The quick brown\n    fox jumps over\n    the lazy dog.",
    "[test_linter] The\n    quick brown fox jumps over the lazy\n    dog.",
    "[test_linter] The quick brown fox jumps over the lazy dog."
  )

  patrick::with_parameters_test_that(
    "format.lint, format.lints support optional message wrapping",
    {
      expect_match(format(lint, width = width), wrapped_string, fixed = TRUE)
      expect_match(format(lints, width = width), wrapped_string, fixed = TRUE)

      withr::with_options(c(lintr.format_width = width), {
        expect_match(format(lint), wrapped_string, fixed = TRUE)
        expect_match(format(lints), wrapped_string, fixed = TRUE)
      })
    },
    .test_name = test_names,
    width = widths,
    wrapped_string = wrapped_strings
  )
})

test_that("GitHub Actions functionality works", {
  withr::local_envvar(list(GITHUB_ACTIONS = "true", IN_PKGDOWN = "false"))
  withr::local_options(lintr.rstudio_source_markers = FALSE)
  tmp <- withr::local_tempfile(lines = "x <- 1:nrow(y)")

  l <- lint(tmp)
  expect_output(print(l), "::warning file", fixed = TRUE)
})

test_that("GitHub Actions functionality works in a subdirectory", {
  pkg_path <- test_path("dummy_packages", "assignmentLinter")
  withr::local_envvar(list(GITHUB_ACTIONS = "true"))
  withr::local_options(lintr.rstudio_source_markers = FALSE, lintr.github_annotation_project_dir = pkg_path)

  l <- lint_package(
    pkg_path,
    linters = list(assignment_linter()),
    parse_settings = FALSE
  )
  expect_output(
    print(l),
    paste0("::warning file=", file.path(pkg_path, "R(/|\\\\)abc\\.R"))
  )
})

patrick::with_parameters_test_that(
  "GitHub Actions - error on lint works",
  {
    withr::local_envvar(list(GITHUB_ACTIONS = "true", IN_PKGDOWN = "", LINTR_ERROR_ON_LINT = env_var_value))
    withr::local_options(lintr.rstudio_source_markers = FALSE)
    tmp <- withr::local_tempfile(lines = "x <- 1:nrow(y)")

    l <- lint(tmp)

    local_mocked_bindings(quit = \(...) cat("Tried to quit.\n"))
    expect_output(print(l), "::warning file", fixed = TRUE)
  },
  env_var_value = list("T", "true")
)

patrick::with_parameters_test_that(
  "GitHub Actions - env var for error on lint is converted to logical",
  {
    withr::local_envvar(list(GITHUB_ACTIONS = "true", LINTR_ERROR_ON_LINT = env_var_value))
    withr::local_options(lintr.rstudio_source_markers = FALSE)
    tmp <- withr::local_tempfile(lines = "x <- 1:nrow(y)")

    l <- lint(tmp)

    expect_output(print(l), "::warning file", fixed = TRUE)
  },
  env_var_value = list("", "F", NA, NULL)
)

test_that("GitHub Actions log is skipped in pkgdown websites", {
  withr::local_envvar(list(GITHUB_ACTIONS = "true", IN_PKGDOWN = "true"))
  withr::local_options(lintr.rstudio_source_markers = FALSE)
  tmp <- withr::local_tempfile(lines = "x <- 1:nrow(y)")

  l <- lint(tmp, linters = seq_linter())
  expect_output(print(l), "warning: [seq_linter]", fixed = TRUE)
})

# Fixtures

fixtures <- list()

fixtures$retrieve_lint <- function() {
  file_name <- "R/test.R"
  lines <- c("foobar1", "foobar2", "foobar3")
  lints <- list(
    Lint(file_name, 1L, line = "foobar1"),
    Lint(file_name, 2L, line = "foobar2"),
    Lint(file_name, 3L, line = "foobar3")
  )
  expr <- list(content = paste(collapse = "\n", lines))
  list(
    lines = lines,
    linters = list(),
    lints = lints,
    expr = expr
  )
}

# Run tests with a temporary cache directory, so we don't leave files behind
# after running
withr::local_options(lintr.cache_directory = withr::local_tempdir())

# Helper functions

fhash <- function(filename) {
  digest::digest(filename, algo = "sha1")
}

# Tests

# `clear_cache`

test_that("clear_cache deletes the directory if no file is given", {
  local_mocked_bindings(
    read_settings = \(...) invisible(...),
    unlink = \(...) list(...)
  )

  expect_identical(clear_cache(file = NULL, path = "."), list(".", recursive = TRUE))
})

test_that("lint with cache uses the provided relative cache directory", { # nofuzz: assignment
  path <- withr::local_tempdir("my_cache_dir")
  linter <- assignment_linter()

  # create the cache
  expect_no_lint("a <- 1", linter, cache = path)
  expect_true(dir.exists(path))
  expect_length(list.files(file.path(path)), 1L)

  # read the cache
  expect_no_lint("a <- 1", linter, cache = path)
  expect_true(dir.exists(path))
})

test_that("it works outside of a package", { # nofuzz: assignment
  linter <- assignment_linter()

  local_mocked_bindings(find_package = \(...) NULL)
  path <- withr::local_tempfile(pattern = "my_cache_dir_")
  expect_false(dir.exists(path))
  expect_no_lint("a <- 1", linter, cache = path)
  expect_true(dir.exists(path))
  expect_length(list.files(path), 1L)
  expect_no_lint("a <- 1", linter, cache = path)
  expect_true(dir.exists(path))
})

test_that("cache = TRUE workflow works", {
  # Need a test structure with a safe to load .lintr
  withr::local_dir(file.path("dummy_packages", "package"))
  withr::local_options(lintr.linter_file = "lintr_test_config")
  files <- normalize_path(list.files(recursive = TRUE, full.names = TRUE))

  # Manually clear cache (that function is exported)
  for (f in files) {
    clear_cache(file = f)
  }
  l1 <- lint_package(cache = TRUE)
  l2 <- lint_package(cache = TRUE)
  expect_identical(l1, l2)
})

test_that("cache = TRUE works with nolint", {
  linters <- list(infix_spaces_linter())
  file <- withr::local_tempfile()

  writeLines("1+1\n", file)
  expect_length(lint(file, linters, cache = TRUE), 1L)

  writeLines("1+1 # nolint\n", file)
  expect_length(lint(file, linters, cache = TRUE), 0L)

  writeLines("1+1\n", file)
  expect_length(lint(file, linters, cache = TRUE), 1L)

  writeLines("1+1 # nolint\n", file)
  expect_length(lint(file, linters, cache = TRUE), 0L)
})

test_that("return lint report as checkstyle xml", {
  lints <- list(
    Lint(
      filename = "test_file",
      line_number = 1L,
      column_number = 2L,
      type = "error",
      line = "a line",
      message = "foo"
    ),
    Lint(
      filename = "test_file",
      line_number = 2L,
      column_number = 1L,
      type = "style",
      line = "another line",
      message = "bar"
    ),
    Lint(
      filename = "test_file2",
      line_number = 1L,
      column_number = 1L,
      type = "warning",
      line = "yet another line",
      message = "baz"
    )
  )
  class(lints) <- "lints"
  tmp <- withr::local_tempfile()
  checkstyle_output(lints, tmp)

  # The second line is the checkstyle version, so we ignore it during the
  # check, so we don't have to update the version every release.
  expect_identical(readLines(tmp)[-2L], readLines(test_path("checkstyle.xml"))[-2L])
})

# When called from inside a package:
# > lint_package(".")
# .. should give the same results as when called from outside the package
# with:
# > lint_package(path_to_package)

# Template packages for use in testing are stored in
# tests/testthat/dummy_packages/<pkgName>
# These packages should not have a .lintr file:  Hardcoding a .lintr in a
# dummy package throws problems during `R CMD check` (they are flagged as
# hidden files, but can't be added to RBuildIgnore since they should be
# available during `R CMD check` tests)

test_that(
  "`lint_package` does not depend on path to pkg - no excluded files",
  {
    withr::local_options(lintr.linter_file = "lintr_test_config")

    # This dummy package does not have a .lintr file, so no files / lines should
    # be excluded from analysis
    pkg_path <- test_path("dummy_packages", "assignmentLinter")

    expected_lines <- c(
      # from abc.R
      "abc = 123",
      # from jkl.R
      "jkl = 456",
      "mno = 789",
      # from exec/script.R
      "x = 1:4"
    )

    lints_from_outside <- lint_package(
      pkg_path,
      linters = list(assignment_linter())
    )
    lints_from_pkg_root <- withr::with_dir(
      pkg_path,
      lint_package(".", linters = list(assignment_linter()), parse_settings = FALSE)
    )
    lints_from_a_subdir <- withr::with_dir(
      file.path(pkg_path, "R"),
      lint_package("..", linters = list(assignment_linter()), parse_settings = FALSE)
    )

    expect_identical(
      as.data.frame(lints_from_outside)[["line"]],
      expected_lines
    )
    expect_identical(
      as.data.frame(lints_from_outside),
      as.data.frame(lints_from_pkg_root),
      info = paste(
        "lint_package() finds the same lints from pkg-root as from outside a pkg",
        "(no .lintr config present)"
      )
    )
    expect_identical(
      as.data.frame(lints_from_outside),
      as.data.frame(lints_from_a_subdir),
      info = paste(
        "lint_package() finds the same lints from a subdir as from outside a pkg",
        "(no .lintr config present)"
      )
    )
  }
)

test_that(
  "`lint_package` does not depend on path to pkg - with excluded files",
  {
    # Since excluded regions can be specified in two ways
    # list(
    #   filename = line_numbers, # approach 1
    #   filename                 # approach 2
    # ),
    # the test checks both approaches

    pkg_path <- test_path("dummy_packages", "assignmentLinter")

    # Add a .lintr that excludes the whole of `abc.R` and the first line of
    # `jkl.R` (and remove it on finishing this test)
    local_config("exclusions: list('R/abc.R', 'R/jkl.R' = 1)", pkg_path)

    expected_lines <- c("mno = 789", "x = 1:4")
    lints_from_outside <- lint_package(
      pkg_path,
      linters = list(assignment_linter())
    )
    lints_from_pkg_root <- withr::with_dir(
      pkg_path,
      lint_package(".", linters = list(assignment_linter()))
    )
    lints_from_a_subdir <- withr::with_dir(
      file.path(pkg_path, "R"),
      lint_package(".", linters = list(assignment_linter()))
    )
    lints_from_a_subsubdir <- withr::with_dir(
      file.path(pkg_path, "tests", "testthat"),
      lint_package(".", linters = list(assignment_linter()))
    )

    expect_identical(
      as.data.frame(lints_from_outside)[["line"]],
      expected_lines
    )
    expect_identical(
      as.data.frame(lints_from_outside),
      as.data.frame(lints_from_pkg_root),
      info = paste(
        "lint_package() finds the same lints from pkg-root as from outside a pkg",
        "(.lintr config present)"
      )
    )
    expect_identical(
      as.data.frame(lints_from_outside),
      as.data.frame(lints_from_a_subdir),
      info = paste(
        "lint_package() finds the same lints from a subdir as from outside a pkg",
        "(.lintr config present)"
      )
    )
    expect_identical(
      as.data.frame(lints_from_outside),
      as.data.frame(lints_from_a_subsubdir),
      info = paste(
        "lint_package() finds the same lints from a sub-subdir as from outside a pkg",
        "(.lintr config present)"
      )
    )
  }
)

test_that("lint_package returns early if no package is found", {
  temp_pkg <- withr::local_tempdir("dir")

  expect_warning(
    {
      l <- lint_package(temp_pkg)
    },
    "Didn't find any R package",
    fixed = TRUE
  )
  expect_null(l)

  # ignore a folder named DESCRIPTION, #702
  file.copy(test_path("dummy_packages", "desc_dir_pkg"), temp_pkg, recursive = TRUE)

  expect_warning(
    lint_package(file.path(temp_pkg, "desc_dir_pkg", "DESCRIPTION", "R")),
    "Didn't find any R package",
    fixed = TRUE
  )
})

test_that("length(path)>1 is not supported", {
  expect_error(lint_package(letters), "one package at a time", fixed = TRUE)
})

test_that(
  "`lint_package` will use a `.lintr` file in `.github/linters/` directory the same as the package root",
  {
    withr::local_options(lintr.linter_file = "lintr_test_config")

    pkg_path <- test_path("dummy_packages", "github_lintr_file")

    # First, ensure that the package has lint messages in the absence of a
    # custom configuration:

    pkg_lints_before <- withr::with_dir(
      pkg_path,
      lint_package(".", linters = list(quotes_linter()))
    )

    expect_identical(
      as.data.frame(pkg_lints_before)[["line"]],
      c("'abc'", "'abc'"),
      "linting the `github_lintr_file` package should fail"
    )

    # In `github/linters`add a `.lintr` file
    dir.create(
      path = file.path(pkg_path, ".github", "linters/"),
      recursive = TRUE
    )
    on.exit(unlink(file.path(pkg_path, ".github"), recursive = TRUE), add = TRUE)

    local_config(
      "linters: linters_with_defaults(quotes_linter(\"'\"))",
      file.path(pkg_path, ".github", "linters"),
      filename = "lintr_test_config"
    )

    pkg_lints <- withr::with_dir(pkg_path, lint_package("."))
    expect_length(pkg_lints, 0L)

    subdir_lints <- withr::with_dir(pkg_path, lint_dir("tests/testthat"))
    expect_length(subdir_lints, 0L)
  }
)

test_that("package using .lintr.R config lints correctly", {
  withr::local_options(lintr.linter_file = "lintr_test_config")

  r_config_pkg <- test_path("dummy_packages", "RConfig")

  lints <- as.data.frame(lint_package(r_config_pkg))
  expect_identical(unique(basename(lints$filename)), "lint_me.R")
  expect_identical(lints$linter, c("infix_spaces_linter", "any_duplicated_linter"))

  # config has bad R syntax
  expect_error(
    lint_package(test_path("dummy_packages", "RConfigInvalid")),
    "Malformed config file (lintr_test_config.R), ensure it is valid R syntax",
    fixed = TRUE
  )

  # config produces unused variables
  withr::local_options(lintr.linter_file = "lintr_test_config_extraneous")
  expect_warning(
    expect_length(lint_package(r_config_pkg), 2L),
    "Found unused settings in config file",
    fixed = TRUE
  )

  # R is preferred if multiple matched configs
  withr::local_options(lintr.linter_file = "lintr_test_config_conflict")
  lints <- as.data.frame(lint_package(r_config_pkg))
  expect_identical(unique(basename(lints$filename)), "testthat.R")
  expect_identical(lints$linter, c("expect_null_linter", "trailing_blank_lines_linter"))
})

test_that("lintr need not be attached for .lintr.R configs to use lintr functions", {
  exprs <- paste(
    'options(lintr.linter_file = "lintr_test_config")',
    sprintf('lints <- lintr::lint_package("%s")', test_path("dummy_packages", "RConfig")),
    # simplify output to be read from stdout
    'cat(paste(as.data.frame(lints)$linter, collapse = "|"), "\n", sep = "")',
    sep = "; "
  )
  if (tolower(Sys.info()[["sysname"]]) == "windows") {
    rscript <- file.path(R.home("bin"), "Rscript.exe")
  } else {
    rscript <- file.path(R.home("bin"), "Rscript")
  }
  expect_identical(
    system2(rscript, c("-e", shQuote(exprs)), stdout = TRUE),
    "infix_spaces_linter|any_duplicated_linter"
  )
})

#' Require usage of `expect_false(x)` over `expect_true(!x)`
#'
#' [testthat::expect_false()] exists specifically for testing that an output is
#'   `FALSE`. [testthat::expect_true()] can also be used for such tests by
#'   negating the output, but it is better to use the tailored function instead.
#'   The reverse is also true -- use `expect_false(A)` instead of
#'   `expect_true(!A)`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "expect_true(!x)",
#'   linters = expect_not_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "expect_false(x)",
#'   linters = expect_not_linter()
#' )
#'
#' @evalRd rd_tags("expect_not_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
expect_not_linter <- make_linter_from_function_xpath(
  function_names = c("expect_true", "expect_false"),
  xpath = "
  following-sibling::expr[OP-EXCLAMATION]
    /parent::expr
  ",
  lint_message = "expect_false(x) is better than expect_true(!x), and vice versa."
)

test_that("lint all files in a directory", {
  # NB: not using .lintr in the test packages because
  #   R CMD check doesn't like hidden files in any subdirectory
  withr::local_options(lintr.linter_file = "lintr_test_config")
  the_dir <- test_path("dummy_packages", "package", "vignettes")
  files <- list.files(the_dir)

  lints <- lint_dir(the_dir, parse_settings = FALSE)
  linted_files <- unique(names(lints))

  expect_s3_class(lints, "lints")
  expect_identical(sort(linted_files), sort(files))
})

test_that("lint all relevant directories in a package", {
  withr::local_options(lintr.linter_file = "lintr_test_config")
  the_pkg <- test_path("dummy_packages", "package")
  files <- setdiff(
    list.files(the_pkg, recursive = TRUE),
    c("package.Rproj", "DESCRIPTION", "NAMESPACE", "lintr_test_config")
  )

  lints <- lint_package(the_pkg, parse_settings = FALSE)
  linted_files <- unique(names(lints))

  # lintr paths contain backslash on windows, list.files uses forward slash.
  linted_files <- gsub("\\", "/", linted_files, fixed = TRUE)

  expect_s3_class(lints, "lints")
  expect_identical(sort(linted_files), sort(files))

  # Code coverage is not detected for default_linters.
  # We want to ensure that object_name_linter uses namespace_imports correctly.
  # assignment_linter is needed to cause a lint in all vignettes.
  linters <- list(assignment_linter(), object_name_linter())
  lints <- lint_package(the_pkg, linters = linters, parse_settings = FALSE)
  linted_files <- unique(names(lints))

  # lintr paths contain backslash on windows, list.files uses forward slash.
  linted_files <- gsub("\\", "/", linted_files, fixed = TRUE)

  expect_s3_class(lints, "lints")
  expect_identical(sort(linted_files), sort(files))
})

test_that("respects directory exclusions", {
  the_dir <- withr::local_tempdir()

  the_excluded_dir <- file.path(the_dir, "exclude-me")
  dir.create(the_excluded_dir)

  defaults_path <- test_path("default_linter_testcode.R")

  file.copy(defaults_path, the_dir)
  file.copy(defaults_path, the_excluded_dir)
  file.copy(defaults_path, file.path(the_excluded_dir, "bad2.R"))

  lints <- lint_dir(the_dir, exclusions = "exclude-me")
  linted_files <- unique(names(lints))
  expect_length(linted_files, 1L)
  expect_identical(linted_files, "default_linter_testcode.R")

  lints_norm <- lint_dir(the_dir, exclusions = "exclude-me", relative_path = FALSE)
  linted_files <- unique(names(lints_norm))
  expect_length(linted_files, 1L)
  expect_identical(linted_files, normalize_path(file.path(the_dir, "default_linter_testcode.R")))
})

test_that("respect directory exclusions from settings", {
  the_dir <- withr::local_tempdir()

  the_excluded_dir <- file.path(the_dir, "exclude-me")
  dir.create(the_excluded_dir)

  defaults_path <- test_path("default_linter_testcode.R")

  file.copy(defaults_path, the_dir)
  file.copy(defaults_path, the_excluded_dir)
  file.copy(defaults_path, file.path(the_excluded_dir, "bad2.R"))
  cat("exclusions:\n  'exclude-me'\n", file = file.path(the_dir, ".lintr"))

  lints <- lint_dir(the_dir)
  linted_files <- unique(names(lints))
  expect_length(linted_files, 1L)
})

test_that("lint_dir works with specific linters without specifying other arguments", {
  withr::local_options(lintr.linter_file = "lintr_test_config")
  the_dir <- test_path("dummy_packages", "package", "vignettes")
  expect_length(lint_dir(the_dir, assignment_linter(), parse_settings = FALSE), 14L)
  expect_length(lint_dir(the_dir, commented_code_linter(), parse_settings = FALSE), 0L)
})

test_that("typo in argument name gives helpful error", {
  expect_error(lint_dir(litners = identity), "Found unknown arguments in `...`: `litners`")
})

test_that("linting empty directory passes", {
  expect_length(lint_dir(withr::local_tempdir(), any_duplicated_linter()), 0L)
})

test_that("linters", {
  # non-empty named list of functions
  x <- default_linters
  expect_type(x, "list")
  expect_gt(length(x), 0L)
  expect_true(all(nzchar(names(x))))
  expect_true(all(vapply(x, inherits, logical(1L), "linter")))
  expect_true(all(vapply(x, is.function, logical(1L))))
})

test_that("undesirable functions and operators", {
  # non-empty named list of NAs and character strings
  vars <- list(
    all_undesirable_functions, default_undesirable_functions,
    all_undesirable_operators, default_undesirable_operators
  )

  for (x in vars) {
    expect_type(x, "list")
    expect_gt(length(x), 0L)
    expect_true(all(nzchar(names(x))))
    expect_true(all(vapply(x, \(x) is.na(x) || is.character(x), logical(1L))))
    expect_true(all(lengths(x) == 1L))
  }
})

test_that("settings", {
  # non-empty named list
  x <- default_settings
  expect_type(default_settings, "list")
  expect_gt(length(default_settings), 0L)
  expect_true(all(nzchar(names(default_settings))))
})

patrick::with_parameters_test_that(
  "all default linters throw lints with their name on 'default_linter_testcode.R'",
  {
    lints <- lint(
      test_path("default_linter_testcode.R"),
      linters = default_linters[[linter_name]],
      parse_settings = FALSE
    )
    lint_df <- as.data.frame(lints)
    expect_gt(nrow(lint_df), 0L)
    reported_linter_name <- lint_df[["linter"]][1L]
    expect_identical(reported_linter_name, linter_name)
  },
  linter_name = names(default_linters),
  .test_name = names(default_linters)
)

test_that("it returns markers which match lints", {
  skip_if_not_installed("rstudioapi")
  local_mocked_bindings(
    callFun = \(...) list(...),
    executeCommand = \(...) NULL,
    .package = "rstudioapi"
  )

  lint1 <- list(Lint(
    filename = "test_file",
    line_number = 1L,
    column_number = 2L,
    type = "error",
    line = "a line",
    message = "hi"
  ))
  class(lint1) <- "lints"
  lint1[[1L]]$linter <- "linter_name"

  marker1 <- rstudio_source_markers(lint1)
  expect_identical(marker1$name, "lintr")
  expect_identical(marker1$markers[[1L]]$type, lint1[[1L]]$type)
  expect_identical(marker1$markers[[1L]]$file, lint1[[1L]]$filename)
  expect_identical(marker1$markers[[1L]]$line, lint1[[1L]]$line_number)
  expect_identical(marker1$markers[[1L]]$column, lint1[[1L]]$column_number)
  expect_identical(marker1$markers[[1L]]$message, paste0("[", lint1[[1L]]$linter, "] ", lint1[[1L]]$message))

  lint2 <- list(
    Lint(
      filename = "test_file",
      line_number = 1L,
      column_number = 2L,
      type = "error",
      line = "a line",
      message = "hi"
    ),
    Lint(
      filename = "test_file2",
      line_number = 10L,
      column_number = 1L,
      type = "warning",
      message = "test a message"
    )
  )
  class(lint2) <- "lints"
  lint2[[1L]]$linter <- "linter_name"
  lint2[[2L]]$linter <- "linter_name"
  marker2 <- rstudio_source_markers(lint2)
  expect_identical(marker2$name, "lintr")
  expect_identical(marker2$markers[[1L]]$type, lint2[[1L]]$type)
  expect_identical(marker2$markers[[1L]]$file, lint2[[1L]]$filename)
  expect_identical(marker2$markers[[1L]]$line, lint2[[1L]]$line_number)
  expect_identical(marker2$markers[[1L]]$column, lint2[[1L]]$column_number)
  expect_identical(marker2$markers[[1L]]$message, paste0("[", lint2[[1L]]$linter, "] ", lint2[[1L]]$message))
})

test_that("it prepends the package path if it exists", {
  skip_if_not_installed("rstudioapi")
  local_mocked_bindings(
    callFun = \(...) list(...),
    executeCommand = \(...) NULL,
    .package = "rstudioapi"
  )

  lint3 <- list(Lint(
    filename = "test_file",
    line_number = 1L,
    column_number = 2L,
    type = "error",
    line = "a line",
    message = "hi"
  ))
  class(lint3) <- "lints"
  attr(lint3, "path") <- "test"
  lint3[[1L]]$linter <- "linter_name"
  marker3 <- rstudio_source_markers(lint3)
  expect_identical(marker3$name, "lintr")
  expect_identical(marker3$basePath, "test")
  expect_identical(marker3$markers[[1L]]$type, lint3[[1L]]$type)
  expect_identical(marker3$markers[[1L]]$file, file.path("test", lint3[[1L]]$filename))
  expect_identical(marker3$markers[[1L]]$line, lint3[[1L]]$line_number)
  expect_identical(marker3$markers[[1L]]$column, lint3[[1L]]$column_number)
  expect_identical(marker3$markers[[1L]]$message, paste0("[", lint3[[1L]]$linter, "] ", lint3[[1L]]$message))
})

test_that("it returns an empty list of markers if there are no lints", {
  skip_if_not_installed("rstudioapi")
  local_mocked_bindings(
    callFun = \(...) list(...),
    executeCommand = \(...) NULL,
    .package = "rstudioapi"
  )

  lint4 <- `class<-`(list(), "lints")
  marker4 <- rstudio_source_markers(lint4)
  expect_identical(marker4$name, "lintr")
  expect_identical(marker4$markers, list())
})

test_that("rstudio_source_markers apply to print within rstudio", {
  withr::local_options(lintr.rstudio_source_markers = TRUE)

  tmp <- withr::local_tempfile(lines = "1:ncol(x)")
  empty <- withr::local_tempfile(lines = character(0L))

  skip_if_not_installed("rstudioapi")
  local_mocked_bindings(
    hasFun = \(...) TRUE,
    .package = "rstudioapi"
  )
  local_mocked_bindings(rstudio_source_markers = \(x) cat("matched\n"))

  l <- lint(tmp, seq_linter())
  expect_output(print(l), "matched", fixed = TRUE)

  l <- lint(empty, seq_linter())
  expect_message(expect_output(print(l), "matched", fixed = TRUE), "No lints found")
})

test_that("Lint() errors on invalid input", {
  dummy_line <- "abc"
  expect_error(
    Lint("dummy.R", line = dummy_line, column_number = NA_integer_),
    "`column_number` must be an integer between 0 and 4 (`nchar(line) + 1`)",
    fixed = TRUE
  )
  expect_error(
    Lint("dummy.R", line = dummy_line, line_number = 0L),
    "`line_number` must be a positive integer",
    fixed = TRUE
  )
  expect_error(
    Lint("dummy.R", ranges = c(1L, 3L)),
    "`ranges` must be `NULL` or a list",
    fixed = TRUE
  )
  expect_error(
    Lint("dummy.R", ranges = list(1L)),
    "`ranges` must only contain integer vectors of length 2 without `NA`s.",
    fixed = TRUE
  )
  expect_error(
    Lint("dummy.R", ranges = list(c(1L, NA_integer_))),
    "`ranges` must only contain integer vectors of length 2 without `NA`s.",
    fixed = TRUE
  )
  expect_error(
    Lint("dummy.R", line = dummy_line, ranges = list(c(1L, 2L), c(1L, 5L))),
    "`ranges` must satisfy 0 <= range[1L] <= range[2L] <= 4 (nchar(line) + 1).",
    fixed = TRUE
  )
  expect_error(
    Lint("dummy.R", line = 1L),
    "`line` must be a character string",
    fixed = TRUE
  )
  expect_error(
    Lint("dummy.R", message = 1L),
    "`message` must be a character string",
    fixed = TRUE
  )
  expect_error(
    Lint("dummy.R", message = letters),
    "`message` must be a character string",
    fixed = TRUE
  )
  expect_warning(
    Lint("dummy.R", message = glue("a")),
    "`message` must be a simple string",
    fixed = TRUE
  )
})

#' Use lintr in your project
#'
#' Create a minimal lintr config file as a starting point for customization
#'
#' @param path Path to project root, where a `.lintr` file should be created.
#' If the `.lintr` file already exists, an error will be thrown.
#' @param type What kind of configuration to create?
#'
#'  * `tidyverse` creates a minimal lintr config, based on the default linters ([linters_with_defaults()]).
#'    These are suitable for following [the tidyverse style guide](https://style.tidyverse.org/).
#'  * `full` creates a lintr config using all available linters via [all_linters()].
#'
#' @return Path to the generated configuration, invisibly.
#'
#' @export
#' @seealso `vignette("lintr")` for detailed introduction to using and configuring lintr.
#' @examples
#' if (FALSE) {
#'   # use the default set of linters
#'   lintr::use_lintr()
#'   # or try all linters
#'   lintr::use_lintr(type = "full")
#'
#'   # then
#'   lintr::lint_dir()
#' }
use_lintr <- function(path = ".", type = c("tidyverse", "full")) {
  config_file <- normalize_path(file.path(path, lintr_option("linter_file")), mustWork = FALSE)
  if (file.exists(config_file)) {
    cli_abort("Found an existing configuration file at {.file {config_file}}.")
  }
  type <- match.arg(type)
  the_config <- switch(type,
    tidyverse = list(
      linters = 'linters_with_defaults() # see vignette("lintr")',
      encoding = '"UTF-8"'
    ),
    full = list(
      linters = 'all_linters(packages = "lintr") # see vignette("lintr")',
      encoding = '"UTF-8"',
      exclusions = 'list("renv", "packrat") # see ?lintr::exclude'
    )
  )
  write.dcf(the_config, config_file, width = Inf)

  pkg_path <- find_package(path)

  if (is.null(pkg_path)) {
    return(invisible(config_file))
  }

  rbuildignore_path <- file.path(pkg_path, ".Rbuildignore")
  rel_path <- xfun::relative_path(config_file, pkg_path)
  escaped_config_path <- rex(start, rel_path, end)

  if (!file.exists(rbuildignore_path)) {
    writeLines(escaped_config_path, rbuildignore_path)
    cli_inform("Added {.val {escaped_config_path}} to {.code .Rbuildignore}.")
    return(invisible(config_file))
  }

  ignored <- readLines(rbuildignore_path, warn = FALSE)

  if (escaped_config_path %in% ignored) {
    cli_inform("Configuration file {.val {escaped_config_path}} is already ignored in {.code .Rbuildignore}.")
    return(invisible(config_file))
  }

  writeLines(c(ignored, escaped_config_path), rbuildignore_path)
  cli_inform("Added {.val {escaped_config_path}} to {.code .Rbuildignore}.")

  invisible(config_file)
}

#' Block usage of `file.path()` with `system.file()`
#'
#' [system.file()] has a `...` argument which, internally, is passed to
#'   [file.path()], so including it in user code is repetitive.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'system.file(file.path("path", "to", "data"), package = "foo")',
#'   linters = system_file_linter()
#' )
#'
#' lint(
#'   text = 'file.path(system.file(package = "foo"), "path", "to", "data")',
#'   linters = system_file_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'system.file("path", "to", "data", package = "foo")',
#'   linters = system_file_linter()
#' )
#'
#' @evalRd rd_tags("system_file_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
system_file_linter <- function() {
  # either system.file(file.path(...)) or file.path(system.file(...))
  file_path_xpath <- "
  self::*[following-sibling::expr/expr/SYMBOL_FUNCTION_CALL[text() = 'system.file']]
    /parent::expr
  "
  system_file_xpath <- "
  self::*[following-sibling::expr/expr/SYMBOL_FUNCTION_CALL[text() = 'file.path']]
    /parent::expr
  "

  Linter(linter_level = "expression", function(source_expression) {
    file_path_calls <- source_expression$xml_find_function_calls("file.path")
    system_file_calls <- source_expression$xml_find_function_calls("system.file")

    bad_expr <- combine_nodesets(
      xml_find_all_(file_path_calls, file_path_xpath),
      xml_find_all_(system_file_calls, system_file_xpath)
    )

    outer_call <- xp_call_name(bad_expr)
    lint_message <- paste(
      "Use the `...` argument of system.file() to expand paths,",
      'e.g. system.file("data", "model.csv", package = "myrf") instead of',
      ifelse(
        outer_call == "system.file",
        'system.file(file.path("data", "model.csv"), package = "myrf")',
        'file.path(system.file(package = "myrf"), "data", "model.csv")'
      )
    )

    xml_nodes_to_lints(bad_expr, source_expression, lint_message, type = "warning")
  })
}

#' Prohibit close() from terminating a function definition
#'
#' Functions that end in `close(x)` are almost always better written by using
#'   `on.exit(close(x))` close to where `x` is defined and/or opened.
#'
#' @examples
#' # will produce lints
#' code <- paste(
#'   "f <- function(fl) {",
#'   "  conn <- file(fl, open = 'r')",
#'   "  readLines(conn)",
#'   "  close(conn)",
#'   "}",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = terminal_close_linter()
#' )
#'
#' # okay
#' code <- paste(
#'   "f <- function(fl) {",
#'   "  conn <- file(fl, open = 'r')",
#'   "  on.exit(close(conn))",
#'   "  readLines(conn)",
#'   "}",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = terminal_close_linter()
#' )
#'
#' @evalRd rd_tags("terminal_close_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
terminal_close_linter <- make_linter_from_xpath(
  xpath = "
  (//FUNCTION | //OP-LAMBDA)
    /following-sibling::expr
    /expr[last()][
      expr/SYMBOL_FUNCTION_CALL[text() = 'close']
      or expr[
        SYMBOL_FUNCTION_CALL[text() = 'return']
        and following-sibling::expr/expr/SYMBOL_FUNCTION_CALL[text() = 'close']
      ]
    ]
  ",
  lint_message = "Use on.exit(close(x)) to close connections instead of running it as the last call in a function."
)

#' Missing package linter
#'
#' Check for missing packages in `library()`, `require()`, `loadNamespace()`, and `requireNamespace()` calls.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "library(xyzxyz)",
#'   linters = missing_package_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "library(stats)",
#'   linters = missing_package_linter()
#' )
#'
#' @evalRd rd_tags("missing_package_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
missing_package_linter <- function() {
  library_require_xpath <- "
  parent::expr[
    expr[2][STR_CONST]
    or (
      expr[2][SYMBOL]
      and not(
        SYMBOL_SUB[text() = 'character.only']
        /following-sibling::expr[1]
        /NUM_CONST[text() = 'TRUE' or text() = 'T']
      )
    )
  ]"
  load_require_namespace_xpath <- "
  following-sibling::expr[1][STR_CONST]
    /parent::expr
  "

  Linter(linter_level = "file", function(source_expression) {
    library_require_calls <- source_expression$xml_find_function_calls(c("library", "require"))
    load_require_namespace_calls <- source_expression$xml_find_function_calls(c("loadNamespace", "requireNamespace"))
    pkg_calls <- combine_nodesets(
      xml_find_all_(library_require_calls, library_require_xpath),
      xml_find_all_(load_require_namespace_calls, load_require_namespace_xpath)
    )
    pkg_names <- get_r_string(xml_find_all_(
      pkg_calls,
      "OP-LEFT-PAREN[1]/following-sibling::expr[1][SYMBOL | STR_CONST]"
    ))

    # run here, not in the factory, to allow for run- vs. "compile"-time differences in available packages
    installed_packges <- .packages(all.available = TRUE)
    missing_pkgs <- !(pkg_names %in% installed_packges)

    xml_nodes_to_lints(
      pkg_calls[missing_pkgs],
      source_expression = source_expression,
      lint_message = sprintf("Package '%s' is not installed.", pkg_names[missing_pkgs]),
      type = "warning"
    )
  })
}

generate_top_level_map <- function(pc) {
  if (is.null(pc)) {
    return(NULL)
  }
  tl_ids <- pc$id[pc$parent <= 0L]
  tl_parent <- pc$parent
  tl_parent[pc$parent <= 0L] <- tl_ids
  i_not_assigned <- which(!tl_parent %in% tl_ids)
  while ((prev_length <- length(i_not_assigned)) > 0L) { # nolint: implicit_assignment_linter. TODO(#2015): remove this.
    tl_parent[i_not_assigned] <- pc$parent[match(tl_parent[i_not_assigned], pc$id)]
    i_not_assigned <- which(!tl_parent %in% tl_ids)
    # nocov start
    if (length(i_not_assigned) >= prev_length) {
      cli_abort_internal("Logical error: unassigned set did not shrink. Check file syntax.")
    }
    # nocov end
  }
  tl_parent
}

lag <- function(x) {
  c(NA, x[seq_len(length(x) - 1L)])
}

#' Require usage of boolean operators over equivalent arithmetic
#'
#' `length(which(x == y)) == 0` is the same as `!any(x == y)`, but the latter
#'   is more readable and more efficient.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "length(which(x == y)) == 0L",
#'   linters = boolean_arithmetic_linter()
#' )
#'
#' lint(
#'   text = "sum(grepl(pattern, x)) == 0",
#'   linters = boolean_arithmetic_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "!any(x == y)",
#'   linters = boolean_arithmetic_linter()
#' )
#'
#' lint(
#'   text = "!any(grepl(pattern, x))",
#'   linters = boolean_arithmetic_linter()
#' )
#'
#' @evalRd rd_tags("boolean_arithmetic_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
boolean_arithmetic_linter <- function() {
  # TODO(#1581): extend to include all()-alike expressions
  zero_expr <- "(EQ or NE or GT or LE) and expr/NUM_CONST[text() = '0' or text() = '0L']"
  one_expr <- "(LT or GE) and expr/NUM_CONST[text() = '1' or text() = '1L']"
  length_xpath <- glue("
  parent::expr
    /parent::expr[
      expr/SYMBOL_FUNCTION_CALL[text() = 'length']
      and parent::expr[ ({zero_expr}) or ({one_expr})]
    ]
  ")
  known_logical_calls <- c(
    "grepl", "str_detect", "nzchar", "startsWith", "endsWith",
    "xor", "is.element", "duplicated",
    "is.na", "is.nan", "is.finite", "is.infinite",
    NULL
  )
  sum_xpath <- glue("
  parent::expr[
    expr[
      expr/SYMBOL_FUNCTION_CALL[{xp_text_in_table(known_logical_calls)}]
      or (EQ or NE or GT or LT or GE or LE or AND or OR or OP-EXCLAMATION)
      or SPECIAL[text() = '%in%' or text() = '%chin%']
    ]
    and parent::expr[ ({zero_expr}) or ({one_expr})]
  ]")

  Linter(linter_level = "expression", function(source_expression) {
    length_calls <- source_expression$xml_find_function_calls(c("which", "grep"))
    sum_calls <- source_expression$xml_find_function_calls("sum")
    any_expr <- c(
      xml_find_all_(length_calls, length_xpath),
      xml_find_all_(sum_calls, sum_xpath)
    )

    xml_nodes_to_lints(
      any_expr,
      source_expression = source_expression,
      # TODO(#2464): customize this?
      lint_message = paste(
        "Use any() to express logical aggregations.",
        "For example, replace length(which(x == y)) == 0 with !any(x == y)."
      ),
      type = "warning"
    )
  })
}

#' Assignment linter
#'
#' Check that the specified operator is used for assignment.
#'
#' @param operator Character vector of valid assignment operators. Defaults to allowing only `<-`; other valid
#'   options are `=`, `<<-`, `->`, `->>`, `%<>%`; use `"any"` to denote "allow all operators", in which case
#    this linter only considers `allow_trailing` for generating lints.
#' @param allow_trailing Logical, default `TRUE`. If `FALSE` then assignments aren't allowed at end of lines.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x = mean(x)",
#'   linters = assignment_linter()
#' )
#'
#' code_lines <- "1 -> x\n2 ->> y\nz <<- 3"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = assignment_linter()
#' )
#'
#' lint(
#'   text = "x %<>% as.character()",
#'   linters = assignment_linter()
#' )
#'
#' lint(
#'   text = "x <- 1",
#'   linters = assignment_linter(operator = "=")
#' )
#'
#' # okay
#' lint(
#'   text = "x <- mean(x)",
#'   linters = assignment_linter()
#' )
#'
#' code_lines <- "x <- 1\ny <- 2\nz <- 3"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = assignment_linter()
#' )
#'
#' # customizing using arguments
#' code_lines <- "1 -> x\n2 ->> y"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = assignment_linter(operator = "->")
#' )
#'
#' lint(
#'   text = "x <<- 1",
#'   linters = assignment_linter(operator = "<-")
#' )
#'
#' writeLines("foo(bar = \n 1)")
#' lint(
#'   text = "foo(bar = \n 1)",
#'   linters = assignment_linter(allow_trailing = FALSE)
#' )
#'
#' lint(
#'   text = "x %<>% as.character()",
#'   linters = assignment_linter(operator = "%<>%")
#' )
#'
#' lint(
#'   text = "x = 1",
#'   linters = assignment_linter(operator = "=")
#' )
#'
#' @evalRd rd_tags("assignment_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#assignment-1>
#' - <https://style.tidyverse.org/pipes.html#assignment-2>
#' @export
assignment_linter <- function(operator = "<-",
                              allow_trailing = TRUE) {
  all_operators <- c("<-", "=", "->", "<<-", "->>", "%<>%")
  if ("any" %in% operator) {
    operator <- all_operators
  } else {
    operator <- match.arg(operator, all_operators, several.ok = TRUE)
  }
  no_cascading <- !any(c("<<-", "->>") %in% operator)
  trailing_assign_xpath <- paste0(
    collapse = " | ",
    c(
      "//LEFT_ASSIGN",
      "//RIGHT_ASSIGN",
      "//EQ_ASSIGN",
      "//EQ_SUB",
      "//EQ_FORMALS",
      "//SPECIAL[text() = '%<>%']"
    ),
    "[@line1 < following-sibling::expr[1]/@line1]"
  )

  op_xpath_parts <- c(
    if (!"=" %in% operator) "//EQ_ASSIGN",
    # -> and ->> are both 'RIGHT_ASSIGN'
    glue("//RIGHT_ASSIGN[{ xp_text_in_table(setdiff(c('->', '->>'), operator)) }]"),
    # <-, :=, and <<- are all 'LEFT_ASSIGN'; check the text if blocking <<-.
    # NB: := is not linted because of (1) its common usage in rlang/data.table and
    #   (2) it's extremely uncommon as a normal assignment operator
    glue("//LEFT_ASSIGN[{ xp_text_in_table(setdiff(c('<-', '<<-'), operator)) }]"),
    if (!"%<>%" %in% operator) "//SPECIAL[text() = '%<>%']"
  )
  if (!is.null(op_xpath_parts)) {
    implicit_assignment_xpath <- "[not(ancestor::expr[
      preceding-sibling::*[
        self::expr/SYMBOL_FUNCTION_CALL
        or self::IF
        or self::WHILE
        or self::IN
      ]
      and not(descendant-or-self::expr/*[1][self::OP-LEFT-PAREN])
    ])]"
    op_xpath <- paste0(op_xpath_parts, implicit_assignment_xpath, collapse = " | ")
  } else {
    op_xpath <- NULL
  }

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    lints <- NULL
    if (!is.null(op_xpath)) {
      op_expr <- xml_find_all_(xml, op_xpath)

      op_text <- xml_text(op_expr)
      allowed_op_list <- if (length(operator) > 1L) paste("one of", toString(operator)) else operator
      op_lint_message_fmt <- rep(
        "Use %s for assignment, not %s.",
        length(op_text)
      )
      if (no_cascading) {
        op_lint_message_fmt[op_text %in% c("<<-", "->>")] <-
          "Replace %2$s by assigning to a specific environment (with assign() or <-) to avoid hard-to-predict behavior."
      }
      op_lint_message_fmt[op_text == "%<>%"] <-
        "Avoid the assignment pipe %2$s; prefer pipes and assignment in separate steps."

      # suppress "one argument not used by format" warnings for the special cased messages that don't include
      # allowed_op_list.
      op_lint_message <- suppressWarnings(sprintf(op_lint_message_fmt, allowed_op_list, op_text))
      lints <- xml_nodes_to_lints(op_expr, source_expression, op_lint_message, type = "style")
    }

    if (!allow_trailing) {
      trailing_assign_expr <- xml_find_all_(xml, trailing_assign_xpath)
      trailing_assign_text <- xml_text(trailing_assign_expr)
      trailing_assign_msg_fmt <- "Assignment %s should not be trailing at the end of a line."
      trailing_assign_msg <- sprintf(trailing_assign_msg_fmt, trailing_assign_text)
      lints <- c(lints,
        xml_nodes_to_lints(trailing_assign_expr, source_expression, trailing_assign_msg, type = "style")
      )
    }

    lints
  })
}

#' Require usage of `startsWith()` and `endsWith()` over `grepl()`/`substr()` versions
#'
#' [base::startsWith()] is used to detect fixed initial substrings; it is more
#'   readable and more efficient than equivalents using [grepl()] or [substr()].
#'   c.f. `startsWith(x, "abc")`, `grepl("^abc", x)`,
#'   `substr(x, 1L, 3L) == "abc"`.
#'
#' Ditto for using [base::endsWith()] to detect fixed terminal substrings.
#'
#' Note that there is a difference in behavior between how `grepl()` and `startsWith()`
#'   (and `endsWith()`) handle missing values. In particular, for `grepl()`, `NA` inputs
#'   are considered `FALSE`, while for `startsWith()`, `NA` inputs have `NA` outputs.
#'   That means the strict equivalent of `grepl("^abc", x)` is
#'   `!is.na(x) & startsWith(x, "abc")`.
#'
#' We lint `grepl()` usages by default because the `!is.na()` version is more explicit
#'   with respect to `NA` handling -- though documented, the way `grepl()` handles
#'   missing inputs may be surprising to some users.
#'
#' @param allow_grepl Logical, default `FALSE`. If `TRUE`, usages with `grepl()`
#'   are ignored. Some authors may prefer the conciseness offered by `grepl()` whereby
#'   `NA` input maps to `FALSE` output, which doesn't have a direct equivalent
#'   with `startsWith()` or `endsWith()`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'grepl("^a", x)',
#'   linters = string_boundary_linter()
#' )
#'
#' lint(
#'   text = 'grepl("z$", x)',
#'   linters = string_boundary_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'startsWith(x, "a")',
#'   linters = string_boundary_linter()
#' )
#'
#' lint(
#'   text = 'endsWith(x, "z")',
#'   linters = string_boundary_linter()
#' )
#'
#' # If missing values are present, the suggested alternative wouldn't be strictly
#' # equivalent, so this linter can also be turned off in such cases.
#' lint(
#'   text = 'grepl("z$", x)',
#'   linters = string_boundary_linter(allow_grepl = TRUE)
#' )
#'
#' @evalRd rd_tags("string_boundary_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
string_boundary_linter <- function(allow_grepl = FALSE) {
  str_cond <- xp_and(
    "string-length(text()) > 3",
    "contains(text(), '^') or contains(text(), '$')"
  )
  str_detect_xpath <- glue("
  following-sibling::expr[2]
    /STR_CONST[ {str_cond} ]
  ")
  str_detect_message_map <- c(
    both = "Use == to check for an exact string match.",
    initial = "Use startsWith() to detect a fixed initial substring.",
    terminal = "Use endsWith() to detect a fixed terminal substring."
  )

  if (!allow_grepl) {
    grepl_xpath <- glue("
    parent::expr[
      not(SYMBOL_SUB[
        text() = 'ignore.case'
        and not(following-sibling::expr[1][NUM_CONST[text() = 'FALSE'] or SYMBOL[text() = 'F']])
      ])
      and not(SYMBOL_SUB[
        text() = 'fixed'
        and not(following-sibling::expr[1][NUM_CONST[text() = 'FALSE'] or SYMBOL[text() = 'F']])
      ])
    ]
      /expr[2]
      /STR_CONST[ {str_cond} ]
    ")
    grepl_lint_fmt <- paste(
      "Use !is.na(x) & %1$s(x, string) to detect a fixed %2$s substring, or,",
      "if missingness is not a concern, just %1$s()."
    )
    grepl_message_map <- c(
      both = "Use == to check for an exact string match.",
      initial = sprintf(grepl_lint_fmt, "startsWith", "initial"),
      terminal = sprintf(grepl_lint_fmt, "endsWith", "terminal")
    )
  }

  get_regex_lint_data <- function(xml, xpath) {
    expr <- xml_find_all_(xml, xpath)
    patterns <- get_r_string(expr)
    initial_anchor <- startsWith(patterns, "^")
    terminal_anchor <- endsWith(patterns, "$")
    search_start <- 1L + initial_anchor
    search_end <- nchar(patterns) - terminal_anchor
    should_lint <- (initial_anchor | terminal_anchor) &
      is_not_regex(substr(patterns, search_start, search_end))
    initial_anchor <- initial_anchor[should_lint]
    terminal_anchor <- terminal_anchor[should_lint]

    lint_type <- character(length(initial_anchor))

    lint_type[initial_anchor & terminal_anchor] <- "both"
    lint_type[initial_anchor & !terminal_anchor] <- "initial"
    lint_type[!initial_anchor & terminal_anchor] <- "terminal"
    list(lint_expr = expr[should_lint], lint_type = lint_type)
  }

  string_comparison_xpath <- "self::*[(EQ or NE) and expr/STR_CONST]"
  substr_xpath <- glue("
  self::*[expr/expr[
    (
      position() = 3
      and NUM_CONST[text() = '1' or text() = '1L']
    ) or (
      position() = 4
      and expr[1][SYMBOL_FUNCTION_CALL[text() = 'nchar']]
      and expr[position() = 2] = preceding-sibling::expr[2]
    )
  ]]")

  substr_arg2_xpath <- "string(./expr[expr[1][SYMBOL_FUNCTION_CALL]]/expr[3])"

  Linter(linter_level = "expression", function(source_expression) {
    lints <- list()

    str_detect_lint_data <- get_regex_lint_data(
      source_expression$xml_find_function_calls("str_detect"),
      str_detect_xpath
    )
    str_detect_lint_message <- str_detect_message_map[str_detect_lint_data$lint_type]

    lints <- c(lints, xml_nodes_to_lints(
      str_detect_lint_data$lint_expr,
      source_expression = source_expression,
      lint_message = paste(str_detect_lint_message, "Doing so is more readable and more efficient."),
      type = "warning"
    ))

    if (!allow_grepl) {
      grepl_lint_data <- get_regex_lint_data(source_expression$xml_find_function_calls("grepl"), grepl_xpath)
      grepl_lint_message <- grepl_message_map[grepl_lint_data$lint_type]

      lints <- c(lints, xml_nodes_to_lints(
        grepl_lint_data$lint_expr,
        source_expression = source_expression,
        lint_message = paste(grepl_lint_message, "Doing so is more readable and more efficient."),
        type = "warning"
      ))
    }

    substr_calls <- xml_find_all_(
      source_expression$xml_find_function_calls(c("substr", "substring")),
      "parent::*/parent::*"
    )
    is_str_comparison <- !is.na(xml_find_first_(substr_calls, string_comparison_xpath))
    substr_calls <- strip_comments_from_subtree(substr_calls[is_str_comparison])
    substr_expr <- xml_find_all_(substr_calls, substr_xpath)
    substr_one <- xml_find_chr_(substr_expr, substr_arg2_xpath) %in% c("1", "1L")
    substr_lint_message <- paste(
      ifelse(
        substr_one,
        "Use startsWith() to detect an initial substring.",
        "Use endsWith() to detect a terminal substring."
      ),
      "Doing so is more readable and more efficient."
    )

    lints <- c(lints, xml_nodes_to_lints(substr_expr, source_expression, substr_lint_message, type = "warning"))
    lints
  })
}

#' TODO comment linter
#'
#' Check that the source contains no TODO comments (case-insensitive).
#'
#' @param todo Vector of case-insensitive strings that identify TODO comments.
#' @param except_regex Vector of case-sensitive regular expressions that identify
#'    _valid_ TODO comments.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x + y # TOODOO",
#'   linters = todo_comment_linter(todo = "toodoo")
#' )
#'
#' lint(
#'   text = "pi <- 1.0 # FIIXMEE",
#'   linters = todo_comment_linter(todo = "fiixmee")
#' )
#'
#' lint(
#'   text = "x <- TRUE # TOODOO(#1234): Fix this hack.",
#'   linters = todo_comment_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x + y # my informative comment",
#'   linters = todo_comment_linter()
#' )
#'
#' lint(
#'   text = "pi <- 3.14",
#'   linters = todo_comment_linter()
#' )
#'
#' lint(
#'   text = "x <- TRUE",
#'   linters = todo_comment_linter()
#' )
#'
#' lint(
#'   text = "x <- TRUE # TODO(#1234): Fix this hack.",
#'   linters = todo_comment_linter(except_regex = "TODO\\(#[0-9]+\\):")
#' )
#'
#' @evalRd rd_tags("todo_comment_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
todo_comment_linter <- function(todo = c("todo", "fixme"), except_regex = NULL) {
  todo_comment_regex <- rex(one_or_more("#"), maybe("'"), any_spaces, or(todo))
  valid_todo_regex <- NULL
  if (!is.null(except_regex)) {
    valid_todo_regex <- paste0("#+'?", rex::shortcuts$any_spaces, "(?:", paste(except_regex, collapse = "|"), ")")
  }

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    comment_expr <- xml_find_all_(xml, "//COMMENT")
    comment_text <- xml_text(comment_expr)
    invalid_todo <- re_matches_logical(comment_text, todo_comment_regex, ignore.case = TRUE)
    if (!is.null(valid_todo_regex)) {
      invalid_todo <- invalid_todo & !re_matches_logical(comment_text, valid_todo_regex)
    }

    xml_nodes_to_lints(
      comment_expr[invalid_todo],
      source_expression = source_expression,
      lint_message = "Remove TODO comments.",
      type = "style"
    )
  })
}

#' Block usage of anonymous functions in iteration functions when unnecessary
#'
#' Using an anonymous function in, e.g., [lapply()] is not always necessary,
#'   e.g. `lapply(DF, sum)` is the same as `lapply(DF, function(x) sum(x))` and
#'   the former is more readable.
#'
#' Cases like `lapply(x, \(xi) grep("ptn", xi))` are excluded because, though
#'   the anonymous function _can_ be avoided, doing so is not always more
#'   readable.
#'
#' @param allow_comparison Logical, default `FALSE`. If `TRUE`, lambdas like
#'   `function(x) foo(x) == 2`, where `foo` can be extracted to the "mapping"
#'   function and `==` vectorized instead of called repeatedly, are linted.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "lapply(list(1:3, 2:4), function(xi) sum(xi))",
#'   linters = unnecessary_lambda_linter()
#' )
#'
#' lint(
#'   text = "sapply(x, function(xi) xi == 2)",
#'   linters = unnecessary_lambda_linter()
#' )
#'
#' lint(
#'   text = "sapply(x, function(xi) sum(xi) > 0)",
#'   linters = unnecessary_lambda_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "lapply(list(1:3, 2:4), sum)",
#'   linters = unnecessary_lambda_linter()
#' )
#'
#' lint(
#'   text = 'lapply(x, function(xi) grep("ptn", xi))',
#'   linters = unnecessary_lambda_linter()
#' )
#'
#' lint(
#'   text = "lapply(x, function(xi) data.frame(col = xi))",
#'   linters = unnecessary_lambda_linter()
#' )
#'
#' lint(
#'   text = "sapply(x, function(xi) xi == 2)",
#'   linters = unnecessary_lambda_linter(allow_comparison = TRUE)
#' )
#'
#' lint(
#'   text = "sapply(x, function(xi) sum(xi) > 0)",
#'   linters = unnecessary_lambda_linter(allow_comparison = TRUE)
#' )
#'
#' lint(
#'   text = "sapply(x, function(xi) sum(abs(xi)) > 10)",
#'   linters = unnecessary_lambda_linter()
#' )
#'
#' lint(
#'   text = "sapply(x, sum) > 0",
#'   linters = unnecessary_lambda_linter()
#' )
#'
#' @evalRd rd_tags("unnecessary_lambda_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
unnecessary_lambda_linter <- function(allow_comparison = FALSE) {
  # include any base function like those where FUN is an argument
  #   and ... follows positionally directly afterwards (with ...
  #   being passed on to FUN). That excludes functions like
  #   Filter/Reduce (which don't accept ...), as well as functions
  #   like sweep() (where check.margin comes between FUN and ...,
  #   and thus would need to be supplied in order to replicate
  #   positional arguments). Technically, these latter could
  #   be part of the linter logic (e.g., detect if the anonymous
  #   call is using positional or keyword arguments -- we can
  #   throw a lint for sweep() lambdas where the following arguments
  #   are all named) but for now it seems like overkill.
  apply_funs <- c(
    "lapply", "sapply", "vapply", "apply",
    "tapply", "rapply", "eapply", "dendrapply",
    "mapply", "by", "outer",
    "mclapply", "mcmapply", "parApply", "parCapply", "parLapply",
    "parLapplyLB", "parRapply", "parSapply", "parSapplyLB", "pvec",
    purrr_mappers
  )

  # OP-PLUS: condition for complex literal, e.g. 0+2i.
  # NB: this includes 0+3 and TRUE+FALSE, which are also fine.
  inner_comparison_xpath <- glue("
  parent::expr
    /expr[FUNCTION or OP-LAMBDA]
    /expr[
      ({ xp_or(infix_metadata$xml_tag[infix_metadata$comparator]) })
      and expr[
        expr/SYMBOL_FUNCTION_CALL
        and expr/SYMBOL
      ]
      and expr[
        NUM_CONST
        or STR_CONST
        or (OP-PLUS and count(expr/NUM_CONST) = 2)
      ]
    ]
  ")

  # outline:
  #   1. match one of the identified mappers
  #   2. match an anonymous function that can be "symbol-ized"
  #     a. it's a one-variable function [TODO(#2477): relax this]
  #     b. the function is a single call
  #     c. that call's _first_ argument is just the function argument (a SYMBOL)
  #       - and it has to be passed positionally (not as a keyword)
  #     d. the function argument doesn't appear elsewhere in the call
  unary_op <- infix_metadata$xml_tag[infix_metadata$unary]
  default_fun_xpath <- glue("
  following-sibling::expr[(FUNCTION or OP-LAMBDA) and count(SYMBOL_FORMALS) = 1]
    /expr[last()][
      count(.//SYMBOL[self::* = preceding::SYMBOL_FORMALS[1]]) = 1
      and count(.//SYMBOL_FUNCTION_CALL[text() != 'return']) = 1
      and not(
        ({ xp_or(unary_op) })
        or (
          *[not(self::COMMENT)][1][self::OP-LEFT-PAREN or self::OP-LEFT-BRACE]
          and expr[{ xp_or(unary_op) }]
        )
      )
      and preceding-sibling::SYMBOL_FORMALS =
        .//expr[
          position() = 2
          and preceding-sibling::expr/SYMBOL_FUNCTION_CALL
          and not(preceding-sibling::*[not(self::COMMENT)][1][self::EQ_SUB])
          and not(parent::expr[
            preceding-sibling::expr[not(SYMBOL_FUNCTION_CALL)]
            or following-sibling::*[not(
              self::OP-RIGHT-PAREN
              or self::OP-RIGHT-BRACE
              or self::COMMENT
            )]
          ])
        ]/SYMBOL
    ]
    /parent::expr
  ")

  # purrr-style inline formulas-as-functions, e.g. ~foo(.x)
  # logic is basically the same as that above, except we need
  #   1. a formula (OP-TILDE)
  #   2. the lone argument marker `.x` or `.`
  purrr_symbol <- "SYMBOL[text() = '.x' or text() = '.']"
  purrr_fun_xpath <- glue("
  following-sibling::expr[
    OP-TILDE
    and expr
      /OP-LEFT-PAREN
      /following-sibling::expr[1][
        not(preceding-sibling::*[not(self::COMMENT)][2][self::SYMBOL_SUB])
      ]
      /{purrr_symbol}
    and not(expr/OP-LEFT-PAREN/following-sibling::expr[position() > 1]//{purrr_symbol})
  ]")

  # path to calling function symbol from the matched expressions
  fun_xpath <- "./parent::expr/expr/SYMBOL_FUNCTION_CALL"
  # path to the symbol of the simpler function that avoids a lambda
  symbol_xpath <- "expr[last()]//expr[SYMBOL_FUNCTION_CALL[text() != 'return']]"

  Linter(linter_level = "expression", function(source_expression) {
    default_calls <- source_expression$xml_find_function_calls(apply_funs)
    default_fun_expr <- xml_find_all_(default_calls, default_fun_xpath)

    # TODO(#2478): Give a specific recommendation in the message.
    default_call_fun <- xml_text(xml_find_first_(default_fun_expr, fun_xpath))
    default_symbol <- xml_text(xml_find_first_(default_fun_expr, symbol_xpath))
    default_fun_lints <- xml_nodes_to_lints(
      default_fun_expr,
      source_expression = source_expression,
      lint_message = paste0(
        "Pass ", default_symbol, " directly as a symbol to ", default_call_fun, "() ",
        "instead of wrapping it in an unnecessary anonymous function. ",
        "For example, prefer lapply(DF, sum) to lapply(DF, function(x) sum(x))."
      ),
      type = "warning"
    )

    inner_comparison_lints <- NULL
    if (!allow_comparison) {
      sapply_vapply_calls <- source_expression$xml_find_function_calls(c("sapply", "vapply"))
      inner_comparison_expr <- xml_find_all_(sapply_vapply_calls, inner_comparison_xpath)

      mapper <- xp_call_name(xml_find_first_(inner_comparison_expr, "parent::expr/parent::expr"))
      if (length(mapper) > 0L) fun_value <- if (mapper == "sapply") "" else ", FUN.VALUE = <intermediate>"

      inner_comparison_lints <- xml_nodes_to_lints(
        inner_comparison_expr,
        source_expression = source_expression,
        lint_message = sprintf(
          paste(
            "Compare to a constant after calling %1$s() to get the full benefits of vectorization.",
            "Prefer %1$s(x, foo%2$s) == 2 over %1$s(x, function(xi) foo(xi) == 2, logical(1L))."
          ),
          mapper, fun_value
        ),
        type = "warning"
      )
    }

    purrr_calls <- source_expression$xml_find_function_calls(purrr_mappers)
    purrr_fun_expr <- xml_find_all_(purrr_calls, purrr_fun_xpath)

    purrr_call_fun <- xml_text(xml_find_first_(purrr_fun_expr, fun_xpath))
    purrr_symbol <- xml_text(xml_find_first_(purrr_fun_expr, symbol_xpath))
    purrr_fun_lints <- xml_nodes_to_lints(
      purrr_fun_expr,
      source_expression = source_expression,
      lint_message = paste0(
        "Pass ", purrr_symbol, " directly as a symbol to ", purrr_call_fun, "() ",
        "instead of wrapping it in an unnecessary anonymous function. ",
        "For example, prefer purrr::map(DF, sum) to purrr::map(DF, ~sum(.x))."
      ),
      type = "warning"
    )

    c(default_fun_lints, inner_comparison_lints, purrr_fun_lints)
  })
}

#' Check that indentation is consistent
#'
#' @param indent Number of spaces, that a code block should be indented by relative to its parent code block.
#'   Used for multi-line code blocks (`{ ... }`), function calls (`( ... )`) and extractions (`[ ... ]`, `[[ ... ]]`).
#'   Defaults to 2.
#' @param hanging_indent_style Indentation style for multi-line function calls with arguments in their first line.
#'   Defaults to tidyverse style, i.e. a block indent is used if the function call terminates with `)` on a separate
#'   line and a hanging indent if not.
#'   Note that multi-line function calls without arguments on their first line will always be expected to have
#'   block-indented arguments.
#'
#'   ```r
#'   # complies to any style
#'   map(
#'     x,
#'     f,
#'     additional_arg = 42
#'   )
#'
#'   # complies to "tidy" and "never"
#'   map(x, f,
#'     additional_arg = 42
#'   )
#'
#'   # complies to "always"
#'   map(x, f,
#'       additional_arg = 42
#'   )
#'
#'   # complies to "tidy" and "always"
#'   map(x, f,
#'       additional_arg = 42)
#'
#'   # complies to "never"
#'   map(x, f,
#'     additional_arg = 42)
#'
#'   # complies to "tidy"
#'   function(
#'     a,
#'     b
#'   ) {
#'     # body
#'   }
#'   ```
#' @param assignment_as_infix Treat `<-` as a regular (i.e. left-associative) infix operator?
#'   This means, that infix operators on the right hand side of an assignment do not trigger a second level of
#'   indentation:
#'   ```r
#'   # complies to any style
#'   variable <- a %+%
#'     b %+%
#'     c
#'
#'   # complies to assignment_as_infix = TRUE
#'   variable <-
#'     a %+%
#'     b %+%
#'     c
#'
#'   # complies to assignment_as_infix = FALSE
#'   variable <-
#'     a %+%
#'       b %+%
#'       c
#'   ```
#'
#' @examples
#' # will produce lints
#' code_lines <- "if (TRUE) {\n1 + 1\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = indentation_linter()
#' )
#'
#' code_lines <- "if (TRUE) {\n    1 + 1\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = indentation_linter()
#' )
#'
#' code_lines <- "map(x, f,\n  additional_arg = 42\n)"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = indentation_linter(hanging_indent_style = "always")
#' )
#'
#' code_lines <- "map(x, f,\n    additional_arg = 42)"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = indentation_linter(hanging_indent_style = "never")
#' )
#'
#' # okay
#' code_lines <- "map(x, f,\n  additional_arg = 42\n)"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = indentation_linter()
#' )
#'
#' code_lines <- "if (TRUE) {\n    1 + 1\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = indentation_linter(indent = 4)
#' )
#'
#' @evalRd rd_tags("indentation_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#indenting>
#' - <https://style.tidyverse.org/functions.html#long-lines-1>
#'
#' @export
indentation_linter <- function(indent = 2L, hanging_indent_style = c("tidy", "always", "never"),
                               assignment_as_infix = TRUE) {
  paren_tokens_left_no_brace <- c("OP-LEFT-PAREN", "OP-LEFT-BRACKET", "LBB")
  paren_tokens_left <- c("OP-LEFT-BRACE", paren_tokens_left_no_brace)
  paren_tokens_right <- c("OP-RIGHT-BRACE", "OP-RIGHT-PAREN", "OP-RIGHT-BRACKET", "OP-RIGHT-BRACKET")
  infix_tokens <- setdiff(infix_metadata$xml_tag, c("OP-LEFT-BRACE", "OP-COMMA", paren_tokens_left))
  no_paren_keywords <- c("ELSE", "REPEAT")
  keyword_tokens <- c("FUNCTION", "OP-LAMBDA", "IF", "WHILE")

  xp_last_on_line <- "@line1 != following-sibling::*[not(self::COMMENT)][1]/@line1"

  hanging_indent_style <- match.arg(hanging_indent_style)

  find_indent_type <- switch(hanging_indent_style,
    tidy = build_indentation_style_tidy(),
    always = build_indentation_style_always(),
    never = \(change) "block"
  )

  if (isTRUE(assignment_as_infix)) {
    suppressing_tokens <- c("LEFT_ASSIGN", "EQ_ASSIGN", "EQ_SUB", "EQ_FORMALS")
    xp_suppress <- glue("preceding-sibling::{suppressing_tokens}[{xp_last_on_line}]")

    restoring_tokens <- c("expr[SYMBOL_FUNCTION_CALL]", "OP-LEFT-BRACE")
    xp_restore <- glue("preceding-sibling::{restoring_tokens}")

    # match the first ancestor expr that is either
    #  * a suppressing token (<- or =) or
    #  * a restoring token (braces or a function call)
    # suppress the indent if the matched ancestor is a suppressing token
    infix_condition <- glue("
      and not(ancestor::expr[{xp_or(c(xp_suppress, xp_restore))}][1][{xp_or(xp_suppress)}])
    ")
  } else {
    infix_condition <- ""
  }

  xp_block_ends <- paste0(
    "number(",
    paste(
      c(
        glue("self::{paren_tokens_left}/following-sibling::{paren_tokens_right}/preceding-sibling::*[1]/@line2"),
        glue("
          self::*[{xp_and(paste0('not(self::', paren_tokens_left, ')'))}]
            /following-sibling::*[
              self::SYMBOL_FUNCTION_CALL
              or self::SLOT[parent::expr/following-sibling::OP-LEFT-PAREN]
            ]
            /parent::expr
            /following-sibling::expr[1]
            /@line2
        "),
        glue("
          self::*[
            {xp_and(paste0('not(self::', paren_tokens_left, ')'))}
            and not(following-sibling::*[
              self::SYMBOL_FUNCTION_CALL
              or self::SLOT[parent::expr/following-sibling::OP-LEFT-PAREN]
            ])
          ]
            /following-sibling::*[not(self::COMMENT)][1]
            /@line2
        ")
      ),
      collapse = " | "
    ),
    ")"
  )

  global_nodes <- function(nodes) paste0("//", nodes, collapse = "|")
  xp_indent_changes <- paste(
    c(
      glue("//{paren_tokens_left}[not(
        @line1 = following-sibling::expr[
          @line2 > @line1 and
          ({xp_or(paste0('descendant::', paren_tokens_left, '[', xp_last_on_line, ']'))})
        ]/@line1
      )]"),
      glue("({ global_nodes(infix_tokens) })[{xp_last_on_line}{infix_condition}]"),
      glue("({ global_nodes(no_paren_keywords) })[{xp_last_on_line}]"),
      glue("
        ({ global_nodes(keyword_tokens) })
          /following-sibling::OP-RIGHT-PAREN[
            {xp_last_on_line} and
            not(following-sibling::expr[1][OP-LEFT-BRACE])
          ]
      "),
      # FOR loop is strange, #2564
      glue("
        //forcond[
          OP-RIGHT-PAREN/{xp_last_on_line}
          and not(following-sibling::expr[1]/OP-LEFT-BRACE)
        ]
      ")
    ),
    collapse = " | "
  )

  xp_multiline_string <- "//STR_CONST[@line1 < @line2]"
  build_line_metadata <- function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    indent_changes <- xml_find_all_(xml, xp_indent_changes)
    change_tags <- xml_name_(indent_changes)
    change_types <- vapply(indent_changes, find_indent_type, character(1L))
    change_begins <- as.integer(xml_attr_(indent_changes, "line1")) + 1L
    change_ends <- xml_find_num_(indent_changes, xp_block_ends)
    col2s <- as.integer(xml_attr_(indent_changes, "col2"))

    check_idx <- which(change_begins <= change_ends)

    line_metadata <- data.frame(
      line = source_expression$file_lines,
      number = as.integer(names(source_expression$file_lines))
    )
    line_metadata$indent_level <- re_matches(line_metadata$line, rex(start, any_spaces), locations = TRUE)[, "end"]
    line_metadata$in_str_const <- logical(nrow(line_metadata)) # FALSE, but also logical() for 0-row case

    multiline_strings <- xml_find_all_(xml, xp_multiline_string)
    string_line1 <- as.integer(xml_attr_(multiline_strings, "line1"))
    string_line2 <- as.integer(xml_attr_(multiline_strings, "line2"))

    line_metadata$in_str_const[unlist(Map(`:`, string_line1 + 1L, string_line2))] <- TRUE

    line_metadata$expected_level <- integer(nrow(line_metadata))
    line_metadata$is_hanging <- logical(nrow(line_metadata))
    line_metadata$hanging_cols <- integer(nrow(line_metadata))

    for (ii in check_idx) {
      to_indent <- seq(change_begins[ii], change_ends[ii])
      line_metadata$expected_level[to_indent] <- find_new_indent(
        current_indent = line_metadata$expected_level[to_indent],
        change_type = change_types[ii],
        indent = indent,
        hanging_indent = col2s[ii]
      )
      line_metadata$is_hanging[to_indent] <- change_types[ii] == "hanging"
      is_paren_block <- change_types[ii] == "block" && change_tags[ii] %in% paren_tokens_left_no_brace
      line_metadata$hanging_cols[to_indent] <- if (is_paren_block) col2s[ii] else 0L
    }

    # Only lint non-empty lines if the indentation level doesn't match.
    line_metadata$is_bad <- find_bad_lines(line_metadata)

    line_metadata[line_metadata$is_bad, ]
  }

  Linter(linter_level = "file", function(source_expression) {
    # must run on file level because a line can contain multiple expressions, losing indentation information, e.g.
    #
    #> fun(
    #    a) # comment
    #
    # will have "# comment" as a separate expression

    # Indentation increases by 1 for:
    #  - { } blocks that span multiple lines
    #  - ( ), [ ], or [[ ]] calls that span multiple lines
    #     + if a token follows (, a hanging indent is required until )
    #     + if there is no token following ( on the same line, a block indent is required until )
    #  - binary operators where the second arguments starts on a new line

    lint_line_df <- build_line_metadata(source_expression)

    if (nrow(lint_line_df) == 0L) {
      return(list())
    }

    lint_metadata <- indent_lint_metadata(lint_line_df)

    Map(
      Lint,
      filename = source_expression$filename,
      line_number = lint_metadata$line_numbers,
      column_number = lint_metadata$column_numbers,
      type = "style",
      message = lint_metadata$lint_messages,
      line = lint_metadata$lines,
      ranges = lint_metadata$ranges
    )
  })
}

find_new_indent <- function(current_indent, change_type, indent, hanging_indent) {
  switch(change_type,
    suppress = current_indent,
    hanging = hanging_indent,
    block = current_indent + indent
  )
}

find_bad_lines <- function(line_metadata) {
  is_bad <-
    with(line_metadata, !is.na(line) & indent_level != expected_level & nzchar(trimws(line, "left")) & !in_str_const)
  if (!any(is_bad)) {
    return(is_bad)
  }

  # Suppress consecutive lints with the same indentation difference, to not generate an excessive number of lints
  indent_diff <- line_metadata$expected_level - line_metadata$indent_level
  indent_diff[is.na(indent_diff)] <- 0L

  is_consecutive <- c(FALSE, diff(is_bad) == 0L)
  is_same_diff <- c(FALSE, diff(indent_diff) == 0L)

  is_bad & !(is_consecutive & is_same_diff)
}

indent_lint_metadata <- function(line_metadata) {
  lint_messages <- with(line_metadata, sprintf(
    "%s should be %d spaces but is %d spaces%s.",
    ifelse(is_hanging, "Hanging indent", "Indentation"),
    expected_level,
    indent_level,
    ifelse(hanging_cols > 0L & indent_level == hanging_cols, " (or start argument on previous line)", "")
  ))

  ranges <- with(line_metadata, Map(\(x, y) list(c(x, y)),
    # when indent_levels==0, need to start ranges at column 1.
    pmax(pmin(expected_level + 1L, indent_level), 1L),
    # If the expected indent is larger than the current line width, the lint range would become invalid.
    # Therefore, limit range end to end of line.
    pmin(pmax(expected_level, indent_level), nchar(line) + 1L)
  ))

  list(
    lines = line_metadata$line,
    line_numbers = line_metadata$number,
    column_numbers = line_metadata$indent_level,
    lint_messages = lint_messages,
    ranges = ranges
  )
}

build_indentation_style_tidy <- function() {
  paren_tokens_left <- c("OP-LEFT-BRACE", "OP-LEFT-PAREN", "OP-LEFT-BRACKET", "LBB")
  paren_tokens_right <- c("OP-RIGHT-BRACE", "OP-RIGHT-PAREN", "OP-RIGHT-BRACKET", "OP-RIGHT-BRACKET")
  xp_last_on_line <- "@line1 != following-sibling::*[not(self::COMMENT)][1]/@line1"
  xp_inner_expr <- "preceding-sibling::*[1][self::expr and expr[SYMBOL_FUNCTION_CALL]]/*[not(self::COMMENT)]"

  xp_suppress <- paste(
    glue("
        self::{paren_tokens_left}[
          @line1 = following-sibling::{paren_tokens_right}/{xp_inner_expr}[position() = 1]/@line1
        ]/following-sibling::{paren_tokens_right}[
          @line1 > {xp_inner_expr}[position() = last() - 1]/@line2
        ]"),
    collapse = " | "
  )

  xp_is_not_hanging <- paste(
    c(
      glue("
        self::{paren_tokens_left}
          /following-sibling::{paren_tokens_right}[@line1 > preceding-sibling::*[1]/@line2]
      "),
      glue("self::*[{xp_and(paste0('not(self::', paren_tokens_left, ')'))} and {xp_last_on_line}]"),
      glue("self::{paren_tokens_left}[parent::expr[FUNCTION or OP-LAMBDA] and {xp_last_on_line}]")
    ),
    collapse = "\n|  "
  )

  function(change) {
    if (length(xml_find_first_(change, xp_suppress)) > 0L) {
      "suppress"
    } else if (length(xml_find_first_(change, xp_is_not_hanging)) == 0L) {
      "hanging"
    } else {
      "block"
    }
  }
}

build_indentation_style_always <- function() {
  paren_tokens_left <- c("OP-LEFT-BRACE", "OP-LEFT-PAREN", "OP-LEFT-BRACKET", "LBB")
  paren_tokens_right <- c("OP-RIGHT-BRACE", "OP-RIGHT-PAREN", "OP-RIGHT-BRACKET", "OP-RIGHT-BRACKET")
  xp_last_on_line <- "@line1 != following-sibling::*[not(self::COMMENT)][1]/@line1"

  xp_is_not_hanging <- paste(
    c(
      glue("
        self::{paren_tokens_left}[{xp_last_on_line}]/
          following-sibling::{paren_tokens_right}[@line1 > preceding-sibling::*[1]/@line2]
      "),
      glue("self::*[{xp_and(paste0('not(self::', paren_tokens_left, ')'))} and {xp_last_on_line}]")
    ),
    collapse = " | "
  )

  function(change) {
    if (length(xml_find_first_(change, xp_is_not_hanging)) == 0L) {
      "hanging"
    } else {
      "block"
    }
  }
}

#' Object usage linter
#'
#' Check that closures have the proper usage using [codetools::checkUsage()].
#' Note that this runs [base::eval()] on the code, so **do not use with untrusted code**.
#'
#' @param interpret_glue (Defunct)
#' @param interpret_extensions Character vector of extensions to interpret. These are meant to cover known cases where
#'   variables may be used in ways understood by the reader but not by `checkUsage()` to avoid false positives.
#'   Currently `"glue"` and `"rlang"` are supported, both of which are in the default.
#'   - For `glue`, examine [glue::glue()] calls.
#'   - For `rlang`, examine `.env$key` usages.
#' @param skip_with A logical. If `TRUE` (default), code in `with()` expressions
#'   will be skipped. This argument will be passed to `skipWith` argument of
#'   `codetools::checkUsage()`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "foo <- function() { x <- 1 }",
#'   linters = object_usage_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "foo <- function(x) { x <- 1 }",
#'   linters = object_usage_linter()
#' )
#'
#' lint(
#'   text = "foo <- function() { x <- 1; return(x) }",
#'   linters = object_usage_linter()
#' )
#' @evalRd rd_linters("package_development")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
object_usage_linter <- function(interpret_glue = NULL, interpret_extensions = c("glue", "rlang"), skip_with = TRUE) {
  if (!is.null(interpret_glue)) {
    lintr_deprecated(
      "interpret_glue",
      '"glue" in interpret_extensions',
      version = "3.3.0",
      type = "Argument",
      signal = "stop"
    )
  }

  if (length(interpret_extensions) > 0L) {
    interpret_extensions <- match.arg(interpret_extensions, several.ok = TRUE)
  }

  # NB: difference across R versions in how EQ_ASSIGN is represented in the AST
  #   (under <expr_or_assign_or_help> or <equal_assign>)
  # NB: the repeated expr[2][FUNCTION] XPath has no performance impact, so the different direct assignment XPaths are
  #   split for better readability, see PR#1197
  # TODO(#1106): use //[...] to capture assignments in more scopes
  fun_node <- "FUNCTION or OP-LAMBDA"
  xpath_function_assignment <- glue("
    expr[LEFT_ASSIGN or EQ_ASSIGN]/expr[2][{fun_node}]
    | expr_or_assign_or_help[EQ_ASSIGN]/expr[2][{fun_node}]
    | equal_assign[EQ_ASSIGN]/expr[2][{fun_node}]
    | //SYMBOL_FUNCTION_CALL[text() = 'assign']/parent::expr/following-sibling::expr[2][{fun_node}]
    | //SYMBOL_FUNCTION_CALL[text() = 'setMethod']/parent::expr/following-sibling::expr[3][{fun_node}]
  ")

  # code like:
  #   foo <- \ #comment
  #     (x) x
  # is technically valid, but won't parse unless the lambda is in a bigger expression (here '<-').
  #   the same doesn't apply to 'function', which is acknowledged as "not worth a breaking change to fix":
  #   https://bugs.r-project.org/show_bug.cgi?id=18924. If we find such code (which has only ever
  #   arisen in content fuzzing where we inject comments at random to the AST), we have to avoid parsing
  #   it as a standalone expression.
  xpath_unsafe_lambda <- "OP-LAMBDA[@line1 = following-sibling::*[1][self::COMMENT]/@line1]"

  # not all instances of linted symbols are potential sources for the observed violations -- see #1914
  symbol_exclude_cond <- "preceding-sibling::OP-DOLLAR or preceding-sibling::OP-AT or ancestor::expr[OP-TILDE]"
  xpath_culprit_symbol <- glue("
    descendant::SYMBOL[not( {symbol_exclude_cond} )]
    | descendant::SYMBOL_FUNCTION_CALL[not( {symbol_exclude_cond} )]
    | descendant::SPECIAL
    | descendant::LEFT_ASSIGN[text() = ':=']
  ")

  Linter(linter_level = "file", function(source_expression) {
    pkg_name <- pkg_name(find_package(dirname(source_expression$filename)))

    declared_globals <- try_silently(globalVariables(package = pkg_name %||% globalenv()))

    xml <- source_expression$full_xml_parsed_content

    # Catch missing packages and report them as lints
    outer_env <- new.env(parent = emptyenv())
    outer_env$library_lints <- list()
    library_lint_hook <- function(lint_node, lint_msg) {
      outer_env$library_lints[[length(outer_env$library_lints) + 1L]] <- xml_nodes_to_lints(
        lint_node, source_expression = source_expression, lint_message = lint_msg, type = "warning"
      )
    }

    # run the following at run-time, not "compile" time to allow package structure to change
    env <- make_check_env(pkg_name, xml, library_lint_hook)

    fun_assignments <- xml_find_all_(xml, xpath_function_assignment)

    lapply(fun_assignments, function(fun_assignment) {
      # this will mess with the source line numbers. but I don't think anybody cares.
      needs_braces <- !is.na(xml_find_first_(fun_assignment, xpath_unsafe_lambda))
      code <- get_content(lines = source_expression$content, fun_assignment, needs_braces = needs_braces)
      fun <- try_silently(eval(
        envir = env,
        parse(
          text = code,
          keep.source = TRUE
        )
      ))

      if (inherits(fun, "try-error")) {
        return()
      }
      known_used_symbols <- known_used_symbols(fun_assignment, interpret_extensions = interpret_extensions)
      res <- parse_check_usage(
        fun,
        known_used_symbols = known_used_symbols,
        declared_globals = declared_globals,
        start_line = as.integer(xml_attr_(fun_assignment, "line1")),
        end_line = as.integer(xml_attr_(fun_assignment, "line2")),
        skip_with = skip_with
      )

      res$name <- re_substitutes(res$name, rex("<-"), "")

      lintable_symbols <- xml_find_all_(fun_assignment, xpath_culprit_symbol)

      lintable_symbol_names <- gsub("^`|`$", "", xml_text(lintable_symbols))
      lintable_symbol_lines <- as.integer(xml_attr_(lintable_symbols, "line1"))

      matched_symbol <- vapply(
        seq_len(nrow(res)),
        function(i) {
          match(
            TRUE,
            lintable_symbol_names == res$name[i] &
              lintable_symbol_lines >= res$line1[i] &
              lintable_symbol_lines <= res$line2[i]
          )
        },
        integer(1L)
      )
      nodes <- unclass(lintable_symbols)[matched_symbol]

      # fallback to line based matching if no symbol is found
      missing_symbol <- is.na(matched_symbol)
      nodes[missing_symbol] <- lapply(which(missing_symbol), function(i) {
        line_based_match <- xml_find_first_(
          fun_assignment,
          glue::glue_data(res[i, ], "descendant::expr[@line1 = {line1} and @line2 = {line2}]")
        )
        if (is.na(line_based_match)) fun_assignment else line_based_match
      })

      c(
        outer_env$library_lints,
        xml_nodes_to_lints(nodes, source_expression = source_expression, lint_message = res$message, type = "warning")
      )
    })
  })
}

make_check_env <- function(pkg_name, xml, library_lint_hook) {
  if (!is.null(pkg_name)) {
    parent_env <- try_silently(getNamespace(pkg_name))
  }
  if (is.null(pkg_name) || inherits(parent_env, "try-error")) {
    parent_env <- globalenv()
  }
  env <- new.env(parent = parent_env)

  symbols <- c(
    get_assignment_symbols(xml),
    get_imported_symbols(xml, library_lint_hook)
  )

  # Just assign them an empty function
  for (symbol in symbols) {
    assign(symbol, \(...) invisible(), envir = env)
  }
  env
}

get_assignment_symbols <- function(xml) {
  get_r_string(xml_find_all_(
    xml,
    "
      expr[LEFT_ASSIGN or EQ_ASSIGN]/expr[1]/SYMBOL[1] |
      expr[RIGHT_ASSIGN]/expr[2]/SYMBOL[1] |
      equal_assign/expr[1]/SYMBOL[1] |
      expr_or_assign_or_help/expr[1]/SYMBOL[1] |
      expr[expr[1][SYMBOL_FUNCTION_CALL/text() = 'assign']]/expr[2]/* |
      expr[expr[1][SYMBOL_FUNCTION_CALL/text() = 'setMethod']]/expr[2]/*
    "
  ))
}

get_check_usage_results <- function(expression, known_used_symbols, declared_globals, skip_with) {
  report_env <- new.env(parent = emptyenv())
  report_env$vals <- character()
  report <- function(x) report_env$vals <- c(report_env$vals, x)
  old <- options(useFancyQuotes = FALSE)
  on.exit(options(old))
  try(
    codetools::checkUsage(
      expression,
      report = report,
      suppressLocalUnused = known_used_symbols,
      suppressUndefined = declared_globals,
      skipWith = skip_with
    )
  )
  report_env$vals
}

parse_check_usage <- function(expression,
                              known_used_symbols = character(),
                              declared_globals = character(),
                              start_line = 1L,
                              end_line = 1L,
                              skip_with = TRUE) {
  vals <- get_check_usage_results(expression, known_used_symbols, declared_globals, skip_with)

  function_name <- rex(anything, ": ")
  line_info <- rex(
    " ", "(", capture(name = "path", non_spaces), ":",
    capture(name = "line1", digits), maybe("-", capture(name = "line2", digits)), ")"
  )

  res <- re_matches(
    vals,
    rex(
      function_name,
      capture(
        name = "message",
        zero_or_more(any, type = "lazy"),
        maybe(
          "'",
          capture(name = "name", anything),
          "'",
          zero_or_more(any, type = "lazy")
        )
      ),
      or(line_info, end)
    )
  )

  res$line1 <- ifelse(
    nzchar(res$line1),
    as.integer(res$line1) + start_line - 1L,
    NA_integer_
  )
  res$line2 <- ifelse(
    nzchar(res$line2),
    as.integer(res$line2) + start_line - 1L,
    res$line1
  )

  res$line1[is.na(res$line1)] <- start_line
  res$line2[is.na(res$line2)] <- end_line

  res
}

get_imported_symbols <- function(xml, library_lint_hook) {
  import_exprs_xpath <- "
  //SYMBOL_FUNCTION_CALL[text() = 'library' or text() = 'require']
    /parent::expr
    /parent::expr[
      not(SYMBOL_SUB[
        text() = 'character.only' and
        following-sibling::expr[1][NUM_CONST[text() = 'TRUE'] or SYMBOL[text() = 'T']]
      ])
      or expr[2][STR_CONST]
    ]
    /expr[STR_CONST or SYMBOL][1]
  "
  import_exprs <- xml_find_all_(xml, import_exprs_xpath)
  imported_pkgs <- get_r_string(import_exprs)

  # nolint next: undesirable_function_name_linter.
  unlist(Map(pkg = imported_pkgs, expr = xml_parent(import_exprs), function(pkg, expr) {
    tryCatch(
      getNamespaceExports(pkg),
      error = function(e) {
        lib_paths <- .libPaths() # nolint: undesirable_function_name. .libPaths() is necessary here.
        lib_noun <- if (length(lib_paths) == 1L) "library" else "libraries"
        lint_msg <- paste0(
          "Could not find exported symbols for package \"", pkg, "\" in ", lib_noun, " ",
          toString(shQuote(lib_paths)), " (", conditionMessage(e), "). This may lead to false positives."
        )
        library_lint_hook(expr, lint_msg)
        character()
      }
    )
  }))
}

known_used_symbols <- function(fun_assignment, interpret_extensions) {
  unique(c(
    if ("rlang" %in% interpret_extensions) extract_env_symbols(fun_assignment),
    if ("glue" %in% interpret_extensions) extract_glued_symbols(fun_assignment)
  ))
}

extract_env_symbols <- function(fun_assignment) {
  env_names <- xml_find_all_(
    fun_assignment,
    "
    .//SYMBOL[text() = '.env']
      /parent::expr
      /following-sibling::OP-DOLLAR
      /following-sibling::SYMBOL
    | .//SYMBOL[text() = '.env']
      /parent::expr
      /following-sibling::LBB
      /following-sibling::expr[STR_CONST]
    "
  )
  get_r_string(env_names)
}

downstream <- function(x) {
  a_really_really_long_local_object_name <- 1
  aSilLLyObjEctNaME <- 2

  myFancyUpstream::upstreams_really_long_imported_object_name()
  myFancyUpstream::uPStreams_SilLY.objecT.Name()
}

# importFrom() functions in unknown namespace
a_veritably_dilettantish_function.my_class <- function(x, ...) x
sPoNgEbOb_cAsE.FuNcTiOn.my_class <- function(x, ...) x

#' Whitespace linter
#'
#' Check that the correct character is used for indentation.
#'
#' Currently, only supports linting in the presence of tabs.
#'
#' Much ink has been spilled on this topic, and we encourage you to check
#'   out references for more information.
#'
#' @include make_linter_from_regex.R
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "\tx",
#'   linters = whitespace_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "  x",
#'   linters = whitespace_linter()
#' )
#'
#' @evalRd rd_tags("whitespace_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#'
#' @references
#' - https://www.jwz.org/doc/tabs-vs-spaces.html
#' - https://blog.codinghorror.com/death-to-the-space-infidels/
#' @export
whitespace_linter <- make_linter_from_regex(
  regex = rex(start, zero_or_more(regex("\\s")), one_or_more("\t")),
  lint_type = "style",
  lint_msg = "Use spaces to indent, not tabs."
)

#' Require usage of direct methods for subsetting strings via regex
#'
# TODO(R>=4.5.0): Just use [grepv()] directly. Need this while ?grepv doesn't exist.
#' Using [`grepv()`][grep] returns the subset of the input that matches the pattern,
#'   e.g. `grepv("[a-m]", letters)` will return the first 13 elements (`a` through `m`).
#'
#' `letters[grep("[a-m]", letters)]` and `letters[grepl("[a-m]", letters)]`
#'   both return the same thing, but more circuitously and more verbosely.
#'
#' The `stringr` package also provides an even more readable alternative,
#'   namely `str_subset()`, which should be preferred to versions using
#'   `str_detect()` and `str_which()`.
#'
#' @section Exceptions:
#'   Note that `x[grep(pattern, x)]` and `grepv(pattern, x)` are not
#'    _completely_ interchangeable when `x` is not character (most commonly,
#'   when `x` is a factor), because the output of the latter will be a
#'   character vector while the former remains a factor. It still may be
#'   preferable to refactor such code, as it may be faster to match the
#'   pattern on `levels(x)` and use that to subset instead.
#'
#' @evalRd rd_tags("regex_subset_linter")
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x[grep(pattern, x)]",
#'   linters = regex_subset_linter()
#' )
#'
#' lint(
#'   text = "x[stringr::str_which(x, pattern)]",
#'   linters = regex_subset_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "grepv(pattern, x)",
#'   linters = regex_subset_linter()
#' )
#'
#' lint(
#'   text = "stringr::str_subset(x, pattern)",
#'   linters = regex_subset_linter()
#' )
#'
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
regex_subset_linter <- function() {
  xpath_fmt <- "
  self::*[
    not(LEFT_ASSIGN or EQ_ASSIGN or RIGHT_ASSIGN)
  ]
    /expr[
      OP-LEFT-BRACKET
      and expr[1] = expr/expr[position() = {arg_pos} ]
    ]
  "
  grep_xpath <- glue(xpath_fmt, arg_pos = 3L)
  stringr_xpath <- glue(xpath_fmt, arg_pos = 2L)

  Linter(linter_level = "expression", function(source_expression) {
    grep_calls <- xml_find_all_(
      source_expression$xml_find_function_calls(c("grepl", "grep")),
      "parent::*/parent::*/parent::*"
    )
    grep_calls <- strip_comments_from_subtree(grep_calls)
    grep_expr <- xml_find_all_(grep_calls, grep_xpath)

    grep_lints <- xml_nodes_to_lints(
      grep_expr,
      source_expression = source_expression,
      lint_message = paste(
        "Prefer grepv(pattern, x, ...) over x[grep(pattern, x, ...)] and x[grepl(pattern, x, ...)].",
        "Code required to run on R versions before 4.5.0 can use grep(pattern, x, ..., value = TRUE)."
      ),
      type = "warning"
    )

    stringr_calls <- xml_find_all_(
      source_expression$xml_find_function_calls(c("str_detect", "str_which")),
      "parent::*/parent::*/parent::*"
    )
    stringr_calls <- strip_comments_from_subtree(stringr_calls)
    stringr_expr <- xml_find_all_(stringr_calls, stringr_xpath)

    stringr_lints <- xml_nodes_to_lints(
      stringr_expr,
      source_expression = source_expression,
      lint_message =
        "Prefer stringr::str_subset(x, pattern) over x[str_detect(x, pattern)] and x[str_which(x, pattern)].",
      type = "warning"
    )

    c(grep_lints, stringr_lints)
  })
}

#' Is this an expression- or a file-level source object?
#'
#' Helper for determining whether the current `source_expression` contains
#'   all expressions in the current file, or just a single expression.
#'
#' @param source_expression A parsed expression object, i.e., an element
#'   of the object returned by [get_source_expressions()].
#' @param level Which level of expression is being tested? `"expression"`
#'   means an individual expression, while `"file"` means all expressions
#'   in the current file are available.
#'
#' @examples
#' tmp <- tempfile()
#' writeLines(c("x <- 1", "y <- x + 1"), tmp)
#' source_exprs <- get_source_expressions(tmp)
#' is_lint_level(source_exprs$expressions[[1L]], level = "expression")
#' is_lint_level(source_exprs$expressions[[1L]], level = "file")
#' is_lint_level(source_exprs$expressions[[3L]], level = "expression")
#' is_lint_level(source_exprs$expressions[[3L]], level = "file")
#' unlink(tmp)
#'
#' @export
is_lint_level <- function(source_expression, level = c("expression", "file")) {
  level <- match.arg(level)
  required_key <- switch(level, file = "full_parsed_content", expression = "parsed_content")
  required_key %in% names(source_expression)
}

#' Determine whether a linter needs to run for a given source_expression level
#'
#' Used by [lint()] to avoid unnecessary calls to linters.
#'
#' @param linter A linter.
#' @param level Which level of expression is being tested? `"expression"`
#'   means an individual expression, while `"file"` means all expressions
#'   in the current file are available.
#'
#' @keywords internal
#' @noRd
is_linter_level <- function(linter, level = c("expression", "file")) {
  linter_level <- attr(linter, "linter_level", exact = TRUE)
  if (is.null(linter_level) || is.na(linter_level)) {
    return(TRUE)
  }
  level <- match.arg(level)
  identical(linter_level, level)
}

test_that("is_lint_level works", {
  pc <- list(parsed_content = 1L)
  expect_true(is_lint_level(pc, "expression"))
  expect_false(is_lint_level(pc, "file"))

  names(pc) <- "full_parsed_content"
  expect_true(is_lint_level(pc, "file"))
  expect_false(is_lint_level(pc, "expression"))
})

#' Block usage of pipes nested inside other calls
#'
#' Nesting pipes harms readability; extract sub-steps to separate variables,
#'   append further pipeline steps, or otherwise refactor such usage away.
#'
#' @param allow_inline Logical, default `TRUE`, in which case only "inner"
#'   pipelines which span more than one line are linted. If `FALSE`, even
#'   "inner" pipelines that fit in one line are linted.
#' @param allow_outer_calls Character vector dictating which "outer"
#'   calls to exempt from the requirement to unnest (see examples). Defaults
#'   to [try()], [tryCatch()], and [withCallingHandlers()].
#'
#' @examples
#' # will produce lints
#' code <- "df1 %>%\n  inner_join(df2 %>%\n    select(a, b)\n  )"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = nested_pipe_linter()
#' )
#'
#' lint(
#'   text = "df1 %>% inner_join(df2 %>% select(a, b))",
#'   linters = nested_pipe_linter(allow_inline = FALSE)
#' )
#'
#' lint(
#'   text = "tryCatch(x %>% filter(grp == 'a'), error = identity)",
#'   linters = nested_pipe_linter(allow_outer_calls = character())
#' )
#'
#' # okay
#' lint(
#'   text = "df1 %>% inner_join(df2 %>% select(a, b))",
#'   linters = nested_pipe_linter()
#' )
#'
#' code <- "df1 %>%\n  inner_join(df2 %>%\n    select(a, b)\n  )"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = nested_pipe_linter(allow_outer_calls = "inner_join")
#' )
#'
#' lint(
#'   text = "tryCatch(x %>% filter(grp == 'a'), error = identity)",
#'   linters = nested_pipe_linter()
#' )
#'
#' @evalRd rd_tags("nested_pipe_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
nested_pipe_linter <- function(
  allow_inline = TRUE,
  allow_outer_calls = c("try", "tryCatch", "withCallingHandlers")
) {
  multiline_and <- if (allow_inline) "@line1 != @line2 and" else ""
  xpath <- glue("
  (//PIPE | //SPECIAL[{ xp_text_in_table(magrittr_pipes) }])
    /parent::expr[{multiline_and} preceding-sibling::expr/SYMBOL_FUNCTION_CALL[
      not({ xp_text_in_table(allow_outer_calls) })
      and (
        text() != 'switch'
        or parent::expr
          /following-sibling::expr[1]
          /*[self::PIPE or self::SPECIAL[{ xp_text_in_table(magrittr_pipes) }]]
      )
    ]]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = "Don't nest pipes inside other calls.",
      type = "warning"
    )
  })
}

#' Library call linter
#'
#' This linter covers several rules related to [library()] calls:
#'  - Enforce such calls to all be at the top of the script.
#'  - Block usage of argument `character.only`, in particular
#'    for loading packages in a loop.
#'  - Block consecutive calls to `suppressMessages(library(.))`
#'    in favor of using [suppressMessages()] only once to suppress
#'    messages from all `library()` calls. Ditto [suppressPackageStartupMessages()].
#'
#' @param allow_preamble Logical, default `TRUE`. If `FALSE`,
#'   no code is allowed to precede the first `library()` call,
#'   otherwise some setup code is allowed, but all `library()`
#'   calls must follow consecutively after the first one.
#' @examples
#' # will produce lints
#'
#' code <- "library(dplyr)\nprint('test')\nlibrary(tidyr)"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = library_call_linter()
#' )
#'
#' lint(
#'   text = "library('dplyr', character.only = TRUE)",
#'   linters = library_call_linter()
#' )
#'
#' code <- paste(
#'   "pkg <- c('dplyr', 'tibble')",
#'   "sapply(pkg, library, character.only = TRUE)",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = library_call_linter()
#' )
#'
#' code <- "suppressMessages(library(dplyr))\nsuppressMessages(library(tidyr))"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = library_call_linter()
#' )
#'
#' # okay
#' code <- "library(dplyr)\nprint('test')"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = library_call_linter()
#' )
#'
#' code <- "# comment\nlibrary(dplyr)"
#' lint(
#'   text = code,
#'   linters = library_call_linter()
#' )
#'
#' code <- paste(
#'   "foo <- function(pkg) {",
#'   "  sapply(pkg, library, character.only = TRUE)",
#'   "}",
#'   sep = "\n"
#' )
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = library_call_linter()
#' )
#'
#' code <- "suppressMessages({\n  library(dplyr)\n  library(tidyr)\n})"
#' writeLines(code)
#' lint(
#'   text = code,
#'   linters = library_call_linter()
#' )
#'
#' @evalRd rd_tags("library_call_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
library_call_linter <- function(allow_preamble = TRUE) {
  attach_calls <- c("library", "require")
  attach_call_cond <- xp_text_in_table(attach_calls)
  suppress_call_cond <- xp_text_in_table(c("suppressMessages", "suppressPackageStartupMessages"))

  unsuppressed_call_cond <- glue("not( {xp_or(attach_call_cond, suppress_call_cond)} )")
  if (allow_preamble) {
    unsuppressed_call_cond <- xp_and(
      unsuppressed_call_cond,
      glue("@line1 > //SYMBOL_FUNCTION_CALL[{ attach_call_cond }][1]/@line1")
    )
  }
  upfront_call_xpath <- glue("
    //SYMBOL_FUNCTION_CALL[{ attach_call_cond }][last()]
      /preceding::expr
      /*[
        (self::SYMBOL_FUNCTION_CALL or self::SLOT[parent::expr/following-sibling::OP-LEFT-PAREN])
        and ({ unsuppressed_call_cond })
      ][last()]
      /following::expr[SYMBOL_FUNCTION_CALL[{ attach_call_cond }]]
      /parent::expr
  ")

  # STR_CONST: block library|require("..."), i.e., supplying a string literal
  # ancestor::expr[FUNCTION]: Skip usages inside functions a la {knitr}
  char_only_direct_xpath <- glue("
  //SYMBOL_FUNCTION_CALL[{attach_call_cond}]
    /parent::expr
    /parent::expr[
      expr[2][STR_CONST]
      or (
        SYMBOL_SUB[text() = 'character.only']
        and not(ancestor::expr[FUNCTION or OP-LAMBDA])
      )
    ]
  ")

  bad_indirect_funs <- c("do.call", "lapply", "sapply", "map", "walk")
  call_symbol_cond <- glue("SYMBOL[{attach_call_cond}] or STR_CONST")
  char_only_indirect_xpath <- glue("
  //SYMBOL_FUNCTION_CALL[{ xp_text_in_table(bad_indirect_funs) }]
    /parent::expr
    /parent::expr[
      not(ancestor::expr[FUNCTION or OP-LAMBDA])
      and expr[{ call_symbol_cond }]
    ]
  ")
  call_symbol_path <- glue("./expr[{call_symbol_cond}]")

  attach_expr_cond <- glue("expr[expr/SYMBOL_FUNCTION_CALL[{attach_call_cond}]]")

  # Use `calls` in the first condition, not in the second, to prevent, e.g.,
  #   the first call matching calls[1] but the second matching calls[2].
  #   That is, ensure that calls[i] only matches a following call to calls[i].
  # match on the expr, not the SYMBOL_FUNCTION_CALL, to ensure
  #   namespace-qualified calls only match if the namespaces do.
  consecutive_suppress_xpath <- glue("
  //SYMBOL_FUNCTION_CALL[{ suppress_call_cond }]
    /parent::expr
    /parent::expr[
      expr[SYMBOL_FUNCTION_CALL[{ suppress_call_cond }]] =
        following-sibling::expr[1][{attach_expr_cond}]/expr
      and {attach_expr_cond}
    ]
  ")

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    upfront_call_expr <- xml_find_all_(xml, upfront_call_xpath)

    upfront_call_name <- xp_call_name(upfront_call_expr)

    upfront_call_lints <- xml_nodes_to_lints(
      upfront_call_expr,
      source_expression = source_expression,
      lint_message = sprintf("Move all %s calls to the top of the script.", upfront_call_name),
      type = "warning"
    )

    char_only_direct_expr <- xml_find_all_(xml, char_only_direct_xpath)
    char_only_direct_calls <- xp_call_name(char_only_direct_expr)
    character_only <-
      xml_find_first_(char_only_direct_expr, "./SYMBOL_SUB[text() = 'character.only']")
    char_only_direct_msg_fmt <- ifelse(
      is.na(character_only),
      "Use symbols, not strings, in %s calls.",
      "Use symbols in %s calls to avoid the need for 'character.only'."
    )
    char_only_direct_msg <-
      sprintf(as.character(char_only_direct_msg_fmt), char_only_direct_calls)
    char_only_direct_lints <- xml_nodes_to_lints(
      char_only_direct_expr,
      source_expression = source_expression,
      lint_message = char_only_direct_msg,
      type = "warning"
    )

    char_only_indirect_expr <- xml_find_all_(xml, char_only_indirect_xpath)
    char_only_indirect_lib_calls <- vapply(
      char_only_indirect_expr,
      function(expr) {
        calls <- get_r_string(xml_find_all_(expr, call_symbol_path))
        calls <- calls[calls %in% attach_calls]
        if (length(calls) == 1L) calls else NA_character_
      },
      character(1L)
    )

    # For STR_CONST entries, the XPath doesn't check the string value -- we use
    #   get_r_string() here to do that filter more robustly.
    is_attach_call <- !is.na(char_only_indirect_lib_calls)
    char_only_indirect_expr <- char_only_indirect_expr[is_attach_call]
    char_only_indirect_lib_calls <- char_only_indirect_lib_calls[is_attach_call]

    char_only_indirect_loop_calls <- xp_call_name(char_only_indirect_expr)
    char_only_indirect_msg <- sprintf(
      "Call %s() directly, not vectorized with %s().",
      char_only_indirect_lib_calls, char_only_indirect_loop_calls
    )
    char_only_indirect_lints <- xml_nodes_to_lints(
      char_only_indirect_expr,
      source_expression = source_expression,
      lint_message = char_only_indirect_msg,
      type = "warning"
    )

    consecutive_suppress_expr <- xml_find_all_(xml, consecutive_suppress_xpath)
    consecutive_suppress_call_text <- xp_call_name(consecutive_suppress_expr)
    consecutive_suppress_message <- paste0(
      "Unify consecutive calls to ", consecutive_suppress_call_text, "(). ",
      "You can do so by writing all of the calls in one braced expression ",
      "like ", consecutive_suppress_call_text, "({...})."
    )
    consecutive_suppress_lints <- xml_nodes_to_lints(
      consecutive_suppress_expr,
      source_expression = source_expression,
      lint_message = consecutive_suppress_message,
      type = "warning"
    )

    c(upfront_call_lints, char_only_direct_lints, char_only_indirect_lints, consecutive_suppress_lints)
  })
}

#' @include utils.R
#' @include xp_utils.R
#' @include make_linter_from_xpath.R
NULL

#' Available linters
#' @name linters
#'
#' @description A variety of linters are available in \pkg{lintr}. The most popular ones are readily
#' accessible through [default_linters()].
#'
#' Within a [lint()] function call, the linters in use are initialized with the provided
#' arguments and fed with the source file (provided by [get_source_expressions()]).
#'
#' A data frame of all available linters can be retrieved using [available_linters()].
#' Documentation for linters is structured into tags to allow for easier discovery;
#' see also [available_tags()].
#'
#' @evalRd rd_taglist()
#' @evalRd rd_linterlist()
NULL

# need to register rex shortcuts as globals to avoid CRAN check errors
rex::register_shortcuts("lintr")

globalVariables(
  c(
    "line1", "col1", "line2", "col2", # columns of parsed_content
    "id", "parent", "token", "terminal", "text" # ditto
  ),
  "lintr"
)

#' Lintr
#'
#' Checks adherence to a given style, syntax errors, and possible semantic
#'   issues.
#' Supports on the fly checking of R code edited with Emacs, Vim, and
#'   Sublime Text.
#'
#' @seealso [lint()], [lint_package()], [lint_dir()], [linters]
#' @keywords internal
"_PACKAGE"

## lintr namespace: start
#' @importFrom cli cli_inform cli_abort cli_warn qty
#' @importFrom glue glue glue_collapse
#' @importFrom rex rex regex re_matches re_substitutes character_class
#' @importFrom stats complete.cases na.omit
#' @importFrom tools R_user_dir
#' @importFrom utils capture.output getParseData  globalVariables head
#' @importFrom utils relist tail
#' @importFrom xml2 xml_attr xml_attrs xml_child xml_children xml_find_all
#' @importFrom xml2 xml_find_chr xml_find_first xml_find_lgl xml_find_num
#' @importFrom xml2 xml_missing xml_name xml_parent xml_text
## lintr namespace: end
NULL

# make binding available for mock testing
# ref: https://testthat.r-lib.org/dev/reference/local_mocked_bindings.html#base-functions
# nolint start: object_name_linter. These will be copied from base style.
requireNamespace <- NULL
system.file <- NULL
unlink <- NULL
quit <- NULL
# nolint end: object_name_linter.

#' Recommend usage of `call. = FALSE` in conditions
#'
#' This linter, with the default `display_call = FALSE`, enforces the
#' recommendation of the tidyverse design guide regarding displaying error
#' calls.
#'
#' @param display_call Logical specifying expected behavior regarding `call.`
#' argument in conditions.
#'   - `NA` forces providing `call. =` but ignores its value (this can be used in
#'     cases where you expect a mix of `call. = FALSE` and `call. = TRUE`)
#'   - `TRUE` lints `call. = FALSE`
#'   - `FALSE` forces `call. = FALSE` (lints `call. = TRUE` or missing `call. =` value)
#'
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "stop('test')",
#'   linters = condition_call_linter()
#' )
#'
#' lint(
#'   text = "stop('test', call. = TRUE)",
#'   linters = condition_call_linter()
#' )
#'
#' lint(
#'   text = "stop('test', call. = FALSE)",
#'   linters = condition_call_linter(display_call = TRUE)
#' )
#'
#' lint(
#'   text = "stop('this is a', 'test', call. = FALSE)",
#'   linters = condition_call_linter(display_call = TRUE)
#' )
#'
#' # okay
#' lint(
#'   text = "stop('test', call. = FALSE)",
#'   linters = condition_call_linter()
#' )
#'
#' lint(
#'   text = "stop('this is a', 'test', call. = FALSE)",
#'   linters = condition_call_linter()
#' )
#'
#' lint(
#'   text = "stop('test', call. = TRUE)",
#'   linters = condition_call_linter(display_call = TRUE)
#' )
#'
#' @evalRd rd_tags("condition_call_linter")
#' @seealso
#'   - [linters] for a complete list of linters available in lintr.
#'   - <https://design.tidyverse.org/err-call.html>>
#' @export
condition_call_linter <- function(display_call = FALSE) {

  str_fct <- c("sprintf", "gettextf", "paste", "paste0", "format", "tr_")
  str_literal_or_fct <- glue("
    STR_CONST
    or expr/SYMBOL_FUNCTION_CALL[ { xp_text_in_table(str_fct) }]
  ")

  call_xpath <- glue("
    following-sibling::expr[ {str_literal_or_fct} ]
    and following-sibling::SYMBOL_SUB[text() = 'call.']
      /following-sibling::expr[1]
      /NUM_CONST[text() = '{!display_call}']
  ")
  no_call_xpath <- glue("
    following-sibling::expr[ {str_literal_or_fct} ]
    and parent::expr[not(SYMBOL_SUB[text() = 'call.'])]
  ")

  if (is.na(display_call)) {
    call_cond <- no_call_xpath
    msg_fmt <- "Provide an explicit value for `call.` in %s()."
  } else if (display_call) {
    call_cond <- call_xpath
    msg_fmt <- "Use %s(.) to display the call in an error message."
  } else {
    # call. = TRUE can be expressed in two way:
    #  - either explicitly with call. = TRUE
    #  - or by implicitly relying on the default
    call_cond <- xp_or(call_xpath, no_call_xpath)
    msg_fmt <- "Use %s(., call. = FALSE) not to display the call in an error message."
  }

  xpath <- glue("./self::*[{call_cond}]/parent::expr")

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(c("stop", "warning"))

    bad_expr <- xml_find_all_(xml_calls, xpath)

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = sprintf(msg_fmt, xp_call_name(bad_expr)),
      type = "warning"
    )
  })
}

#' Create a linter from an XPath
#'
#' @inheritParams xml_nodes_to_lints
#' @inheritParams is_lint_level
#' @param xpath Character string, an XPath identifying R code to lint.
#'   For `make_linter_from_function_xpath()`, the XPath is relative to the `parent::expr` of the
#'   `SYMBOL_FUNCTION_CALL` nodes of the selected functions.
#'   See [xmlparsedata::xml_parse_data()] and [get_source_expressions()].
#'
#' @examples
#' number_linter <- make_linter_from_xpath("//NUM_CONST", "This is a number.")
#' lint(text = "1 + 2", linters = number_linter())
#' @export
make_linter_from_xpath <- function(xpath,
                                   lint_message,
                                   type = c("warning", "style", "error"),
                                   level = c("expression", "file")) {
  type <- match.arg(type)
  level <- match.arg(level)

  if (!is.character(xpath) || length(xpath) != 1L || is.na(xpath)) {
    cli_abort("{.arg xpath} must be a character string.")
  }
  if (missing(lint_message)) {
    cli_abort("{.arg lint_message} is required.")
  }

  xml_key <- if (level == "expression") "xml_parsed_content" else "full_xml_parsed_content"

  function() {
    Linter(linter_level = level, function(source_expression) {
      xml <- source_expression[[xml_key]]


      expr <- xml_find_all_(xml, xpath)

      xml_nodes_to_lints(
        expr,
        source_expression = source_expression,
        lint_message = lint_message,
        type = type
      )
    })
  }
}

#' @rdname make_linter_from_xpath
#' @param function_names Character vector, names of functions whose calls to examine..
#' @export
# nolint next: object_length.
make_linter_from_function_xpath <- function(function_names,
                                            xpath,
                                            lint_message,
                                            type = c("warning", "style", "error"),
                                            level = c("expression", "file")) {
  type <- match.arg(type)
  level <- match.arg(level)

  if (!is.character(function_names) || length(function_names) == 0L) {
    cli_abort("{.arg function_names} must be a character vector.")
  }
  if (!is.character(xpath) || length(xpath) != 1L || is.na(xpath)) {
    cli_abort("{.arg xpath} must be a character string.")
  }
  if (missing(lint_message)) {
    cli_abort("{.arg lint_message} is required.")
  }

  function() {
    Linter(linter_level = level, function(source_expression) {
      call_xml <- source_expression$xml_find_function_calls(function_names)

      expr <- xml_find_all_(call_xml, xpath)

      xml_nodes_to_lints(
        expr,
        source_expression = source_expression,
        lint_message = lint_message,
        type = type
      )
    })
  }
}

#' Block usage of `paste()` and `paste0()` with messaging functions using `...`
#'
#' @description
#' This linter discourages combining condition functions like [stop()] with string concatenation
#'   functions [base::paste()] and [base::paste0()]. This is because
#'
#'  - `stop(paste0(...))` is redundant as it is exactly equivalent to `stop(...)`
#'  - `stop(paste(...))` is similarly equivalent to `stop(...)` with separators (see examples)
#'
#' The same applies to the other default condition functions as well, i.e., [warning()], [message()],
#'   and [packageStartupMessage()].
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'stop(paste("a string", "another"))',
#'   linters = condition_message_linter()
#' )
#'
#' lint(
#'   text = 'warning(paste0("a string", " another"))',
#'   linters = condition_message_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'stop("a string", " another")',
#'   linters = condition_message_linter()
#' )
#'
#' lint(
#'   text = 'warning("a string", " another")',
#'   linters = condition_message_linter()
#' )
#'
#' lint(
#'   text = 'warning(paste("a string", "another", sep = "-"))',
#'   linters = condition_message_linter()
#' )
#'
#' @evalRd rd_tags("condition_message_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
condition_message_linter <- function() {
  translators <- c("packageStartupMessage", "message", "warning", "stop")
  xpath <- glue("
  self::*[SYMBOL_FUNCTION_CALL[
    not(preceding-sibling::OP-DOLLAR or preceding-sibling::OP-AT)
  ]]
    /following-sibling::expr[
      expr[1][SYMBOL_FUNCTION_CALL[text() = 'paste' or text() = 'paste0']]
      and not(SYMBOL_SUB[text() = 'collapse'])
    ]
    /parent::expr
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(translators)
    bad_expr <- xml_find_all_(xml_calls, xpath)
    sep_value <- get_r_string(bad_expr, xpath = "./expr/SYMBOL_SUB[text() = 'sep']/following-sibling::expr/STR_CONST")

    bad_expr <- bad_expr[is.na(sep_value) | sep_value %in% c("", " ")]
    outer_call <- xp_call_name(bad_expr)
    inner_call <- xp_call_name(bad_expr, depth = 2L)
    lint_message <- paste(
      "Don't use", inner_call, "to build", outer_call, "strings.",
      "Instead use the fact that these functions build condition message strings from their input",
      '(using "" as a separator). For translatable strings, prefer using gettextf().'
    )
    xml_nodes_to_lints(bad_expr, source_expression = source_expression, lint_message = lint_message, type = "warning")
  })
}

#' Extract row by ID
#'
#' @describeIn ids_with_token
#' Return the row of the `parsed_content` entry of the `[get_source_expressions]()` object. Typically used in
#' conjunction with `ids_with_token` to iterate over rows containing desired tokens.
#'
#' @param id Integer. The index corresponding to the desired row
#' of `parsed_content`.
#' @return `with_id`: A data frame corresponding to the row(s) specified in `id`.
#' @export
with_id <- function(source_expression, id) {
  if (!is_lint_level(source_expression, "expression")) {
    return(data.frame())
  }
  source_expression$parsed_content[id, ]
}

#' Require usage of sample.int(n, m, ...) over sample(1:n, m, ...)
#'
#' [sample.int()] is preferable to `sample()` for the case of sampling numbers
#'   between 1 and `n`. `sample` calls `sample.int()` "under the hood".
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "sample(1:10, 2)",
#'   linters = sample_int_linter()
#' )
#'
#' lint(
#'   text = "sample(seq(4), 2)",
#'   linters = sample_int_linter()
#' )
#'
#' lint(
#'   text = "sample(seq_len(8), 2)",
#'   linters = sample_int_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "sample(seq(1, 5, by = 2), 2)",
#'   linters = sample_int_linter()
#' )
#'
#' lint(
#'   text = "sample(letters, 2)",
#'   linters = sample_int_linter()
#' )
#'
#' @evalRd rd_tags("sample_int_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
sample_int_linter <- function() {
  # looking for anything like sample(1: that doesn't come after a $ extraction
  # exclude TRUE/FALSE for sample(replace = TRUE, ...) usage. better
  #   would be match.arg() but this also works.
  xpath <- glue("
  self::*[not(OP-DOLLAR or OP-AT)]
    /following-sibling::expr[1][
      (
        expr[1]/NUM_CONST[text() = '1' or text() = '1L']
        and OP-COLON
      )
      or expr/SYMBOL_FUNCTION_CALL[text() = 'seq_len']
      or (
        expr/SYMBOL_FUNCTION_CALL[text() = 'seq']
        and (
          count(expr) = 2
          or (
            expr[2]/NUM_CONST[text() = '1' or text() = '1L']
            and not(SYMBOL_SUB[
              text() = 'by'
              and not(following-sibling::expr[1]/NUM_CONST[text() = '1' or text() = '1L'])
            ])
          )
        )
      )
      or NUM_CONST[not(text() = 'TRUE' or text() = 'FALSE')]
    ]
    /parent::expr
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls("sample")
    bad_expr <- xml_find_all_(xml_calls, xpath)

    first_call <- xp_call_name(bad_expr, depth = 2L)
    original <- sprintf("%s(n)", first_call)
    original[!is.na(xml_find_first_(bad_expr, "expr[2]/OP-COLON"))] <- "1:n"
    original[!is.na(xml_find_first_(bad_expr, "expr[2]/NUM_CONST"))] <- "n"

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste0("sample.int(n, m, ...) is preferable to sample(", original, ", m, ...)."),
      type = "warning"
    )
  })
}

#' Block usage of pipeline placeholders if unnecessary
#'
#' The argument placeholder `.` in magrittr pipelines is unnecessary if
#'   passed as the first positional argument; using it can cause confusion
#'   and impacts readability.
#'
#' This is true for forward (`%>%`), assignment (`%<>%`), and tee (`%T>%`) operators.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x %>% sum(., na.rm = TRUE)",
#'   linters = unnecessary_placeholder_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x %>% sum(na.rm = TRUE)",
#'   linters = unnecessary_placeholder_linter()
#' )
#'
#' lint(
#'   text = "x %>% lm(data = ., y ~ z)",
#'   linters = unnecessary_placeholder_linter()
#' )
#'
#' lint(
#'   text = "x %>% outer(., .)",
#'   linters = unnecessary_placeholder_linter()
#' )
#'
#' @evalRd rd_tags("unnecessary_placeholder_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
unnecessary_placeholder_linter <- function() {
  # NB: Native placeholder '_' must be used with a named argument, so it's not relevant here.
  xpath <- glue("
  //SPECIAL[{ xp_text_in_table(magrittr_pipes) }]
    /following-sibling::expr[
      expr/SYMBOL_FUNCTION_CALL
      and not(expr[
        position() > 2
        and descendant-or-self::expr/SYMBOL[text() = '.']
      ])
    ]
    /expr[2][
      SYMBOL[text() = '.']
      and not(preceding-sibling::*[not(self::COMMENT)][1][self::EQ_SUB])
    ]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste(
        "Don't use the placeholder (`.`) when it's not needed,",
        "i.e., when it's only used as the first positional argument in a pipeline step."
      ),
      type = "warning"
    )
  })
}

# This file contains all documentation for linter tags in lintr except for default_linters.

#' Style linters
#' @name style_linters
#' @description
#' Linters highlighting code style issues.
#' @evalRd rd_linters("style")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Robustness linters
#' @name robustness_linters
#' @description
#' Linters highlighting code robustness issues, such as possibly wrong edge case behavior.
#' @evalRd rd_linters("robustness")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Best practices linters
#' @name best_practices_linters
#' @description
#' Linters checking the use of coding best practices, such as explicit typing of numeric constants.
#' @evalRd rd_linters("best_practices")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Consistency linters
#' @name consistency_linters
#' @description
#' Linters checking enforcing a consistent alternative if there are multiple syntactically valid ways to write
#' something.
#' @evalRd rd_linters("consistency")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Readability linters
#' @name readability_linters
#' @description
#' Linters highlighting readability issues, such as missing whitespace.
#' @evalRd rd_linters("readability")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Correctness linters
#' @name correctness_linters
#' @description
#' Linters highlighting possible programming mistakes, such as unused variables.
#' @evalRd rd_linters("correctness")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Common mistake linters
#' @name common_mistakes_linters
#' @description
#' Linters highlighting common mistakes, such as duplicate arguments.
#' @evalRd rd_linters("common_mistakes")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Efficiency linters
#' @name efficiency_linters
#' @description
#' Linters highlighting code efficiency problems, such as unnecessary function calls.
#' @evalRd rd_linters("efficiency")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Configurable linters
#' @name configurable_linters
#' @description
#' Generic linters which support custom configuration to your needs.
#' @evalRd rd_linters("configurable")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Package development linters
#' @name package_development_linters
#' @description
#' Linters useful to package developers, for example for writing consistent tests.
#' @evalRd rd_linters("package_development")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Deprecated linters
#' @name deprecated_linters
#' @description
#' Linters that are deprecated and provided for backwards compatibility only.
#' These linters will be excluded from [linters_with_tags()] by default.
#' @evalRd rd_linters("deprecated")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Code executing linters
#' @name executing_linters
#' @description
#' Linters that evaluate parts of the linted code, such as loading referenced packages.
#' These linters should not be used with untrusted code, and may need dependencies of the linted package or project to
#' be available in order to function correctly. For package authors, note that this includes loading the package itself,
#'   e.g. with `pkgload::load_all()` or installing and attaching the package.
#' @evalRd rd_linters("executing")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Testthat linters
#' @name pkg_testthat_linters
#' @description
#' Linters encouraging best practices within testthat suites.
#' @evalRd rd_linters("pkg_testthat")
#' @seealso
#'  - [linters] for a complete list of linters available in lintr.
#'  - <https://testthat.r-lib.org>
#'  - <https://r-pkgs.org/testing-basics.html>
NULL

#' Regular expression linters
#' @name regex_linters
#' @description
#' Linters that examine the usage of regular expressions and functions executing them in user code.
#' @evalRd rd_linters("regex")
#' @seealso [linters] for a complete list of linters available in lintr.
NULL

#' Tidyverse design linters
#' @name tidy_design_linters
#' @description
#' Linters based on guidelines described in the 'Tidy design principles' book.
#' @evalRd rd_linters("tidy_design")
#' @seealso
#'  - [linters] for a complete list of linters available in lintr.
#'  - <https://design.tidyverse.org/>
NULL

#' Require usage of nzchar where appropriate
#'
#' [nzchar()] efficiently determines which of a vector of strings are empty
#'   (i.e., are `""`). It should in most cases be used instead of
#'   constructions like `string == ""` or `nchar(string) == 0`.
#'
#' One crucial difference is in the default handling of `NA_character_`, i.e.,
#'   missing strings. `nzchar(NA_character_)` is `TRUE`, while `NA_character_ == ""`
#'   and `nchar(NA_character_) == 0` are both `NA`. Therefore, for strict
#'   compatibility, use `nzchar(x, keepNA = TRUE)`. If the input is known to be
#'   complete (no missing entries), this argument can be dropped for conciseness.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x[x == '']",
#'   linters = nzchar_linter()
#' )
#'
#' lint(
#'   text = "x[nchar(x) > 0]",
#'   linters = nzchar_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x[!nzchar(x, keepNA = TRUE)]",
#'   linters = nzchar_linter()
#' )
#'
#' lint(
#'   text = "x[nzchar(x, keepNA = TRUE)]",
#'   linters = nzchar_linter()
#' )
#'
#' @evalRd rd_tags("nzchar_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
nzchar_linter <- function() {
  comparator_nodes <- infix_metadata$xml_tag[infix_metadata$comparator]

  # use string-length to capture both "" and ''
  # if (any(x == "")) is not treated like it's part of if(), but
  #   any(if (x == "") y else z) _is_ treated so. this condition looks for the
  #   expr to be inside a call that's _not_ above an IF/WHILE.
  comparison_xpath <- glue("
  //STR_CONST[string-length(text()) = 2]
    /parent::expr
    /parent::expr[
      ({ xp_or(comparator_nodes) })
      and (
        not(ancestor-or-self::expr[
          preceding-sibling::IF
          or preceding-sibling::WHILE
        ])
        or ancestor-or-self::expr[
          (
            preceding-sibling::expr/SYMBOL_FUNCTION_CALL
            or preceding-sibling::OP-LEFT-BRACKET
          ) and not(
            descendant-or-self::expr[IF or WHILE]
          )
        ]
      )
    ]
  ")

  comparison_msg_map <- c(
    GT = 'Use nzchar(x) instead of x > "". ',
    NE = 'Use nzchar(x) instead of x != "". ',
    LE = 'Use !nzchar(x) instead of x <= "". ',
    EQ = 'Use !nzchar(x) instead of x == "". ',
    GE = 'x >= "" is always true, maybe you want nzchar(x)? ',
    LT = 'x < "" is always false, maybe you want !nzchar(x)? '
  )

  # nchar(., type="width") not strictly compatible with nzchar
  # nchar(., allowNA=TRUE) neither
  nchar_xpath <- glue("
  parent::expr
    /parent::expr[
      ({ xp_or(comparator_nodes) })
      and not(expr/SYMBOL_SUB[
        (
          text() = 'type'
          and following-sibling::expr[1]/STR_CONST[contains(text(), 'width')]
        ) or (
          text() = 'allowNA'
          and following-sibling::expr[1]/NUM_CONST[text() = 'TRUE']
        )
      ])
      and expr/NUM_CONST[text() = '0' or text() = '0L' or text() = '0.0']
    ]
  ")

  nchar_msg_map <- c(
    GT = "Use nzchar(x) instead of nchar(x) > 0. ",
    NE = "Use nzchar(x) instead of nchar(x) != 0. ",
    LE = "Use !nzchar(x) instead of nchar(x) <= 0. ",
    EQ = "Use !nzchar(x) instead of nchar(x) == 0. ",
    GE = "nchar(x) >= 0 is always true, maybe you want nzchar(x)? ",
    LT = "nchar(x) < 0 is always false, maybe you want !nzchar(x)? "
  )

  keepna_note <- paste(
    "Whenever missing data is possible,",
    "please take care to use nzchar(., keepNA = TRUE);",
    "nzchar(NA) is TRUE by default."
  )

  # For ordered operators like '>', we need to give the message for
  #   its "opposite" (not inverse) if the bad usage is on the RHS,
  #   e.g. 0 < nchar(x) has to be treated as nchar(x) > 0.
  op_for_msg <- function(expr, const) {
    op <- xml_name_(xml_find_first_(expr, "*[not(self::COMMENT)][2]"))
    maybe_needs_flip <-
      !is.na(xml_find_first_(expr, sprintf("*[not(self::COMMENT)][1][%s]", const)))

    ordered_ops <- c("GT", "GE", "LE", "LT")
    ordered_idx <- match(op, ordered_ops)

    needs_flip <- maybe_needs_flip & !is.na(ordered_idx)
    # un-benchmarked, but should be faster (though less readable) as
    # > ordered_ops[5L - ordered_idx[needs_flip]]
    op[needs_flip] <- rev(ordered_ops)[ordered_idx[needs_flip]]
    op
  }

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    comparison_expr <- xml_find_all_(xml, comparison_xpath)
    comparison_op <- op_for_msg(comparison_expr, const = "STR_CONST")
    comparison_lints <- xml_nodes_to_lints(
      comparison_expr,
      source_expression = source_expression,
      lint_message = paste0(
        comparison_msg_map[comparison_op],
        "Note that unlike nzchar(), ", comparison_op, " coerces to character, ",
        "so you'll have to use as.character() if x is a factor. ",
        keepna_note
      ),
      type = "warning"
    )

    xml_calls <- source_expression$xml_find_function_calls("nchar")
    nchar_expr <- xml_find_all_(xml_calls, nchar_xpath)
    nchar_op <- op_for_msg(nchar_expr, const = "NUM_CONST")
    nchar_lints <- xml_nodes_to_lints(
      nchar_expr,
      source_expression = source_expression,
      lint_message = paste0(nchar_msg_map[nchar_op], keepna_note),
      type = "warning"
    )

    c(comparison_lints, nchar_lints)
  })
}

#' Require usage of `!any(x)` over `all(!x)`, `!all(x)` over `any(!x)`
#'
#' `any(!x)` is logically equivalent to `!all(x)`; ditto for the equivalence of
#'   `all(!x)` and `!any(x)`. Negating after aggregation only requires inverting
#'   one logical value, and is typically more readable.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "all(!x)",
#'   linters = outer_negation_linter()
#' )
#'
#' lint(
#'   text = "any(!x)",
#'   linters = outer_negation_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "!any(x)",
#'   linters = outer_negation_linter()
#' )
#'
#' lint(
#'   text = "!all(x)",
#'   linters = outer_negation_linter()
#' )
#'
#' @evalRd rd_tags("outer_negation_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
outer_negation_linter <- function() {
  # NB: the double negation is a bit hairy, but it's what we need to check if
  #   _all_ of the inputs to any(..., na.rm=na.rm) are negated, i.e., there are
  #   _not_ any entries that are _not_ negated. IINM that's what we're stuck
  #   with in xpath if we want to guarantee a condition on _all_ <expr>
  #   coming after any( and before na.rm= .
  # NB: requirement that count(expr)>1 is to prevent any() from linting
  #   e.g. in magrittr pipelines.
  xpath <- "
  self::*[following-sibling::expr]
    /parent::expr[
      not(expr[
        position() > 1
        and not(OP-EXCLAMATION)
        and not(preceding-sibling::*[not(self::COMMENT)][1][self::EQ_SUB])
      ])
    ]
  "

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(c("any", "all"))
    bad_expr <- xml_find_all_(xml_calls, xpath)

    matched_call <- xp_call_name(bad_expr)
    inverse_call <- ifelse(matched_call == "any", "all", "any")
    lint_message <- paste(
      sprintf("!%s(x) is better than %s(!x).", inverse_call, matched_call),
      "The former applies negation only once after aggregation instead of many times for each element of x."
    )

    xml_nodes_to_lints(bad_expr, source_expression, lint_message, type = "warning")
  })
}

#' Unneeded concatenation linter
#'
#' Check that the [c()] function is not used without arguments nor with a single constant.
#'
#' @param allow_single_expression Logical, default `TRUE`. If `FALSE`, one-expression
#'   usages of `c()` are always linted, e.g. `c(x)` and `c(matrix(...))`. In some such
#'   cases, `c()` is being used for its side-effect of stripping non-name attributes;
#'   it is usually preferable to use the more readable [as.vector()] instead.
#'   [as.vector()] is not always preferable, for example with environments
#'   (especially, `R6` objects), in which case `list()` is the better alternative.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x <- c()",
#'   linters = unnecessary_concatenation_linter()
#' )
#'
#' lint(
#'   text = "x <- c(TRUE)",
#'   linters = unnecessary_concatenation_linter()
#' )
#'
#' lint(
#'   text = "x <- c(1.5 + 2.5)",
#'   linters = unnecessary_concatenation_linter(allow_single_expression = FALSE)
#' )
#'
#' # okay
#' lint(
#'   text = "x <- NULL",
#'   linters = unnecessary_concatenation_linter()
#' )
#'
#' # In case the intent here was to seed a vector of known size
#' lint(
#'   text = "x <- integer(4L)",
#'   linters = unnecessary_concatenation_linter()
#' )
#'
#' lint(
#'   text = "x <- TRUE",
#'   linters = unnecessary_concatenation_linter()
#' )
#'
#' lint(
#'   text = "x <- c(1.5 + 2.5)",
#'   linters = unnecessary_concatenation_linter(allow_single_expression = TRUE)
#' )
#'
#' @evalRd rd_tags("unnecessary_concatenation_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
unnecessary_concatenation_linter <- function(allow_single_expression = TRUE) { # nolint: object_length_linter.
  if (!is.logical(allow_single_expression) || length(allow_single_expression) != 1L) {
    cli_abort("{.arg allow_single_expression} must be a single logical value.")
  }

  msg_empty <-
    "Replace unnecessary c() by NULL or, whenever possible, vector() seeded with the correct type and/or length."

  msg_const <- "Remove unnecessary c() of a constant."

  non_constant_cond <- "SYMBOL or (expr and not(OP-COLON and count(expr[SYMBOL or expr]) != 2))"

  pipes <- setdiff(magrittr_pipes, "%$%")
  to_pipe_xpath <- glue("
    ./preceding-sibling::*[not(self::COMMENT)][1][
      self::PIPE or
      self::SPECIAL[{ xp_text_in_table(pipes) }]
    ]
  ")
  if (allow_single_expression) {
    zero_arg_cond <-
      glue("count(expr) = 1 and not( {to_pipe_xpath} / preceding-sibling::expr[ {non_constant_cond} ])")
    one_arg_cond <-
      glue("count(expr) = 2 and not(expr[2][ {non_constant_cond} ])")
  } else {
    zero_arg_cond <- glue("count(expr) = 1 and not( {to_pipe_xpath} )")
    one_arg_cond <- "count(expr) = 2 and not(expr[2]/SYMBOL[text() = '...'])"
    path_to_non_constant <- glue("./expr[2][ {non_constant_cond} ]")

    msg_const_expr <- paste(
      "Remove unnecessary c() of a constant expression.",
      "Replace with as.vector() if c() is used to strip attributes, e.g. in converting an array to a vector."
    )
  }
  call_xpath <- glue("
  parent::expr[
    not(EQ_SUB)
    and ( {xp_or(zero_arg_cond, one_arg_cond)} )
  ]")
  num_args_xpath <- "count(./expr) - 1"

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls("c")
    c_calls <- xml_find_all_(xml_calls, call_xpath)

    # bump count(args) by 1 if inside a pipeline
    num_args <- as.integer(xml_find_num_(c_calls, num_args_xpath)) +
      as.integer(!is.na(xml_find_first_(c_calls, to_pipe_xpath)))
    # NB: the xpath guarantees num_args is 0, 1, or 2. 2 comes
    #   in "a" %>% c("b").
    # TODO(#2476): Push this logic back into the XPath.
    is_unneeded <- num_args <= 1L
    c_calls <- c_calls[is_unneeded]
    num_args <- num_args[is_unneeded]
    msg <- ifelse(num_args == 0L, msg_empty, msg_const)
    if (!allow_single_expression) {
      is_single_expression <- !is.na(xml_find_first_(c_calls, path_to_non_constant))
      msg[is_single_expression] <- msg_const_expr
    }

    xml_nodes_to_lints(
      c_calls,
      source_expression = source_expression,
      lint_message = msg
    )
  })
}

#' Require usage of rep_len(x, n) over rep(x, length.out = n)
#'
#' `rep(x, length.out = n)` calls `rep_len(x, n)` "under the hood". The latter
#'   is thus more direct and equally readable.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "rep(1:3, length.out = 10)",
#'   linters = rep_len_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "rep_len(1:3, 10)",
#'   linters = rep_len_linter()
#' )
#'
#' lint(
#'   text = "rep(1:3, each = 2L, length.out = 10L)",
#'   linters = rep_len_linter()
#' )
#'
#' @evalRd rd_tags("rep_len_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
rep_len_linter <- make_linter_from_function_xpath(
  function_names = "rep",
  # count(expr) is for cases using positional matching; see ?rep.
  xpath = "
    parent::expr[
      (
        SYMBOL_SUB[text() = 'length.out']
        or (not(SYMBOL_SUB) and count(expr) = 4)
      )
      and not(SYMBOL_SUB[text() = 'each'] or count(expr) = 5)
    ]
  ",
  lint_message = "Use rep_len(x, n) instead of rep(x, length.out = n)."
)

#' Identify cases where `stringsAsFactors` should be supplied explicitly
#'
#' Designed for code bases written for versions of R before 4.0 seeking to upgrade to R >= 4.0, where
#'   one of the biggest pain points will surely be the flipping of the
#'   default value of `stringsAsFactors` from `TRUE` to `FALSE`.
#'
#' It's not always possible to tell statically whether the change will break
#'   existing code because R is dynamically typed -- e.g. in `data.frame(x)`
#'   if `x` is a string, this code will be affected, but if `x` is a number,
#'   this code will be unaffected. However, in `data.frame(x = "a")`, the
#'   output will unambiguously be affected. We can instead supply
#'   `stringsAsFactors = TRUE`, which will make this code backwards-compatible.
#'
#' See <https://developer.r-project.org/Blog/public/2020/02/16/stringsasfactors/>.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'data.frame(x = "a")',
#'   linters = strings_as_factors_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'data.frame(x = "a", stringsAsFactors = TRUE)',
#'   linters = strings_as_factors_linter()
#' )
#'
#' lint(
#'   text = 'data.frame(x = "a", stringsAsFactors = FALSE)',
#'   linters = strings_as_factors_linter()
#' )
#'
#' lint(
#'   text = "data.frame(x = 1.2)",
#'   linters = strings_as_factors_linter()
#' )
#'
#' @evalRd rd_tags("strings_as_factors_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
strings_as_factors_linter <- local({
  # a call to c() with only literal string inputs,
  #   e.g. c("a") or c("a", "b"), but not c("a", b)
  c_combine_strings <- "
    expr[1][SYMBOL_FUNCTION_CALL[text() = 'c']]
    and expr[STR_CONST]
    and not(expr[SYMBOL or expr])
  "

  # like data.frame(character()) or data.frame(as.character(1))
  known_character_funs <- c(
    "character", "as.character", "paste", "sprintf",
    "format", "formatC", "prettyNum", "toString", "encodeString"
  )

  # four inclusions of arguments to data.frame():
  #   (1) "a"                 (but exclude argument names, e.g. in c("a b" = 1), #1036)
  #   (2) c("a", "b")
  #   (3) rep("a", 2)
  #   (4) rep(c("a", "b"), 2)
  # two exclusions
  #   (1) above argument is to row.names=
  #   (2) stringsAsFactors is manually supplied (with any value)
  xpath <- glue("
  parent::expr[
    expr[
      (
        STR_CONST[not(following-sibling::*[not(self::COMMENT)][1][self::EQ_SUB])]
        or ( {c_combine_strings} )
        or expr[1][
          SYMBOL_FUNCTION_CALL[text() = 'rep']
          and following-sibling::expr[1][STR_CONST or ({c_combine_strings})]
        ]
        or expr[1][SYMBOL_FUNCTION_CALL[ {xp_text_in_table(known_character_funs)} ]]
      )
      and not(preceding-sibling::*[not(self::COMMENT)][2][self::SYMBOL_SUB and text() = 'row.names'])
    ]
    and not(SYMBOL_SUB[text() = 'stringsAsFactors'])
  ]")

  make_linter_from_function_xpath(
    function_names = "data.frame",
    xpath = xpath,
    lint_message =
      "Supply an explicit value for stringsAsFactors for this code to work before and after R version 4.0.",
    type = "warning"
  )
})

#' Force consecutive calls to assertions into just one when possible
#'
#' [base::stopifnot()] accepts any number of tests, so sequences like
#'   `stopifnot(x); stopifnot(y)` are redundant. Ditto for tests using
#'   `assertthat::assert_that()` without specifying `msg=`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "stopifnot(x); stopifnot(y)",
#'   linters = consecutive_assertion_linter()
#' )
#'
#' lint(
#'   text = "assert_that(x); assert_that(y)",
#'   linters = consecutive_assertion_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "stopifnot(x, y)",
#'   linters = consecutive_assertion_linter()
#' )
#'
#' lint(
#'   text = 'assert_that(x, msg = "Bad x!"); assert_that(y)',
#'   linters = consecutive_assertion_linter()
#' )
#'
#' @evalRd rd_tags("consecutive_assertion_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
consecutive_assertion_linter <- function() {
  # annoying expr-but-not-really nodes
  next_expr <- "following-sibling::*[self::expr or self::expr_or_assign_or_help or self::equal_assign][1]"

  stopifnot_xpath <- glue("
  parent::expr[
    expr[1]/SYMBOL_FUNCTION_CALL = {next_expr}/expr[1]/SYMBOL_FUNCTION_CALL
  ]")
  assert_that_xpath <- glue("
  parent::expr[
    not(SYMBOL_SUB[text() = 'msg'])
    and not(following-sibling::expr[1]/SYMBOL_SUB[text() = 'msg'])
    and expr[1]/SYMBOL_FUNCTION_CALL = {next_expr}/expr[1]/SYMBOL_FUNCTION_CALL
  ]")

  Linter(linter_level = "file", function(source_expression) {
    # need the full file to also catch usages at the top level
    stopifnot_calls <- source_expression$xml_find_function_calls("stopifnot")
    assert_that_calls <- source_expression$xml_find_function_calls("assert_that")
    bad_expr <- combine_nodesets(
      xml_find_all_(stopifnot_calls, stopifnot_xpath),
      xml_find_all_(assert_that_calls, assert_that_xpath)
    )

    matched_function <- xp_call_name(bad_expr)
    xml_nodes_to_lints(
      bad_expr,
      source_expression,
      lint_message = sprintf("Unify consecutive calls to %s().", matched_function),
      type = "warning"
    )
  })
}

#' Semicolon linter
#'
#' Check that no semicolons terminate expressions.
#'
#' @param allow_compound Logical, default `FALSE`. If `TRUE`, "compound"
#'   semicolons (e.g. as in `x; y`, i.e., on the same line of code) are allowed.
#' @param allow_trailing Logical, default `FALSE`. If `TRUE`, "trailing"
#'   semicolons (i.e., those that terminate lines of code) are allowed.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "a <- 1;",
#'   linters = semicolon_linter()
#' )
#'
#' lint(
#'   text = "a <- 1; b <- 1",
#'   linters = semicolon_linter()
#' )
#'
#' lint(
#'   text = "function() { a <- 1; b <- 1 }",
#'   linters = semicolon_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "a <- 1",
#'   linters = semicolon_linter()
#' )
#'
#' lint(
#'   text = "a <- 1;",
#'   linters = semicolon_linter(allow_trailing = TRUE)
#' )
#'
#' code_lines <- "a <- 1\nb <- 1"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = semicolon_linter()
#' )
#'
#' lint(
#'   text = "a <- 1; b <- 1",
#'   linters = semicolon_linter(allow_compound = TRUE)
#' )
#'
#' code_lines <- "function() { \n  a <- 1\n  b <- 1\n}"
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = semicolon_linter()
#' )
#'
#' @evalRd rd_tags("semicolon_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#semicolons>
#' @export
semicolon_linter <- function(allow_compound = FALSE, allow_trailing = FALSE) {
  msg_trailing <- "Remove trailing semicolons."
  msg_compound <- "Replace compound semicolons by a newline."

  if (allow_compound && allow_trailing) {
    cli_abort(c(
      x = "At least one of {.arg allow_compound} or {.arg allow_trailing} must be `FALSE`.",
      i = "No lints can be generated otherwise."
    ))
  } else if (allow_compound && !allow_trailing) {
    # lint only trailing
    xpath <- "//OP-SEMICOLON[not(@line1 = following-sibling::*[1]/@line1)]"
    msg <- msg_trailing # nolint: object_usage. (usage in linter, since need_detection == FALSE)
    need_detection <- FALSE
  } else if (!allow_compound && allow_trailing) {
    # lint only compound
    xpath <- "//OP-SEMICOLON[@line1 = following-sibling::*[1]/@line1]"
    msg <- msg_compound # nolint: object_usage. (usage in linter, since need_detection == FALSE)
    need_detection <- FALSE
  } else {
    # lint all
    xpath <- "//OP-SEMICOLON"
    need_detection <- TRUE
  }
  compound_xpath <- "self::*[@line1 = following-sibling::*[1]/@line1]"

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    bad_exprs <- xml_find_all_(xml, xpath)
    if (need_detection) {
      is_trailing <- is.na(xml_find_first_(bad_exprs, compound_xpath))
      msg <- ifelse(is_trailing, msg_trailing, msg_compound)
    }

    xml_nodes_to_lints(
      bad_exprs,
      source_expression = source_expression,
      lint_message = msg
    )
  })
}

#' Block usages like !(x == y) where a direct relational operator is appropriate
#'
#' `!(x == y)` is more readably expressed as `x != y`. The same is true of
#'   other negations of simple comparisons like `!(x > y)` and `!(x <= y)`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "!x == 2",
#'   linters = comparison_negation_linter()
#' )
#'
#' lint(
#'   text = "!(x > 2)",
#'   linters = comparison_negation_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "!(x == 2 & y > 2)",
#'   linters = comparison_negation_linter()
#' )
#'
#' lint(
#'   text = "!(x & y)",
#'   linters = comparison_negation_linter()
#' )
#'
#' lint(
#'   text = "x != 2",
#'   linters = comparison_negation_linter()
#' )
#'
#' @evalRd rd_tags("comparison_negation_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
comparison_negation_linter <- function() {
  comparators <- infix_metadata$xml_tag[infix_metadata$comparator]

  comparator_inverses <- c(EQ = "!=", NE = "==", GE = "<", LE = ">", GT = "<=", LT = ">=")

  # outline
  #   - match !
  #   - avoid false positives like `!!x > y`
  #     + two conditions: one for the "outer" ! and one for the "inner" !
  #   - avoid false positives like !any(x > y)
  #   - avoid false positives like !x[f == g]
  #   - make sure to catch both !(x == y) and !x == y
  xpath <- glue("
  //OP-EXCLAMATION
    /parent::expr[
      not(parent::expr[OP-EXCLAMATION])
      and expr[
        not(expr[SYMBOL_FUNCTION_CALL] or OP-EXCLAMATION or OP-LEFT-BRACKET)
        and (
          expr[ {xp_or(comparators)} ]
          or {xp_or(comparators)}
        )
      ]
    ]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)

    comparator_node <- xml_find_first_(bad_expr, "expr/expr/*[not(self::COMMENT)][2]")
    comparator_name <- xml_name_(comparator_node)

    # "typical" case is assumed to be !(x == y), so try that first, and back
    #   up to the less nested case. there may be a cleaner way to do this...
    unnested <- !comparator_name %in% names(comparator_inverses)
    comparator_node[unnested] <- xml_find_first_(bad_expr[unnested], "expr/*[not(self::COMMENT)][2]")
    comparator_name[unnested] <- xml_name_(comparator_node[unnested])

    comparator_text <- xml_text(comparator_node)
    inverse <- comparator_inverses[comparator_name]

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = paste("Use x", inverse, "y, not !(x", comparator_text, "y)."),
      type = "warning"
    )
  })
}

#' Recommend usage of a portable `mode` value for downloading files
#'
#' `mode = "w"` (the default) or `mode = "a"` in `download.file()` can generate broken files
#' on Windows. Instead, [utils::download.file()] recommends the usage of `mode = "wb"`
#' and `mode = "ab"`.
#' If `method = "curl"` or `method = "wget"`, no `mode` should be provided as it will be ignored.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "download.file(x = my_url)",
#'   linters = download_file_linter()
#' )
#'
#' lint(
#'   text = "download.file(x = my_url, mode = 'w')",
#'   linters = download_file_linter()
#' )
#'
#' lint(
#'   text = "download.file(x = my_url, method = 'curl', mode = 'wb')",
#'   linters = download_file_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "download.file(x = my_url, mode = 'wb')",
#'   linters = download_file_linter()
#' )
#'
#' lint(
#'   text = "download.file(x = my_url, method = 'curl')",
#'   linters = download_file_linter()
#' )
#'
#' @evalRd rd_tags("download_file_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
download_file_linter <- function() {

  Linter(linter_level = "expression", function(source_expression) {
    download_file_calls <- source_expression$xml_find_function_calls("download.file")

    download_file_mode <- get_r_string(
      download_file_calls,
      "parent::expr[1]/
        SYMBOL_SUB[text() = 'mode']/
        following-sibling::expr[1]/STR_CONST
      "
    )

    has_implicit_mode <- is.na(download_file_mode)

    # According to download.file() documentation, mode is ignored when method
    # is "curl" or "wget" so if it's explicitly specified, we should let the
    # user know they might not get the desired effect
    download_file_method <- get_r_string(
      download_file_calls,
      "parent::expr[1]/
        SYMBOL_SUB[text() = 'method']/
        following-sibling::expr[1]/STR_CONST
      "
    )
    has_ignored_mode <- !is.na(download_file_method) & download_file_method %in% c("wget", "curl")

    ignored_mode_lint <- xml_nodes_to_lints(
      download_file_calls[!has_implicit_mode & has_ignored_mode],
      source_expression = source_expression,
      lint_message = sprintf(
        "mode argument value is ignored for download.file(method = '%s').",
        download_file_method[has_ignored_mode]
      ),
      type = "warning"
    )

    implicit_mode_lint <- xml_nodes_to_lints(
      download_file_calls[has_implicit_mode & !has_ignored_mode],
      source_expression = source_expression,
      lint_message = "download.file() should use mode = 'wb' or ('ab') rather than relying on the default mode = 'w'.",
      type = "warning"
    )

    wrong_mode <- !has_implicit_mode & !has_ignored_mode & download_file_mode %in% c("a", "w")

    explicit_wrong_mode_lint <- xml_nodes_to_lints(
      download_file_calls[wrong_mode],
      source_expression = source_expression,
      lint_message = sprintf(
        "download.file() should use mode = '%1$sb' rather than mode = '%1$s' for portability on Windows.",
        download_file_mode[wrong_mode]
      ),
      type = "warning"
    )

    c(implicit_mode_lint, ignored_mode_lint, explicit_wrong_mode_lint)
  })

}

#' Require usage of `anyDuplicated(x) > 0` over `any(duplicated(x))`
#'
#' [anyDuplicated()] exists as a replacement for `any(duplicated(.))`, which is
#'   more efficient for simple objects, and is at worst equally efficient.
#'   Therefore, it should be used in all situations instead of the latter.
#'
#' Also match usage like `length(unique(x$col)) == nrow(x)`, which can
#'   be replaced by `anyDuplicated(x$col) == 0L`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "any(duplicated(x), na.rm = TRUE)",
#'   linters = any_duplicated_linter()
#' )
#'
#' lint(
#'   text = "length(unique(x)) == length(x)",
#'   linters = any_duplicated_linter()
#' )
#'
#' lint(
#'   text = "DT[, uniqueN(col) == .N]",
#'   linters = any_duplicated_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "anyDuplicated(x)",
#'   linters = any_duplicated_linter()
#' )
#'
#' lint(
#'   text = "anyDuplicated(x) == 0L",
#'   linters = any_duplicated_linter()
#' )
#'
#' lint(
#'   text = "anyDuplicated(DT, by = 'col') == 0L",
#'   linters = any_duplicated_linter()
#' )
#'
#' @evalRd rd_tags("any_duplicated_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
any_duplicated_linter <- function() {
  any_duplicated_xpath <- "
  following-sibling::expr[1][expr[1]/SYMBOL_FUNCTION_CALL[text() = 'duplicated']]
    /parent::expr[
      count(expr) = 2
      or (count(expr) = 3 and SYMBOL_SUB[text() = 'na.rm'])
    ]
  "

  # outline:
  #   EQ/NE/GT/LT: ensure we're in a comparison clause
  #   check for length(unique()) matching:
  #     - length(unique( _ )) == length( _ )
  #     - length(unique( << _$col or _[["col"]] >> )) == nrow( _ )
  # NB: parent::expr/.../following-sibling::expr is the path from
  #  the expr of the unique() call to the call that needs to match.
  #  the final parent::expr/expr gets us to the expr on the other side of EQ;
  #  this lets us match on either side of EQ, where following-sibling
  #  assumes we are before EQ, preceding-sibling assumes we are after EQ.
  length_unique_xpath_parts <- glue("
  expr/expr[
    expr[1]/SYMBOL_FUNCTION_CALL[text() = 'length']
    and expr/expr[1][
      SYMBOL_FUNCTION_CALL[text() = 'unique']
      and (
        following-sibling::expr =
          parent::expr
            /parent::expr
            /parent::expr
            /expr
            /expr[1][SYMBOL_FUNCTION_CALL[text() = 'length']]
            /following-sibling::expr
        or following-sibling::expr[OP-DOLLAR or LBB]/expr[1] =
          parent::expr
            /parent::expr
            /parent::expr
            /expr
            /expr[1][SYMBOL_FUNCTION_CALL[text() = 'nrow']]
            /following-sibling::expr
        or parent::expr
          /parent::expr
          /parent::expr
          /expr[
            SYMBOL[text() = '.N']
            or (expr/SYMBOL_FUNCTION_CALL[text() = 'n'] and count(expr) = 1)
          ]
      )
    ]
  ]
  ")
  length_unique_xpath <- paste(length_unique_xpath_parts, collapse = " | ")

  distinct_xpath <- glue("
  expr/expr[
    expr[1][
      SYMBOL_FUNCTION_CALL[text() = 'uniqueN' or text() = 'n_distinct']
      and (
        following-sibling::expr =
          parent::expr
            /parent::expr
            /expr
            /expr[1][SYMBOL_FUNCTION_CALL[text() = 'length' or text() = 'nrow']]
            /following-sibling::expr
        or following-sibling::expr[OP-DOLLAR or LBB]/expr[1] =
          parent::expr
            /parent::expr
            /expr
            /expr[1][SYMBOL_FUNCTION_CALL[text() = 'nrow']]
            /following-sibling::expr
        or parent::expr
          /parent::expr
          /expr[
            SYMBOL[text() = '.N']
            or (expr/SYMBOL_FUNCTION_CALL[text() = 'n'] and count(expr) = 1)
          ]
      )
    ]
  ]
  ")

  uses_nrow_xpath <- "./parent::expr/expr/expr[1]/SYMBOL_FUNCTION_CALL[text() = 'nrow']"
  uses_dtn_xpath <- "./parent::expr/expr/SYMBOL[text() = '.N']"
  uses_dplyr_xpath <- "./parent::expr/expr/expr[1]/SYMBOL_FUNCTION_CALL[text() = 'n']"

  Linter(linter_level = "expression", function(source_expression) {
    # NB: need two parents given three parent::expr in XPath and stripped comments.
    xml <- source_expression$xml_parsed_content |>
      xml_find_all_("(//EQ | //NE | //GT | //LT)/parent::*/parent::*") |>
      strip_comments_from_subtree()

    xml_calls <- source_expression$xml_find_function_calls("any")

    any_duplicated_expr <- xml_find_all_(xml_calls, any_duplicated_xpath)
    any_duplicated_lints <- xml_nodes_to_lints(
      any_duplicated_expr,
      source_expression = source_expression,
      lint_message = "anyDuplicated(x, ...) > 0 is better than any(duplicated(x), ...).",
      type = "warning"
    )

    length_unique_expr <- xml_find_all_(xml, length_unique_xpath)
    length_unique_lint_message <- character(length(length_unique_expr))
    length_unique_lint_message[] <- "anyDuplicated(x) == 0L is better than length(unique(x)) == length(x)."
    length_unique_lint_message[!is.na(xml_find_first_(length_unique_expr, uses_nrow_xpath))] <-
      "anyDuplicated(DF$col) == 0L is better than length(unique(DF$col)) == nrow(DF)"
    length_unique_lint_message[!is.na(xml_find_first_(length_unique_expr, uses_dtn_xpath))] <-
      "anyDuplicated(x) == 0L is better than length(unique(x)) == .N"
    length_unique_lint_message[!is.na(xml_find_first_(length_unique_expr, uses_dplyr_xpath))] <-
      "anyDuplicated(x) == 0L is better than length(unique(x)) == n()."
    length_unique_lints <- xml_nodes_to_lints(
      length_unique_expr,
      source_expression = source_expression,
      lint_message = length_unique_lint_message,
      type = "warning"
    )

    distinct_expr <- xml_find_all_(xml, distinct_xpath)
    distinct_lint_message_fmt <- character(length(distinct_expr))
    distinct_lint_message_fmt[] <- "anyDuplicated(x) == 0L is better than %s(x) == length(x)."
    distinct_lint_message_fmt[!is.na(xml_find_first_(distinct_expr, uses_nrow_xpath))] <-
      "anyDuplicated(DF$col) == 0L is better than %s(DF$col) == nrow(DF)"
    distinct_lint_message_fmt[!is.na(xml_find_first_(distinct_expr, uses_dtn_xpath))] <-
      "anyDuplicated(x) == 0L is better than %s(x) == .N"
    distinct_lint_message_fmt[!is.na(xml_find_first_(distinct_expr, uses_dplyr_xpath))] <-
      "anyDuplicated(x) == 0L is better than %s(x) == n()."

    distinct_lint_message <- sprintf(distinct_lint_message_fmt, xp_call_name(distinct_expr))
    distinct_lints <- xml_nodes_to_lints(
      distinct_expr,
      source_expression = source_expression,
      lint_message = distinct_lint_message,
      type = "warning"
    )

    c(any_duplicated_lints, length_unique_lints, distinct_lints)
  })
}

#' Object name linter
#'
#' Check that object names conform to a naming style.
#' The default naming styles are "snake_case" and "symbols".
#'
#' Quotes (`` `"' ``) and specials (`%` and trailing `<-`) are not considered part of the object name.
#'
#' Note when used in a package, in order to ignore objects imported
#'   from other namespaces, this linter will attempt [getNamespaceExports()]
#'   whenever an `import(PKG)` or `importFrom(PKG, ...)` statement is found
#'   in your NAMESPACE file. If [requireNamespace()] fails (e.g., the package
#'   is not yet installed), the linter won't be able to ignore some usages
#'   that would otherwise be allowed.
#'
#' Suppose, for example, you have `import(upstream)` in your NAMESPACE,
#'   which makes available its exported S3 generic function
#'   `a_really_quite_long_function_name` that you then extend in your package
#'   by defining a corresponding method for your class `my_class`.
#'   Then, if `upstream` is not installed when this linter runs, a lint
#'   will be thrown on this object (even though you don't "own" its full name).
#'
#' The best way to get lintr to work correctly is to install the package so
#'   that it's available in the session where this linter is running.
#'
#' @param styles A subset of
#'   \Sexpr[stage=render, results=rd]{lintr:::regexes_rd}. A name should
#'   match at least one of these styles. The `"symbols"` style refers to
#'   names containing *only* non-alphanumeric characters; e.g., defining `%+%`
#'   from ggplot2 or `%>%` from magrittr would not generate lint markers,
#'   whereas `%m+%` from lubridate (containing both alphanumeric *and*
#'   non-alphanumeric characters) would.
#'
#' @param regexes A (possibly named) character vector specifying a custom naming convention.
#'   If named, the names will be used in the lint message. Otherwise, the regexes enclosed by `/` will be used in the
#'   lint message.
#'   Note that specifying `regexes` overrides the default `styles`. So if you want to combine `regexes` and `styles`,
#'   both need to be explicitly specified.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "my_var <- 1L",
#'   linters = object_name_linter(styles = "CamelCase")
#' )
#'
#' lint(
#'   text = "xYz <- 1L",
#'   linters = object_name_linter(styles = c("UPPERCASE", "lowercase"))
#' )
#'
#' lint(
#'   text = "MyVar <- 1L",
#'   linters = object_name_linter(styles = "dotted.case")
#' )
#'
#' lint(
#'   text = "asd <- 1L",
#'   linters = object_name_linter(regexes = c(my_style = "F$", "f$"))
#' )
#'
#' # okay
#' lint(
#'   text = "my_var <- 1L",
#'   linters = object_name_linter(styles = "snake_case")
#' )
#'
#' lint(
#'   text = "xyz <- 1L",
#'   linters = object_name_linter(styles = "lowercase")
#' )
#'
#' lint(
#'   text = "my.var <- 1L; myvar <- 2L",
#'   linters = object_name_linter(styles = c("dotted.case", "lowercase"))
#' )
#'
#' lint(
#'   text = "asdf <- 1L; asdF <- 1L",
#'   linters = object_name_linter(regexes = c(my_style = "F$", "f$"))
#' )
#'
#' @evalRd rd_tags("object_name_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
object_name_linter <- function(styles = c("snake_case", "symbols"), regexes = character()) {
  if ((!missing(styles) || missing(regexes)) && length(styles) > 0L) {
    # Allow `object_name_linter(NULL, "my_regex")`
    styles <- match.arg(styles, names(style_regexes), several.ok = TRUE)
    style_list <- style_regexes[styles]
  } else {
    style_list <- list()
  }
  if (length(regexes) > 0L) {
    if (!is.character(regexes)) {
      cli_abort("{.arg regexes} must be a {.cls character} vector.")
    }
    rx_names <- names2(regexes)
    missing_name <- !nzchar(rx_names)
    rx_names[missing_name] <- paste0("/", regexes[missing_name], "/") # auto-name regex "asd" -> /asd/
    names(regexes) <- rx_names

    style_list <- c(style_list, as.list(regexes))
  }
  if (length(style_list) == 0L) {
    cli_abort("At least one style must be specified using {.arg styles} or {.arg regexes}.")
  }

  lint_message <- paste0(
    "Variable and function name style should match ",
    glue_collapse(unique(names(style_list)), sep = ", ", last = " or "), "."
  )

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    assignments <- xml_find_all_(xml, object_name_xpath)

    # Retrieve assigned name
    nms <- strip_names(
      xml_text(assignments)
    )

    # run namespace_imports at run-time, not "compile" time to allow package structure to change
    pkg <- find_package(source_expression$filename)
    generics <- c(
      declared_s3_generics(xml),
      imported_s3_generics(namespace_imports(pkg))$fun,
      exported_s3_generics(pkg)$fun,
      .base_s3_generics
    )
    generics <- unique(generics[nzchar(generics)])

    style_matches <- lapply(style_list, function(style) {
      check_style(nms, style, generics)
    })

    matches_a_style <- Reduce(`|`, style_matches)

    xml_nodes_to_lints(
      assignments[!matches_a_style],
      source_expression,
      lint_message = lint_message,
      type = "style"
    )
  })
}

check_style <- function(nms, style, generics = character()) {
  conforming <- re_matches_logical(nms, style)

  # mark empty or NA names as conforming
  conforming <- is.na(nms) | !nzchar(nms) | conforming

  if (!all(conforming)) {
    possible_s3 <- re_matches(
      nms[!conforming],
      rex(start, capture(name = "generic", or(generics)), ".", capture(name = "method", something), end)
    )
    if (!all(is.na(possible_s3))) {
      has_generic <- possible_s3$generic %in% generics

      # If they are not conforming, but are S3 methods then ignore them
      conforming[!conforming][has_generic] <- TRUE
    }
    # exclude namespace hooks like .onLoad, .Last.lib, etc (#500) and ...
    is_special <- is_special_function(nms[!conforming]) | nms[!conforming] == "..."
    conforming[!conforming][is_special] <- TRUE
  }
  conforming
}

loweralnum <- rex(one_of(lower, digit))
upperalnum <- rex(one_of(upper, digit))

style_regexes <- list(
  symbols     = rex(start, zero_or_more(none_of(alnum)), end),
  CamelCase   = rex(start, maybe("."), upper, zero_or_more(alnum), end),
  camelCase   = rex(start, maybe("."), lower, zero_or_more(alnum), end),
  snake_case  = rex(start, maybe("."), some_of(lower, digit), any_of("_", lower, digit), end),
  SNAKE_CASE  = rex(start, maybe("."), some_of(upper, digit), any_of("_", upper, digit), end),
  dotted.case = rex(start, maybe("."), one_or_more(loweralnum), zero_or_more(dot, one_or_more(loweralnum)), end),
  lowercase   = rex(start, maybe("."), one_or_more(loweralnum), end),
  UPPERCASE   = rex(start, maybe("."), one_or_more(upperalnum), end)
)

regexes_rd <- toString(paste0("\\sQuote{", names(style_regexes), "}"))

# Each of the default linters should throw at least one lint on this file

# assignment
# function_left_parentheses
# closed_curly
# commas
# paren_brace
f = function (x,y = 1){}

# commented_code
# some <- commented("out code")

# cyclocomp
# equals_na
# infix_spaces
# line_length
# object_length
# object_name
# object_usage
# open_curly
# T_and_F_symbol
someComplicatedFunctionWithALongCamelCaseName <- function(x)
{
  y <- 1
  if (1 > 2 && 2 > 3 && 3 > 4 && 4 > 5 && 5*10 > 6 && 5 > 6 && 6 > 7 && x == NA) {T} else {F}
}

# no_tab
# pipe_continuation
# seq_linter
# spaces_inside
x <- 1:10
x[ 2]
1:length(x) %>% lapply(function(x) x*2) %>%
	head()

# single_quotes
message('single_quotes')

# spaces_left_parentheses
# trailing_whitespace
# semicolon
x <- 42; y <- 2 +(1:10) 

# trailing_blank_lines

#' Block usage like x %in% "a"
#'
#' `vector %in% set` is appropriate for matching a vector to a set, but if
#'   that set has size 1, `==` is more appropriate. However, if `vector` has
#'   also size 1 and can be `NA`, the use of `==` should be accompanied by extra
#'   protection for the missing case (for example, `isTRUE(NA == "arg")` or
#'   `!is.na(x) && x == "arg"`).
#'
#' `scalar %in% vector` is OK, because the alternative (`any(vector == scalar)`)
#'   is more circuitous & potentially less clear.
#'
#' @param in_operators Character vector of additional infix operators that behave like the `%in%` operator,
#'   e.g. `{data.table}`'s `%chin%` operator.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x %in% 1L",
#'   linters = scalar_in_linter()
#' )
#'
#' lint(
#'   text = "x %chin% 'a'",
#'   linters = scalar_in_linter(in_operators = "%chin%")
#' )
#'
#' # okay
#' lint(
#'   text = "x %in% 1:10",
#'   linters = scalar_in_linter()
#' )
#'
#' @evalRd rd_tags("scalar_in_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
scalar_in_linter <- function(in_operators = NULL) {
  # TODO(#2085): Extend to include other cases where the RHS is clearly a scalar
  # NB: all of logical, integer, double, hex, complex are parsed as NUM_CONST
  xpath <- glue("
  //SPECIAL[{xp_text_in_table(c('%in%', {in_operators}))}]
    /following-sibling::expr[NUM_CONST[not(starts-with(text(), 'NA'))] or STR_CONST]
    /parent::expr
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)
    in_op <- xml_find_chr_(bad_expr, "string(SPECIAL)")
    lint_msg <- paste0(
      "Use comparison operators (e.g. ==, !=, etc.) to match length-1 scalars instead of ", in_op, ". ",
      "Note that if x can be NA, x == 'arg' is NA whereas x ", in_op, " 'arg' is FALSE, ",
      "so consider extra protection for the missing case in your code."
    )

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = lint_msg,
      type = "warning"
    )
  })
}

# config excludes assignment_linter() so this doesn't lint
a = 1
# default config includes infix_spaces_linter() so this lints
b=a + 2
# config extends defaults with any_duplicated_linter() so this lints
any(duplicated(b))
# custom exclude setting is also picked up so this doesn't lint
1+1 # NOLINT

#' Get Linter metadata from a package
#'
#' `available_linters()` obtains a tagged list of all Linters available in a package.
#'
#' @param packages A character vector of packages to search for linters.
#' @param tags Optional character vector of tags to search. Only linters with at least one matching tag will be
#'   returned. If `tags` is `NULL`, all linters will be returned. See `available_tags("lintr")` to find out what
#'   tags are already used by lintr.
#' @param exclude_tags Tags to exclude from the results. Linters with at least one matching tag will not be returned.
#'   If `exclude_tags` is `NULL`, no linters will be excluded. Note that `tags` takes priority, meaning that any
#'   tag found in both `tags` and `exclude_tags` will be included, not excluded. Note that linters with tag `"defunct"`
#'   (which do not work and can no longer be run) cannot be queried directly.
#'
#' @section Package Authors:
#'
#' To implement `available_linters()` for your package, include a file `inst/lintr/linters.csv` in your
#' package.
#' The CSV file must contain the columns 'linter' and 'tags', and be UTF-8 encoded.
#' Additional columns will be silently ignored if present and the columns are identified by name.
#' Each row describes a linter by
#'
#'  1. its function name (e.g. `"assignment_linter"`) in the column 'linter'.
#'  2. space-separated tags associated with the linter (e.g. `"style consistency default"`) in the column 'tags'.
#'
#' Tags should be snake_case.
#'
#' See `available_tags("lintr")` to find out what tags are already used by lintr.
#'
#' @return
#' `available_linters` returns a data frame with columns 'linter', 'package' and 'tags':
#'
#' \describe{
#' \item{linter}{A character column naming the function associated with the linter.}
#' \item{package}{A character column containing the name of the package providing the linter.}
#' \item{tags}{A list column containing tags associated with the linter.}
#' }
#'
#' @examples
#' lintr_linters <- available_linters()
#'
#' # If the package doesn't exist or isn't installed, an empty data frame will be returned
#' available_linters("does-not-exist")
#'
#' lintr_linters2 <- available_linters(c("lintr", "does-not-exist"))
#' identical(lintr_linters, lintr_linters2)
#' @seealso
#'  - [linters] for a complete list of linters available in lintr.
#'  - [available_tags()] to retrieve the set of valid tags.
#' @export
available_linters <- function(packages = "lintr", tags = NULL, exclude_tags = "deprecated") {
  if (!is.character(packages)) {
    cli_abort("{.arg packages} must be a {.cls character} vector.")
  }
  if (!is.null(tags) && !is.character(tags)) {
    cli_abort("{.arg tags} must be a {.cls character} vector.")
  }
  if (!is.null(exclude_tags) && !is.character(exclude_tags)) {
    cli_abort("{.arg exclude_tags} must be a {.cls character} vector.")
  }

  # any tags specified explicitly will not be excluded (#1959)
  # never include defunct linters, which don't work / error on instantiation (#2284).
  exclude_tags <- unique(c(setdiff(exclude_tags, tags), "defunct"))

  # Handle multiple packages
  if (length(packages) > 1L) {
    return(do.call(rbind, lapply(packages, available_linters, tags = tags, exclude_tags = exclude_tags)))
  }

  csv_file <- system.file("lintr", "linters.csv", package = packages)

  if (!file.exists(csv_file)) {
    return(empty_linters())
  }
  available <- utils::read.csv(csv_file, encoding = "UTF-8", as.is = TRUE)

  if (!validate_linter_db(available, packages)) {
    return(empty_linters())
  }

  build_available_linters(available, packages, tags, exclude_tags)
}

build_available_linters <- function(available, package, tags, exclude_tags) {
  available_df <- data.frame(linter = available[["linter"]], package)
  available_df$tags <- strsplit(available[["tags"]], split = " ", fixed = TRUE)
  if (!is.null(tags)) {
    matches_tags <- vapply(available_df$tags, \(linter_tags) any(linter_tags %in% tags), logical(1L))
    available_df <- available_df[matches_tags, ]
  }
  if (!is.null(exclude_tags)) {
    matches_exclude <- vapply(available_df$tags, \(linter_tags) any(linter_tags %in% exclude_tags), logical(1L))
    available_df <- available_df[!matches_exclude, ]
  }

  # Due to removal of deprecated linters in the returned data frame, there can be gaps in row numbers.
  # To avoid this inconsistency, regenerate row names.
  rownames(available_df) <- NULL
  available_df
}

#' Make sure we always return a valid data frame
#'
#' `data.frame` constructors don't handle zero-row list-columns properly, so supply `tags` afterwards.
#' @noRd
empty_linters <- function() {
  empty_df <- data.frame(linter = character(), package = character())
  empty_df$tags <- list()
  empty_df
}

validate_linter_db <- function(available, package) {
  # Check that the csv file contains two character columns, named 'linter' and 'tags'.
  # Otherwise, fallback to an empty data frame.
  if (!all(c("linter", "tags") %in% colnames(available))) {
    cli_warn(c(
      i = "{.file linters.csv} must contain the columns {.val {c('linter', 'tags')}}.",
      x = "Package {.pkg package} is missing {.str {setdiff(c('linter', 'tags'), names(available))}}."
    ))
    return(FALSE)
  }
  nrow(available) > 0L
}

#' @rdname available_linters
#'
#' @description
#' `available_tags()` searches for available tags.
#'
#' @return `available_tags` returns a character vector of linter tags used by the packages.
#' @export
#' @examples
#' available_tags()
available_tags <- function(packages = "lintr") {
  available_linters(packages = packages, exclude_tags = NULL)[["tags"]] |>
    unlist() |>
    unique() |>
    platform_independent_sort()
}

# nocov start

#' Generate Rd fragment for the Tags section of a linter
#'
#' @param linter_name Name of the linter to generate Rd code for.
#'
#' @noRd
rd_tags <- function(linter_name) {
  linters <- available_linters(exclude_tags = NULL)
  tags <- platform_independent_sort(linters[["tags"]][[match(linter_name, linters[["linter"]])]])
  if (length(tags) == 0L) {
    cli_abort("Tags are required, but found none for {.fn {linter_name}}.")
  }

  c(
    "\\section{Tags}{",
    paste0("\\link[=", tags, "_linters]{", tags, "}", collapse = ", "),
    "}"
  )
}

#' Generate Rd fragment for the Linters section of a tag
#'
#' @param tag_name Name of the tag to generate Rd code for.
#'
#' @noRd
rd_linters <- function(tag_name) {
  linters <- available_linters(tags = tag_name)
  tagged <- platform_independent_sort(linters[["linter"]])
  if (length(tagged) == 0L) {
    section_body <- paste0("There are not currently any linters tagged with '", tag_name, "'.")
  } else {
    section_body <- c(
      paste0("The following linters are tagged with '", tag_name, "':"),
      "\\itemize{",
      paste0("\\item{\\code{\\link{", tagged, "}}}"),
      "}"
    )
  }

  c("\\section{Linters}{", section_body, "}")
}

#' Generate Rd fragment for the main help page, listing all tags
#'
#' @noRd
rd_taglist <- function() {
  linters <- available_linters(exclude_tags = NULL)
  # don't count tags on deprecated linters to the counts of other tags
  linters$tags <- lapply(linters$tags, \(x) if ("deprecated" %in% x) "deprecated" else x)

  tag_table <- table(unlist(linters[["tags"]]))
  tags <- platform_independent_sort(names(tag_table))
  # re-order
  tag_table <- tag_table[tags]

  c(
    "\\section{Tags}{",
    "The following tags exist:",
    "\\itemize{",
    vapply(tags, function(tag) {
      paste0("\\item{\\link[=", tag, "_linters]{", tag, "} (", tag_table[[tag]], " linters)}")
    }, character(1L)),
    "}", # itemize
    "}" # section
  )
}

#' Generate Rd fragment for the main help page, listing all linters
#'
#' @noRd
rd_linterlist <- function() {
  linters <- available_linters()
  linter_names <- platform_independent_sort(linters[["linter"]])

  c(
    "\\section{Linters}{",
    "The following linters exist:",
    "\\itemize{",
    vapply(linter_names, function(linter_name) {
      tags <- platform_independent_sort(linters[["tags"]][[match(linter_name, linters[["linter"]])]])
      paste0("\\item{\\code{\\link{", linter_name, "}} (tags: ", toString(tags), ")}")
    }, character(1L)),
    "}", # itemize
    "}" # section
  )
}

# nocov end

test_that("input validation for available_linters works as expected", {
  expect_error(available_linters(1L), "`packages` must be a <character> vector.")
  expect_error(available_linters(tags = 1L), "`tags` must be a <character> vector.")
  expect_error(available_linters(exclude_tags = 1L), "`exclude_tags` must be a <character> vector.")
})

test_that("available_linters returns a data frame", {
  avail <- available_linters()
  avail2 <- available_linters(c("lintr", "not-a-package"))
  empty <- available_linters("not-a-package")

  expect_s3_class(avail, "data.frame")
  expect_identical(avail, avail2)
  expect_identical(nrow(empty), 0L)
  expect_named(avail, c("linter", "package", "tags"))
  expect_type(avail[["linter"]], "character")
  expect_type(avail[["package"]], "character")
  expect_type(avail[["tags"]], "list")
  expect_type(avail[["tags"]][[1L]], "character")
})

test_that("available_tags returns a character vector", {
  tags <- available_tags()
  tags2 <- available_tags(c("lintr", "not-a-package"))
  empty <- available_tags("not-a-package")

  expect_type(tags, "character")
  expect_identical(tags, tags2)
  expect_length(empty, 0L)
  expect_true(all(available_linters()$tags[[1L]] %in% tags))
  expect_true(all(unlist(available_linters()$tags) %in% tags))
  expect_identical(anyDuplicated(tags), 0L)
})

test_that("default_linters and default tag match up", {
  avail <- available_linters()
  tagged_default <- avail[["linter"]][vapply(avail[["tags"]], \(tags) "default" %in% tags, logical(1L))]
  expect_identical(tagged_default, names(default_linters))
})

test_that("warnings occur only for deprecated linters", {
  skip_if_not_installed("cyclocomp") # actually we expect a warning there

  expect_error(linters_with_tags(), "`tags` was not specified. Available tags:")
  expect_silent(linters_with_tags(tags = NULL))
  num_deprecated_linters <- nrow(available_linters(tags = "deprecated", exclude_tags = NULL))
  outer_env <- new.env(parent = emptyenv())
  outer_env$deprecation_warns_seen <- 0L
  expect_silent({
    withCallingHandlers(
      linters_with_tags(tags = "deprecated", exclude_tags = NULL),
      warning = function(w) {
        if (grepl("was deprecated", conditionMessage(w), fixed = TRUE)) {
          outer_env$deprecation_warns_seen <- outer_env$deprecation_warns_seen + 1L
          invokeRestart("muffleWarning")
        } else {
          w
        }
      }
    )
  })
  expect_identical(outer_env$deprecation_warns_seen, num_deprecated_linters)
})

test_that("available_linters matches the set of linters available from lintr", {
  lintr_db <- available_linters(exclude_tags = NULL)
  defunct_linters <-
    subset(read.csv(system.file("lintr", "linters.csv", package = "lintr"), as.is = TRUE), tags == "defunct")$linter
  ignore <- c("is_linter", defunct_linters)
  linters_in_namespace <- setdiff(ls(asNamespace("lintr"), pattern = "_linter$"), ignore)
  # ensure that the contents of inst/lintr/linters.csv covers all _linter objects in our namespace
  expect_identical(sort(lintr_db$linter), sort(linters_in_namespace))
  # ensure that all _linter objects in our namespace are also exported
  exported_linters <- setdiff(grep("_linter$", getNamespaceExports("lintr"), value = TRUE), ignore)
  expect_identical(sort(linters_in_namespace), sort(exported_linters))
})

test_that("rownames for available_linters data frame doesn't have missing entries", {
  lintr_db <- available_linters()
  expect_identical(
    tail(rownames(lintr_db), 1L),
    as.character(nrow(lintr_db))
  )

  lintr_db2 <- available_linters(exclude_tags = NULL)
  expect_identical(
    tail(rownames(lintr_db2), 1L),
    as.character(nrow(lintr_db2))
  )
})

# See the roxygen helpers in R/linter_tags.R for the code used to generate the docs.
#   This test helps ensure the documentation is up to date with the available_linters() database
test_that("lintr help files are up to date", {
  help_db <- safe_load_help_db()

  lintr_db <- available_linters(exclude_tags = NULL)
  lintr_db$package <- NULL
  lintr_db$tags <- lapply(lintr_db$tags, \(x) if ("deprecated" %in% x) "deprecated" else sort(x))
  lintr_db <- lintr_db[order(lintr_db$linter), ]

  expect_true("linters.Rd" %in% names(help_db), info = "?linters exists")

  # NB: objects in help_env are class Rd, see ?as.character.Rd (part of 'tools')
  rd_as_string <- function(rd) paste(as.character(rd), collapse = "")

  # Get a character string with the contents of ?linters
  linter_help_text <- rd_as_string(help_db$linters.Rd)

  # Test three things about ?linters
  #   (1) the complete list of linters and tags matches that in available_linters()
  #   (2) the tabulation of tags & corresponding count of linters matches that in available_linters()
  #   (3) the 'configurable' tag applies if and only if the linter has parameters

  # Extract linters & associated tags from ?linters text
  # Rd markup for items looks like \item{\code{\link{...}} (tags: ...)}
  #                                                  [1]          [2]
  # [1]: linter name; [2]: associated tags
  help_linters <- rex::re_matches(
    linter_help_text,
    rex::rex(
      "\\item{\\code{\\link{",
      capture(some_of(letter, number, "_", "."), name = "linter"),
      "}} (tags: ",
      capture(some_of(letter, "_", ",", " "), name = "tags"),
      ")}"
    ),
    global = TRUE
  )[[1L]]
  help_linters$tags <- lapply(strsplit(help_linters$tags, ", ", fixed = TRUE), sort)
  help_linters <- help_linters[order(help_linters$linter), ]

  # (1) the complete list of linters and tags matches that in available_linters()
  expect_identical(
    help_linters,
    # deprecated linters are excluded from this page
    lintr_db[!vapply(lintr_db$tags, identical, "deprecated", FUN.VALUE = NA), ],
    info = "Database implied by ?linters is the same as is available_linters()",
    ignore_attr = "row.names"
  )

  # Counts of tags from available_linters()
  #   NB: as.data.frame.table returns stringsAsFactors=TRUE default in R>4
  db_tag_table <- as.data.frame(
    table(tag = unlist(lintr_db$tags)),
    responseName = "n_linters",
    stringsAsFactors = FALSE
  )
  # In ?linters, entries in the enumeration of tags look like
  #   \item{\link[=${TAG}_linters]{${TAG}} (${N_LINTERS_WITH_TAG} linters)}
  help_tag_table <- rex::re_matches(
    linter_help_text,
    rex::rex(
      "\\item{\\link[=",
      capture(some_of(letter, "_"), name = "tag_page"),
      "_linters]{",
      capture(some_of(letter, "_"), name = "tag"),
      "} (",
      capture(numbers, name = "n_linters"),
      " linters)}"
    ),
    global = TRUE
  )[[1L]]
  # consistency/sanity check
  expect_identical(help_tag_table$tag_page, help_tag_table$tag)
  help_tag_table$tag_page <- NULL
  help_tag_table$n_linters <- as.integer(help_tag_table$n_linters)

  # (2) the tabulation of tags & corresponding count of linters matches that in available_linters()
  expect_identical(
    help_tag_table[order(help_tag_table$tag), ],
    db_tag_table[order(db_tag_table$tag), ],
    ignore_attr = "row.names",
    info = "Tags and corresponding counts in ?linters is the same as in available_linters()"
  )

  # Now test an analogue to (1) from above for each tag's help page
  for (tag in db_tag_table$tag) {
    expect_true(paste0(tag, "_linters.Rd") %in% names(help_db), info = paste0("?", tag, "_linters exists"))

    tag_help <- help_db[[paste0(tag, "_linters.Rd")]]
    tag_help_text <- rd_as_string(tag_help)

    # linters listed in ?${TAG}_linter
    help_tag_linters <- rex::re_matches(
      tag_help_text,
      rex::rex("\\item{\\code{\\link{", capture(some_of(letter, number, "_", "."), name = "linter"), "}}}"),
      global = TRUE
    )[[1L]]

    # those entries in available_linters() with the current tag
    db_linter_has_tag <- vapply(lintr_db$tags, \(linter_tag) any(tag %in% linter_tag), logical(1L))
    expected <- lintr_db$linter[db_linter_has_tag]

    expect_identical(
      sort(help_tag_linters$linter),
      sort(expected),
      info = paste0("?", tag, "_linters lists all linters with that tag in available_linters()")
    )
  }

  # (3) the 'configurable' tag applies if and only if the linter has parameters
  has_args <- 0L < lengths(Map(
    \(linter, tags) if ("deprecated" %in% tags) NULL else formals(match.fun(linter)),
    lintr_db$linter,
    lintr_db$tags
  ))
  has_configurable_tag <- vapply(lintr_db$tags, \(tags) "configurable" %in% tags, logical(1L))

  expect_identical(has_configurable_tag, unname(has_args))
})

test_that("available_linters gives precedence to included tags", {
  expect_true("style" %in% unlist(available_linters(tags = "style", exclude_tags = "style")$tags))
  # also for the default case
  expect_identical(
    available_linters(tags = "deprecated"),
    available_linters(tags = "deprecated", exclude_tags = NULL)
  )
})

test_that("all linters have at least one tag", {
  expect_true(all(lengths(available_linters()$tags) > 0L))
})

test_that("other packages' linters can be included", {
  db_loc <- withr::local_tempdir()
  local_mocked_bindings(
    system.file = \(..., package) file.path(db_loc, package, ...)
  )

  custom_db <- data.frame(
    linter = c("linter1", "linter2"),
    package = "myPkg",
    tags = c("tag1 tag2", "tag1 tag3")
  )
  custom_db$tags <- strsplit(custom_db$tags, " ", fixed = TRUE)
  custom_dir <- file.path(db_loc, "myPkg", "lintr")
  custom_csv <- file.path(custom_dir, "linters.csv")
  dir.create(custom_dir, recursive = TRUE)
  write_csv <- function(x, file) write.csv(x, file, row.names = FALSE, quote = FALSE)
  write_csv(
    within(custom_db, {
      tags <- vapply(tags, paste, collapse = " ", FUN.VALUE = character(1L))
      rm(package)
    }),
    custom_csv
  )

  expect_identical(
    available_linters(packages = "myPkg"),
    custom_db
  )

  # edge case
  write_csv(custom_db[0L, c("linter", "tags")], custom_csv)
  expect_identical(
    available_linters(packages = "myPkg"),
    custom_db[0L, ]
  )

  write_csv(data.frame(a = 1L, b = 1L), custom_csv)
  expect_warning(
    expect_identical(
      available_linters(packages = "myPkg"),
      custom_db[0L, ]
    ),
    'must contain the columns "linter" and "tags"'
  )
})

#' Block usage of comparison operators with known-list() functions like lapply
#'
#' Usage like `lapply(x, sum) > 10` is awkward because the list must first
#'   be coerced to a vector for comparison. A function like [vapply()]
#'   should be preferred.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "lapply(x, sum) > 10",
#'   linters = list_comparison_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "unlist(lapply(x, sum)) > 10",
#'   linters = list_comparison_linter()
#' )
#'
#' @evalRd rd_tags("list_comparison_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
list_comparison_linter <- function() {
  list_mapper_alternatives <- c(
    lapply = "vapply(x, FUN, character(1L))",
    map = "map_chr(x, FUN)",
    Map = "mapply()",
    .mapply = "mapply()"
  )

  # NB: anchor to the comparison expr so that we can easily include the comparator
  #   in the lint message.
  xpath <- glue("
  parent::expr
    /parent::expr[{ xp_or(infix_metadata$xml_tag[infix_metadata$comparator]) }]
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls(names(list_mapper_alternatives))
    bad_expr <- xml_find_all_(xml_calls, xpath)

    list_mapper <- xp_call_name(bad_expr, depth = 2L)

    vector_mapper <- list_mapper_alternatives[list_mapper]
    # we are at `x ? y` in which the comparator ? comes 2nd
    comparator <- xml_find_chr_(bad_expr, "string(*[2])")

    lint_message <- as.character(glue(
      "The output of {list_mapper}(), a list(), is being ",
      "coerced for comparison by `{comparator}`. ",
      "Instead, use a mapper that generates a vector with the correct type ",
      "directly, for example {vector_mapper} if the output is a string."
    ))
    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = lint_message,
      type = "warning"
    )
  })
}

# Helpers for lintr tests

single_quote <- function(x) paste0("'", x, "'")
double_quote <- function(x) paste0('"', x, '"')

#' Trim some leading whitespace
#'
#' Trim `num` characters from the start of every line in `x`, or auto-detect to remove the maximum number whitespace
#' from all lines while preserving relative indentation
#'
#' @param x a string containing newlines
#' @param num number of characters to strip from the start of each line.
#' `NULL` will auto-detect this based on the minimum number of leading spaces greater than one.
#'
#' @return A modified version of `x` with `num` characters removed from the start of every line and with a possible
#' leading and trailing blank line removed.
#'
#' @examples
#' my_var <- local({
#'   out <- trim_some("
#'     This will be the content
#'     of the file where
#'     only the following line
#'       is indented by two spaces.
#'   ")
#' })
#'
#' stopifnot(identical(
#'   my_var,
#'   "This will be the content\nof the file where\nonly the following line\n  is indented by two spaces."
#' ))
trim_some <- function(x, num = NULL) {
  x <- rex::re_substitutes(
    x,
    rex::rex(list(start, any_blanks, newline) %or% list(newline, any_blanks, end)),
    replacement = "",
    global = TRUE
  )

  if (is.null(num)) {
    ms <- rex::re_matches(x, "^\\s+", locations = TRUE, global = TRUE, options = "multi-line")[[1L]]
    num <- min(ms$end - ms$start) + 1L
  }

  rex::re_substitutes(x, rex::rex(start, n_times(any, num)), "", global = TRUE, options = "multi-line")
}

local_config <- function(contents, config_dir = ".", filename = ".lintr", .local_envir = parent.frame()) {
  config_path <- file.path(config_dir, filename)
  writeLines(contents, config_path)
  withr::defer(unlink(config_path), envir = .local_envir)
  config_path
}

with_config <- function(contents, code, config_dir = ".", filename = ".lintr") {
  local_config(contents, config_dir, filename)
  code
}

skip_if_not_utf8_locale <- function() {
  testthat::skip_if_not(l10n_info()[["UTF-8"]], "Not a UTF-8 locale")
}

safe_load_help_db <- function() {
  help_db <- tryCatch(tools::Rd_db("lintr"), error = \(e) NULL)
  # e.g. in dev under pkgload::load_all()
  if (length(help_db) == 0L) {
    help_db <- tryCatch(tools::Rd_db(dir = testthat::test_path("..", "..")), error = \(e) NULL)
    testthat::skip_if_not(length(help_db) > 0L, message = "Package help corrupted or not installed")
  }
  help_db
}

pipes <- function(exclude = NULL) {
  all_pipes <- c(
    standard = "%>%",
    greedy = "%!>%",
    tee = "%T>%",
    assignment = "%<>%",
    extraction = "%$%",
    native = "|>"
  )
  all_pipes[!all_pipes %in% exclude]
}

test_that("returns the correct linting", {
  lint_msg <- rex::escape("Do not use absolute paths.")

  # strict mode
  linter <- absolute_path_linter(lax = FALSE)
  non_absolute_path_strings <- c(
    "..",
    "./blah",
    encodeString("blah\\file.txt")
  )
  for (path in non_absolute_path_strings) {
    expect_no_lint(single_quote(path), linter)
    expect_no_lint(double_quote(path), linter)
  }

  expect_no_lint("\"'/'\"", linter) # nested quotes

  absolute_path_strings <- c(
    "/",
    "/blah/file.txt",
    encodeString("d:\\"),
    "E:/blah/file.txt",
    encodeString("\\\\"),
    encodeString("\\\\server\\path"),
    "~",
    "~james.hester/blah/file.txt",
    encodeString("/a\nsdf"),
    "/as:df"
  )
  for (path in absolute_path_strings) {
    expect_lint(single_quote(path), lint_msg, linter)
    expect_lint(double_quote(path), lint_msg, linter)
  }

  # lax mode: no check for strings that are likely not paths (too short or with special characters)
  linter <- absolute_path_linter(lax = TRUE)
  unlikely_path_strings <- c(
    "/",
    encodeString("/a\nsdf/bar"),
    "/as:df/bar"
  )
  for (path in unlikely_path_strings) {
    expect_no_lint(single_quote(path), linter)
    expect_no_lint(double_quote(path), linter)
  }
})

test_that("raw strings are handled correctly", {
  expect_no_lint('R"(./blah)"', absolute_path_linter(lax = FALSE))
  expect_lint(
    "R'--[/blah/file.txt]--'",
    rex::rex("Do not use absolute paths."),
    absolute_path_linter(lax = FALSE)
  )
})

test_that("lints vectorize", {
  lint_msg <- rex::rex("Do not use absolute paths.")

  expect_lint(
    trim_some("{
      '/'
      '/blah/file.txt'
      'abcdefg'
      '~'
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L),
      list(lint_msg, line_number = 5L)
    ),
    absolute_path_linter(lax = FALSE)
  )
})

test_that("absolute_path_linter() handles unmarked UTF-8 text input", {
  skip_if_not_utf8_locale()

  tf <- withr::local_tempfile()
  writeLines('x <- "\u00e4"', tf, useBytes = TRUE)
  text_native <- readLines(tf)

  expect_no_error(
    expect_length(lint(text = text_native, linters = absolute_path_linter()), 0L)
  )
})

test_that("nonportable_path_linter skips allowed usages", {
  linter <- nonportable_path_linter(lax = FALSE)

  # various strings
  non_path_strings <- c(
    "foo",
    "https://cran.r-project.org/web/packages/lintr/",
    encodeString("hello\nthere!")
  )
  for (path in non_path_strings) {
    expect_no_lint(single_quote(path), linter)
    expect_no_lint(double_quote(path), linter)
  }

  expect_no_lint("\"'/foo'\"", linter) # nested quotes

  # system root
  root_path_strings <- c("/", "~", "c:", ".")
  for (path in root_path_strings) {
    expect_no_lint(single_quote(path), linter)
    expect_no_lint(double_quote(path), linter)
  }
})

test_that("nonportable_path_linter blocks disallowed usages", {
  linter <- nonportable_path_linter(lax = FALSE)
  lint_msg <- rex::escape("Use file.path() to construct portable file paths.")

  # paths with (back)slashes
  slash_path_strings <- c(
    "~/",
    "c:/",
    encodeString("D:\\"),
    "../",
    "/foo",
    "foo/",
    "foo/bar",
    encodeString("foo\\bar"),
    "/as:df",
    encodeString("/a\nsdf")
  )
  for (path in slash_path_strings) {
    expect_lint(single_quote(path), lint_msg, linter)
    expect_lint(double_quote(path), lint_msg, linter)
  }
})

test_that("nonportable_path_linter's lax argument works", {
  # lax mode: no check for strings that are likely not paths (too short or with special characters)
  linter <- nonportable_path_linter(lax = TRUE)

  unlikely_path_strings <- c(
    "/foo", encodeString("/a\nsdf/bar"), "/as:df/bar"
  )
  for (path in unlikely_path_strings) {
    expect_no_lint(single_quote(path), linter)
    expect_no_lint(double_quote(path), linter)
  }
})

test_that("lints vectorize", {
  lint_msg <- rex::escape("Use file.path() to construct portable file paths.")

  expect_lint(
    trim_some("{
      '~/'
      'C:/'
    }"),
    list(
      list(lint_msg, line_number = 2L),
      list(lint_msg, line_number = 3L)
    ),
    nonportable_path_linter(lax = FALSE)
  )
})

test_that("All linter help files have examples", {
  help_db <- safe_load_help_db()
  linter_db <- help_db[endsWith(names(help_db), "_linter.Rd")]
  rd_has_examples <- function(rd) any(vapply(rd, attr, "Rd_tag", FUN.VALUE = character(1L)) == "\\examples")
  linter_has_examples <- vapply(linter_db, rd_has_examples, logical(1L))
  for (ii in seq_along(linter_has_examples)) {
    expect_true(linter_has_examples[ii], label = paste("Linter", names(linter_db)[ii], "has examples"))
  }
})

#' Missing argument linter
#'
#' Check for missing arguments in function calls (e.g. `stats::median(1:10, )`).
#'
#' @param except a character vector of function names as exceptions.
#' @param allow_trailing always allow trailing empty arguments?
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = 'tibble(x = "a", )',
#'   linters = missing_argument_linter()
#' )
#'
#' # okay
#' lint(
#'   text = 'tibble(x = "a")',
#'   linters = missing_argument_linter()
#' )
#'
#' lint(
#'   text = 'tibble(x = "a", )',
#'   linters = missing_argument_linter(except = "tibble")
#' )
#'
#' lint(
#'   text = 'tibble(x = "a", )',
#'   linters = missing_argument_linter(allow_trailing = TRUE)
#' )
#'
#' @evalRd rd_tags("missing_argument_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
missing_argument_linter <- function(except = c("alist", "quote", "switch"), allow_trailing = FALSE) {
  conds <- c(
    "self::OP-COMMA[preceding-sibling::*[not(self::COMMENT)][1][self::OP-LEFT-PAREN or self::OP-COMMA]]",
    "self::EQ_SUB[following-sibling::*[not(self::COMMENT)][1][self::OP-RIGHT-PAREN or self::OP-COMMA]]"
  )
  if (!allow_trailing) {
    conds <- c(conds,
      "self::OP-RIGHT-PAREN[preceding-sibling::*[not(self::COMMENT)][1][self::OP-LEFT-PAREN or self::OP-COMMA]]"
    )
  }

  # require >3 children to exclude foo(), which is <expr><OP-LEFT-PAREN><OP-RIGHT-PAREN>,
  #   possibly with intervening comments too, #2741
  xpath <- glue("
  parent::expr[count(*) - count(COMMENT) > 3]
    /*[{xp_or(conds)}]
  ")

  Linter(linter_level = "file", function(source_expression) {
    xml_targets <- source_expression$xml_find_function_calls(NULL, keep_names = TRUE, include_s4_slots = TRUE)
    xml_targets <- xml_targets[!names(xml_targets) %in% except]

    missing_args <- xml_find_all_(xml_targets, xpath)

    named_idx <- xml_name_(missing_args) == "EQ_SUB"
    arg_id <- character(length(missing_args))
    arg_id[named_idx] <- sQuote(xml_find_chr_(missing_args[named_idx], "string(preceding-sibling::SYMBOL_SUB[1])"), "'")
    # TODO(#2452): use xml_find_int_() instead
    arg_id[!named_idx] <- xml_find_num_(missing_args[!named_idx], "count(preceding-sibling::OP-COMMA)") + 1.0

    xml_nodes_to_lints(
      missing_args,
      source_expression = source_expression,
      lint_message = sprintf("Missing argument %s in function call.", arg_id)
    )
  })
}

#' Infix spaces linter
#'
#' Check that infix operators are surrounded by spaces. Enforces the corresponding Tidyverse style guide rule;
#'   see <https://style.tidyverse.org/syntax.html#infix-operators>.
#'
#' @param exclude_operators Character vector of operators to exclude from consideration for linting.
#'   Default is to include the following "low-precedence" operators:
#'   `+`, `-`, `~`, `>`, `>=`, `<`, `<=`, `==`, `!=`, `&`, `&&`, `|`, `||`, `<-`, `:=`, `<<-`, `->`, `->>`,
#'   `=`, `/`, `*`, and any infix operator (exclude infixes by passing `"%%"`). Note that `"="` here includes
#'   three different operators, from the parser's point of view. To lint only some of these, pass the
#'   corresponding parse tags (i.e., some of `"EQ_ASSIGN"`, `"EQ_SUB"`, and `"EQ_FORMALS"`; see
#'   [utils::getParseData()]).
#' @param allow_multiple_spaces Logical, default `TRUE`. If `FALSE`, usage like `x  =  2` will also be linted;
#'   excluded by default because such usage can sometimes be used for better code alignment, as is allowed
#'   by the style guide.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "x<-1L",
#'   linters = infix_spaces_linter()
#' )
#'
#' lint(
#'   text = "1:4 %>%sum()",
#'   linters = infix_spaces_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "x <- 1L",
#'   linters = infix_spaces_linter()
#' )
#'
#' lint(
#'   text = "1:4 %>% sum()",
#'   linters = infix_spaces_linter()
#' )
#'
#' code_lines <- "
#' ab     <- 1L
#' abcdef <- 2L
#' "
#' writeLines(code_lines)
#' lint(
#'   text = code_lines,
#'   linters = infix_spaces_linter(allow_multiple_spaces = TRUE)
#' )
#'
#' lint(
#'   text = "a||b",
#'   linters = infix_spaces_linter(exclude_operators = "||")
#' )
#'
#' lint(
#'   text = "sum(1:10, na.rm=TRUE)",
#'   linters = infix_spaces_linter(exclude_operators = "EQ_SUB")
#' )
#'
#' @evalRd rd_tags("infix_spaces_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#infix-operators>
#' @export
infix_spaces_linter <- function(exclude_operators = NULL, allow_multiple_spaces = TRUE) {
  if (allow_multiple_spaces) {
    op <- "<"
    lint_message <- "Put spaces around all infix operators."
  } else {
    op <- "!="
    lint_message <- "Put exactly one space on each side of infix operators."
  }

  infix_tokens <- infix_metadata$xml_tag_exact[
    infix_metadata$low_precedence &
      !infix_metadata$string_value %in% exclude_operators &
      # parse_tag, not xml_tag, since the former is easier for the user to discover with getParseData()
      !infix_metadata$parse_tag %in% exclude_operators
  ]

  # NB: preceding-sibling::* and not preceding-sibling::expr because
  #   of the foo(a=1) case, where the tree is <SYMBOL_SUB><EQ_SUB><expr>
  # NB: parent::*[count(expr | SYMBOL_SUB)) > 1] for the unary case, e.g. x[-1]
  #  SYMBOL_SUB for case with missing argument like alist(a =)
  # NB: the last not() disables lints inside box::use() declarations
  global_xpath <- paste0("//", infix_tokens, collapse = "|")
  xpath <- glue("({global_xpath})[
    parent::*[count(expr | SYMBOL_SUB) > 1]
    and (
      (
        @line1 = preceding-sibling::*[1]/@line2
        and @start {op} preceding-sibling::*[1]/@end + 2
      ) or (
        @line1 = following-sibling::*[1]/@line1
        and following-sibling::*[1]/@start {op} @end + 2
      )
    )
    and not(
      self::OP-SLASH[
        ancestor::expr/preceding-sibling::OP-LEFT-PAREN/preceding-sibling::expr[
          ./SYMBOL_PACKAGE[text() = 'box'] and
          ./SYMBOL_FUNCTION_CALL[text() = 'use']
        ]
      ]
    )
  ]")

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    bad_expr <- xml_find_all_(xml, xpath)

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = lint_message,
      type = "style"
    )
  })
}

#' Object length linter
#'
#' Check that object names are not too long.
#' The length of an object name is defined as the length in characters, after removing extraneous parts:
#'  * generic prefixes for implementations of S3 generics, e.g. `as.data.frame.my_class` has length 8.
#'  * leading `.`, e.g. `.my_hidden_function` has length 18.
#'  * "%%" for infix operators, e.g. `%my_op%` has length 5.
#'  * trailing `<-` for assignment functions, e.g. `my_attr<-` has length 7.
#'
#' Note that this behavior relies in part on having packages in your Imports available;
#'   see the detailed note in [object_name_linter()] for more details.
#'
#' @param length maximum variable name length allowed.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "very_very_long_variable_name <- 1L",
#'   linters = object_length_linter(length = 10L)
#' )
#'
#' # okay
#' lint(
#'   text = "very_very_long_variable_name <- 1L",
#'   linters = object_length_linter(length = 30L)
#' )
#'
#' lint(
#'   text = "var <- 1L",
#'   linters = object_length_linter(length = 10L)
#' )
#'
#' @evalRd rd_tags("object_length_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
object_length_linter <- function(length = 30L) {
  lint_message <- paste("Variable and function names should not be longer than", length, "characters.")

  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    assignments <- xml_find_all_(xml, object_name_xpath)

    # Retrieve assigned name
    nms <- strip_names(
      xml_text(assignments)
    )

    # run namespace_imports at run-time, not "compile" time to allow package structure to change
    pkg <- find_package(source_expression$filename)
    ns_imports <- namespace_imports(pkg)
    generics <- strip_names(c(
      declared_s3_generics(xml),
      imported_s3_generics(ns_imports)$fun,
      exported_s3_generics(pkg)$fun,
      .base_s3_generics
    ))
    generics <- unique(generics[nzchar(generics)])

    # Remove generic function names from generic implementations
    # This only lints S3 implementations if the class names are too long, still lints generics if they are too long.
    nms_stripped <- re_substitutes(nms, rex(start, or(generics), "."), "")

    too_long <- nchar(nms_stripped) > length

    xml_nodes_to_lints(
      assignments[too_long],
      source_expression = source_expression,
      lint_message = lint_message,
      type = "style"
    )
  })
}

#' Require `c()` to be applied before relatively expensive vectorized functions
#'
#' `as.Date(c(a, b))` is logically equivalent to `c(as.Date(a), as.Date(b))`.
#'   The same equivalence holds for several other vectorized functions like
#'   [as.POSIXct()] and math functions like [sin()]. The former is to be
#'   preferred so that the most expensive part of the operation ([as.Date()])
#'   is applied only once.
#'
#' Note that [strptime()] has one idiosyncrasy to be aware of, namely that
#'   auto-detected `format=` is set by the first matching input, which means
#'   that a case like `c(as.POSIXct("2024-01-01"), as.POSIXct("2024-01-01 01:02:03"))`
#'   gives different results to `as.POSIXct(c("2024-01-01", "2024-01-01 01:02:03"))`.
#'   This false positive is rare; a workaround where possible is to use
#'   consistent formatting, i.e., `"2024-01-01 00:00:00"` in the example.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "c(log10(x), log10(y), log10(z))",
#'   linters = inner_combine_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "log10(c(x, y, z))",
#'   linters = inner_combine_linter()
#' )
#'
#' lint(
#'   text = "c(log(x, base = 10), log10(x, base = 2))",
#'   linters = inner_combine_linter()
#' )
#'
#' @evalRd rd_tags("inner_combine_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
inner_combine_linter <- function() {
  # these don't take any other arguments (except maybe by non-default
  #   methods), so don't need to check equality of other arguments
  no_arg_vectorized_funs <- c(
    "sin", "cos", "tan", "sinpi", "cospi", "tanpi", "asin", "acos", "atan",
    "log2", "log10", "log1p", "exp", "expm1",
    "sqrt", "abs"
  )

  # TODO(#2468): Try and make this XPath less brittle/more extensible.

  # See ?as.Date, ?as.POSIXct. tryFormats is not explicitly in any default
  #   POSIXct method, but it is in as.Date.character and as.POSIXlt.character --
  #   the latter is what actually gets invoked when running as.POSIXct
  #   on a character. So it is indeed an argument by pass-through.
  date_args <- c("format", "origin", "tz", "tryFormats")
  date_funs <- c("as.Date", "as.POSIXct", "as.POSIXlt")

  # See ?log. Only these two take a 'base' argument.
  log_funs <- c("log", "logb")
  log_args <- "base"

  # See ?lubridate::ymd and ?lubridate::ymd_hms
  lubridate_args <- c("quiet", "tz", "locale", "truncated")
  lubridate_funs <- c(
    "ymd", "ydm", "mdy", "myd", "dmy", "dym",
    "yq", "ym", "my",
    "ymd_hms", "ymd_hm", "ymd_h", "dmy_hms", "dmy_hm", "dmy_h",
    "mdy_hms", "mdy_hm", "mdy_h", "ydm_hms", "ydm_hm", "ydm_h"
  )

  date_args_cond <- build_arg_condition(date_funs, date_args)
  log_args_cond <- build_arg_condition(log_funs, log_args)
  lubridate_args_cond <- build_arg_condition(lubridate_funs, lubridate_args)

  c_expr_cond <- xp_and(
    sprintf(
      "expr[1][SYMBOL_FUNCTION_CALL[%s]]",
      xp_text_in_table(c(no_arg_vectorized_funs, date_funs, log_funs, lubridate_funs))
    ),
    "not(following-sibling::expr[not(expr[1][SYMBOL_FUNCTION_CALL])])",
    "not(expr[1]/SYMBOL_FUNCTION_CALL != following-sibling::expr/expr[1]/SYMBOL_FUNCTION_CALL)",
    date_args_cond,
    log_args_cond,
    lubridate_args_cond
  )
  xpath <- glue("
  self::*[count(following-sibling::expr) > 1]
    /following-sibling::expr[1][ {c_expr_cond} ]
    /parent::expr
  ")

  Linter(linter_level = "expression", function(source_expression) {
    xml_calls <- source_expression$xml_find_function_calls("c")
    bad_expr <- xml_find_all_(xml_calls, xpath)

    matched_call <- xp_call_name(bad_expr, depth = 2L)
    lint_message <- paste(
      "Combine inputs to vectorized functions first to take full advantage of vectorization, e.g.,",
      sprintf(
        "%1$s(c(x, y)) only runs the more expensive %1$s() once as compared to c(%1$s(x), %1$s(y)).",
        matched_call
      )
    )
    xml_nodes_to_lints(bad_expr, source_expression = source_expression, lint_message, type = "warning")
  })
}

#' Make the XPath condition ensuring an argument matches across calls
#'
#' @param arg Character scalar naming an argument
#' @noRd
arg_match_condition <- function(arg) {
  this_symbol <- sprintf("SYMBOL_SUB[text() = '%s']", arg)
  following_symbol <- sprintf("following-sibling::expr/%s", this_symbol)
  next_expr <- "following-sibling::expr[1]"
  xp_or(
    sprintf("not(%s) and not(%s)", this_symbol, following_symbol),
    xp_and(
      this_symbol,
      following_symbol,
      sprintf(
        "not(%1$s/%3$s != %2$s/%3$s)",
        this_symbol, following_symbol, next_expr
      )
    )
  )
}

build_arg_condition <- function(calls, arguments) {
  xp_or(
    sprintf("not(expr[1][SYMBOL_FUNCTION_CALL[%s]])", xp_text_in_table(calls)),
    "not(EQ_SUB) and not(following-sibling::expr/EQ_SUB)",
    xp_and(vapply(arguments, arg_match_condition, character(1L)))
  )
}

#' Commas linter
#'
#' Check that all commas are followed by spaces, but do not have spaces before them.
#'
#' @param allow_trailing If `TRUE`, the linter allows a comma to be followed
#' directly by a closing bracket without a space.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "switch(op , x = foo, y = bar)",
#'   linters = commas_linter()
#' )
#'
#' lint(
#'   text = "mean(x,trim = 0.2,na.rm = TRUE)",
#'   linters = commas_linter()
#' )
#'
#' lint(
#'   text = "x[ ,, drop=TRUE]",
#'   linters = commas_linter()
#' )
#'
#' lint(
#'   text = "x[1,]",
#'   linters = commas_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "switch(op, x = foo, y = bar)",
#'   linters = commas_linter()
#' )
#'
#' lint(
#'   text = "switch(op, x = , y = bar)",
#'   linters = commas_linter()
#' )
#'
#' lint(
#'   text = "mean(x, trim = 0.2, na.rm = TRUE)",
#'   linters = commas_linter()
#' )
#'
#' lint(
#'   text = "a[1, , 2, , 3]",
#'   linters = commas_linter()
#' )
#'
#' lint(
#'   text = "x[1,]",
#'   linters = commas_linter(allow_trailing = TRUE)
#' )
#'
#' @evalRd rd_tags("commas_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - <https://style.tidyverse.org/syntax.html#commas>
#' @export
commas_linter <- function(allow_trailing = FALSE) {
  # conditions are in carefully-chosen order for performance --
  #   an expression like c(a,b,c,....) with many elements can have
  #   a huge number of preceding-siblings and the performance of
  #   preceding-sibling::*[1][not(self::OP-COMMA)] is terrible.
  #   This approach exits early on most nodes ('and' condition)
  #   to avoid this. See #1340.
  xpath_before <- "
  //OP-COMMA[
    @col1 != preceding-sibling::*[1]/@col2 + 1 and
    @line1 = preceding-sibling::*[1]/@line1 and
    not(preceding-sibling::*[1][self::OP-COMMA or self::EQ_SUB])
  ]"
  xpath_after <- paste0(
    "//OP-COMMA[@line1 = following-sibling::*[1]/@line1 and @col1 = following-sibling::*[1]/@col1 - 1 ",
    if (allow_trailing) "and not(following-sibling::*[1][self::OP-RIGHT-BRACKET or self::RBB or self::OP-RIGHT-PAREN])",
    "]"
  )

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    before_lints <- xml_nodes_to_lints(
      xml_find_all_(xml, xpath_before),
      source_expression = source_expression,
      lint_message = "Remove spaces before a comma.",
      range_start_xpath = "number(./preceding-sibling::*[1]/@col2 + 1)", # start after preceding expression
      range_end_xpath = "number(./@col1 - 1)" # end before comma
    )

    after_lints <- xml_nodes_to_lints(
      xml_find_all_(xml, xpath_after),
      source_expression = source_expression,
      lint_message = "Put a space after a comma.",
      range_start_xpath = "number(./@col2 + 1)", # start and end after comma
      range_end_xpath = "number(./@col2 + 1)"
    )

    c(before_lints, after_lints)
  })
}

#' Non-portable path linter
#'
#' Check that [file.path()] is used to construct safe and portable paths.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "'abcdefg/hijklmnop/qrst/uv/wxyz'",
#'   linters = nonportable_path_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "file.path('abcdefg', 'hijklmnop', 'qrst', 'uv', 'wxyz')",
#'   linters = nonportable_path_linter()
#' )
#'
#' @inheritParams absolute_path_linter
#' @evalRd rd_tags("nonportable_path_linter")
#' @seealso
#' - [linters] for a complete list of linters available in lintr.
#' - [absolute_path_linter()]
#' @export
nonportable_path_linter <- function(lax = TRUE) {
  path_linter_factory(
    path_function = function(path) {
      is_path(path) && is_valid_long_path(path, lax) && path != "/" &&
        re_matches(path, rex(one_of("/", "\\")))
    },
    message = "Use file.path() to construct portable file paths."
  )
}

#' Block usage of print() for logging
#'
#' The default print method for character vectors is appropriate for interactively inspecting objects,
#'   not for logging messages. Thus checked-in usage like `print(paste('Data has', nrow(DF), 'rows.'))`
#'   is better served by using [cat()], e.g. `cat(sprintf('Data has %d rows.\n', nrow(DF)))`. Note that
#'   using `cat()` entails supplying your own line returns; [glue::glue()] might be preferable
#'   to [sprintf()] for constructing templated strings. Alternatively, [writeLines()] can be used to
#'   print character vectors with their own line returns. Lastly, note that [message()] differs slightly
#'   from `cat()` in that it prints to `stderr` by default, not `stdout`, but is still a good option
#'   to consider for logging purposes.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "print('a')",
#'   linters = print_linter()
#' )
#'
#' lint(
#'   text = "print(paste(x, 'y'))",
#'   linters = print_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "print(x)",
#'   linters = print_linter()
#' )
#'
#' @evalRd rd_tags("print_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
print_linter <- make_linter_from_function_xpath(
  function_names = "print",
  xpath = "
    parent::expr[expr[2][
      STR_CONST
      or expr/SYMBOL_FUNCTION_CALL[
        text() = 'paste' or text() = 'paste0' or text() = 'sprintf'
      ]
    ]]
  ",
  lint_message =
    "Use cat() instead of print() logging messages. Use message() in cases calling for a signalled condition."
)

#' Redirect `is.numeric(x) || is.integer(x)` to just use `is.numeric(x)`
#'
#' [is.numeric()] returns `TRUE` when `typeof(x)` is `double` or `integer` --
#'   testing `is.numeric(x) || is.integer(x)` is thus redundant.
#'
#' NB: This linter plays well with [class_equals_linter()], which can help
#'   avoid further `is.numeric()` equivalents like
#'   `any(class(x) == c("numeric", "integer"))`.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "is.numeric(y) || is.integer(y)",
#'   linters = is_numeric_linter()
#' )
#'
#' lint(
#'   text = 'class(z) %in% c("numeric", "integer")',
#'   linters = is_numeric_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "is.numeric(y) || is.factor(y)",
#'   linters = is_numeric_linter()
#' )
#'
#' lint(
#'   text = 'class(z) %in% c("numeric", "integer", "factor")',
#'   linters = is_numeric_linter()
#' )
#'
#' @evalRd rd_tags("is_numeric_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
is_numeric_linter <- function() {
  # TODO(#2469): This should also cover is.double(x) || is.integer(x).
  # TODO(#1636): is.numeric(x) || is.integer(x) || is.factor(x) is also redundant.
  # TODO(#2470): Consider usages with class(), typeof(), or inherits().
  is_numeric_expr <- "expr[1][SYMBOL_FUNCTION_CALL[text() = 'is.numeric']]"
  is_integer_expr <- "expr[1][SYMBOL_FUNCTION_CALL[text() = 'is.integer']]"

  # testing things like is.numeric(x) || is.integer(x)
  or_xpath <- glue("
  self::*[
    expr/{is_numeric_expr}
    and expr/{is_integer_expr}
    and
      expr/{is_numeric_expr}/following-sibling::expr[1]
      = expr/{is_integer_expr}/following-sibling::expr[1]
  ]")

  # testing class(x) %in% c("numeric", "integer")
  class_xpath <- "
  //SPECIAL[
    text() = '%in%'
    and following-sibling::expr[
      expr/SYMBOL_FUNCTION_CALL[text() = 'c']
      and count(expr/STR_CONST) = 2
      and count(expr) = 3
    ]
    and preceding-sibling::expr/expr/SYMBOL_FUNCTION_CALL[text() = 'class']
  ]
    /parent::expr
  "

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    or_expr <- xml |>
      xml_find_all_("//OR2/parent::*") |>
      strip_comments_from_subtree() |>
      xml_find_all_(or_xpath)
    or_lints <- xml_nodes_to_lints(
      or_expr,
      source_expression = source_expression,
      lint_message = paste(
        "Use `is.numeric(x)` instead of the equivalent `is.numeric(x) || is.integer(x)`.",
        "Use is.double(x) to test for objects stored as 64-bit floating point."
      ),
      type = "warning"
    )

    class_expr <- xml_find_all_(xml, class_xpath)
    if (length(class_expr) > 0L) {
      class_strings <- c(
        get_r_string(class_expr, "expr[2]/expr[2]/STR_CONST"),
        get_r_string(class_expr, "expr[2]/expr[3]/STR_CONST")
      )
      is_lintable <- "integer" %in% class_strings && "numeric" %in% class_strings
      class_expr <- class_expr[is_lintable]
    }
    class_lints <- xml_nodes_to_lints(
      class_expr,
      source_expression = source_expression,
      lint_message = paste(
        'Use is.numeric(x) instead of class(x) %in% c("integer", "numeric").',
        "Use is.double(x) to test for objects stored as 64-bit floating point."
      ),
      type = "warning"
    )

    c(or_lints, class_lints)
  })
}

#' Backport linter
#'
#' Check for usage of unavailable functions. Not reliable for testing r-devel dependencies.
#'
#' @param r_version Minimum R version to test for compatibility. Defaults to
#'  the R version currently in use. The version can be specified as a version
#'  number, or as a version alias (such as `"devel"`, `"oldrel"`, `"oldrel-1"`).
#' @param except Character vector of functions to be excluded from linting.
#'  Use this to list explicitly defined backports, e.g. those imported from the `{backports}` package or manually
#'  defined in your package.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "trimws(x)",
#'   linters = backport_linter("3.0.0")
#' )
#'
#' lint(
#'   text = "str2lang(x)",
#'   linters = backport_linter("3.2.0")
#' )
#'
#' lint(
#'   text = "deparse1(expr)",
#'   linters = backport_linter("3.6.0")
#' )
#'
#' # okay
#' lint(
#'   text = "trimws(x)",
#'   linters = backport_linter("3.6.0")
#' )
#'
#' lint(
#'   text = "str2lang(x)",
#'   linters = backport_linter("3.2.0", except = "str2lang")
#' )
#'
#' # Version aliases instead of numbers can also be passed to `r_version`
#' lint(
#'   text = "deparse1(expr)",
#'   linters = backport_linter("release")
#' )
#'
#' @evalRd rd_tags("backport_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
backport_linter <- function(r_version = getRversion(), except = character()) {
  r_version <- normalize_r_version(r_version)

  if (all(r_version >= R_system_version(names(backports)))) {
    return(Linter(\(source_expression) list(), linter_level = "file"))
  }

  backport_blacklist <- backports[r_version < R_system_version(names(backports))]
  backport_blacklist <- lapply(backport_blacklist, setdiff, except)
  backport_index <- rep(names(backport_blacklist), times = lengths(backport_blacklist))
  names(backport_index) <- unlist(backport_blacklist)

  Linter(linter_level = "expression", function(source_expression) {
    xml <- source_expression$xml_parsed_content

    used_symbols <- xml_find_all_(xml, "//SYMBOL | //SPECIAL")
    used_symbols <- used_symbols[xml_text(used_symbols) %in% names(backport_index)]

    used_calls <- source_expression$xml_find_function_calls(names(backport_index))
    all_names_nodes <- combine_nodesets(
      xml_find_first_(used_calls, "SYMBOL_FUNCTION_CALL"),
      used_symbols
    )
    all_names <- xml_text(all_names_nodes)

    bad_versions <- unname(backport_index[all_names])

    lint_message <- sprintf(
      "%s (R %s) is not always available for requested dependency (R >= %s).",
      all_names,
      bad_versions,
      r_version
    )
    xml_nodes_to_lints(
      all_names_nodes,
      source_expression = source_expression,
      lint_message = lint_message,
      type = "warning"
    )
  })
}

normalize_r_version <- function(r_version) {
  rx_release_spec <- rex(
    start,
    "release" %or%
      list("oldrel", maybe("-", digits)) %or%
      "devel",
    end
  )
  if (is.character(r_version) && re_matches(r_version, rx_release_spec)) {
    # Support devel, release, oldrel, oldrel-1, ...
    if (r_version == "oldrel") {
      r_version <- "oldrel-1"
    }

    all_versions <- names(backports)
    minor_versions <- unique(re_substitutes(all_versions, rex(".", digits, end), ""))
    version_names <- c("devel", "release", paste0("oldrel-", seq_len(length(minor_versions) - 2L)))
    if (!r_version %in% version_names) {
      # This can only trip if e.g. oldrel-99 is requested
      cli_abort(c(
        "{.arg r_version} is not valid:",
        i = "It must be a version number or one of {.str {version_names}}.",
        x = "You entered {.str {r_version}} instead."
      ))
    }
    requested_version <- minor_versions[match(r_version, table = version_names)]
    available_patches <- all_versions[startsWith(all_versions, requested_version)]
    selected_patch <- which.max(as.integer(
      substr(available_patches, start = nchar(requested_version) + 2L, stop = nchar(available_patches))
    ))

    r_version <- R_system_version(available_patches[selected_patch])
  } else if (is.character(r_version)) {
    r_version <- R_system_version(r_version, strict = TRUE)
  } else if (!inherits(r_version, "R_system_version")) {
    cli_abort("{.arg r_version} must be an R version number, returned by {.fun R_system_version}, or a string.")
  }
  if (r_version < "3.0.0") {
    cli_warn(c(
      x = "Depending on an R version older than {.val 3.0.0} is not recommended.",
      i = "Resetting {.arg r_version} to {.val 3.0.0}."
    ))
    r_version <- R_system_version("3.0.0")
  }
  r_version
}

backports <- list(
  `4.6.0` = c(
    # base
    "%notin%", "messageCondition", "mtfrm.Date",
    # grDevices
    "glyphFontVariation",
    # stats
    "free1way", "power.free1way.test", "ppplot", "rfree1way",
    # tcltk
    "tkgrid.child", "tkpack.child", "tkplace.child",
    # tools
    "checkSHA256sums", "verifySHA256signature",
    # utils
    "bitstring"
  ),
  `4.5.1` = "analyze_license", # tools
  `4.5.0` = c(
    # base
    ".set_ops_need_as_vector", "grepv", "summary.difftime", "zstdfile",
    # methods
    "matrixOps",
    # parallel
    "registerClusterType",
    # stats
    "qr.influence",
    # tools
    "CRAN_aliases_db", "CRAN_archive_db", "CRAN_authors_db",
    "CRAN_current_db", "CRAN_rdxrefs_db", "base_aliases_db",
    "base_rdxrefs_db", "check_package_dois", "check_package_urls",
    "parse_URI_reference", "sha256sum",
    # utils
    ".checkHT"
  ),
  `4.4.3` = character(), # need character() entries for oldrel specifications
  `4.4.1` = "R", # tools
  `4.4.0` = c(
    # base
    "%||%", ".formula2varlist", ".internalGenerics", ".rangeNum", "Exec",
    "Tailcall", "array2DF", "chooseOpsMethod", "chooseOpsMethod.default",
    "declare", "mtfrm.POSIXct", "nameOfClass", "nameOfClass.default",
    "range.POSIXct", "sort_by", "sort_by.data.frame", "sort_by.default",
    "use",
    # tools
    "pkg2HTML", "standard_package_names",
    # utils
    ".AtNames"
  ),
  `4.3.3` = character(),
  `4.3.1` = c(
    # parallel
    "closeNode", "recvData", "recvOneData", "sendData",
    # utils
    "findMatches"
  ),
  `4.3.0` = c(
    # base
    "$<-.POSIXlt", ".check_tzones", "R_compiled_by", "balancePOSIXlt",
    "is.finite.POSIXlt", "is.infinite.POSIXlt", "is.nan.POSIXlt",
    "unCfillPOSIXlt",
    # grDevices
    "embedGlyphs", "glyphAnchor", "glyphFont", "glyphFontList", "glyphHeight",
    "glyphHeightBottom", "glyphInfo", "glyphJust", "glyphWidth",
    "glyphWidthLeft",
    # grid
    "glyphGrob", "grid.glyph", "isClosed",
    # stats
    "toeplitz2",
    # tools
    "as.Rconcordance", "followConcordance", "matchConcordance"
  ),
  `4.2.3` = character(),
  `4.2.1` = c(
    # stats
    "psmirnov", "qsmirnov", "rsmirnov",
    # utils
    "findCRANmirror"
  ),
  `4.2.0` = c(
    # base
    ".LC.categories", ".pretty", "Sys.setLanguage", "as.vector.POSIXlt",
    "as.vector.data.frame", "mtfrm", "mtfrm.default",
    # grDevices
    ".clipPath", ".defineGroup", ".devUp", ".mask", ".opIndex", ".ruleIndex",
    ".useGroup",
    # grid
    "as.mask", "as.path", "defineGrob", "defnRotate", "defnScale",
    "defnTranslate", "emptyGTreeCoords", "emptyGrobCoords", "fillGrob",
    "fillStrokeGrob", "grid.define", "grid.fill", "grid.fillStroke",
    "grid.group", "grid.stroke", "grid.use", "gridCoords", "gridGTreeCoords",
    "gridGrobCoords", "groupFlip", "groupGrob", "groupRotate", "groupScale",
    "groupShear", "groupTranslate", "strokeGrob", "useGrob", "useRotate",
    "useScale", "useTranslate", "viewportRotate", "viewportScale",
    "viewportTransform", "viewportTranslate",
    # utils
    "clrhash", "gethash", "hashtab", "is.hashtab", "maphash", "numhash",
    "remhash", "sethash", "typhash"
  ),
  `4.1.3` = character(),
  `4.1.0` = c(
    # base
    "...names", "[<-.difftime", "all.equal.function", "c.factor", "gregexec",
    "isa", "numToBits", "numToInts", "rep.difftime", "xtfrm.data.frame",
    # grDevices
    ".linearGradientPattern", ".radialGradientPattern", ".setClipPath",
    ".setMask", ".setPattern", ".tilingPattern",
    # grid
    "editViewport", "linearGradient", "pattern", "radialGradient",
    # tools
    "checkRdContents",
    # utils
    "RtangleFinish", "RtangleRuncode", "charClass"
  ),
  `4.0.5` = character(),
  `4.0.1` = c(
    # base
    "activeBindingFunction",
    # grDevices
    "cairoSymbolFont"
  ),
  `4.0.0` = c(
    # base
    ".S3method", ".class2", "anyNA.data.frame", "as.list.difftime",
    "deparse1", "globalCallingHandlers", "infoRDS", "list2DF", "marginSums",
    "plot", "proportions", "sequence.default", "serverSocket", "socketAccept",
    "socketTimeout", "tryInvokeRestart",
    # grDevices
    "palette.colors", "palette.pals",
    # grid
    "unitType",
    # stats
    "Pair",
    # tools
    "R_user_dir"
  ),
  `3.6.3` = character(),
  `3.6.1` = c(
    # base
    "str2expression", "str2lang",
    # tools
    "update_PACKAGES"
  ),
  `3.6.0` = c(
    # base
    ".doSortWrap", "[[<-.POSIXlt", "allowInterrupts", "asplit",
    "conflictRules", "errorCondition", "length<-.difftime", "mem.maxNSize",
    "mem.maxVSize", "nullfile", "packageNotFoundError", "suspendInterrupts",
    "warningCondition",
    # grDevices
    "hcl.colors", "hcl.pals",
    # grid
    "delayGrob", "deviceDim", "deviceLoc", "emptyCoords", "grobCoords",
    "grobPoints", "isEmptyCoords", "unit.psum",
    # stats
    "DF2formula",
    # tools
    "vignetteInfo",
    # utils
    "osVersion"
  ),
  `3.5.3` = character(),
  `3.5.0` = c(
    # base
    "...elt", "...length", "..deparseOpts", ".Date", ".S3_methods_table",
    ".col", ".doWrap", ".fixupGFortranStderr", ".fixupGFortranStdout",
    ".makeSortEnum", ".row", ".rowNamesDF<-", "[.DLLInfoList", "[[.POSIXlt",
    "as.list.POSIXlt", "isFALSE", "isWrappable", "length<-.Date",
    "length<-.POSIXct", "length<-.POSIXlt", "print.summary.warnings",
    "summary.warnings",
    # grid
    "recordGrob",
    # parallel
    "getDefaultCluster",
    # stats
    ".vcov.aliased",
    # utils
    "asDateBuilt", "askYesNo", "packageDate", "warnErrList"
  ),
  `3.4.4` = character(),
  `3.4.1` = "CRAN_check_issues", # tools
  `3.4.0` = c(
    # base
    ".tryResumeInterrupt", ".valid.factor", "La_library", "diff.difftime",
    "duplicated.warnings", "print.eigen", "withAutoprint",
    # methods
    ".debugMethod", ".isMethodDebugged", ".undebugMethod", "isRematched",
    # tools
    "CRAN_check_details", "CRAN_check_results", "CRAN_memtest_notes",
    "CRAN_package_db", "check_packages_in_dir_details",
    "package_native_routine_registration_skeleton",
    # utils
    ".RtangleCodeLabel", ".romans", "debugcall", "hasName", "isS3stdGeneric",
    "strcapture", "undebugcall"
  ),
  `3.3.3` = character(),
  `3.3.1` = c(
    # tools
    "Rcmd", "makevars_site", "makevars_user", "summarize_CRAN_check_status"
  ),
  `3.3.0` = c(
    # base
    ".format.zeros", ".rmpkg", ".traceback", "[.table", "c.difftime",
    "chkDots", "endsWith", "grouping", "startsWith", "strrep", "validEnc",
    "validUTF8",
    # methods
    "externalRefMethod",
    # stats
    "sigma",
    # tcltk
    "tkimage.delete", "tkimage.height", "tkimage.inuse", "tkimage.type",
    "tkimage.types", "tkimage.width", "ttkscale", "ttkspinbox",
    # tools
    "langElts", "nonS3methods",
    # utils
    "isS3method"
  ),
  `3.2.5` = character(),
  `3.2.0` = c(
    # base
    ".getNamespaceInfo", ".maskedMsg", "[.Dlist", "[<-.numeric_version",
    "all.equal.envRefClass", "all.equal.environment", "as.data.frame.noquote",
    "c.warnings", "curlGetHeaders", "debuggingState", "dir.exists", "dynGet",
    "extSoftVersion", "file.mode", "file.mtime", "file.size", "forceAndCall",
    "get0", "is.na<-.numeric_version", "isNamespaceLoaded", "lengths",
    "libcurlVersion", "print.Dlist", "returnValue", "trimws",
    "unique.warnings",
    # grDevices
    "grSoftVersion",
    # methods
    ".S4methods",
    # stats
    ".nknots.smspl",
    # tcltk
    "tclVersion",
    # tools
    "check_packages_in_dir_changes", "loadPkgRdMacros", "loadRdMacros",
    "toTitleCase",
    # utils
    ".S3methods", "hsearch_db", "hsearch_db_concepts", "hsearch_db_keywords"
  ),
  `3.1.3` = "pcre_config", # base
  `3.1.2` = "icuGetCollate", # base
  `3.1.1` = "promptImport", # utils
  `3.1.0` = c(
    # base
    "$.data.frame", "..getNamespace", "OlsonNames", "[.warnings", "agrepl",
    "all.equal.POSIXt", "anyNA", "anyNA.POSIXlt", "anyNA.numeric_version",
    "cospi", "dontCheck", "sinpi", "tanpi",
    # grid
    "current.parent", "current.rotation", "explode", "forceGrob", "grid.grep",
    "legendGrob", "resolveHJust", "resolveRasterSize", "resolveVJust",
    # stats
    ".lm.fit", ".preformat.ts", "confint.lm", "dummy.coef.lm",
    # tools
    "buildVignette", "find_gs_cmd",
    # utils
    "changedFiles", "fileSnapshot", "suppressForeignCheck"
  ),
  `3.0.3` = "La_version", # base
  `3.0.2` = c(
    # parallel
    "mcMap",
    # tools
    "assertCondition", "assertError", "assertWarning", "getVignetteInfo"
  ),
  `3.0.1` = "depth", # grid
  `3.0.0` = c(
    # base
    ".External2", ".bincode", ".detach", ".getNamespace", ".kappa_tri",
    ".mapply", "@<-", "bitwAnd", "bitwNot", "bitwOr", "bitwShiftL",
    "bitwShiftR", "bitwXor", "clearPushBack", "numeric", "provideDimnames",
    "rep_len", "summary.proc_time",
    # grDevices
    "cairo_pdf", "cairo_ps", "svg",
    # grid
    "grid.delay", "grid.force", "grid.reorder", "grid.revert", "makeContent",
    "makeContext", "reorderGrob",
    # methods
    "checkAtAssignment", "evalOnLoad", "evalqOnLoad", "getLoadActions",
    "hasLoadAction", "insertClassMethods", "setLoadAction", "setLoadActions",
    # tools
    ".print.via.format", "checkPoFile", "checkPoFiles",
    "check_packages_in_dir", "getBibstyle", "make_translations_pkg",
    "summarize_check_packages_in_dir_depends",
    "summarize_check_packages_in_dir_results",
    "summarize_check_packages_in_dir_timings", "update_pkg_po",
    "vignetteEngine",
    # utils
    "aspell_package_C_files", "aspell_package_R_files", "cite", "citeNatbib",
    "getParseData", "getParseText", "globalVariables", "packageName",
    "process.events"
  )
)

#' Namespace linter
#'
#' Check for missing packages and symbols in namespace calls.
#' Note that using `check_exports=TRUE` or `check_nonexports=TRUE` will load packages used in user code so it could
#' potentially change the global state.
#'
#' @param check_exports Check if `symbol` is exported from `namespace` in `namespace::symbol` calls.
#' @param check_nonexports Check if `symbol` exists in `namespace` in `namespace:::symbol` calls.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "xyzxyz::sd(c(1, 2, 3))",
#'   linters = namespace_linter()
#' )
#'
#' lint(
#'   text = "stats::ssd(c(1, 2, 3))",
#'   linters = namespace_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "stats::sd(c(1, 2, 3))",
#'   linters = namespace_linter()
#' )
#'
#' lint(
#'   text = "stats::ssd(c(1, 2, 3))",
#'   linters = namespace_linter(check_exports = FALSE)
#' )
#'
#' lint(
#'   text = "stats:::ssd(c(1, 2, 3))",
#'   linters = namespace_linter(check_nonexports = FALSE)
#' )
#'
#' @evalRd rd_tags("namespace_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
namespace_linter <- function(check_exports = TRUE, check_nonexports = TRUE) {
  Linter(linter_level = "file", function(source_expression) {
    xml <- source_expression$full_xml_parsed_content

    ns_nodes <- xml_find_all_(xml, "//NS_GET | //NS_GET_INT")

    if (length(ns_nodes) == 0L) {
      return(list())
    }

    ## Case 1: pkg is uninstalled in pkg::foo

    package_nodes <- xml_find_all_(ns_nodes, "preceding-sibling::*[1]")
    packages <- get_r_string(package_nodes)

    lints <- list()

    # run here, not in the factory, to allow for run- vs. "compile"-time differences in available packages
    installed_packages <- .packages(all.available = TRUE)
    installed <- packages %in% installed_packages

    if (!all(installed)) {
      lints <- c(lints, xml_nodes_to_lints(
        package_nodes[!installed],
        source_expression = source_expression,
        lint_message = sprintf("Package '%s' is not installed.", packages[!installed]),
        type = "warning"
      ))

      ns_nodes <- ns_nodes[installed]
      packages <- packages[installed]
      package_nodes <- package_nodes[installed]
    }

    if (length(packages) == 0L) {
      return(lints)
    }

    ## Case 2/3/4/5: problems with foo in pkg::foo / pkg:::foo

    # run here, not in the factory, to allow for run- vs. "compile"-time differences in package structure
    namespaces <- lapply(packages, \(package) tryCatch(getNamespace(package), packageNotFoundError = identity))
    failed_namespace <- vapply(namespaces, inherits, "packageNotFoundError", FUN.VALUE = logical(1L))

    # nocov start
    if (any(failed_namespace)) {
      cli_abort_internal("Failed to retrieve namespaces for one or more of the packages used with `::` or `:::`. ")
    }
    # nocov end

    ns_get <- xml_text(ns_nodes) == "::"
    symbol_nodes <- xml_find_all_(ns_nodes, "following-sibling::*[1]")
    symbols <- get_r_string(symbol_nodes)
    symbols <- sub("^`(.*)`$", "\\1", symbols)

    if (check_nonexports) {
      lints <- c(lints, build_ns_get_int_lints(
        packages[!ns_get],
        symbols[!ns_get],
        symbol_nodes[!ns_get],
        namespaces[!ns_get],
        source_expression
      ))
    }

    if (check_exports) {
      lints <- c(lints, build_ns_get_lints(
        packages[ns_get],
        symbols[ns_get],
        symbol_nodes[ns_get],
        namespaces[ns_get],
        source_expression
      ))
    }

    pkg_path <- find_package(source_expression$filename)
    ns_imports <- namespace_imports(pkg_path)
    if (nrow(ns_imports) == 0L) {
      return(lints)
    }

    is_imported <- vapply(
      seq_along(symbols),
      \(ii) any(ns_imports$pkg == packages[ii] & ns_imports$fun == symbols[ii]),
      logical(1L)
    )
    if (any(is_imported)) {
      lints <- c(lints, xml_nodes_to_lints(
        symbol_nodes[is_imported],
        source_expression = source_expression,
        lint_message = sprintf(
          "Don't use `%s` to access %s, which is already imported from %s.",
          ifelse(ns_get[is_imported], "::", ":::"), symbols[is_imported], packages[is_imported]
        ),
        type = "warning"
      ))
    }

    lints
  })
}

namespace_symbols <- function(ns, exported = TRUE) {
  if (exported) {
    c(getNamespaceExports(ns), names(.getNamespaceInfo(ns, "lazydata")))
  } else {
    ls(ns, all.names = TRUE)
  }
}
is_in_pkg <- function(symbols, namespaces, exported = TRUE) {
  vapply(
    seq_along(symbols),
    \(ii) symbols[[ii]] %in% namespace_symbols(namespaces[[ii]], exported = exported),
    logical(1L)
  )
}

build_ns_get_int_lints <- function(packages, symbols, symbol_nodes, namespaces, source_expression) {
  ## Case 2: foo does not exist in pkg:::foo
  non_symbols <- !is_in_pkg(symbols, namespaces, exported = FALSE)
  non_symbols_lints <- xml_nodes_to_lints(
    symbol_nodes[non_symbols],
    source_expression = source_expression,
    lint_message = sprintf("'%s' does not exist in {%s}.", symbols[non_symbols], packages[non_symbols]),
    type = "warning"
  )

  packages <- packages[!non_symbols]
  symbols <- symbols[!non_symbols]
  symbol_nodes <- symbol_nodes[!non_symbols]
  namespaces <- namespaces[!non_symbols]

  ## Case 3: foo is already exported pkg:::foo
  exported <- is_in_pkg(symbols, namespaces)
  exported_lints <- xml_nodes_to_lints(
    symbol_nodes[exported],
    source_expression = source_expression,
    lint_message =
      sprintf("Don't use `:::` to access %s, which is exported from %s.", symbols[exported], packages[exported]),
    type = "warning"
  )

  c(non_symbols_lints, exported_lints)
}

build_ns_get_lints <- function(packages, symbols, symbol_nodes, namespaces, source_expression) {
  ## Case 4: foo is not an export in pkg::foo
  unexported <- !is_in_pkg(symbols, namespaces)
  xml_nodes_to_lints(
    symbol_nodes[unexported],
    source_expression = source_expression,
    lint_message = sprintf("'%s' is not exported from {%s}.", symbols[unexported], packages[unexported]),
    type = "warning"
  )
}

#' Block usage of for loops directly overwriting the indexing variable
#'
#' `for (x in x)` is a poor choice of indexing variable. This overwrites
#'   `x` in the calling scope and is confusing to read.
#'
#' @examples
#' # will produce lints
#' lint(
#'   text = "for (x in x) { TRUE }",
#'   linters = for_loop_index_linter()
#' )
#'
#' lint(
#'   text = "for (x in foo(x, y)) { TRUE }",
#'   linters = for_loop_index_linter()
#' )
#'
#' # okay
#' lint(
#'   text = "for (xi in x) { TRUE }",
#'   linters = for_loop_index_linter()
#' )
#'
#' lint(
#'   text = "for (col in DF$col) { TRUE }",
#'   linters = for_loop_index_linter()
#' )
#'
#' @evalRd rd_tags("for_loop_index_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
for_loop_index_linter <- make_linter_from_xpath(
  xpath = "
  //forcond
    /SYMBOL[text() =
      following-sibling::expr
        //SYMBOL[not(parent::expr[OP-DOLLAR or OP-AT or preceding-sibling::OP-LEFT-BRACKET])]
        /text()
    ]
  ",
  lint_message = "Don't re-use any sequence symbols as the index symbol in a for loop."
)

# This file is in 'exclusions' & nothing lints under R config.
# Under DCF config, '# SKIP_LINT' is the exclusion & this line won't lint
1+1 # SKIP_LINT
# This is included as a linter in the DCF, thus this should lint
expect_equal(foo(x), NULL)

# trailing blank line next will lint under DCF

---
title: "Continuous integration"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Continuous integration}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

You can configure `lintr` to run as part of continuous integration (either for a package or a general project containing R files) in order to automatically check that commits and pull requests do not deteriorate code style.

## For packages

First, take special note of the proviso in `?executing_linters` about the need to have your package and its dependencies installed or loaded (e.g. with `pkgload::load_all()`) in order for certain linters (e.g. `object_usage_linter()`) to function as intended.

### GitHub Actions

If your package is on GitHub, the easiest way to do this is with GitHub Actions.
The workflow configuration files use YAML syntax.
The `usethis` package has some great functionality that can help you with workflow files.
The most straightforward way to add a `lintr` workflow to your package is to use the [r-lib/actions](https://github.com/r-lib/actions/)'s [`lint` example](https://github.com/r-lib/actions/tree/v2-branch/examples#lint-workflow).
To do this with `usethis`, you need to call

```r
usethis::use_github_action("lint")
```

This will create a workflow file called `lint.yaml` and place it in the correct location, namely in the `.github/workflows` directory of your repository.
This file configures all the steps required to run `lintr::lint_package()` on your package.

Alternatively you can use the eponymous [`lint-changed-files.yaml`](https://github.com/r-lib/actions/blob/v2-branch/examples/lint-changed-files.yaml) to only lint any changed files:

```r
usethis::use_github_action("lint-changed-files")
```

Comments to the commit or pull request will be printed as
[annotations](https://docs.github.com/en/free-pro-team@latest/github/collaborating-with-issues-and-pull-requests/about-status-checks#types-of-status-checks-on-github)
along side the status check on GitHub.
If you want the builds to produce an error instead of just a warning, you can set the environment variable `LINTR_ERROR_ON_LINT=true`.
This is set by default for both [r-lib/actions](https://github.com/r-lib/actions/)'s `lint.yaml` and `lint-changed-files.yaml`. 
Note that this will kill the R process in case of a lint.

If your project is in a subdirectory and you would like to use GitHub Actions annotations, you can set
`options(lintr.github_annotation_project_dir = "path/to/project")` which will make sure that the annotations point to the correct paths.

## For projects

You are not limited to using `lintr` for packages -- you can use it in combination with continuous integration for any other project.

### GitHub Actions

If your project is on GitHub, you could take advantage of GitHub Actions and the `usethis` functionality.
[r-lib/actions](https://github.com/r-lib/actions/) includes a [`lint-project` example](https://github.com/r-lib/actions/tree/v2-branch/examples#lint-project-workflow), which you can use by calling:

``` r
usethis::use_github_action("lint-project")
```

### Super-Linter

`lintr` powers R lints for [Super-Linter](https://github.com/github/super-linter) and [MegaLinter](https://megalinter.io/latest/),
which provide a unified linting experience across many languages.
Specifically, they execute `lintr::lint()` on the R and R Markdown files included in a given project.

---
title: "Editor setup"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Editor setup}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

in_pkgdown <- identical(Sys.getenv("IN_PKGDOWN"), "true")

maybe_still <- function(url) {
  if (in_pkgdown) {
    url
  } else {
    gsub("\\.gif$", "-still.gif", url)
  }
}
```

```{r, echo = FALSE, results = 'asis'}
if (!in_pkgdown) {
  cat(
    "Note: This vignette is best viewed [online](https://lintr.r-lib.org/articles/editors.html),",
    "where we can render full animations of editor flows.\n"
  )
}
```

## RStudio

lintr lints are automatically displayed in the RStudio Markers pane (RStudio versions \> v0.99.206).

![RStudio Example](rstudio.png "Rstudio Example")

In order to show the "Markers" pane in RStudio: Menu "Tools" -\> "Global Options...", a window with title "Options" will pop up.
In that window: click "Code" on the left; click "Diagnostics" tab; check "Show diagnostics for R".

To lint a source file `test.R` type in the Console `lintr::lint("test.R")` and look at the result in the "Markers" pane.

This package also includes two addins for linting the current source and package.
To bind the addin to a keyboard shortcut navigate to Tools \> addins \> Browse Addins \> Keyboard Shortcuts.
It's recommended to use Alt+Shift+L for linting the current source lint and Ctrl+Shift+Alt+L to code the package.
These are easy to remember as you are Alt+Shift+L(int) ;)

## Emacs

lintr has [built-in integration](http://www.flycheck.org/en/latest/languages.html#r) with [flycheck](https://github.com/flycheck/flycheck) versions greater than `0.23`.
![Emacs Example](`r maybe_still("emacs.gif")` "Emacs Example")

### Installation

lintr is fully integrated into flycheck when using [ESS](http://ess.r-project.org/).
See the installation documentation for those packages for more information.

### Configuration

You can also configure what linters are used.
e.g. using a different line length cutoff.
- `M-x customize-option` -\> `flycheck-lintr-linters` -\> `linters_with_defaults(line_length_linter(120))`

## Vim - syntastic

lintr can be integrated with [syntastic](https://github.com/vim-syntastic/syntastic) for on-the-fly linting.

![Vim Example](`r maybe_still("vim-syntastic.gif")` "Vim Example")

### Installation

Put the file [syntastic/lintr.vim](https://raw.githubusercontent.com/r-lib/lintr/v2.0.1/inst/syntastic/lintr.vim) in `syntastic/syntax_checkers/r`.
If you are using [pathogen](https://github.com/tpope/vim-pathogen) this directory is `~/.vim/bundles/syntastic/syntax_checkers/r`.

You will also need to add the following lines to your `.vimrc`.

``` vim
let g:syntastic_enable_r_lintr_checker = 1
let g:syntastic_r_checkers = ['lintr']
```

### Configuration

You can also configure what linters are used.
e.g. using a different line length cutoff.

``` vim
let g:syntastic_r_lintr_linters = "linters_with_defaults(line_length_linter(120))"
```

## Vim - ALE

lintr can be integrated with [ALE](https://github.com/dense-analysis/ale) for on the fly linting.

### Installation

lintr is integrated with ALE and requires no additional installation.

### Configuration

You can configure what linters are used, e.g. using a different line length cutoff.

``` vim
let g:ale_r_lintr_options = "linters_with_defaults(line_length_linter(120))"
```

You can also configure whether `lint` or `lint_package` is used.
Set to 1 for `lint_package` and 0 (default) for `lint`.

``` vim
let g:ale_r_lintr_lint_package = 1
```

See `:h ale_r_lintr` for more information.

Note that configuration through `.lintr` files are not supported.

There is a work around that can be used to read the contents of a `.lintr` file
in the root of the working directory. This would allow the use of configuration
through `.lintr` files.

``` vim
if filereadable(".lintr")
  let g:ale_r_lintr_options = join(readfile('.lintr'))
endif
```

## Sublime Text 3

lintr can be integrated with [Sublime Linter](https://github.com/SublimeLinter/SublimeLinter) for on-the-fly linting.

![Sublime Example](`r maybe_still("sublime.gif")` "Sublime Example")

### Installation

Simply install `sublimeLinter-contrib-lintr` using [Package Control](https://packagecontrol.io/).

For more information see [Sublime Linter Docs](https://sublimelinter.readthedocs.io/en/latest/installation.html#installing-via-pc)

### Configuration

You can also configure what linters are used.
e.g. disabling the assignment linter and using a different line length cutoff.
In the SublimeLinter User Settings

```json
    {
      "linters": {
        "lintr": {
          "linters": "linters_with_defaults(assignment_linter = NULL, line_length_linter(120))"
        }
      }
    }
```

## Atom

lintr can be integrated with [Linter](https://github.com/steelbrain/linter) for on the fly linting.

![Atom Example](atom.png "Atom Example")

### Installation

Simply install `linter-lintr` from within Atom or on the command line with:

``` bash
apm install linter-lintr
```

For more information and bug reports see [Atom linter-lintr](https://github.com/AtomLinter/linter-lintr).

## Visual Studio Code

In Visual Studio Code, [vscode-R](https://github.com/REditorSupport/vscode-R#r-extension-for-visual-studio-code) presents the lintr diagnostics from [languageserver](https://github.com/REditorSupport/languageserver).

![VS Code Example](vscode.png "VS Code Example")

### Installation

Installing `languageserver` package in R and `vscode-R` extension in VS Code will enable lintr in VS Code by default or run the following command lines:

``` bash
Rscript -e 'install.packages("languageserver")'
code --install-extension reditorsupport.r
```

---
title: "Editor setup"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Editor setup}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

in_pkgdown <- identical(Sys.getenv("IN_PKGDOWN"), "true")

maybe_still <- function(url) {
  if (in_pkgdown) {
    url
  } else {
    gsub("\\.gif$", "-still.gif", url)
  }
}
```

```{r, echo = FALSE, results = 'asis'}
if (!in_pkgdown) {
  cat(
    "Note: This vignette is best viewed [online](https://lintr.r-lib.org/articles/editors.html),",
    "where we can render full animations of editor flows.\n"
  )
}
```

## RStudio

lintr lints are automatically displayed in the RStudio Markers pane (RStudio versions \> v0.99.206).

![RStudio Example](rstudio.png "Rstudio Example")

In order to show the "Markers" pane in RStudio: Menu "Tools" -\> "Global Options...", a window with title "Options" will pop up.
In that window: click "Code" on the left; click "Diagnostics" tab; check "Show diagnostics for R".

To lint a source file `test.R` type in the Console `lintr::lint("test.R")` and look at the result in the "Markers" pane.

This package also includes two addins for linting the current source and package.
To bind the addin to a keyboard shortcut navigate to Tools \> addins \> Browse Addins \> Keyboard Shortcuts.
It's recommended to use Alt+Shift+L for linting the current source lint and Ctrl+Shift+Alt+L to code the package.
These are easy to remember as you are Alt+Shift+L(int) ;)

## Emacs

lintr has [built-in integration](http://www.flycheck.org/en/latest/languages.html#r) with [flycheck](https://github.com/flycheck/flycheck) versions greater than `0.23`.
![Emacs Example](`r maybe_still("emacs.gif")` "Emacs Example")

### Installation

lintr is fully integrated into flycheck when using [ESS](http://ess.r-project.org/).
See the installation documentation for those packages for more information.

### Configuration

You can also configure what linters are used.
e.g. using a different line length cutoff.
- `M-x customize-option` -\> `flycheck-lintr-linters` -\> `linters_with_defaults(line_length_linter(120))`

## Vim - syntastic

lintr can be integrated with [syntastic](https://github.com/vim-syntastic/syntastic) for on-the-fly linting.

![Vim Example](`r maybe_still("vim-syntastic.gif")` "Vim Example")

### Installation

Put the file [syntastic/lintr.vim](https://raw.githubusercontent.com/r-lib/lintr/v2.0.1/inst/syntastic/lintr.vim) in `syntastic/syntax_checkers/r`.
If you are using [pathogen](https://github.com/tpope/vim-pathogen) this directory is `~/.vim/bundles/syntastic/syntax_checkers/r`.

You will also need to add the following lines to your `.vimrc`.

``` vim
let g:syntastic_enable_r_lintr_checker = 1
let g:syntastic_r_checkers = ['lintr']
```

### Configuration

You can also configure what linters are used.
e.g. using a different line length cutoff.

``` vim
let g:syntastic_r_lintr_linters = "linters_with_defaults(line_length_linter(120))"
```

## Vim - ALE

lintr can be integrated with [ALE](https://github.com/dense-analysis/ale) for on the fly linting.

### Installation

lintr is integrated with ALE and requires no additional installation.

### Configuration

You can configure what linters are used, e.g. using a different line length cutoff.

``` vim
let g:ale_r_lintr_options = "linters_with_defaults(line_length_linter(120))"
```

You can also configure whether `lint` or `lint_package` is used.
Set to 1 for `lint_package` and 0 (default) for `lint`.

``` vim
let g:ale_r_lintr_lint_package = 1
```

See `:h ale_r_lintr` for more information.

Note that configuration through `.lintr` files are not supported.

There is a work around that can be used to read the contents of a `.lintr` file
in the root of the working directory. This would allow the use of configuration
through `.lintr` files.

``` vim
if filereadable(".lintr")
  let g:ale_r_lintr_options = join(readfile('.lintr'))
endif
```

## Sublime Text 3

lintr can be integrated with [Sublime Linter](https://github.com/SublimeLinter/SublimeLinter) for on-the-fly linting.

![Sublime Example](`r maybe_still("sublime.gif")` "Sublime Example")

### Installation

Simply install `sublimeLinter-contrib-lintr` using [Package Control](https://packagecontrol.io/).

For more information see [Sublime Linter Docs](http://sublimelinter.readthedocs.io/en/latest/installation.html#installing-via-pc)

### Configuration

You can also configure what linters are used.
e.g. disabling the assignment linter and using a different line length cutoff.
In the SublimeLinter User Settings

```json
    {
      "linters": {
        "lintr": {
          "linters": "linters_with_defaults(assignment_linter = NULL, line_length_linter(120))"
        }
      }
    }
```

## Atom

lintr can be integrated with [Linter](https://github.com/steelbrain/linter) for on the fly linting.

![Atom Example](atom.png "Atom Example")

### Installation

Simply install `linter-lintr` from within Atom or on the command line with:

``` bash
apm install linter-lintr
```

For more information and bug reports see [Atom linter-lintr](https://github.com/AtomLinter/linter-lintr).

## Visual Studio Code

In Visual Studio Code, [vscode-R](https://github.com/REditorSupport/vscode-R#r-extension-for-visual-studio-code) presents the lintr diagnostics from [languageserver](https://github.com/REditorSupport/languageserver).

![VS Code Example](vscode.png "VS Code Example")

### Installation

Installing `languageserver` package in R and `vscode-R` extension in VS Code will enable lintr in VS Code by default or run the following command lines:

``` bash
Rscript -e 'install.packages("languageserver")'
code --install-extension reditorsupport.r
```

---
title: "Creating new linters"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Creating new linters}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

This vignette describes the steps necessary to create a new linter.

See the last section for some details specific to writing new linters for `{lintr}`.

A good example of a simple linter is the `pipe_call_linter`.

```r
#' Pipe call linter
#'
#' Force explicit calls in magrittr pipes, e.g.,
#' `1:3 %>% sum()` instead of `1:3 %>% sum`.
#'
#' @evalRd rd_tags("pipe_call_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
pipe_call_linter <- function() {
  xpath <- "//expr[preceding-sibling::SPECIAL[text() = '%>%'] and *[1][self::SYMBOL]]"

  Linter(linter_level = "expression", function(source_expression) {
    if (!is_lint_level(source_expression, "expression")) {
      return(list())
    }

    xml <- source_expression$xml_parsed_content
    if (is.null(xml)) return(list())

    bad_expr <- xml2::xml_find_all(xml, xpath)

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = "Use explicit calls in magrittr pipes, i.e., `a %>% foo` should be `a %>% foo()`.",
      type = "warning"
    )
  })
}
```

Let's walk through the parts of the linter individually.

## Writing the linter ##

```r
#' Pipe call linter
#'
#' Force explicit calls in magrittr pipes, e.g.,
#' `1:3 %>% sum()` instead of `1:3 %>% sum`.
```

Describe the linter, giving it a title and briefly covering the usages that are discouraged when the linter is active.

```r
#' @evalRd rd_tags("pipe_call_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
```

These lines (1) generate a Tags section in the documentation for the linter^[NB: this is a helper function for generating custom Rd styling. See R/linter_tags.R.]; (2) link to the full table of available linters; and (3) mark the function for export. The most unfamiliar here is probably (1), which can be skipped outside of `lintr` itself.

```r
pipe_call_linter <- function() {
```

Next, we define the name of the new linter.  The convention is to suffix all
linter names with `_linter`. All `_linter` functions are function factories that return a
closure that will do the actual linting function. We could define additional parameters
that are useful for the linter in this function declaration (see, e.g. `assignment_linter`), but
`pipe_call_linter` requires no additional arguments.

```r
xpath <- "//expr[preceding-sibling::SPECIAL[text() = '%>%'] and *[1][self::SYMBOL]]"
```

Here is the core linter logic. `xpath` is an XPath expression for expressions matching the discouraged usage.
`xpath` is saved inside the factory code (as opposed to inside the linter itself) for efficiency. Often,
the `xpath` will be somewhat complicated / involve some assembly code using `paste()` or `glue::glue()`[^See
`infix_spaces_linter()` for an example of this], in which case it is preferable to execute this code only
once when creating the linter; the cached XPath is then re-used on each expression in each file where the
linter is run.

Let's examine the XPath a bit more closely:

```xpath
//expr                  # global search (//) for 'expr' nodes (R expressions), at any nesting level
[                       # node[...] looks for any 'node' satisfying conditions in ...
  preceding-sibling::   # "siblings" are at the same nesting level in XML
    SPECIAL[            # 'SPECIAL' is the parse token for infix operators like %% or %+%
      text() = '%>%'    # text() returns the string associated with this node
    ]                   #
  and                   # combine conditions with 'and'
  *                     # match any node
  [1]                   # match the first such node
  [self::SYMBOL]        # match if the current node is a 'SYMBOL' (i.e., a 'name' in R)
]                       #
```

Taken together, that means we want to match `expr` nodes preceded by the `%>%` infix operator whose first child node is a `name`.
That maps pretty closely to the description of what the `pipe_call_linter` is looking for, but there is subtlety in mapping
between the R code you're used to and how they show up in the XML representation. `expr` nodes in particular take some practice
to get accustomed to -- use the plentiful XPath-based linters in `lintr` as a guide to get extra
practice^[The W3schools tutorials are also quite helpful; see https://www.w3schools.com/xml/xpath_intro.asp]. Note: `xml2`
implements XPath 1.0, which lacks some helpful features available in XPath 2.0.


```r
Linter(function(source_expression) {
```

This is the closure. It will be called on the `source_expression` variable that
contains the top level expressions in the file to be linted. The call to
`Linter()` gives this closure the class 'linter' (it also guesses the name of the linter; see `?Linter` for more details).

The raw text of the expression is available from `source_file$content`. However,
it is not generally possible to implement linters from the raw text -- consider
`equals_na_linter`. If we just look for `== NA` in the text of the file, we'll
generate many false positives, e.g. in comments (such as `# note: is.na() is the proper way to check == NA`)
or inside character literals (such as `warning("don't use == NA to check missingness")`).
We're also likely to generate false negatives, for example when `==` and `NA` appear on different lines!
Working around these issues using only the un-parsed text in every situation amounts to re-implementing the parser.

Therefore it is recommended to work with the tokens from `source_file$parsed_content` or `source_file$xml_parsed_content`,
as they are tokenized from the `R` parser. These tokens are obtained
from `parse()` and `utils::getParseData()` calls done prior to calling the new linter.
`getParseData()` returns a `data.frame` with information from the source parse
tree of the file being linted. A list of tokens is available from
[r-source/src/main/gram.y](https://github.com/r-devel/r-svn/blob/master/src/main/gram.y#L395-L412).

`source_file$xml_parsed_content` uses `xmlparsedata::xml_parse_data()` to convert the `getParseData()` output into an XML tree,
which enables writing linter logic in [XPath](https://www.w3schools.com/xml/xpath_intro.asp), a powerful language
for expressing paths within the nested XML data structure. Most linters in `lintr` are built using XPath because it is a
powerful language for computation on the abstract syntax tree / parse tree.

```r
if (!is_lint_level(source_expression, "expression")) {
  return(list())
}
```

Here, we return early if `source_expression` is not the expression-level object. `get_source_expression()` returns an
object that parses the input file in two ways -- once is done expression-by-expression, the other contains all of the
expressions in the file. This is done to facilitate caching. Suppose your package has a long source file (e.g., 100s of expressions) --
rather than run linters on every expression every time the file is updated, when caching is activated `lintr` will
only run the linter again on expressions that have changed.

Note that this check is unnecessary because we provided `linter_level = "expression"` which guarantees that `source_expression` will be at the expression level and not at the file level.

Therefore, it is preferable to write expression-level linters whenever possible. Two types of exceptions observed in `lintr` are
(1) when several or all expressions are _required_ to ensure the linter logic applies (e.g., `conjunct_test_linter` looks for
consecutive calls to `stopifnot()`, which will typically appear on consecutive expressions) or (2) when the linter logic
is very simple & fast to compute, so that the overhead of re-running the linter is low (e.g., `single_quotes_linter`). In those cases,
use `is_lint_level(source_expression, "file")`.

```r
xml <- source_expression$xml_parsed_content

bad_expr <- xml2::xml_find_all(xml, xpath)
```

`source_expression$xml_parsed_content` is copied to a local variable
(this is not strictly necessary but facilitates debugging). Then `xml2::xml_find_all()` is used to execute the XPath on
this particular expression. Keep in mind that it is typically possible for a single expression to generate more than one
lint -- for example, `x %>% na.omit %>% sum` will trigger the `pipe_call_linter()` twice^[This is particularly important
if you want the `message` field in the resulting `Lint()` to vary depending on the exact violation that's found. For
`pipe_call_linter()`, the message is always the same. See `assignment_linter()` for an example where the `message` can vary.].

```r
xml_nodes_to_lints(
  bad_expr,
  source_expression = source_expression,
  lint_message = "Use explicit calls in magrittr pipes, i.e., `a %>% foo` should be `a %>% foo()`.",
  type = "warning"
)
```

Finally, we pass the matching XML node(s) to `xml_nodes_to_lints()`, which returns `Lint` objects corresponding to any "bad" usages
found in `source_expression`. See `?Lint` and `?xml_nodes_to_lints` for details about the arguments.
Note that here, the `message` for the lint is always the same, but for many linters,
the message is customized to more closely match the observed usage. In such cases, `xml_nodes_to_lint()` can conveniently
accept a function in `lint_message` which takes a node as input and converts it to a customized message.
See, for example, `seq_linter`.

```r
linter_level = "expression"
```

This is a more efficient way to implement the condition `if (!is_lint_level(source_expression, "expression"))`.

## Writing linter tests

(NB: this section uses the `assignment_linter()` which has simpler examples than `pipe_continuation_linter()`.)

`{lintr}` works best inside the `{testthat}` unit testing framework, in particular, `{lintr}`
exports `lintr::expect_lint()` which is designed as a companion to other testthat expectations.

You can define tests inside separate `test_that` calls.  Most of the linters use the same default form.

```r
test_that("returns the correct linting", {
```

You then test a series of expectations for the linter using `expect_lint`.
Please see `?expect_lint` for a full description of the parameters.

The main three aspects to test are:

1. Linter returns no lints when there is nothing to lint, e.g.

```r
expect_no_lint("blah", assignment_linter())
```

2. Linter returns a lint when there is something to lint, e.g.

```r
expect_lint("blah=1",
  rex("Use <-, not =, for assignment."),
  assignment_linter()
)
```

3. As many edge cases as you can think of that might break it, e.g.

```r
expect_lint("fun((blah = fun(1)))",
  rex("Use <-, not =, for assignment."),
  assignment_linter()
)
```

You may want to test that additional `lint` attributes are correct,
such as the type, line number, column number, e.g.

```r
expect_lint("blah=1",
  list(message = "assignment", line_number = 1, column_number = 5, type = "style"),
  assignment_linter()
)
```

Finally, it is a good idea to test that your linter reports multiple lints if
needed, e.g.

```r
expect_lint("blah=1; blah=2",
  list(
    list(line_number = 1, column_number = 5),
    list(line_number = 1, column_number = 13),
  ),
  assignment_linter()
)
```

It is always better to write too many tests rather than too few.

## Other utilities for writing custom linters

Besides `is_lint_level()`, `{lintr}` also exports some other helpers generally useful
for writing custom linters; these are used a lot in the internals of our own helpers,
and so they've been tested and demonstrated their utility already.

 * `get_r_string()`: Whenever your linter needs to examine the value of a character
   literal (e.g., whether an argument value is set to some string), use this to
   extract the string exactly as R will see it. This is especially important to make
   your logic robust to R-4-style raw strings like `R"-(hello)-"`, which is otherwise
   difficult to express, for example as an XPath.
 * `xml_find_function_calls()`: Whenever your linter needs to query R function calls,
   e.g. via the XPath `//SYMBOL_FUNCTION_CALL[text() = 'myfun']`, use this member of
   `source_expression` to obtain the function call nodes more efficiently.
   Instead of
   ```r
   xml <- source_expression$xml_parsed_content
   xpath <- "//SYMBOL_FUNCTION_CALL[text() = 'myfun']/parent::expr/some/cond"
   xml_find_all(xml, xpath)
   ```
   use
   ```r
   xml_calls <- source_expression$xml_find_function_calls("myfun")
   call_xpath <- "some/cond"
   xml_find_all(xml_calls, call_xpath)
   ```
 * `make_linter_from_xpath()` and `make_linter_from_function_xpath()`: Whenever your
   linter can be expressed by a static XPath and a static message, use `make_linter_from_xpath()`
   or, if the XPath starts with `//SYMBOL_FUNCTION_CALL`, use `make_linter_from_function_xpath()`.
   Instead of
   `make_linter_from_xpath(xpath = "//SYMBOL_FUNCTION_CALL[text() = 'foo' or text() = 'bar']")`,
   use
   `make_linter_from_function_xpath(function_names = c("foo", "bar"), xpath = "SYMBOL_FUNCTION_CALL/cond")`.
   Very often, such XPaths look like `//SYMBOL_FUNCTION_CALL/parent::expr/cond2`,
   in which case the `xpath=` is simpler: `xpath = "cond2"`. Another common case is
   `//SYMBOL_FUNCTION_CALL/parent::expr[exprCond]/furtherCond`, which becomes
   `xpath = "self::*[exprCond]/furtherCond"`.

## Contributing to `{lintr}`

### More details about writing tests for new `{lintr}` linters

The `{lintr}` package uses [testthat](https://github.com/r-lib/testthat) for
testing. You can run all of the currently available tests using
`devtools::test()`.  If you want to run only the tests in a given file use the
`filter` argument to `devtools::test()`.

Linter tests should be put in the
[tests/testthat/](https://github.com/r-lib/lintr/tree/main/tests/testthat)
folder.  The test filename should be the linter name prefixed by `test-`, e.g.
`test-pipe_continuation_linter.R`.

### Adding your linter to the default_linters ##

If your linter implements part of the tidyverse style guide you can add it to `default_linters`.
This object is created in the file `zzz.R` (this name ensures that it will always run after all
the linters are defined). Add your linter name to the `default_linters`
list before the `NULL` at the end, and add a corresponding test case to the test script
`./tests/testthat/default_linter_testcode.R`.

### Submit pull request ##

Push your changes to a branch of your fork of the
[lintr](https://github.com/r-lib/lintr) repository, and submit a pull
request to get your linter merged into lintr!

---
title: "Creating new linters"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Creating new linters}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

This vignette describes the steps necessary to create a new linter.

See the last section for some details specific to writing new linters for `{lintr}`.

A good example of a simple linter is the `pipe_call_linter`.

```r
#' Pipe call linter
#'
#' Force explicit calls in magrittr pipes, e.g.,
#' `1:3 %>% sum()` instead of `1:3 %>% sum`.
#'
#' @evalRd rd_tags("pipe_call_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
pipe_call_linter <- function() {
  xpath <- "//expr[preceding-sibling::SPECIAL[text() = '%>%'] and *[1][self::SYMBOL]]"

  Linter(linter_level = "expression", function(source_expression) {
    if (!is_lint_level(source_expression, "expression")) {
      return(list())
    }

    xml <- source_expression$xml_parsed_content
    if (is.null(xml)) return(list())

    bad_expr <- xml2::xml_find_all(xml, xpath)

    xml_nodes_to_lints(
      bad_expr,
      source_expression = source_expression,
      lint_message = "Use explicit calls in magrittr pipes, i.e., `a %>% foo` should be `a %>% foo()`.",
      type = "warning"
    )
  })
}
```

Let's walk through the parts of the linter individually.

## Writing the linter ##

```r
#' Pipe call linter
#'
#' Force explicit calls in magrittr pipes, e.g.,
#' `1:3 %>% sum()` instead of `1:3 %>% sum`.
```

Describe the linter, giving it a title and briefly covering the usages that are discouraged when the linter is active.

```r
#' @evalRd rd_tags("pipe_call_linter")
#' @seealso [linters] for a complete list of linters available in lintr.
#' @export
```

These lines (1) generate a Tags section in the documentation for the linter^[NB: this is a helper function for generating custom Rd styling. See R/linter_tags.R.]; (2) link to the full table of available linters; and (3) mark the function for export. The most unfamiliar here is probably (1), which can be skipped outside of `lintr` itself.

```r
pipe_call_linter <- function() {
```

Next, we define the name of the new linter.  The convention is to suffix all
linter names with `_linter`. All `_linter` functions are function factories that return a
closure that will do the actual linting function. We could define additional parameters
that are useful for the linter in this function declaration (see, e.g. `assignment_linter`), but
`pipe_call_linter` requires no additional arguments.

```r
xpath <- "//expr[preceding-sibling::SPECIAL[text() = '%>%'] and *[1][self::SYMBOL]]"
```

Here is the core linter logic. `xpath` is an XPath expression for expressions matching the discouraged usage.
`xpath` is saved inside the factory code (as opposed to inside the linter itself) for efficiency. Often,
the `xpath` will be somewhat complicated / involve some assembly code using `paste()` or `glue::glue()`[^See
`infix_spaces_linter()` for an example of this], in which case it is preferable to execute this code only
once when creating the linter; the cached XPath is then re-used on each expression in each file where the
linter is run.

Let's examine the XPath a bit more closely:

```xpath
//expr                  # global search (//) for 'expr' nodes (R expressions), at any nesting level
[                       # node[...] looks for any 'node' satisfying conditions in ...
  preceding-sibling::   # "siblings" are at the same nesting level in XML
    SPECIAL[            # 'SPECIAL' is the parse token for infix operators like %% or %+%
      text() = '%>%'    # text() returns the string associated with this node
    ]                   #
  and                   # combine conditions with 'and'
  *                     # match any node
  [1]                   # match the first such node
  [self::SYMBOL]        # match if the current node is a 'SYMBOL' (i.e., a 'name' in R)
]                       #
```

Taken together, that means we want to match `expr` nodes preceded by the `%>%` infix operator whose first child node is a `name`.
That maps pretty closely to the description of what the `pipe_call_linter` is looking for, but there is subtlety in mapping
between the R code you're used to and how they show up in the XML representation. `expr` nodes in particular take some practice
to get accustomed to -- use the plentiful XPath-based linters in `lintr` as a guide to get extra
practice^[The W3schools tutorials are also quite helpful; see https://www.w3schools.com/xml/xpath_intro.asp]. Note: `xml2`
implements XPath 1.0, which lacks some helpful features available in XPath 2.0.


```r
Linter(function(source_expression) {
```

This is the closure. It will be called on the `source_expression` variable that
contains the top level expressions in the file to be linted. The call to
`Linter()` gives this closure the class 'linter' (it also guesses the name of the linter; see `?Linter` for more details).

The raw text of the expression is available from `source_file$content`. However,
it is not generally possible to implement linters from the raw text -- consider
`equals_na_linter`. If we just look for `== NA` in the text of the file, we'll
generate many false positives, e.g. in comments (such as `# note: is.na() is the proper way to check == NA`)
or inside character literals (such as `warning("don't use == NA to check missingness")`).
We're also likely to generate false negatives, for example when `==` and `NA` appear on different lines!
Working around these issues using only the un-parsed text in every situation amounts to re-implementing the parser.

Therefore it is recommended to work with the tokens from `source_file$parsed_content` or `source_file$xml_parsed_content`,
as they are tokenized from the `R` parser. These tokens are obtained
from `parse()` and `utils::getParseData()` calls done prior to calling the new linter.
`getParseData()` returns a `data.frame` with information from the source parse
tree of the file being linted. A list of tokens is available from
[r-source/src/main/gram.y](https://github.com/r-devel/r-svn/blob/master/src/main/gram.y#L395-L412).

`source_file$xml_parsed_content` uses `xmlparsedata::xml_parse_data()` to convert the `getParseData()` output into an XML tree,
which enables writing linter logic in [XPath](https://www.w3schools.com/xml/xpath_intro.asp), a powerful language
for expressing paths within the nested XML data structure. Most linters in `lintr` are built using XPath because it is a
powerful language for computation on the abstract syntax tree / parse tree.

```r
if (!is_lint_level(source_expression, "expression")) {
  return(list())
}
```

Here, we return early if `source_expression` is not the expression-level object. `get_source_expression()` returns an
object that parses the input file in two ways -- once is done expression-by-expression, the other contains all of the
expressions in the file. This is done to facilitate caching. Suppose your package has a long source file (e.g., 100s of expressions) --
rather than run linters on every expression every time the file is updated, when caching is activated `lintr` will
only run the linter again on expressions that have changed.

Note that this check is unnecessary because we provided `linter_level = "expression"` which guarantees that `source_expression` will be at the expression level and not at the file level.

Therefore, it is preferable to write expression-level linters whenever possible. Two types of exceptions observed in `lintr` are
(1) when several or all expressions are _required_ to ensure the linter logic applies (e.g., `conjunct_test_linter` looks for
consecutive calls to `stopifnot()`, which will typically appear on consecutive expressions) or (2) when the linter logic
is very simple & fast to compute, so that the overhead of re-running the linter is low (e.g., `quotes_linter`). In those cases,
use `is_lint_level(source_expression, "file")`.

```r
xml <- source_expression$xml_parsed_content

bad_expr <- xml2::xml_find_all(xml, xpath)
```

`source_expression$xml_parsed_content` is copied to a local variable
(this is not strictly necessary but facilitates debugging). Then `xml2::xml_find_all()` is used to execute the XPath on
this particular expression. Keep in mind that it is typically possible for a single expression to generate more than one
lint -- for example, `x %>% na.omit %>% sum` will trigger the `pipe_call_linter()` twice^[This is particularly important
if you want the `message` field in the resulting `Lint()` to vary depending on the exact violation that's found. For
`pipe_call_linter()`, the message is always the same. See `assignment_linter()` for an example where the `message` can vary.].

```r
xml_nodes_to_lints(
  bad_expr,
  source_expression = source_expression,
  lint_message = "Use explicit calls in magrittr pipes, i.e., `a %>% foo` should be `a %>% foo()`.",
  type = "warning"
)
```

Finally, we pass the matching XML node(s) to `xml_nodes_to_lints()`, which returns `Lint` objects corresponding to any "bad" usages
found in `source_expression`. See `?Lint` and `?xml_nodes_to_lints` for details about the arguments.
Note that here, the `message` for the lint is always the same, but for many linters,
the message is customized to more closely match the observed usage. In such cases, `xml_nodes_to_lint()` can conveniently
accept a function in `lint_message` which takes a node as input and converts it to a customized message.
See, for example, `seq_linter`.

```r
linter_level = "expression"
```

This is a more efficient way to implement the condition `if (!is_lint_level(source_expression, "expression"))`.

## Writing linter tests

(NB: this section uses the `assignment_linter()` which has simpler examples than `pipe_continuation_linter()`.)

`{lintr}` works best inside the `{testthat}` unit testing framework, in particular, `{lintr}`
exports `lintr::expect_lint()` which is designed as a companion to other testthat expectations.

You can define tests inside separate `test_that` calls.  Most of the linters use the same default form.

```r
test_that("returns the correct linting", {
```

You then test a series of expectations for the linter using `expect_lint`.
Please see `?expect_lint` for a full description of the parameters.

The main three aspects to test are:

1. Linter returns no lints when there is nothing to lint, e.g.

```r
expect_no_lint("blah", assignment_linter())
```

2. Linter returns a lint when there is something to lint, e.g.

```r
expect_lint("blah=1",
  rex("Use <-, not =, for assignment."),
  assignment_linter()
)
```

3. As many edge cases as you can think of that might break it, e.g.

```r
expect_lint("fun((blah = fun(1)))",
  rex("Use <-, not =, for assignment."),
  assignment_linter()
)
```

You may want to test that additional `lint` attributes are correct,
such as the type, line number, column number, e.g.

```r
expect_lint("blah=1",
  list(message = "assignment", line_number = 1, column_number = 5, type = "style"),
  assignment_linter()
)
```

Finally, it is a good idea to test that your linter reports multiple lints if
needed, e.g.

```r
expect_lint("blah=1; blah=2",
  list(
    list(line_number = 1, column_number = 5),
    list(line_number = 1, column_number = 13),
  ),
  assignment_linter()
)
```

It is always better to write too many tests rather than too few.

## Other utilities for writing custom linters

Besides `is_lint_level()`, `{lintr}` also exports some other helpers generally useful
for writing custom linters; these are used a lot in the internals of our own helpers,
and so they've been tested and demonstrated their utility already.

 * `get_r_string()`: Whenever your linter needs to examine the value of a character
   literal (e.g., whether an argument value is set to some string), use this to
   extract the string exactly as R will see it. This is especially important to make
   your logic robust to R-4-style raw strings like `R"-(hello)-"`, which is otherwise
   difficult to express, for example as an XPath.
 * `xml_find_function_calls()`: Whenever your linter needs to query R function calls,
   e.g. via the XPath `//SYMBOL_FUNCTION_CALL[text() = 'myfun']`, use this member of
   `source_expression` to obtain the function call nodes more efficiently.
   Instead of
   ```r
   xml <- source_expression$xml_parsed_content
   xpath <- "//SYMBOL_FUNCTION_CALL[text() = 'myfun']/parent::expr/some/cond"
   xml_find_all(xml, xpath)
   ```
   use
   ```r
   xml_calls <- source_expression$xml_find_function_calls("myfun")
   call_xpath <- "some/cond"
   xml_find_all(xml_calls, call_xpath)
   ```
 * `make_linter_from_xpath()` and `make_linter_from_function_xpath()`: Whenever your
   linter can be expressed by a static XPath and a static message, use `make_linter_from_xpath()`
   or, if the XPath starts with `//SYMBOL_FUNCTION_CALL`, use `make_linter_from_function_xpath()`.
   Instead of
   `make_linter_from_xpath(xpath = "//SYMBOL_FUNCTION_CALL[text() = 'foo' or text() = 'bar']")`,
   use
   `make_linter_from_function_xpath(function_names = c("foo", "bar"), xpath = "SYMBOL_FUNCTION_CALL/cond")`.
   Very often, such XPaths look like `//SYMBOL_FUNCTION_CALL/parent::expr/cond2`,
   in which case the `xpath=` is simpler: `xpath = "cond2"`. Another common case is
   `//SYMBOL_FUNCTION_CALL/parent::expr[exprCond]/furtherCond`, which becomes
   `xpath = "self::*[exprCond]/furtherCond"`.

## Contributing to `{lintr}`

### More details about writing tests for new `{lintr}` linters

The `{lintr}` package uses [testthat](https://github.com/r-lib/testthat) for
testing. You can run all of the currently available tests using
`devtools::test()`.  If you want to run only the tests in a given file use the
`filter` argument to `devtools::test()`.

Linter tests should be put in the
[tests/testthat/](https://github.com/r-lib/lintr/tree/main/tests/testthat)
folder.  The test filename should be the linter name prefixed by `test-`, e.g.
`test-pipe_continuation_linter.R`.

### Adding your linter to the default_linters ##

If your linter implements part of the tidyverse style guide you can add it to `default_linters`.
This object is created in the file `zzz.R` (this name ensures that it will always run after all
the linters are defined). Add your linter name to the `default_linters`
list before the `NULL` at the end, and add a corresponding test case to the test script
`./tests/testthat/default_linter_testcode.R`.

### Submit pull request ##

Push your changes to a branch of your fork of the
[lintr](https://github.com/r-lib/lintr) repository, and submit a pull
request to get your linter merged into lintr!

---
documentclass: book
output: bookdown::html_document2
---

# Examples

```{definition}
The characteristic function of a random variable $X$ is defined by
$$\varphi _{X}(t)=\operatorname {E} \left[e^{itX}\right], \; t\in\mathcal{R}$$
```

```{r}
a = 1
```

# print.lint, print.lints support optional message wrapping : width = 10

    Code
      print(lints, width = width)
    Output
      <text>:1:1:
          warning:
          [test_linter]
          The
          quick
          brown
          fox
          jumps
          over
          the
          lazy
          dog.
      a
      ^

---

    Code
      print(lints)
    Output
      <text>:1:1:
          warning:
          [test_linter]
          The
          quick
          brown
          fox
          jumps
          over
          the
          lazy
          dog.
      a
      ^

# print.lint, print.lints support optional message wrapping : width = 20

    Code
      print(lints, width = width)
    Output
      <text>:1:1:
          warning:
          [test_linter]
          The quick brown
          fox jumps over
          the lazy dog.
      a
      ^

---

    Code
      print(lints)
    Output
      <text>:1:1:
          warning:
          [test_linter]
          The quick brown
          fox jumps over
          the lazy dog.
      a
      ^

# print.lint, print.lints support optional message wrapping : width = 40

    Code
      print(lints, width = width)
    Output
      <text>:1:1: warning: [test_linter] The
          quick brown fox jumps over the lazy
          dog.
      a
      ^

---

    Code
      print(lints)
    Output
      <text>:1:1: warning: [test_linter] The
          quick brown fox jumps over the lazy
          dog.
      a
      ^

# print.lint, print.lints support optional message wrapping : width = 80

    Code
      print(lints, width = width)
    Output
      <text>:1:1: warning: [test_linter] The quick brown fox jumps over the lazy dog.
      a
      ^

---

    Code
      print(lints)
    Output
      <text>:1:1: warning: [test_linter] The quick brown fox jumps over the lazy dog.
      a
      ^

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cyclocomp_linter.R
\name{cyclocomp_linter}
\alias{cyclocomp_linter}
\title{Cyclomatic complexity linter}
\usage{
cyclocomp_linter(complexity_limit = 15L)
}
\arguments{
\item{complexity_limit}{Maximum cyclomatic complexity, default \code{15}. Expressions more complex
than this are linted.}
}
\description{
Check for overly complicated expressions. See \code{cyclocomp()} function from \code{{cyclocomp}}.
}
\examples{
\dontshow{if (requireNamespace("cyclocomp", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# will produce lints
lint(
  text = "if (TRUE) 1 else 2",
  linters = cyclocomp_linter(complexity_limit = 1L)
)

# okay
lint(
  text = "if (TRUE) 1 else 2",
  linters = cyclocomp_linter(complexity_limit = 2L)
)
\dontshow{\}) # examplesIf}
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/with.R
\name{modify_defaults}
\alias{modify_defaults}
\title{Modify lintr defaults}
\usage{
modify_defaults(defaults, ...)
}
\arguments{
\item{defaults}{named list of elements to modify.}

\item{...}{arguments of elements to change. If unnamed, the argument is automatically named.
If the named argument already exists in \code{defaults}, it is replaced by the new element.
If it does not exist, it is added. If the value is \code{NULL}, the element is removed.}
}
\value{
A modified list of elements, sorted by name. To achieve this sort in a platform-independent way, two
transformations are applied to the names: (1) replace \verb{_} with \code{0} and (2) convert \code{\link[=tolower]{tolower()}}.
}
\description{
Modify a list of defaults by name, allowing for replacement, deletion and addition of new elements.
}
\examples{
# custom list of undesirable functions:
#    remove `sapply` (using `NULL`)
#    add `cat` (with an accompanying message),
#    add `print` (unnamed, i.e. with no accompanying message)
#    add `source` (as taken from `all_undesirable_functions`)
my_undesirable_functions <- modify_defaults(
  defaults = default_undesirable_functions,
  sapply = NULL, "cat" = "No cat allowed", "print", all_undesirable_functions[["source"]]
)

# list names of functions specified as undesirable
names(my_undesirable_functions)
}
\seealso{
\itemize{
\item \link{linters_with_defaults} for basing off lintr's set of default linters.
\item \link{all_linters} for basing off all available linters in lintr.
\item \link{linters_with_tags} for basing off tags attached to linters, possibly across multiple packages.
\item \link{available_linters} to get a data frame of available linters.
\item \link{linters} for a complete list of linters available in lintr.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conjunct_test_linter.R
\name{conjunct_test_linter}
\alias{conjunct_test_linter}
\title{Force \code{&&} conditions to be written separately where appropriate}
\usage{
conjunct_test_linter(
  allow_named_stopifnot = TRUE,
  allow_filter = c("never", "not_dplyr", "always")
)
}
\arguments{
\item{allow_named_stopifnot}{Logical, \code{TRUE} by default. If \code{FALSE}, "named" calls to \code{stopifnot()} are also linted.}

\item{allow_filter}{Character naming the method for linting calls to \code{filter()}. The default, \code{"never"}, means
\code{filter()} and \code{dplyr::filter()} calls are linted; \code{"not_dplyr"} means only \code{dplyr::filter()} calls are linted;
and \code{"always"} means no calls to \code{filter()} are linted. Calls like \code{stats::filter()} are never linted.}
}
\description{
For readability of test outputs, testing only one thing per call to
\code{\link[testthat:expect_true]{testthat::expect_true()}} is preferable, i.e.,
\verb{expect_true(A); expect_true(B)} is better than \code{expect_true(A && B)}, and
\verb{expect_false(A); expect_false(B)} is better than \code{expect_false(A || B)}.
}
\details{
Similar reasoning applies to \code{&&} usage inside \code{\link[base:stopifnot]{base::stopifnot()}} and \code{assertthat::assert_that()} calls.

Relatedly, \code{dplyr::filter(DF, A & B)} is the same as \code{dplyr::filter(DF, A, B)}, but the latter will be more readable
/ easier to format for long conditions. Note that this linter assumes usages of \code{filter()} are \code{dplyr::filter()};
if you're using another function named \code{filter()}, e.g. \code{\link[stats:filter]{stats::filter()}}, please namespace-qualify it to avoid
false positives. You can omit linting \code{filter()} expressions altogether via \code{allow_filter = TRUE}.
}
\examples{
# will produce lints
lint(
  text = "expect_true(x && y)",
  linters = conjunct_test_linter()
)

lint(
  text = "expect_false(x || (y && z))",
  linters = conjunct_test_linter()
)

lint(
  text = "stopifnot('x must be a logical scalar' = length(x) == 1 && is.logical(x) && !is.na(x))",
  linters = conjunct_test_linter(allow_named_stopifnot = FALSE)
)

lint(
  text = "dplyr::filter(mtcars, mpg > 20 & vs == 0)",
  linters = conjunct_test_linter()
)

lint(
  text = "filter(mtcars, mpg > 20 & vs == 0)",
  linters = conjunct_test_linter()
)

# okay
lint(
  text = "expect_true(x || (y && z))",
  linters = conjunct_test_linter()
)

lint(
  text = 'stopifnot("x must be a logical scalar" = length(x) == 1 && is.logical(x) && !is.na(x))',
  linters = conjunct_test_linter(allow_named_stopifnot = TRUE)
)

lint(
  text = "dplyr::filter(mtcars, mpg > 20 & vs == 0)",
  linters = conjunct_test_linter(allow_filter = "always")
)

lint(
  text = "filter(mtcars, mpg > 20 & vs == 0)",
  linters = conjunct_test_linter(allow_filter = "not_dplyr")
)

lint(
  text = "stats::filter(mtcars$cyl, mtcars$mpg > 20 & mtcars$vs == 0)",
  linters = conjunct_test_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=package_development_linters]{package_development}, \link[=pkg_testthat_linters]{pkg_testthat}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missing_package_linter.R
\name{missing_package_linter}
\alias{missing_package_linter}
\title{Missing package linter}
\usage{
missing_package_linter()
}
\description{
Check for missing packages in \code{library()}, \code{require()}, \code{loadNamespace()}, and \code{requireNamespace()} calls.
}
\examples{
# will produce lints
lint(
  text = "library(xyzxyz)",
  linters = missing_package_linter()
)

# okay
lint(
  text = "library(stats)",
  linters = missing_package_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=common_mistakes_linters]{common_mistakes}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missing_argument_linter.R
\name{missing_argument_linter}
\alias{missing_argument_linter}
\title{Missing argument linter}
\usage{
missing_argument_linter(
  except = c("alist", "quote", "switch"),
  allow_trailing = FALSE
)
}
\arguments{
\item{except}{a character vector of function names as exceptions.}

\item{allow_trailing}{always allow trailing empty arguments?}
}
\description{
Check for missing arguments in function calls (e.g. \code{stats::median(1:10, )}).
}
\examples{
# will produce lints
lint(
  text = 'tibble(x = "a", )',
  linters = missing_argument_linter()
)

# okay
lint(
  text = 'tibble(x = "a")',
  linters = missing_argument_linter()
)

lint(
  text = 'tibble(x = "a", )',
  linters = missing_argument_linter(except = "tibble")
)

lint(
  text = 'tibble(x = "a", )',
  linters = missing_argument_linter(allow_trailing = TRUE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=common_mistakes_linters]{common_mistakes}, \link[=configurable_linters]{configurable}, \link[=correctness_linters]{correctness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_boundary_linter.R
\name{string_boundary_linter}
\alias{string_boundary_linter}
\title{Require usage of \code{startsWith()} and \code{endsWith()} over \code{grepl()}/\code{substr()} versions}
\usage{
string_boundary_linter(allow_grepl = FALSE)
}
\arguments{
\item{allow_grepl}{Logical, default \code{FALSE}. If \code{TRUE}, usages with \code{grepl()}
are ignored. Some authors may prefer the conciseness offered by \code{grepl()} whereby
\code{NA} input maps to \code{FALSE} output, which doesn't have a direct equivalent
with \code{startsWith()} or \code{endsWith()}.}
}
\description{
\code{\link[base:startsWith]{base::startsWith()}} is used to detect fixed initial substrings; it is more
readable and more efficient than equivalents using \code{\link[=grepl]{grepl()}} or \code{\link[=substr]{substr()}}.
c.f. \code{startsWith(x, "abc")}, \code{grepl("^abc", x)},
\code{substr(x, 1L, 3L) == "abc"}.
}
\details{
Ditto for using \code{\link[base:endsWith]{base::endsWith()}} to detect fixed terminal substrings.

Note that there is a difference in behavior between how \code{grepl()} and \code{startsWith()}
(and \code{endsWith()}) handle missing values. In particular, for \code{grepl()}, \code{NA} inputs
are considered \code{FALSE}, while for \code{startsWith()}, \code{NA} inputs have \code{NA} outputs.
That means the strict equivalent of \code{grepl("^abc", x)} is
\code{!is.na(x) & startsWith(x, "abc")}.

We lint \code{grepl()} usages by default because the \code{!is.na()} version is more explicit
with respect to \code{NA} handling -- though documented, the way \code{grepl()} handles
missing inputs may be surprising to some users.
}
\examples{
# will produce lints
lint(
  text = 'grepl("^a", x)',
  linters = string_boundary_linter()
)

lint(
  text = 'grepl("z$", x)',
  linters = string_boundary_linter()
)

# okay
lint(
  text = 'startsWith(x, "a")',
  linters = string_boundary_linter()
)

lint(
  text = 'endsWith(x, "z")',
  linters = string_boundary_linter()
)

# If missing values are present, the suggested alternative wouldn't be strictly
# equivalent, so this linter can also be turned off in such cases.
lint(
  text = 'grepl("z$", x)',
  linters = string_boundary_linter(allow_grepl = TRUE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}, \link[=regex_linters]{regex}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lintr-package.R
\docType{package}
\name{lintr-package}
\alias{lintr}
\alias{lintr-package}
\title{Lintr}
\description{
Checks adherence to a given style, syntax errors, and possible semantic
issues.
Supports on the fly checking of R code edited with Emacs, Vim, and
Sublime Text.
}
\seealso{
\code{\link[=lint]{lint()}}, \code{\link[=lint_package]{lint_package()}}, \code{\link[=lint_dir]{lint_dir()}}, \link{linters}
}
\author{
\strong{Maintainer}: Michael Chirico \email{michaelchirico4@gmail.com} (\href{https://orcid.org/0000-0003-0787-087X}{ORCID})

Authors:
\itemize{
  \item Michael Chirico \email{michaelchirico4@gmail.com} (\href{https://orcid.org/0000-0003-0787-087X}{ORCID})
  \item Jim Hester
  \item Florent Angly (GitHub: fangly)
  \item Russ Hyde
  \item Kun Ren
  \item Alexander Rosenstock (GitHub: AshesITR)
  \item Indrajeet Patil \email{patilindrajeet.science@gmail.com} (\href{https://orcid.org/0000-0003-1995-6531}{ORCID})
  \item Hugo Gruson (\href{https://orcid.org/0000-0002-4094-1476}{ORCID})
  \item Emmanuel Ferdman
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{correctness_linters}
\alias{correctness_linters}
\title{Correctness linters}
\description{
Linters highlighting possible programming mistakes, such as unused variables.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'correctness':
\itemize{
\item{\code{\link{duplicate_argument_linter}}}
\item{\code{\link{equals_na_linter}}}
\item{\code{\link{missing_argument_linter}}}
\item{\code{\link{namespace_linter}}}
\item{\code{\link{object_usage_linter}}}
\item{\code{\link{package_hooks_linter}}}
\item{\code{\link{sprintf_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seq_linter.R
\name{seq_linter}
\alias{seq_linter}
\title{Sequence linter}
\usage{
seq_linter()
}
\description{
This linter checks for \code{1:length(...)}, \code{1:nrow(...)}, \code{1:ncol(...)},
\code{1:NROW(...)} and \code{1:NCOL(...)} expressions in base-R, or their usage in
conjunction with \code{seq()} (e.g., \code{seq(length(...))}, \code{seq(nrow(...))}, etc.).
}
\details{
Additionally, it checks for \code{1:n()} (from \code{{dplyr}}) and \code{1:.N} (from \code{{data.table}}).

These often cause bugs when the right-hand side is zero.
Instead, it is safer to use \code{\link[base:seq_len]{base::seq_len()}} (to create a sequence of a specified \emph{length}) or
\code{\link[base:seq_along]{base::seq_along()}} (to create a sequence \emph{along} an object).
}
\examples{
# will produce lints
lint(
  text = "seq(length(x))",
  linters = seq_linter()
)

lint(
  text = "1:nrow(x)",
  linters = seq_linter()
)

lint(
  text = "dplyr::mutate(x, .id = 1:n())",
  linters = seq_linter()
)

lint(
  text = "seq_len(length(x))",
  linters = seq_linter()
)

lint(
  text = "unlist(lapply(x, seq_len))",
  linters = seq_linter()
)

# okay
lint(
  text = "seq_along(x)",
  linters = seq_linter()
)

lint(
  text = "seq_len(nrow(x))",
  linters = seq_linter()
)

lint(
  text = "dplyr::mutate(x, .id = seq_len(n()))",
  linters = seq_linter()
)

lint(
  text = "seq_along(x)",
  linters = seq_linter()
)

lint(
  text = "sequence(x)",
  linters = seq_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=default_linters]{default}, \link[=efficiency_linters]{efficiency}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inner_combine_linter.R
\name{inner_combine_linter}
\alias{inner_combine_linter}
\title{Require \code{c()} to be applied before relatively expensive vectorized functions}
\usage{
inner_combine_linter()
}
\description{
\code{as.Date(c(a, b))} is logically equivalent to \code{c(as.Date(a), as.Date(b))}.
The same equivalence holds for several other vectorized functions like
\code{\link[=as.POSIXct]{as.POSIXct()}} and math functions like \code{\link[=sin]{sin()}}. The former is to be
preferred so that the most expensive part of the operation (\code{\link[=as.Date]{as.Date()}})
is applied only once.
}
\details{
Note that \code{\link[=strptime]{strptime()}} has one idiosyncrasy to be aware of, namely that
auto-detected \verb{format=} is set by the first matching input, which means
that a case like \code{c(as.POSIXct("2024-01-01"), as.POSIXct("2024-01-01 01:02:03"))}
gives different results to \code{as.POSIXct(c("2024-01-01", "2024-01-01 01:02:03"))}.
This false positive is rare; a workaround where possible is to use
consistent formatting, i.e., \code{"2024-01-01 00:00:00"} in the example.
}
\examples{
# will produce lints
lint(
  text = "c(log10(x), log10(y), log10(z))",
  linters = inner_combine_linter()
)

# okay
lint(
  text = "log10(c(x, y, z))",
  linters = inner_combine_linter()
)

lint(
  text = "c(log(x, base = 10), log10(x, base = 2))",
  linters = inner_combine_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=consistency_linters]{consistency}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use_lintr.R
\name{use_lintr}
\alias{use_lintr}
\title{Use lintr in your project}
\usage{
use_lintr(path = ".", type = c("tidyverse", "full"))
}
\arguments{
\item{path}{Path to project root, where a \code{.lintr} file should be created.
If the \code{.lintr} file already exists, an error will be thrown.}

\item{type}{What kind of configuration to create?
\itemize{
\item \code{tidyverse} creates a minimal lintr config, based on the default linters (\code{\link[=linters_with_defaults]{linters_with_defaults()}}).
These are suitable for following \href{https://style.tidyverse.org/}{the tidyverse style guide}.
\item \code{full} creates a lintr config using all available linters via \code{\link[=all_linters]{all_linters()}}.
}}
}
\value{
Path to the generated configuration, invisibly.
}
\description{
Create a minimal lintr config file as a starting point for customization
}
\examples{
if (FALSE) {
  # use the default set of linters
  lintr::use_lintr()
  # or try all linters
  lintr::use_lintr(type = "full")

  # then
  lintr::lint_dir()
}
}
\seealso{
\code{vignette("lintr")} for detailed introduction to using and configuring lintr.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spaces_left_parentheses_linter.R
\name{spaces_left_parentheses_linter}
\alias{spaces_left_parentheses_linter}
\title{Spaces before parentheses linter}
\usage{
spaces_left_parentheses_linter()
}
\description{
Check that all left parentheses have a space before them unless they are in a function call.
}
\examples{
# will produce lints
lint(
  text = "if(TRUE) x else y",
  linters = spaces_left_parentheses_linter()
)

# okay
lint(
  text = "if (TRUE) x else y",
  linters = spaces_left_parentheses_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#parentheses}
\item \code{\link[=function_left_parentheses_linter]{function_left_parentheses_linter()}}
}
}
\section{Tags}{
\link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/implicit_assignment_linter.R
\name{implicit_assignment_linter}
\alias{implicit_assignment_linter}
\title{Avoid implicit assignment in function calls}
\usage{
implicit_assignment_linter(
  except = c("bquote", "expression", "expr", "quo", "quos", "quote"),
  allow_lazy = FALSE,
  allow_scoped = FALSE,
  allow_paren_print = FALSE
)
}
\arguments{
\item{except}{A character vector of functions to be excluded from linting.}

\item{allow_lazy}{logical, default \code{FALSE}. If \code{TRUE}, assignments that only
trigger conditionally (e.g. in the RHS of \code{&&} or \code{||} expressions) are skipped.}

\item{allow_scoped}{Logical, default \code{FALSE}. If \code{TRUE}, "scoped assignments",
where the object is assigned in the statement beginning a branch and used only
within that branch, are skipped.}

\item{allow_paren_print}{Logical, default \code{FALSE}. If \code{TRUE}, assignments using
\code{(} for auto-printing at the top-level are not linted.}
}
\description{
Assigning inside function calls makes the code difficult to read, and should
be avoided, except for functions that capture side-effects (e.g. \code{\link[utils:capture.output]{utils::capture.output()}}).
}
\examples{
# will produce lints
lint(
  text = "if (x <- 1L) TRUE",
  linters = implicit_assignment_linter()
)

lint(
  text = "mean(x <- 1:4)",
  linters = implicit_assignment_linter()
)

lint(
  text = "(x <- 1)",
  linters = implicit_assignment_linter()
)


# okay
lines <- "x <- 1L\nif (x) TRUE"
writeLines(lines)
lint(
  text = lines,
  linters = implicit_assignment_linter()
)

lines <- "x <- 1:4\nmean(x)"
writeLines(lines)
lint(
  text = lines,
  linters = implicit_assignment_linter()
)

lint(
  text = "A && (B <- foo(A))",
  linters = implicit_assignment_linter(allow_lazy = TRUE)
)

lines <- c(
  "if (any(idx <- x < 0)) {",
  "  stop('negative elements: ', toString(which(idx)))",
  "}"
)
writeLines(lines)
lint(
  text = lines,
  linters = implicit_assignment_linter(allow_scoped = TRUE)
)

lint(
  text = "(x <- 1)",
  linters = implicit_assignment_linter(allow_paren_print = TRUE)
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#assignment}
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function_return_linter.R
\name{function_return_linter}
\alias{function_return_linter}
\title{Lint common mistakes/style issues cropping up from return statements}
\usage{
function_return_linter()
}
\description{
\code{return(x <- ...)} is either distracting (because \code{x} is ignored), or
confusing (because assigning to \code{x} has some side effect that is muddled
by the dual-purpose expression).
}
\examples{
# will produce lints
lint(
  text = "foo <- function(x) return(y <- x + 1)",
  linters = function_return_linter()
)

lint(
  text = "foo <- function(x) return(x <<- x + 1)",
  linters = function_return_linter()
)

writeLines("e <- new.env() \nfoo <- function(x) return(e$val <- x + 1)")
lint(
  text = "e <- new.env() \nfoo <- function(x) return(e$val <- x + 1)",
  linters = function_return_linter()
)

# okay
lint(
  text = "foo <- function(x) return(x + 1)",
  linters = function_return_linter()
)

code_lines <- "
foo <- function(x) {
  x <<- x + 1
  return(x)
}
"
lint(
  text = code_lines,
  linters = function_return_linter()
)

code_lines <- "
e <- new.env()
foo <- function(x) {
  e$val <- x + 1
  return(e$val)
}
"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = function_return_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consecutive_assertion_linter.R
\name{consecutive_assertion_linter}
\alias{consecutive_assertion_linter}
\title{Force consecutive calls to assertions into just one when possible}
\usage{
consecutive_assertion_linter()
}
\description{
\code{\link[base:stopifnot]{base::stopifnot()}} accepts any number of tests, so sequences like
\verb{stopifnot(x); stopifnot(y)} are redundant. Ditto for tests using
\code{assertthat::assert_that()} without specifying \verb{msg=}.
}
\examples{
# will produce lints
lint(
  text = "stopifnot(x); stopifnot(y)",
  linters = consecutive_assertion_linter()
)

lint(
  text = "assert_that(x); assert_that(y)",
  linters = consecutive_assertion_linter()
)

# okay
lint(
  text = "stopifnot(x, y)",
  linters = consecutive_assertion_linter()
)

lint(
  text = 'assert_that(x, msg = "Bad x!"); assert_that(y)',
  linters = consecutive_assertion_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_numeric_linter.R
\name{is_numeric_linter}
\alias{is_numeric_linter}
\title{Redirect \code{is.numeric(x) || is.integer(x)} to just use \code{is.numeric(x)}}
\usage{
is_numeric_linter()
}
\description{
\code{\link[=is.numeric]{is.numeric()}} returns \code{TRUE} when \code{typeof(x)} is \code{double} or \code{integer} --
testing \code{is.numeric(x) || is.integer(x)} is thus redundant.
}
\details{
NB: This linter plays well with \code{\link[=class_equals_linter]{class_equals_linter()}}, which can help
avoid further \code{is.numeric()} equivalents like
\code{any(class(x) == c("numeric", "integer"))}.
}
\examples{
# will produce lints
lint(
  text = "is.numeric(y) || is.integer(y)",
  linters = is_numeric_linter()
)

lint(
  text = 'class(z) \%in\% c("numeric", "integer")',
  linters = is_numeric_linter()
)

# okay
lint(
  text = "is.numeric(y) || is.factor(y)",
  linters = is_numeric_linter()
)

lint(
  text = 'class(z) \%in\% c("numeric", "integer", "factor")',
  linters = is_numeric_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assignment_linter.R
\name{assignment_linter}
\alias{assignment_linter}
\title{Assignment linter}
\usage{
assignment_linter(operator = "<-", allow_trailing = TRUE)
}
\arguments{
\item{operator}{Character vector of valid assignment operators. Defaults to allowing only \verb{<-}; other valid
options are \code{=}, \verb{<<-}, \verb{->}, \verb{->>}, \verb{\%<>\%}; use \code{"any"} to denote "allow all operators", in which case}

\item{allow_trailing}{Logical, default \code{TRUE}. If \code{FALSE} then assignments aren't allowed at end of lines.}
}
\description{
Check that the specified operator is used for assignment.
}
\examples{
# will produce lints
lint(
  text = "x = mean(x)",
  linters = assignment_linter()
)

code_lines <- "1 -> x\n2 ->> y\nz <<- 3"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = assignment_linter()
)

lint(
  text = "x \%<>\% as.character()",
  linters = assignment_linter()
)

lint(
  text = "x <- 1",
  linters = assignment_linter(operator = "=")
)

# okay
lint(
  text = "x <- mean(x)",
  linters = assignment_linter()
)

code_lines <- "x <- 1\ny <- 2\nz <- 3"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = assignment_linter()
)

# customizing using arguments
code_lines <- "1 -> x\n2 ->> y"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = assignment_linter(operator = "->")
)

lint(
  text = "x <<- 1",
  linters = assignment_linter(operator = "<-")
)

writeLines("foo(bar = \n 1)")
lint(
  text = "foo(bar = \n 1)",
  linters = assignment_linter(allow_trailing = FALSE)
)

lint(
  text = "x \%<>\% as.character()",
  linters = assignment_linter(operator = "\%<>\%")
)

lint(
  text = "x = 1",
  linters = assignment_linter(operator = "=")
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#assignment-1}
\item \url{https://style.tidyverse.org/pipes.html#assignment-2}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=default_linters]{default}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{efficiency_linters}
\alias{efficiency_linters}
\title{Efficiency linters}
\description{
Linters highlighting code efficiency problems, such as unnecessary function calls.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'efficiency':
\itemize{
\item{\code{\link{any_duplicated_linter}}}
\item{\code{\link{any_is_na_linter}}}
\item{\code{\link{boolean_arithmetic_linter}}}
\item{\code{\link{consecutive_mutate_linter}}}
\item{\code{\link{fixed_regex_linter}}}
\item{\code{\link{if_switch_linter}}}
\item{\code{\link{ifelse_censor_linter}}}
\item{\code{\link{inner_combine_linter}}}
\item{\code{\link{length_test_linter}}}
\item{\code{\link{lengths_linter}}}
\item{\code{\link{list2df_linter}}}
\item{\code{\link{literal_coercion_linter}}}
\item{\code{\link{matrix_apply_linter}}}
\item{\code{\link{nested_ifelse_linter}}}
\item{\code{\link{nrow_subset_linter}}}
\item{\code{\link{nzchar_linter}}}
\item{\code{\link{outer_negation_linter}}}
\item{\code{\link{redundant_equals_linter}}}
\item{\code{\link{redundant_ifelse_linter}}}
\item{\code{\link{regex_subset_linter}}}
\item{\code{\link{routine_registration_linter}}}
\item{\code{\link{sample_int_linter}}}
\item{\code{\link{scalar_in_linter}}}
\item{\code{\link{seq_linter}}}
\item{\code{\link{sort_linter}}}
\item{\code{\link{string_boundary_linter}}}
\item{\code{\link{unnecessary_concatenation_linter}}}
\item{\code{\link{unnecessary_lambda_linter}}}
\item{\code{\link{vector_logic_linter}}}
\item{\code{\link{which_grepl_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redundant_equals_linter.R
\name{redundant_equals_linter}
\alias{redundant_equals_linter}
\title{Block usage of \code{==}, \code{!=} on logical vectors}
\usage{
redundant_equals_linter()
}
\description{
Testing \code{x == TRUE} is redundant if \code{x} is a logical vector. Wherever this is
used to improve readability, the solution should instead be to improve the
naming of the object to better indicate that its contents are logical. This
can be done using prefixes (is, has, can, etc.). For example, \code{is_child},
\code{has_parent_supervision}, \code{can_watch_horror_movie} clarify their logical
nature, while \code{child}, \code{parent_supervision}, \code{watch_horror_movie} don't.
}
\examples{
# will produce lints
lint(
  text = "if (any(x == TRUE)) 1",
  linters = redundant_equals_linter()
)

lint(
  text = "if (any(x != FALSE)) 0",
  linters = redundant_equals_linter()
)

lint(
  text = "dt[is_tall == FALSE, y]",
  linters = redundant_equals_linter()
)

# okay
lint(
  text = "if (any(x)) 1",
  linters = redundant_equals_linter()
)

lint(
  text = "if (!all(x)) 0",
  linters = redundant_equals_linter()
)

# in `{data.table}` semantics, `dt[x]` is a join, `dt[(x)]` is a subset
lint(
  text = "dt[(!is_tall), y]",
  linters = redundant_equals_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \code{\link[=outer_negation_linter]{outer_negation_linter()}}
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=common_mistakes_linters]{common_mistakes}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sort_linter.R
\name{sort_linter}
\alias{sort_linter}
\title{Check for common mistakes around sorting vectors}
\usage{
sort_linter()
}
\description{
This linter checks for some common mistakes when using \code{\link[=order]{order()}} or \code{\link[=sort]{sort()}}.
}
\details{
First, it requires usage of \code{sort()} over \code{.[order(.)]}.

\code{\link[=sort]{sort()}} is the dedicated option to sort a list or vector. It is more legible
and around twice as fast as \code{.[order(.)]}, with the gap in performance
growing with the vector size.

Second, it requires usage of \code{\link[=is.unsorted]{is.unsorted()}} over equivalents using \code{sort()}.

The base function \code{is.unsorted()} exists to test the sortedness of a vector.
Prefer it to inefficient and less-readable equivalents like
\code{x != sort(x)}. The same goes for checking \code{x == sort(x)} -- use
\code{!is.unsorted(x)} instead.

Moreover, use of \code{x == sort(x)} can be risky because \code{\link[=sort]{sort()}} drops missing
elements by default, meaning \code{==} might end up trying to compare vectors
of differing lengths.
}
\examples{
# will produce lints
lint(
  text = "x[order(x)]",
  linters = sort_linter()
)

lint(
  text = "x[order(x, decreasing = TRUE)]",
  linters = sort_linter()
)

lint(
  text = "sort(x) == x",
  linters = sort_linter()
)

lint(
  text = "rev(sort(x))",
  linters = sort_linter()
)

# okay
lint(
  text = "x[sample(order(x))]",
  linters = sort_linter()
)

lint(
  text = "y[order(x)]",
  linters = sort_linter()
)

lint(
  text = "sort(x, decreasing = TRUE) == x",
  linters = sort_linter()
)

lint(
  text = "sort(x, decreasing = TRUE)",
  linters = sort_linter()
)

# If you are sorting several objects based on the order of one of them, such
# as:
x <- sample(1:26)
y <- letters
newx <- x[order(x)]
newy <- y[order(x)]
# This will be flagged by the linter. However, in this very specific case,
# it would be clearer and more efficient to run order() once and assign it
# to an object, rather than mix and match order() and sort()
index <- order(x)
newx <- x[index]
newy <- y[index]

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quotes_linter.R
\name{quotes_linter}
\alias{quotes_linter}
\title{Character string quote linter}
\usage{
quotes_linter(delimiter = c("\\"", "'"))
}
\arguments{
\item{delimiter}{Which quote delimiter to accept. Defaults to the tidyverse
default of \verb{"} (double-quoted strings).}
}
\description{
Check that the desired quote delimiter is used for string constants.
}
\examples{
# will produce lints
lint(
  text = "c('a', 'b')",
  linters = quotes_linter()
)

# okay
lint(
  text = 'c("a", "b")',
  linters = quotes_linter()
)

code_lines <- "paste0(x, '\"this is fine\"')"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = quotes_linter()
)

# okay
lint(
  text = "c('a', 'b')",
  linters = quotes_linter(delimiter = "'")
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#character-vectors}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/if_switch_linter.R
\name{if_switch_linter}
\alias{if_switch_linter}
\title{Require usage of switch() over repeated if/else blocks}
\usage{
if_switch_linter(max_branch_lines = 0L, max_branch_expressions = 0L)
}
\arguments{
\item{max_branch_lines, max_branch_expressions}{Integer, default 0 indicates "no maximum".
If set any \code{if}/\verb{else if}/.../\verb{else} chain where any branch occupies more than
this number of lines (resp. expressions) will not be linted. The conjugate
applies to \code{switch()} statements -- if these parameters are set, any \code{switch()}
statement with any overly-complicated branches will be linted. See examples.}
}
\description{
\code{\link[=switch]{switch()}} statements in R are used to delegate behavior based
on the value of some input scalar string, e.g.
\code{switch(x, a = 1, b = 3, c = 7, d = 8)} will be one of
\code{1}, \code{3}, \code{7}, or \code{8}, depending on the value of \code{x}.
}
\details{
This can also be accomplished by repeated \code{if}/\verb{else} statements like
so: \code{if (x == "a") 1 else if (x == "b") 2 else if (x == "c") 7 else 8}
(implicitly, the last \verb{else} assumes x only takes 4 possible values),
but this is more cluttered and slower (note that \code{switch()} takes the same
time to evaluate regardless of the value of \code{x}, and is faster even
when \code{x} takes the first value (here \code{a}), and that the \code{if}/\verb{else}
approach is roughly linear in the number of conditions that need to
be evaluated, here up to 3 times).
}
\examples{
# will produce lints
lint(
  text = "if (x == 'a') 1 else if (x == 'b') 2 else 3",
  linters = if_switch_linter()
)

code <- paste(
  "if (x == 'a') {",
  "  1",
  "} else if (x == 'b') {",
  "  2",
  "} else if (x == 'c') {",
  "  y <- x",
  "  z <- sqrt(match(y, letters))",
  "  z",
  "}",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = if_switch_linter()
)

code <- paste(
  "if (x == 'a') {",
  "  1",
  "} else if (x == 'b') {",
  "  2",
  "} else if (x == 'c') {",
  "  y <- x",
  "  z <- sqrt(",
  "    match(y, letters)",
  "  )",
  "  z",
  "}",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = if_switch_linter()
)

code <- paste(
  "switch(x,",
  "  a = {",
  "    1",
  "    2",
  "    3",
  "  },",
  "  b = {",
  "    1",
  "    2",
  "  }",
  ")",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = if_switch_linter(max_branch_lines = 2L)
)

# okay
lint(
  text = "switch(x, a = 1, b = 2, 3)",
  linters = if_switch_linter()
)

# switch() version not as clear
lint(
  text = "if (x == 'a') 1 else if (x == 'b' & y == 2) 2 else 3",
  linters = if_switch_linter()
)

code <- paste(
  "if (x == 'a') {",
  "  1",
  "} else if (x == 'b') {",
  "  2",
  "} else if (x == 'c') {",
  "  y <- x",
  "  z <- sqrt(match(y, letters))",
  "  z",
  "}",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = if_switch_linter(max_branch_lines = 2L)
)

code <- paste(
  "if (x == 'a') {",
  "  1",
  "} else if (x == 'b') {",
  "  2",
  "} else if (x == 'c') {",
  "  y <- x",
  "  z <- sqrt(",
  "    match(y, letters)",
  "  )",
  "  z",
  "}",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = if_switch_linter(max_branch_expressions = 2L)
)

code <- paste(
  "switch(x,",
  "  a = {",
  "    1",
  "    2",
  "    3",
  "  },",
  "  b = {",
  "    1",
  "    2",
  "  }",
  ")",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = if_switch_linter(max_branch_lines = 3L)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coalesce_linter.R
\name{coalesce_linter}
\alias{coalesce_linter}
\title{Encourage usage of the null coalescing operator \verb{\%||\%}}
\usage{
coalesce_linter()
}
\description{
The \code{x \%||\% y} is equivalent to
\code{if (is.null(x)) y else x}, but more expressive.
It is exported by R since 4.4.0, and equivalents
have been available in other tidyverse packages
for much longer, e.g. 2008 for ggplot2.
}
\examples{
# will produce lints
lint(
  text = "if (is.null(x)) y else x",
  linters = coalesce_linter()
)

lint(
  text = "if (!is.null(x)) x else y",
  linters = coalesce_linter()
)

lint(
  text = "if (is.null(x[1])) x[2] else x[1]",
  linters = coalesce_linter()
)

# okay
lint(
  text = "x \%||\% y",
  linters = coalesce_linter()
)

lint(
  text = "x \%||\% y",
  linters = coalesce_linter()
)

lint(
  text = "x[1] \%||\% x[2]",
  linters = coalesce_linter()
)


}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/with.R
\name{all_linters}
\alias{all_linters}
\title{Create a linter configuration based on all available linters}
\usage{
all_linters(..., packages = "lintr")
}
\arguments{
\item{...}{Arguments of elements to change. If unnamed, the argument is automatically named.
If the named argument already exists in the list of linters, it is replaced by the new element.
If it does not exist, it is added. If the value is \code{NULL}, the linter is removed.}

\item{packages}{A character vector of packages to search for linters.}
}
\description{
Create a linter configuration based on all available linters
}
\examples{
names(all_linters())

}
\seealso{
\itemize{
\item \link{linters_with_defaults} for basing off lintr's set of default linters.
\item \link{linters_with_tags} for basing off tags attached to linters, possibly across multiple packages.
\item \link{available_linters} to get a data frame of available linters.
\item \link{linters} for a complete list of linters available in lintr.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/length_test_linter.R
\name{length_test_linter}
\alias{length_test_linter}
\title{Check for a common mistake where a size check like 'length' is applied in the wrong place}
\usage{
length_test_linter()
}
\description{
Usage like \code{length(x == 0)} is a mistake. If you intended to check \code{x} is empty,
use \code{length(x) == 0}. Other mistakes are possible, but running \code{length()} on the
outcome of a logical comparison is never the best choice.
}
\details{
The linter also checks for similar usage with \code{nrow()}, \code{ncol()}, \code{NROW()}, and \code{NCOL()}.
}
\examples{
# will produce lints
lint(
  text = "length(x == 0)",
  linters = length_test_linter()
)

lint(
  text = "nrow(x > 0) || ncol(x > 0)",
  linters = length_test_linter()
)

lint(
  text = "NROW(x == 1) && NCOL(y == 1)",
  linters = length_test_linter()
)

# okay
lint(
  text = "length(x) > 0",
  linters = length_test_linter()
)

lint(
  text = "nrow(x) > 0 || ncol(x) > 0",
  linters = length_test_linter()
)

lint(
  text = "NROW(x) == 1 && NCOL(y) == 1",
  linters = length_test_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/one_call_pipe_linter.R
\name{one_call_pipe_linter}
\alias{one_call_pipe_linter}
\title{Block single-call magrittr pipes}
\usage{
one_call_pipe_linter()
}
\description{
Prefer using a plain call instead of a pipe with only one call,
i.e. \code{1:10 \%>\% sum()} should instead be \code{sum(1:10)}. Note that
calls in the first \verb{\%>\%} argument count. \code{rowSums(x) \%>\% max()} is OK
because there are two total calls (\code{rowSums()} and \code{max()}).
}
\details{
Note also that un-"called" steps are \emph{not} counted, since they should
be calls (see \code{\link[=pipe_call_linter]{pipe_call_linter()}}).
}
\examples{
# will produce lints
lint(
  text = "(1:10) \%>\% sum()",
  linters = one_call_pipe_linter()
)

lint(
  text = "DT \%>\% .[grp == 'a', sum(v)]",
  linters = one_call_pipe_linter()
)

# okay
lint(
  text = "rowSums(x) \%>\% mean()",
  linters = one_call_pipe_linter()
)

lint(
  text = "DT[src == 'a', .N, by = grp] \%>\% .[N > 10]",
  linters = one_call_pipe_linter()
)

# assignment pipe is exempted
lint(
  text = "DF \%<>\% mutate(a = 2)",
  linters = one_call_pipe_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/pipes.html#short-pipes}
}
}
\section{Tags}{
\link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_nodes_to_lints.R
\name{xml_nodes_to_lints}
\alias{xml_nodes_to_lints}
\title{Convert an XML node or nodeset into a Lint}
\usage{
xml_nodes_to_lints(
  xml,
  source_expression,
  lint_message,
  type = c("style", "warning", "error"),
  column_number_xpath = range_start_xpath,
  range_start_xpath = "number(./@col1)",
  range_end_xpath = "number(./@col2)"
)
}
\arguments{
\item{xml}{An \code{xml_node} object (to generate one \code{Lint}) or an
\code{xml_nodeset} object (to generate several \code{Lint}s), e.g. as returned by
\code{\link[xml2:xml_find_all]{xml2::xml_find_all()}} or \code{\link[xml2:xml_find_first]{xml2::xml_find_first()}} or a
list of \code{xml_node} objects.}

\item{source_expression}{A source expression object, e.g. as
returned typically by \code{\link[=lint]{lint()}}, or more generally
by \code{\link[=get_source_expressions]{get_source_expressions()}}.}

\item{lint_message}{The message to be included as the \code{message}
to the \code{Lint} object. If \code{lint_message} is a character vector the same length as \code{xml},
the \code{i}-th lint will be given the \code{i}-th message.}

\item{type}{type of lint.}

\item{column_number_xpath}{XPath expression to return the column number location of the lint.
Defaults to the start of the range matched by \code{range_start_xpath}. See details for more information.}

\item{range_start_xpath}{XPath expression to return the range start location of the lint.
Defaults to the start of the expression matched by \code{xml}. See details for more information.}

\item{range_end_xpath}{XPath expression to return the range end location of the lint.
Defaults to the end of the expression matched by \code{xml}. See details for more information.}
}
\value{
For \code{xml_node}s, a \code{lint}. For \code{xml_nodeset}s, \code{lints} (a list of \code{lint}s).
}
\description{
Convenience function for converting nodes matched by XPath-based
linter logic into a \code{\link[=Lint]{Lint()}} object to return.
}
\details{
The location XPaths, \code{column_number_xpath}, \code{range_start_xpath} and \code{range_end_xpath} are evaluated using
\code{\link[xml2:xml_find_num]{xml2::xml_find_num()}} and will usually be of the form \code{"number(./relative/xpath)"}.
Note that the location line number cannot be changed and lints spanning multiple lines will ignore \code{range_end_xpath}.
\code{column_number_xpath} and \code{range_start_xpath} are assumed to always refer to locations on the starting line of the
\code{xml} node.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/implicit_integer_linter.R
\name{implicit_integer_linter}
\alias{implicit_integer_linter}
\title{Implicit integer linter}
\usage{
implicit_integer_linter(allow_colon = FALSE)
}
\arguments{
\item{allow_colon}{Logical, default \code{FALSE}. If \code{TRUE}, expressions involving \code{:}
won't throw a lint regardless of whether the inputs are implicitly integers.}
}
\description{
Check that integers are explicitly typed using the form \code{1L} instead of \code{1}.
}
\examples{
# will produce lints
lint(
  text = "x <- 1",
  linters = implicit_integer_linter()
)

lint(
  text = "x[2]",
  linters = implicit_integer_linter()
)

lint(
  text = "1:10",
  linters = implicit_integer_linter()
)

# okay
lint(
  text = "x <- 1.0",
  linters = implicit_integer_linter()
)

lint(
  text = "x <- 1L",
  linters = implicit_integer_linter()
)

lint(
  text = "x[2L]",
  linters = implicit_integer_linter()
)

lint(
  text = "1:10",
  linters = implicit_integer_linter(allow_colon = TRUE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comparison_negation_linter.R
\name{comparison_negation_linter}
\alias{comparison_negation_linter}
\title{Block usages like !(x == y) where a direct relational operator is appropriate}
\usage{
comparison_negation_linter()
}
\description{
\code{!(x == y)} is more readably expressed as \code{x != y}. The same is true of
other negations of simple comparisons like \code{!(x > y)} and \code{!(x <= y)}.
}
\examples{
# will produce lints
lint(
  text = "!x == 2",
  linters = comparison_negation_linter()
)

lint(
  text = "!(x > 2)",
  linters = comparison_negation_linter()
)

# okay
lint(
  text = "!(x == 2 & y > 2)",
  linters = comparison_negation_linter()
)

lint(
  text = "!(x & y)",
  linters = comparison_negation_linter()
)

lint(
  text = "x != 2",
  linters = comparison_negation_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/yoda_test_linter.R
\name{yoda_test_linter}
\alias{yoda_test_linter}
\title{Block obvious "yoda tests"}
\usage{
yoda_test_linter()
}
\description{
Yoda tests use \verb{(expected, actual)} instead of the more common \verb{(actual, expected)}.
This is not always possible to detect statically; this linter focuses on
the simple case of testing an expression against a literal value, e.g.
\verb{(1L, foo(x))} should be \verb{(foo(x), 1L)}.
}
\examples{
# will produce lints
lint(
  text = "expect_equal(2, x)",
  linters = yoda_test_linter()
)

lint(
  text = 'expect_identical("a", x)',
  linters = yoda_test_linter()
)

# okay
lint(
  text = "expect_equal(x, 2)",
  linters = yoda_test_linter()
)

lint(
  text = 'expect_identical(x, "a")',
  linters = yoda_test_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
\url{https://en.wikipedia.org/wiki/Yoda_conditions}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=package_development_linters]{package_development}, \link[=pkg_testthat_linters]{pkg_testthat}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/with.R
\name{linters_with_tags}
\alias{linters_with_tags}
\title{Create a tag-based linter configuration}
\usage{
linters_with_tags(tags, ..., packages = "lintr", exclude_tags = "deprecated")
}
\arguments{
\item{tags}{Optional character vector of tags to search. Only linters with at least one matching tag will be
returned. If \code{tags} is \code{NULL}, all linters will be returned. See \code{available_tags("lintr")} to find out what
tags are already used by lintr.}

\item{...}{Arguments of elements to change. If unnamed, the argument is automatically named.
If the named argument already exists in the list of linters, it is replaced by the new element.
If it does not exist, it is added. If the value is \code{NULL}, the linter is removed.}

\item{packages}{A character vector of packages to search for linters.}

\item{exclude_tags}{Tags to exclude from the results. Linters with at least one matching tag will not be returned.
If \code{exclude_tags} is \code{NULL}, no linters will be excluded. Note that \code{tags} takes priority, meaning that any
tag found in both \code{tags} and \code{exclude_tags} will be included, not excluded. Note that linters with tag \code{"defunct"}
(which do not work and can no longer be run) cannot be queried directly.}
}
\value{
A modified list of linters.
}
\description{
Make a new list based on all linters provided by \code{packages} and tagged with \code{tags}.
The result of this function is meant to be passed to the \code{linters} argument of \code{lint()},
or to be put in your configuration file.
}
\examples{
# `linters_with_defaults()` and `linters_with_tags("default")` are the same:
all.equal(linters_with_defaults(), linters_with_tags("default"))

# Get all linters useful for package development
linters <- linters_with_tags(tags = c("package_development", "style"))
names(linters)

# Get all linters tagged as "default" from lintr and mypkg
if (FALSE) {
  linters_with_tags("default", packages = c("lintr", "mypkg"))
}
}
\seealso{
\itemize{
\item \link{linters_with_defaults} for basing off lintr's set of default linters.
\item \link{all_linters} for basing off all available linters in lintr.
\item \link{available_linters} to get a data frame of available linters.
\item \link{linters} for a complete list of linters available in lintr.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lint.R
\name{sarif_output}
\alias{sarif_output}
\title{SARIF Report for lint results}
\usage{
sarif_output(lints, filename = "lintr_results.sarif")
}
\arguments{
\item{lints}{the linting results.}

\item{filename}{the name of the output report}
}
\description{
Generate a report of the linting results using the \href{https://sarifweb.azurewebsites.net/}{SARIF} format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{regex_linters}
\alias{regex_linters}
\title{Regular expression linters}
\description{
Linters that examine the usage of regular expressions and functions executing them in user code.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'regex':
\itemize{
\item{\code{\link{fixed_regex_linter}}}
\item{\code{\link{regex_subset_linter}}}
\item{\code{\link{string_boundary_linter}}}
\item{\code{\link{which_grepl_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/object_overwrite_linter.R
\name{object_overwrite_linter}
\alias{object_overwrite_linter}
\title{Block assigning any variables whose name clashes with a \code{base} R function}
\usage{
object_overwrite_linter(
  packages = c("base", "stats", "utils", "tools", "methods", "graphics", "grDevices"),
  allow_names = character()
)
}
\arguments{
\item{packages}{Character vector of packages to search for names that should
be avoided. Defaults to the most common default packages: base, stats,
utils, tools, methods, graphics, and grDevices.}

\item{allow_names}{Character vector of object names to ignore, i.e., which
are allowed to collide with exports from \code{packages}.}
}
\description{
Re-using existing names creates a risk of subtle error best avoided.
Avoiding this practice also encourages using better, more descriptive names.
}
\examples{
# will produce lints
code <- "function(x) {\n  data <- x\n  data\n}"
writeLines(code)
lint(
  text = code,
  linters = object_overwrite_linter()
)

code <- "function(x) {\n  lint <- 'fun'\n  lint\n}"
writeLines(code)
lint(
  text = code,
  linters = object_overwrite_linter(packages = "lintr")
)

# okay
code <- "function(x) {\n  data('mtcars')\n}"
writeLines(code)
lint(
  text = code,
  linters = object_overwrite_linter()
)

code <- "function(x) {\n  data <- x\n  data\n}"
writeLines(code)
lint(
  text = code,
  linters = object_overwrite_linter(packages = "base")
)

# names in function signatures are ignored
lint(
  text = "function(data) data <- subset(data, x > 0)",
  linters = object_overwrite_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#object-names}
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=executing_linters]{executing}, \link[=readability_linters]{readability}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{read_settings}
\alias{read_settings}
\title{Read lintr settings}
\usage{
read_settings(filename, call = parent.frame())
}
\arguments{
\item{filename}{Source file to be linted.}

\item{call}{Passed to malformed to ensure linear trace.}
}
\description{
Lintr searches for settings for a given source file in the following order:
\enumerate{
\item options defined as \code{linter.setting}.
\item \code{linter_file} in the same directory
\item \code{linter_file} in the project directory
\item \code{linter_file} in the user home directory
\item \code{\link[=default_settings]{default_settings()}}
}
}
\details{
The default linter_file name is \code{.lintr} but it can be changed with option \code{lintr.linter_file}
or the environment variable \code{R_LINTR_LINTER_FILE}
This file is a DCF file, see \code{\link[base:read.dcf]{base::read.dcf()}} for details.
Here is an example of a \code{.lintr} file:

\if{html}{\out{<div class="sourceCode">}}\preformatted{linters: linters_with_defaults(
    any_duplicated_linter(),
    any_is_na_linter(),
    backport_linter("oldrel-4", except = c("R_user_dir", "str2lang")),
    line_length_linter(120L),
    missing_argument_linter(),
    unnecessary_concatenation_linter(allow_single_expression = FALSE),
    yoda_test_linter()
  )
exclusions: list(
    "inst/doc/creating_linters.R" = 1,
    "inst/example/bad.R",
    "tests/testthat/default_linter_testcode.R",
    "tests/testthat/dummy_packages"
  )
}\if{html}{\out{</div>}}

Experimentally, we also support keeping the config in a plain R file. By default we look for
a file named \code{.lintr.R} (in the same directories where we search for \code{.lintr}).
We are still deciding the future of config support in lintr, so user feedback is welcome.
The advantage of R is that it maps more closely to how the configs are actually stored,
whereas the DCF approach requires somewhat awkward formatting of parseable R code within
valid DCF key-value pairs. The main disadvantage of the R file is it might be \emph{too} flexible,
with users tempted to write configs with side effects causing hard-to-detect bugs or
like YAML could work, but require new dependencies and are harder to parse
both programmatically and visually.
Here is an example of a \code{.lintr.R} file:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{linters <- linters_with_defaults(
    any_duplicated_linter(),
    any_is_na_linter(),
    backport_linter("oldrel-4", except = c("R_user_dir", "str2lang")),
    line_length_linter(120L),
    missing_argument_linter(),
    unnecessary_concatenation_linter(allow_single_expression = FALSE),
    yoda_test_linter()
  )
exclusions <- list(
    "inst/doc/creating_linters.R" = 1,
    "inst/example/bad.R",
    "tests/testthat/default_linter_testcode.R",
    "tests/testthat/dummy_packages"
  )
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{clear_cache}
\alias{clear_cache}
\title{Clear the lintr cache}
\usage{
clear_cache(file = NULL, path = NULL)
}
\arguments{
\item{file}{filename whose cache to clear. If you pass \code{NULL}, it will delete all of the caches.}

\item{path}{directory to store caches. Reads option 'lintr.cache_directory' as the default.}
}
\value{
0 for success, 1 for failure, invisibly.
}
\description{
Clear the lintr cache
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consecutive_mutate_linter.R
\name{consecutive_mutate_linter}
\alias{consecutive_mutate_linter}
\title{Require consecutive calls to mutate() to be combined when possible}
\usage{
consecutive_mutate_linter(invalid_backends = "dbplyr")
}
\arguments{
\item{invalid_backends}{Character vector of packages providing dplyr backends
which may not be compatible with combining \code{mutate()} calls in all cases.
Defaults to \code{"dbplyr"} since not all SQL backends can handle re-using
a variable defined in the same \code{mutate()} expression.}
}
\description{
\code{dplyr::mutate()} accepts any number of columns, so sequences like
\code{DF \%>\% dplyr::mutate(..1) \%>\% dplyr::mutate(..2)} are redundant --
they can always be expressed with a single call to \code{dplyr::mutate()}.
}
\details{
An exception is for some SQL back-ends, where the translation logic may not be
as sophisticated as that in the default \code{dplyr}, for example in
\code{DF \%>\% mutate(a = a + 1) \%>\% mutate(b = a - 2)}.
}
\examples{
# will produce lints
lint(
  text = "x \%>\% mutate(a = 1) \%>\% mutate(b = 2)",
  linters = consecutive_mutate_linter()
)

# okay
lint(
  text = "x \%>\% mutate(a = 1, b = 2)",
  linters = consecutive_mutate_linter()
)

code <- "library(dbplyr)\nx \%>\% mutate(a = 1) \%>\% mutate(a = a + 1)"
writeLines(code)
lint(
  text = code,
  linters = consecutive_mutate_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{default_settings}
\alias{default_settings}
\alias{settings}
\alias{config}
\alias{lintr-config}
\alias{lintr-settings}
\alias{.lintr}
\title{Default lintr settings}
\usage{
default_settings
}
\description{
The default settings consist of
\itemize{
\item \code{linters}: a list of default linters (see \code{\link[=default_linters]{default_linters()}})
\item \code{encoding}: the character encoding assumed for the file
\item \code{exclude}: pattern used to exclude a line of code
\item \code{exclude_start}, \code{exclude_end}: patterns used to mark start and end of the code block to exclude
\item \code{exclude_linter}, \code{exclude_linter_sep}: patterns used to exclude linters
\item \code{exclusions}: a list of exclusions, see \code{\link[=exclude]{exclude()}} for a complete description of valid values.
\item \code{cache_directory}: location of cache directory
\item \code{error_on_lint}: decides if error should be produced when any lints are found
}

There are no settings without defaults, i.e., this list describes every valid setting.
}
\examples{
# available settings
names(default_settings)

# linters included by default
names(default_settings$linters)

# default values for a few of the other settings
default_settings[c(
  "encoding",
  "exclude",
  "exclude_start",
  "exclude_end",
  "exclude_linter",
  "exclude_linter_sep",
  "exclusions",
  "error_on_lint"
)]

}
\seealso{
\code{\link[=read_settings]{read_settings()}}, \link{default_linters}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backport_linter.R
\name{backport_linter}
\alias{backport_linter}
\title{Backport linter}
\usage{
backport_linter(r_version = getRversion(), except = character())
}
\arguments{
\item{r_version}{Minimum R version to test for compatibility. Defaults to
the R version currently in use. The version can be specified as a version
number, or as a version alias (such as \code{"devel"}, \code{"oldrel"}, \code{"oldrel-1"}).}

\item{except}{Character vector of functions to be excluded from linting.
Use this to list explicitly defined backports, e.g. those imported from the \code{{backports}} package or manually
defined in your package.}
}
\description{
Check for usage of unavailable functions. Not reliable for testing r-devel dependencies.
}
\examples{
# will produce lints
lint(
  text = "trimws(x)",
  linters = backport_linter("3.0.0")
)

lint(
  text = "str2lang(x)",
  linters = backport_linter("3.2.0")
)

lint(
  text = "deparse1(expr)",
  linters = backport_linter("3.6.0")
)

# okay
lint(
  text = "trimws(x)",
  linters = backport_linter("3.6.0")
)

lint(
  text = "str2lang(x)",
  linters = backport_linter("3.2.0", except = "str2lang")
)

# Version aliases instead of numbers can also be passed to `r_version`
lint(
  text = "deparse1(expr)",
  linters = backport_linter("release")
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=package_development_linters]{package_development}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipe_call_linter.R
\name{pipe_call_linter}
\alias{pipe_call_linter}
\title{Pipe call linter}
\usage{
pipe_call_linter()
}
\description{
Force explicit calls in magrittr pipes, e.g., \code{1:3 \%>\% sum()} instead of \code{1:3 \%>\% sum}.
Note that native pipe always requires a function call, i.e. \verb{1:3 |> sum} will produce an error.
}
\examples{
# will produce lints
lint(
  text = "1:3 \%>\% mean \%>\% as.character",
  linters = pipe_call_linter()
)

# okay
lint(
  text = "1:3 \%>\% mean() \%>\% as.character()",
  linters = pipe_call_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/for_loop_index_linter.R
\name{for_loop_index_linter}
\alias{for_loop_index_linter}
\title{Block usage of for loops directly overwriting the indexing variable}
\usage{
for_loop_index_linter()
}
\description{
\verb{for (x in x)} is a poor choice of indexing variable. This overwrites
\code{x} in the calling scope and is confusing to read.
}
\examples{
# will produce lints
lint(
  text = "for (x in x) { TRUE }",
  linters = for_loop_index_linter()
)

lint(
  text = "for (x in foo(x, y)) { TRUE }",
  linters = for_loop_index_linter()
)

# okay
lint(
  text = "for (xi in x) { TRUE }",
  linters = for_loop_index_linter()
)

lint(
  text = "for (col in DF$col) { TRUE }",
  linters = for_loop_index_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=readability_linters]{readability}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/object_name_linter.R
\name{object_name_linter}
\alias{object_name_linter}
\title{Object name linter}
\usage{
object_name_linter(styles = c("snake_case", "symbols"), regexes = character())
}
\arguments{
\item{styles}{A subset of
\Sexpr[stage=render, results=rd]{lintr:::regexes_rd}. A name should
match at least one of these styles. The \code{"symbols"} style refers to
names containing \emph{only} non-alphanumeric characters; e.g., defining \verb{\%+\%}
from ggplot2 or \verb{\%>\%} from magrittr would not generate lint markers,
whereas \verb{\%m+\%} from lubridate (containing both alphanumeric \emph{and}
non-alphanumeric characters) would.}

\item{regexes}{A (possibly named) character vector specifying a custom naming convention.
If named, the names will be used in the lint message. Otherwise, the regexes enclosed by \code{/} will be used in the
lint message.
Note that specifying \code{regexes} overrides the default \code{styles}. So if you want to combine \code{regexes} and \code{styles},
both need to be explicitly specified.}
}
\description{
Check that object names conform to a naming style.
The default naming styles are "snake_case" and "symbols".
}
\details{
Quotes (\verb{`"'}) and specials (\verb{\%} and trailing \verb{<-}) are not considered part of the object name.

Note when used in a package, in order to ignore objects imported
from other namespaces, this linter will attempt \code{\link[=getNamespaceExports]{getNamespaceExports()}}
whenever an \code{import(PKG)} or \code{importFrom(PKG, ...)} statement is found
in your NAMESPACE file. If \code{\link[=requireNamespace]{requireNamespace()}} fails (e.g., the package
is not yet installed), the linter won't be able to ignore some usages
that would otherwise be allowed.

Suppose, for example, you have \code{import(upstream)} in your NAMESPACE,
which makes available its exported S3 generic function
\code{a_really_quite_long_function_name} that you then extend in your package
by defining a corresponding method for your class \code{my_class}.
Then, if \code{upstream} is not installed when this linter runs, a lint
will be thrown on this object (even though you don't "own" its full name).

The best way to get lintr to work correctly is to install the package so
that it's available in the session where this linter is running.
}
\examples{
# will produce lints
lint(
  text = "my_var <- 1L",
  linters = object_name_linter(styles = "CamelCase")
)

lint(
  text = "xYz <- 1L",
  linters = object_name_linter(styles = c("UPPERCASE", "lowercase"))
)

lint(
  text = "MyVar <- 1L",
  linters = object_name_linter(styles = "dotted.case")
)

lint(
  text = "asd <- 1L",
  linters = object_name_linter(regexes = c(my_style = "F$", "f$"))
)

# okay
lint(
  text = "my_var <- 1L",
  linters = object_name_linter(styles = "snake_case")
)

lint(
  text = "xyz <- 1L",
  linters = object_name_linter(styles = "lowercase")
)

lint(
  text = "my.var <- 1L; myvar <- 2L",
  linters = object_name_linter(styles = c("dotted.case", "lowercase"))
)

lint(
  text = "asdf <- 1L; asdF <- 1L",
  linters = object_name_linter(regexes = c(my_style = "F$", "f$"))
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=default_linters]{default}, \link[=executing_linters]{executing}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stopifnot_all_linter.R
\name{stopifnot_all_linter}
\alias{stopifnot_all_linter}
\title{Block usage of all() within stopifnot()}
\usage{
stopifnot_all_linter()
}
\description{
\code{stopifnot(A)} actually checks \code{all(A)} "under the hood" if \code{A} is a vector,
and produces a better error message than \code{stopifnot(all(A))} does.
}
\examples{
# will produce lints
lint(
  text = "stopifnot(all(x > 0))",
  linters = stopifnot_all_linter()
)

lint(
  text = "stopifnot(y > 3, all(x < 0))",
  linters = stopifnot_all_linter()
)

# okay
lint(
  text = "stopifnot(is.null(x) || all(x > 0))",
  linters = stopifnot_all_linter()
)

lint(
  text = "assert_that(all(x > 0))",
  linters = stopifnot_all_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/length_levels_linter.R
\name{length_levels_linter}
\alias{length_levels_linter}
\title{Require usage of nlevels over length(levels(.))}
\usage{
length_levels_linter()
}
\description{
\code{length(levels(x))} is the same as \code{nlevels(x)}, but harder to read.
}
\examples{
# will produce lints
lint(
  text = "length(levels(x))",
  linters = length_levels_linter()
)

# okay
lint(
  text = "length(c(levels(x), levels(y)))",
  linters = length_levels_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_linter_from_xpath.R
\name{make_linter_from_xpath}
\alias{make_linter_from_xpath}
\alias{make_linter_from_function_xpath}
\title{Create a linter from an XPath}
\usage{
make_linter_from_xpath(
  xpath,
  lint_message,
  type = c("warning", "style", "error"),
  level = c("expression", "file")
)

make_linter_from_function_xpath(
  function_names,
  xpath,
  lint_message,
  type = c("warning", "style", "error"),
  level = c("expression", "file")
)
}
\arguments{
\item{xpath}{Character string, an XPath identifying R code to lint.
For \code{make_linter_from_function_xpath()}, the XPath is relative to the \code{parent::expr} of the
\code{SYMBOL_FUNCTION_CALL} nodes of the selected functions.
See \code{\link[xmlparsedata:xml_parse_data]{xmlparsedata::xml_parse_data()}} and \code{\link[=get_source_expressions]{get_source_expressions()}}.}

\item{lint_message}{The message to be included as the \code{message}
to the \code{Lint} object. If \code{lint_message} is a character vector the same length as \code{xml},
the \code{i}-th lint will be given the \code{i}-th message.}

\item{type}{type of lint.}

\item{level}{Which level of expression is being tested? \code{"expression"}
means an individual expression, while \code{"file"} means all expressions
in the current file are available.}

\item{function_names}{Character vector, names of functions whose calls to examine..}
}
\description{
Create a linter from an XPath
}
\examples{
number_linter <- make_linter_from_xpath("//NUM_CONST", "This is a number.")
lint(text = "1 + 2", linters = number_linter())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nrow_subset_linter.R
\name{nrow_subset_linter}
\alias{nrow_subset_linter}
\title{Block usage of \code{nrow(subset(x, .))}}
\usage{
nrow_subset_linter()
}
\description{
Using \code{nrow(subset(x, condition))} to count the instances where \code{condition}
applies inefficiently requires doing a full subset of \code{x} just to
count the number of rows in the resulting subset.
There are a number of equivalent expressions that don't require the full
subset, e.g. \code{with(x, sum(condition))} (or, more generically,
\code{with(x, sum(condition, na.rm = TRUE))}).
}
\examples{
# will produce lints
lint(
  text = "nrow(subset(x, is_treatment))",
  linters = nrow_subset_linter()
)

lint(
  text = "nrow(filter(x, is_treatment))",
  linters = nrow_subset_linter()
)

lint(
  text = "x \%>\% filter(x, is_treatment) \%>\% nrow()",
  linters = nrow_subset_linter()
)

# okay
lint(
  text = "with(x, sum(is_treatment, na.rm = TRUE))",
  linters = nrow_subset_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=efficiency_linters]{efficiency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{configurable_linters}
\alias{configurable_linters}
\title{Configurable linters}
\description{
Generic linters which support custom configuration to your needs.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'configurable':
\itemize{
\item{\code{\link{absolute_path_linter}}}
\item{\code{\link{assignment_linter}}}
\item{\code{\link{backport_linter}}}
\item{\code{\link{brace_linter}}}
\item{\code{\link{commas_linter}}}
\item{\code{\link{condition_call_linter}}}
\item{\code{\link{conjunct_test_linter}}}
\item{\code{\link{consecutive_mutate_linter}}}
\item{\code{\link{cyclocomp_linter}}}
\item{\code{\link{duplicate_argument_linter}}}
\item{\code{\link{fixed_regex_linter}}}
\item{\code{\link{if_not_else_linter}}}
\item{\code{\link{if_switch_linter}}}
\item{\code{\link{implicit_assignment_linter}}}
\item{\code{\link{implicit_integer_linter}}}
\item{\code{\link{indentation_linter}}}
\item{\code{\link{infix_spaces_linter}}}
\item{\code{\link{library_call_linter}}}
\item{\code{\link{line_length_linter}}}
\item{\code{\link{missing_argument_linter}}}
\item{\code{\link{namespace_linter}}}
\item{\code{\link{nested_pipe_linter}}}
\item{\code{\link{nonportable_path_linter}}}
\item{\code{\link{object_length_linter}}}
\item{\code{\link{object_name_linter}}}
\item{\code{\link{object_overwrite_linter}}}
\item{\code{\link{object_usage_linter}}}
\item{\code{\link{paste_linter}}}
\item{\code{\link{pipe_consistency_linter}}}
\item{\code{\link{quotes_linter}}}
\item{\code{\link{redundant_ifelse_linter}}}
\item{\code{\link{return_linter}}}
\item{\code{\link{scalar_in_linter}}}
\item{\code{\link{semicolon_linter}}}
\item{\code{\link{string_boundary_linter}}}
\item{\code{\link{todo_comment_linter}}}
\item{\code{\link{trailing_whitespace_linter}}}
\item{\code{\link{undesirable_function_linter}}}
\item{\code{\link{undesirable_operator_linter}}}
\item{\code{\link{unnecessary_concatenation_linter}}}
\item{\code{\link{unnecessary_lambda_linter}}}
\item{\code{\link{unnecessary_nesting_linter}}}
\item{\code{\link{unreachable_code_linter}}}
\item{\code{\link{unused_import_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vector_logic_linter.R
\name{vector_logic_linter}
\alias{vector_logic_linter}
\title{Enforce usage of scalar logical operators in conditional statements}
\usage{
vector_logic_linter()
}
\description{
Usage of \code{&} in conditional statements is error-prone and inefficient.
\code{condition} in \code{if (condition) expr} must always be of length 1, in which
case \code{&&} is to be preferred. Ditto for \code{|} vs. \code{||}.
}
\details{
This linter covers inputs to \verb{if()} and \verb{while()} conditions and to
\code{\link[testthat:expect_true]{testthat::expect_true()}} and \code{\link[testthat:expect_false]{testthat::expect_false()}}.

Note that because \code{&} and \code{|} are generics, it is possible that
\code{&&} / \code{||} are not perfect substitutes because \code{&} is doing
method dispatch in an incompatible way.

Moreover, be wary of code that may have side effects, most commonly
assignments. Consider \code{if ((a <- foo(x)) | (b <- bar(y))) { ... }}
vs. \code{if ((a <- foo(x)) || (b <- bar(y))) { ... }}. Because \code{||} exits
early, if \code{a} is \code{TRUE},  the second condition will never be evaluated
and \code{b} will not be assigned. Such usage is not allowed by the Tidyverse
style guide, and the code can easily be refactored by pulling the
assignment outside the condition, so using \code{||} is still preferable.
}
\examples{
# will produce lints
lint(
  text = "if (TRUE & FALSE) 1",
  linters = vector_logic_linter()
)

lint(
  text = "if (TRUE && (TRUE | FALSE)) 4",
  linters = vector_logic_linter()
)

lint(
  text = "filter(x, A && B)",
  linters = vector_logic_linter()
)

# okay
lint(
  text = "if (TRUE && FALSE) 1",
  linters = vector_logic_linter()
)

lint(
  text = "if (TRUE && (TRUE || FALSE)) 4",
  linters = vector_logic_linter()
)

lint(
  text = "filter(x, A & B)",
  linters = vector_logic_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#if-statements}
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=common_mistakes_linters]{common_mistakes}, \link[=default_linters]{default}, \link[=efficiency_linters]{efficiency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boolean_arithmetic_linter.R
\name{boolean_arithmetic_linter}
\alias{boolean_arithmetic_linter}
\title{Require usage of boolean operators over equivalent arithmetic}
\usage{
boolean_arithmetic_linter()
}
\description{
\code{length(which(x == y)) == 0} is the same as \code{!any(x == y)}, but the latter
is more readable and more efficient.
}
\examples{
# will produce lints
lint(
  text = "length(which(x == y)) == 0L",
  linters = boolean_arithmetic_linter()
)

lint(
  text = "sum(grepl(pattern, x)) == 0",
  linters = boolean_arithmetic_linter()
)

# okay
lint(
  text = "!any(x == y)",
  linters = boolean_arithmetic_linter()
)

lint(
  text = "!any(grepl(pattern, x))",
  linters = boolean_arithmetic_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{consistency_linters}
\alias{consistency_linters}
\title{Consistency linters}
\description{
Linters checking enforcing a consistent alternative if there are multiple syntactically valid ways to write
something.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'consistency':
\itemize{
\item{\code{\link{assignment_linter}}}
\item{\code{\link{class_equals_linter}}}
\item{\code{\link{coalesce_linter}}}
\item{\code{\link{comparison_negation_linter}}}
\item{\code{\link{condition_message_linter}}}
\item{\code{\link{consecutive_assertion_linter}}}
\item{\code{\link{consecutive_mutate_linter}}}
\item{\code{\link{function_argument_linter}}}
\item{\code{\link{if_not_else_linter}}}
\item{\code{\link{if_switch_linter}}}
\item{\code{\link{implicit_integer_linter}}}
\item{\code{\link{inner_combine_linter}}}
\item{\code{\link{is_numeric_linter}}}
\item{\code{\link{keyword_quote_linter}}}
\item{\code{\link{length_levels_linter}}}
\item{\code{\link{literal_coercion_linter}}}
\item{\code{\link{nested_pipe_linter}}}
\item{\code{\link{nrow_subset_linter}}}
\item{\code{\link{numeric_leading_zero_linter}}}
\item{\code{\link{nzchar_linter}}}
\item{\code{\link{object_name_linter}}}
\item{\code{\link{paste_linter}}}
\item{\code{\link{print_linter}}}
\item{\code{\link{quotes_linter}}}
\item{\code{\link{redundant_ifelse_linter}}}
\item{\code{\link{rep_len_linter}}}
\item{\code{\link{scalar_in_linter}}}
\item{\code{\link{seq_linter}}}
\item{\code{\link{system_file_linter}}}
\item{\code{\link{T_and_F_symbol_linter}}}
\item{\code{\link{unnecessary_nesting_linter}}}
\item{\code{\link{which_grepl_linter}}}
\item{\code{\link{whitespace_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_source_expressions.R
\name{get_source_expressions}
\alias{get_source_expressions}
\title{Parsed sourced file from a filename}
\usage{
get_source_expressions(filename, lines = NULL)
}
\arguments{
\item{filename}{the file to be parsed.}

\item{lines}{a character vector of lines.
If \code{NULL}, then \code{filename} will be read.}
}
\value{
A \code{list} with three components:
\describe{
\item{expressions}{a \code{list} of
\code{n+1} objects. The first \code{n} elements correspond to each expression in
\code{filename}, and consist of a list of 8 elements:
\itemize{
\item{\code{filename} (\code{character}) the name of the file.}
\item{\code{line} (\code{integer}) the line in the file where this expression begins.}
\item{\code{column} (\code{integer}) the column in the file where this expression begins.}
\item{\code{lines} (named \code{character}) vector of all lines spanned by this
expression, named with the corresponding line numbers.}
\item{\code{parsed_content} (\code{data.frame}) as given by \code{\link[utils:getParseData]{utils::getParseData()}} for this expression.}
\item{\code{xml_parsed_content} (\code{xml_document}) the XML parse tree of this expression as given by
\code{\link[xmlparsedata:xml_parse_data]{xmlparsedata::xml_parse_data()}}.}
\item{\code{content} (\code{character}) the same as \code{lines} as a single string (not split across lines).}
\item{\code{xml_find_function_calls(function_names)} (\code{function}) a function that returns all \code{SYMBOL_FUNCTION_CALL}
XML nodes from \code{xml_parsed_content} with specified function names.}
}

The final element of \code{expressions} is a list corresponding to the full file
consisting of 7 elements:
\itemize{
\item{\code{filename} (\code{character}) the name of this file.}
\item{\code{file_lines} (\code{character}) the \code{\link[=readLines]{readLines()}} output for this file.}
\item{\code{content} (\code{character}) for .R files, the same as \code{file_lines};
for .Rmd or .qmd scripts, this is the extracted R source code (as text).}
\item{\code{full_parsed_content} (\code{data.frame}) as given by
\code{\link[utils:getParseData]{utils::getParseData()}} for the full content.}
\item{\code{full_xml_parsed_content} (\code{xml_document}) the XML parse tree of all
expressions as given by \code{\link[xmlparsedata:xml_parse_data]{xmlparsedata::xml_parse_data()}}.}
\item{\code{terminal_newline} (\code{logical}) records whether \code{filename} has a terminal
newline (as determined by \code{\link[=readLines]{readLines()}} producing a corresponding warning).}
\item{\code{xml_find_function_calls(function_names)} (\code{function}) a function that returns all \code{SYMBOL_FUNCTION_CALL}
XML nodes from \code{full_xml_parsed_content} with specified function names.}
}
}
\item{error}{A \code{Lint} object describing any parsing error.}
\item{warning}{A \code{lints} object describing any parsing warning.}
\item{lines}{The \code{\link[=readLines]{readLines()}} output for this file.}
}
}
\description{
This object is given as input to each linter.
}
\details{
The file is read using the \code{encoding} setting.
This setting is found by taking the first valid result from the following locations
\enumerate{
\item The \code{encoding} key from the usual lintr configuration settings.
\item The \code{Encoding} field from a Package \code{DESCRIPTION} file in a parent directory.
\item The \code{Encoding} field from an R Project \code{.Rproj} file in a parent directory.
\item \code{"UTF-8"} as a fallback.
}
}
\examples{
tmp <- tempfile()
writeLines(c("x <- 1", "y <- x + 1"), tmp)
get_source_expressions(tmp)
unlink(tmp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/if_not_else_linter.R
\name{if_not_else_linter}
\alias{if_not_else_linter}
\title{Block statements like if (!A) x else y}
\usage{
if_not_else_linter(exceptions = c("is.null", "is.na", "missing"))
}
\arguments{
\item{exceptions}{Character vector of calls to exclude from linting.
By default, \code{\link[=is.null]{is.null()}}, \code{\link[=is.na]{is.na()}}, and \code{\link[=missing]{missing()}} are excluded
given the common idiom \code{!is.na(x)} as "x is present".}
}
\description{
\code{if (!A) x else y} is the same as \code{if (A) y else x}, but the latter is
easier to reason about in the \verb{else} case. The former requires
double negation that can be avoided by switching the statement order.
}
\details{
This only applies in the simple \verb{if/else} case. Statements like
\code{if (!A) x else if (B) y else z} don't always have a simpler or
more readable form.

It also applies to \code{\link[=ifelse]{ifelse()}} and the package equivalents
\code{dplyr::if_else()} and \code{data.table::fifelse()}.
}
\examples{
# will produce lints
lint(
  text = "if (!A) x else y",
  linters = if_not_else_linter()
)

lint(
  text = "if (!A) x else if (!B) y else z",
  linters = if_not_else_linter()
)

lint(
  text = "ifelse(!is_treatment, x, y)",
  linters = if_not_else_linter()
)

lint(
  text = "if (!is.null(x)) x else 2",
  linters = if_not_else_linter(exceptions = character())
)

# okay
lint(
  text = "if (A) x else y",
  linters = if_not_else_linter()
)

lint(
  text = "if (!A) x else if (B) z else y",
  linters = if_not_else_linter()
)

lint(
  text = "ifelse(is_treatment, y, x)",
  linters = if_not_else_linter()
)

lint(
  text = "if (!is.null(x)) x else 2",
  linters = if_not_else_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_file_linter.R
\name{download_file_linter}
\alias{download_file_linter}
\title{Recommend usage of a portable \code{mode} value for downloading files}
\usage{
download_file_linter()
}
\description{
\code{mode = "w"} (the default) or \code{mode = "a"} in \code{download.file()} can generate broken files
on Windows. Instead, \code{\link[utils:download.file]{utils::download.file()}} recommends the usage of \code{mode = "wb"}
and \code{mode = "ab"}.
If \code{method = "curl"} or \code{method = "wget"}, no \code{mode} should be provided as it will be ignored.
}
\examples{
# will produce lints
lint(
  text = "download.file(x = my_url)",
  linters = download_file_linter()
)

lint(
  text = "download.file(x = my_url, mode = 'w')",
  linters = download_file_linter()
)

lint(
  text = "download.file(x = my_url, method = 'curl', mode = 'wb')",
  linters = download_file_linter()
)

# okay
lint(
  text = "download.file(x = my_url, mode = 'wb')",
  linters = download_file_linter()
)

lint(
  text = "download.file(x = my_url, method = 'curl')",
  linters = download_file_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=common_mistakes_linters]{common_mistakes}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ids_with_token.R, R/with_id.R
\name{ids_with_token}
\alias{ids_with_token}
\alias{with_id}
\title{Get parsed IDs by token}
\usage{
ids_with_token(source_expression, value, fun = `==`)

with_id(source_expression, id)
}
\arguments{
\item{source_expression}{A list of source expressions, the result of a call to \code{\link[=get_source_expressions]{get_source_expressions()}},
for the desired filename.}

\item{value}{Character. String corresponding to the token to search for.
For example:
\itemize{
\item "SYMBOL"
\item "FUNCTION"
\item "EQ_FORMALS"
\item "$"
\item "("
}}

\item{fun}{For additional flexibility, a function to search for in
the \code{token} column of \code{parsed_content}. Typically \code{==} or \code{\%in\%}.}

\item{id}{Integer. The index corresponding to the desired row
of \code{parsed_content}.}
}
\value{
\code{ids_with_token}: The indices of the \code{parsed_content} data frame
entry of the list of source expressions. Indices correspond to the
\emph{rows} where \code{fun} evaluates to \code{TRUE} for the \code{value} in the \emph{token} column.

\code{with_id}: A data frame corresponding to the row(s) specified in \code{id}.
}
\description{
Gets the source IDs (row indices) corresponding to given token.
}
\section{Functions}{
\itemize{
\item \code{with_id()}: Return the row of the \code{parsed_content} entry of the \verb{[get_source_expressions]()} object. Typically used in
conjunction with \code{ids_with_token} to iterate over rows containing desired tokens.

}}
\examples{
tmp <- tempfile()
writeLines(c("x <- 1", "y <- x + 1"), tmp)
source_exprs <- get_source_expressions(tmp)
ids_with_token(source_exprs$expressions[[1L]], value = "SYMBOL")
with_id(source_exprs$expressions[[1L]], 2L)
unlink(tmp)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/namespace_linter.R
\name{namespace_linter}
\alias{namespace_linter}
\title{Namespace linter}
\usage{
namespace_linter(check_exports = TRUE, check_nonexports = TRUE)
}
\arguments{
\item{check_exports}{Check if \code{symbol} is exported from \code{namespace} in \code{namespace::symbol} calls.}

\item{check_nonexports}{Check if \code{symbol} exists in \code{namespace} in \code{namespace:::symbol} calls.}
}
\description{
Check for missing packages and symbols in namespace calls.
Note that using \code{check_exports=TRUE} or \code{check_nonexports=TRUE} will load packages used in user code so it could
potentially change the global state.
}
\examples{
# will produce lints
lint(
  text = "xyzxyz::sd(c(1, 2, 3))",
  linters = namespace_linter()
)

lint(
  text = "stats::ssd(c(1, 2, 3))",
  linters = namespace_linter()
)

# okay
lint(
  text = "stats::sd(c(1, 2, 3))",
  linters = namespace_linter()
)

lint(
  text = "stats::ssd(c(1, 2, 3))",
  linters = namespace_linter(check_exports = FALSE)
)

lint(
  text = "stats:::ssd(c(1, 2, 3))",
  linters = namespace_linter(check_nonexports = FALSE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=correctness_linters]{correctness}, \link[=executing_linters]{executing}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_identical_linter.R
\name{expect_identical_linter}
\alias{expect_identical_linter}
\title{Require usage of \code{expect_identical(x, y)} where appropriate}
\usage{
expect_identical_linter()
}
\description{
This linter enforces the usage of \code{\link[testthat:expect_identical]{testthat::expect_identical()}} as the
default expectation for comparisons in a testthat suite. \code{expect_true(identical(x, y))}
is an equivalent but unadvised method of the same test. Further,
\code{\link[testthat:expect_equal]{testthat::expect_equal()}} should only be used when \code{expect_identical()}
is inappropriate, i.e., when \code{x} and \code{y} need only be \emph{numerically
equivalent} instead of fully identical (in which case, provide the
\verb{tolerance=} argument to \code{expect_equal()} explicitly). This also applies
when it's inconvenient to check full equality (e.g., names can be ignored,
in which case \code{ignore_attr = "names"} should be supplied to
\code{expect_equal()} (or, for 2nd edition, \code{check.attributes = FALSE}).
}
\section{Exceptions}{


The linter allows \code{expect_equal()} in three circumstances:
\enumerate{
\item A named argument is set (e.g. \code{ignore_attr} or \code{tolerance})
\item Comparison is made to an explicit decimal, e.g.
\code{expect_equal(x, 1.0)} (implicitly setting \code{tolerance})
\item \code{...} is passed (wrapper functions which might set
arguments such as \code{ignore_attr} or \code{tolerance})
}
}

\examples{
# will produce lints
lint(
  text = "expect_equal(x, y)",
  linters = expect_identical_linter()
)

lint(
  text = "expect_true(identical(x, y))",
  linters = expect_identical_linter()
)

# okay
lint(
  text = "expect_identical(x, y)",
  linters = expect_identical_linter()
)

lint(
  text = "expect_equal(x, y, check.attributes = FALSE)",
  linters = expect_identical_linter()
)

lint(
  text = "expect_equal(x, y, tolerance = 1e-6)",
  linters = expect_identical_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=package_development_linters]{package_development}, \link[=pkg_testthat_linters]{pkg_testthat}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infix_spaces_linter.R
\name{infix_spaces_linter}
\alias{infix_spaces_linter}
\title{Infix spaces linter}
\usage{
infix_spaces_linter(exclude_operators = NULL, allow_multiple_spaces = TRUE)
}
\arguments{
\item{exclude_operators}{Character vector of operators to exclude from consideration for linting.
Default is to include the following "low-precedence" operators:
\code{+}, \code{-}, \code{~}, \code{>}, \code{>=}, \code{<}, \code{<=}, \code{==}, \code{!=}, \code{&}, \code{&&}, \code{|}, \code{||}, \verb{<-}, \verb{:=}, \verb{<<-}, \verb{->}, \verb{->>},
\code{=}, \code{/}, \code{*}, and any infix operator (exclude infixes by passing \code{"\%\%"}). Note that \code{"="} here includes
three different operators, from the parser's point of view. To lint only some of these, pass the
corresponding parse tags (i.e., some of \code{"EQ_ASSIGN"}, \code{"EQ_SUB"}, and \code{"EQ_FORMALS"}; see
\code{\link[utils:getParseData]{utils::getParseData()}}).}

\item{allow_multiple_spaces}{Logical, default \code{TRUE}. If \code{FALSE}, usage like \code{x  =  2} will also be linted;
excluded by default because such usage can sometimes be used for better code alignment, as is allowed
by the style guide.}
}
\description{
Check that infix operators are surrounded by spaces. Enforces the corresponding Tidyverse style guide rule;
see \url{https://style.tidyverse.org/syntax.html#infix-operators}.
}
\examples{
# will produce lints
lint(
  text = "x<-1L",
  linters = infix_spaces_linter()
)

lint(
  text = "1:4 \%>\%sum()",
  linters = infix_spaces_linter()
)

# okay
lint(
  text = "x <- 1L",
  linters = infix_spaces_linter()
)

lint(
  text = "1:4 \%>\% sum()",
  linters = infix_spaces_linter()
)

code_lines <- "
ab     <- 1L
abcdef <- 2L
"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = infix_spaces_linter(allow_multiple_spaces = TRUE)
)

lint(
  text = "a||b",
  linters = infix_spaces_linter(exclude_operators = "||")
)

lint(
  text = "sum(1:10, na.rm=TRUE)",
  linters = infix_spaces_linter(exclude_operators = "EQ_SUB")
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#infix-operators}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spaces_inside_linter.R
\name{spaces_inside_linter}
\alias{spaces_inside_linter}
\title{Spaces inside linter}
\usage{
spaces_inside_linter()
}
\description{
Check that parentheses and square brackets do not have spaces directly
inside them, i.e., directly following an opening delimiter or directly
preceding a closing delimiter.
}
\examples{
# will produce lints
lint(
  text = "c( TRUE, FALSE )",
  linters = spaces_inside_linter()
)

lint(
  text = "x[ 1L ]",
  linters = spaces_inside_linter()
)

# okay
lint(
  text = "c(TRUE, FALSE)",
  linters = spaces_inside_linter()
)

lint(
  text = "x[1L]",
  linters = spaces_inside_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#parentheses}
}
}
\section{Tags}{
\link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_lint.R
\name{expect_lint_free}
\alias{expect_lint_free}
\title{Test that the package is lint free}
\usage{
expect_lint_free(...)
}
\arguments{
\item{...}{arguments passed to \code{\link[=lint_package]{lint_package()}}}
}
\description{
This function is a thin wrapper around lint_package that simply tests there are no
lints in the package. It can be used to ensure that your tests fail if the package
contains lints.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/condition_message_linter.R
\name{condition_message_linter}
\alias{condition_message_linter}
\title{Block usage of \code{paste()} and \code{paste0()} with messaging functions using \code{...}}
\usage{
condition_message_linter()
}
\description{
This linter discourages combining condition functions like \code{\link[=stop]{stop()}} with string concatenation
functions \code{\link[base:paste]{base::paste()}} and \code{\link[base:paste0]{base::paste0()}}. This is because
\itemize{
\item \code{stop(paste0(...))} is redundant as it is exactly equivalent to \code{stop(...)}
\item \code{stop(paste(...))} is similarly equivalent to \code{stop(...)} with separators (see examples)
}

The same applies to the other default condition functions as well, i.e., \code{\link[=warning]{warning()}}, \code{\link[=message]{message()}},
and \code{\link[=packageStartupMessage]{packageStartupMessage()}}.
}
\examples{
# will produce lints
lint(
  text = 'stop(paste("a string", "another"))',
  linters = condition_message_linter()
)

lint(
  text = 'warning(paste0("a string", " another"))',
  linters = condition_message_linter()
)

# okay
lint(
  text = 'stop("a string", " another")',
  linters = condition_message_linter()
)

lint(
  text = 'warning("a string", " another")',
  linters = condition_message_linter()
)

lint(
  text = 'warning(paste("a string", "another", sep = "-"))',
  linters = condition_message_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exclude.R
\name{exclude}
\alias{exclude}
\title{Exclude lines or files from linting}
\usage{
exclude(lints, exclusions = settings$exclusions, linter_names = NULL, ...)
}
\arguments{
\item{lints}{that need to be filtered.}

\item{exclusions}{manually specified exclusions}

\item{linter_names}{character vector of names of the active linters, used for parsing inline exclusions.}

\item{...}{additional arguments passed to \code{\link[=parse_exclusions]{parse_exclusions()}}}
}
\description{
Exclude lines or files from linting
}
\details{
Exclusions can be specified in three different ways.
\enumerate{
\item Single line in the source file. default: \verb{# nolint}, possibly followed by a listing of linters to exclude.
If the listing is missing, all linters are excluded on that line. The default listing format is
\verb{# nolint: linter_name, linter2_name.}. There may not be anything between the colon and the line exclusion tag
and the listing must be terminated with a full stop (\code{.}) for the linter list to be respected.
\item Line range in the source file. default: \verb{# nolint start}, \verb{# nolint end}. \verb{# nolint start} accepts linter
lists in the same form as \verb{# nolint}.
\item Exclusions parameter, a list with named and/or unnamed entries.
Outer elements have the following characteristics:
\enumerate{
\item Unnamed elements specify filenames or directories.
\item Named elements are a vector or list of line numbers, with \code{Inf} indicating 'all lines'.
The name gives a path relative to the config.
\enumerate{
\item Unnamed elements denote exclusion of all linters in the given path or directory.
\item Named elements, where the name specifies a linter, denote exclusion for that linter.
For convenience, a vector can be used in place of a list whenever it would not introduce ambiguity, e.g.
a character vector of files to exclude or a vector of lines to exclude.
}
}

Note also that:
\enumerate{
\item All paths are interpreted as globs (\code{\link[=Sys.glob]{Sys.glob()}}), so that e.g. \code{*} does pattern expansion.
\item For \code{exclusions} provided \emph{via} a config file, all paths are taken as
relative \emph{to the config file itself}, not the path where \code{lint()} is run.
}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{deprecated_linters}
\alias{deprecated_linters}
\title{Deprecated linters}
\description{
Linters that are deprecated and provided for backwards compatibility only.
These linters will be excluded from \code{\link[=linters_with_tags]{linters_with_tags()}} by default.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
There are not currently any linters tagged with 'deprecated'.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unnecessary_nesting_linter.R
\name{unnecessary_nesting_linter}
\alias{unnecessary_nesting_linter}
\title{Block instances of unnecessary nesting}
\usage{
unnecessary_nesting_linter(
  allow_assignment = TRUE,
  allow_functions = c("switch", "try", "tryCatch", "withCallingHandlers", "quote",
    "expression", "bquote", "substitute", "with_parameters_test_that", "reactive",
    "observe", "observeEvent", "renderCachedPlot", "renderDataTable", "renderImage",
    "renderPlot", "renderPrint", "renderTable", "renderText", "renderUI"),
  branch_exit_calls = character()
)
}
\arguments{
\item{allow_assignment}{Logical, default \code{TRUE}, in which case
braced expressions consisting only of a single assignment are skipped.
if \code{FALSE}, all braced expressions with only one child expression are linted.
The \code{TRUE} case facilitates interaction with \code{\link[=implicit_assignment_linter]{implicit_assignment_linter()}}
for certain cases where an implicit assignment is necessary, so a braced
assignment is used to further distinguish the assignment. See examples.}

\item{allow_functions}{Character vector of functions which always allow
one-child braced expressions. \code{testthat::test_that()} is always allowed because
testthat requires a braced expression in its \code{code} argument. The other defaults
similarly compute on expressions in a way which is worth highlighting by
em-bracing them, even if there is only one expression, while \code{\link[=switch]{switch()}} is allowed
for its use as a control flow analogous to \code{if}/\verb{else}.]}

\item{branch_exit_calls}{Character vector of functions which are considered
as "exiting" a branch for the purpose of recommending removing nesting in a branch
\emph{lacking} an exit call when the other branch terminates with one. Calls which
always interrupt or quit the current call or R session,
e.g. \code{\link[=stop]{stop()}} and \code{\link[=q]{q()}}, are always included.}
}
\description{
Excessive nesting harms readability. Use helper functions or early returns
to reduce nesting wherever possible.
}
\examples{
# will produce lints
code <- "if (A) {\n  stop('A is bad!')\n} else {\n  do_good()\n}"
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter()
)

code <- "tryCatch(\n  {\n    foo()\n  },\n  error = identity\n)"
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter()
)

code <- "expect_warning(\n  {\n    x <- foo()\n  },\n  'warned'\n)"
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter(allow_assignment = FALSE)
)

code <- "if (x) { \n  if (y) { \n   return(1L) \n  } \n}"
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter()
)

lint(
  text = "my_quote({x})",
  linters = unnecessary_nesting_linter()
)

code <- paste(
  "if (A) {",
  "  stop('A is bad because a.')",
  "} else {",
  "  warning('!A requires caution.')",
  "}",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter()
)

# okay
code <- "if (A) {\n  stop('A is bad because a.')\n} else {\n  stop('!A is bad too.')\n}"
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter()
)

code <- "capture.output({\n  foo()\n})"
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter()
)

code <- "expect_warning(\n  {\n    x <- foo()\n  },\n  'warned'\n)"
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter()
)

code <- "if (x && y) { \n  return(1L) \n}"
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter()
)

code <- "if (x) { \n  y <- x + 1L\n  if (y) { \n   return(1L) \n  } \n}"
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter()
)

lint(
  text = "my_quote({x})",
  linters = unnecessary_nesting_linter(allow_functions = "my_quote")
)

code <- paste(
  "if (A) {",
  "  stop('A is bad because a.')",
  "} else {",
  "  warning('!A requires caution.')",
  "}",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = unnecessary_nesting_linter(branch_exit_calls = c("stop", "warning"))
)

}
\seealso{
\itemize{
\item \code{\link[=cyclocomp_linter]{cyclocomp_linter()}} for another linter that penalizes overly complex code.
\item \link{linters} for a complete list of linters available in lintr.
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exclude.R
\name{normalize_exclusions}
\alias{normalize_exclusions}
\title{Normalize lint exclusions}
\usage{
normalize_exclusions(x, normalize_path = TRUE, root = getwd(), pattern = NULL)
}
\arguments{
\item{x}{Exclusion specification
\itemize{
\item A character vector of filenames or directories relative to \code{root}. Interpreted as globs, see \code{\link[=Sys.glob]{Sys.glob()}}.
\item A named list of integers specifying lines to be excluded per file
\item A named list of named lists specifying linters and lines to be excluded for the linters per file.
}}

\item{normalize_path}{Should the names of the returned exclusion list be normalized paths?
If \code{FALSE}, they will be relative to \code{root}.}

\item{root}{Base directory for relative filename resolution.}

\item{pattern}{If non-NULL, only exclude files in excluded directories if they match
\code{pattern}. Passed to \code{\link[=list.files]{list.files()}} if a directory is excluded.}
}
\value{
A named list of file exclusions.
The names of the list specify the filenames to be excluded.

Each file exclusion is a possibly named list containing line numbers to exclude, or the sentinel \code{Inf} for
completely excluded files. If the an entry is named, the exclusions only take effect for the linter with
the same name.

If \code{normalize_path} is \code{TRUE}, file names will be normalized relative to \code{root}.
Otherwise the paths are left as provided (relative to \code{root} or absolute). That also means
existence is not checked.
}
\description{
Normalize lint exclusions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipe_consistency_linter.R
\name{pipe_consistency_linter}
\alias{pipe_consistency_linter}
\title{Pipe consistency linter}
\usage{
pipe_consistency_linter(pipe = c("|>", "\%>\%", "auto"))
}
\arguments{
\item{pipe}{Which pipe operator is valid (either \code{"\%>\%"} or \code{"|>"}). The default
is the native pipe (\verb{|>}). \code{"auto"} will instead
only enforce consistency, i.e., that in any given file there is only one pipe.}
}
\description{
Check that the recommended pipe operator is used, or more conservatively that
pipes are consistent by file.
}
\examples{
# will produce lints
lint(
  text = "1:3 |> mean() \%>\% as.character()",
  linters = pipe_consistency_linter()
)

lint(
  text = "1:3 \%>\% mean() \%>\% as.character()",
  linters = pipe_consistency_linter("|>")
)

# okay
lint(
  text = "1:3 |> mean() |> as.character()",
  linters = pipe_consistency_linter()
)

lint(
  text = "1:3 \%>\% mean() \%>\% as.character()",
  linters = pipe_consistency_linter("\%>\%")
)
}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/pipes.html#magrittr}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/library_call_linter.R
\name{library_call_linter}
\alias{library_call_linter}
\title{Library call linter}
\usage{
library_call_linter(allow_preamble = TRUE)
}
\arguments{
\item{allow_preamble}{Logical, default \code{TRUE}. If \code{FALSE},
no code is allowed to precede the first \code{library()} call,
otherwise some setup code is allowed, but all \code{library()}
calls must follow consecutively after the first one.}
}
\description{
This linter covers several rules related to \code{\link[=library]{library()}} calls:
\itemize{
\item Enforce such calls to all be at the top of the script.
\item Block usage of argument \code{character.only}, in particular
for loading packages in a loop.
\item Block consecutive calls to \code{suppressMessages(library(.))}
in favor of using \code{\link[=suppressMessages]{suppressMessages()}} only once to suppress
messages from all \code{library()} calls. Ditto \code{\link[=suppressPackageStartupMessages]{suppressPackageStartupMessages()}}.
}
}
\examples{
# will produce lints

code <- "library(dplyr)\nprint('test')\nlibrary(tidyr)"
writeLines(code)
lint(
  text = code,
  linters = library_call_linter()
)

lint(
  text = "library('dplyr', character.only = TRUE)",
  linters = library_call_linter()
)

code <- paste(
  "pkg <- c('dplyr', 'tibble')",
  "sapply(pkg, library, character.only = TRUE)",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = library_call_linter()
)

code <- "suppressMessages(library(dplyr))\nsuppressMessages(library(tidyr))"
writeLines(code)
lint(
  text = code,
  linters = library_call_linter()
)

# okay
code <- "library(dplyr)\nprint('test')"
writeLines(code)
lint(
  text = code,
  linters = library_call_linter()
)

code <- "# comment\nlibrary(dplyr)"
lint(
  text = code,
  linters = library_call_linter()
)

code <- paste(
  "foo <- function(pkg) {",
  "  sapply(pkg, library, character.only = TRUE)",
  "}",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = library_call_linter()
)

code <- "suppressMessages({\n  library(dplyr)\n  library(tidyr)\n})"
writeLines(code)
lint(
  text = code,
  linters = library_call_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outer_negation_linter.R
\name{outer_negation_linter}
\alias{outer_negation_linter}
\title{Require usage of \code{!any(x)} over \code{all(!x)}, \code{!all(x)} over \code{any(!x)}}
\usage{
outer_negation_linter()
}
\description{
\code{any(!x)} is logically equivalent to \code{!all(x)}; ditto for the equivalence of
\code{all(!x)} and \code{!any(x)}. Negating after aggregation only requires inverting
one logical value, and is typically more readable.
}
\examples{
# will produce lints
lint(
  text = "all(!x)",
  linters = outer_negation_linter()
)

lint(
  text = "any(!x)",
  linters = outer_negation_linter()
)

# okay
lint(
  text = "!any(x)",
  linters = outer_negation_linter()
)

lint(
  text = "!all(x)",
  linters = outer_negation_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{robustness_linters}
\alias{robustness_linters}
\title{Robustness linters}
\description{
Linters highlighting code robustness issues, such as possibly wrong edge case behavior.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'robustness':
\itemize{
\item{\code{\link{absolute_path_linter}}}
\item{\code{\link{all_equal_linter}}}
\item{\code{\link{backport_linter}}}
\item{\code{\link{class_equals_linter}}}
\item{\code{\link{download_file_linter}}}
\item{\code{\link{equals_na_linter}}}
\item{\code{\link{for_loop_index_linter}}}
\item{\code{\link{missing_package_linter}}}
\item{\code{\link{namespace_linter}}}
\item{\code{\link{nonportable_path_linter}}}
\item{\code{\link{object_overwrite_linter}}}
\item{\code{\link{routine_registration_linter}}}
\item{\code{\link{sample_int_linter}}}
\item{\code{\link{seq_linter}}}
\item{\code{\link{strings_as_factors_linter}}}
\item{\code{\link{T_and_F_symbol_linter}}}
\item{\code{\link{terminal_close_linter}}}
\item{\code{\link{undesirable_function_linter}}}
\item{\code{\link{undesirable_operator_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_lint_level.R
\name{is_lint_level}
\alias{is_lint_level}
\title{Is this an expression- or a file-level source object?}
\usage{
is_lint_level(source_expression, level = c("expression", "file"))
}
\arguments{
\item{source_expression}{A parsed expression object, i.e., an element
of the object returned by \code{\link[=get_source_expressions]{get_source_expressions()}}.}

\item{level}{Which level of expression is being tested? \code{"expression"}
means an individual expression, while \code{"file"} means all expressions
in the current file are available.}
}
\description{
Helper for determining whether the current \code{source_expression} contains
all expressions in the current file, or just a single expression.
}
\examples{
tmp <- tempfile()
writeLines(c("x <- 1", "y <- x + 1"), tmp)
source_exprs <- get_source_expressions(tmp)
is_lint_level(source_exprs$expressions[[1L]], level = "expression")
is_lint_level(source_exprs$expressions[[1L]], level = "file")
is_lint_level(source_exprs$expressions[[3L]], level = "expression")
is_lint_level(source_exprs$expressions[[3L]], level = "file")
unlink(tmp)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unnecessary_placeholder_linter.R
\name{unnecessary_placeholder_linter}
\alias{unnecessary_placeholder_linter}
\title{Block usage of pipeline placeholders if unnecessary}
\usage{
unnecessary_placeholder_linter()
}
\description{
The argument placeholder \code{.} in magrittr pipelines is unnecessary if
passed as the first positional argument; using it can cause confusion
and impacts readability.
}
\details{
This is true for forward (\verb{\%>\%}), assignment (\verb{\%<>\%}), and tee (\verb{\%T>\%}) operators.
}
\examples{
# will produce lints
lint(
  text = "x \%>\% sum(., na.rm = TRUE)",
  linters = unnecessary_placeholder_linter()
)

# okay
lint(
  text = "x \%>\% sum(na.rm = TRUE)",
  linters = unnecessary_placeholder_linter()
)

lint(
  text = "x \%>\% lm(data = ., y ~ z)",
  linters = unnecessary_placeholder_linter()
)

lint(
  text = "x \%>\% outer(., .)",
  linters = unnecessary_placeholder_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brace_linter.R
\name{brace_linter}
\alias{brace_linter}
\title{Brace linter}
\usage{
brace_linter(
  allow_single_line = FALSE,
  function_bodies = c("multi_line", "always", "not_inline", "never")
)
}
\arguments{
\item{allow_single_line}{If \code{TRUE}, allow an open and closed curly pair on the same line.}

\item{function_bodies}{When to require function bodies to be wrapped in curly braces. One of
\itemize{
\item \code{"always"} to require braces around all function bodies, including inline functions,
\item \code{"not_inline"} to require braces when a function body does not start on the same line as its signature,
\item \code{"multi_line"} (the default) to require braces when a function definition spans multiple lines,
\item \code{"never"} to never require braces in function bodies.
}}
}
\description{
Perform various style checks related to placement and spacing of curly braces:
\itemize{
\item Opening curly braces are never on their own line and are always followed by a newline.
\item Opening curly braces have a space before them.
\item Closing curly braces are on their own line unless they are followed by an \verb{else}.
\item Closing curly braces in \code{if} conditions are on the same line as the corresponding \verb{else}.
\item Either both or neither branch in \code{if}/\verb{else} use curly braces, i.e., either both branches use \code{{...}} or neither
does.
\item Function bodies are wrapped in curly braces.
}
}
\examples{
# will produce lints
lint(
  text = "f <- function() { 1 }",
  linters = brace_linter()
)

writeLines("if (TRUE) {\n return(1) }")
lint(
  text = "if (TRUE) {\n return(1) }",
  linters = brace_linter()
)

# okay
writeLines("f <- function() {\n  1\n}")
lint(
  text = "f <- function() {\n  1\n}",
  linters = brace_linter()
)

writeLines("if (TRUE) { \n return(1) \n}")
lint(
  text = "if (TRUE) { \n return(1) \n}",
  linters = brace_linter()
)

# customizing using arguments
writeLines("if (TRUE) { return(1) }")
lint(
  text = "if (TRUE) { return(1) }",
  linters = brace_linter(allow_single_line = TRUE)
)
}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#indenting}
\item \url{https://style.tidyverse.org/syntax.html#if-statements}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/todo_comment_linter.R
\name{todo_comment_linter}
\alias{todo_comment_linter}
\title{TODO comment linter}
\usage{
todo_comment_linter(todo = c("todo", "fixme"), except_regex = NULL)
}
\arguments{
\item{todo}{Vector of case-insensitive strings that identify TODO comments.}

\item{except_regex}{Vector of case-sensitive regular expressions that identify
\emph{valid} TODO comments.}
}
\description{
Check that the source contains no TODO comments (case-insensitive).
}
\examples{
# will produce lints
lint(
  text = "x + y # TOODOO",
  linters = todo_comment_linter(todo = "toodoo")
)

lint(
  text = "pi <- 1.0 # FIIXMEE",
  linters = todo_comment_linter(todo = "fiixmee")
)

lint(
  text = "x <- TRUE # TOODOO(#1234): Fix this hack.",
  linters = todo_comment_linter()
)

# okay
lint(
  text = "x + y # my informative comment",
  linters = todo_comment_linter()
)

lint(
  text = "pi <- 3.14",
  linters = todo_comment_linter()
)

lint(
  text = "x <- TRUE",
  linters = todo_comment_linter()
)

lint(
  text = "x <- TRUE # TODO(#1234): Fix this hack.",
  linters = todo_comment_linter(except_regex = "TODO\\\\(#[0-9]+\\\\):")
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lint.R
\name{lint-s3}
\alias{lint-s3}
\alias{Lint}
\title{Create a \code{lint} object}
\usage{
Lint(
  filename,
  line_number = 1L,
  column_number = 1L,
  type = c("style", "warning", "error"),
  message = "",
  line = "",
  ranges = NULL
)
}
\arguments{
\item{filename}{path to the source file that was linted.}

\item{line_number}{line number where the lint occurred.}

\item{column_number}{column number where the lint occurred.}

\item{type}{type of lint.}

\item{message}{message used to describe the lint error}

\item{line}{code source where the lint occurred}

\item{ranges}{a list of ranges on the line that should be emphasized.}
}
\value{
an object of class \code{c("lint", "list")}.
}
\description{
Create a \code{lint} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{best_practices_linters}
\alias{best_practices_linters}
\title{Best practices linters}
\description{
Linters checking the use of coding best practices, such as explicit typing of numeric constants.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'best_practices':
\itemize{
\item{\code{\link{absolute_path_linter}}}
\item{\code{\link{any_duplicated_linter}}}
\item{\code{\link{any_is_na_linter}}}
\item{\code{\link{boolean_arithmetic_linter}}}
\item{\code{\link{class_equals_linter}}}
\item{\code{\link{coalesce_linter}}}
\item{\code{\link{commented_code_linter}}}
\item{\code{\link{condition_call_linter}}}
\item{\code{\link{condition_message_linter}}}
\item{\code{\link{conjunct_test_linter}}}
\item{\code{\link{cyclocomp_linter}}}
\item{\code{\link{download_file_linter}}}
\item{\code{\link{empty_assignment_linter}}}
\item{\code{\link{expect_comparison_linter}}}
\item{\code{\link{expect_length_linter}}}
\item{\code{\link{expect_named_linter}}}
\item{\code{\link{expect_not_linter}}}
\item{\code{\link{expect_null_linter}}}
\item{\code{\link{expect_s3_class_linter}}}
\item{\code{\link{expect_s4_class_linter}}}
\item{\code{\link{expect_true_false_linter}}}
\item{\code{\link{expect_type_linter}}}
\item{\code{\link{fixed_regex_linter}}}
\item{\code{\link{for_loop_index_linter}}}
\item{\code{\link{function_argument_linter}}}
\item{\code{\link{function_return_linter}}}
\item{\code{\link{if_switch_linter}}}
\item{\code{\link{ifelse_censor_linter}}}
\item{\code{\link{implicit_assignment_linter}}}
\item{\code{\link{implicit_integer_linter}}}
\item{\code{\link{is_numeric_linter}}}
\item{\code{\link{length_levels_linter}}}
\item{\code{\link{lengths_linter}}}
\item{\code{\link{library_call_linter}}}
\item{\code{\link{list_comparison_linter}}}
\item{\code{\link{literal_coercion_linter}}}
\item{\code{\link{nonportable_path_linter}}}
\item{\code{\link{nrow_subset_linter}}}
\item{\code{\link{nzchar_linter}}}
\item{\code{\link{object_overwrite_linter}}}
\item{\code{\link{outer_negation_linter}}}
\item{\code{\link{paste_linter}}}
\item{\code{\link{pipe_return_linter}}}
\item{\code{\link{print_linter}}}
\item{\code{\link{redundant_equals_linter}}}
\item{\code{\link{redundant_ifelse_linter}}}
\item{\code{\link{regex_subset_linter}}}
\item{\code{\link{rep_len_linter}}}
\item{\code{\link{routine_registration_linter}}}
\item{\code{\link{scalar_in_linter}}}
\item{\code{\link{seq_linter}}}
\item{\code{\link{sort_linter}}}
\item{\code{\link{stopifnot_all_linter}}}
\item{\code{\link{system_file_linter}}}
\item{\code{\link{T_and_F_symbol_linter}}}
\item{\code{\link{terminal_close_linter}}}
\item{\code{\link{undesirable_function_linter}}}
\item{\code{\link{undesirable_operator_linter}}}
\item{\code{\link{unnecessary_lambda_linter}}}
\item{\code{\link{unnecessary_nesting_linter}}}
\item{\code{\link{unnecessary_placeholder_linter}}}
\item{\code{\link{unreachable_code_linter}}}
\item{\code{\link{unused_import_linter}}}
\item{\code{\link{vector_logic_linter}}}
\item{\code{\link{yoda_test_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/whitespace_linter.R
\name{whitespace_linter}
\alias{whitespace_linter}
\title{Whitespace linter}
\usage{
whitespace_linter()
}
\description{
Check that the correct character is used for indentation.
}
\details{
Currently, only supports linting in the presence of tabs.

Much ink has been spilled on this topic, and we encourage you to check
out references for more information.
}
\examples{
# will produce lints
lint(
  text = "\tx",
  linters = whitespace_linter()
)

# okay
lint(
  text = "  x",
  linters = whitespace_linter()
)

}
\references{
\itemize{
\item https://www.jwz.org/doc/tabs-vs-spaces.html
\item https://blog.codinghorror.com/death-to-the-space-infidels/
}
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=consistency_linters]{consistency}, \link[=default_linters]{default}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trailing_blank_lines_linter.R
\name{trailing_blank_lines_linter}
\alias{trailing_blank_lines_linter}
\title{Trailing blank lines linter}
\usage{
trailing_blank_lines_linter()
}
\description{
Check that there are no trailing blank lines in source code.
}
\examples{
# will produce lints
f <- tempfile()
cat("x <- 1\n\n", file = f)
writeLines(readChar(f, file.size(f)))
lint(
  filename = f,
  linters = trailing_blank_lines_linter()
)
unlink(f)

# okay
cat("x <- 1\n", file = f)
writeLines(readChar(f, file.size(f)))
lint(
  filename = f,
  linters = trailing_blank_lines_linter()
)
unlink(f)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=default_linters]{default}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paste_linter.R
\name{paste_linter}
\alias{paste_linter}
\title{Raise lints for several common poor usages of \code{paste()}}
\usage{
paste_linter(
  allow_empty_sep = FALSE,
  allow_to_string = FALSE,
  allow_file_path = c("double_slash", "always", "never")
)
}
\arguments{
\item{allow_empty_sep}{Logical, default \code{FALSE}. If \code{TRUE}, usage of
\code{paste()} with \code{sep = ""} is not linted.}

\item{allow_to_string}{Logical, default \code{FALSE}. If \code{TRUE}, usage of
\code{paste()} and \code{paste0()} with \code{collapse = ", "} is not linted.}

\item{allow_file_path}{String, one of \code{"never"}, \code{"double_slash"}, or \code{"always"}; \code{"double_slash"} by default.
If \code{"always"}, usage of \code{paste()} and \code{paste0()} to construct file paths is not linted. If \code{"double_slash"},
strings containing consecutive forward slashes will not lint. The main use case here is for URLs -- "paths" like
\code{"https://"} will not induce lints, since constructing them with \code{file.path()} might be deemed unnatural.
Lastly, if \code{"never"}, strings with consecutive forward slashes will also lint. Note that \code{"//"} is never linted
when it comes at the beginning or end of the input, to avoid requiring empty inputs like
\code{file.path("", ...)} or \code{file.path(..., "")}.}
}
\description{
The following issues are linted by default by this linter
(see arguments for which can be de-activated optionally):
\enumerate{
\item Block usage of \code{\link[base:paste]{base::paste()}} with \code{sep = ""}. \code{\link[base:paste0]{base::paste0()}} is a faster, more concise alternative;
this is valid unless \code{paste} occurs inside \link[base:expression]{base::expression}, which according to \link[grDevices:plotmath]{grDevices::plotmath}
does not support the \code{sep} argument.
\item Block usage of \code{paste()} or \code{paste0()} with \code{collapse = ", "}. \code{\link[=toString]{toString()}} is a direct
wrapper for this, and alternatives like \code{\link[glue:glue_collapse]{glue::glue_collapse()}} might give better messages for humans.
\item Block usage of \code{paste0()} that supplies \verb{sep=} -- this is not a formal argument to \code{paste0}, and
is likely to be a mistake.
\item Block usage of \code{paste()} / \code{paste0()} combined with \code{\link[=rep]{rep()}} that could be replaced by
\code{\link[base:strrep]{base::strrep()}}. \code{strrep()} can handle the task of building a block of repeated strings
(e.g. often used to build "horizontal lines" for messages). This is both more readable and
skips the (likely small) overhead of putting two strings into the global string cache when only one is needed.

Only target scalar usages -- \code{strrep} can handle more complicated cases (e.g. \code{strrep(letters, 26:1)},
but those aren't as easily translated from a \code{paste(collapse=)} call.
\item Block usage of \code{paste()} or \code{paste0()} to collapse the output of \code{\link[base:deparse]{base::deparse()}}, e.g.
\code{paste(deparse(x), collapse = " ")}. \code{\link[base:deparse1]{base::deparse1()}} is a more readable equivalent,
i.e. \code{deparse1(x)}.
}
}
\examples{
# will produce lints
lint(
  text = 'paste("a", "b", sep = "")',
  linters = paste_linter()
)

lint(
  text = 'paste(c("a", "b"), collapse = ", ")',
  linters = paste_linter()
)

lint(
  text = 'paste0(c("a", "b"), sep = " ")',
  linters = paste_linter()
)

lint(
  text = 'paste0(rep("*", 10L), collapse = "")',
  linters = paste_linter()
)

lint(
  text = 'paste0("http://site.com/", path)',
  linters = paste_linter(allow_file_path = "never")
)

lint(
  text = 'paste0(x, collapse = "")',
  linters = paste_linter()
)

lint(
  text = 'expression(paste("a", "b", sep = ""))',
  linters = paste_linter()
)

lint(
  text = 'paste(deparse(x), collapse = " ")',
  linters = paste_linter()
)

# okay
lint(
  text = 'paste0("a", "b")',
  linters = paste_linter()
)

lint(
  text = 'paste("a", "b", sep = "")',
  linters = paste_linter(allow_empty_sep = TRUE)
)

lint(
  text = 'toString(c("a", "b"))',
  linters = paste_linter()
)

lint(
  text = 'paste(c("a", "b"), collapse = ", ")',
  linters = paste_linter(allow_to_string = TRUE)
)

lint(
  text = 'paste(c("a", "b"))',
  linters = paste_linter()
)

lint(
  text = 'strrep("*", 10L)',
  linters = paste_linter()
)

lint(
  text = 'paste0(year, "/", month, "/", day)',
  linters = paste_linter(allow_file_path = "always")
)

lint(
  text = 'paste0("http://site.com/", path)',
  linters = paste_linter()
)

lint(
  text = 'paste(x, collapse = "")',
  linters = paste_linter()
)

lint(
  text = 'expression(paste("a", "b"))',
  linters = paste_linter()
)

lint(
  text = "deparse1(x)",
  linters = paste_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trailing_whitespace_linter.R
\name{trailing_whitespace_linter}
\alias{trailing_whitespace_linter}
\title{Trailing whitespace linter}
\usage{
trailing_whitespace_linter(allow_empty_lines = FALSE, allow_in_strings = TRUE)
}
\arguments{
\item{allow_empty_lines}{Suppress lints for lines that contain only whitespace.}

\item{allow_in_strings}{Suppress lints for trailing whitespace in string constants.}
}
\description{
Check that there are no space characters at the end of source lines.
}
\examples{
# will produce lints
lint(
  text = "x <- 1.2  ",
  linters = trailing_whitespace_linter()
)

code_lines <- "a <- TRUE\n \nb <- FALSE"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = trailing_whitespace_linter()
)

# okay
lint(
  text = "x <- 1.2",
  linters = trailing_whitespace_linter()
)

lint(
  text = "x <- 1.2  # comment about this assignment",
  linters = trailing_whitespace_linter()
)

code_lines <- "a <- TRUE\n \nb <- FALSE"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = trailing_whitespace_linter(allow_empty_lines = TRUE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{style_linters}
\alias{style_linters}
\title{Style linters}
\description{
Linters highlighting code style issues.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'style':
\itemize{
\item{\code{\link{assignment_linter}}}
\item{\code{\link{brace_linter}}}
\item{\code{\link{commas_linter}}}
\item{\code{\link{commented_code_linter}}}
\item{\code{\link{condition_call_linter}}}
\item{\code{\link{consecutive_assertion_linter}}}
\item{\code{\link{cyclocomp_linter}}}
\item{\code{\link{function_argument_linter}}}
\item{\code{\link{function_left_parentheses_linter}}}
\item{\code{\link{implicit_assignment_linter}}}
\item{\code{\link{implicit_integer_linter}}}
\item{\code{\link{indentation_linter}}}
\item{\code{\link{infix_spaces_linter}}}
\item{\code{\link{keyword_quote_linter}}}
\item{\code{\link{library_call_linter}}}
\item{\code{\link{line_length_linter}}}
\item{\code{\link{numeric_leading_zero_linter}}}
\item{\code{\link{object_length_linter}}}
\item{\code{\link{object_name_linter}}}
\item{\code{\link{object_usage_linter}}}
\item{\code{\link{one_call_pipe_linter}}}
\item{\code{\link{package_hooks_linter}}}
\item{\code{\link{paren_body_linter}}}
\item{\code{\link{pipe_call_linter}}}
\item{\code{\link{pipe_consistency_linter}}}
\item{\code{\link{pipe_continuation_linter}}}
\item{\code{\link{quotes_linter}}}
\item{\code{\link{repeat_linter}}}
\item{\code{\link{return_linter}}}
\item{\code{\link{semicolon_linter}}}
\item{\code{\link{spaces_inside_linter}}}
\item{\code{\link{spaces_left_parentheses_linter}}}
\item{\code{\link{T_and_F_symbol_linter}}}
\item{\code{\link{todo_comment_linter}}}
\item{\code{\link{trailing_blank_lines_linter}}}
\item{\code{\link{trailing_whitespace_linter}}}
\item{\code{\link{undesirable_function_linter}}}
\item{\code{\link{undesirable_operator_linter}}}
\item{\code{\link{unnecessary_concatenation_linter}}}
\item{\code{\link{whitespace_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unreachable_code_linter.R
\name{unreachable_code_linter}
\alias{unreachable_code_linter}
\title{Block unreachable code and comments following return statements}
\usage{
unreachable_code_linter(
  allow_comment_regex = getOption("covr.exclude_end", "# nocov end")
)
}
\arguments{
\item{allow_comment_regex}{Character vector of regular expressions which identify
comments to exclude when finding unreachable terminal comments. By default, this
includes the default "skip region" end marker for \code{{covr}}
(option "covr.exclude_end", or \code{"# nocov end"} if unset).
The end marker for \code{{lintr}} (\code{settings$exclude_end}) is always included.
Note that the regexes should include the initial comment character \verb{#}.}
}
\description{
Code after e.g. a \code{\link[=return]{return()}} or \code{\link[=stop]{stop()}}
or in deterministically false conditional loops like \verb{if (FALSE)} can't be reached;
typically this is vestigial code left after refactoring or sandboxing code, which
is fine for exploration, but shouldn't ultimately be checked in. Comments
meant for posterity should be placed \emph{before} the final \code{return()}.
}
\examples{
# will produce lints
code_lines <- "f <- function() {\n  return(1 + 1)\n  2 + 2\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unreachable_code_linter()
)

code_lines <- "if (FALSE) {\n 2 + 2\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unreachable_code_linter()
)

code_lines <- "while (FALSE) {\n 2 + 2\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unreachable_code_linter()
)

code_lines <- "f <- function() {\n  return(1)\n  # end skip\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unreachable_code_linter()
)

# okay
code_lines <- "f <- function() {\n  return(1 + 1)\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unreachable_code_linter()
)

code_lines <- "if (foo) {\n 2 + 2\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unreachable_code_linter()
)

code_lines <- "while (foo) {\n 2 + 2\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unreachable_code_linter()
)

code_lines <- "f <- function() {\n  return(1)\n  # end skip\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unreachable_code_linter(allow_comment_regex = "# end skip")
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/undesirable_operator_linter.R
\name{undesirable_operator_linter}
\alias{undesirable_operator_linter}
\title{Undesirable operator linter}
\usage{
undesirable_operator_linter(
  op = default_undesirable_operators,
  call_is_undesirable = TRUE
)
}
\arguments{
\item{op}{Character vector of undesirable operators. Input can be any of three types:
\itemize{
\item Unnamed entries must be a character string specifying an undesirable operator.
\item For named entries, the name specifies the undesirable operator.
\itemize{
\item If the entry is a character string, it is used as a description of
why a given operator is undesirable
\item Otherwise, entries should be missing (\code{NA})
A generic message that the named operator is undesirable is used if no
specific description is provided.
Input can also be a list of character strings for convenience.
}
}

Defaults to \link{default_undesirable_operators}. To make small customizations to this list,
use \code{\link[=modify_defaults]{modify_defaults()}}.}

\item{call_is_undesirable}{Logical, default \code{TRUE}. Should lints also be produced
for prefix-style usage of the operators provided in \code{op}?}
}
\description{
Report the use of undesirable operators, e.g. \code{\link[base:ns-dblcolon]{:::}} or
\code{\link[base:assignOps]{<<-}} and suggest an alternative.
}
\examples{
# defaults for which functions are considered undesirable
names(default_undesirable_operators)

# will produce lints
lint(
  text = "a <<- log(10)",
  linters = undesirable_operator_linter()
)

lint(
  text = "mtcars$wt",
  linters = undesirable_operator_linter(op = c("$" = "As an alternative, use the `[[` accessor."))
)

lint(
  text = "`:::`(utils, hasName)",
  linters = undesirable_operator_linter()
)

lint(
  text = "mtcars$wt",
  linters = undesirable_operator_linter("$")
)

# okay
lint(
  text = "a <- log(10)",
  linters = undesirable_operator_linter()
)
lint(
  text = 'mtcars[["wt"]]',
  linters = undesirable_operator_linter(op = c("$" = NA))
)

lint(
  text = 'mtcars[["wt"]]',
  linters = undesirable_operator_linter(op = c("$" = "As an alternative, use the `[[` accessor."))
)

lint(
  text = "`:::`(utils, hasName)",
  linters = undesirable_operator_linter(call_is_undesirable = FALSE)
)

lint(
  text = 'mtcars[["wt"]]',
  linters = undesirable_operator_linter("$")
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=robustness_linters]{robustness}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/numeric_leading_zero_linter.R
\name{numeric_leading_zero_linter}
\alias{numeric_leading_zero_linter}
\title{Require usage of a leading zero in all fractional numerics}
\usage{
numeric_leading_zero_linter()
}
\description{
While .1 and 0.1 mean the same thing, the latter is easier to read due
to the small size of the '.' glyph.
}
\examples{
# will produce lints
lint(
  text = "x <- .1",
  linters = numeric_leading_zero_linter()
)

lint(
  text = "x <- -.1",
  linters = numeric_leading_zero_linter()
)

# okay
lint(
  text = "x <- 0.1",
  linters = numeric_leading_zero_linter()
)

lint(
  text = "x <- -0.1",
  linters = numeric_leading_zero_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xp_utils.R
\name{xp_call_name}
\alias{xp_call_name}
\title{Get the name of the function matched by an XPath}
\usage{
xp_call_name(expr, depth = 1L, condition = NULL)
}
\arguments{
\item{expr}{An \code{xml_node} or \code{xml_nodeset}, e.g. from \code{\link[xml2:xml_find_all]{xml2::xml_find_all()}}.}

\item{depth}{Integer, default \code{1L}. How deep in the AST represented by \code{expr}
should we look to find the call? By default, we assume \code{expr} is matched
to an \verb{<expr>} node under which the corresponding \verb{<SYMBOL_FUNCTION_CALL>}
node is found directly. \code{depth = 0L} means \code{expr} is matched directly
to the \code{SYMBOL_FUNCTION_CALL}; \code{depth > 1L} means \code{depth} total \verb{<expr>}
nodes must be traversed before finding the call.}

\item{condition}{An additional (XPath condition on the \code{SYMBOL_FUNCTION_CALL}
required for a match. The default (\code{NULL}) is no condition. See examples.}
}
\description{
Often, it is more helpful to tailor the \code{message} of a lint to record
which function was matched by the lint logic. This function encapsulates
the logic to pull out the matched call in common situations.
}
\examples{
xml_from_code <- function(str) {
  xml2::read_xml(xmlparsedata::xml_parse_data(parse(text = str, keep.source = TRUE)))
}
xml <- xml_from_code("sum(1:10)")
xp_call_name(xml, depth = 2L)

xp_call_name(xml2::xml_find_first(xml, "expr"))

xml <- xml_from_code(c("sum(1:10)", "sd(1:10)"))
xp_call_name(xml, depth = 2L, condition = "text() = 'sum'")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/equals_na_linter.R
\name{equals_na_linter}
\alias{equals_na_linter}
\title{Equality check with NA linter}
\usage{
equals_na_linter()
}
\description{
Check for \code{x == NA}, \code{x != NA} and \code{x \%in\% NA}. Such usage is almost surely incorrect --
checks for missing values should be done with \code{\link[=is.na]{is.na()}}.
}
\examples{
# will produce lints
lint(
  text = "x == NA",
  linters = equals_na_linter()
)

lint(
  text = "x != NA",
  linters = equals_na_linter()
)

lint(
  text = "x \%in\% NA",
  linters = equals_na_linter()
)

# okay
lint(
  text = "is.na(x)",
  linters = equals_na_linter()
)

lint(
  text = "!is.na(x)",
  linters = equals_na_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=common_mistakes_linters]{common_mistakes}, \link[=correctness_linters]{correctness}, \link[=default_linters]{default}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/strings_as_factors_linter.R
\name{strings_as_factors_linter}
\alias{strings_as_factors_linter}
\title{Identify cases where \code{stringsAsFactors} should be supplied explicitly}
\usage{
strings_as_factors_linter()
}
\description{
Designed for code bases written for versions of R before 4.0 seeking to upgrade to R >= 4.0, where
one of the biggest pain points will surely be the flipping of the
default value of \code{stringsAsFactors} from \code{TRUE} to \code{FALSE}.
}
\details{
It's not always possible to tell statically whether the change will break
existing code because R is dynamically typed -- e.g. in \code{data.frame(x)}
if \code{x} is a string, this code will be affected, but if \code{x} is a number,
this code will be unaffected. However, in \code{data.frame(x = "a")}, the
output will unambiguously be affected. We can instead supply
\code{stringsAsFactors = TRUE}, which will make this code backwards-compatible.

See \url{https://developer.r-project.org/Blog/public/2020/02/16/stringsasfactors/}.
}
\examples{
# will produce lints
lint(
  text = 'data.frame(x = "a")',
  linters = strings_as_factors_linter()
)

# okay
lint(
  text = 'data.frame(x = "a", stringsAsFactors = TRUE)',
  linters = strings_as_factors_linter()
)

lint(
  text = 'data.frame(x = "a", stringsAsFactors = FALSE)',
  linters = strings_as_factors_linter()
)

lint(
  text = "data.frame(x = 1.2)",
  linters = strings_as_factors_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/any_is_na_linter.R
\name{any_is_na_linter}
\alias{any_is_na_linter}
\title{Require usage of \code{anyNA(x)} over \code{any(is.na(x))}}
\usage{
any_is_na_linter()
}
\description{
\code{\link[base:anyNA]{base::anyNA()}} exists as a replacement for \code{any(is.na(x))} which is more efficient
for simple objects, and is at worst equally efficient.
Therefore, it should be used in all situations instead of the latter.
}
\details{
This also lints \code{NA \%in\% x}, which is also more readably expressed as \code{anyNA(x)}. In
this case, note a subtle difference of behavior, namely that \code{is.na(NaN)} is \code{TRUE}
while \code{NA \%in\% NaN} is \code{FALSE}.
}
\examples{
# will produce lints
lint(
  text = "any(is.na(x), na.rm = TRUE)",
  linters = any_is_na_linter()
)

lint(
  text = "any(is.na(foo(x)))",
  linters = any_is_na_linter()
)

# okay
lint(
  text = "anyNA(x)",
  linters = any_is_na_linter()
)

lint(
  text = "anyNA(foo(x))",
  linters = any_is_na_linter()
)

lint(
  text = "any(!is.na(x), na.rm = TRUE)",
  linters = any_is_na_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=efficiency_linters]{efficiency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/literal_coercion_linter.R
\name{literal_coercion_linter}
\alias{literal_coercion_linter}
\title{Require usage of correctly-typed literals over literal coercions}
\usage{
literal_coercion_linter()
}
\description{
\code{as.integer(1)} (or \code{rlang::int(1)}) is the same as \code{1L} but the latter is
more concise and gets typed correctly at compilation.
}
\details{
The same applies to missing sentinels like \code{NA} -- typically, it is not
necessary to specify the storage type of \code{NA}, but when it is, prefer
using the typed version (e.g. \code{NA_real_}) instead of a coercion
(like \code{as.numeric(NA)}).
}
\examples{
# will produce lints
lint(
  text = "int(1)",
  linters = literal_coercion_linter()
)

lint(
  text = "as.character(NA)",
  linters = literal_coercion_linter()
)

lint(
  text = "rlang::lgl(1L)",
  linters = literal_coercion_linter()
)

# okay
lint(
  text = "1L",
  linters = literal_coercion_linter()
)

lint(
  text = "NA_character_",
  linters = literal_coercion_linter()
)

lint(
  text = "TRUE",
  linters = literal_coercion_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=efficiency_linters]{efficiency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_not_linter.R
\name{expect_not_linter}
\alias{expect_not_linter}
\title{Require usage of \code{expect_false(x)} over \code{expect_true(!x)}}
\usage{
expect_not_linter()
}
\description{
\code{\link[testthat:expect_false]{testthat::expect_false()}} exists specifically for testing that an output is
\code{FALSE}. \code{\link[testthat:expect_true]{testthat::expect_true()}} can also be used for such tests by
negating the output, but it is better to use the tailored function instead.
The reverse is also true -- use \code{expect_false(A)} instead of
\code{expect_true(!A)}.
}
\examples{
# will produce lints
lint(
  text = "expect_true(!x)",
  linters = expect_not_linter()
)

# okay
lint(
  text = "expect_false(x)",
  linters = expect_not_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=package_development_linters]{package_development}, \link[=pkg_testthat_linters]{pkg_testthat}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/absolute_path_linter.R
\name{absolute_path_linter}
\alias{absolute_path_linter}
\title{Absolute path linter}
\usage{
absolute_path_linter(lax = TRUE)
}
\arguments{
\item{lax}{Less stringent linting, leading to fewer false positives.
If \code{TRUE}, only lint path strings, which
\itemize{
\item contain at least two path elements, with one having at least two characters and
\item contain only alphanumeric chars (including UTF-8), spaces, and win32-allowed punctuation
}}
}
\description{
Check that no absolute paths are used (e.g. "/var", "C:\\System", "~/docs").
}
\examples{
# will produce lints
lint(
  text = 'R"(/blah/file.txt)"',
  linters = absolute_path_linter()
)

# okay
lint(
  text = 'R"(./blah)"',
  linters = absolute_path_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \code{\link[=nonportable_path_linter]{nonportable_path_linter()}}
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{all_undesirable_functions}
\alias{all_undesirable_functions}
\alias{default_undesirable_functions}
\alias{all_undesirable_operators}
\alias{default_undesirable_operators}
\title{Default undesirable functions and operators}
\format{
A named list of character strings.
}
\usage{
all_undesirable_functions

default_undesirable_functions

all_undesirable_operators

default_undesirable_operators
}
\description{
Lists of function names and operators for \code{\link[=undesirable_function_linter]{undesirable_function_linter()}} and \code{\link[=undesirable_operator_linter]{undesirable_operator_linter()}}.
There is a list for the default elements and another that contains all available elements.
Use \code{\link[=modify_defaults]{modify_defaults()}} to produce a custom list.
}
\details{
The following functions are sometimes regarded as undesirable:
\itemize{
\item \code{\link[=.libPaths]{.libPaths()}} As an alternative, use \code{\link[withr:with_libpaths]{withr::with_libpaths()}} for a temporary change instead of permanently modifying the library location.
\item \code{\link[=attach]{attach()}} As an alternative, use roxygen2's @importFrom statement in packages, or \code{::} in scripts. \code{\link[=attach]{attach()}} modifies the global search path.
\item \code{\link[=browser]{browser()}} As an alternative, remove this likely leftover from debugging. It pauses execution when run.
\item \code{\link[=debug]{debug()}} As an alternative, remove this likely leftover from debugging. It traps a function and causes execution to pause when that function is run.
\item \code{\link[=debugcall]{debugcall()}} As an alternative, remove this likely leftover from debugging. It traps a function and causes execution to pause when that function is run.
\item \code{\link[=debugonce]{debugonce()}} As an alternative, remove this likely leftover from debugging. It traps a function and causes execution to pause when that function is run.
\item \code{\link[=detach]{detach()}} As an alternative, avoid modifying the global search path. Detaching environments from the search path is rarely necessary in production code.
\item \code{\link[=library]{library()}} As an alternative, use roxygen2's @importFrom statement in packages and \code{::} in scripts, instead of modifying the global search path.
\item \code{\link[=mapply]{mapply()}} As an alternative, use \code{\link[=Map]{Map()}} to guarantee a list is returned and simplify accordingly.
\item \code{\link[=options]{options()}} As an alternative, use \code{\link[withr:with_options]{withr::with_options()}} for a temporary change instead of permanently modifying the session options.
\item \code{\link[=par]{par()}} As an alternative, use \code{\link[withr:with_par]{withr::with_par()}} for a temporary change instead of permanently modifying the graphics device parameters.
\item \code{\link[=require]{require()}} As an alternative, use roxygen2's @importFrom statement in packages and \code{\link[=library]{library()}} or \code{::} in scripts, instead of modifying the global search path.
\item \code{\link[=sapply]{sapply()}} As an alternative, use \code{\link[=vapply]{vapply()}} with an appropriate \code{FUN.VALUE=} argument to obtain type-stable simplification.
\item \code{\link[=setwd]{setwd()}} As an alternative, use \code{\link[withr:with_dir]{withr::with_dir()}} for a temporary change instead of modifying the global working directory.
\item \code{\link[=sink]{sink()}} As an alternative, use \code{\link[withr:with_sink]{withr::with_sink()}} for a temporary redirection instead of permanently redirecting output.
\item \code{\link[=source]{source()}} As an alternative, manage dependencies through packages. \code{\link[=source]{source()}} loads code into the global environment unless \code{local = TRUE} is used, which can cause hard-to-predict behavior.
\item \code{\link[=structure]{structure()}} As an alternative, Use \code{class<-}, \code{names<-}, and \code{attr<-} to set attributes.
\item \code{\link[=Sys.setenv]{Sys.setenv()}} As an alternative, use \code{\link[withr:with_envvar]{withr::with_envvar()}} for a temporary change instead of permanently modifying global environment variables.
\item \code{\link[=Sys.setlocale]{Sys.setlocale()}} As an alternative, use \code{\link[withr:with_locale]{withr::with_locale()}} for a temporary change instead of permanently modifying the session locale.
\item \code{\link[=trace]{trace()}} As an alternative, remove this likely leftover from debugging. It traps a function and causes execution of arbitrary code when that function is run.
\item \code{\link[=undebug]{undebug()}} As an alternative, remove this likely leftover from debugging. It is only useful for interactive debugging with \code{\link[=debug]{debug()}}.
\item \code{\link[=untrace]{untrace()}} As an alternative, remove this likely leftover from debugging. It is only useful for interactive debugging with \code{\link[=trace]{trace()}}.
}

The following operators are sometimes regarded as undesirable:
\itemize{
\item \code{\link[base:assignOps]{->>}}. It assigns outside the current environment in a way that can be hard to reason about. Prefer fully-encapsulated functions wherever possible, or, if necessary, assign to a specific environment with \code{\link[=assign]{assign()}}. Recall that you can create an environment at the desired scope with \code{\link[=new.env]{new.env()}}.
\item \code{\link[base:ns-dblcolon]{:::}}. It accesses non-exported functions inside packages. Code relying on these is likely to break in future versions of the package because the functions are not part of the public interface and may be changed or removed by the maintainers without notice. Use public functions via \code{::} instead.
\item \code{\link[base:assignOps]{<<-}}. It assigns outside the current environment in a way that can be hard to reason about. Prefer fully-encapsulated functions wherever possible, or, if necessary, assign to a specific environment with \code{\link[=assign]{assign()}}. Recall that you can create an environment at the desired scope with \code{\link[=new.env]{new.env()}}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rep_len_linter.R
\name{rep_len_linter}
\alias{rep_len_linter}
\title{Require usage of rep_len(x, n) over rep(x, length.out = n)}
\usage{
rep_len_linter()
}
\description{
\code{rep(x, length.out = n)} calls \code{rep_len(x, n)} "under the hood". The latter
is thus more direct and equally readable.
}
\examples{
# will produce lints
lint(
  text = "rep(1:3, length.out = 10)",
  linters = rep_len_linter()
)

# okay
lint(
  text = "rep_len(1:3, 10)",
  linters = rep_len_linter()
)

lint(
  text = "rep(1:3, each = 2L, length.out = 10L)",
  linters = rep_len_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/object_length_linter.R
\name{object_length_linter}
\alias{object_length_linter}
\title{Object length linter}
\usage{
object_length_linter(length = 30L)
}
\arguments{
\item{length}{maximum variable name length allowed.}
}
\description{
Check that object names are not too long.
The length of an object name is defined as the length in characters, after removing extraneous parts:
\itemize{
\item generic prefixes for implementations of S3 generics, e.g. \code{as.data.frame.my_class} has length 8.
\item leading \code{.}, e.g. \code{.my_hidden_function} has length 18.
\item "\%\%" for infix operators, e.g. \verb{\%my_op\%} has length 5.
\item trailing \verb{<-} for assignment functions, e.g. \verb{my_attr<-} has length 7.
}
}
\details{
Note that this behavior relies in part on having packages in your Imports available;
see the detailed note in \code{\link[=object_name_linter]{object_name_linter()}} for more details.
}
\examples{
# will produce lints
lint(
  text = "very_very_long_variable_name <- 1L",
  linters = object_length_linter(length = 10L)
)

# okay
lint(
  text = "very_very_long_variable_name <- 1L",
  linters = object_length_linter(length = 30L)
)

lint(
  text = "var <- 1L",
  linters = object_length_linter(length = 10L)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=executing_linters]{executing}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_null_linter.R
\name{expect_null_linter}
\alias{expect_null_linter}
\title{Require usage of \code{expect_null} for checking \code{NULL}}
\usage{
expect_null_linter()
}
\description{
Require usage of \code{expect_null(x)} over \code{expect_equal(x, NULL)} and similar
usages.
}
\details{
\code{\link[testthat:expect_null]{testthat::expect_null()}} exists specifically for testing for \code{NULL} objects.
\code{\link[testthat:expect_equal]{testthat::expect_equal()}}, \code{\link[testthat:expect_identical]{testthat::expect_identical()}}, and
\code{\link[testthat:expect_true]{testthat::expect_true()}} can also be used for such tests,
but it is better to use the tailored function instead.
}
\examples{
# will produce lints
lint(
  text = "expect_equal(x, NULL)",
  linters = expect_null_linter()
)

lint(
  text = "expect_identical(x, NULL)",
  linters = expect_null_linter()
)

lint(
  text = "expect_true(is.null(x))",
  linters = expect_null_linter()
)


# okay
lint(
  text = "expect_null(x)",
  linters = expect_null_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=package_development_linters]{package_development}, \link[=pkg_testthat_linters]{pkg_testthat}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_s3_class_linter.R
\name{expect_s3_class_linter}
\alias{expect_s3_class_linter}
\title{Require usage of \code{expect_s3_class()}}
\usage{
expect_s3_class_linter()
}
\description{
\code{\link[testthat:expect_s3_class]{testthat::expect_s3_class()}} exists specifically for testing the class
of S3 objects. \code{\link[testthat:expect_equal]{testthat::expect_equal()}}, \code{\link[testthat:expect_identical]{testthat::expect_identical()}},
and \code{\link[testthat:expect_true]{testthat::expect_true()}} can also be used for such tests,
but it is better to use the tailored function instead.
}
\examples{
# will produce lints
lint(
  text = 'expect_equal(class(x), "data.frame")',
  linters = expect_s3_class_linter()
)

lint(
  text = 'expect_equal(class(x), "numeric")',
  linters = expect_s3_class_linter()
)

# okay
lint(
  text = 'expect_s3_class(x, "data.frame")',
  linters = expect_s3_class_linter()
)

lint(
  text = 'expect_type(x, "double")',
  linters = expect_s3_class_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \code{\link[=expect_s4_class_linter]{expect_s4_class_linter()}}
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=package_development_linters]{package_development}, \link[=pkg_testthat_linters]{pkg_testthat}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matrix_apply_linter.R
\name{matrix_apply_linter}
\alias{matrix_apply_linter}
\title{Require usage of \code{colSums(x)} or \code{rowSums(x)} over \code{apply(x, ., sum)}}
\usage{
matrix_apply_linter()
}
\description{
\code{\link[=colSums]{colSums()}} and \code{\link[=rowSums]{rowSums()}} are clearer and more performant alternatives to
\code{apply(x, 2, sum)} and \code{apply(x, 1, sum)} respectively in the case of 2D
arrays, or matrices
}
\examples{
# will produce lints
lint(
  text = "apply(x, 1, sum)",
  linters = matrix_apply_linter()
)

lint(
  text = "apply(x, 2, sum)",
  linters = matrix_apply_linter()
)

lint(
  text = "apply(x, 2, sum, na.rm = TRUE)",
  linters = matrix_apply_linter()
)

lint(
  text = "apply(x, 2:4, sum)",
  linters = matrix_apply_linter()
)

# okay
lint(
  text = "rowSums(x)",
  linters = matrix_apply_linter()
)

lint(
  text = "colSums(x)",
  linters = matrix_apply_linter()
)

lint(
  text = "colSums(x, na.rm = TRUE)",
  linters = matrix_apply_linter()
)

lint(
  text = "rowSums(colSums(x), dims = 3)",
  linters = matrix_apply_linter()
)


}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list2df_linter.R
\name{list2df_linter}
\alias{list2df_linter}
\title{Recommend direct usage of \code{data.frame()} to create a data.frame from a list}
\usage{
list2df_linter()
}
\description{
\code{\link[base:list2DF]{base::list2DF()}} is the preferred way to turn a list of columns into a data.frame.
Note that it doesn't support recycling; if that's required, use \code{\link[=data.frame]{data.frame()}}.
}
\examples{
# will produce lints
lint(
  text = "do.call(cbind.data.frame, x)",
  linters = list2df_linter()
)

lint(
  text = "do.call('cbind.data.frame', x)",
  linters = list2df_linter()
)

lint(
  text = "do.call(cbind.data.frame, list(a = 1, b = 1:10))",
  linters = list2df_linter()
)

# okay
lint(
  text = "list2df(x)",
  linters = list2df_linter()
)

lint(
  text = "data.frame(list(a = 1, b = 1:10))",
  linters = list2df_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repeat_linter.R
\name{repeat_linter}
\alias{repeat_linter}
\title{Repeat linter}
\usage{
repeat_linter()
}
\description{
Check that \verb{while (TRUE)} is not used for infinite loops. While this is valid
R code, using \code{repeat {}} is more explicit.
}
\examples{
# will produce lints
lint(
  text = "while (TRUE) { }",
  linters = repeat_linter()
)


# okay
lint(
  text = "repeat { }",
  linters = repeat_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semicolon_linter.R
\name{semicolon_linter}
\alias{semicolon_linter}
\title{Semicolon linter}
\usage{
semicolon_linter(allow_compound = FALSE, allow_trailing = FALSE)
}
\arguments{
\item{allow_compound}{Logical, default \code{FALSE}. If \code{TRUE}, "compound"
semicolons (e.g. as in \verb{x; y}, i.e., on the same line of code) are allowed.}

\item{allow_trailing}{Logical, default \code{FALSE}. If \code{TRUE}, "trailing"
semicolons (i.e., those that terminate lines of code) are allowed.}
}
\description{
Check that no semicolons terminate expressions.
}
\examples{
# will produce lints
lint(
  text = "a <- 1;",
  linters = semicolon_linter()
)

lint(
  text = "a <- 1; b <- 1",
  linters = semicolon_linter()
)

lint(
  text = "function() { a <- 1; b <- 1 }",
  linters = semicolon_linter()
)

# okay
lint(
  text = "a <- 1",
  linters = semicolon_linter()
)

lint(
  text = "a <- 1;",
  linters = semicolon_linter(allow_trailing = TRUE)
)

code_lines <- "a <- 1\nb <- 1"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = semicolon_linter()
)

lint(
  text = "a <- 1; b <- 1",
  linters = semicolon_linter(allow_compound = TRUE)
)

code_lines <- "function() { \n  a <- 1\n  b <- 1\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = semicolon_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#semicolons}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/terminal_close_linter.R
\name{terminal_close_linter}
\alias{terminal_close_linter}
\title{Prohibit close() from terminating a function definition}
\usage{
terminal_close_linter()
}
\description{
Functions that end in \code{close(x)} are almost always better written by using
\code{on.exit(close(x))} close to where \code{x} is defined and/or opened.
}
\examples{
# will produce lints
code <- paste(
  "f <- function(fl) {",
  "  conn <- file(fl, open = 'r')",
  "  readLines(conn)",
  "  close(conn)",
  "}",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = terminal_close_linter()
)

# okay
code <- paste(
  "f <- function(fl) {",
  "  conn <- file(fl, open = 'r')",
  "  on.exit(close(conn))",
  "  readLines(conn)",
  "}",
  sep = "\n"
)
writeLines(code)
lint(
  text = code,
  linters = terminal_close_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/object_usage_linter.R
\name{object_usage_linter}
\alias{object_usage_linter}
\title{Object usage linter}
\usage{
object_usage_linter(
  interpret_glue = NULL,
  interpret_extensions = c("glue", "rlang"),
  skip_with = TRUE
)
}
\arguments{
\item{interpret_glue}{(Defunct)}

\item{interpret_extensions}{Character vector of extensions to interpret. These are meant to cover known cases where
variables may be used in ways understood by the reader but not by \code{checkUsage()} to avoid false positives.
Currently \code{"glue"} and \code{"rlang"} are supported, both of which are in the default.
\itemize{
\item For \code{glue}, examine \code{\link[glue:glue]{glue::glue()}} calls.
\item For \code{rlang}, examine \code{.env$key} usages.
}}

\item{skip_with}{A logical. If \code{TRUE} (default), code in \code{with()} expressions
will be skipped. This argument will be passed to \code{skipWith} argument of
\code{codetools::checkUsage()}.}
}
\description{
Check that closures have the proper usage using \code{\link[codetools:checkUsage]{codetools::checkUsage()}}.
Note that this runs \code{\link[base:eval]{base::eval()}} on the code, so \strong{do not use with untrusted code}.
}
\examples{
# will produce lints
lint(
  text = "foo <- function() { x <- 1 }",
  linters = object_usage_linter()
)

# okay
lint(
  text = "foo <- function(x) { x <- 1 }",
  linters = object_usage_linter()
)

lint(
  text = "foo <- function() { x <- 1; return(x) }",
  linters = object_usage_linter()
)
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'package_development':
\itemize{
\item{\code{\link{backport_linter}}}
\item{\code{\link{conjunct_test_linter}}}
\item{\code{\link{expect_comparison_linter}}}
\item{\code{\link{expect_identical_linter}}}
\item{\code{\link{expect_length_linter}}}
\item{\code{\link{expect_named_linter}}}
\item{\code{\link{expect_not_linter}}}
\item{\code{\link{expect_null_linter}}}
\item{\code{\link{expect_s3_class_linter}}}
\item{\code{\link{expect_s4_class_linter}}}
\item{\code{\link{expect_true_false_linter}}}
\item{\code{\link{expect_type_linter}}}
\item{\code{\link{package_hooks_linter}}}
\item{\code{\link{yoda_test_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{Linter}
\alias{Linter}
\title{Create a \code{linter} closure}
\usage{
Linter(
  fun,
  name = linter_auto_name(),
  linter_level = c(NA_character_, "file", "expression")
)
}
\arguments{
\item{fun}{A function that takes a source file and returns \code{lint} objects.}

\item{name}{Default name of the Linter.
Lints produced by the linter will be labelled with \code{name} by default.}

\item{linter_level}{Which level of expression is the linter working with?
\code{"expression"} means an individual expression in \code{xml_parsed_content}, while \code{"file"} means all expressions
in the current file are available in \code{full_xml_parsed_content}.
\code{NA} means the linter will be run with both, expression-level and file-level source expressions.}
}
\value{
The same function with its class set to 'linter'.
}
\description{
Create a \code{linter} closure
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unnecessary_lambda_linter.R
\name{unnecessary_lambda_linter}
\alias{unnecessary_lambda_linter}
\title{Block usage of anonymous functions in iteration functions when unnecessary}
\usage{
unnecessary_lambda_linter(allow_comparison = FALSE)
}
\arguments{
\item{allow_comparison}{Logical, default \code{FALSE}. If \code{TRUE}, lambdas like
\code{function(x) foo(x) == 2}, where \code{foo} can be extracted to the "mapping"
function and \code{==} vectorized instead of called repeatedly, are linted.}
}
\description{
Using an anonymous function in, e.g., \code{\link[=lapply]{lapply()}} is not always necessary,
e.g. \code{lapply(DF, sum)} is the same as \code{lapply(DF, function(x) sum(x))} and
the former is more readable.
}
\details{
Cases like \verb{lapply(x, \\(xi) grep("ptn", xi))} are excluded because, though
the anonymous function \emph{can} be avoided, doing so is not always more
readable.
}
\examples{
# will produce lints
lint(
  text = "lapply(list(1:3, 2:4), function(xi) sum(xi))",
  linters = unnecessary_lambda_linter()
)

lint(
  text = "sapply(x, function(xi) xi == 2)",
  linters = unnecessary_lambda_linter()
)

lint(
  text = "sapply(x, function(xi) sum(xi) > 0)",
  linters = unnecessary_lambda_linter()
)

# okay
lint(
  text = "lapply(list(1:3, 2:4), sum)",
  linters = unnecessary_lambda_linter()
)

lint(
  text = 'lapply(x, function(xi) grep("ptn", xi))',
  linters = unnecessary_lambda_linter()
)

lint(
  text = "lapply(x, function(xi) data.frame(col = xi))",
  linters = unnecessary_lambda_linter()
)

lint(
  text = "sapply(x, function(xi) xi == 2)",
  linters = unnecessary_lambda_linter(allow_comparison = TRUE)
)

lint(
  text = "sapply(x, function(xi) sum(xi) > 0)",
  linters = unnecessary_lambda_linter(allow_comparison = TRUE)
)

lint(
  text = "sapply(x, function(xi) sum(abs(xi)) > 10)",
  linters = unnecessary_lambda_linter()
)

lint(
  text = "sapply(x, sum) > 0",
  linters = unnecessary_lambda_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commented_code_linter.R
\name{commented_code_linter}
\alias{commented_code_linter}
\title{Commented code linter}
\usage{
commented_code_linter()
}
\description{
Check that there is no commented code outside roxygen blocks.
}
\examples{
# will produce lints
lint(
  text = "# x <- 1",
  linters = commented_code_linter()
)

lint(
  text = "x <- f() # g()",
  linters = commented_code_linter()
)

lint(
  text = "x + y # + z[1, 2]",
  linters = commented_code_linter()
)

# okay
lint(
  text = "x <- 1; x <- f(); x + y",
  linters = commented_code_linter()
)

lint(
  text = "#' x <- 1",
  linters = commented_code_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redundant_ifelse_linter.R
\name{redundant_ifelse_linter}
\alias{redundant_ifelse_linter}
\title{Prevent \code{ifelse()} from being used to produce \code{TRUE}/\code{FALSE} or \code{1}/\code{0}}
\usage{
redundant_ifelse_linter(allow10 = FALSE)
}
\arguments{
\item{allow10}{Logical, default \code{FALSE}. If \code{TRUE}, usage like
\code{ifelse(x, 1, 0)} is allowed, i.e., only usage like
\code{ifelse(x, TRUE, FALSE)} is linted.}
}
\description{
Expressions like \code{ifelse(x, TRUE, FALSE)} and \code{ifelse(x, FALSE, TRUE)} are
redundant; just \code{x} or \code{!x} suffice in R code where logical vectors are a
core data structure. \code{ifelse(x, 1, 0)} is also \code{as.numeric(x)}, but even
this should be needed only rarely.
}
\examples{
# will produce lints
lint(
  text = "ifelse(x >= 2.5, TRUE, FALSE)",
  linters = redundant_ifelse_linter()
)

lint(
  text = "ifelse(x < 2.5, 1L, 0L)",
  linters = redundant_ifelse_linter()
)

# okay
lint(
  text = "x >= 2.5",
  linters = redundant_ifelse_linter()
)

# Note that this is just to show the strict equivalent of the example above;
# converting to integer is often unnecessary and the logical vector itself
# should suffice.
lint(
  text = "as.integer(x < 2.5)",
  linters = redundant_ifelse_linter()
)

lint(
  text = "ifelse(x < 2.5, 1L, 0L)",
  linters = redundant_ifelse_linter(allow10 = TRUE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=efficiency_linters]{efficiency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_equal_linter.R
\name{all_equal_linter}
\alias{all_equal_linter}
\title{Warn about invalid usage of \code{all.equal()}}
\usage{
all_equal_linter()
}
\description{
\code{\link[=all.equal]{all.equal()}} returns \code{TRUE} in the absence of differences but return a
character string (not \code{FALSE}) in the presence of differences.
Usage of \code{all.equal()} without wrapping it in \code{isTRUE()} in \code{if} clauses, or
preceded by the negation operator \code{!}, are thus likely to generate unexpected
errors if the compared objects have differences.
An alternative is to use \code{identical()} to compare vector of strings or when
exact equality is expected.
}
\examples{
# lints
lint(
  text = 'if (all.equal(a, b)) message("equal")',
  linters = all_equal_linter()
)

lint(
  text = '!all.equal(a, b)',
  linters = all_equal_linter()
)

lint(
  text = 'isFALSE(all.equal(a, b))',
  linters = all_equal_linter()
)

# okay
lint(
  text = 'if (isTRUE(all.equal(a, b))) message("equal")',
  linters = all_equal_linter()
)

lint(
  text = '!identical(a, b)',
  linters = all_equal_linter()
)

lint(
  text = "!isTRUE(all.equal(a, b))",
  linters = all_equal_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=common_mistakes_linters]{common_mistakes}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tags.R
\name{available_linters}
\alias{available_linters}
\alias{available_tags}
\title{Get Linter metadata from a package}
\usage{
available_linters(packages = "lintr", tags = NULL, exclude_tags = "deprecated")

available_tags(packages = "lintr")
}
\arguments{
\item{packages}{A character vector of packages to search for linters.}

\item{tags}{Optional character vector of tags to search. Only linters with at least one matching tag will be
returned. If \code{tags} is \code{NULL}, all linters will be returned. See \code{available_tags("lintr")} to find out what
tags are already used by lintr.}

\item{exclude_tags}{Tags to exclude from the results. Linters with at least one matching tag will not be returned.
If \code{exclude_tags} is \code{NULL}, no linters will be excluded. Note that \code{tags} takes priority, meaning that any
tag found in both \code{tags} and \code{exclude_tags} will be included, not excluded. Note that linters with tag \code{"defunct"}
(which do not work and can no longer be run) cannot be queried directly.}
}
\value{
\code{available_linters} returns a data frame with columns 'linter', 'package' and 'tags':

\describe{
\item{linter}{A character column naming the function associated with the linter.}
\item{package}{A character column containing the name of the package providing the linter.}
\item{tags}{A list column containing tags associated with the linter.}
}

\code{available_tags} returns a character vector of linter tags used by the packages.
}
\description{
\code{available_linters()} obtains a tagged list of all Linters available in a package.

\code{available_tags()} searches for available tags.
}
\section{Package Authors}{


To implement \code{available_linters()} for your package, include a file \code{inst/lintr/linters.csv} in your
package.
The CSV file must contain the columns 'linter' and 'tags', and be UTF-8 encoded.
Additional columns will be silently ignored if present and the columns are identified by name.
Each row describes a linter by
\enumerate{
\item its function name (e.g. \code{"assignment_linter"}) in the column 'linter'.
\item space-separated tags associated with the linter (e.g. \code{"style consistency default"}) in the column 'tags'.
}

Tags should be snake_case.

See \code{available_tags("lintr")} to find out what tags are already used by lintr.
}

\examples{
lintr_linters <- available_linters()

# If the package doesn't exist or isn't installed, an empty data frame will be returned
available_linters("does-not-exist")

lintr_linters2 <- available_linters(c("lintr", "does-not-exist"))
identical(lintr_linters, lintr_linters2)
available_tags()
}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \code{\link[=available_tags]{available_tags()}} to retrieve the set of valid tags.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{readability_linters}
\alias{readability_linters}
\title{Readability linters}
\description{
Linters highlighting readability issues, such as missing whitespace.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'readability':
\itemize{
\item{\code{\link{boolean_arithmetic_linter}}}
\item{\code{\link{brace_linter}}}
\item{\code{\link{coalesce_linter}}}
\item{\code{\link{commas_linter}}}
\item{\code{\link{commented_code_linter}}}
\item{\code{\link{comparison_negation_linter}}}
\item{\code{\link{conjunct_test_linter}}}
\item{\code{\link{consecutive_assertion_linter}}}
\item{\code{\link{consecutive_mutate_linter}}}
\item{\code{\link{cyclocomp_linter}}}
\item{\code{\link{empty_assignment_linter}}}
\item{\code{\link{expect_length_linter}}}
\item{\code{\link{expect_named_linter}}}
\item{\code{\link{expect_not_linter}}}
\item{\code{\link{expect_true_false_linter}}}
\item{\code{\link{fixed_regex_linter}}}
\item{\code{\link{for_loop_index_linter}}}
\item{\code{\link{function_left_parentheses_linter}}}
\item{\code{\link{function_return_linter}}}
\item{\code{\link{if_not_else_linter}}}
\item{\code{\link{if_switch_linter}}}
\item{\code{\link{implicit_assignment_linter}}}
\item{\code{\link{indentation_linter}}}
\item{\code{\link{infix_spaces_linter}}}
\item{\code{\link{inner_combine_linter}}}
\item{\code{\link{is_numeric_linter}}}
\item{\code{\link{keyword_quote_linter}}}
\item{\code{\link{length_levels_linter}}}
\item{\code{\link{lengths_linter}}}
\item{\code{\link{library_call_linter}}}
\item{\code{\link{line_length_linter}}}
\item{\code{\link{list2df_linter}}}
\item{\code{\link{matrix_apply_linter}}}
\item{\code{\link{nested_ifelse_linter}}}
\item{\code{\link{nested_pipe_linter}}}
\item{\code{\link{numeric_leading_zero_linter}}}
\item{\code{\link{object_length_linter}}}
\item{\code{\link{object_overwrite_linter}}}
\item{\code{\link{object_usage_linter}}}
\item{\code{\link{one_call_pipe_linter}}}
\item{\code{\link{outer_negation_linter}}}
\item{\code{\link{paren_body_linter}}}
\item{\code{\link{pipe_call_linter}}}
\item{\code{\link{pipe_consistency_linter}}}
\item{\code{\link{pipe_continuation_linter}}}
\item{\code{\link{quotes_linter}}}
\item{\code{\link{redundant_equals_linter}}}
\item{\code{\link{rep_len_linter}}}
\item{\code{\link{repeat_linter}}}
\item{\code{\link{sample_int_linter}}}
\item{\code{\link{scalar_in_linter}}}
\item{\code{\link{semicolon_linter}}}
\item{\code{\link{sort_linter}}}
\item{\code{\link{spaces_inside_linter}}}
\item{\code{\link{spaces_left_parentheses_linter}}}
\item{\code{\link{stopifnot_all_linter}}}
\item{\code{\link{string_boundary_linter}}}
\item{\code{\link{system_file_linter}}}
\item{\code{\link{T_and_F_symbol_linter}}}
\item{\code{\link{unnecessary_concatenation_linter}}}
\item{\code{\link{unnecessary_lambda_linter}}}
\item{\code{\link{unnecessary_nesting_linter}}}
\item{\code{\link{unnecessary_placeholder_linter}}}
\item{\code{\link{unreachable_code_linter}}}
\item{\code{\link{which_grepl_linter}}}
\item{\code{\link{yoda_test_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system_file_linter.R
\name{system_file_linter}
\alias{system_file_linter}
\title{Block usage of \code{file.path()} with \code{system.file()}}
\usage{
system_file_linter()
}
\description{
\code{\link[=system.file]{system.file()}} has a \code{...} argument which, internally, is passed to
\code{\link[=file.path]{file.path()}}, so including it in user code is repetitive.
}
\examples{
# will produce lints
lint(
  text = 'system.file(file.path("path", "to", "data"), package = "foo")',
  linters = system_file_linter()
)

lint(
  text = 'file.path(system.file(package = "foo"), "path", "to", "data")',
  linters = system_file_linter()
)

# okay
lint(
  text = 'system.file("path", "to", "data", package = "foo")',
  linters = system_file_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fixed_regex_linter.R
\name{fixed_regex_linter}
\alias{fixed_regex_linter}
\title{Require usage of \code{fixed=TRUE} in regular expressions where appropriate}
\usage{
fixed_regex_linter(allow_unescaped = FALSE)
}
\arguments{
\item{allow_unescaped}{Logical, default \code{FALSE}. If \code{TRUE}, only patterns that
require regex escapes (e.g. \code{"\\\\$"} or \code{"[$]"}) will be linted. See examples.}
}
\description{
Invoking a regular expression engine is overkill for cases when the search
pattern only involves static patterns.
}
\details{
NB: for \code{stringr} functions, that means wrapping the pattern in \code{stringr::fixed()}.

NB: this linter is likely not able to distinguish every possible case when
a fixed regular expression is preferable, rather it seeks to identify
likely cases. It should \emph{never} report false positives, however; please
report false positives as an error.
}
\examples{
# will produce lints
code_lines <- 'gsub("\\\\\\\\.", "", x)'
writeLines(code_lines)
lint(
  text = code_lines,
  linters = fixed_regex_linter()
)

lint(
  text = 'grepl("a[*]b", x)',
  linters = fixed_regex_linter()
)

lint(
  text = 'grepl("a[*]b", x)',
  linters = fixed_regex_linter(allow_unescaped = TRUE)
)

code_lines <- 'stringr::str_subset(x, "\\\\\\\\$")'
writeLines(code_lines)
lint(
  text = code_lines,
  linters = fixed_regex_linter()
)

lint(
  text = 'grepl("Munich", address)',
  linters = fixed_regex_linter()
)

# okay
code_lines <- 'gsub("\\\\\\\\.", "", x, fixed = TRUE)'
writeLines(code_lines)
lint(
  text = code_lines,
  linters = fixed_regex_linter()
)

lint(
  text = 'grepl("a*b", x, fixed = TRUE)',
  linters = fixed_regex_linter()
)

lint(
  text = 'stringr::str_subset(x, stringr::fixed("$"))',
  linters = fixed_regex_linter()
)

lint(
  text = 'grepl("Munich", address, fixed = TRUE)',
  linters = fixed_regex_linter()
)

lint(
  text = 'grepl("Munich", address)',
  linters = fixed_regex_linter(allow_unescaped = TRUE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}, \link[=regex_linters]{regex}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nzchar_linter.R
\name{nzchar_linter}
\alias{nzchar_linter}
\title{Require usage of nzchar where appropriate}
\usage{
nzchar_linter()
}
\description{
\code{\link[=nzchar]{nzchar()}} efficiently determines which of a vector of strings are empty
(i.e., are \code{""}). It should in most cases be used instead of
constructions like \code{string == ""} or \code{nchar(string) == 0}.
}
\details{
One crucial difference is in the default handling of \code{NA_character_}, i.e.,
missing strings. \code{nzchar(NA_character_)} is \code{TRUE}, while \code{NA_character_ == ""}
and \code{nchar(NA_character_) == 0} are both \code{NA}. Therefore, for strict
compatibility, use \code{nzchar(x, keepNA = TRUE)}. If the input is known to be
complete (no missing entries), this argument can be dropped for conciseness.
}
\examples{
# will produce lints
lint(
  text = "x[x == '']",
  linters = nzchar_linter()
)

lint(
  text = "x[nchar(x) > 0]",
  linters = nzchar_linter()
)

# okay
lint(
  text = "x[!nzchar(x, keepNA = TRUE)]",
  linters = nzchar_linter()
)

lint(
  text = "x[nzchar(x, keepNA = TRUE)]",
  linters = nzchar_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=efficiency_linters]{efficiency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lint.R
\name{gitlab_output}
\alias{gitlab_output}
\title{GitLab Report for lint results}
\usage{
gitlab_output(lints, filename = "lintr_results.json")
}
\arguments{
\item{lints}{The linting results}

\item{filename}{The file name of the output report}
}
\description{
Generate a report of the linting results using the
\href{https://docs.gitlab.com/ci/testing/code_quality/#code-quality-report-format}{GitLab} format.
}
\details{
lintr only supports three severity types ("style", "warning", and "error")
while the GitLab format supports five ("info", "minor", "major", "critical",
and "blocker"). The types "style", "warning", and "error" are mapped to
the GitLab types "info", "major", and "blocker", respectively. The GitLab
types "minor" and "critical" are ignored.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commas_linter.R
\name{commas_linter}
\alias{commas_linter}
\title{Commas linter}
\usage{
commas_linter(allow_trailing = FALSE)
}
\arguments{
\item{allow_trailing}{If \code{TRUE}, the linter allows a comma to be followed
directly by a closing bracket without a space.}
}
\description{
Check that all commas are followed by spaces, but do not have spaces before them.
}
\examples{
# will produce lints
lint(
  text = "switch(op , x = foo, y = bar)",
  linters = commas_linter()
)

lint(
  text = "mean(x,trim = 0.2,na.rm = TRUE)",
  linters = commas_linter()
)

lint(
  text = "x[ ,, drop=TRUE]",
  linters = commas_linter()
)

lint(
  text = "x[1,]",
  linters = commas_linter()
)

# okay
lint(
  text = "switch(op, x = foo, y = bar)",
  linters = commas_linter()
)

lint(
  text = "switch(op, x = , y = bar)",
  linters = commas_linter()
)

lint(
  text = "mean(x, trim = 0.2, na.rm = TRUE)",
  linters = commas_linter()
)

lint(
  text = "a[1, , 2, , 3]",
  linters = commas_linter()
)

lint(
  text = "x[1,]",
  linters = commas_linter(allow_trailing = TRUE)
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#commas}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_linter.R
\name{print_linter}
\alias{print_linter}
\title{Block usage of print() for logging}
\usage{
print_linter()
}
\description{
The default print method for character vectors is appropriate for interactively inspecting objects,
not for logging messages. Thus checked-in usage like \code{print(paste('Data has', nrow(DF), 'rows.'))}
is better served by using \code{\link[=cat]{cat()}}, e.g. \code{cat(sprintf('Data has \%d rows.\\n', nrow(DF)))}. Note that
using \code{cat()} entails supplying your own line returns; \code{\link[glue:glue]{glue::glue()}} might be preferable
to \code{\link[=sprintf]{sprintf()}} for constructing templated strings. Alternatively, \code{\link[=writeLines]{writeLines()}} can be used to
print character vectors with their own line returns. Lastly, note that \code{\link[=message]{message()}} differs slightly
from \code{cat()} in that it prints to \code{stderr} by default, not \code{stdout}, but is still a good option
to consider for logging purposes.
}
\examples{
# will produce lints
lint(
  text = "print('a')",
  linters = print_linter()
)

lint(
  text = "print(paste(x, 'y'))",
  linters = print_linter()
)

# okay
lint(
  text = "print(x)",
  linters = print_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regex_subset_linter.R
\name{regex_subset_linter}
\alias{regex_subset_linter}
\title{Require usage of direct methods for subsetting strings via regex}
\usage{
regex_subset_linter()
}
\description{
Using \code{\link[=grep]{grepv()}} returns the subset of the input that matches the pattern,
e.g. \code{grepv("[a-m]", letters)} will return the first 13 elements (\code{a} through \code{m}).
}
\details{
\code{letters[grep("[a-m]", letters)]} and \code{letters[grepl("[a-m]", letters)]}
both return the same thing, but more circuitously and more verbosely.

The \code{stringr} package also provides an even more readable alternative,
namely \code{str_subset()}, which should be preferred to versions using
\code{str_detect()} and \code{str_which()}.
}
\section{Exceptions}{

Note that \code{x[grep(pattern, x)]} and \code{grepv(pattern, x)} are not
\emph{completely} interchangeable when \code{x} is not character (most commonly,
when \code{x} is a factor), because the output of the latter will be a
character vector while the former remains a factor. It still may be
preferable to refactor such code, as it may be faster to match the
pattern on \code{levels(x)} and use that to subset instead.
}

\examples{
# will produce lints
lint(
  text = "x[grep(pattern, x)]",
  linters = regex_subset_linter()
)

lint(
  text = "x[stringr::str_which(x, pattern)]",
  linters = regex_subset_linter()
)

# okay
lint(
  text = "grepv(pattern, x)",
  linters = regex_subset_linter()
)

lint(
  text = "stringr::str_subset(x, pattern)",
  linters = regex_subset_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=efficiency_linters]{efficiency}, \link[=regex_linters]{regex}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{default_linters}
\alias{default_linters}
\title{Default linters}
\usage{
default_linters
}
\description{
List of default linters for \code{\link[=lint]{lint()}}. Use
\code{\link[=linters_with_defaults]{linters_with_defaults()}} to customize it. Most of the default linters
are based on \href{https://style.tidyverse.org/}{the tidyverse style guide}.

The set of default linters is as follows (any parameterized linters, e.g., \code{line_length_linter} use their default
argument(s), see \verb{?<linter_name>} for details):
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'default':
\itemize{
\item{\code{\link{assignment_linter}}}
\item{\code{\link{brace_linter}}}
\item{\code{\link{commas_linter}}}
\item{\code{\link{commented_code_linter}}}
\item{\code{\link{equals_na_linter}}}
\item{\code{\link{function_left_parentheses_linter}}}
\item{\code{\link{indentation_linter}}}
\item{\code{\link{infix_spaces_linter}}}
\item{\code{\link{line_length_linter}}}
\item{\code{\link{object_length_linter}}}
\item{\code{\link{object_name_linter}}}
\item{\code{\link{object_usage_linter}}}
\item{\code{\link{paren_body_linter}}}
\item{\code{\link{pipe_consistency_linter}}}
\item{\code{\link{pipe_continuation_linter}}}
\item{\code{\link{quotes_linter}}}
\item{\code{\link{return_linter}}}
\item{\code{\link{semicolon_linter}}}
\item{\code{\link{seq_linter}}}
\item{\code{\link{spaces_inside_linter}}}
\item{\code{\link{spaces_left_parentheses_linter}}}
\item{\code{\link{T_and_F_symbol_linter}}}
\item{\code{\link{trailing_blank_lines_linter}}}
\item{\code{\link{trailing_whitespace_linter}}}
\item{\code{\link{vector_logic_linter}}}
\item{\code{\link{whitespace_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_comparison_linter.R
\name{expect_comparison_linter}
\alias{expect_comparison_linter}
\title{Require usage of \code{expect_gt(x, y)} over \code{expect_true(x > y)} (and similar)}
\usage{
expect_comparison_linter()
}
\description{
\code{\link[testthat:expect_gt]{testthat::expect_gt()}}, \code{\link[testthat:expect_gte]{testthat::expect_gte()}}, \code{\link[testthat:expect_lt]{testthat::expect_lt()}},
\code{\link[testthat:expect_lte]{testthat::expect_lte()}}, and \code{\link[testthat:expect_equal]{testthat::expect_equal()}} exist specifically
for testing comparisons between two objects. \code{\link[testthat:expect_true]{testthat::expect_true()}} can
also be used for such tests, but it is better to use the tailored function
instead.
}
\examples{
# will produce lints
lint(
  text = "expect_true(x > y)",
  linters = expect_comparison_linter()
)

lint(
  text = "expect_true(x <= y)",
  linters = expect_comparison_linter()
)

lint(
  text = "expect_true(x == (y == 2))",
  linters = expect_comparison_linter()
)

# okay
lint(
  text = "expect_gt(x, y)",
  linters = expect_comparison_linter()
)

lint(
  text = "expect_lte(x, y)",
  linters = expect_comparison_linter()
)

lint(
  text = "expect_identical(x, y == 2)",
  linters = expect_comparison_linter()
)

lint(
  text = "expect_true(x < y | x > y^2)",
  linters = expect_comparison_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=package_development_linters]{package_development}, \link[=pkg_testthat_linters]{pkg_testthat}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duplicate_argument_linter.R
\name{duplicate_argument_linter}
\alias{duplicate_argument_linter}
\title{Duplicate argument linter}
\usage{
duplicate_argument_linter(except = c("mutate", "transmute"))
}
\arguments{
\item{except}{A character vector of function names as exceptions. Defaults to
functions that allow sequential updates to variables, currently \code{dplyr::mutate()}
and \code{dplyr::transmute()}.}
}
\description{
Check for duplicate arguments in function calls. Some cases are run-time errors
(e.g. \code{mean(x = 1:5, x = 2:3)}), otherwise this linter is used to discourage
explicitly providing duplicate names to objects (e.g. \code{c(a = 1, a = 2)}).
Duplicate-named objects are hard to work with programmatically and
should typically be avoided.
}
\examples{
# will produce lints
lint(
  text = "list(x = 1, x = 2)",
  linters = duplicate_argument_linter()
)

lint(
  text = "fun(arg = 1, arg = 2)",
  linters = duplicate_argument_linter()
)

# okay
lint(
  text = "list(x = 1, x = 2)",
  linters = duplicate_argument_linter(except = "list")
)

lint(
  text = "df \%>\% dplyr::mutate(x = a + b, x = x + d)",
  linters = duplicate_argument_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=common_mistakes_linters]{common_mistakes}, \link[=configurable_linters]{configurable}, \link[=correctness_linters]{correctness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paren_body_linter.R
\name{paren_body_linter}
\alias{paren_body_linter}
\title{Parenthesis before body linter}
\usage{
paren_body_linter()
}
\description{
Check that there is a space between right parenthesis and a body expression.
}
\examples{
# will produce lints
lint(
  text = "function(x)x + 1",
  linters = paren_body_linter()
)

# okay
lint(
  text = "function(x) x + 1",
  linters = paren_body_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#parentheses}
}
}
\section{Tags}{
\link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_true_false_linter.R
\name{expect_true_false_linter}
\alias{expect_true_false_linter}
\title{Require usage of \code{expect_true(x)} over \code{expect_equal(x, TRUE)}}
\usage{
expect_true_false_linter()
}
\description{
\code{\link[testthat:expect_true]{testthat::expect_true()}} and \code{\link[testthat:expect_false]{testthat::expect_false()}} exist specifically
for testing the \code{TRUE}/\code{FALSE} value of an object.
\code{\link[testthat:expect_equal]{testthat::expect_equal()}} and \code{\link[testthat:expect_identical]{testthat::expect_identical()}} can also be
used for such tests, but it is better to use the tailored function instead.
}
\examples{
# will produce lints
lint(
  text = "expect_equal(x, TRUE)",
  linters = expect_true_false_linter()
)

lint(
  text = "expect_equal(x, FALSE)",
  linters = expect_true_false_linter()
)

# okay
lint(
  text = "expect_true(x)",
  linters = expect_true_false_linter()
)

lint(
  text = "expect_false(x)",
  linters = expect_true_false_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=package_development_linters]{package_development}, \link[=pkg_testthat_linters]{pkg_testthat}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/with.R
\name{linters_with_defaults}
\alias{linters_with_defaults}
\title{Create a linter configuration based on defaults}
\usage{
linters_with_defaults(..., defaults = default_linters)
}
\arguments{
\item{...}{Arguments of elements to change. If unnamed, the argument is automatically named.
If the named argument already exists in the list of linters, it is replaced by the new element.
If it does not exist, it is added. If the value is \code{NULL}, the linter is removed.}

\item{defaults}{Default list of linters to modify. Must be named.}
}
\description{
Make a new list based on \pkg{lintr}'s default linters.
The result of this function is meant to be passed to the \code{linters} argument of \code{lint()},
or to be put in your configuration file.
}
\examples{
# When using interactively you will usually pass the result onto `lint` or `lint_package()`
f <- tempfile()
writeLines("my_slightly_long_variable_name <- 2.3", f)
lint(f, linters = linters_with_defaults(line_length_linter = line_length_linter(120L)))
unlink(f)

# the default linter list with a different line length cutoff
my_linters <- linters_with_defaults(line_length_linter = line_length_linter(120L))

# omit the argument name if you are just using different arguments
my_linters <- linters_with_defaults(defaults = my_linters, object_name_linter("camelCase"))

# remove assignment checks (with NULL), add absolute path checks
my_linters <- linters_with_defaults(
  defaults = my_linters,
  assignment_linter = NULL,
  absolute_path_linter()
)

# checking the included linters
names(my_linters)

}
\seealso{
\itemize{
\item \link{linters_with_tags} for basing off tags attached to linters, possibly across multiple packages.
\item \link{all_linters} for basing off all available linters in lintr.
\item \link{available_linters} to get a data frame of available linters.
\item \link{linters} for a complete list of linters available in lintr.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keyword_quote_linter.R
\name{keyword_quote_linter}
\alias{keyword_quote_linter}
\title{Block unnecessary quoting in calls}
\usage{
keyword_quote_linter()
}
\description{
Any valid symbol can be used as a keyword argument to an R function call.
Sometimes, it is necessary to quote (or backtick) an argument that is
not an otherwise valid symbol (e.g. creating a vector whose names have
spaces); besides this edge case, quoting should not be done.
}
\details{
The most common source of violation for this is creating named vectors,
lists, or data.frame-alikes, but it can be observed in other calls as well.

Similar reasoning applies to extractions with \code{$} or \code{@}.
}
\examples{
# will produce lints
lint(
  text = 'data.frame("a" = 1)',
  linters = keyword_quote_linter()
)

lint(
  text = "data.frame(`a` = 1)",
  linters = keyword_quote_linter()
)

lint(
  text = 'my_list$"key"',
  linters = keyword_quote_linter()
)

lint(
  text = 's4obj@"key"',
  linters = keyword_quote_linter()
)

# okay
lint(
  text = "data.frame(`a b` = 1)",
  linters = keyword_quote_linter()
)

lint(
  text = "my_list$`a b`",
  linters = keyword_quote_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lint.R
\name{lint}
\alias{lint}
\alias{lint_dir}
\alias{lint_package}
\title{Lint a file, directory, or package}
\usage{
lint(
  filename,
  linters = NULL,
  ...,
  cache = FALSE,
  parse_settings = !inline_data,
  text = NULL
)

lint_dir(
  path = ".",
  ...,
  relative_path = TRUE,
  exclusions = list("renv", "packrat"),
  pattern = "(?i)[.](r|rmd|qmd|rnw|rhtml|rrst|rtex|rtxt)$",
  parse_settings = TRUE,
  show_progress = NULL
)

lint_package(
  path = ".",
  ...,
  relative_path = TRUE,
  exclusions = list("R/RcppExports.R"),
  parse_settings = TRUE,
  show_progress = NULL
)
}
\arguments{
\item{filename}{Either the filename for a file to lint, or a character string of inline R code for linting.
The latter (inline data) applies whenever \code{filename} has a newline character (\\n).}

\item{linters}{A named list of linter functions to apply. See \link{linters} for a full list of default and available
linters.}

\item{...}{Provide additional arguments to be passed to:
\itemize{
\item \code{\link[=exclude]{exclude()}} (in case of \code{lint()}; e.g. \code{lints} or \code{exclusions})
\item \code{\link[=lint]{lint()}} (in case of \code{lint_dir()} and \code{lint_package()}; e.g. \code{linters} or \code{cache})
}}

\item{cache}{When logical, toggle caching of lint results. If passed a character string, store the cache in this
directory.}

\item{parse_settings}{Logical. Whether to try and parse the \link[=read_settings]{settings}. Otherwise, the
\code{\link[=default_settings]{default_settings()}} are used. \code{TRUE} by default when linting files, as opposed to \verb{text=}.}

\item{text}{Optional argument for supplying a string or lines directly, e.g. if the file is already in memory or
linting is being done ad hoc.}

\item{path}{For the base directory of the project (for \code{lint_dir()}) or
package (for \code{lint_package()}).}

\item{relative_path}{if \code{TRUE}, file paths are printed using their path relative to the base directory.
If \code{FALSE}, use the full absolute path.}

\item{exclusions}{exclusions for \code{\link[=exclude]{exclude()}}, relative to the package path.}

\item{pattern}{pattern for files, by default it will take files with any of the extensions
.R, .Rmd, .qmd, .Rnw, .Rhtml, .Rrst, .Rtex, .Rtxt allowing for lowercase r (.r, ...).}

\item{show_progress}{Logical controlling whether to show linting progress with
\code{\link[cli:cli_progress_along]{cli::cli_progress_along()}}. The default behavior is to show progress in \code{\link[=interactive]{interactive()}} sessions
not running a testthat suite.}
}
\value{
An object of class \code{c("lints", "list")}, each element of which is a \code{"list"} object.
}
\description{
\itemize{
\item \code{lint()} lints a single file.
\item \code{lint_dir()} lints all files in a directory.
\item \code{lint_package()} lints all likely locations for R files in a package, i.e.
\verb{R/}, \verb{tests/}, \verb{inst/}, \verb{vignettes/}, \verb{data-raw/}, \verb{demo/}, and \verb{exec/}.
}
}
\details{
Read \code{vignette("lintr")} to learn how to configure which linters are run
by default.
Note that if files contain unparseable encoding problems, only the encoding problem will be linted to avoid
unintelligible error messages from other linters.
}
\examples{
# linting inline-code
lint("a = 123\n")
lint(text = "a = 123")

# linting a file
f <- tempfile()
writeLines("a=1", f)
lint(f)
unlink(f)

if (FALSE) {
  lint_dir()

  lint_dir(
    linters = list(semicolon_linter()),
    exclusions = list(
      "inst/doc/creating_linters.R" = 1,
      "inst/example/bad.R",
      "renv"
    )
  )
}
if (FALSE) {
  lint_package()

  lint_package(
    linters = linters_with_defaults(semicolon_linter = semicolon_linter()),
    exclusions = list("inst/doc/creating_linters.R" = 1, "inst/example/bad.R")
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/any_duplicated_linter.R
\name{any_duplicated_linter}
\alias{any_duplicated_linter}
\title{Require usage of \code{anyDuplicated(x) > 0} over \code{any(duplicated(x))}}
\usage{
any_duplicated_linter()
}
\description{
\code{\link[=anyDuplicated]{anyDuplicated()}} exists as a replacement for \code{any(duplicated(.))}, which is
more efficient for simple objects, and is at worst equally efficient.
Therefore, it should be used in all situations instead of the latter.
}
\details{
Also match usage like \code{length(unique(x$col)) == nrow(x)}, which can
be replaced by \code{anyDuplicated(x$col) == 0L}.
}
\examples{
# will produce lints
lint(
  text = "any(duplicated(x), na.rm = TRUE)",
  linters = any_duplicated_linter()
)

lint(
  text = "length(unique(x)) == length(x)",
  linters = any_duplicated_linter()
)

lint(
  text = "DT[, uniqueN(col) == .N]",
  linters = any_duplicated_linter()
)

# okay
lint(
  text = "anyDuplicated(x)",
  linters = any_duplicated_linter()
)

lint(
  text = "anyDuplicated(x) == 0L",
  linters = any_duplicated_linter()
)

lint(
  text = "anyDuplicated(DT, by = 'col') == 0L",
  linters = any_duplicated_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=efficiency_linters]{efficiency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unnecessary_concatenation_linter.R
\name{unnecessary_concatenation_linter}
\alias{unnecessary_concatenation_linter}
\title{Unneeded concatenation linter}
\usage{
unnecessary_concatenation_linter(allow_single_expression = TRUE)
}
\arguments{
\item{allow_single_expression}{Logical, default \code{TRUE}. If \code{FALSE}, one-expression
usages of \code{c()} are always linted, e.g. \code{c(x)} and \code{c(matrix(...))}. In some such
cases, \code{c()} is being used for its side-effect of stripping non-name attributes;
it is usually preferable to use the more readable \code{\link[=as.vector]{as.vector()}} instead.
\code{\link[=as.vector]{as.vector()}} is not always preferable, for example with environments
(especially, \code{R6} objects), in which case \code{list()} is the better alternative.}
}
\description{
Check that the \code{\link[=c]{c()}} function is not used without arguments nor with a single constant.
}
\examples{
# will produce lints
lint(
  text = "x <- c()",
  linters = unnecessary_concatenation_linter()
)

lint(
  text = "x <- c(TRUE)",
  linters = unnecessary_concatenation_linter()
)

lint(
  text = "x <- c(1.5 + 2.5)",
  linters = unnecessary_concatenation_linter(allow_single_expression = FALSE)
)

# okay
lint(
  text = "x <- NULL",
  linters = unnecessary_concatenation_linter()
)

# In case the intent here was to seed a vector of known size
lint(
  text = "x <- integer(4L)",
  linters = unnecessary_concatenation_linter()
)

lint(
  text = "x <- TRUE",
  linters = unnecessary_concatenation_linter()
)

lint(
  text = "x <- c(1.5 + 2.5)",
  linters = unnecessary_concatenation_linter(allow_single_expression = TRUE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exclude.R
\name{parse_exclusions}
\alias{parse_exclusions}
\title{read a source file and parse all the excluded lines from it}
\usage{
parse_exclusions(
  file,
  exclude = settings$exclude,
  exclude_next = settings$exclude_next,
  exclude_start = settings$exclude_start,
  exclude_end = settings$exclude_end,
  exclude_linter = settings$exclude_linter,
  exclude_linter_sep = settings$exclude_linter_sep,
  lines = NULL,
  linter_names = NULL
)
}
\arguments{
\item{file}{R source file}

\item{exclude}{Regular expression used to mark lines to exclude.}

\item{exclude_next}{Regular expression used to mark lines immediately preceding excluded lines.}

\item{exclude_start}{Regular expression used to mark the start of an excluded range.}

\item{exclude_end}{Regular expression used to mark the end of an excluded range.}

\item{exclude_linter}{Regular expression used to capture a list of to-be-excluded linters immediately following a
\code{exclude} or \code{exclude_start} marker.}

\item{exclude_linter_sep}{Regular expression used to split a linter list into individual linter names for exclusion.}

\item{lines}{A character vector of the content lines of \code{file}.}

\item{linter_names}{Names of active linters.}
}
\value{
A possibly named list of excluded lines, possibly for specific linters.
}
\description{
read a source file and parse all the excluded lines from it
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AAA.R
\name{linters}
\alias{linters}
\title{Available linters}
\description{
A variety of linters are available in \pkg{lintr}. The most popular ones are readily
accessible through \code{\link[=default_linters]{default_linters()}}.

Within a \code{\link[=lint]{lint()}} function call, the linters in use are initialized with the provided
arguments and fed with the source file (provided by \code{\link[=get_source_expressions]{get_source_expressions()}}).

A data frame of all available linters can be retrieved using \code{\link[=available_linters]{available_linters()}}.
Documentation for linters is structured into tags to allow for easier discovery;
see also \code{\link[=available_tags]{available_tags()}}.
}
\section{Tags}{
The following tags exist:
\itemize{
\item{\link[=best_practices_linters]{best_practices} (65 linters)}
\item{\link[=common_mistakes_linters]{common_mistakes} (13 linters)}
\item{\link[=configurable_linters]{configurable} (44 linters)}
\item{\link[=consistency_linters]{consistency} (33 linters)}
\item{\link[=correctness_linters]{correctness} (7 linters)}
\item{\link[=default_linters]{default} (26 linters)}
\item{\link[=efficiency_linters]{efficiency} (30 linters)}
\item{\link[=executing_linters]{executing} (6 linters)}
\item{\link[=package_development_linters]{package_development} (14 linters)}
\item{\link[=pkg_testthat_linters]{pkg_testthat} (12 linters)}
\item{\link[=readability_linters]{readability} (66 linters)}
\item{\link[=regex_linters]{regex} (4 linters)}
\item{\link[=robustness_linters]{robustness} (19 linters)}
\item{\link[=style_linters]{style} (40 linters)}
\item{\link[=tidy_design_linters]{tidy_design} (1 linters)}
}
}
\section{Linters}{
The following linters exist:
\itemize{
\item{\code{\link{absolute_path_linter}} (tags: best_practices, configurable, robustness)}
\item{\code{\link{all_equal_linter}} (tags: common_mistakes, robustness)}
\item{\code{\link{any_duplicated_linter}} (tags: best_practices, efficiency)}
\item{\code{\link{any_is_na_linter}} (tags: best_practices, efficiency)}
\item{\code{\link{assignment_linter}} (tags: configurable, consistency, default, style)}
\item{\code{\link{backport_linter}} (tags: configurable, package_development, robustness)}
\item{\code{\link{boolean_arithmetic_linter}} (tags: best_practices, efficiency, readability)}
\item{\code{\link{brace_linter}} (tags: configurable, default, readability, style)}
\item{\code{\link{class_equals_linter}} (tags: best_practices, consistency, robustness)}
\item{\code{\link{coalesce_linter}} (tags: best_practices, consistency, readability)}
\item{\code{\link{commas_linter}} (tags: configurable, default, readability, style)}
\item{\code{\link{commented_code_linter}} (tags: best_practices, default, readability, style)}
\item{\code{\link{comparison_negation_linter}} (tags: consistency, readability)}
\item{\code{\link{condition_call_linter}} (tags: best_practices, configurable, style, tidy_design)}
\item{\code{\link{condition_message_linter}} (tags: best_practices, consistency)}
\item{\code{\link{conjunct_test_linter}} (tags: best_practices, configurable, package_development, pkg_testthat, readability)}
\item{\code{\link{consecutive_assertion_linter}} (tags: consistency, readability, style)}
\item{\code{\link{consecutive_mutate_linter}} (tags: configurable, consistency, efficiency, readability)}
\item{\code{\link{cyclocomp_linter}} (tags: best_practices, configurable, readability, style)}
\item{\code{\link{download_file_linter}} (tags: best_practices, common_mistakes, robustness)}
\item{\code{\link{duplicate_argument_linter}} (tags: common_mistakes, configurable, correctness)}
\item{\code{\link{empty_assignment_linter}} (tags: best_practices, readability)}
\item{\code{\link{equals_na_linter}} (tags: common_mistakes, correctness, default, robustness)}
\item{\code{\link{expect_comparison_linter}} (tags: best_practices, package_development, pkg_testthat)}
\item{\code{\link{expect_identical_linter}} (tags: package_development, pkg_testthat)}
\item{\code{\link{expect_length_linter}} (tags: best_practices, package_development, pkg_testthat, readability)}
\item{\code{\link{expect_named_linter}} (tags: best_practices, package_development, pkg_testthat, readability)}
\item{\code{\link{expect_not_linter}} (tags: best_practices, package_development, pkg_testthat, readability)}
\item{\code{\link{expect_null_linter}} (tags: best_practices, package_development, pkg_testthat)}
\item{\code{\link{expect_s3_class_linter}} (tags: best_practices, package_development, pkg_testthat)}
\item{\code{\link{expect_s4_class_linter}} (tags: best_practices, package_development, pkg_testthat)}
\item{\code{\link{expect_true_false_linter}} (tags: best_practices, package_development, pkg_testthat, readability)}
\item{\code{\link{expect_type_linter}} (tags: best_practices, package_development, pkg_testthat)}
\item{\code{\link{fixed_regex_linter}} (tags: best_practices, configurable, efficiency, readability, regex)}
\item{\code{\link{for_loop_index_linter}} (tags: best_practices, readability, robustness)}
\item{\code{\link{function_argument_linter}} (tags: best_practices, consistency, style)}
\item{\code{\link{function_left_parentheses_linter}} (tags: default, readability, style)}
\item{\code{\link{function_return_linter}} (tags: best_practices, readability)}
\item{\code{\link{if_not_else_linter}} (tags: configurable, consistency, readability)}
\item{\code{\link{if_switch_linter}} (tags: best_practices, configurable, consistency, efficiency, readability)}
\item{\code{\link{ifelse_censor_linter}} (tags: best_practices, efficiency)}
\item{\code{\link{implicit_assignment_linter}} (tags: best_practices, configurable, readability, style)}
\item{\code{\link{implicit_integer_linter}} (tags: best_practices, configurable, consistency, style)}
\item{\code{\link{indentation_linter}} (tags: configurable, default, readability, style)}
\item{\code{\link{infix_spaces_linter}} (tags: configurable, default, readability, style)}
\item{\code{\link{inner_combine_linter}} (tags: consistency, efficiency, readability)}
\item{\code{\link{is_numeric_linter}} (tags: best_practices, consistency, readability)}
\item{\code{\link{keyword_quote_linter}} (tags: consistency, readability, style)}
\item{\code{\link{length_levels_linter}} (tags: best_practices, consistency, readability)}
\item{\code{\link{length_test_linter}} (tags: common_mistakes, efficiency)}
\item{\code{\link{lengths_linter}} (tags: best_practices, efficiency, readability)}
\item{\code{\link{library_call_linter}} (tags: best_practices, configurable, readability, style)}
\item{\code{\link{line_length_linter}} (tags: configurable, default, readability, style)}
\item{\code{\link{list2df_linter}} (tags: efficiency, readability)}
\item{\code{\link{list_comparison_linter}} (tags: best_practices, common_mistakes)}
\item{\code{\link{literal_coercion_linter}} (tags: best_practices, consistency, efficiency)}
\item{\code{\link{matrix_apply_linter}} (tags: efficiency, readability)}
\item{\code{\link{missing_argument_linter}} (tags: common_mistakes, configurable, correctness)}
\item{\code{\link{missing_package_linter}} (tags: common_mistakes, robustness)}
\item{\code{\link{namespace_linter}} (tags: configurable, correctness, executing, robustness)}
\item{\code{\link{nested_ifelse_linter}} (tags: efficiency, readability)}
\item{\code{\link{nested_pipe_linter}} (tags: configurable, consistency, readability)}
\item{\code{\link{nonportable_path_linter}} (tags: best_practices, configurable, robustness)}
\item{\code{\link{nrow_subset_linter}} (tags: best_practices, consistency, efficiency)}
\item{\code{\link{numeric_leading_zero_linter}} (tags: consistency, readability, style)}
\item{\code{\link{nzchar_linter}} (tags: best_practices, consistency, efficiency)}
\item{\code{\link{object_length_linter}} (tags: configurable, default, executing, readability, style)}
\item{\code{\link{object_name_linter}} (tags: configurable, consistency, default, executing, style)}
\item{\code{\link{object_overwrite_linter}} (tags: best_practices, configurable, executing, readability, robustness)}
\item{\code{\link{object_usage_linter}} (tags: configurable, correctness, default, executing, readability, style)}
\item{\code{\link{one_call_pipe_linter}} (tags: readability, style)}
\item{\code{\link{outer_negation_linter}} (tags: best_practices, efficiency, readability)}
\item{\code{\link{package_hooks_linter}} (tags: correctness, package_development, style)}
\item{\code{\link{paren_body_linter}} (tags: default, readability, style)}
\item{\code{\link{paste_linter}} (tags: best_practices, configurable, consistency)}
\item{\code{\link{pipe_call_linter}} (tags: readability, style)}
\item{\code{\link{pipe_consistency_linter}} (tags: configurable, default, readability, style)}
\item{\code{\link{pipe_continuation_linter}} (tags: default, readability, style)}
\item{\code{\link{pipe_return_linter}} (tags: best_practices, common_mistakes)}
\item{\code{\link{print_linter}} (tags: best_practices, consistency)}
\item{\code{\link{quotes_linter}} (tags: configurable, consistency, default, readability, style)}
\item{\code{\link{redundant_equals_linter}} (tags: best_practices, common_mistakes, efficiency, readability)}
\item{\code{\link{redundant_ifelse_linter}} (tags: best_practices, configurable, consistency, efficiency)}
\item{\code{\link{regex_subset_linter}} (tags: best_practices, efficiency, regex)}
\item{\code{\link{rep_len_linter}} (tags: best_practices, consistency, readability)}
\item{\code{\link{repeat_linter}} (tags: readability, style)}
\item{\code{\link{return_linter}} (tags: configurable, default, style)}
\item{\code{\link{routine_registration_linter}} (tags: best_practices, efficiency, robustness)}
\item{\code{\link{sample_int_linter}} (tags: efficiency, readability, robustness)}
\item{\code{\link{scalar_in_linter}} (tags: best_practices, configurable, consistency, efficiency, readability)}
\item{\code{\link{semicolon_linter}} (tags: configurable, default, readability, style)}
\item{\code{\link{seq_linter}} (tags: best_practices, consistency, default, efficiency, robustness)}
\item{\code{\link{sort_linter}} (tags: best_practices, efficiency, readability)}
\item{\code{\link{spaces_inside_linter}} (tags: default, readability, style)}
\item{\code{\link{spaces_left_parentheses_linter}} (tags: default, readability, style)}
\item{\code{\link{sprintf_linter}} (tags: common_mistakes, correctness)}
\item{\code{\link{stopifnot_all_linter}} (tags: best_practices, readability)}
\item{\code{\link{string_boundary_linter}} (tags: configurable, efficiency, readability, regex)}
\item{\code{\link{strings_as_factors_linter}} (tags: robustness)}
\item{\code{\link{system_file_linter}} (tags: best_practices, consistency, readability)}
\item{\code{\link{T_and_F_symbol_linter}} (tags: best_practices, consistency, default, readability, robustness, style)}
\item{\code{\link{terminal_close_linter}} (tags: best_practices, robustness)}
\item{\code{\link{todo_comment_linter}} (tags: configurable, style)}
\item{\code{\link{trailing_blank_lines_linter}} (tags: default, style)}
\item{\code{\link{trailing_whitespace_linter}} (tags: configurable, default, style)}
\item{\code{\link{undesirable_function_linter}} (tags: best_practices, configurable, robustness, style)}
\item{\code{\link{undesirable_operator_linter}} (tags: best_practices, configurable, robustness, style)}
\item{\code{\link{unnecessary_concatenation_linter}} (tags: configurable, efficiency, readability, style)}
\item{\code{\link{unnecessary_lambda_linter}} (tags: best_practices, configurable, efficiency, readability)}
\item{\code{\link{unnecessary_nesting_linter}} (tags: best_practices, configurable, consistency, readability)}
\item{\code{\link{unnecessary_placeholder_linter}} (tags: best_practices, readability)}
\item{\code{\link{unreachable_code_linter}} (tags: best_practices, configurable, readability)}
\item{\code{\link{unused_import_linter}} (tags: best_practices, common_mistakes, configurable, executing)}
\item{\code{\link{vector_logic_linter}} (tags: best_practices, common_mistakes, default, efficiency)}
\item{\code{\link{which_grepl_linter}} (tags: consistency, efficiency, readability, regex)}
\item{\code{\link{whitespace_linter}} (tags: consistency, default, style)}
\item{\code{\link{yoda_test_linter}} (tags: best_practices, package_development, pkg_testthat, readability)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ifelse_censor_linter.R
\name{ifelse_censor_linter}
\alias{ifelse_censor_linter}
\title{Block usage of \code{ifelse()} where \code{pmin()} or \code{pmax()} is more appropriate}
\usage{
ifelse_censor_linter()
}
\description{
\code{ifelse(x > M, M, x)} is the same as \code{pmin(x, M)}, but harder
to read and requires several passes over the vector.
}
\details{
The same goes for other similar ways to censor a vector, e.g.
\code{ifelse(x <= M, x, M)} is \code{pmin(x, M)},
\code{ifelse(x < m, m, x)} is \code{pmax(x, m)}, and
\code{ifelse(x >= m, x, m)} is \code{pmax(x, m)}.
}
\examples{
# will produce lints
lint(
  text = "ifelse(5:1 < pi, 5:1, pi)",
  linters = ifelse_censor_linter()
)

lint(
  text = "ifelse(x > 0, x, 0)",
  linters = ifelse_censor_linter()
)

# okay
lint(
  text = "pmin(5:1, pi)",
  linters = ifelse_censor_linter()
)

lint(
  text = "pmax(x, 0)",
  linters = ifelse_censor_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=efficiency_linters]{efficiency}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/undesirable_function_linter.R
\name{undesirable_function_linter}
\alias{undesirable_function_linter}
\title{Undesirable function linter}
\usage{
undesirable_function_linter(
  fun = default_undesirable_functions,
  symbol_is_undesirable = TRUE
)
}
\arguments{
\item{fun}{Character vector of undesirable function names. Input can be any of three types:
\itemize{
\item Unnamed entries must be a character string specifying an undesirable function.
\item For named entries, the name specifies the undesirable function.
\itemize{
\item If the entry is a character string, it is used as a description of
why a given function is undesirable
\item Otherwise, entries should be missing (\code{NA})
A generic message that the named function is undesirable is used if no
specific description is provided.
Input can also be a list of character strings for convenience.
}
}

Defaults to \link{default_undesirable_functions}. To make small customizations to this list,
use \code{\link[=modify_defaults]{modify_defaults()}}.}

\item{symbol_is_undesirable}{Whether to consider the use of an undesirable function
name as a symbol undesirable or not.}
}
\description{
Report the use of undesirable functions and suggest an alternative.
}
\examples{
# defaults for which functions are considered undesirable
names(default_undesirable_functions)

# will produce lints
lint(
  text = "sapply(x, mean)",
  linters = undesirable_function_linter()
)

lint(
  text = "log10(x)",
  linters = undesirable_function_linter(fun = c("log10" = NA))
)

lint(
  text = "log10(x)",
  linters = undesirable_function_linter(fun = c("log10" = "use log()"))
)

lint(
  text = 'dir <- "path/to/a/directory"',
  linters = undesirable_function_linter(fun = c("dir" = NA))
)


lint(
  text = 'dir <- "path/to/a/directory"',
  linters = undesirable_function_linter(fun = "dir")
)

# okay
lint(
  text = "vapply(x, mean, FUN.VALUE = numeric(1))",
  linters = undesirable_function_linter()
)

lint(
  text = "log(x, base = 10)",
  linters = undesirable_function_linter(fun = c("log10" = "use log()"))
)

lint(
  text = 'dir <- "path/to/a/directory"',
  linters = undesirable_function_linter(fun = c("dir" = NA), symbol_is_undesirable = FALSE)
)

lint(
  text = 'dir <- "path/to/a/directory"',
  linters = undesirable_function_linter(fun = "dir", symbol_is_undesirable = FALSE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=robustness_linters]{robustness}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/empty_assignment_linter.R
\name{empty_assignment_linter}
\alias{empty_assignment_linter}
\title{Block assignment of \code{{}}}
\usage{
empty_assignment_linter()
}
\description{
Assignment of \code{{}} is the same as assignment of \code{NULL}; use the latter
for clarity. Closely related: \code{\link[=unnecessary_concatenation_linter]{unnecessary_concatenation_linter()}}.
}
\examples{
# will produce lints
lint(
  text = "x <- {}",
  linters = empty_assignment_linter()
)

writeLines("x = {\n}")
lint(
  text = "x = {\n}",
  linters = empty_assignment_linter()
)

# okay
lint(
  text = "x <- { 3 + 4 }",
  linters = empty_assignment_linter()
)

lint(
  text = "x <- NULL",
  linters = empty_assignment_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sprintf_linter.R
\name{sprintf_linter}
\alias{sprintf_linter}
\title{Require correct \code{sprintf()} calls}
\usage{
sprintf_linter()
}
\description{
Check for an inconsistent number of arguments or arguments with incompatible types (for literal arguments) in
\code{\link[=sprintf]{sprintf()}} calls.
}
\details{
\code{\link[=gettextf]{gettextf()}} calls are also included, since \code{gettextf()} is a thin wrapper around \code{sprintf()}.
}
\examples{
# will produce lints
lint(
  text = 'sprintf("hello \%s \%s \%d", x, y)',
  linters = sprintf_linter()
)

lint(
  text = 'sprintf("hello")',
  linters = sprintf_linter()
)

# okay
lint(
  text = 'sprintf("hello \%s \%s \%d", x, y, z)',
  linters = sprintf_linter()
)

lint(
  text = 'sprintf("hello \%s \%s \%d", x, y, ...)',
  linters = sprintf_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=common_mistakes_linters]{common_mistakes}, \link[=correctness_linters]{correctness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scalar_in_linter.R
\name{scalar_in_linter}
\alias{scalar_in_linter}
\title{Block usage like x \%in\% "a"}
\usage{
scalar_in_linter(in_operators = NULL)
}
\arguments{
\item{in_operators}{Character vector of additional infix operators that behave like the \code{\%in\%} operator,
e.g. \code{{data.table}}'s \verb{\%chin\%} operator.}
}
\description{
\code{vector \%in\% set} is appropriate for matching a vector to a set, but if
that set has size 1, \code{==} is more appropriate. However, if \code{vector} has
also size 1 and can be \code{NA}, the use of \code{==} should be accompanied by extra
protection for the missing case (for example, \code{isTRUE(NA == "arg")} or
\code{!is.na(x) && x == "arg"}).
}
\details{
\code{scalar \%in\% vector} is OK, because the alternative (\code{any(vector == scalar)})
is more circuitous & potentially less clear.
}
\examples{
# will produce lints
lint(
  text = "x \%in\% 1L",
  linters = scalar_in_linter()
)

lint(
  text = "x \%chin\% 'a'",
  linters = scalar_in_linter(in_operators = "\%chin\%")
)

# okay
lint(
  text = "x \%in\% 1:10",
  linters = scalar_in_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nested_ifelse_linter.R
\name{nested_ifelse_linter}
\alias{nested_ifelse_linter}
\title{Block usage of nested \code{ifelse()} calls}
\usage{
nested_ifelse_linter()
}
\description{
Calling \code{\link[=ifelse]{ifelse()}} in nested calls is problematic for two main reasons:
\enumerate{
\item It can be hard to read -- mapping the code to the expected output
for such code can be a messy task/require a lot of mental bandwidth,
especially for code that nests more than once
\item It is inefficient -- \code{ifelse()} can evaluate \emph{all} of its arguments at
both yes and no (see \url{https://stackoverflow.com/q/16275149}); this issue
is exacerbated for nested calls
}
}
\details{
Users can instead rely on a more readable alternative modeled after SQL
CASE WHEN statements.

Let's say this is our original code:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ifelse(
  x == "a",
  2L,
  ifelse(x == "b", 3L, 1L)
)
}\if{html}{\out{</div>}}

Here are a few ways to avoid nesting and make the code more readable:
\itemize{
\item Use \code{data.table::fcase()}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{data.table::fcase(
  x == "a", 2L,
  x == "b", 3L,
  default = 1L
)
}\if{html}{\out{</div>}}
\item Use \code{dplyr::case_match()}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{dplyr::case_match(
  x,
  "a" ~ 2L,
  "b" ~ 3L,
  .default = 1L
)
}\if{html}{\out{</div>}}
\item Use a look-up-and-merge approach (build a mapping table between values
and outputs and merge this to the input)

\if{html}{\out{<div class="sourceCode r">}}\preformatted{default <- 1L
values <- data.frame(
  a = 2L,
  b = 3L
)
found_value <- values[[x]]
ifelse(is.null(found_value), default, found_value)
}\if{html}{\out{</div>}}
}
}
\examples{
# will produce lints
lint(
  text = 'ifelse(x == "a", 1L, ifelse(x == "b", 2L, 3L))',
  linters = nested_ifelse_linter()
)

# okay
lint(
  text = 'dplyr::case_when(x == "a" ~ 1L, x == "b" ~ 2L, TRUE ~ 3L)',
  linters = nested_ifelse_linter()
)

lint(
  text = 'data.table::fcase(x == "a", 1L, x == "b", 2L, default = 3L)',
  linters = nested_ifelse_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/routine_registration_linter.R
\name{routine_registration_linter}
\alias{routine_registration_linter}
\title{Identify unregistered native routines}
\usage{
routine_registration_linter()
}
\description{
It is preferable to register routines for efficiency and safety.
}
\examples{
# will produce lints
lint(
  text = '.Call("cpp_routine", PACKAGE = "mypkg")',
  linters = routine_registration_linter()
)

lint(
  text = '.Fortran("f_routine", PACKAGE = "mypkg")',
  linters = routine_registration_linter()
)

# okay
lint(
  text = ".Call(cpp_routine)",
  linters = routine_registration_linter()
)

lint(
  text = ".Fortran(f_routine)",
  linters = routine_registration_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Registering-native-routines}
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=efficiency_linters]{efficiency}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/return_linter.R
\name{return_linter}
\alias{return_linter}
\title{Return linter}
\usage{
return_linter(
  return_style = c("implicit", "explicit"),
  allow_implicit_else = TRUE,
  return_functions = NULL,
  except = NULL,
  except_regex = NULL
)
}
\arguments{
\item{return_style}{Character string naming the return style. \code{"implicit"},
the default, enforces the Tidyverse guide recommendation to leave terminal
returns implicit. \code{"explicit"} style requires that \code{return()} always be
explicitly supplied.}

\item{allow_implicit_else}{Logical, default \code{TRUE}. If \code{FALSE}, functions with a terminal
\code{if} clause must always have an \verb{else} clause, making the \code{NULL} alternative explicit
if necessary. Similarly, functions with terminal \code{\link[=switch]{switch()}} statements must have an
explicit default case.}

\item{return_functions}{Character vector of functions that are accepted as terminal calls
when \code{return_style = "explicit"}. These are in addition to exit functions
from base that are always allowed: \code{\link[=stop]{stop()}}, \code{\link[=q]{q()}}, \code{\link[=quit]{quit()}}, \code{\link[=invokeRestart]{invokeRestart()}},
\code{tryInvokeRestart()}, \code{\link[=UseMethod]{UseMethod()}}, \code{\link[=NextMethod]{NextMethod()}}, \code{\link[=standardGeneric]{standardGeneric()}},
\code{\link[=callNextMethod]{callNextMethod()}}, \code{\link[=.C]{.C()}}, \code{\link[=.Call]{.Call()}}, \code{\link[=.External]{.External()}}, and \code{\link[=.Fortran]{.Fortran()}}.}

\item{except, except_regex}{Character vector of functions that are not checked when
\code{return_style = "explicit"}. These are in addition to namespace hook functions
that are never checked: \code{.onLoad()}, \code{.onUnload()}, \code{.onAttach()}, \code{.onDetach()},
\code{.Last.lib()}, \code{.First()} and \code{.Last()}. \code{except} matches function names exactly,
while \code{except_regex} does exclusion by pattern matching with \code{\link[rex:re_matches]{rex::re_matches()}}.}
}
\description{
This linter checks functions' \code{\link[=return]{return()}} expressions.
}
\examples{
# will produce lints
code <- "function(x) {\n  return(x + 1)\n}"
writeLines(code)
lint(
  text = code,
  linters = return_linter()
)

code <- "function(x) {\n  x + 1\n}"
writeLines(code)
lint(
  text = code,
  linters = return_linter(return_style = "explicit")
)

code <- "function(x) {\n  if (x > 0) 2\n}"
writeLines(code)
lint(
  text = code,
  linters = return_linter(allow_implicit_else = FALSE)
)

# okay
code <- "function(x) {\n  x + 1\n}"
writeLines(code)
lint(
  text = code,
  linters = return_linter()
)

code <- "function(x) {\n  return(x + 1)\n}"
writeLines(code)
lint(
  text = code,
  linters = return_linter(return_style = "explicit")
)

code <- "function(x) {\n  if (x > 0) 2 else NULL\n}"
writeLines(code)
lint(
  text = code,
  linters = return_linter(allow_implicit_else = FALSE)
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/functions.html?q=return#return}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/which_grepl_linter.R
\name{which_grepl_linter}
\alias{which_grepl_linter}
\title{Require usage of grep over which(grepl(.))}
\usage{
which_grepl_linter()
}
\description{
\code{which(grepl(pattern, x))} is the same as \code{grep(pattern, x)}, but harder
to read and requires two passes over the vector.
}
\examples{
# will produce lints
lint(
  text = "which(grepl('^a', x))",
  linters = which_grepl_linter()
)

# okay
lint(
  text = "which(grepl('^a', x) | grepl('^b', x))",
  linters = which_grepl_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=consistency_linters]{consistency}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}, \link[=regex_linters]{regex}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample_int_linter.R
\name{sample_int_linter}
\alias{sample_int_linter}
\title{Require usage of sample.int(n, m, ...) over sample(1:n, m, ...)}
\usage{
sample_int_linter()
}
\description{
\code{\link[=sample.int]{sample.int()}} is preferable to \code{sample()} for the case of sampling numbers
between 1 and \code{n}. \code{sample} calls \code{sample.int()} "under the hood".
}
\examples{
# will produce lints
lint(
  text = "sample(1:10, 2)",
  linters = sample_int_linter()
)

lint(
  text = "sample(seq(4), 2)",
  linters = sample_int_linter()
)

lint(
  text = "sample(seq_len(8), 2)",
  linters = sample_int_linter()
)

# okay
lint(
  text = "sample(seq(1, 5, by = 2), 2)",
  linters = sample_int_linter()
)

lint(
  text = "sample(letters, 2)",
  linters = sample_int_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}, \link[=robustness_linters]{robustness}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function_argument_linter.R
\name{function_argument_linter}
\alias{function_argument_linter}
\title{Function argument linter}
\usage{
function_argument_linter()
}
\description{
Check that arguments with defaults come last in all function declarations,
as per the tidyverse design guide.

Changing the argument order can be a breaking change. An alternative to changing the argument order
is to instead set the default for such arguments to \code{NULL}.
}
\examples{
# will produce lints
lint(
  text = "function(y = 1, z = 2, x) {}",
  linters = function_argument_linter()
)

lint(
  text = "function(x, y, z = 1, ..., w) {}",
  linters = function_argument_linter()
)

# okay
lint(
  text = "function(x, y = 1, z = 2) {}",
  linters = function_argument_linter()
)

lint(
  text = "function(x, y, w, z = 1, ...) {}",
  linters = function_argument_linter()
)

lint(
  text = "function(y = 1, z = 2, x = NULL) {}",
  linters = function_argument_linter()
)

lint(
  text = "function(x, y, z = 1, ..., w = NULL) {}",
  linters = function_argument_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://design.tidyverse.org/required-no-defaults.html}
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unused_import_linter.R
\name{unused_import_linter}
\alias{unused_import_linter}
\title{Check that imported packages are actually used}
\usage{
unused_import_linter(
  allow_ns_usage = FALSE,
  except_packages = c("bit64", "data.table", "tidyverse"),
  interpret_glue = TRUE
)
}
\arguments{
\item{allow_ns_usage}{Suppress lints for packages only used via namespace.
This is \code{FALSE} by default because \code{pkg::fun()} doesn't require \code{library(pkg)}.
You can use \link[=requireNamespace]{requireNamespace("pkg")} to ensure a package is
installed without attaching it.}

\item{except_packages}{Character vector of packages that are ignored.
These are usually attached for their side effects.}

\item{interpret_glue}{If \code{TRUE}, interpret \code{\link[glue:glue]{glue::glue()}} calls to avoid false positives caused by local variables
which are only used in a glue expression.}
}
\description{
Check that imported packages are actually used
}
\examples{
# will produce lints
code_lines <- "library(dplyr)\n1 + 1"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unused_import_linter()
)

code_lines <- "library(dplyr)\ndplyr::tibble(a = 1)"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unused_import_linter()
)

# okay
code_lines <- "library(dplyr)\ntibble(a = 1)"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unused_import_linter()
)

code_lines <- "library(dplyr)\ndplyr::tibble(a = 1)"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = unused_import_linter(allow_ns_usage = TRUE)
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=common_mistakes_linters]{common_mistakes}, \link[=configurable_linters]{configurable}, \link[=executing_linters]{executing}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indentation_linter.R
\name{indentation_linter}
\alias{indentation_linter}
\title{Check that indentation is consistent}
\usage{
indentation_linter(
  indent = 2L,
  hanging_indent_style = c("tidy", "always", "never"),
  assignment_as_infix = TRUE
)
}
\arguments{
\item{indent}{Number of spaces, that a code block should be indented by relative to its parent code block.
Used for multi-line code blocks (\code{{ ... }}), function calls (\code{( ... )}) and extractions (\verb{[ ... ]}, \verb{[[ ... ]]}).
Defaults to 2.}

\item{hanging_indent_style}{Indentation style for multi-line function calls with arguments in their first line.
Defaults to tidyverse style, i.e. a block indent is used if the function call terminates with \verb{)} on a separate
line and a hanging indent if not.
Note that multi-line function calls without arguments on their first line will always be expected to have
block-indented arguments.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# complies to any style
map(
  x,
  f,
  additional_arg = 42
)

# complies to "tidy" and "never"
map(x, f,
  additional_arg = 42
)

# complies to "always"
map(x, f,
    additional_arg = 42
)

# complies to "tidy" and "always"
map(x, f,
    additional_arg = 42)

# complies to "never"
map(x, f,
  additional_arg = 42)

# complies to "tidy"
function(
  a,
  b
) \{
  # body
\}
}\if{html}{\out{</div>}}}

\item{assignment_as_infix}{Treat \verb{<-} as a regular (i.e. left-associative) infix operator?
This means, that infix operators on the right hand side of an assignment do not trigger a second level of
indentation:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# complies to any style
variable <- a \%+\%
  b \%+\%
  c

# complies to assignment_as_infix = TRUE
variable <-
  a \%+\%
  b \%+\%
  c

# complies to assignment_as_infix = FALSE
variable <-
  a \%+\%
    b \%+\%
    c
}\if{html}{\out{</div>}}}
}
\description{
Check that indentation is consistent
}
\examples{
# will produce lints
code_lines <- "if (TRUE) {\n1 + 1\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = indentation_linter()
)

code_lines <- "if (TRUE) {\n    1 + 1\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = indentation_linter()
)

code_lines <- "map(x, f,\n  additional_arg = 42\n)"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = indentation_linter(hanging_indent_style = "always")
)

code_lines <- "map(x, f,\n    additional_arg = 42)"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = indentation_linter(hanging_indent_style = "never")
)

# okay
code_lines <- "map(x, f,\n  additional_arg = 42\n)"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = indentation_linter()
)

code_lines <- "if (TRUE) {\n    1 + 1\n}"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = indentation_linter(indent = 4)
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#indenting}
\item \url{https://style.tidyverse.org/functions.html#long-lines-1}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nested_pipe_linter.R
\name{nested_pipe_linter}
\alias{nested_pipe_linter}
\title{Block usage of pipes nested inside other calls}
\usage{
nested_pipe_linter(
  allow_inline = TRUE,
  allow_outer_calls = c("try", "tryCatch", "withCallingHandlers")
)
}
\arguments{
\item{allow_inline}{Logical, default \code{TRUE}, in which case only "inner"
pipelines which span more than one line are linted. If \code{FALSE}, even
"inner" pipelines that fit in one line are linted.}

\item{allow_outer_calls}{Character vector dictating which "outer"
calls to exempt from the requirement to unnest (see examples). Defaults
to \code{\link[=try]{try()}}, \code{\link[=tryCatch]{tryCatch()}}, and \code{\link[=withCallingHandlers]{withCallingHandlers()}}.}
}
\description{
Nesting pipes harms readability; extract sub-steps to separate variables,
append further pipeline steps, or otherwise refactor such usage away.
}
\examples{
# will produce lints
code <- "df1 \%>\%\n  inner_join(df2 \%>\%\n    select(a, b)\n  )"
writeLines(code)
lint(
  text = code,
  linters = nested_pipe_linter()
)

lint(
  text = "df1 \%>\% inner_join(df2 \%>\% select(a, b))",
  linters = nested_pipe_linter(allow_inline = FALSE)
)

lint(
  text = "tryCatch(x \%>\% filter(grp == 'a'), error = identity)",
  linters = nested_pipe_linter(allow_outer_calls = character())
)

# okay
lint(
  text = "df1 \%>\% inner_join(df2 \%>\% select(a, b))",
  linters = nested_pipe_linter()
)

code <- "df1 \%>\%\n  inner_join(df2 \%>\%\n    select(a, b)\n  )"
writeLines(code)
lint(
  text = code,
  linters = nested_pipe_linter(allow_outer_calls = "inner_join")
)

lint(
  text = "tryCatch(x \%>\% filter(grp == 'a'), error = identity)",
  linters = nested_pipe_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=consistency_linters]{consistency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/T_and_F_symbol_linter.R
\name{T_and_F_symbol_linter}
\alias{T_and_F_symbol_linter}
\title{\code{T} and \code{F} symbol linter}
\usage{
T_and_F_symbol_linter()
}
\description{
Although they can be synonyms, avoid the symbols \code{T} and \code{F}, and use \code{TRUE} and \code{FALSE}, respectively, instead.
\code{T} and \code{F} are not reserved keywords and can be assigned to any other values.
}
\examples{
# will produce lints
lint(
  text = "x <- T; y <- F",
  linters = T_and_F_symbol_linter()
)

lint(
  text = "T = 1.2; F = 2.4",
  linters = T_and_F_symbol_linter()
)

# okay
lint(
  text = "x <- c(TRUE, FALSE)",
  linters = T_and_F_symbol_linter()
)

lint(
  text = "t = 1.2; f = 2.4",
  linters = T_and_F_symbol_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#logical-vectors}
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=robustness_linters]{robustness}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_comparison_linter.R
\name{list_comparison_linter}
\alias{list_comparison_linter}
\title{Block usage of comparison operators with known-list() functions like lapply}
\usage{
list_comparison_linter()
}
\description{
Usage like \code{lapply(x, sum) > 10} is awkward because the list must first
be coerced to a vector for comparison. A function like \code{\link[=vapply]{vapply()}}
should be preferred.
}
\examples{
# will produce lints
lint(
  text = "lapply(x, sum) > 10",
  linters = list_comparison_linter()
)

# okay
lint(
  text = "unlist(lapply(x, sum)) > 10",
  linters = list_comparison_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=common_mistakes_linters]{common_mistakes}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{executing_linters}
\alias{executing_linters}
\title{Code executing linters}
\description{
Linters that evaluate parts of the linted code, such as loading referenced packages.
These linters should not be used with untrusted code, and may need dependencies of the linted package or project to
be available in order to function correctly. For package authors, note that this includes loading the package itself,
e.g. with \code{pkgload::load_all()} or installing and attaching the package.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'executing':
\itemize{
\item{\code{\link{namespace_linter}}}
\item{\code{\link{object_length_linter}}}
\item{\code{\link{object_name_linter}}}
\item{\code{\link{object_overwrite_linter}}}
\item{\code{\link{object_usage_linter}}}
\item{\code{\link{unused_import_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function_left_parentheses_linter.R
\name{function_left_parentheses_linter}
\alias{function_left_parentheses_linter}
\title{Function left parentheses linter}
\usage{
function_left_parentheses_linter()
}
\description{
Check that all left parentheses in a function call do not have spaces before them
(e.g. \code{mean  (1:3)}). Although this is syntactically valid, it makes the code
difficult to read.
}
\details{
Exceptions are made for control flow functions (\code{if}, \code{for}, etc.).
}
\examples{
# will produce lints
lint(
  text = "mean (x)",
  linters = function_left_parentheses_linter()
)

lint(
  text = "stats::sd(c (x, y, z))",
  linters = function_left_parentheses_linter()
)

# okay
lint(
  text = "mean(x)",
  linters = function_left_parentheses_linter()
)

lint(
  text = "stats::sd(c(x, y, z))",
  linters = function_left_parentheses_linter()
)

lint(
  text = "foo <- function(x) (x + 1)",
  linters = function_left_parentheses_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#parentheses}
\item \code{\link[=spaces_left_parentheses_linter]{spaces_left_parentheses_linter()}}
}
}
\section{Tags}{
\link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_named_linter.R
\name{expect_named_linter}
\alias{expect_named_linter}
\title{Require usage of \code{expect_named(x, n)} over \code{expect_equal(names(x), n)}}
\usage{
expect_named_linter()
}
\description{
\code{\link[testthat:expect_named]{testthat::expect_named()}} exists specifically for testing the \code{\link[=names]{names()}} of
an object. \code{\link[testthat:expect_equal]{testthat::expect_equal()}} can also be used for such tests,
but it is better to use the tailored function instead.
}
\examples{
# will produce lints
lint(
  text = 'expect_equal(names(x), "a")',
  linters = expect_named_linter()
)

# okay
lint(
  text = 'expect_named(x, "a")',
  linters = expect_named_linter()
)

lint(
  text = 'expect_equal(colnames(x), "a")',
  linters = expect_named_linter()
)

lint(
  text = 'expect_equal(dimnames(x), "a")',
  linters = expect_named_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=package_development_linters]{package_development}, \link[=pkg_testthat_linters]{pkg_testthat}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/line_length_linter.R
\name{line_length_linter}
\alias{line_length_linter}
\title{Line length linter}
\usage{
line_length_linter(length = 80L, ignore_string_bodies = FALSE)
}
\arguments{
\item{length}{Maximum line length allowed. Default is \code{80L} (Hollerith limit).}

\item{ignore_string_bodies}{Logical, default \code{FALSE}. If \code{TRUE}, the contents
of string literals are ignored. The quotes themselves are included, so this
mainly affects wide multiline strings, e.g. SQL queries.}
}
\description{
Check that the line length of both comments and code is less than \code{length}.
}
\examples{
# will produce lints
lint(
  text = strrep("x", 23L),
  linters = line_length_linter(length = 20L)
)

# the trailing ' is counted towards line length, so this still lints
lint(
  text = "'a long single-line string'",
  linters = line_length_linter(length = 15L, ignore_string_bodies = TRUE)
)

lines <- paste(
  "query <- '",
  "  SELECT *",
  "  FROM MyTable",
  "  WHERE profit > 0",
  "'",
  sep = "\n"
)
writeLines(lines)
lint(
  text = lines,
  linters = line_length_linter(length = 10L)
)

# okay
lint(
  text = strrep("x", 21L),
  linters = line_length_linter(length = 40L)
)

lines <- paste(
  "paste(",
  "  'a long',",
  "  'single-line',",
  "  'string'",
  ")",
  sep = "\n"
)
writeLines(lines)
lint(
  text = lines,
  linters = line_length_linter(length = 15L, ignore_string_bodies = TRUE)
)

lines <- paste(
  "query <- '",
  "  SELECT *",
  "  FROM MyTable",
  "  WHERE profit > 0",
  "'",
  sep = "\n"
)
writeLines(lines)
lint(
  text = lines,
  linters = line_length_linter(length = 10L, ignore_string_bodies = TRUE)
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/syntax.html#long-lines}
}
}
\section{Tags}{
\link[=configurable_linters]{configurable}, \link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipe_continuation_linter.R
\name{pipe_continuation_linter}
\alias{pipe_continuation_linter}
\title{Pipe continuation linter}
\usage{
pipe_continuation_linter()
}
\description{
Check that each step in a pipeline is on a new line, or the entire pipe fits on one line.
}
\examples{
# will produce lints
code_lines <- "1:3 \%>\%\n mean() \%>\% as.character()"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = pipe_continuation_linter()
)

code_lines <- "1:3 |> mean() |>\n as.character()"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = pipe_continuation_linter()
)

# okay
lint(
  text = "1:3 \%>\% mean() \%>\% as.character()",
  linters = pipe_continuation_linter()
)

code_lines <- "1:3 \%>\%\n mean() \%>\%\n as.character()"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = pipe_continuation_linter()
)

lint(
  text = "1:3 |> mean() |> as.character()",
  linters = pipe_continuation_linter()
)

code_lines <- "1:3 |>\n mean() |>\n as.character()"
writeLines(code_lines)
lint(
  text = code_lines,
  linters = pipe_continuation_linter()
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://style.tidyverse.org/pipes.html#long-lines-2}
}
}
\section{Tags}{
\link[=default_linters]{default}, \link[=readability_linters]{readability}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lengths_linter.R
\name{lengths_linter}
\alias{lengths_linter}
\title{Require usage of \code{lengths()} where possible}
\usage{
lengths_linter()
}
\description{
\code{\link[base:lengths]{base::lengths()}} is a function that was added to base R in version 3.2.0 to
get the length of each element of a list. It is equivalent to
\code{sapply(x, length)}, but faster and more readable.
}
\examples{
# will produce lints
lint(
  text = "sapply(x, length)",
  linters = lengths_linter()
)

lint(
  text = "vapply(x, length, integer(1L))",
  linters = lengths_linter()
)

lint(
  text = "purrr::map_int(x, length)",
  linters = lengths_linter()
)

# okay
lint(
  text = "lengths(x)",
  linters = lengths_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=efficiency_linters]{efficiency}, \link[=readability_linters]{readability}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package_hooks_linter.R
\name{package_hooks_linter}
\alias{package_hooks_linter}
\title{Package hooks linter}
\usage{
package_hooks_linter()
}
\description{
Check various common "gotchas" in \code{\link[=.onLoad]{.onLoad()}}, \code{\link[=.onAttach]{.onAttach()}}, \code{\link[=.Last.lib]{.Last.lib()}}, \code{\link[=.onDetach]{.onDetach()}}, and \code{\link[=.onUnload]{.onUnload()}}
namespace hooks that will cause \verb{R CMD check} issues. See Writing R Extensions for details.
}
\details{
\enumerate{
\item \code{.onLoad()} shouldn't call \code{\link[=cat]{cat()}}, \code{\link[=message]{message()}}, \code{\link[=print]{print()}}, \code{\link[=writeLines]{writeLines()}}, \code{\link[=packageStartupMessage]{packageStartupMessage()}},
\code{\link[=require]{require()}}, \code{\link[=library]{library()}}, or \code{\link[=installed.packages]{installed.packages()}}.
\item \code{.onAttach()} shouldn't call \code{cat()}, \code{message()}, \code{print()}, \code{writeLines()}, \code{\link[=library.dynam]{library.dynam()}},
\code{require()}, \code{library()}, or \code{installed.packages()}.
\item \code{.Last.lib()} and \code{.onDetach()} shouldn't call \code{\link[=library.dynam.unload]{library.dynam.unload()}}.
\item \code{.onLoad()} and \code{.onAttach()} should take two arguments, with names matching \verb{^lib} and \verb{^pkg};
\code{.Last.lib()}, \code{.onDetach()}, and \code{.onUnload()} should take one argument with name matching \verb{^lib}.
}
}
\examples{
# will produce lints
lint(
  text = ".onLoad <- function(lib, ...) { }",
  linters = package_hooks_linter()
)

lint(
  text = ".onAttach <- function(lib, pkg) { require(foo) }",
  linters = package_hooks_linter()
)

lint(
  text = ".onDetach <- function(pkg) { }",
  linters = package_hooks_linter()
)

lint(
  text = ".onUnload <- function() { }",
  linters = package_hooks_linter()
)

# okay
lint(
  text = ".onLoad <- function(lib, pkg) { }",
  linters = package_hooks_linter()
)

lint(
  text = '.onAttach <- function(lib, pkg) { loadNamespace("foo") }',
  linters = package_hooks_linter()
)

lint(
  text = ".onDetach <- function(lib) { }",
  linters = package_hooks_linter()
)

lint(
  text = ".onUnload <- function(libpath) { }",
  linters = package_hooks_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=correctness_linters]{correctness}, \link[=package_development_linters]{package_development}, \link[=style_linters]{style}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{tidy_design_linters}
\alias{tidy_design_linters}
\title{Tidyverse design linters}
\description{
Linters based on guidelines described in the 'Tidy design principles' book.
}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://design.tidyverse.org/}
}
}
\section{Linters}{
The following linters are tagged with 'tidy_design':
\itemize{
\item{\code{\link{condition_call_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/condition_call_linter.R
\name{condition_call_linter}
\alias{condition_call_linter}
\title{Recommend usage of \code{call. = FALSE} in conditions}
\usage{
condition_call_linter(display_call = FALSE)
}
\arguments{
\item{display_call}{Logical specifying expected behavior regarding \code{call.}
argument in conditions.
\itemize{
\item \code{NA} forces providing \verb{call. =} but ignores its value (this can be used in
cases where you expect a mix of \code{call. = FALSE} and \code{call. = TRUE})
\item \code{TRUE} lints \code{call. = FALSE}
\item \code{FALSE} forces \code{call. = FALSE} (lints \code{call. = TRUE} or missing \verb{call. =} value)
}}
}
\description{
This linter, with the default \code{display_call = FALSE}, enforces the
recommendation of the tidyverse design guide regarding displaying error
calls.
}
\examples{
# will produce lints
lint(
  text = "stop('test')",
  linters = condition_call_linter()
)

lint(
  text = "stop('test', call. = TRUE)",
  linters = condition_call_linter()
)

lint(
  text = "stop('test', call. = FALSE)",
  linters = condition_call_linter(display_call = TRUE)
)

lint(
  text = "stop('this is a', 'test', call. = FALSE)",
  linters = condition_call_linter(display_call = TRUE)
)

# okay
lint(
  text = "stop('test', call. = FALSE)",
  linters = condition_call_linter()
)

lint(
  text = "stop('this is a', 'test', call. = FALSE)",
  linters = condition_call_linter()
)

lint(
  text = "stop('test', call. = TRUE)",
  linters = condition_call_linter(display_call = TRUE)
)

}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://design.tidyverse.org/err-call.html}>
}
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=configurable_linters]{configurable}, \link[=style_linters]{style}, \link[=tidy_design_linters]{tidy_design}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_lint.R
\name{expect_lint}
\alias{expect_lint}
\alias{expect_no_lint}
\title{Lint expectation}
\usage{
expect_lint(
  content,
  checks,
  ...,
  file = NULL,
  language = "en",
  ignore_order = FALSE
)

expect_no_lint(content, ..., file = NULL, language = "en")
}
\arguments{
\item{content}{A character vector for the file content to be linted, each vector element representing a line of
text.}

\item{checks}{Checks to be performed:
\describe{
\item{NULL}{check that no lints are returned.}
\item{single string or regex object}{check that the single lint returned has a matching message.}
\item{named list}{check that the single lint returned has fields that match. Accepted fields are the same as those
taken by \code{\link[=Lint]{Lint()}}.}
\item{list of named lists}{for each of the multiple lints returned, check that it matches the checks in the
corresponding named list (as described in the point above).}
}
Named vectors are also accepted instead of named lists, but this is a compatibility feature that
is not recommended for new code.}

\item{...}{Arguments passed to \code{\link[=lint]{lint()}}, e.g. the linters or cache to use.}

\item{file}{If not \code{NULL}, read content from the specified file rather than from \code{content}.}

\item{language}{Temporarily override Rs \code{LANGUAGE} envvar, controlling localization of base R error messages.
This makes testing them reproducible on all systems irrespective of their native R language setting.}

\item{ignore_order}{Logical, default \code{FALSE}. If \code{TRUE}, the order of the \code{checks} does not matter, e.g.
lints with higher line numbers can come before those with lower line numbers, and the order of linters
affecting the same line is also irrelevant.}
}
\value{
\code{NULL}, invisibly.
}
\description{
These are expectation functions to test specified linters on sample code in the \code{testthat} testing framework.
\itemize{
\item \code{expect_lint} asserts that specified lints are generated.
\item \code{expect_no_lint} asserts that no lints are generated.
}
}
\examples{
# no expected lint
expect_no_lint("a", trailing_blank_lines_linter())

# one expected lint
expect_lint("a\n", "trailing blank", trailing_blank_lines_linter())
expect_lint("a\n", list(message = "trailing blank", line_number = 2), trailing_blank_lines_linter())

# several expected lints
expect_lint("a\n\n", list("trailing blank", "trailing blank"), trailing_blank_lines_linter())
expect_lint(
  "a\n\n",
  list(
    list(message = "trailing blank", line_number = 2),
    list(message = "trailing blank", line_number = 3)
  ),
  trailing_blank_lines_linter()
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_r_string}
\alias{get_r_string}
\title{Extract text from \code{STR_CONST} nodes}
\usage{
get_r_string(s, xpath = NULL)
}
\arguments{
\item{s}{An input string or strings. If \code{s} is an \code{xml_node} or \code{xml_nodeset} and \code{xpath} is \code{NULL},
extract its string value with \code{\link[xml2:xml_text]{xml2::xml_text()}}. If \code{s} is an \code{xml_node} or \code{xml_nodeset}
and \code{xpath} is specified, it is extracted with \code{\link[xml2:xml_find_chr]{xml2::xml_find_chr()}}.}

\item{xpath}{An XPath, passed on to \code{\link[xml2:xml_find_chr]{xml2::xml_find_chr()}} after wrapping with \code{string()}.}
}
\description{
Convert \code{STR_CONST} \code{text()} values into R strings. This is useful to account for arbitrary
character literals, e.g. \code{R"------[hello]------"}, which is parsed in R as \code{"hello"}.
It is quite cumbersome to write XPaths allowing for strings like this, so whenever your
linter logic requires testing a \code{STR_CONST} node's value, use this function.
NB: this is also properly vectorized on \code{s}, and accepts a variety of inputs. Empty inputs
will become \code{NA} outputs, which helps ensure that \code{length(get_r_string(s)) == length(s)}.
}
\examples{
tmp <- tempfile()
writeLines("c('a', 'b')", tmp)
expr_as_xml <- get_source_expressions(tmp)$expressions[[1L]]$xml_parsed_content
writeLines(as.character(expr_as_xml))
get_r_string(expr_as_xml, "expr[2]")
get_r_string(expr_as_xml, "expr[3]")
unlink(tmp)

# more importantly, extract raw strings correctly
tmp_raw <- tempfile()
writeLines("c(R'(a\\\\b)', R'--[a\\\\\"\'\"\\\\b]--')", tmp_raw)
expr_as_xml_raw <- get_source_expressions(tmp_raw)$expressions[[1L]]$xml_parsed_content
writeLines(as.character(expr_as_xml_raw))
get_r_string(expr_as_xml_raw, "expr[2]")
get_r_string(expr_as_xml_raw, "expr[3]")
unlink(tmp_raw)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{pkg_testthat_linters}
\alias{pkg_testthat_linters}
\title{Testthat linters}
\description{
Linters encouraging best practices within testthat suites.
}
\seealso{
\itemize{
\item \link{linters} for a complete list of linters available in lintr.
\item \url{https://testthat.r-lib.org}
\item \url{https://r-pkgs.org/testing-basics.html}
}
}
\section{Linters}{
The following linters are tagged with 'pkg_testthat':
\itemize{
\item{\code{\link{conjunct_test_linter}}}
\item{\code{\link{expect_comparison_linter}}}
\item{\code{\link{expect_identical_linter}}}
\item{\code{\link{expect_length_linter}}}
\item{\code{\link{expect_named_linter}}}
\item{\code{\link{expect_not_linter}}}
\item{\code{\link{expect_null_linter}}}
\item{\code{\link{expect_s3_class_linter}}}
\item{\code{\link{expect_s4_class_linter}}}
\item{\code{\link{expect_true_false_linter}}}
\item{\code{\link{expect_type_linter}}}
\item{\code{\link{yoda_test_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipe_return_linter.R
\name{pipe_return_linter}
\alias{pipe_return_linter}
\title{Block usage of return() in magrittr pipelines}
\usage{
pipe_return_linter()
}
\description{
\code{\link[=return]{return()}} inside a magrittr pipeline does not actually execute \code{return()}
like you'd expect:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{bad_usage <- function(x) \{
  x \%>\%
    return()
  FALSE
\}
}\if{html}{\out{</div>}}
}
\details{
\code{bad_usage(TRUE)} will return \code{FALSE}! It will technically work "as expected"
if this is the final statement in the function body, but such usage is misleading.
Instead, assign the pipe outcome to a variable and return that.
}
\examples{
# will produce lints
lint(
  text = "function(x) x \%>\% return()",
  linters = pipe_return_linter()
)

# okay
code <- "function(x) {\n  y <- sum(x)\n  return(y)\n}"
writeLines(code)
lint(
  text = code,
  linters = pipe_return_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=common_mistakes_linters]{common_mistakes}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linter_tag_docs.R
\name{common_mistakes_linters}
\alias{common_mistakes_linters}
\title{Common mistake linters}
\description{
Linters highlighting common mistakes, such as duplicate arguments.
}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Linters}{
The following linters are tagged with 'common_mistakes':
\itemize{
\item{\code{\link{all_equal_linter}}}
\item{\code{\link{download_file_linter}}}
\item{\code{\link{duplicate_argument_linter}}}
\item{\code{\link{equals_na_linter}}}
\item{\code{\link{length_test_linter}}}
\item{\code{\link{list_comparison_linter}}}
\item{\code{\link{missing_argument_linter}}}
\item{\code{\link{missing_package_linter}}}
\item{\code{\link{pipe_return_linter}}}
\item{\code{\link{redundant_equals_linter}}}
\item{\code{\link{sprintf_linter}}}
\item{\code{\link{unused_import_linter}}}
\item{\code{\link{vector_logic_linter}}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class_equals_linter.R
\name{class_equals_linter}
\alias{class_equals_linter}
\title{Block comparison of class with \code{==}}
\usage{
class_equals_linter()
}
\description{
Usage like \code{class(x) == "character"} is prone to error since class in R
is in general a vector. The correct version for S3 classes is \code{\link[=inherits]{inherits()}}:
\code{inherits(x, "character")}. Often, class \code{k} will have an \code{is.} equivalent,
for example \code{\link[=is.character]{is.character()}} or \code{\link[=is.data.frame]{is.data.frame()}}.
}
\details{
Similar reasoning applies for \code{class(x) \%in\% "character"}.
}
\examples{
# will produce lints
lint(
  text = 'is_lm <- class(x) == "lm"',
  linters = class_equals_linter()
)

lint(
  text = 'if ("lm" \%in\% class(x)) is_lm <- TRUE',
  linters = class_equals_linter()
)

# okay
lint(
  text = 'is_lm <- inherits(x, "lm")',
  linters = class_equals_linter()
)

lint(
  text = 'if (inherits(x, "lm")) is_lm <- TRUE',
  linters = class_equals_linter()
)

}
\seealso{
\link{linters} for a complete list of linters available in lintr.
}
\section{Tags}{
\link[=best_practices_linters]{best_practices}, \link[=consistency_linters]{consistency}, \link[=robustness_linters]{robustness}
}
